
#pragma once

/*
 * MULTI-CHANNEL SIGNED DISTANCE FIELD GENERATOR v1.1 (2016-05-08)
 * ---------------------------------------------------------------
 * A utility by Viktor Chlumsky, (c) 2014 - 2016
 *
 * The technique used to generate multi-channel distance fields in this code
 * has been developed by Viktor Chlumsky in 2014 for his master's thesis,
 * "Shape Decomposition for Multi-Channel Distance Fields". It provides improved
 * quality of sharp corners in glyphs and other 2D shapes in comparison to monochrome
 * distance fields. To reconstruct an image of the shape, apply the median of three
 * operation on the triplet of sampled distance field values.
 *
 */

#include <vector>
#include <cstdlib>
#include <cmath>

namespace msdfgen {

// "core/arithmetics.hpp"
/// Returns the smaller of the arguments.

template <typename T>
inline T min(T a, T b) {
    return b < a ? b : a;
}

/// Returns the larger of the arguments.
template <typename T>
inline T max(T a, T b) {
    return a < b ? b : a;
}

/// Returns the middle out of three values
template <typename T>
inline T median(T a, T b, T c) {
    return max(min(a, b), min(max(a, b), c));
}

/// Returns the weighted average of a and b.
template <typename T, typename S>
inline T mix(T a, T b, S weight) {
    return T((S(1)-weight)*a+weight*b);
}

/// Clamps the number to the interval from 0 to 1.
template <typename T>
inline T clamp(T n) {
    return n >= T(0) && n <= T(1) ? n : T(n > T(0));
}

/// Clamps the number to the interval from 0 to b.
template <typename T>
inline T clamp(T n, T b) {
    return n >= T(0) && n <= b ? n : T(n > T(0))*b;
}

/// Clamps the number to the interval from a to b.
template <typename T>
inline T clamp(T n, T a, T b) {
    return n >= a && n <= b ? n : n < a ? a : b;
}

/// Returns 1 for non-negative values and -1 for negative values.
template <typename T>
inline int nonZeroSign(T n) {
    return 2*(n > T(0))-1;
}

// "core/Vector2.h"
/**
* A 2-dimensional euclidean vector with double precision.
* Implementation based on the Vector2 template from Artery Engine.
* @author Viktor Chlumsky
*/
struct Vector2 {

    double x, y;

    Vector2(double val = 0);
    Vector2(double x, double y);
    /// Sets the vector to zero.
    void reset();
    /// Sets individual elements of the vector.
    void set(double x, double y);
    /// Returns the vector's length.
    double length() const;
    /// Returns the angle of the vector in radians (atan2).
    double direction() const;
    /// Returns the normalized vector - one that has the same direction but unit length.
    Vector2 normalize(bool allowZero = false) const;
    /// Returns a vector with the same length that is orthogonal to this one.
    Vector2 getOrthogonal(bool polarity = true) const;
    /// Returns a vector with unit length that is orthogonal to this one.
    Vector2 getOrthonormal(bool polarity = true, bool allowZero = false) const;
    /// Returns a vector projected along this one.
    Vector2 project(const Vector2 &vector, bool positive = false) const;
    operator const void *() const;
    bool operator!() const;
    bool operator==(const Vector2 &other) const;
    bool operator!=(const Vector2 &other) const;
    Vector2 operator+() const;
    Vector2 operator-() const;
    Vector2 operator+(const Vector2 &other) const;
    Vector2 operator-(const Vector2 &other) const;
    Vector2 operator*(const Vector2 &other) const;
    Vector2 operator/(const Vector2 &other) const;
    Vector2 operator*(double value) const;
    Vector2 operator/(double value) const;
    Vector2 & operator+=(const Vector2 &other);
    Vector2 & operator-=(const Vector2 &other);
    Vector2 & operator*=(const Vector2 &other);
    Vector2 & operator/=(const Vector2 &other);
    Vector2 & operator*=(double value);
    Vector2 & operator/=(double value);
    /// Dot product of two vectors.
    friend double dotProduct(const Vector2 &a, const Vector2 &b);
    /// A special version of the cross product for 2D vectors (returns scalar value).
    friend double crossProduct(const Vector2 &a, const Vector2 &b);
    friend Vector2 operator*(double value, const Vector2 &vector);
    friend Vector2 operator/(double value, const Vector2 &vector);

};

/// A vector may also represent a point, which shall be differentiated semantically using the alias Point2.
typedef Vector2 Point2;


// "core/equation-solveer.h"
// ax^2 + bx + c = 0
int solveQuadratic(double x[2], double a, double b, double c);

// ax^3 + bx^2 + cx + d = 0
int solveCubic(double x[3], double a, double b, double c, double d);


// "core/SignedDistance.h"
/// Represents a signed distance and alignment, which together can be compared to uniquely determine the closest edge segment.
class SignedDistance {

public:
    static const SignedDistance INFINITE;

    double distance;
    double dot;

    SignedDistance();
    SignedDistance(double dist, double d);

    friend bool operator<(SignedDistance a, SignedDistance b);
    friend bool operator>(SignedDistance a, SignedDistance b);
    friend bool operator<=(SignedDistance a, SignedDistance b);
    friend bool operator>=(SignedDistance a, SignedDistance b);

};


// "core/EdgeColor.h"
/// Edge color specifies which color channels an edge belongs to.
enum EdgeColor {
    BLACK = 0,
    RED = 1,
    GREEN = 2,
    YELLOW = 3,
    BLUE = 4,
    MAGENTA = 5,
    CYAN = 6,
    WHITE = 7
};


// "core/edge-segments.h"
// Parameters for iterative search of closest point on a cubic Bezier curve. Increase for higher precision.
#define MSDFGEN_CUBIC_SEARCH_STARTS 4
#define MSDFGEN_CUBIC_SEARCH_STEPS 4

/// An abstract edge segment.
class EdgeSegment {

public:
    EdgeColor color;

    EdgeSegment(EdgeColor edgeColor = WHITE) : color(edgeColor) { }
    virtual ~EdgeSegment() { }
    /// Creates a copy of the edge segment.
    virtual EdgeSegment * clone() const = 0;
    /// Returns the point on the edge specified by the parameter (between 0 and 1).
    virtual Point2 point(double param) const = 0;
    /// Returns the direction the edge has at the point specified by the parameter.
    virtual Vector2 direction(double param) const = 0;
    /// Returns the minimum signed distance between origin and the edge.
    virtual SignedDistance signedDistance(Point2 origin, double &param) const = 0;
    /// Converts a previously retrieved signed distance from origin to pseudo-distance.
    virtual void distanceToPseudoDistance(SignedDistance &distance, Point2 origin, double param) const;
    /// Adjusts the bounding box to fit the edge segment.
    virtual void bounds(double &l, double &b, double &r, double &t) const = 0;

    /// Moves the start point of the edge segment.
    virtual void moveStartPoint(Point2 to) = 0;
    /// Moves the end point of the edge segment.
    virtual void moveEndPoint(Point2 to) = 0;
    /// Splits the edge segments into thirds which together represent the original edge.
    virtual void splitInThirds(EdgeSegment *&part1, EdgeSegment *&part2, EdgeSegment *&part3) const = 0;

};

/// A line segment.
class LinearSegment : public EdgeSegment {

public:
    Point2 p[2];

    LinearSegment(Point2 p0, Point2 p1, EdgeColor edgeColor = WHITE);
    LinearSegment * clone() const;
    Point2 point(double param) const;
    Vector2 direction(double param) const;
    SignedDistance signedDistance(Point2 origin, double &param) const;
    void bounds(double &l, double &b, double &r, double &t) const;

    void moveStartPoint(Point2 to);
    void moveEndPoint(Point2 to);
    void splitInThirds(EdgeSegment *&part1, EdgeSegment *&part2, EdgeSegment *&part3) const;

};

/// A quadratic Bezier curve.
class QuadraticSegment : public EdgeSegment {

public:
    Point2 p[3];

    QuadraticSegment(Point2 p0, Point2 p1, Point2 p2, EdgeColor edgeColor = WHITE);
    QuadraticSegment * clone() const;
    Point2 point(double param) const;
    Vector2 direction(double param) const;
    SignedDistance signedDistance(Point2 origin, double &param) const;
    void bounds(double &l, double &b, double &r, double &t) const;

    void moveStartPoint(Point2 to);
    void moveEndPoint(Point2 to);
    void splitInThirds(EdgeSegment *&part1, EdgeSegment *&part2, EdgeSegment *&part3) const;

};

/// A cubic Bezier curve.
class CubicSegment : public EdgeSegment {

public:
    Point2 p[4];

    CubicSegment(Point2 p0, Point2 p1, Point2 p2, Point2 p3, EdgeColor edgeColor = WHITE);
    CubicSegment * clone() const;
    Point2 point(double param) const;
    Vector2 direction(double param) const;
    SignedDistance signedDistance(Point2 origin, double &param) const;
    void bounds(double &l, double &b, double &r, double &t) const;

    void moveStartPoint(Point2 to);
    void moveEndPoint(Point2 to);
    void splitInThirds(EdgeSegment *&part1, EdgeSegment *&part2, EdgeSegment *&part3) const;

};


// "core/EdgeHolder.h"
/// Container for a single edge of dynamic type.
class EdgeHolder {

public:
    EdgeHolder();
    EdgeHolder(EdgeSegment *segment);
    EdgeHolder(Point2 p0, Point2 p1, EdgeColor edgeColor = WHITE);
    EdgeHolder(Point2 p0, Point2 p1, Point2 p2, EdgeColor edgeColor = WHITE);
    EdgeHolder(Point2 p0, Point2 p1, Point2 p2, Point2 p3, EdgeColor edgeColor = WHITE);
    EdgeHolder(const EdgeHolder &orig);
#ifdef MSDFGEN_USE_CPP11
    EdgeHolder(EdgeHolder &&orig);
#endif
    ~EdgeHolder();
    EdgeHolder & operator=(const EdgeHolder &orig);
#ifdef MSDFGEN_USE_CPP11
    EdgeHolder & operator=(EdgeHolder &&orig);
#endif
    EdgeSegment & operator*();
    const EdgeSegment & operator*() const;
    EdgeSegment * operator->();
    const EdgeSegment * operator->() const;
    operator EdgeSegment *();
    operator const EdgeSegment *() const;

private:
    EdgeSegment *edgeSegment;

};


// "core/Contour.h"
/// A single closed contour of a shape.
class Contour {

public:
    /// The sequence of edges that make up the contour.
    std::vector<EdgeHolder> edges;

    /// Adds an edge to the contour.
    void addEdge(const EdgeHolder &edge);
#ifdef MSDFGEN_USE_CPP11
    void addEdge(EdgeHolder &&edge);
#endif
    /// Creates a new edge in the contour and returns its reference.
    EdgeHolder & addEdge();
    /// Computes the bounding box of the contour.
    void bounds(double &l, double &b, double &r, double &t) const;

};


// "core/Shape.h"
/// Vector shape representation.
class Shape {

public:
    /// The list of contours the shape consists of.
    std::vector<Contour> contours;
    /// Specifies whether the shape uses bottom-to-top (false) or top-to-bottom (true) Y coordinates.
    bool inverseYAxis;

    Shape();
    /// Adds a contour.
    void addContour(const Contour &contour);
#ifdef MSDFGEN_USE_CPP11
    void addContour(Contour &&contour);
#endif
    /// Adds a blank contour and returns its reference.
    Contour & addContour();
    /// Normalizes the shape geometry for distance field generation.
    void normalize();
    /// Performs basic checks to determine if the object represents a valid shape.
    bool validate() const;
    /// Computes the shape's bounding box.
    void bounds(double &l, double &b, double &r, double &t) const;

};


// "core/Bitmap.h"
/// A floating-point RGB pixel.
struct FloatRGB {
    float r, g, b;
};

/// A 2D image bitmap.
template <typename T>
class Bitmap {

public:
    Bitmap();
    Bitmap(int width, int height);
    Bitmap(const Bitmap<T> &orig);
#ifdef MSDFGEN_USE_CPP11
    Bitmap(Bitmap<T> &&orig);
#endif
    ~Bitmap();
    Bitmap<T> & operator=(const Bitmap<T> &orig);
#ifdef MSDFGEN_USE_CPP11
    Bitmap<T> & operator=(Bitmap<T> &&orig);
#endif
    /// Bitmap width in pixels.
    int width() const;
    /// Bitmap height in pixels.
    int height() const;
    T & operator()(int x, int y);
    const T & operator()(int x, int y) const;

private:
    T *content;
    int w, h;

};


// "core/edge-coloring.h"
/** Assigns colors to edges of the shape in accordance to the multi-channel distance field technique.
 *  May split some edges if necessary.
 *  angleThreshold specifies the maximum angle (in radians) to be considered a corner, for example 3 (~172 degrees).
 *  Values below 1/2 PI will be treated as the external angle.
 */
void edgeColoringSimple(Shape &shape, double angleThreshold);


// "msdfgen.h"
#define MSDFGEN_VERSION "1.1"

/// Generates a conventional single-channel signed distance field.
void generateSDF(Bitmap<float> &output, const Shape &shape, double range, const Vector2 &scale, const Vector2 &translate);

/// Generates a single-channel signed pseudo-distance field.
void generatePseudoSDF(Bitmap<float> &output, const Shape &shape, double range, const Vector2 &scale, const Vector2 &translate);

/// Generates a multi-channel signed distance field. Edge colors must be assigned first! (see edgeColoringSimple)
void generateMSDF(Bitmap<FloatRGB> &output, const Shape &shape, double range, const Vector2 &scale, const Vector2 &translate, double edgeThreshold = 1.00000001);

}

