
// AssetDB
const char *AssetTypeNameFromAssetType(AssetType type) {
    if (type > ENUM_VALUECOUNT(AssetType)) {
        return NULL;
    }

    return AssetNameTable[type];
}

Schema *AssetSchemaFromAssetType(AssetType type) {
    if (const char *schemaName = AssetTypeNameFromAssetType(type)) {
        return Game.GetSchema(schemaName);
    }

    return NULL;
}

typedef b32 AssetLoadFileFunc(AssetDB *assets, const char *path, void *data);
typedef b32 AssetLoadMainThreadFunc(void *data);

typedef void AssetUnloadFileFunc(AssetDB *assets, void *data);

#include "assets.generated.cpp"

#define DECLARE_ASSETDATA_UNION_FIELD(name) name _##name;

struct AssetData {
    AssetType type;

    union {
        FOREACH_ASSET(DECLARE_ASSETDATA_UNION_FIELD);
    } data;

    AssetDataState state;

    AssetData *nextQueued;
};


 b32 AssetHandleIsValid(AssetHandle handle) {
    return (handle.type != AssetType_Invalid && handle.asset != NULL);
}

b32 AssetHandleIsProcedural(AssetHandle handle) {
    b32 result = (AssetHandleIsValid(handle) &&
                  handle.asset->source.type == AssetSourceType_Procedural);
    return result;
}


void AllocateAssetDB(AssetDB *assets, const char *searchPath) {
    Plat.ClearMem(assets, sizeof(AssetDB));

    assets->searchPath = searchPath;

    AllocateHashTable(&assets->assetNames, 1024);

    AllocateMemoryPool(&assets->assetPool, 64);
    AllocateMemoryPool(&assets->assetDataPool, 64);

    RegisterAssetTypeExtensions(assets);

#if ETC_DEBUG
    AllocateHashTable(&assets->assetPaths, 1024);
#endif
}

void DeallocateAssetDB(AssetDB *assets) {
    DeallocateHashTable(&assets->assetNames);

    DeallocateMemoryPool(&assets->assetPool);
    DeallocateMemoryPool(&assets->assetDataPool);

    // TODO: loop through all Assets and clear out allocated names/paths
    // TODO: probably wise to clear out all loaded assets explicitly (including GPU calls)
}

Asset *AllocateAsset(AssetDB *assets, const char *name, AssetType type) {

    Asset *asset = Allocate_(&assets->assetPool);

    asset->name = StringCopy(name);
    asset->type = type;

    HashTableInsert(&assets->assetNames, asset->name, asset);

    return asset;
}

AssetData *AllocateAssetData(AssetDB *assets, Asset *asset) {
    AssetData *assetData = Allocate_(&assets->assetDataPool);

    assetData->type = asset->type;
    assetData->state = AssetDataState_Loading;

    if (asset->assetDataQueueTail != NULL) {
        asset->assetDataQueueTail->nextQueued = assetData;
    }

    asset->assetDataQueueTail = assetData;
    if (asset->assetDataQueueHead == NULL) {
        asset->assetDataQueueHead = assetData;
    }

    return assetData;
}

void DeallocateAssetData(AssetDB *assets, AssetData *assetData) {
    Deallocate_(&assets->assetDataPool, assetData);
}


Asset *GetAssetByName(AssetDB *assets, const char *name) {
    Asset *asset = NULL;
    if (name != NULL) {
        HashTableGet(&assets->assetNames, name, &asset);
    }
    
    return asset;
}

const char *FindPathForAsset(AssetDB *assets, const char *name, AssetType type) {
    DynamicArray<const char *> extensions = assets->assetTypeExtensions[type];

    const char *path = Plat.FindFileInDirectory(assets->searchPath, name, extensions.count, extensions.data);

    return path;
}

void *GetAssetDataData(AssetDB *assets, Asset *asset, AssetType type) {
    void *result = NULL;

    AcquireLock(&assets->lock);

    if (asset != NULL && asset->type == type) {
        AssetData *head = asset->assetDataQueueHead;
        if (head != NULL && head->state == AssetDataState_Active) {
            result = &head->data;
        }
    }

    ReleaseLock(&assets->lock);

    return result;
}


void UnloadAssetFile(AssetDB *assets, Asset *asset, AssetData *assetData) {
    if (asset->source.type == AssetSourceType_File) {
        AssetUnloadFileFunc *unloadFileFunc = AssetUnloadFileFuncTable[assetData->type];
        if (unloadFileFunc != NULL) {
            unloadFileFunc(assets, &assetData->data);
        }
    }
}

// @NOTE: it's assumed that this function is called while the AssetDB lock is held
void FlushCompletedAssetDataLoads(AssetDB *assets, Asset *asset) {

    AssetData *current = asset->assetDataQueueHead;
    AssetData *prev = NULL;

    while (current != NULL) {
        AssetData *next = current->nextQueued;

        switch (current->state) {
            case AssetDataState_Active: {
            } break;

            case AssetDataState_Loading: {
                goto endloop;
            } break;

            case AssetDataState_Loaded: {
                // Unload all previous AssetDatas in the chain
                {
                    AssetData *head = asset->assetDataQueueHead;
                    while (head != current) {
                        AssetData *_next = head->nextQueued;

                        UnloadAssetFile(assets, asset, head);
                        DeallocateAssetData(assets, head);

                        head = _next;
                    }
                }

                Plat.DebugPrint("Load: %s", asset->name);

                asset->assetDataQueueHead = current;
                current->state = AssetDataState_Active;
            } break;

            case AssetDataState_LoadFailed: {
                Plat.DebugPrint("Load FAILED: %s", asset->name);

                UnloadAssetFile(assets, asset, current);
                DeallocateAssetData(assets, current);

                if (prev != NULL) {
                    prev->nextQueued = next;
                    current = prev;
                }
                else {
                    asset->assetDataQueueHead = next;
                    current = NULL;
                }

                if (next == NULL) {
                    asset->assetDataQueueTail = prev;
                }
            } break;
        }

        prev = current;
        current = next;
    }

endloop: ;
}


// @TODO: weird, this payload size is maximum 24 bytes! (see Task struct in task.h)
struct AssetDataLoadTransaction {
    AssetDB *assets;
    Asset *asset;
    AssetData *assetData;
};

// @NOTE: it's assumed that this function is called while the AssetDB lock is held
b32 LoadAssetMainThread_(AssetDB *assets, Asset *asset, AssetData *assetData) {
    b32 loadedMainThread = true;

    switch (assetData->state) {
        case AssetDataState_Loading: {
            AssetLoadMainThreadFunc *loadMainThreadFunc = AssetLoadMainThreadFuncTable[asset->type];
            if (loadMainThreadFunc != NULL) {
                loadedMainThread = loadMainThreadFunc(&assetData->data);
            }

            if (loadedMainThread) {
                assetData->state = AssetDataState_Loaded;
            }
            else {
                assetData->state = AssetDataState_LoadFailed;
            }
        } break;

        case AssetDataState_LoadFailed: {
            loadedMainThread = false;
        } break;

        default: {
            ASSERT(false);
        } break;
    }

    FlushCompletedAssetDataLoads(assets, asset);

    return loadedMainThread;
}

void LoadAssetMainThreadProc(AssetDataLoadTransaction *load) {
    AssetDB *assets = load->assets;
    Asset *asset = load->asset;
    AssetData *assetData = load->assetData;

    AcquireLock(&assets->lock);

    LoadAssetMainThread_(load->assets, load->asset, load->assetData);

    ReleaseLock(&assets->lock);
}


// @NOTE: it's assumed that this function is called while the AssetDB lock is _NOT_ held
b32 LoadAssetFromPath_(AssetDB *assets, Asset *asset, AssetData *assetData, const char *path) {
    b32 loadedFromFile = false;

    ASSERT(asset->source.type == AssetSourceType_File);

    if (path != NULL) {
        asset->source.path = path;

        AssetLoadFileFunc *loadFileFunc = AssetLoadFileFuncTable[asset->type];
        if (loadFileFunc != NULL) {
            loadedFromFile = loadFileFunc(assets, path, &assetData->data);
        }
        else if (Schema *schema = AssetSchemaFromAssetType(asset->type)) {
            EtcTxtParseState file;

            if (OpenFileForRead(&file, path)) {
                const char *name = NULL;
                if (ParseIdentifier(&file, &name) &&
                    StringEquals(name, schema->name)) {
                    loadedFromFile = DeserializeValue(&file, schema, &assetData->data);
                }

                CloseFile(&file);
            }
        }

#if ETC_DEBUG
        AcquireLock(&assets->lock);

        // Even if the asset load failed, the Asset reference will
        // still be valid. We'll want to always register the path to
        // handle the case where the first load attempt fails.
        AssetDB::AssetDebugRecord record;
        record.asset = asset;
        record.lastModifiedTime = Plat.FileLastModifiedTime(path);

        HashTableInsert(&assets->assetPaths, path, record, true);

        ReleaseLock(&assets->lock);
#endif
    }

    // Caller is responsible for cleanup after load failure
    return loadedFromFile;
}

void LoadAssetFileProc(AssetDataLoadTransaction *load) {
    AssetDB *assets = load->assets;
    Asset *asset = load->asset;
    AssetData *assetData = load->assetData;

    ASSERT(assetData->state == AssetDataState_Loading);

    const char *path = FindPathForAsset(assets, asset->name, asset->type);
    asset->source.type = AssetSourceType_File;

    if (!LoadAssetFromPath_(assets, asset, assetData, path)) {
        assetData->state = AssetDataState_LoadFailed;
    }

    Task *task = AllocateTask(Plat.mainThreadScheduler);
    task->fn = (TaskFunc *)LoadAssetMainThreadProc;
    Plat.CopyMem(task->data, load, sizeof(AssetDataLoadTransaction));

    SubmitTask(Plat.mainThreadScheduler, task);
}

AssetHandle LoadAsset(AssetDB *assets, AssetType type, const char *name, b32 forceReload) {
    AssetHandle result = {};

    if (name == NULL) {
        return result;
    }

    if (strlen(name) == 0) {
        return result;
    }

    AcquireLock(&assets->lock);

    Asset *asset = GetAssetByName(assets, name);
    if (asset == NULL || asset->type != type || asset->source.type != AssetSourceType_File) {
        asset = AllocateAsset(assets, name, type);
    }

    if (forceReload || asset->assetDataQueueHead == NULL) {
        AssetData *assetData = AllocateAssetData(assets, asset);
        assetData->state = AssetDataState_Loading;

        Task *task = AllocateTask(Plat.scheduler);
        task->fn = (TaskFunc *)LoadAssetFileProc;

        AssetDataLoadTransaction *load = (AssetDataLoadTransaction *)task->data;
        load->assets = assets;
        load->asset = asset;
        load->assetData = assetData;

        SubmitTask(Plat.scheduler, task);
    }

    ReleaseLock(&assets->lock);

    result.type = type;
    result.asset = asset;

    return result;
}

AssetHandle LoadProceduralAsset(AssetDB *assets, AssetType type, const char *name, void *data) {
    AcquireLock(&assets->lock);

    Asset *asset = GetAssetByName(assets, name);

    // Reuse the asset unless if there's already data in it
    if (asset == NULL || asset->assetDataQueueHead != NULL) {
        asset = AllocateAsset(assets, name, type);
    }

    asset->source.type = AssetSourceType_Procedural;

    AssetData *assetData = AllocateAssetData(assets, asset);

    uint32 dataSize = 0;

    if (Schema *assetSchema = AssetSchemaFromAssetType(asset->type)) {
        dataSize = assetSchema->sizeInBytes;
    }
    else {
        return AssetHandleInvalid();
    }

    Plat.CopyMem(&assetData->data, data, dataSize);
    assetData->state = AssetDataState_Loading;


    AssetDataLoadTransaction load;
    load.assets = assets;
    load.asset = asset;
    load.assetData = assetData;

    Task *task = AllocateTask(Plat.mainThreadScheduler);
    task->fn = (TaskFunc *)LoadAssetMainThreadProc;
    Plat.CopyMem(task->data, &load, sizeof(AssetDataLoadTransaction));

    SubmitTask(Plat.mainThreadScheduler, task);


    ReleaseLock(&assets->lock);

    AssetHandle handle = {};
    handle.type = type;
    handle.asset = asset;

    return handle;
}

AssetHandle SyncLoadProceduralAsset(AssetDB *assets, AssetType type, const char *name, void *data) {
    ASSERT(data != NULL && name != NULL && type != AssetType_Invalid);

    uint32 dataSize = 0;
    if (Schema *assetSchema = AssetSchemaFromAssetType(type)) {
        dataSize = assetSchema->sizeInBytes;
    }
    else {
        return AssetHandleInvalid();
    }

    
    AcquireLock(&assets->lock);

    Asset *asset = AllocateAsset(assets, name, type);

    AssetData *assetData = AllocateAssetData(assets, asset);
    asset->source.type = AssetSourceType_Procedural;

    Plat.CopyMem(&assetData->data, data, dataSize);
    assetData->state = AssetDataState_Loading;

    LoadAssetMainThread_(assets, asset, assetData);

    ReleaseLock(&assets->lock);

    AssetHandle handle = {};
    handle.type = type;
    handle.asset = asset;

    return handle;
}

AssetHandle SyncLoadAssetFromPath(AssetDB *assets, AssetType type, const char *path) {
    ASSERT(path != NULL);

    AcquireLock(&assets->lock);

    const char *name;
    SplitFileExtension(path, &name, NULL);

    int32 lastSlashIndex = Max(FindLastSubstring(path, "/"), FindLastSubstring(path, "\\"));
    if (lastSlashIndex >= 0) {
        name = &name[lastSlashIndex + 1];
    }

    Asset *asset = AllocateAsset(assets, name, type);
    asset->source.type = AssetSourceType_File;

    AssetData *assetData = AllocateAssetData(assets, asset);
    assetData->state = AssetDataState_Loading;

    ReleaseLock(&assets->lock);

    if (!LoadAssetFromPath_(assets, asset, assetData, path)) {
        assetData->state = AssetDataState_LoadFailed;
    }

    AcquireLock(&assets->lock);

    LoadAssetMainThread_(assets, asset, assetData);

    ReleaseLock(&assets->lock);

    AssetHandle handle = {};
    handle.type = type;
    handle.asset = asset;

    return handle;
}

AssetHandle SyncLoadAsset(AssetDB *assets, AssetType type, const char *name) {
    const char *path = FindPathForAsset(assets, name, type);

    if (path != NULL) {
        return SyncLoadAssetFromPath(assets, type, path);
    }
    return AssetHandleInvalid();
}


#if ETC_DEBUG
void FindModifiedAssetPaths(AssetDB *assets) {
    HashTable<const char *, AssetDB::AssetDebugRecord> *table = &assets->assetPaths;

    if (table->count == 0 || table->capacity == 0) {
        return;
    }

    static const uint32 numAssetsToCheck = 32;
    static const uint32 numRecordsToCheck = 512;
    uint32 assetsChecked = 0;

    uint32 iteration = assets->modifiedAssetIteration;

    for (int i = 0; i < numRecordsToCheck; i++) {
        if (table->occupied[iteration]) {
            const char *path = table->keys[iteration];
            AssetDB::AssetDebugRecord *record = &table->values[iteration];

            uint64 lastModifiedTime = Plat.FileLastModifiedTime(path);

            if (lastModifiedTime > record->lastModifiedTime) {

                // Write out the lastModifiedTime to ensure we don't
                // get duplicate reloads for the same modification
                record->lastModifiedTime = lastModifiedTime;

                RegisterModifiedAssetPath(assets, path);
            }

            assetsChecked++;

            if (assetsChecked >= numAssetsToCheck) {
                break;
            }
        }

        iteration = (iteration + 1) % table->capacity;
    }

    assets->modifiedAssetIteration = iteration;
}

void ReloadAssetFileProc(AssetDataLoadTransaction *load) {
    AssetDB *assets = load->assets;
    Asset *asset = load->asset;
    AssetData *assetData = load->assetData;

    ASSERT(assetData->state == AssetDataState_Loading);

    const char *path = asset->source.path;
    ASSERT(path != NULL);

    if (!LoadAssetFromPath_(assets, asset, assetData, path)) {
        assetData->state = AssetDataState_LoadFailed;
    }

    Task *task = AllocateTask(Plat.mainThreadScheduler);
    task->fn = (TaskFunc *)LoadAssetMainThreadProc;
    Plat.CopyMem(task->data, load, sizeof(AssetDataLoadTransaction));

    SubmitTask(Plat.mainThreadScheduler, task);
}

void ReloadModifiedAssets(AssetDB *assets) {
    for (int i = 0; i < assets->modifiedFilePaths.count; i++) {
        const char *path = assets->modifiedFilePaths[i];

        AssetDB::AssetDebugRecord record;
        if (HashTableGet(&assets->assetPaths, path, &record)) {
            AssetData *assetData = AllocateAssetData(assets, record.asset);
            assetData->state = AssetDataState_Loading;

            Task *task = AllocateTask(Plat.scheduler);
            task->fn = (TaskFunc *)ReloadAssetFileProc;

            AssetDataLoadTransaction *load = (AssetDataLoadTransaction *)task->data;
            load->assets = assets;
            load->asset = record.asset;
            load->assetData = assetData;

            SubmitTask(Plat.scheduler, task);
        }
    }

    DynamicArrayClear(&assets->modifiedFilePaths);
}
#endif


////////////////////////////////////////////////
void AllocateTexture(Texture *texture, uint16 width, uint16 height, TextureFormat format) {
    ASSERT(width > 0);
    ASSERT(height > 0);

    texture->width = width;
    texture->height = height;
    texture->format = format;

    texture->bufferSize = width * height * BytesPerPixel(format);
    texture->data.data = Plat.AllocateMem(texture->bufferSize);
    Plat.ClearMem(texture->data.data, texture->bufferSize);
}

void DeallocateTexture(Texture *texture) {
    Plat.DeallocateMem(texture->data.data);
}

void AllocateCubemap(Cubemap *cubemap, uint16 width, uint16 height, uint8 components) {
    ASSERT(width > 0);
    ASSERT(height > 0);

    cubemap->width = width;
    cubemap->height = height;
    cubemap->components = components;

    uint32 faceDataSize = sizeof(uint8) * width * height * components;
    for (int i = 0; i < 6; i++) {
        cubemap->data[i] = (uint8 *)Plat.AllocateMem(faceDataSize);
    }
}

void DeallocateCubemap(Cubemap *cubemap) {
    for (int i = 0; i < 6; i++) {
        Plat.DeallocateMem(cubemap->data[i]);
    }
}


void MeshWriteFile(const char *path, Mesh *mesh) {
    FileHandle file;
    u32 fileSize = mesh->vertexBufferSize * 8;
    if (Plat.OpenFileForWrite(path, &file, fileSize)) {

        uint32 version = 0;
        WriteUInt32(&file, version);

        const char *primitiveTypeName = ENUM_STRINGVALUE(PrimitiveType, mesh->primitiveType);
        uint32 primitiveTypeNameLength = StringLength(primitiveTypeName);

        WriteUInt32(&file, primitiveTypeNameLength);
        WriteBytes(&file, (void *)primitiveTypeName, primitiveTypeNameLength);

        uint32 attribCount = 0;
        const char *attribs[6] = {};

        if (mesh->positions != NULL) { attribs[attribCount++] = "Position"; }
        if (mesh->texcoords != NULL) { attribs[attribCount++] = "TexCoord"; }
        if (mesh->normals != NULL) { attribs[attribCount++] = "Normal"; }
        if (mesh->colors != NULL) { attribs[attribCount++] = "Color"; }
        if (mesh->boneIndices != NULL) { attribs[attribCount++] = "BoneIndices"; }
        if (mesh->boneWeights != NULL) { attribs[attribCount++] = "BoneWeights"; }

        WriteUInt32(&file, attribCount);
        for (int i = 0; i < attribCount; i++) {
            const char *attribName = attribs[i];
            uint32 attribNameLength = StringLength(attribName);

            WriteUInt32(&file, attribNameLength);
            WriteBytes(&file, (void *)attribName, attribNameLength);
        }


        WriteUInt32(&file, mesh->vertexCount);

        WriteUInt32(&file, mesh->vertexBufferSize);
        WriteUInt32(&file, mesh->indexCount);
        WriteUInt32(&file, mesh->submeshCount);

        WriteBytes(&file, mesh->vertices, mesh->vertexBufferSize);
        WriteBytes(&file, mesh->indices, sizeof(uint32) * mesh->indexCount);
        WriteBytes(&file, mesh->submeshLengths, sizeof(uint32) * mesh->submeshCount);

        for (int i = 0; i < mesh->submeshCount; i++) {
            WriteReal32(&file, mesh->materials[i].diffuseColor.r);
            WriteReal32(&file, mesh->materials[i].diffuseColor.g);
            WriteReal32(&file, mesh->materials[i].diffuseColor.b);
            WriteReal32(&file, mesh->materials[i].diffuseColor.a);
        }

        Plat.CloseFile(&file);
    }
}


// Constants in little-endian form
#define WAV_RIFF_VALUE 0x46464952
#define WAV_WAVE_VALUE 0x45564157
#define WAV_FMT_VALUE 0x20746d66
#define WAV_DATA_VALUE 0x61746164

enum WavFormat {
    WAV_PCM                = 0x01,
    WAV_IEEEFloatingPoint  = 0x03,
    WAV_ALaw               = 0x06,
    WAV_MuLaw              = 0x07,
    WAV_IMAADPCM           = 0x11,
    WAV_YamahaITUG723ADPCM = 0x16,
    WAV_GSM610             = 0x31,
    WAV_ITUG721ADPCM       = 0x40,
    WAV_MPEG               = 0x50,
    WAV_Extensible         = 0xFFFE
};

b32 LoadSoundClipFile(const char *path, SoundClip *sound) {
    ASSERT(sound != NULL);

    b32 success = true;
    FileHandle file;

    if (Plat.OpenFileForRead(path, &file)) {
        WavFormat format;

        int32 riffID = ReadInt32(&file);
        int32 chunkSize = ReadInt32(&file);
        int32 riffFormat = ReadInt32(&file);

        if (riffID == WAV_RIFF_VALUE && riffFormat == WAV_WAVE_VALUE) {

            int16 numChannels = 0;
            int16 bitsPerSample = 0;
            int32 sampleRate = 0;
            
            b32 atData = false;
            while (!atData) {
                int32 subchunkID = ReadInt32(&file);

                switch (subchunkID) {
                    case WAV_FMT_VALUE: {
                        int32 subchunkSize = ReadInt32(&file);
                        uint32 nextSubchunkOffset = file.offset + (uint32)subchunkSize;

                        format = (WavFormat)ReadInt16(&file);
                        numChannels = ReadInt16(&file);
                        sampleRate = ReadInt32(&file);
                        int32 byteRate = ReadInt32(&file);
                        int16 blockAlign = ReadInt16(&file);
                        bitsPerSample = ReadInt16(&file);
                        
                        FileSeek(&file, nextSubchunkOffset);
                    } break;
                    case WAV_DATA_VALUE: {
                        atData = true;
                    } break;
                    default: {
                        int32 subchunkSize = ReadInt32(&file);
                        FileSeek(&file, file.offset + (uint32)subchunkSize);
                    } break;
                }
            }

            ASSERT(numChannels > 0);
            ASSERT(numChannels <= 2);
            ASSERT(sampleRate > 0);

            uint32 dataSizeInBytes = ReadInt32(&file);
            uint32 sampleCount = dataSizeInBytes / (bitsPerSample / 8) / numChannels;

            sound->sampleRate = sampleRate;
            sound->sampleCount = sampleCount;
            sound->channels = numChannels;

            switch (format) {
                case WAV_PCM: {
                    sound->data = (real32 *)Plat.AllocateMem(sizeof(real32) * sampleCount * numChannels);
                
                    // Always read in little-endian!
                    for (int i = 0; i < sampleCount * numChannels; i++) {
                        int16 sample = ReadInt16(&file);
                        sound->data[i] = sample / (real32)32768.0f;
                    }

                } break;
                default: {
                    success = false;
                } break;
            }
        }

        Plat.CloseFile(&file);
    }
    else {
        success = false;
    }

    return success;
}

b32 LoadSoundStreamFile(const char *path, SoundStream *soundStream) {
    ASSERT(soundStream);

    b32 success = false;

    FileHandle file;
    if (Plat.OpenFileForRead(path, &file)) {

        soundStream->data = (uint8 *)Plat.AllocateMem(file.fileSize);
        soundStream->dataSize = file.fileSize;
        Plat.CopyMem(soundStream->data, file.data, file.fileSize);
        
        Plat.CloseFile(&file);


        real32 **output;
        int32 bytesConsumed;
        int32 totalBytesConsumed = 0;
        int32 error;
        stb_vorbis *vorbis = stb_vorbis_open_memory(soundStream->data, soundStream->dataSize, &error, NULL);
        if (vorbis) {
            stb_vorbis_info info = stb_vorbis_get_info(vorbis);
            soundStream->sampleRate = info.sample_rate;
            soundStream->channels = info.channels;
            soundStream->sampleCount = stb_vorbis_stream_length_in_samples(vorbis);

            success = true;
        }

        stb_vorbis_close(vorbis);
    }

    return success;
}


// Texture Loading
b32 TextureLoadMainThread(Texture *texture) {
    CreateTexture2DOpenGL(texture);

    b32 success = texture->textureID != 0;
    return success;
}

b32 TextureLoadFile(AssetDB *assets, const char *path, Texture *texture) {
    b32 success = false;

    int width, height, loadedComponents;
    uint8 *data = stbi_load(path, &width, &height, &loadedComponents, 0);

    if (data != NULL) {
        success = true;

        TextureFormat format;
        if (loadedComponents == 1) {
            format = TextureFormat_R8;
        }
        else if (loadedComponents == 2) {
            format = TextureFormat_RG8;
        }
        else if (loadedComponents == 3) {
            format = TextureFormat_RGB8;
        }
        else if (loadedComponents == 4) {
            format = TextureFormat_RGBA8;
        }
        else {
            ASSERT(false);
        }

        AllocateTexture(texture, width, height, format);
        uint32 bytesPerPixel = BytesPerPixel(format);

        // Flip the image because stb_image assumes origin in upper left
        for (int y = 0; y < height; y++) {
            for (int x = 0; x < width; x++) {

                uint32 srcIndex = loadedComponents * ((height - 1 - y) * width + x);
                uint32 destIndex = loadedComponents * (y * width + x);

                Plat.CopyMem(texture->data.u8 + destIndex, data + srcIndex, bytesPerPixel);
            }
        }

        stbi_image_free(data);
    }

    return success;
}

void TextureWriteFile(const char *path, Texture *texture) {
    uint32 width = texture->width;
    uint32 height = texture->height;
    uint32 components = ComponentCount(texture->format);

    uint8 *convertedTexture = PushArray(Game.frameMem, uint8, texture->width * texture->height * components);
    uint32 destIndex = 0;
    switch (texture->format) {
        case TextureFormat_R8:
        case TextureFormat_RG8:
        case TextureFormat_RGB8:
        case TextureFormat_RGBA8: {
            u8 *srcData = texture->data.u8;

            for (int y = height - 1; y >= 0; y--) {
                for (int x = 0; x < width; x++) {
                    uint8 *src = srcData + (y * width * components);
                    Plat.CopyMem(convertedTexture + destIndex, src, width * components);
                    destIndex += width * components;
                }
            }
        } break;

        case TextureFormat_R32F:
        case TextureFormat_RG32F:
        case TextureFormat_RGB32F:
        case TextureFormat_RGBA32F: {
            real32 *srcData = texture->data.f32;

            for (int y = height - 1; y >= 0; y--) {
                for (int x = 0; x < width; x++) {
                    for (int c = 0; c < components; c++) {
                        int32 sourceIndex = (y * width * components) + x * components + c;
                        convertedTexture[destIndex++] = (uint8)(srcData[sourceIndex] * 255);
                    }
                }
            }
        } break;
    }

    if (HasFileExtension(path, "jpg") || HasFileExtension(path, "jpeg")) {
        stbi_write_jpg(path, width, height, components, convertedTexture, 75);
    }
    else if (HasFileExtension(path, "png")) {
        uint32 stride = sizeof(uint8) * components * width;
        stbi_write_png(path, width, height, components, convertedTexture, stride);
    }
    else if (HasFileExtension(path, "bmp")) {
        stbi_write_bmp(path, width, height, components, convertedTexture);
    }
    else if (HasFileExtension(path, "tga")) {
        stbi_write_tga(path, width, height, components, convertedTexture);
    }
}

void TextureUnloadFile(AssetDB *assets, Texture *texture) {
    DeleteTextureOpenGL(texture);
    Plat.DeallocateMem(texture->data.u8);
}

// Cubemap Loading
b32 CubemapLoadMainThread(Cubemap *cubemap) {
    CreateCubemapOpenGL(cubemap);

    b32 success = cubemap->textureID != 0;
    return success;
}

b32 CubemapLoadFile(AssetDB *assets, const char *path, Cubemap *cubemap) {
    b32 success = false;

    int32 totalWidth, totalHeight, loadedComponents;
    uint8 *data = stbi_load(path, &totalWidth, &totalHeight, &loadedComponents, 0);
    if (data != NULL) {
        success = true;

        int32 width = totalWidth / 2;
        int32 height = totalHeight / 3;
        AllocateCubemap(cubemap, width, height, loadedComponents);

        for (int face = 0; face < 6; face++) {
            
            vec2i upperLeft = {};

            switch (face) {
                case Cubemap_PosX: {
                    upperLeft.x = width;
                    upperLeft.y = 0;
                } break;
                case Cubemap_NegX: {
                    upperLeft.x = 0;
                    upperLeft.y = 0;
                } break;
                case Cubemap_PosY: {
                    upperLeft.x = width;
                    upperLeft.y = height;
                } break;
                case Cubemap_NegY: {
                    upperLeft.x = 0;
                    upperLeft.y = height;
                } break;
                case Cubemap_PosZ: {
                    upperLeft.x = width;
                    upperLeft.y = height * 2;
                } break;
                case Cubemap_NegZ: {
                    upperLeft.x = 0;
                    upperLeft.y = height * 2;
                } break;
            }

            for (int y = 0; y < height; y++) {
                for (int x = 0; x < width; x++) {
                
                    uint32 srcIndex = loadedComponents * ((upperLeft.y + y) * totalWidth + (upperLeft.x + x));
                    uint32 destIndex = loadedComponents * (y * width + x);

                    Plat.CopyMem(cubemap->data[face] + destIndex, data + srcIndex, loadedComponents);
                }
            }
        }

        stbi_image_free(data);
    }

    return success;
}

void CubemapUnloadFile(AssetDB *assets, Cubemap *cubemap) {
    DeleteCubemapOpenGL(cubemap);

    for (int i = 0; i < 6; i++) {
        Plat.DeallocateMem(cubemap->data[i]);
    }
}


// ShaderProgram Loading
b32 ShaderProgramLoadMainThread(ShaderProgram *program) {
    CreateShaderProgramOpenGL(program);

    b32 success = program->programID != 0;
    return success;
}

b32 ShaderProgramLoadFile(AssetDB *assets, const char *path, ShaderProgram *program) {
    ParsedGLSLSource glslSource;
    b32 success = false;

    if (ParseGLSL(assets->searchPath, path, &glslSource)) {
        success = true;

        program->vertexSource = glslSource.vertexSource;
        program->vertexSourceSize = glslSource.vertexSourceSize;

        program->geometrySource = glslSource.geometrySource;
        program->geometrySourceSize = glslSource.geometrySourceSize;

        program->fragmentSource = glslSource.fragmentSource;
        program->fragmentSourceSize = glslSource.fragmentSourceSize;

        program->computeSource = glslSource.computeSource;
        program->computeSourceSize = glslSource.computeSourceSize;
    }

    return success;
}

void ShaderProgramUnloadFile(AssetDB *assets, ShaderProgram *program) {
    DeleteShaderProgramOpenGL(program);

    // @TODO: deallocate uniforms list

    if (program->vertexSource != NULL) {
        Plat.DeallocateMem((char *)program->vertexSource);
    }
    if (program->fragmentSource != NULL) {
        Plat.DeallocateMem((char *)program->fragmentSource);
    }
    if (program->geometrySource != NULL) {
        Plat.DeallocateMem((char *)program->geometrySource);
    }
}


// FontTable Loading
b32 FontTableLoadMainThread(FontTable *font) {
    glGenTextures(1, &font->texcoordsMapID);
    glBindTexture(GL_TEXTURE_1D, font->texcoordsMapID);
    glTexImage1D(GL_TEXTURE_1D, 0, GL_RGBA32F, font->glyphCount, 0, GL_RGBA, GL_FLOAT, font->texcoordsMapData);
    glBindTexture(GL_TEXTURE_1D, 0);

    CreateTexture2DOpenGL(&font->atlas);

    b32 success = font->texcoordsMapID != 0 && font->atlas.textureID != 0;
    return success;
}

b32 FontTableLoadFile(AssetDB *assets, const char *path, FontTable *font) {
    FileHandle file;
    if (Plat.OpenFileForRead(path, &file)) {
        font->glyphCount = ReadInt32(&file);

        font->msdfRange = ReadReal32(&file);
        font->ascent = ReadReal32(&file);
        font->descent = ReadReal32(&file);
        font->lineHeight = ReadReal32(&file);
        font->unitsPerEM = ReadReal32(&file);

        AllocateHashTable(&font->codepointMap, font->glyphCount);
    
        for (int i = 0; i < font->glyphCount; i++) {
            uint32 unicode = (uint32)ReadInt32(&file); 
            uint32 codepoint = (uint32)ReadInt32(&file);

            if (unicode < 128) {
                font->ascii[unicode] = codepoint;
            }
            else {
                HashTableInsert(&font->codepointMap, unicode, codepoint);
            }
        }

        font->glyphs = (GlyphInfo *)Plat.AllocateMem(sizeof(GlyphInfo) * font->glyphCount);
        font->texcoordsMapData = (vec4 *)Plat.AllocateMem(sizeof(vec4) * font->glyphCount);

        for (int i = 0; i < font->glyphCount; i++) {
            font->glyphs[i].lowerLeft.x = ReadReal32(&file);
            font->glyphs[i].lowerLeft.y = ReadReal32(&file);

            font->glyphs[i].size.x = ReadReal32(&file);
            font->glyphs[i].size.y = ReadReal32(&file);

            font->glyphs[i].advance = ReadReal32(&file);


            // Spaces appear on some fonts to have nan sizes and
            // lowerlefts, so correct it here to make some layout
            // calculations easier
            if (isnan(font->glyphs[i].lowerLeft.x)) {
                font->glyphs[i].lowerLeft.x = 0;
            }

            if (isnan(font->glyphs[i].lowerLeft.y)) {
                font->glyphs[i].lowerLeft.y = 0;
            }

            if (isnan(font->glyphs[i].size.x)) {
                font->glyphs[i].size.x = 0;
            }

            if (isnan(font->glyphs[i].size.y)) {
                font->glyphs[i].size.y = 0;
            }
        }

        font->kernAdvances = (real32 *)Plat.AllocateMem(sizeof(real32) * font->glyphCount * font->glyphCount);
        for (int i = 0; i < font->glyphCount * font->glyphCount; i++) {
            font->kernAdvances[i] = ReadReal32(&file);
        }

        
        for (int i = 0; i < font->glyphCount; i++) {
            vec2 uvMin, uvMax;
            uvMin.x = ReadReal32(&file);
            uvMin.y = ReadReal32(&file);

            uvMax.x = ReadReal32(&file);
            uvMax.y = ReadReal32(&file);

            font->texcoordsMapData[i] = V4(uvMin.x, uvMin.y, uvMax.x, uvMax.y);
        }

        font->atlas.width = ReadInt32(&file);
        font->atlas.height = ReadInt32(&file);
        uint32 loadedComponents = ReadInt8(&file);

        if (loadedComponents == 1) {
            font->atlas.format = TextureFormat_R8;
        }
        else if (loadedComponents == 2) {
            font->atlas.format = TextureFormat_RG8;
        }
        else if (loadedComponents == 3) {
            font->atlas.format = TextureFormat_RGB8;
        }
        else if (loadedComponents == 4) {
            font->atlas.format = TextureFormat_RGBA8;
        }
        else {
            ASSERT(false);
        }

        uint32 bytesPerPixel = BytesPerPixel(font->atlas.format);

        font->atlas.bufferSize = font->atlas.width * font->atlas.height * bytesPerPixel;
        font->atlas.data.data = Plat.AllocateMem(font->atlas.bufferSize);
        ReadBytes(&file, font->atlas.bufferSize, font->atlas.data.u8);

        Plat.CloseFile(&file);
        
        return true;
    }

    return false;
}

void FontTableUnloadFile(AssetDB *assets, FontTable *font) {
}


// SoundClip Loading
b32 SoundClipLoadFile(AssetDB *assets, const char *path, SoundClip *soundClip) {
    b32 success = LoadSoundClipFile(path, soundClip);
    return success;
}

void SoundClipUnloadFile(AssetDB *assets, SoundClip *soundClip) {
}


// SoundStream Loading
b32 SoundStreamLoadFile(AssetDB *assets, const char *path, SoundStream *soundStream) {
    b32 success = LoadSoundStreamFile(path, soundStream);
    return success;
}

void SoundStreamUnloadFile(AssetDB *assets, AssetData *assetData) {
}

