
#ifndef _ETC_ASSETS_H
#define _ETC_ASSETS_H


struct AssetDB;

enum AssetType : uint32;
struct Asset;
struct AssetData;

struct AssetHandle {
    AssetType type;
    Asset *asset;
};

inline AssetHandle AssetHandleInvalid() {
    AssetHandle handle = {};
    return handle;
}

b32 AssetHandleIsValid(AssetHandle handle);
b32 AssetHandleIsProcedural(AssetHandle handle);

inline b32 operator==(AssetHandle a, AssetHandle b) {
    b32 result = a.type == b.type && a.asset == b.asset;

    return result;
}

inline b32 operator!=(AssetHandle a, AssetHandle b) {
    b32 result = a.type != b.type || a.asset != b.asset;

    return result;
}

AssetHandle LoadAsset(AssetDB *assets, AssetType type, const char *name, b32 forceReload = false);
AssetHandle LoadProceduralAsset(AssetDB *assets, AssetType type, const char *name, void *data);
AssetHandle SyncLoadProceduralAsset(AssetDB *assets, AssetType type, const char *name, void *data);
AssetHandle SyncLoadAssetFromPath(AssetDB *assets, AssetType type, const char *path);
AssetHandle SyncLoadAsset(AssetDB *assets, AssetType type, const char *name);

void *GetAssetDataData(AssetDB *assets, Asset *asset, AssetType type);

// @NOTE: used by the UI system to validate an asset handle
const char *FindPathForAsset(AssetDB *assets, const char *name, AssetType type);


#include "assets.generated.h"

const char *AssetTypeNameFromAssetType(AssetType type);
Schema *AssetSchemaFromAssetType(AssetType type);

enum AssetDataState {
    AssetDataState_Loading,
    AssetDataState_Loaded,
    AssetDataState_LoadFailed,

    AssetDataState_Active,
};

enum AssetSourceType {
    AssetSourceType_Invalid,
    AssetSourceType_File,
    AssetSourceType_Procedural,
};

struct AssetSource {
    AssetSourceType type;

    const char *path;
};

struct AssetData;

struct Asset {
    const char *name;
    AssetType type;
    AssetSource source;

    AssetData *assetDataQueueHead;
    AssetData *assetDataQueueTail;
};

struct AssetDB {
    SpinLock lock;

    const char *searchPath;
    DynamicArray<const char *> assetTypeExtensions[ENUM_VALUECOUNT(AssetType)];

    // Maps from name -> asset
    HashTable<const char *, Asset *> assetNames;

    MemoryPool<Asset> assetPool;
    MemoryPool<AssetData> assetDataPool;


#if ETC_DEBUG
    DynamicArray<const char *> modifiedFilePaths;

    // Maps from path -> asset
    struct AssetDebugRecord {
        Asset *asset;
        uint64 lastModifiedTime;
    };

    HashTable<const char *, AssetDebugRecord> assetPaths;
    uint32 modifiedAssetIteration;
#endif
};

void AllocateAssetDB(AssetDB *assets, const char *searchPath);
void DeallocateAssetDB(AssetDB *assets);


#if ETC_DEBUG
inline void RegisterModifiedAssetPath(AssetDB *assets, const char *path) {
    if (!Contains(&assets->modifiedFilePaths, path)) {
        PushBack(&assets->modifiedFilePaths, path);
    }
}

void FindModifiedAssetPaths(AssetDB *assets);
void ReloadModifiedAssets(AssetDB *assets);
#endif


// User-defined Assets (@TODO: it's now free to move anywhere!)
////////////////////////////////////////////////

ENUM()
enum TextureFormat : uint32 {
    TextureFormat_R8,
    TextureFormat_RG8,
    TextureFormat_RGB8,
    TextureFormat_RGBA8,

    TextureFormat_R32F,
    TextureFormat_RG32F,
    TextureFormat_RGB32F,
    TextureFormat_RGBA32F,
};

inline uint32 ComponentCount(TextureFormat format) {
    switch (format) {
        case TextureFormat_R8: {
            return 1;
        } break;
        case TextureFormat_RG8: {
            return 2;
        } break;
        case TextureFormat_RGB8: {
            return 3;
        } break;
        case TextureFormat_RGBA8: {
            return 4;
        } break;

        case TextureFormat_R32F: {
            return 1;
        } break;
        case TextureFormat_RG32F: {
            return 2;
        } break;
        case TextureFormat_RGB32F: {
            return 3;
        } break;
        case TextureFormat_RGBA32F: {
            return 4;
        } break;
    }

    return 0;
}

inline uint32 BytesPerPixel(TextureFormat format) {
    switch (format) {
        case TextureFormat_R8: {
            return 1 * sizeof(uint8);
        } break;
        case TextureFormat_RG8: {
            return 2 * sizeof(uint8);
        } break;
        case TextureFormat_RGB8: {
            return 3 * sizeof(uint8);
        } break;
        case TextureFormat_RGBA8: {
            return 4 * sizeof(uint8);
        } break;

        case TextureFormat_R32F: {
            return 1 * sizeof(real32);
        } break;
        case TextureFormat_RG32F: {
            return 2 * sizeof(real32);
        } break;
        case TextureFormat_RGB32F: {
            return 3 * sizeof(real32);
        } break;
        case TextureFormat_RGBA32F: {
            return 4 * sizeof(real32);
        } break;
    }

    return 0;
}

// @NOTE: this is just a hack because the preprocessor doesn't support anonymous unions inside STRUCT/ASSETS!
union _TextureData {
    void *data;
    uint8 *u8;
    RGBA8 *rgba8;

    real32 *f32;
    vec3 *v2;
    vec3 *v3;
    vec4 *v4;
};

ASSET({"jpg", "jpeg", "png", "bmp", "tga"}, Serializable = false)
struct Texture {
    TextureFormat format;

    uint32 width;
    uint32 height;

    uint32 bufferSize;
    _TextureData data;

    uint32 textureID;
};

ASSET({"jpg", "jpeg", "png", "bmp", "tga"}, Serializable = false)
struct Cubemap {
    uint32 width;
    uint32 height;
    uint8 components;

    uint8 *data[6];

    uint32 textureID;
};


STRUCT()
struct ShaderUniform {
    const char *name;
    int32 location;
};

ASSET({"glsl"}, Serialize = false)
struct ShaderProgram {
    const char *vertexSource;
    uint32 vertexSourceSize;

    const char *fragmentSource;
    uint32 fragmentSourceSize;

    const char *geometrySource;
    uint32 geometrySourceSize;

    const char *computeSource;
    uint32 computeSourceSize;
    
    uint32 programID;
    uint32 vertexID;
    uint32 fragmentID;
    uint32 geometryID;
    uint32 computeID;

    DynamicArray<ShaderUniform> uniforms;
};

inline int32 UniformLocation(ShaderProgram *program, const char *uniform) {
    int32 result = -1;

    for (int i = 0; i < program->uniforms.count; i++) {
        if (StringEquals(program->uniforms[i].name, uniform)) {
            result = program->uniforms[i].location;
        }
    }

    return result;
}


STRUCT()
struct GlyphInfo {
    vec2 lowerLeft;
    vec2 size;

    real32 advance;
};

ASSET({"etc_font"}, Serializable = false)
struct FontTable {
    uint32 glyphCount;

    GlyphInfo *glyphs;
    real32 *kernAdvances;

    real32 msdfRange;
    
    real32 ascent;
    real32 descent;
    real32 lineHeight;
    real32 unitsPerEM;

    // Maps from UCS to indices into glyph array
    uint32 ascii[128];
    HashTable<uint32, uint32> codepointMap;

    uint32 texcoordsMapID;
    vec4 *texcoordsMapData;
    
    Texture atlas;
};

inline uint32 GetInternalCodepoint(FontTable *font, uint32 unicode) {
    uint32 result = 0;
    if (unicode < 128) {
        result = font->ascii[unicode];
    }
    else if (!HashTableGet(&font->codepointMap, unicode, &result)) {
        ASSERT(false);
    }

    return result;
}

inline uint32 ConvertInternalCodepointToUCS(FontTable *font, uint32 codepoint) {
    uint32 result;
    for (int i = 0; i < font->codepointMap.capacity; i++) {
        if (font->codepointMap.occupied[i]) {
            if (font->codepointMap.values[i] == codepoint) {
                result = font->codepointMap.keys[i];
            }
        }
    }
    return result;
}

inline GlyphInfo GetGlyphInfoUCS(FontTable *font, uint32 unicode) {
    uint32 glyphIndex = GetInternalCodepoint(font, unicode);
    GlyphInfo result = font->glyphs[glyphIndex];

    return result;
}

inline GlyphInfo GetGlyphInfo(FontTable *font, uint32 codepoint) {
    ASSERT(codepoint < font->glyphCount);
    GlyphInfo result = font->glyphs[codepoint];

    return result;
}

// codepointA and codepointB are internal codepoints
inline real32 GetKernAdvance(FontTable *font, uint32 codepointA, uint32 codepointB) {
    ASSERT(codepointA < font->glyphCount);
    ASSERT(codepointB < font->glyphCount);

    uint32 index = codepointA * (font->glyphCount) + codepointB;
    real32 result = font->kernAdvances[index];

    return result;
}


// Sounds
ASSET({"wav"}, Serializable = false)
struct SoundClip {
    uint32 sampleRate;
    uint32 channels;
    uint32 sampleCount;

    // channels are interleaved
    real32 *data;
};

ASSET({"ogg"}, Serializable = false)
struct SoundStream {
    uint32 sampleRate;
    uint32 channels;
    uint32 sampleCount;

    // this is a bytestream that is processed by stb_vorbis (the contents of the file)
    uint8 *data;
    uint32 dataSize;
};

#endif

