
void AddCameraComponent(EntityDB *entityDB, CameraSystem *system, uint32 count, EntityID *entityIDs) {
    for (int i = 0; i < count; i++) {
        if (!Contains(&system->entityIDs, entityIDs[i])) {
            AddEntityToSystem(system, entityIDs[i]);
        
            CameraComponent *cam = PushBackPtr(&system->cameras);
            Plat.ClearMem(cam, sizeof(CameraComponent));

            // The first camera that gets added is set as the main camera
            if (system->mainCameraID == EntityID_Invalid) {
                system->mainCameraID = entityIDs[i];
            }
        }
    }
}

void RemoveCameraComponent(EntityDB *entityDB, CameraSystem *system, EntityID entityID) {
    uint32 index;
    if (Contains(&system->entityIDs, entityID, &index)) {
        RemoveAtIndexBySwap(&system->entityIDs, index);
        RemoveAtIndexBySwap(&system->cameras, index);
    }
}

b32 GetCameraComponentValue(EntityDB *entityDB, CameraSystem *system, EntityID entityID, CameraComponent *component) {
    b32 result = false;

    uint32 index;
    if (Contains(&system->entityIDs, entityID, &index)) {
        *component = system->cameras[index];
        result = true;
    }

    return result;
}

void SetCameraComponentValue(EntityDB *entityDB, CameraSystem *system, EntityID entityID, CameraComponent *component) {
    uint32 index;
    if (Contains(&system->entityIDs, entityID, &index)) {

        real32 aspect = Game.graphics->aspect;
        if (component->overrideAspect) {
            aspect = component->aspect;
        }

        switch (component->cameraType) {
            case CameraType_Perspective: {
                if (component->useFiniteFarClip) {
                    component->projection = Perspective(component->fov, aspect,
                                                        component->nearClip, component->farClip);
                }
                else {
                    component->projection = PerspectiveInfiniteFarPlane(component->fov, aspect,
                                                                        component->nearClip);
                }
            } break;
            case CameraType_Orthographic: {
                component->projection = Orthographic(component->width * -0.5f,
                                                     component->width * 0.5f,
                                                     component->height * -0.5f,
                                                     component->height * 0.5f,
                                                     component->nearClip,
                                                     component->farClip);
            } break;
        }

        system->cameras[index] = *component;
    }
}

mat4 GetViewProjection(CameraSystem *system, EntityID entityID) {
    mat4 result;
    uint32 index;
    if (Contains(&system->entityIDs, entityID, &index)) {
        result = system->cameras[index].viewProjection;
    }
    else {
        result = {};
    }

    return result;
}

CameraInfo GetCameraInfo(EntityDB *entityDB, EntityID id) {
    CameraInfo info = {};

    CameraComponent cameraComponent = GetComponentValue(entityDB, Camera, id);
    info.view = cameraComponent.view;
    info.projection = cameraComponent.projection;
    info.viewProjection = cameraComponent.projection * cameraComponent.view;

    info.worldTransform = GetWorldTransform(GetSystem(entityDB, Transform), id);

    info.cameraType = cameraComponent.cameraType;
    info.width = cameraComponent.width;
    info.height = cameraComponent.height;
    info.fov = cameraComponent.fov;

    if (cameraComponent.overrideAspect) {
        info.aspect = cameraComponent.aspect;
    }
    else {
        info.aspect = Game.graphics->aspect;
    }

    info.nearClip = cameraComponent.nearClip;
    info.farClip = cameraComponent.farClip;

    return info;
}

CameraInfo GetActiveCameraInfo() {
    if (SceneData *scene = GetActiveScene()) {
#if ETC_DEBUG
        if (scene->usingEditorCamera) {
            return GetEditorCameraInfo(scene->editorCamera);
        }
#endif

        CameraSystem *cameraSystem = GetSystem(&scene->entityDB, Camera);
        if (cameraSystem->mainCameraID != EntityID_Invalid) {
            return GetCameraInfo(&scene->entityDB, cameraSystem->mainCameraID);
        }
    }

    CameraInfo info = {};
    return info;
}

void SetMainCamera(CameraSystem *system, EntityID mainCameraID) {
    ASSERT(mainCameraID != EntityID_Invalid);

    for (int i = 0; i < system->cameras.count; i++) {
        if (system->entityIDs[i] == mainCameraID) {
            system->mainCameraID = mainCameraID;

            // @NOTE: do this here so that a call to SetMainCamera, no
            // matter if it's in an update or tick will tell the debug
            // system the appropriate camera to return to
#if ETC_DEBUG
            if (SceneData *activeScene = GetScene(Game.sceneManager->activeScene)) {
                if (activeScene->usingEditorCamera) {
                    EditorCamera *editorCamera = &activeScene->editorCamera;
                    if (system->mainCameraID != editorCamera->entityID) {
                        editorCamera->cameraIDToReturnTo = mainCameraID;
                    }

                    system->mainCameraID = editorCamera->entityID;
                }
            }
#endif

            break;
        }
    }
}

void TickCameraSystem(EtcGameMemory *mem, EntityDB *entityDB, CameraSystem *system) {
    CameraComponent *camera;
    for (int i = 0; i < system->cameras.count; i++) {
        camera = system->cameras.data + i;

        mat4 camWorld = GetWorldTransformMatrix(GetSystem(entityDB, Transform), system->entityIDs[i]);
        camera->view = OrthogonalInverse(camWorld);
        camera->viewProjection = camera->projection * camera->view;

        if (system->entityIDs[i] == system->mainCameraID) {
            CameraUniformBuffer *cameraUniformBuffer = &GetActiveScene()->cameraUniformBuffer;
            cameraUniformBuffer->view = camera->view;
            cameraUniformBuffer->projection = camera->projection;
            cameraUniformBuffer->viewProjection = camera->viewProjection;

            cameraUniformBuffer->invView = OrthogonalInverse(camera->view);
            Inverse(camera->projection, &cameraUniformBuffer->invProjection);

            cameraUniformBuffer->vFov = camera->fov;
            if (camera->overrideAspect) {
                cameraUniformBuffer->aspect = camera->aspect;
            }
            else {
                cameraUniformBuffer->aspect = Game.graphics->aspect;
            }
            cameraUniformBuffer->nearClip = camera->nearClip;
            cameraUniformBuffer->farClip = camera->farClip;
        }
    }
}

// @TODO: won't need this version too much longer
Ray MousePositionToWorldRay(CameraComponent camera, TransformComponent transform, vec2 mousePos) {
    Ray result = {};

    real32 aspect = Game.graphics->aspect;
    if (camera.overrideAspect) {
        aspect = camera.aspect;
    }

    real32 halfHeightAtOne = tanf(camera.fov * 0.5f);
    real32 halfWidthAtOne = halfHeightAtOne * aspect;

    real32 viewXAtOne = Lerp(halfWidthAtOne, -halfWidthAtOne, mousePos.x);
    real32 viewYAtOne = Lerp(-halfHeightAtOne, halfHeightAtOne, mousePos.y);
    vec3 viewRayDirection = Normalize(V3(viewXAtOne, viewYAtOne, 1.0f));

    result.origin = transform.position;
    result.direction = Rotate(transform.rotation, viewRayDirection);

    return result;
}

Ray MousePositionToWorldRay(CameraInfo cameraInfo, vec2 mousePos) {
    Ray result = {};

    switch (cameraInfo.cameraType) {
        case CameraType_Perspective: {
            // @TODO: this should really be doing an unprojection
            real32 halfHeightAtOne = tanf(cameraInfo.fov * 0.5f);
            real32 halfWidthAtOne = halfHeightAtOne * cameraInfo.aspect;

            real32 viewXAtOne = Lerp(halfWidthAtOne, -halfWidthAtOne, mousePos.x);
            real32 viewYAtOne = Lerp(-halfHeightAtOne, halfHeightAtOne, mousePos.y);
            vec3 viewRayDirection = Normalize(V3(viewXAtOne, viewYAtOne, 1.0f));

            result.origin = cameraInfo.worldTransform.position;
            result.direction = Rotate(cameraInfo.worldTransform.rotation, viewRayDirection);
        } break;

        case CameraType_Orthographic: {
            ASSERT(false);
        } break;
    }

    return result;
}

#if ETC_DEBUG
// EditorCamera
void InitEditorCamera(EditorCamera *camera, EntityDB *entityDB, vec3 position, quaternion rotation) {

    CameraSystem *camSys = GetSystem(entityDB, Camera);
    camera->cameraIDToReturnTo = camSys->mainCameraID;

    camera->entityDB = entityDB;
    camera->entityID = PushEntityID(entityDB);

    camera->accelerationScale = 2500.0f;

    AddComponent_(entityDB, ComponentType_Transform, 1, &camera->entityID);
    SetWorldTransform(GetSystem(camera->entityDB, Transform), camera->entityID, position, rotation, V3(1, 1, 1));

    CameraComponent cameraComponent = {};
    cameraComponent.cameraType = CameraType_Perspective;
    cameraComponent.nearClip = 0.1f;
    cameraComponent.fov = DegToRad(85.0f);
    cameraComponent.aspect = Game.graphics->aspect;

    AddComponent_(entityDB, ComponentType_Camera, 1, &camera->entityID);
    SetComponentValue(entityDB, Camera, camera->entityID, &cameraComponent);

    camera->speedMultiplier = 1.0f;
}

void ResetEditorCamera(EditorCamera *camera, vec3 position, quaternion rotation) {
    SetWorldTransform(GetSystem(camera->entityDB, Transform), camera->entityID, position, rotation, V3(1, 1, 1));

    camera->state = EditorCameraState_None;
    camera->linearVelocity = V3(0, 0, 0);
    camera->yawVelocity = 0.0f;
    camera->pitchVelocity = 0.0f;
}

void ResetEditorCameraVelocity(EditorCamera *camera) {
    camera->state = EditorCameraState_None;
    camera->linearVelocity = V3(0, 0, 0);
    camera->yawVelocity = 0.0f;
    camera->pitchVelocity = 0.0f;
}

b32 AttemptStartCameraSnap_(EditorCamera *camera, quaternion rotation) {
    b32 cameraSnapped = true;

    // -Y
    if (InputActionPressed(Game.inputActions, InputAction_EditorSnapNegY)) {
        real32 heading = 0.0f;
        real32 attitude = 0.0f;
        real32 bank = 0.0f;

        ToEulerAngles(rotation, &heading, &attitude, &bank);

        heading = NormalizeAngleRad(heading);

        real32 targetHeading = _PI;
        for (int i = 0; i <= 4; i++) {
            real32 potentialTargetHeading = 2.0f * _PI * (i / 4.0f);

            if (fabsf(heading - potentialTargetHeading) < fabsf(heading - targetHeading)) {
                targetHeading = potentialTargetHeading;
            }
        }


        quaternion q = AxisAngle(V3(0, 1, 0), targetHeading);

        if (Dot(q, rotation) < 0.0f) {
            q = -q;
        }

        camera->snapRotationTarget = q * AxisAngle(V3(1, 0, 0), _PI * 0.5f);
    }
    // +Y
    else if (InputActionPressed(Game.inputActions, InputAction_EditorSnapPosY)) {
        real32 heading = 0.0f;
        real32 attitude = 0.0f;
        real32 bank = 0.0f;

        ToEulerAngles(rotation, &heading, &attitude, &bank);

        heading = NormalizeAngleRad(heading);

        real32 targetHeading = _PI;
        for (int i = 0; i <= 4; i++) {
            real32 potentialTargetHeading = 2.0f * _PI * (i / 4.0f);

            if (fabsf(heading - potentialTargetHeading) < fabsf(heading - targetHeading)) {
                targetHeading = potentialTargetHeading;
            }
        }


        quaternion q = AxisAngle(V3(0, 1, 0), targetHeading);

        if (Dot(q, rotation) < 0.0f) {
            q = -q;
        }

        camera->snapRotationTarget = q * AxisAngle(V3(1, 0, 0), -_PI * 0.5f);
    }
    // -X
    else if (InputActionPressed(Game.inputActions, InputAction_EditorSnapNegX)) {
        camera->snapRotationTarget = AxisAngle(V3(0, 1, 0), -_PI * 0.5f);
    }
    // +X
    else if (InputActionPressed(Game.inputActions, InputAction_EditorSnapPosX)) {
        camera->snapRotationTarget = AxisAngle(V3(0, 1, 0), _PI * 0.5f);
    }
    // -Z
    else if (InputActionPressed(Game.inputActions, InputAction_EditorSnapNegZ)) {
        camera->snapRotationTarget = AxisAngle(V3(0, 1, 0), 0);
    }
    // +Z
    else if (InputActionPressed(Game.inputActions, InputAction_EditorSnapPosZ)) {
        camera->snapRotationTarget = AxisAngle(V3(0, 1, 0), _PI);
    }
    else {
        cameraSnapped = false;
    }

    if (cameraSnapped) {
        camera->snapRotationStartTime = Game.time;
        camera->snapRotationStart = rotation;

        camera->yawVelocity = 0.0f;
        camera->pitchVelocity = 0.0f;
    }

    return cameraSnapped;
}

void UpdateEditorCamera(EditorCamera *camera) {

    const real32 linearDamping = 0.8f;
    const real32 angularDamping = 0.8f;
    const real32 speedScale = 0.016f;

    const real32 scrollScale = 1.0f;

    const real32 turboBoost = 16.0f;
    const real32 antiTurboBoost = 0.25f;


    CameraSystem *cameraSystem = GetSystem(camera->entityDB, Camera);
    SetMainCamera(cameraSystem, camera->entityID);

    TransformSystem *transformSystem = GetSystem(camera->entityDB, Transform);
    Transform transform = GetWorldTransform(transformSystem, camera->entityID);

    CameraInfo cameraInfo = GetEditorCameraInfo(*camera);

    switch (camera->state) {
        case EditorCameraState_None: {
            if (AttemptStartCameraSnap_(camera, transform.rotation)) {
                camera->state = EditorCameraState_SnappingRotation;
            }
            else if (InputActionPressed(Game.inputActions, InputAction_EditorRotateCamera)) {

                camera->state = EditorCameraState_DraggingRotation;
                camera->yawVelocity = 0.0f;
                camera->pitchVelocity = 0.0f;

                camera->originalDragRotation = transform.rotation;
                camera->originalMousePosNorm = InputAnalogue2D(Game.mouse, Etc_MousePositionXYNorm, true);
            }
            else {
                real32 yawAcceleration = 0.0f;
                real32 pitchAcceleration = 0.0f;

                yawAcceleration = 25.0f * InputActionAnalogue(Game.inputActions, InputAction_EditorRotateYaw);
                pitchAcceleration = 25.0f * InputActionAnalogue(Game.inputActions, InputAction_EditorRotatePitch);

                yawAcceleration *= Game.mem->userPreferences.cameraSpeed;
                pitchAcceleration *= Game.mem->userPreferences.cameraSpeed;


                camera->yawVelocity = camera->yawVelocity + yawAcceleration * speedScale;
                camera->yawVelocity *= angularDamping;

                camera->pitchVelocity = camera->pitchVelocity + pitchAcceleration * speedScale;
                camera->pitchVelocity *= angularDamping;

                transform.rotation = AxisAngle(V3(0, 1, 0), camera->yawVelocity * speedScale) * transform.rotation;
                vec3 right = Rotate(transform.rotation, V3(-1, 0, 0));
                transform.rotation = AxisAngle(right, camera->pitchVelocity * speedScale) * transform.rotation;
            }
        } break;

        case EditorCameraState_DraggingRotation: {
            if (InputActionHold(Game.inputActions, InputAction_EditorRotateCamera)) {
                // @TODO: screenwrap mouse
                // @TODO: lock the mouse to whatever the original position is and calculate a virtual position
                // so the mouse isnt going all over the place and leaving the screen
                vec2 originalMousePosNorm = camera->originalMousePosNorm;
                vec2 mousePosNorm = InputAnalogue2D(Game.mouse, Etc_MousePositionXYNorm, true);

                real32 halfHeightAtOne = tanf(cameraInfo.fov * 0.5f);
                real32 halfWidthAtOne = halfHeightAtOne * cameraInfo.aspect;

                real32 originalYaw = atanf(Lerp(-halfWidthAtOne, halfWidthAtOne, originalMousePosNorm.x));
                real32 originalPitch = atanf(Lerp(-halfHeightAtOne, halfHeightAtOne, originalMousePosNorm.y));

                real32 yaw = atanf(Lerp(-halfWidthAtOne, halfWidthAtOne, mousePosNorm.x));
                real32 pitch = atanf(Lerp(-halfHeightAtOne, halfHeightAtOne, mousePosNorm.y));

                real32 yawDiff = yaw - originalYaw;
                real32 pitchDiff = originalPitch - pitch;

                transform.rotation = AxisAngle(V3(0, 1, 0), yawDiff) * camera->originalDragRotation;
                vec3 right = Rotate(transform.rotation, V3(-1, 0, 0));
                transform.rotation = AxisAngle(right, pitchDiff) * transform.rotation;
            }
            else {
                camera->state = EditorCameraState_None;
            }
        } break;

        case EditorCameraState_SnappingRotation: {
            static real32 snapRotationDuration = 0.25f;

            real32 t = (Game.time - camera->snapRotationStartTime) / snapRotationDuration;
            if (t >= 1.0f) {
                camera->state = EditorCameraState_None;
            }

            t = Clamp(t, 0.0f, 1.0f);
            transform.rotation = Lerp(camera->snapRotationStart, camera->snapRotationTarget, t);
        } break;
    }


    real32 changeDir = InputActionAnalogue(Game.inputActions, InputAction_EditorChangeCameraAcceleration);
    if (changeDir != 0) {
        camera->accelerationScale += 500.0f * changeDir;
    }

    real32 forwardAcceleration = camera->accelerationScale * InputActionAnalogue(Game.inputActions, InputAction_EditorMoveZ);
    real32 strafeAcceleration = camera->accelerationScale * InputActionAnalogue(Game.inputActions, InputAction_EditorMoveX);
    real32 upAcceleration = camera->accelerationScale * InputActionAnalogue(Game.inputActions, InputAction_EditorMoveY);

    real32 speedMultiplier = camera->speedMultiplier;
    if (InputActionHold(Game.inputActions, InputAction_EditorBoost, true)) {
        speedMultiplier *= 10;
    }
    if (InputActionHold(Game.inputActions, InputAction_EditorAntiBoost, true)) {
        speedMultiplier *= 0.1f;
    }

    upAcceleration *= speedMultiplier;
    forwardAcceleration *= speedMultiplier;
    strafeAcceleration *= speedMultiplier;

    vec3 right = Rotate(transform.rotation, V3(-1, 0, 0));
    vec3 up = Rotate(transform.rotation, V3(0, 1, 0));
    vec3 forward = Rotate(transform.rotation, V3(0, 0, 1));


    vec3 linearAcceleration = (right * strafeAcceleration +
                               up * upAcceleration +
                               forward * forwardAcceleration) * speedScale;
    
    camera->linearVelocity = camera->linearVelocity + linearAcceleration * speedScale;
    camera->linearVelocity = camera->linearVelocity * linearDamping;

    transform.position = transform.position + camera->linearVelocity * speedScale;

    real32 scroll = scrollScale * InputActionAnalogue(Game.inputActions, InputAction_EditorMoveZScroll);
    transform.position = transform.position + forward * scroll;


    SetWorldTransform(transformSystem, camera->entityID, transform);


    CameraComponent cameraComponent = GetComponentValue(camera->entityDB, Camera, camera->entityID);
    cameraComponent.view = OrthogonalInverse(ToMat4(transform));
    cameraComponent.viewProjection = cameraComponent.projection * cameraComponent.view;
    SetComponentValue(camera->entityDB, Camera, camera->entityID, &cameraComponent);
}

void GetEditorCameraTransforms(EditorCamera *camera, mat4 *view, mat4 *projection) {
    CameraComponent cameraComponent = GetComponentValue(camera->entityDB, Camera, camera->entityID);
    *view = cameraComponent.view;
    *projection = cameraComponent.projection;
}

Transform GetEditorCameraWorldTransform(EditorCamera camera) {
    Transform transform = GetWorldTransform(GetSystem(camera.entityDB, Transform), camera.entityID);
    return transform;
}

CameraInfo GetEditorCameraInfo(EditorCamera camera) {
    return GetCameraInfo(camera.entityDB, camera.entityID);
}

#endif

