 
#ifndef _ETCETERA_CHUNKED_ARRAY_H
#define _ETCETERA_CHUNKED_ARRAY_H

// @TODO: serialize! Not sure what the best way to handle allocators is tho...
// We can serialize the type, and then pass in allocators to the DeserializeSchemaValue func
// We could have allocator names/ids which get serialized along with the type and can be accessed
// globally so we make sure we deserialize into the correct allocator
// Might not need to be that crazy tho, heap is obv, arena can belong to the scene (but maybe
// it would be on the system sometimes...)
// Name seems better to me than ID, but there's still the issue of having access to them
// so we'd need to keep a map of allocators going from name->pointer

struct ArrayChunk {
    uint8 data[0];
};

struct ChunkedArray_Untyped {
    EtcAllocator *allocator;

    uint32 count;

    // @NOTE: this does not include any extra data stored in ArrayChunk
    uint32 chunkSize;

    // @GACK: we have to heap allocate the pointers!
    DynamicArray<ArrayChunk *> chunks;
};

template <typename T>
struct ChunkedArray {
    EtcAllocator *allocator;

    uint32 count;

    // @NOTE: this does not include any extra data stored in ArrayChunk
    uint32 chunkSize;

    // @GACK: we have to heap allocate the pointers!
    DynamicArray<ArrayChunk *> chunks;

    inline T& operator[](const int index) const {
        // (index / chunkCount) to get the chunk
        // then index % chunkCount to get the index
        return *(T *)((uint8 *)chunks[index / chunkSize]->data + (sizeof(T) * (index % chunkSize)));
    }
};

template <typename T>
void ChunkedArrayAllocateChunk(ChunkedArray<T> *array) {
    ArrayChunk *newChunk = (ArrayChunk *)Allocate_(array->allocator, sizeof(ArrayChunk) + (array->chunkSize * sizeof(T)));
    PushBack(&array->chunks, newChunk);
}

template <typename T>
void ChunkedArrayDeallocateChunk(ChunkedArray<T> *array) {
    ArrayChunk *newChunk = Deallocate_(array->allocator);
}

template <typename T>
void ChunkedArrayEnsureCapacity(ChunkedArray<T> *array, uint32 capacity) {

    ASSERT(array->allocator != NULL);
    ASSERT(array->chunkSize > 0);
    
    if (array->chunks.count * array->chunkSize < capacity) {
        uint32 chunksToAdd = ((capacity / array->chunkSize) - array->chunks.count);
        
        if (array->chunks.count > 0) {
            chunksToAdd++;
        }

        for (int i = 0; i < chunksToAdd; i++) {
            ChunkedArrayAllocateChunk(array);
        }
    }
}

template <typename T>
inline ChunkedArray<T> MakeChunkedArray(EtcAllocator *allocator, uint32 chunkSize, uint32 chunkCount = 1) {
    ChunkedArray<T> array = {};
    array.allocator = allocator;
    array.chunkSize = chunkSize;
    ChunkedArrayEnsureCapacity(&array, chunkCount * chunkSize);
    return array;
}

template <typename T>
inline void DeallocateChunkedArray(ChunkedArray<T> *array) {
    for (int i = 0; i < array->chunks.count; i++) {
        Deallocate_(array->allocator, array->chunks[i]);
    }

    DeallocateDynamicArray(&array->chunks);
}

template <typename T>
inline uint32 PushBack(ChunkedArray<T> *array, T elem)  {
    ChunkedArrayEnsureCapacity(array, array->count + 1);
    uint32 index = array->count;
    array->count++;
    (*array)[index] = elem;
    return index;
}

template <typename T>
inline T *PushBackPtr(ChunkedArray<T> *array)  {
    ChunkedArrayEnsureCapacity(array, array->count + 1);
    uint32 index = array->count;
    array->count++;
    T *result = &(*array)[index];
    Plat.ClearMem(result, sizeof(T));
    return result;
}

template <typename T>
inline void PushBackUnique(ChunkedArray<T> *array, T elem) {
    if (!Contains(array, elem)) {
        PushBack(array, elem);
    }
}

template <typename T>
inline b32 PopBack(ChunkedArray<T> *array, T *element = NULL) {
    b32 result = false;

    if (array->count > 0) {
        if (element != NULL) {
            int32 index = array->count - 1;
            *element = (*array)[index];
        }
        array->count--;
        result = true;
    }

    return result;
}

template <typename T>
inline void ChunkedArrayClear(ChunkedArray<T> *array) {
    array->count = 0;
}

template <typename T>
inline void ChunkedArrayClearToCount(ChunkedArray<T> *array, uint32 count) {
    ChunkedArrayEnsureCapacity(array, count);

    for (int i = 0; i < array->chunks.count; i++) {
        Plat.ClearMem(array->chunks[i], array->chunkSize * sizeof(T));
    }
    
    array->count = count;
};

template <typename T>
inline void ExtendChunkedArray(ChunkedArray<T> *array, ChunkedArray<T> from) {
    for (int i = 0; i < from.count; i++) {
        PushBack(array, from[i]);
    }
}

template <typename T>
inline void RemoveAtIndex(ChunkedArray<T> *array, uint32 index) {
    ASSERT(index < array->count);
    ASSERT(array->chunkSize != 0);
    ASSERT(array->chunks.count > 0);

    uint32 lastChunkIndex = Min(array->count / array->chunkSize, array->chunks.count - 1);

    {
        uint32 chunkIndex = index / array->chunkSize;
        uint32 indexInChunk = index % array->chunkSize;
        uint32 elementsInChunk = chunkIndex < lastChunkIndex ?
            array->chunkSize :
            (array->count - (array->chunkSize * lastChunkIndex));

        uint32 elementsToMoveInChunk = elementsInChunk - indexInChunk - 1;
        if (elementsToMoveInChunk > 0) {
            ArrayChunk *chunk = array->chunks[chunkIndex];
            Plat.MoveMem(chunk->data + sizeof(T) * indexInChunk,
                         chunk->data + sizeof(T) * (indexInChunk + 1),
                         sizeof(T) * elementsToMoveInChunk);
        }
    }

    // At this point there is a one element "hole" at the end of the
    // first chunk. Fill it in from the first value in the next chunk,
    // then shift the elements in next chunk, which creates a hole at
    // the end of the next chunk, and repeat...
    for (uint32 chunkIndex = (index / array->chunkSize) + 1; chunkIndex <= lastChunkIndex; chunkIndex++) {
        ArrayChunk *prevChunk = array->chunks[chunkIndex - 1];
        ArrayChunk *chunk = array->chunks[chunkIndex];

        uint32 elementsInChunk = chunkIndex < lastChunkIndex ?
            array->chunkSize :
            (array->count % array->chunkSize);

        if (elementsInChunk == 0) {
            break;
        }

        Plat.CopyMem(prevChunk->data + sizeof(T) * (array->chunkSize - 1), chunk->data, sizeof(T));
        //prevChunk->data[array->chunkSize - 1] = chunk->data[0];
        Plat.MoveMem(chunk->data, chunk->data + sizeof(T), sizeof(T) * (elementsInChunk - 1));
    }

    array->count--;
}

template <typename T>
inline void InsertAtIndex(ChunkedArray<T> *array, uint32 index, T element) {
    ASSERT(index < array->count);
    ASSERT(array->chunkSize != 0);
    ASSERT(array->chunks.count > 0);

    uint32 lastChunkIndex = Min(array->count / array->chunkSize, array->chunks.count - 1);

    ChunkedArrayEnsureCapacity(array, array->count + 1);

    T lastElem = (*array)[array->count - 1];

    array->count++;
    for (int i = array->count - 1; i > index; i--) {
        (*array)[i] = (*array)[i - 1];
    }

    (*array)[index] = element;
}


template <typename T>
inline void RemoveAtIndexBySwap(ChunkedArray<T> *array, uint32 index) {
    ASSERT(index < array->count);

    if (index != array->count - 1) {
        (*array)[index] = (*array)[array->count - 1];
    }

    array->count--;
}

template <typename T>
inline b32 Contains(ChunkedArray<T> *array, T element, uint32 *index = NULL) {
    b32 result = false;

    for (int i = 0; i < array->count; i++) {
        if ((*array)[i] == element) {
            if (index != NULL) { *index = i; }
            result = true;
            break;
        }
    }

    return result;
}

template <typename T>
inline T Last(ChunkedArray<T> *array) {
    ASSERT(array->count > 0);

    T result = (*array)[array->count - 1];
    return result;
}

template <typename T>
inline T *LastPtr(ChunkedArray<T> *array) {
    ASSERT(array->count > 0);

    T *result = &(*array)[array->count - 1];
    return result;
}

template <typename T>
inline void RemoveLast(ChunkedArray<T> *array) {
    ASSERT(array->count > 0);
    RemoveAtIndex(array, array->count - 1);
}

template <typename T>
int32 ChunkedArrayQuicksortPartition_(ChunkedArray<T> *array, int32 low, int32 high, int32 (* comparator)(T a, T b)) {
    T pivot = (*array)[low];

    int32 i = low;
    int32 j = high;

    while (true) {        
        while (comparator((*array)[i], pivot) < 0) {
            i++;
        }

        while (comparator((*array)[j], pivot) > 0) {
            j--;            
        }

        if (i >= j) {
            return j;
        }

        if (i != j) {
            T temp = (*array)[i];
            (*array)[i] = (*array)[j];
            (*array)[j] = temp;
        }

        i++;
        j--;
    }
}

template <typename T>
void Quicksort_(ChunkedArray<T> *array, int32 low, int32 high, int32 (* comparator)(T a, T b)) {
    if (low >= high) {
        return;
    }
    int32 mid = ChunkedArrayQuicksortPartition_(array, low, high, comparator);
    Quicksort_(array, low, mid, comparator);
    Quicksort_(array, mid + 1, high, comparator);
}

template <typename T>
void Quicksort(ChunkedArray<T> *array, int32 (* comparator)(T a, T b)) {
    Quicksort_(array, 0, array->count - 1, comparator);
}

#endif

