#ifndef _ETC_COMMONLIB_H
#define _ETC_COMMONLIB_H

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdio.h>

#if _WIN32
#include <wtypes.h>
#include <windef.h>
#include <intrin.h>

// @TODO: some of these don't seem to work
#define EtcCompareAndSwap(dest, comparand, exchange) InterlockedCompareExchange((LONG volatile *)dest, (LONG)exchange, (LONG)comparand)
#define EtcCompareAndSwap64(dest, comparand, exchange) InterlockedCompareExchange64((LONG64 volatile *)dest, (LONG64)exchange, (LONG64)comparand)
#define EtcAtomicAdd(value, addend) _InterlockedExchangeAdd64((LONGLONG volatile *)value, (LONGLONG)addend)
#define EtcAtomicExchange(ptr, value) InterlockedExchange(ptr, value)
#define EtcAtomicExchange64(ptr, value) InterlockedExchange64((LONG64 *)ptr, (LONG64)value)
#define EtcThreadYield() SwitchToThread()

#define EtcCycleCount() __rdtsc()

#elif __APPLE__
#include <sched.h>
#include <libkern/OSAtomic.h>

#define EtcCompareAndSwap(dest, comparand, exchange) __sync_val_compare_and_swap(dest, comparand, exchange)
#define EtcAtomicAdd(value, addend) __sync_fetch_and_add(value, addend)
#define EtcAtomicExchange(ptr, value) __sync_lock_test_and_set(ptr, value)
#define EtcThreadYield() sched_yield()

inline uint64 EtcCycleCount() {
    uint32 eax = 0;
    uint32 edx;

    __asm__ __volatile__("cpuid;"
                         "rdtsc;"
                         : "+a" (eax), "=d" (edx)
                         :
                         : "%rcx", "%rbx", "memory");

    __asm__ __volatile__("xorl %%eax, %%eax;"
                         "cpuid;"
                         :
                         :
                         : "%rax", "%rbx", "%rcx", "%rdx", "memory");

    return (((uint64)edx << 32) | eax);
}

#elif __linux__
#include <sched.h>
#include <x86intrin.h>
#define EtcCompareAndSwap(dest, comparand, exchange) __sync_val_compare_and_swap(dest, comparand, exchange)
#define EtcCompareAndSwap64(dest, comparand, exchange) __sync_val_compare_and_swap(dest, comparand, exchange)
#define EtcAtomicAdd(value, addend) __sync_fetch_and_add(value, addend)
#define EtcAtomicExchange(ptr, value) __sync_lock_test_and_set(ptr, value)
#define EtcAtomicExchange64(ptr, value) __sync_lock_test_and_set(ptr, value)
#define EtcThreadYield() sched_yield()
#define EtcCycleCount() _rdtsc()

#endif

#if _WIN32
#undef FIELD_OFFSET
#endif

// @SOURCE: https://software.intel.com/en-us/forums/intel-threading-building-blocks/topic/287925
// @NOTE: for my dumbass, this works because we the int/real pair share the same memory, so when we add
// to the newReal, that changes newInt also, and when we successfully swap the old and newInt, we just return
// the newReal
real64 EtcAtomicAddReal64(real64 *value, real64 addend) {
    union {
        volatile int64 oldInt;
        real64 oldReal;
    };

    union {
        volatile int64 newInt;
        real64 newReal;
    };
    
    while (true) {
        oldInt = *((int64 *)value);
        newReal = oldReal + addend;

        if (EtcCompareAndSwap64((int64 *)value, oldInt, newInt) == oldInt) {
            return newReal;
        }
    }

    return *value;
}


#define FIELD_OFFSET(type, field) ((uint64)&(((type *)NULL)->field))

#define POINTER_OFFSET(a, b) ((int32)((uint8 *)(b) - (uint8 *)(a)))
#define BUFFER_OFFSET(buffer, offset) ((uint8 *)(buffer) + (offset))
#define STRIDED_POINTER(type, buffer, stride, offset) (type *)((uint8 *)(buffer) + (stride) * (offset))
#define STRIDED_READ(type, buffer, stride, offset) (*(type *)((uint8 *)(buffer) + (stride) * (offset)))

#define ARRAY_LENGTH(type, array) (sizeof(array) / sizeof(type))



#if ETC_DEBUG
#include <assert.h>
#define ASSERT(...) assert(__VA_ARGS__)

#else
#define ASSERT(...)
#endif

#define HasFlag(bitfield, flag) ((bitfield) & (flag))
#define HasFlags(bitfield, flag) ((flag) != 0 && ((bitfield) & (flag)) == (flag))
#define AddFlag(bitfield, flag) ((bitfield) | (flag))
#define ClearFlag(bitfield, flag) ((bitfield) & ~(flag))


#define SetBit_(bitfield, nth) (((bitfield)[nth / 32]) |= (1 << (nth % 32)))
#define ClearBit_(bitfield, nth) (((bitfield)[nth / 32]) &= ~(1 << (nth % 32)))
#define TestBit_(bitfield, nth) (((bitfield)[nth / 32]) & (1 << (nth % 32)))

// @TODO: these should take a schema/name and calculate the chunkCount for you
inline void SetBits(u32 *bitsA, u32 *bitsB, u32 chunkCount) {
    for (int i = 0; i < chunkCount; i++) {
        bitsA[i] |= bitsB[i];
    }
}

inline void ClearBits(u32 *bitsA, u32 *bitsB, u32 chunkCount) {
    for (int i = 0; i < chunkCount; i++) {
        bitsA[i] &= ~bitsB[i];
    }
}

inline bool TestBits(u32 *bits, u32 *indices, u32 indexCount) {
    bool result = true;
    for (int i = 0; i < indexCount; i++) {
        result &= (TestBit_(bits, indices[i]) != 0);
    }

    return result;
}

// @NOTE: makes sure that all set bits in B are set in A
inline bool TestAllBits(u32 *bitsA, u32 *bitsB, u32 chunkCount) {
    bool result = true;
    for (int i = 0; i < chunkCount * 32; i++) {
        if (TestBit_(bitsB, i)) {
            result &= (TestBit_(bitsA, i) != 0);
        }
    }

    return result;
}


inline void *AllocateClearedMemory(uint64 size) {
    void *result = Plat.AllocateMem(size);
    Plat.ClearMem(result, size);

    return result;
}



#define ALLOCATE_STRUCT(type) (type *)Plat.AllocateMem(sizeof(type))
#define ALLOCATE_CLEARED_STRUCT(type) (type *)AllocateClearedMemory(sizeof(type))

#define ALLOCATE_ARRAY(type, length) (type *)Plat.AllocateMem(sizeof(type) * (length))
#define ALLOCATE_CLEARED_ARRAY(type, length) (type *)AllocateClearedMemory(sizeof(type) * (length))
#define RESIZE_ARRAY(array, type, length) (type *)Plat.ResizeMem(array, sizeof(type) * (length))

struct MemoryArena;
struct MemoryArenaList;
inline void *PushSize(MemoryArenaList *arenaList, uint64 size);

inline b32 IsSpace(uint32 codepoint) {
    return (codepoint == 32);
}

inline b32 IsNewline(uint32 codepoint) {
    return (codepoint == 10 || codepoint == 13);
}

inline b32 IsTab(uint32 codepoint) {
    return (codepoint == 9);
}

inline b32 IsBackspace(uint32 codepoint) {
    return (codepoint == 8);
}

inline b32 IsWhitespace(uint32 codepoint) {
    return IsSpace(codepoint) || IsNewline(codepoint);
}

inline b32 IsControlCode(uint32 codepoint) {
    return codepoint <= 32 || (codepoint >= 127 && codepoint <= 159);
}

inline b32 IsPrintable(uint32 codepoint) {
    return !(codepoint < 32 || (codepoint >= 128 && codepoint <= 159));
}

inline b32 IsDigit(uint32 codepoint) {
    return codepoint >= 48 && codepoint <= 57;
}

inline b32 IsNumeric(uint32 codepoint) {
    return IsDigit(codepoint) || (codepoint >= 45 && codepoint <= 46);
}

inline b32 IsUppercaseASCII(uint32 codepoint) {
    return (codepoint >= 'A') && (codepoint <= 'Z');
}

inline b32 IsLowercaseASCII(uint32 codepoint) {
    return (codepoint >= 'a') && (codepoint <= 'z');
}

inline b32 IsAlphaASCII(uint32 codepoint) {
    return (((codepoint >= 'a') && (codepoint <= 'z')) ||
            ((codepoint >= 'A') && (codepoint <= 'Z')));
}

inline uint32 UppercaseASCII(uint32 codepoint) {
    if (IsLowercaseASCII(codepoint)) {
        return codepoint - 32;
    }

    return codepoint;
}

inline uint32 LowercaseASCII(uint32 codepoint) {
    if (IsUppercaseASCII(codepoint)) {
        return codepoint + 32;
    }

    return codepoint;
}

inline b32 ValidDecimalInput(uint32 codepoint) {
    return IsBackspace(codepoint) || IsNumeric(codepoint);
}

inline b32 ValidAlphaNumericInput(uint32 codepoint) {
    return IsAlphaASCII(codepoint) || IsBackspace(codepoint) || IsDigit(codepoint);
}

inline b32 ValidAlphaNumericLowercaseInput(uint32 codepoint) {
    return (!IsUppercaseASCII(codepoint) && IsAlphaASCII(codepoint)) || IsBackspace(codepoint) || IsDigit(codepoint);
}

inline uint32 StringLength(const char *str) {
    return strlen(str);
}

inline uint32 StringLength(const char *str, uint32 maxLength) {
    uint32 length = 0;

    while (length < maxLength) {
        if (str[length] == '\0') {
            break;
        }

        length++;
    }

    return length;
}

inline const char *StringCopy(const char *str, uint32 length) {
    ASSERT(str != NULL);
    char *copied = (char *)Plat.AllocateMem(sizeof(char) * (length + 1));
    Plat.CopyMem(copied, str, length);
    copied[length] = 0;

    return copied;
}

inline char *StringCopy_(const char *str, uint32 length) {
    ASSERT(str != NULL);
    char *copied = (char *)Plat.AllocateMem(sizeof(char) * (length + 1));
    Plat.CopyMem(copied, str, length);
    copied[length] = 0;

    return copied;
}


inline const char *StringCopy(const char *str) {
    ASSERT(str != NULL);

    const char *copied = StringCopy(str, strlen(str));
    return copied;
}

inline uint32 StringCopy(char *dest, uint32 destLength, const char *src) {
    ASSERT(dest != NULL);
    ASSERT(src != NULL);

    uint32 charsCopied = 0;
    while (src[charsCopied] != 0) {
        if (charsCopied >= (destLength - 1)) {
            break;
        }

        dest[charsCopied] = src[charsCopied];
        charsCopied++;
    }

    dest[charsCopied] = 0;

    return charsCopied;
}

inline char *StringCopy(MemoryArenaList *arena, const char *str, uint32 length) {
    ASSERT(str != NULL);
    char *copied = (char *)PushSize(arena, sizeof(char) * (length + 1));
    Plat.CopyMem(copied, str, length);
    copied[length] = 0;

    return copied;
}

inline char *StringCopy(MemoryArenaList *arena, const char *str) {
    return StringCopy(arena, str, StringLength(str));
}

char *String(const char *str, uint32 length) {
    return (char *)StringCopy(str, length);
}

char *String(const char *str) {
    return (char *)StringCopy(str);
}

char *String(MemoryArenaList *arena, const char *str) {
    return StringCopy(arena, str);
}

const char *StringConcat(const char *str1, const char *str2) {
    uint32 length1 = strlen(str1);
    uint32 length2 = strlen(str2);
    char *result = (char *)Plat.AllocateMem(length1 + length2 + 1);

    Plat.CopyMem(result, str1, length1);
    Plat.CopyMem(result + length1, str2, length2 + 1);
    result[length1 + length2] = 0;

    return result;
}

const char *StringConcat(const char *str1, uint32 str1Length, const char *str2) {
    uint32 length1 = str1Length;
    uint32 length2 = strlen(str2);
    char *result = (char *)Plat.AllocateMem(length1 + length2 + 1);

    Plat.CopyMem(result, str1, length1);
    Plat.CopyMem(result + length1, str2, length2 + 1);
    result[length1 + length2] = 0;

    return result;
}

const char *StringConcat(const char *str1, uint32 str1Length, const char *str2, uint32 str2Length) {
    uint32 length1 = str1Length;
    uint32 length2 = str2Length;
    char *result = (char *)Plat.AllocateMem(length1 + length2 + 1);

    Plat.CopyMem(result, str1, length1);
    Plat.CopyMem(result + length1, str2, length2 + 1);
    result[length1 + length2] = 0;

    return result;
}


char *StringConcat(const char *str1, const char *str2, MemoryArenaList *arena) {
    uint32 length1 = strlen(str1);
    uint32 length2 = strlen(str2);
    char *result = (char *)PushSize(arena, length1 + length2 + 1);

    Plat.CopyMem(result, str1, length1);
    Plat.CopyMem(result + length1, str2, length2 + 1);
    result[length1 + length2] = 0;

    return result;
}


char *StringConcat(const char *str1, const char *str2, uint32 str2Length, MemoryArenaList *arena) {
    uint32 length1 = strlen(str1);
    uint32 length2 = str2Length;
    char *result = (char *)PushSize(arena, length1 + length2 + 1);

    Plat.CopyMem(result, str1, length1);
    Plat.CopyMem(result + length1, str2, length2 + 1);
    result[length1 + length2] = 0;

    return result;
}

char *StringConcat(MemoryArenaList *arena, uint32 count, ...) {
    ASSERT(count >= 2);
    va_list args;
    va_start(args, count);

    char *result = NULL;
    char *str1 = va_arg(args, char *);
    char *str2 = va_arg(args, char *);
    result = StringConcat(str1, str2, arena);
    
    for (int i = 2; i < count; i++) {
        char *str = va_arg(args, char *);
        result = StringConcat(result, str, arena);
    }

    va_end(args);

    return result;
}

char *StringToUpper(const char *a) {
    uint32 len = StringLength(a);
    char *result = (char *)Plat.AllocateMem(len);
    
    for (int i = 0; i < len; i++) {
        char c = a[i];
        if (IsAlphaASCII(c) && !IsUppercaseASCII(c)) {
            c = UppercaseASCII(a[i]);
        }
        result[i] = c;
    }
    return result;
}

char *StringToLower(const char *a) {
    uint32 len = StringLength(a);
    char *result = (char *)Plat.AllocateMem(len);
    
    for (int i = 0; i < len; i++) {
        char c = a[i];
        if (IsAlphaASCII(c) && !IsLowercaseASCII(c)) {
            c = LowercaseASCII(a[i]);
        }
        result[i] = c;
    }
    return result;
}

inline bool StringEquals(const char *a, const char *b) {
    ASSERT(a != NULL);
    ASSERT(b != NULL);

    b32 result = strcmp(a, b) == 0;
    return result;
}

inline bool StringEquals(void *a, void *b) {
    return StringEquals(*(const char **)a, (const char *)b);
}

// @WARNING: this will return true if b contains a with its first length characters
inline bool StringEquals(const char *a, const char *b, uint32 length) {
    ASSERT(a != NULL);
    ASSERT(b != NULL);

    b32 result = memcmp(a, b, length) == 0;
    return result;
}

inline bool StringEquals(const char *a, const char *b, uint32 lengthA, uint32 lengthB) {
    ASSERT(a != NULL);
    ASSERT(b != NULL);

    if (lengthA != lengthB) {
        return false;
    }

    b32 result = memcmp(a, b, lengthA) == 0;
    return result;
}

inline bool StringEquals(char *a, char *b) {
    ASSERT(a != NULL);
    ASSERT(b != NULL);

    b32 result = strcmp(a, b) == 0;
    return result;
}

inline bool StringEqualsCaseInsensitive(char *a, char *b) {
    char *aLower = StringToLower(a);
    char *bLower = StringToLower(b);
    bool result = strcmp(aLower, bLower) == 0;
    return result;
}

inline bool StringEmpty(const char *a) {
    ASSERT(a != NULL);
    return a[0] == 0;
}

inline const char UnescapeCharacter(char escapedCharacter) {
    char result;

    switch (escapedCharacter) {
        case 'r': { result = '\r'; } break;
        case 'n': { result = '\n'; } break;
        case 't': { result = '\t'; } break;
        case '0': { result = '\0'; } break;

        default: {
            result = escapedCharacter;
        } break;
    }

    return result;
}

// @TODO: octal/hex literals? arbitrary unicode values?
inline const char *UnescapeString(const char *a, uint32 length) {
    char *unescaped = ALLOCATE_ARRAY(char, length + 1);
    uint32 unescapedLength = 0;
    uint32 escapedCount = 0;

    for (int i = 0; i < length; i++) {
        char c = a[i];

        if (c == '\0') {
            break;
        }
        else if (c == '\\') {
            if (i + 1 >= length) {
                break;
            }
            else {
                unescaped[i - escapedCount] = UnescapeCharacter(a[i + 1]);
                escapedCount++;
                i++;
            }
        }
        else {
            unescaped[i - escapedCount] = c;
            unescapedLength++;
        }
    }

    unescaped[unescapedLength] = '\0';

    return unescaped;
}

inline const char *EscapeString(const char *a, uint32 length) {
    char *result = NULL;

    // @TODO: ...

    return result;
}

inline void PrettyPrintFloat(char *buffer, uint32 bufferLength, real32 value) {
    stbsp_snprintf(buffer, bufferLength, "%f", value);

    int32 decimalPointIndex = -1;
    int32 trailingZeroBeginIndex = -1;

    for (int i = 0; i < bufferLength; i++) {
        char c = buffer[i];

        if (c == '\0') {
            break;
        }
        else if (c == '.') {
            decimalPointIndex = i;
        }
        // If there's an exponent, assume the rest of the buffer is valid
        else if (c == 'e' || c == 'E') {
            break;
        }
        // The first zero after the decimal point is valid
        else if (decimalPointIndex >= 0 && i > decimalPointIndex + 1) {
            if (c == '0') {
                if (trailingZeroBeginIndex < 0) {
                    trailingZeroBeginIndex = i;
                }
            }
            else {
                trailingZeroBeginIndex = -1;
            }
        }
    }

    if (trailingZeroBeginIndex >= 0) {
        buffer[trailingZeroBeginIndex] = 0;
    }
}

inline b32 FindString(void *list, uint32 listCount, const char *str, uint32 strLength, uint32 *index, uint32 stride = 0) {
    b32 result = false;

    if (stride == 0) {
        stride = sizeof(const char *);
    }

    for (int i = 0; i < listCount; i++) {
        const char *element = *(const char **)((uint8 *)list + i * stride);

        if (StringEquals(element, str, strLength) && element[strLength] == '\0') {
            if (index != NULL) {
                *index = i;
            }
            result = true;
            break;
        }
    }

    return result;
}

inline b32 FindString(void *list, uint32 listCount, const char *str, uint32 *index, uint32 stride = 0) {
    b32 result = FindString(list, listCount, str, strlen(str), index, stride);

    return result;
}

inline int32 FindSubstring(const char *string, uint32 stringLength, const char *substring, uint32 substringLength) {
    if (stringLength < substringLength) {
        return -1;
    }

    int32 index = -1;

    for (int i = 0; i <= stringLength - substringLength; i++) {
        if (StringEquals(string + i, substring, substringLength)) {
            index = i;
            break;
        }
    }

    return index;
}

inline int32 FindSubstring(const char *string, uint32 stringLength, const char *substring) {
    return FindSubstring(string, stringLength, substring, StringLength(substring));
}

inline int32 FindSubstring(const char *string, const char *substring) {
    return FindSubstring(string, StringLength(string), substring, StringLength(substring));
}

inline int32 FindLastSubstring(const char *string, uint32 stringLength, const char *substring, uint32 substringLength) {
    int32 index = FindSubstring(string, stringLength, substring, substringLength);

    while (index >= 0) {
        int32 nextIndex = FindSubstring(string + index + 1, stringLength - index - 1, substring, substringLength);

        if (nextIndex > 0) {
            index = index + nextIndex + 1;
        }
        else {
            break;
        }
    }

    return index;
}

inline int32 FindLastSubstring(const char *string, uint32 stringLength, const char *substring) {
    return FindLastSubstring(string, stringLength, substring, StringLength(substring));
}

inline int32 FindLastSubstring(const char *string, const char *substring) {
    return FindLastSubstring(string, StringLength(string), substring, StringLength(substring));
}

inline const char *ScanQuotedString(const char *string, uint32 stringLength, uint32 *charactersConsumed = NULL) {
    const char *result = NULL;

    int32 startQuote = -1;
    for (int i = 0; i < stringLength; i++) {
        if (string[i] == '\n') {
            break;
        }
        else if (string[i] == '"') {
            startQuote = i;
            break;
        }
        else if (!isspace(string[i])) {
            break;
        }
    }

    if (startQuote >= 0) {
        int32 endQuote = -1;
        for (int i = startQuote + 1; i < stringLength; i++) {
            if (string[i] == '\n') {
                break;
            }
            else if (string[i] == '"') {
                endQuote = i;
                break;
            }
        }

        if (endQuote >= 0) {
            uint32 length = (endQuote - startQuote - 1);
            result = StringCopy(string + startQuote + 1, length);

            if (charactersConsumed != NULL) {
                *charactersConsumed = endQuote + 1;
            }
        }
    }

    return result;
}

inline b32 StringContainsCharacter(const char *a, char character) {
    b32 result = false;

    while (*a != '\0') {
        if (*a == character) {
            break;
        }

        a++;
    }
    
    return result;
}

inline b32 StringContainsCapitals(const char *a) {
    b32 result = false;

    while (*a != '\0') {
        if (isupper(*a)) {
            result = true;
            break;
        }

        a++;
    }

    return result;
}

inline const char *RestOfString(const char *str, char c) {
    const char *rest = str;
    
    while (*rest != c) {
        rest++;
    }

    return rest;
}

inline char *SplitString(MemoryArenaList *arena, const char *str, char c) {
    char *result = (char *)PushSize(arena, StringLength(str) + 1);
    Plat.CopyMem(result, str, StringLength(str));
    result[StringLength(str)] = 0;
    s32 index = -1;

    const char *at = str;
    while (*at != c) {
        index++;
        at++;
    }

    result[index + 1] = 0;

    return result;
}

#include "fuzzy-native/score_match.h"

inline real32 StringFuzzyMatchScore(const char *target, const char *input, b32 matchBeginning = false) {
    ASSERT(target != NULL);
    ASSERT(input != NULL);

    if (matchBeginning && tolower(*input) != tolower(*target)) {
        return 0;
    }

    // @TODO: need to frameMem these temp strings (but Game is not available here)!
    
    uint32 targetLength = StringLength(target);
    //char *targetLower = PushArray(Game.frameMem, char, targetLength);
    char *targetLower = ALLOCATE_ARRAY(char, targetLength);

    for (int i = 0; i < targetLength; i++) {
        targetLower[i] = tolower(target[i]);
    }

    uint32 inputLength = StringLength(input);
    //char *inputLower = PushArray(Game.frameMem, char, inputLength);
    char *inputLower = ALLOCATE_ARRAY(char, inputLength);

    for (int i = 0; i < inputLength; i++) {
        inputLower[i] = tolower(input[i]);
    }

    MatchOptions options;
    options.case_sensitive = false;
    options.smart_case = true;
    options.max_gap = inputLength;

    //int32 *matchIndices = PushArray(Game.frameMem, int32, inputLength);
    //int32 matchIndicesCount;
    real32 score = score_match(target, targetLower, input, inputLower, options);

    Plat.DeallocateMem(targetLower);
    Plat.DeallocateMem(inputLower);
    
    return score;
}

inline b32 StringFuzzyMatches(const char *target, const char *input, b32 matchBeginning = false) {
    return StringFuzzyMatchScore(target, input, matchBeginning) > 0;
}

inline void FuzzySortStrings(const char **strings, uint32 stringCount, const char *input, uint32 *matchingIndices, uint32 *matchingCount, b32 matchBeginning = false) {

    // @TODO: need to frameMem this temp array (but Game is not available here)!
    real32 *scores = ALLOCATE_ARRAY(real32, stringCount);
    *matchingCount = 0;

    for (int i = 0; i < stringCount; i++) {
        real32 score = StringFuzzyMatchScore(strings[i], input, matchBeginning);

        if (score > 0) {

            for (int j = *matchingCount; j >= 0; j--) {
                if (j == 0 || scores[j - 1] >= score) {
                    matchingIndices[j] = i;
                    scores[j] = score;
                    *matchingCount += 1;
                    break;
                }

                matchingIndices[j] = matchingIndices[j - 1];
                scores[j] = scores[j - 1];
            }
        }
    }

    Plat.DeallocateMem(scores);
}

inline char *PathConcat(const char *head, const char *tail) {
    ASSERT(head != NULL);
    ASSERT(tail != NULL);

#if _WIN32
    char directoryDelimiter = '\\';
#else
    char directoryDelimiter = '/';
#endif

    uint32 headLength = strlen(head);
    uint32 tailLength = strlen(tail);
    uint32 pathLength = headLength + tailLength;

    if (head[headLength - 1] != directoryDelimiter) {
        pathLength += 1;
    }

    char *path = (char *)Plat.AllocateMem(sizeof(char) * (pathLength + 1));
    Plat.CopyMem(path, head, headLength);

    if (head[headLength - 1] != directoryDelimiter) {
        path[headLength] = directoryDelimiter;
        Plat.CopyMem(path + headLength + 1, tail, tailLength);
    }
    else {
        Plat.CopyMem(path + headLength, tail, tailLength);
    }

    path[pathLength] = 0;

    return path;
}

char *PathConcat(const char *head, const char *tail, MemoryArenaList *arena) {
    ASSERT(head != NULL);
    ASSERT(tail != NULL);

#if _WIN32
    const char *directoryDelimiter = "\\";
#else
    const char *directoryDelimiter = "/";
#endif
    
    char *path = (char *)StringConcat(arena, 3, head, directoryDelimiter, tail);
    return path;
}

char *PathConcat(MemoryArenaList *arena, uint32 count, ...) {

#if _WIN32
    const char *directoryDelimiter = "\\";
#else
    const char *directoryDelimiter = "/";
#endif

    ASSERT(count >= 2);
    va_list args;
    va_start(args, count);

    char *result = NULL;
    char *str1 = va_arg(args, char *);
    char *str2 = va_arg(args, char *);
    result = StringConcat(arena, 3, str1, directoryDelimiter, str2);
    
    for (int i = 2; i < count; i++) {
        char *str = va_arg(args, char *);
        result = StringConcat(arena, 3, result, directoryDelimiter, str);
    }

    va_end(args);

    return result;
}

bool HasFileExtension(const char *filename, const char *extension) {
    bool result = false;

    const char *extractedExtension = NULL;
    for (int i = 0; i < strlen(filename); i++) {
        if (filename[i] == '.') {
            extractedExtension = filename + i;
        }
    }

    // If the extension passed in doesn't start with '.' then the
    // comparison below breaks unless we check against the extension
    // with the dot
    if (extension[0] != '.' && extractedExtension != NULL) {
        extractedExtension = extractedExtension + 1;
    }

    if (extractedExtension != NULL) {
        if (StringEquals(extension, extractedExtension)) {
            result = true;
        }
    }

    return result;
}

void RemoveTrailingSlashes(char *str) {
    uint32 strLength = strlen(str);
    for (int i = strLength - 1; i >= 0; i--) {
        if (str[i] == '/') {
            str[i] = 0;
        }
        else {
            break;
        }
    }
}

inline char *AppendFileExtension(const char *filename, const char *extension) {
    uint32 filenameLength = strlen(filename);
    uint32 extensionLength = strlen(extension);
    uint32 filenameFullLength = filenameLength + extensionLength + 1;

    char *filenameFull = (char *)Plat.AllocateMem(sizeof(char) * (filenameFullLength + 1));
    Plat.CopyMem(filenameFull, filename, filenameLength);
    filenameFull[filenameLength] = '.';
    Plat.CopyMem(filenameFull + filenameLength + 1, extension, extensionLength);
    filenameFull[filenameFullLength] = 0;
    
    return filenameFull;
}

inline char *AppendFileExtension(const char *filename, const char *extension, MemoryArenaList *arena) {
    char *filenameFull = (char *)StringConcat(arena, 3, filename, ".", extension);
    return filenameFull;
}

inline b32 SplitFileExtension(const char *path, const char **filename, const char **extension) {
    b32 result = false;

    int32 index = FindLastSubstring(path, ".");
    if (index >= 0) {
        if (filename != NULL) {
            *filename = StringCopy(path, index);
        }
        if (extension != NULL) {
            *extension = StringCopy(path + index + 1);
        }

        result = true;
    }

    return result;
}

inline void SplitPath(const char *path, const char **directoryName, const char **fileName) {

    uint32 pathLength = StringLength(path);

    int32 lastSlash = FindLastSubstring(path, pathLength, "/", 1);
    if (lastSlash < 0) {
        *fileName = StringCopy(path, pathLength);
    }

    *directoryName = StringCopy(path, lastSlash + 1);
    *fileName = StringCopy(path + lastSlash + 1, pathLength - (lastSlash + 1));
}

b32 ValidNumericString(char *string, b32 isSigned, b32 isReal) {
    b32 result = true;
    
    uint32 signCount = 0;
    int32 signIndex = -1;
    
    uint32 dotCount = 0;
    int32 dotIndex = -1;
    
    int32 firstDigit = -1;

    for (int i = 0; i < strlen(string); i++) {
        if (string[i] == '-') {
            if (signIndex < 0) {
                signIndex = i;
            }
            signCount++;
        }
        else if (string[i] == '.') {
            if (dotIndex < 0) {
                dotIndex = i;
            }
            dotCount++;
        }
        else if (firstDigit < 0) {
            firstDigit = i;
            if (!isReal && string[firstDigit] == '0') {
                result = false;
                break;
            }
            else if (!IsDigit(string[i])) {
                result = false;
                break;
            }
        }
        else if (!IsDigit(string[i])) {
            result = false;
            break;
        }

        if ((isSigned && signCount > 1) ||
            (!isSigned && signCount > 0) ||
            (signCount > 0 && signIndex != 0) ||
            (isReal && dotCount > 1) ||
            (!isReal && dotCount > 0)) {
            
            result = false;
            break;
        }
    }
    
    return result;
}

inline b32 ValidUIntString(char *string) {
    return ValidNumericString(string, false, false);
}

inline b32 ValidIntString(char *string) {
    return ValidNumericString(string, true, false);
}

inline b32 ValidFloatString(char *string) {
    return ValidNumericString(string, true, true);
}

int32 QuicksortPartition_(uint32 *array, uint32 elemSize, int32 low, int32 high) {
    uint32 pivot = array[low];

    int32 i = low;
    int32 j = high;

    while (true) {
        while (array[i] < pivot) {
            i++;
        }

        while (array[j] > pivot) {
            j--;            
        }

        if (i >= j) {
            return j;
        }

        if (i != j) {
            void *temp = alloca(elemSize);
            Plat.CopyMem(temp, (uint8 *)array + (elemSize * i), elemSize);
            Plat.CopyMem((uint8 *)array + (elemSize * i),
                         (uint8 *)array + (elemSize * j), elemSize);
            Plat.CopyMem((uint8 *)array + (elemSize * j), temp, elemSize);
        }

        i++;
        j--;
    }
}

void Quicksort_(uint32 *array, uint32 elemSize, int32 low, int32 high) {
    if (low >= high) {
        return;
    }
    int32 mid = QuicksortPartition_(array, elemSize, low, high);
    Quicksort_(array, elemSize, low, mid);
    Quicksort_(array, elemSize, mid + 1, high);
}

void Quicksort(uint32 *array, uint32 elemSize, int32 count) {
    Quicksort_(array, elemSize, 0, count - 1);
}


// @NOTE: the comparator returns 1 if a > b, -1 if b < a, and 0 if they're equal
int32 QuicksortPartition_(void *array, uint32 elemSize, int32 low, int32 high, int32 (* comparator)(const void *a, const void *b)) {
    void *pivot = alloca(elemSize);
    Plat.CopyMem(pivot, (uint8 *)array + (elemSize * low), elemSize);

    int32 i = low;
    int32 j = high;

    while (true) {
        while (comparator((uint8 *)array + (elemSize * i),  pivot) < 0) {
            i++;
        }

        while (comparator((uint8 *)array + (elemSize * j), pivot) > 0) {
            j--;            
        }

        if (i >= j) {
            return j;
        }

        if (i != j) {
            void *temp = alloca(elemSize);
            Plat.CopyMem(temp, (uint8 *)array + (elemSize * i), elemSize);
            Plat.CopyMem((uint8 *)array + (elemSize * i),
                         (uint8 *)array + (elemSize * j), elemSize);
            Plat.CopyMem((uint8 *)array + (elemSize * j), temp, elemSize);
        }

        i++;
        j--;
    }
}

void Quicksort_(void *array, uint32 elemSize, int32 low, int32 high, int32 (* comparator)(const void *a, const void *b)) {
    if (low >= high) {
        return;
    }
    int32 mid = QuicksortPartition_(array, elemSize, low, high, comparator);
    Quicksort_(array, elemSize, low, mid, comparator);
    Quicksort_(array, elemSize, mid + 1, high, comparator);
}

void Quicksort(void *array, uint32 elemSize, int32 count, int32 (* comparator)(const void *a, const void *b)) {
    Quicksort_(array, elemSize, 0, count - 1, comparator);
}


int32 QuicksortPartition_(void *array, uint32 elemSize, int32 low, int32 high, int32 (* comparator)(const void *a, const void *b, void *data), void *data) {
    void *pivot = alloca(elemSize);
    Plat.CopyMem(pivot, (uint8 *)array + (elemSize * low), elemSize);

    int32 i = low;
    int32 j = high;

    while (true) {
        while (comparator((uint8 *)array + (elemSize * i),  pivot, data) < 0) {
            i++;
        }

        while (comparator((uint8 *)array + (elemSize * j), pivot, data) > 0) {
            j--;            
        }

        if (i >= j) {
            return j;
        }

        if (i != j) {
            void *temp = alloca(elemSize);
            Plat.CopyMem(temp, (uint8 *)array + (elemSize * i), elemSize);
            Plat.CopyMem((uint8 *)array + (elemSize * i),
                         (uint8 *)array + (elemSize * j), elemSize);
            Plat.CopyMem((uint8 *)array + (elemSize * j), temp, elemSize);
        }

        i++;
        j--;
    }
}

void Quicksort_(void *array, uint32 elemSize, int32 low, int32 high, int32 (* comparator)(const void *a, const void *b, void *data), void *data) {
    if (low >= high) {
        return;
    }
    int32 mid = QuicksortPartition_(array, elemSize, low, high, comparator, data);
    Quicksort_(array, elemSize, low, mid, comparator, data);
    Quicksort_(array, elemSize, mid + 1, high, comparator, data);
}

void Quicksort(void *array, uint32 elemSize, int32 count, int32 (* comparator)(const void *a, const void *b, void *data), void *data) {
    Quicksort_(array, elemSize, 0, count - 1, comparator, data);
}



// @Note: index, value, length is of array, so the number of pairs is length / 2
void QuicksortIndexValuePair(u32 *array, uint32 length) {
    if (length < 4) {
        return;
    }
    u32 pairCount = length / 2;
    u32 pairIndex = pairCount;
    // @NOTE: the pair indices are always even, so if pairCount isnt even, we need to make it so.
    if (pairCount % 2 != 0) {
        pairIndex--;
    }

    u32 pivot = array[pairIndex  + 1];
    u32 i = 0;
    u32 j = length - 2;
    while(true) {
        while(array[i + 1] < pivot) {
            i += 2;
        }
        while(array[j + 1] > pivot) {
            j -= 2;
        }
        if (i >= j) {
            break;
        }

        if (array[i + 1] != array[j + 1]) {
            u32 tempIndex = array[i];
            u32 tempValue = array[i + 1];
            array[i] = array[j];
            array[i + 1] = array[j + 1];
            array[j] = tempIndex;
            array[j + 1] = tempValue;
        }
        
        i += 2;
        j -= 2;
    }

    QuicksortIndexValuePair(array, i);
    QuicksortIndexValuePair(array + i, length - i);
}

typedef bool Comparator(void *, void *);

b32 ElementInBufferContainsField_(void *buffer, uint32 count, uint32 elementSize,uint32 elementOffset, void *element, Comparator *comparator, uint32 *index = NULL) {

    for (int i = 0; i < count; i++) {
        if (comparator((void *)((uint8 *)buffer + (i * elementSize) + elementOffset), element)) {
            if (index != NULL) {
                *index = i;
            }
            return true;
        }
    }

    return false;
}

#define ElementInBufferContainsField(buffer, count, elemName, fieldName, element, Comparator, index) ElementInBufferContainsField_(buffer, count, sizeof(elemName), FIELD_OFFSET(elemName, fieldName), element, Comparator, index)

bool Contains(void *buffer, u32 count, u32 size, void *elem, Comparator *comparator, u32 *index) {
    return ElementInBufferContainsField_(buffer, count, size, 0, elem, comparator, index);
}


// TODO: move spin lock
struct SpinLock {
    uint32 value;
};

inline void AcquireLock(SpinLock *lock) {
    while (true) {
        if (0 == EtcCompareAndSwap(&lock->value, 0, 1)) {
            break;
        }
    }
}

inline void ReleaseLock(SpinLock *lock) {
    lock->value = 0;
}

struct Str {
    uint32 length;
    char *string;

    Str(const char *s, uint32 l) {
        length = l;
        string = StringCopy_(s, length);
    }

    Str(const char *s, uint32 l, MemoryArenaList *list) {
        length = l;
        string = StringCopy(list, s, length);
    }

    Str(const char *s) {
        Str(s, StringLength(s));
    }

};

#endif

