
#ifndef _ETC_DEBUG_H
#define _ETC_DEBUG_H

#define DEBUG_MAX_INPUTCHARACTERS 16
#define DEBUG_TEXT_MAX_LENGTH 1024
#define DEBUG_ARENA_SIZE Megabytes(8)

#define DEBUG_ASSET_RELOADS_PER_FRAME 128


// Profiling
struct DebugRecordLabel {
    const char *label;
    ChunkedArray<uint32> recordIndices;
};

struct DebugRecord {
    const char* label;
    
    real64 startTime;
    real64 timeElapsed;
    const char *fileName;
    const char *functionName;
    uint32 lineNumber;
    uint32 stackDepth;

    // highest 32 bits are the hit count and lowest 32 are cycle count
    uint64 hitAndCycleCount;

    uint32 lastNonZeroCycleCount;
    real64 lastNonZeroElapsedTime;
    uint32 maxCycleCount;
    uint32 minCycleCount;
};

enum DebugStatsState {
    DebugStatsState_Hidden,
    DebugStatsState_Visible,
};


// Debug Log
#define MAX_LOG_CHARACTERS 256

enum DebugLogState {
    DebugLogState_Hidden,
    DebugLogState_Visible,
};

enum DebugLogDisplayState {
    DebugLogDisplayState_Filtered, // everything from one location gets printed on same line.
    DebugLogDisplayState_Scrolling,  
    
};

struct LogData {
    char buffer[MAX_LOG_CHARACTERS];
    const char *fileName;
    uint32 lineNum;
    uint32 instance;
    uint32 frame;
};

struct DebugLog {
    DebugLogState state;
    DebugLogDisplayState displayState;

    bool snapToBottom;

    BufferedList logData;

    EtcRect windowRect;
};

inline void DebugLogSubmit(DebugLog *log, const char *name, uint32 lineNum, uint32 instance, const char *fmt...) {
    LogData logData = {};

    logData.frame = Game.frame;

    char buffer[MAX_LOG_CHARACTERS];
    va_list args;
    va_start(args, fmt);
    vsnprintf(buffer, MAX_LOG_CHARACTERS, fmt, args);
    

    // add the header + all the log data
    uint32 len = strlen(buffer);
    for (int i = 0; i < len; i++) {
        logData.buffer[i] = buffer[i];
    }
    logData.buffer[len] = 0;
    
    logData.fileName = name;
    logData.lineNum = lineNum;
    logData.instance = instance;
    
    void *ptr = PushBackPtr(&log->logData, sizeof(logData));
    Plat.CopyMem(ptr, &logData, sizeof(logData));

    va_end(args);
}

inline void DebugLogSubmitVector(DebugLog *log, const char *name, uint32 lineNum, uint32 instance, const char *label, vec2 vector) {
    
    DebugLogSubmit(log, name, lineNum, instance, "%s x: %f y: %f", label, vector.x, vector.y);
}

inline void DebugLogSubmitVector(DebugLog *log, const char *name, uint32 lineNum, uint32 instance, const char *label, vec3 vector) {
    
    DebugLogSubmit(log, name, lineNum, instance, "%s x: %f y: %f z: %f", label, vector.x, vector.y, vector.z);
}

inline void DebugLogSubmitVector(DebugLog *log, const char *name, uint32 lineNum, uint32 instance, const char *label, vec4 vector) {
    
    DebugLogSubmit(log, name, lineNum, instance, "%s x: %f y: %f z: %f w: %f", label, vector.x, vector.y, vector.z, vector.w);
}

inline void DebugLogSubmitColor(DebugLog *log, const char *name, uint32 lineNum, uint32 instance, const char *label, vec4 color) {
    
    DebugLogSubmit(log, name, lineNum, instance, "%s r: %f g: %f b: %f a: %f", label, color.r, color.g, color.b, color.a);
}


#define LOG(string) DebugLogSubmit(&DebugMem->log, __FILE__, __LINE__, 0, string)
#define LOGF(string, ...) DebugLogSubmit(&DebugMem->log, __FILE__, __LINE__, 0, string, __VA_ARGS__)
#define LOGI(i, string, ...) DebugLogSubmit(&DebugMem->log, __FILE__, __LINE__, i, string, __VA_ARGS__)

#define LOGV(label, vec) DebugLogSubmitVector(&DebugMem->log, __FILE__, __LINE__, 0, label, (vec));
#define LOGC(label, vec) DebugLogSubmitColor(&DebugMem->log, __FILE__, __LINE__, 0, label, (vec));


// Debug Draw
struct DebugDrawLineVertex {
    vec3 position;
    vec4 color;
    real32 dashParam;
    real32 depthFailScale;
};

struct DebugDrawLineBuffer {
    DebugDrawLineBuffer *next;

    uint32 capacity;
    uint32 count;

    uint32 bufferID;
    uint32 bufferSize;
    DebugDrawLineVertex *buffer;
};

struct DebugDrawMeshInstanceData {
    mat4 transform;
    vec4 color;
    real32 depthFailScale;
};

struct DebugDrawMeshInstanceBuffer {
    DebugDrawMeshInstanceBuffer *next;

    MeshHandle meshHandle;
    b32 wireframe;

    uint32 capacity;
    uint32 count;

    uint32 bufferID;
    uint32 bufferSize;
    DebugDrawMeshInstanceData *buffer;
};

vec2 GetDebugGlyphDimensions(real32 size);

struct DebugDrawTextInstanceData {
    uint32 codepoint;
    vec3 position;
    quaternion rotation;
    vec2 scale;
    vec4 color;
};

struct DebugDrawTextInstanceBuffer {
    DebugDrawTextInstanceBuffer *next;

    const char *text;

    real32 size;

    uint32 capacity;
    uint32 count;
    
    uint32 bufferID;
    uint32 bufferSize;

    mat4 model;
    
    DebugDrawTextInstanceData *buffer;
};

struct DebugDraw {
    DebugDrawLineBuffer *lineBufferHead;
    DebugDrawLineBuffer *currentLineBuffer;

    DebugDrawLineBuffer *screenLineBufferHead;
    DebugDrawLineBuffer *currentScreenLineBuffer;

    DebugDrawMeshInstanceBuffer *meshInstanceBufferHead;
    DebugDrawMeshInstanceBuffer *screenMeshInstanceBufferHead;

    DebugDrawTextInstanceBuffer *textInstanceBufferHead;
    DebugDrawTextInstanceBuffer *screenTextInstanceBufferHead;

    ShaderProgramHandle lineShader;
    ShaderProgramHandle meshShader;
    ShaderProgramHandle textShader;
};

void DebugDrawLine(vec3 a, vec3 b, vec4 color = WHITE, real32 dashLength = 0.0f, real32 dashOffset = 0.0f, real32 depthFailScale = DEFAULT_DEPTH_FAIL_SCALE);

void DebugDrawRay(Ray ray, vec4 color = WHITE, real32 dashLength = 0.0f, real32 dashOffset = 0.0f, real32 depthFailScale = DEFAULT_DEPTH_FAIL_SCALE);

void DebugDrawLineScreen(vec3 a, vec3 b, vec4 color = WHITE, real32 dashLength = 0.0f, real32 dashOffset = 0.0f, real32 depthFailScale = DEFAULT_DEPTH_FAIL_SCALE);
void DebugDrawLineScreen(vec2 a, vec2 b, vec4 color = WHITE, real32 dashLength = 0.0f, real32 dashOffset = 0.0f, real32 depthFailScale = DEFAULT_DEPTH_FAIL_SCALE);

void DebugDrawLineNorm(vec3 a, vec3 b, vec4 color = WHITE, real32 dashLength = 0.0f, real32 dashOffset = 0.0f, real32 depthFailScale = DEFAULT_DEPTH_FAIL_SCALE);
void DebugDrawLineScreenNorm(vec2 a, vec2 b, vec4 color = WHITE, real32 dashLength = 0.0f, real32 dashOffset = 0.0f, real32 depthFailScale = DEFAULT_DEPTH_FAIL_SCALE);

void DebugDrawPoint(vec3 a, vec4 color = WHITE, real32 scale = 1.0f, real32 depthFailScale = DEFAULT_DEPTH_FAIL_SCALE);

void DebugDrawPointScreen(vec2 a, vec4 color = WHITE, real32 scale = 1.0f, real32 depthFailScale = DEFAULT_DEPTH_FAIL_SCALE);

void DebugDrawCoordinateSystem(vec3 position, vec3 X, vec3 Y, vec3 Z, real32 scale = 1.0f, real32 depthFailScale = DEFAULT_DEPTH_FAIL_SCALE);

void DebugDrawCoordinateSystem(vec3 position, quaternion rotation, real32 scale = 1.0f, real32 depthFailScale = DEFAULT_DEPTH_FAIL_SCALE);

void DebugDrawPlane(vec3 point, vec3 normal, vec4 color, bool showNormal = false, vec2 scale = V2(1, 1));
void DebugDrawPlane(Plane plane, vec4 color, bool showNormal = false, vec2 scale = V2(1, 1));

void DebugDrawTriangle(vec3 a, vec3 b, vec3 c, vec4 color = WHITE);
void DebugDrawTriangle(vec3 *verts, vec4 color = WHITE);

void DebugDrawRect(vec3 position, vec3 X, vec3 Y, vec4 color = WHITE, real32 dashLength = 0.0f, real32 dashOffset = 0.0f, real32 depthFailScale = DEFAULT_DEPTH_FAIL_SCALE);

void DebugDrawRect(vec3 position, quaternion rotation, vec2 scale, vec4 color = WHITE, real32 dashLength = 0.0f, real32 dashOffset = 0.0f, real32 depthFailScale = DEFAULT_DEPTH_FAIL_SCALE);

void DebugDrawCenteredRect(vec3 position, quaternion rotation, vec2 scale, vec4 color = WHITE, real32 dashLength = 0.0f, real32 dashOffset = 0.0f, real32 depthFailScale = DEFAULT_DEPTH_FAIL_SCALE);

void DebugDrawRectScreen(vec2 position, vec2 dimensions, vec4 color = WHITE, real32 dashLength = 0.0f, real32 dashOffset = 0.0f, real32 depthFailScale = DEFAULT_DEPTH_FAIL_SCALE);

void DebugDrawAABB(AABB aabb, Transform transform, vec4 color = WHITE, real32 depthFailScale = DEFAULT_DEPTH_FAIL_SCALE);
void DebugDrawAABB(vec3 center, vec3 halfExtents, vec4 color = WHITE, real32 depthFailScale = DEFAULT_DEPTH_FAIL_SCALE);
void DebugDrawAABB(AABB aabb, vec3 position, vec4 color = WHITE, real32 depthFailScale = DEFAULT_DEPTH_FAIL_SCALE);

void DebugDrawGrid(vec3 origin, vec3 X, vec3 Y, real32 majorSpacing, int32 majorSegments, int32 minorSegments, vec4 majorColor, vec4 minorColor, real32 depthFailScale = DEFAULT_DEPTH_FAIL_SCALE);

void DebugDrawQuadScreen(vec2 position, vec2 scale, vec4 color, real32 depthFailScale = DEFAULT_DEPTH_FAIL_SCALE);

void DebugDrawCircle(vec3 center, real32 radius, vec3 normal, vec4 color = WHITE, uint32 segments = 32, real32 depthFailScale = DEFAULT_DEPTH_FAIL_SCALE);

void DebugDrawCircleScreen(vec2 center, real32 radius, vec4 color = WHITE, uint32 segments = 32);

void DebugDrawArc(vec3 center, real32 radius, vec3 normal, vec3 X, real32 startAngle, real32 endAngle, vec4 color = WHITE, uint32 segments = 32, real32 depthFailScale = DEFAULT_DEPTH_FAIL_SCALE);

void DebugDrawArcScreen(vec2 center, real32 radius, real32 startAngle, real32 endAngle, vec4 color = WHITE, uint32 segments = 32, real32 depthFailScale = DEFAULT_DEPTH_FAIL_SCALE);

void DebugDrawBezier(vec3 a, vec3 b, vec3 c, vec3 d, vec4 color = WHITE, uint32 segments = 16, real32 dashLength = 0.0f, real32 dashOffset = 0.0f, real32 depthFailScale = DEFAULT_DEPTH_FAIL_SCALE);

void DebugDrawBezierScreen(vec2 a, vec2 b, vec2 c, vec2 d, vec4 color, uint32 segments);

void DebugDrawCatmullRom(vec3 a, vec3 b, vec3 c, vec3 d, real32 alpha, vec4 color = WHITE, uint32 segments = 16, real32 depthFailScale = DEFAULT_DEPTH_FAIL_SCALE);

void DebugDrawStickFigure(vec2 position, vec2 scale, vec4 color, bool reloading = false, bool failed = false);

void DebugDrawMesh(MeshHandle mesh, vec3 position, quaternion rotation, real32 scale = 1.0f, vec4 color = WHITE, real32 depthFailScale = DEFAULT_DEPTH_FAIL_SCALE);
void DebugDrawMesh(MeshHandle mesh, vec3 position, real32 scale = 1.0f, vec4 color = WHITE, real32 depthFailScale = DEFAULT_DEPTH_FAIL_SCALE);
void DebugDrawSphere(vec3 position, real32 scale, vec4 color, real32 depthFailScale = DEFAULT_DEPTH_FAIL_SCALE);
void DebugDrawSphereWireframe(vec3 position, real32 scale, vec4 color, real32 depthFailScale = DEFAULT_DEPTH_FAIL_SCALE);

void DebugDrawMesh(MeshHandle mesh, vec3 position, quaternion rotation, vec3 scale = V3(1, 1, 1), vec4 color = WHITE, real32 depthFailScale = DEFAULT_DEPTH_FAIL_SCALE);
void DebugDrawMesh(MeshHandle mesh, vec3 position, vec3 scale = V3(1, 1, 1), vec4 color = WHITE, real32 depthFailScale = DEFAULT_DEPTH_FAIL_SCALE);

void DebugDrawMeshWireframe(MeshHandle mesh, vec3 position, quaternion rotation, real32 scale = 1.0f, vec4 color = WHITE, real32 depthFailScale = DEFAULT_DEPTH_FAIL_SCALE);
void DebugDrawMeshWireframe(MeshHandle mesh, vec3 position, real32 scale = 1.0f, vec4 color = WHITE, real32 depthFailScale = DEFAULT_DEPTH_FAIL_SCALE);

void DebugDrawMeshWireframe(MeshHandle mesh, vec3 position, quaternion rotation, vec3 scale = V3(1, 1, 1), vec4 color = WHITE, real32 depthFailScale = DEFAULT_DEPTH_FAIL_SCALE);
void DebugDrawMeshWireframe(MeshHandle mesh, vec3 position, vec3 scale = V3(1, 1, 1), vec4 color = WHITE, real32 depthFailScale = DEFAULT_DEPTH_FAIL_SCALE);

void DebugDrawMeshNormals(Mesh *mesh, real32 normalScale, vec3 position, quaternion rotation, real32 scale, vec4 color = WHITE, real32 depthFailScale = DEFAULT_DEPTH_FAIL_SCALE);

void DebugDrawMeshScreen(MeshHandle mesh, vec2 position, quaternion rotation, vec3 scale, vec4 color, real32 depthFailScale = DEFAULT_DEPTH_FAIL_SCALE);
void DebugDrawMeshScreen(MeshHandle mesh, vec2 position, vec3 scale, vec4 color, real32 depthFailScale = DEFAULT_DEPTH_FAIL_SCALE);
void DebugDrawMeshScreen(MeshHandle mesh, vec2 position, real32 scale, vec4 color, real32 depthFailScale = DEFAULT_DEPTH_FAIL_SCALE);
void DebugDrawMeshScreenNorm(MeshHandle mesh, vec2 position, real32 scale, vec4 color, real32 depthFailScale = DEFAULT_DEPTH_FAIL_SCALE);

void DebugDrawText(vec3 position, quaternion rotation, real32 size, vec4 color, int32 cursorAt, vec2 *dimensions, const char *format, ...);
void DebugDrawText(vec3 position, quaternion rotation, real32 size, vec4 color, vec2 *dimensions, const char *format, ...);
void DebugDrawText(vec3 position, quaternion rotation, real32 size, vec4 color, int32 cursorAt, const char *format, ...);
void DebugDrawText(vec3 position, quaternion rotation, real32 size, vec4 color, const char *format, ...);
void DebugDrawText(vec3 position, real32 size, vec4 color, const char *format, ...);
void DebugDrawVector(vec3 position, quaternion rotation, real32 size, vec4 color = WHITE);
void DebugDrawTextScreen(vec2 position, real32 size, vec4 color, int32 cursorAt, vec2 *dimensions, const char *format, ...);
void DebugDrawTextScreen(vec2 position, real32 size, vec4 color, int32 cursorAt, const char *format, ...);
void DebugDrawTextScreen(vec2 position, real32 size, vec4 color, const char *format, ...);

void DebugDrawCodepoints(vec3 position, quaternion rotation, real32 size, vec4 color, vec2 *dimensions, vec2 *positions, uint32 *lineCount, uint32 *codepoints, uint32 codepointCount, real32 maxX = INFINITY);
void DebugDrawCodepointsScreen(vec2 position, real32 size, vec4 color, vec2 *dimensions, vec2 *positions, uint32 *lineCount, uint32 *codepoints, uint32 codepointCount, real32 maxX = INFINITY);

void DebugDrawCodepointsScreen(vec2 position, FontTable *font, real32 size, vec4 color, vec2 *dimensions, vec2 *positions, uint32 *lineCount, uint32 *codepoints, uint32 codepointCount, real32 maxX = INFINITY);


// DebugMemory
#define DEBUG_RECORD_COUNT 1024

struct EtcDebugMemory {
    MemoryArena arena;

    uint32 recordStackDepth;
    DynamicArray<const char*> debugRecordLabels;
    DebugRecord debugRecords[DEBUG_RECORD_COUNT];
    DebugStatsState statsState;


    // Debug Draw
    MeshHandle quad;
    MeshHandle cube;
    MeshHandle skyboxCube;
    MeshHandle sphere;
    MeshHandle sphereHiPoly;
    MeshHandle cylinder;

    // @HACK: hella for changes
    MeshHandle octagon;
    MeshHandle hexagon;
    MeshHandle pentagon;
    MeshHandle diamond;
    MeshHandle trigon;


    MeshHandle arrow;
    MeshHandle cone;
    MeshHandle spotlightCone;
    MeshHandle circle;

    TextureHandle white;
    TextureHandle black;
    TextureHandle magenta;
    TextureHandle clear;

    Material missingMaterial;

    ShaderProgramHandle passthroughShader;
    ShaderProgramHandle passthroughDepthShader;
    ShaderProgramHandle passthroughNormalShader;

    ShaderProgramHandle passthroughBlendedShader;
    ShaderProgramHandle passthroughNormalBlendedShader;

    ShaderProgramHandle debugPaintedSkyboxShader;
    

    FontTableHandle debugFontHandle;
    FontTable *debugFont;
    real32 screenTextSize;

    TextureHandle debugUIArrowRight;
    TextureHandle debugUIArrowDown;
    TextureHandle debugUIArrowLeft;
    TextureHandle debugUIArrowUp;

    MeshHandle debugUIColorPickerRing;
    MeshHandle debugUIColorPickerTriangle;
    ShaderProgramHandle debugUIColorPickerRingShaderRGB;
    ShaderProgramHandle debugUIColorPickerRingShaderRYB;
    ShaderProgramHandle debugUIColorPickerTriangleShaderRGB;
    ShaderProgramHandle debugUIColorPickerTriangleShaderRYB;

    ColorPickerWindowData colorPickerWindowData;
    ContextMenuData contextMenuData;

    CommandList debugDrawPass;
    DebugDraw debugDraw;

    b32 shouldOverrideRenderMode;
    RenderMode renderModeOverride;

    uint32 inputCharactersCapacity;
    uint32 inputCharactersCount;
    char *inputCharacters;

    // NOTE: the whole reason to separate is to control when we clear:
    // we dont want to do it in ResetHandles or at the end of DebugUpdate since both are called before
    // the handles are used in the scene update.
    uint32 scriptInputCharactersCapacity;
    uint32 scriptInputCharactersCount;
    char *scriptInputCharacters;

    DebugLog log;
};


void InitDebug(EtcDebugMemory *debugMem);
void DeinitDebug(EtcDebugMemory *debugMem);

void InitDebugDraw(EtcDebugMemory *debugMem);
void DeinitDebugDraw(EtcDebugMemory *debugMem);

void DebugUpdate(struct EtcGameMemory *mem);
void DebugRender(struct EtcGameMemory *mem);


extern "C" EtcDebugMemory *DebugMem;

typedef void RegisterDebugMemoryFunc(EtcDebugMemory *debugMem);
extern "C" DLLEXPORT void RegisterDebugMemory(EtcDebugMemory *debugMem);

typedef EtcDebugMemory *GetDebugMemoryFunc();
extern "C" DLLEXPORT EtcDebugMemory *GetDebugMemory();


// Debug UI
inline void DebugAddInputCharacters(EtcDebugMemory *debugMem, const char *utf8_string) {
    uint32 length = strlen(utf8_string);

    uint32 charactersToCopy = Min(length, debugMem->inputCharactersCapacity - 1 - debugMem->inputCharactersCount);

    Plat.CopyMem(debugMem->inputCharacters + debugMem->inputCharactersCount, utf8_string, length);
    debugMem->inputCharactersCount += charactersToCopy;
}

inline void DebugAddScriptInputCharacters(EtcDebugMemory *debugMem, const char *utf8_string) {
    uint32 length = strlen(utf8_string);

    uint32 charactersToCopy = Min(length, debugMem->scriptInputCharactersCapacity - 1 - debugMem->scriptInputCharactersCount);

    Plat.CopyMem(debugMem->scriptInputCharacters + debugMem->scriptInputCharactersCount, utf8_string, length);
    debugMem->scriptInputCharactersCount += charactersToCopy;
}

inline void DebugClearInputCharacters(EtcDebugMemory *debugMem) {
    Plat.ClearMem(debugMem->inputCharacters, sizeof(char) * debugMem->inputCharactersCount);
    debugMem->inputCharactersCount = 0;
}

inline void DebugClearScriptInputCharacters(EtcDebugMemory *debugMem) {
    Plat.ClearMem(debugMem->scriptInputCharacters, sizeof(debugMem->scriptInputCharacters));
    debugMem->scriptInputCharactersCount = 0;
}


inline void SetOverrideRenderMode(RenderMode renderMode) {
    DebugMem->shouldOverrideRenderMode = true;
    DebugMem->renderModeOverride = renderMode;
}


// @GACK: this needs to be down here because of the DebugMem dependency...
// Profiling (contd.)
#define TIMED_BLOCK__(label, Number, ...) TimedBlock TimedBlock_##Number(label, __COUNTER__, __FILE__, __LINE__, (const char *)__FUNCTION__, ## __VA_ARGS__)
#define TIMED_BLOCK_(label, Number, ...) TIMED_BLOCK__(label, Number, ## __VA_ARGS__)
#define TIMED_BLOCK(...) TIMED_BLOCK_(NULL, __LINE__, ## __VA_ARGS__)
#define TIMED_BLOCK_LABEL(label, ...) TIMED_BLOCK_(label, __LINE__, ## __VA_ARGS__)

struct TimedBlock {
    DebugRecord *record;
    uint64 startCycles;
    uint32 hitCount;

    // @NOTE: On Linux there's an overhead of about 230 cycles for a timed block.
    TimedBlock(const char *label, uint32 counter, const char *fileName, uint32 lineNumber, const char *functionName, uint32 hitCountInit = 1) {
        // @TODO: actually increment hitcount
        hitCount = hitCountInit;

        record = DebugMem->debugRecords + counter;

        if (label != NULL && !Contains(&DebugMem->debugRecordLabels, label, StringEquals)) {
            PushBack(&DebugMem->debugRecordLabels, label);
        }

        // @TODO: multithreading stackdepth

        record->stackDepth = DebugMem->recordStackDepth++;

        record->label = label;
        record->fileName = fileName;
        record->lineNumber = lineNumber;
        record->functionName = functionName;
        record->startTime = Plat.GetTime();
        
        startCycles = EtcCycleCount();
    }

    ~TimedBlock() {
        r64 timeElapsed = Plat.GetTime() - record->startTime;
        EtcAtomicAddReal64(&record->timeElapsed, timeElapsed);
        uint64 delta = (EtcCycleCount() - startCycles) | ((uint64)hitCount << 32);
        EtcAtomicAdd(&record->hitAndCycleCount, delta);

        DebugMem->recordStackDepth--;
    }
};

#endif

