
const char *ComponentTypeNameFromComponentType(ComponentType type) {
    if (type > ENUM_VALUECOUNT(ComponentType)) {
        return NULL;
    }

    return ComponentTypeNameTable[type];
}

Schema *ComponentSchemaFromComponentType(ComponentType type) {
    if (const char *schemaName = ComponentTypeNameFromComponentType(type)) {
        return Game.GetSchema(schemaName);
    }

    return NULL;
}


#define GENERATE_COMPONENT_FIELD(name) name##Component name;

union ComponentValue {
    FOREACH_COMPONENT(GENERATE_COMPONENT_FIELD);
};

ComponentValue GetComponentValueAutocast_(EntityDB *entityDB, ComponentType type, EntityID entityID) {
    ComponentValue result = {};

    b32 foundComponentValue = GetComponentValue_(entityDB, type, entityID, &result);
    ASSERT(foundComponentValue);

    return result;
}

b32 GetComponentValueTest_(EntityDB *entityDB, ComponentType type, EntityID entityID, ComponentValue *componentValue) {
    b32 foundComponentValue = GetComponentValue_(entityDB, type, entityID, componentValue);
    return foundComponentValue;
}



#define ALLOCATE_SYSTEM(name, systemsList) \
    (systemsList)[(uint32)ComponentType_##name] = (name##System *)Plat.AllocateMem(sizeof(name##System)); \
    Plat.ClearMem((systemsList)[(uint32)ComponentType_##name], sizeof(name##System)); \
    (systemsList)[(uint32)ComponentType_##name]->type = ComponentType_##name;

void AllocateEntityDB(EntityDB *entityDB) {
    entityDB->systemCount = ENUM_VALUECOUNT(ComponentType);
    entityDB->systems = (EtcSystem **)Plat.AllocateMem(sizeof(EtcSystem *) * entityDB->systemCount);
    entityDB->systems[0] = NULL;

    entityDB->entitiesToDelete = MakeDynamicArray<EntityID>(256);

    entityDB->nextTransientID = ((uint64)1) << 63;

    FOREACH_COMPONENT_ARGS(ALLOCATE_SYSTEM, entityDB->systems);
}

void DeallocateEntityDB(EntityDB *entityDB) {
    // TODO: neeed deinit functions for systems so they don't leak memory
}

inline bool EntityIDIsTransient(EntityID id) {
    return ((id & (uint64)1 << 63) != 0);
}

EntityID PushEntityID(EntityDB *entityDB) {
    EntityID result;
    if (entityDB->entityCount == 0) {
        entityDB->nextID = 1;
    }

    result = entityDB->nextID++;
    
    // make sure nextID has not wrapped around
    ASSERT(entityDB->nextID != EntityID_Invalid);
    ASSERT(!EntityIDIsTransient(result));
    entityDB->entityCount++;
    
    return result;
}

EntityID PushTransientEntityID(EntityDB *entityDB) {
    EntityID result;

    result = entityDB->nextTransientID++;
    entityDB->entityCount++;
    entityDB->transientEntityCount++;
    
    return result;
}


void DeleteEntity(EntityDB *entityDB, EntityID entityID) {
    ASSERT(entityID != EntityID_Invalid);
    
    if (EntityIDIsTransient(entityID)) {
        ASSERT(entityID < entityDB->nextTransientID);        
    }
    else {
        ASSERT(entityID < entityDB->nextID);
        ASSERT(!EntityIsDeleted(entityDB, entityID));
    }

    PushBack(&entityDB->entitiesToDelete, entityID);

    ASSERT(entityDB->entitiesToDelete.count <= entityDB->entityCount);
}

void DeleteEntityTree(EntityDB *entityDB, EntityID entityID) {
    
    EntityIterator iter(entityDB, entityID);
    EntityID childID;
    while (GetNextEntity(&iter, &childID)) {
        DeleteEntity(entityDB, childID);
    }

    DeleteEntity(entityDB, entityID);
}

void ProcessDeletedEntities(EntityDB *entityDB) {
    for (int ii = 0; ii < entityDB->entitiesToDelete.count; ii++) {
        EntityID entityID = entityDB->entitiesToDelete[ii];

        RemoveAllComponents(entityDB, entityID);

        DynamicArray<EntityIDRange> *deletedRanges = &entityDB->deletedRanges;

        if (EntityIDIsTransient(entityID)) {
            deletedRanges = &entityDB->deletedTransientRanges;
            entityDB->transientEntityCount--;
        }
    
        b32 deletedID = false;

        for (int jj = 0; jj < deletedRanges->count; jj++) {
            EntityIDRange *range = &(*deletedRanges)[jj];

            if (entityID < range->base) {
                if (entityID == (range->base - 1)) {
                    range->base = entityID;
                    range->count += 1;
                    deletedID = true;
                }
                
                if (jj > 0) {
                    EntityIDRange *prevRange = &(*deletedRanges)[jj - 1];
                    if (entityID == (prevRange->base + prevRange->count)) {
                        if (deletedID) {
                            prevRange->count += range->count;
                            RemoveAtIndex(deletedRanges, jj);
                        }
                        else {
                            prevRange->count += 1;
                        }

                        deletedID = true;
                    }
                }

                if (!deletedID) {
                    EntityIDRange newRange;
                    newRange.base = entityID;
                    newRange.count = 1;
                    InsertAtIndex(deletedRanges, jj, newRange);
                    deletedID = true;
                }

                break;
            }
        }

        if (!deletedID) {
            EntityIDRange newRange;
            newRange.base = entityID;
            newRange.count = 1;
            PushBack(deletedRanges, newRange);
            deletedID = true;
        }

        entityDB->entityCount--;
    }

    DynamicArrayClear(&entityDB->entitiesToDelete);
}

void UndeleteEntity(EntityDB *entityDB, EntityID entityID) {

    for (int i = 0; i < entityDB->deletedRanges.count; i++) {
        EntityIDRange *range = &entityDB->deletedRanges[i];

        if (range->base <= entityID && entityID < (range->base + range->count)) {
            if (entityID == range->base) {
                range->base = range->base + 1;
                range->count--;

                if (range->count == 0) {
                    RemoveAtIndex(&entityDB->deletedRanges, i);
                }
                
                return;
            }
            else if (entityID == (range->base + range->count - 1)) {
                range->count--;

                if (range->count == 0) {
                    RemoveAtIndex(&entityDB->deletedRanges, i);
                }

                return;
            }
            else {
                EntityIDRange newRange;
                newRange.base = range->base;
                newRange.count = entityID - newRange.base;

                range->base = entityID + 1;
                range->count = range->count - newRange.count - 1;

                InsertAtIndex(&entityDB->deletedRanges, i, newRange);
                return;
            }
        }
    }

    // @NOTE: should have found the entity in the deleted ranges!
    ASSERT(false);
}

b32 EntityIsDeleted(EntityDB *entityDB, EntityID entityID) {
    for (int i = 0; i < entityDB->deletedRanges.count; i++) {
        EntityIDRange range = entityDB->deletedRanges[i];

        if (range.base <= entityID && entityID < (range.base + range.count)) {
            return true;
        }
    }

    return false;
}

b32 EntityMarkedForDeletion(EntityDB *entityDB, EntityID entityID) {
    for (int i = 0; i < entityDB->entitiesToDelete.count; i++) {
        if (entityDB->entitiesToDelete[i] == entityID) {
            return true;
        }
    }

    return false;
}


EntityID SpawnEntity(EntityDB *entityDB, const char *name, vec3 position, quaternion rotation, vec3 scale) {
    EntityID id = PushEntityID(entityDB);

    if (name != NULL) {
        NameComponent nameComponent = {};
        StringCopy(nameComponent.name, ARRAY_LENGTH(char, nameComponent.name), name);
        AddComponent(entityDB, Name, 1, &id, &nameComponent);
    }

    TransformComponent transformComponent = {};
    transformComponent.position = position;
    transformComponent.rotation = rotation;
    transformComponent.scale = scale;
    AddComponent(entityDB, Transform, 1, &id, &transformComponent);

    return id;
}

void AddComponent_(EntityDB *entityDB, ComponentType type, uint32 count, EntityID *entityIDs, ComponentValue *components) {
    ASSERT(type != ComponentType_Invalid);
    ASSERT(entityIDs != NULL);
    
    EtcSystem *system = entityDB->systems[type];
    AddComponentFunc *addComponent = Game.components.addComponentFuncs[type];
    if (addComponent != NULL) {
        // @HACK: we dont want to ever add a component for EntityID_Invalid and we dont want boilerplate
        // in every function to check for it, so here we just add one entity at a time
        if (components == NULL) {
            for (int i = 0; i < count; i++) {
                if (entityIDs[i] == EntityID_Invalid) {
                    continue;
                }
                addComponent(entityDB, system, 1, &entityIDs[i]);    
            }
        }

        if (components != NULL) {
            SetComponentValueFunc *setComponentValue = Game.components.setComponentValueFuncs[type];
            for (int i = 0; i < count; i++) {
                if (entityIDs[i] == EntityID_Invalid) {
                    continue;
                }
                ComponentValue temp;
                if (!GetComponentValue_(entityDB, type, entityIDs[i], &temp)) {
                    addComponent(entityDB, system, 1, &entityIDs[i]);    
                }
                setComponentValue(entityDB, system, entityIDs[i], components + i);
            }
        }
    }
}

void RemoveComponent(EntityDB *entityDB, ComponentType type, EntityID entityID) {
    ASSERT(type != ComponentType_Invalid);

    if (entityID == EntityID_Invalid) {
        return;
    }

    EtcSystem *system = entityDB->systems[type];
    RemoveComponentFunc *removeComponent = Game.components.removeComponentFuncs[type];
    if (removeComponent != NULL) {
        removeComponent(entityDB, system, entityID);
    }
}

void RemoveAllComponents(EntityDB *entityDB, EntityID entityID) {
    ASSERT(entityID != EntityID_Invalid);

    for (int type = 1; type < ENUM_VALUECOUNT(ComponentType); type++) {
        RemoveComponent(entityDB, (ComponentType)type, entityID);
    }
}

b32 GetComponentValue_(EntityDB *entityDB, ComponentType type, EntityID entityID, ComponentValue *component) {
    ASSERT(type != ComponentType_Invalid);
    ASSERT(component != NULL);

    if (entityID == EntityID_Invalid) {
        return false;
    }

    b32 result = false;

    GetComponentValueFunc *getComponentValue = Game.components.getComponentValueFuncs[type];
    EtcSystem *system = entityDB->systems[type];
    if (getComponentValue != NULL && getComponentValue(entityDB, system, entityID, component)) {
        result = true;
    }
    
    return result;
}

void SetComponentValue_(EntityDB *entityDB, ComponentType type, EntityID entityID, ComponentValue *component) {
    ASSERT(type != ComponentType_Invalid);
    ASSERT(component != NULL);

    if (entityID == EntityID_Invalid) {
        return;
    }

    SetComponentValueFunc *setComponentValue = Game.components.setComponentValueFuncs[type];
    EtcSystem *system = entityDB->systems[type];
    if (setComponentValue != NULL) {
        setComponentValue(entityDB, system, entityID, component);
    }
}

EntityID CopyEntity(EntityDB *entityDB, EntityID entityID, b32 maintainParent) {
    EntityID copy = PushEntityID(entityDB);

    for (int type = 1; type < entityDB->systemCount; type++) {
        ComponentValue componentValue = {};
        if (GetComponentValue_(entityDB, (ComponentType)type, entityID, &componentValue)) {

            if (type == ComponentType_Name) {
                NameComponent *nameComponent = &componentValue.Name;

                char *idName = PushArray(Game.frameMem, char, 32);
                sprintf(idName, "(%" PRIu64 ")", copy);
                const char *newName = SplitString(Game.frameMem, nameComponent->name, '(');

                Plat.ClearMem(nameComponent->name, MAX_ENTITY_NAME_LENGTH);
                
                newName = StringConcat(Game.frameMem, 2, newName, idName);
                Plat.CopyMem(nameComponent->name, newName, StringLength(newName));
            }
            if (type == ComponentType_Transform) {
                TransformComponent *transformComponent = &componentValue.Transform;

                // @NOTE: clear these so that when we add the component we dont modify the tree for the
                // as-yet uncopied children
                if (!maintainParent) {
                    transformComponent->parent = EntityID_Invalid;
                }
                transformComponent->firstChild = EntityID_Invalid;
                transformComponent->nextSibling = EntityID_Invalid;
                transformComponent->prevSibling = EntityID_Invalid;

                AddComponent_(entityDB, (ComponentType)type, 1, &copy, &componentValue);
                
                TransformSystem *transSys = GetSystem(entityDB, Transform);

                s32 index;
                HashTableGet(&transSys->entityMap, entityID, &index);

                s32 firstChildIndex = transSys->firstChildIndices[index];

                if (firstChildIndex >= 0) {
                    EntityID firstChild = transSys->entityIDs[firstChildIndex];

                    u32 childCount = 0;
                    EntityID *children = GetChildren(entityDB, transSys, entityID, &childCount);

                    EntityID *copiedChildren = PushArray(Game.frameMem, EntityID, childCount);

                    for (int i = 0; i < childCount; i++) {
                        TransformComponent childTransform = GetComponentValue(entityDB, Transform, children[i]);
                        EntityID copiedChild = CopyEntity(entityDB, children[i], false);
                        TransformComponent copiedChildTransform = GetComponentValue(entityDB, Transform, copiedChild);
                        copiedChildren[i] = copiedChild;
                    }

                    for (int i = 0; i < childCount; i++) {
                        TransformComponent childTransform = GetComponentValue(entityDB, Transform, copiedChildren[i]);
                        childTransform.parent = copy;

                        // @NOTE: just set the parent and the tree should reform itself
                        SetComponentValue(entityDB, Transform, copiedChildren[i], &childTransform);
                    }

                    TransformComponent copyTransform = GetComponentValue(entityDB, Transform, copy);
                }
            }
            else {
                AddComponent_(entityDB, (ComponentType)type, 1, &copy, &componentValue);
            }
        }
    }

    return copy;
}


void TickSystems(EntityDB *entityDB, EtcGameMemory *mem) {
    for (int i = 0; i < entityDB->systemCount; i++) {
        EtcSystem *system = entityDB->systems[i];

        TickSystemFunc *tick = Game.components.tickSystemFuncs[i];

        if (tick != NULL) {
            tick(mem, entityDB, system);
        }
    }
}


EntityIterator::EntityIterator(EntityDB *entityDB, EntityID rootID, b32 recursive)
    : entityDB(entityDB),
      rootID(rootID),
      rootIndex(-1),
      currentIndex(-1),
      recursive(recursive) {

    transformSystem = GetSystem(entityDB, Transform);
    HashTableGet(&transformSystem->entityMap, rootID, &rootIndex);

    currentID = rootID;
    currentIndex = rootIndex;

    skipNextSubtree = false;
}

EntityIterator::EntityIterator(struct TransformSystem *transformSystem, EntityID rootID, b32 recursive)
    : transformSystem(transformSystem),
      rootID(rootID),
      rootIndex(-1),
      currentIndex(-1),
      recursive(recursive) {

    HashTableGet(&transformSystem->entityMap, rootID, &rootIndex);
    ASSERT(rootIndex >= 0);

    currentID = rootID;
    currentIndex = rootIndex;

    skipNextSubtree = false;
}

EntityID GetFirstEntity(EntityIterator *iterator) {
    EntityDB *entityDB = iterator->entityDB;
    TransformSystem *transformSystem = iterator->transformSystem;

    if (iterator->rootID == EntityID_Invalid) {
        ASSERT(entityDB != NULL);

        // Take the next ID that doesn't have a parent
        for (uint64 id = 1; id < entityDB->nextID; id++) {
            if (EntityIsDeleted(entityDB, id)) {
                continue;
            }

            int32 index;
            if (HashTableGet(&transformSystem->entityMap, iterator->rootID, &index)) {
                if (transformSystem->parentIndices[index] >= 0) {
                    continue;
                }
            }

            return id;
        }

        return EntityID_Invalid;
    }

    if (iterator->rootIndex < 0) {
        return EntityID_Invalid;
    }

    int32 firstChildIndex = transformSystem->firstChildIndices[iterator->rootIndex];
    if (firstChildIndex < 0) {
        return EntityID_Invalid;
    }

    return transformSystem->entityIDs[firstChildIndex];
}

EntityID GetLastEntity(EntityIterator *iterator) {
    EntityID id = GetFirstEntity(iterator);

    if (id == EntityID_Invalid) {
        return EntityID_Invalid;
    }

    EntityID nextID = id;
    while (GetNextEntity(iterator, &nextID)) {
        id = nextID;
    }

    return id;
}

b32 GetNextEntity(EntityIterator *iterator, EntityID *id) {
    EntityDB *entityDB = iterator->entityDB;
    TransformSystem *transformSystem = iterator->transformSystem;

    b32 result = false;

    // Set up the iterator the first time it's called
    if (iterator->currentID == iterator->rootID) {
        if (!iterator->skipNextSubtree) {
            iterator->currentID = GetFirstEntity(iterator);
            HashTableGet(&transformSystem->entityMap, iterator->currentID, &iterator->currentIndex);

            result = (iterator->currentID != EntityID_Invalid);
        }

        goto end;
    }

    if (iterator->currentIndex >= 0) {
        if (iterator->recursive && !iterator->skipNextSubtree) {
            int32 firstChildIndex = transformSystem->firstChildIndices[iterator->currentIndex];
            if (firstChildIndex >= 0) {
                iterator->currentIndex = firstChildIndex;
                iterator->currentID = transformSystem->entityIDs[firstChildIndex];
                result = true;
                goto end;
            }
        }

        int32 nextSiblingIndex = transformSystem->nextSiblingIndices[iterator->currentIndex];
        if (nextSiblingIndex >= 0) {
            iterator->currentIndex = nextSiblingIndex;
            iterator->currentID = transformSystem->entityIDs[nextSiblingIndex];
            result = true;
            goto end;
        }

        if (!iterator->recursive && iterator->rootID != EntityID_Invalid) {
            iterator->currentIndex = -1;
            iterator->currentID = EntityID_Invalid;
            goto end;
        }

        if (iterator->recursive) {
            int32 parentIndex = transformSystem->parentIndices[iterator->currentIndex];

            int32 parentID = EntityID_Invalid;
            if (parentIndex >= 0) {
                parentID = transformSystem->entityIDs[parentIndex];
            }

            while (parentIndex >= 0 && parentID != iterator->rootID) {
                int32 nextSiblingIndex = transformSystem->nextSiblingIndices[parentIndex];
                if (nextSiblingIndex >= 0) {
                    iterator->currentIndex = nextSiblingIndex;
                    iterator->currentID = transformSystem->entityIDs[nextSiblingIndex];
                    result = true;
                    goto end;
                }

                iterator->currentID = parentID;
                parentIndex = transformSystem->parentIndices[parentIndex];

                if (parentIndex >= 0) {
                    parentID = transformSystem->entityIDs[parentIndex];    
                }
            }

            if (iterator->rootID != EntityID_Invalid) {
                iterator->currentIndex = -1;
                iterator->currentID = EntityID_Invalid;
                goto end;
            }
        }
    }

    ASSERT(entityDB != NULL);

    // Find next top level entity
    iterator->currentIndex = -1;

    for (iterator->currentID = iterator->currentID + 1; iterator->currentID < entityDB->nextID; iterator->currentID++) {
        if (EntityIsDeleted(entityDB, iterator->currentID)) {
            continue;
        }

        int32 index;
        if (HashTableGet(&transformSystem->entityMap, iterator->currentID, &index)) {
            if (transformSystem->parentIndices[index] >= 0) {
                continue;
            }

            iterator->currentIndex = index;
        }

        result = true;
        goto end;
    }

    iterator->currentID = EntityID_Invalid;

 end:
    iterator->skipNextSubtree = false;
    *id = iterator->currentID;

    return result;
}

int32 FindLastChildIndex_(TransformSystem *transformSystem, int32 rootIndex) {
    int32 lastChildIndex = transformSystem->firstChildIndices[rootIndex];

    while (lastChildIndex >= 0) {
        int32 nextSiblingIndex = transformSystem->nextSiblingIndices[lastChildIndex];
        if (nextSiblingIndex >= 0) {
            lastChildIndex = nextSiblingIndex;
            continue;
        }

        int32 firstChildIndex = transformSystem->firstChildIndices[lastChildIndex];
        if (firstChildIndex >= 0) {
            lastChildIndex = firstChildIndex;
            continue;
        }

        return lastChildIndex;
    }

    return -1;
}

b32 GetPrevEntity(EntityIterator *iterator, EntityID *id) {
    EntityDB *entityDB = iterator->entityDB;
    TransformSystem *transformSystem = iterator->transformSystem;

    b32 result = false;

    // Set up the iterator the first time it's called
    if (iterator->currentID == iterator->rootID) {
        iterator->currentID = GetLastEntity(iterator);
        HashTableGet(&transformSystem->entityMap, iterator->currentID, &iterator->currentIndex);

        result = (iterator->currentID != EntityID_Invalid);
        goto end;
    }

    if (iterator->currentIndex >= 0) {
        if (iterator->recursive) {
            int32 prevSiblingIndex = transformSystem->prevSiblingIndices[iterator->currentIndex];
            if (prevSiblingIndex >= 0) {
                int32 lastChildIndex = FindLastChildIndex_(transformSystem, prevSiblingIndex);
                if (lastChildIndex >= 0) {
                    iterator->currentIndex = lastChildIndex;
                    iterator->currentID = transformSystem->entityIDs[lastChildIndex];
                    result = true;
                    goto end;
                }

                iterator->currentIndex = prevSiblingIndex;
                iterator->currentID = transformSystem->entityIDs[prevSiblingIndex];
                result = true;
                goto end;
            }
        }

        if (!iterator->recursive && iterator->rootID != EntityID_Invalid) {
            iterator->currentIndex = -1;
            iterator->currentID = EntityID_Invalid;
            goto end;
        }

        if (iterator->recursive) {
            int32 parentIndex = transformSystem->parentIndices[iterator->currentIndex];
            if (parentIndex >= 0) {
                iterator->currentIndex = parentIndex;
                iterator->currentID = transformSystem->entityIDs[parentIndex];
                result = true;
                goto end;
            }
        }
    }

    ASSERT(entityDB != NULL);

    // Find prev top level entity
    for (iterator->currentID = iterator->currentID - 1; iterator->currentID > 0; iterator->currentID--) {
        if (EntityIsDeleted(entityDB, iterator->currentID)) {
            continue;
        }

        int32 index;
        if (HashTableGet(&transformSystem->entityMap, iterator->currentID, &index)) {
            if (transformSystem->parentIndices[index] >= 0) {
                continue;
            }

            if (iterator->recursive) {
                int32 lastChildIndex = FindLastChildIndex_(transformSystem, iterator->currentID);
                if (lastChildIndex >= 0) {
                    iterator->currentIndex = lastChildIndex;
                    iterator->currentID = transformSystem->entityIDs[lastChildIndex];
                }
                else {
                    iterator->currentIndex = index;
                }
            }
            else {
                iterator->currentIndex = index;
            }
        }

        result = true;
        goto end;
    }

    iterator->currentID = EntityID_Invalid;

 end:
    *id = iterator->currentID;
    return result;
}

b32 GetNextEntityWithComponent(EntityIterator *iterator, ComponentType type, EntityID *id, ComponentValue *value) {

    // @BUG: can't pass NULL value to GetComponentValue_()
    ComponentValue tempComp = {};
    if (value == NULL) {
        value = &tempComp;
    }

    while (GetNextEntity(iterator, id)) {
        if (GetComponentValue_(iterator->entityDB, type, *id, value)) {
            return true;
        }
    }

    return false;
}

b32 GetPrevEntityWithComponent(EntityIterator *iterator, ComponentType type, EntityID *id, ComponentValue *value) {

    // @BUG: can't pass NULL value to GetComponentValue_()
    ComponentValue tempComp = {};
    if (value == NULL) {
        value = &tempComp;
    }

    while (GetPrevEntity(iterator, id)) {
        if (GetComponentValue_(iterator->entityDB, type, *id, value)) {
            return true;
        }
    }

    return false;
}

void SkipNextSubtree(EntityIterator *iterator, b32 shouldSkip) {
    iterator->skipNextSubtree = shouldSkip;
}

void SetCurrentEntity(EntityIterator *iterator, EntityID id) {
    ASSERT(id != EntityID_Invalid);

    iterator->currentID = id;
    HashTableGet(&iterator->transformSystem->entityMap, iterator->currentID, &iterator->currentIndex);
}

b32 FindAncestorWithComponent(EntityDB *entityDB, ComponentType type, EntityID id, EntityID *ancestorID, ComponentValue *value) {
    ComponentValue tempComp = {};
    TransformSystem *transformSystem = GetSystem(entityDB, Transform);

    int32 index;
    if (HashTableGet(&transformSystem->entityMap, id, &index)) {
        if (index < 0) {
            return false;
        }

        int32 parentIndex = transformSystem->parentIndices[index];
        while (parentIndex >= 0) {
            EntityID parentID = transformSystem->entityIDs[parentIndex];
            if (GetComponentValue_(entityDB, type, parentID, &tempComp)) {

                if (ancestorID != NULL) {
                    *ancestorID = parentID;
                }

                if (value != NULL) {
                    *value = tempComp;
                }
                
                return true;
            }

            parentIndex = transformSystem->parentIndices[parentIndex];
        }
    }

    return false;
}

b32 RaycastAllowedForType_(RaycastSceneFilter *filters, uint32 filterCount, RaycastSceneResult currentResult, ComponentType type) {

    b32 result = true;
    b32 mustMatchARequiredType = false;
    b32 matchesARequiredType = false;

    for (int i = 0; i < filterCount; i++) {
        RaycastSceneFilter filter = filters[i];
        switch (filter.filterType) {
            case RaycastSceneFilter_RequireType: {
                mustMatchARequiredType = true;

                if (filter.componentType == type) {
                    matchesARequiredType = true;
                }
            } break;

            case RaycastSceneFilter_ExcludeType: {
                if (filter.componentType == type) {
                    result = false;
                    goto end;
                }
            } break;

            case RaycastSceneFilter_PreferType: {
                // The currentResult already has a type of higher precedence
                if (currentResult.componentType == filter.componentType) {
                    result = false;
                    goto end;
                }
            } break;
        }
    }

 end:
    if (mustMatchARequiredType && !matchesARequiredType) {
        return false;
    }

    return result;
}

b32 RaycastAllowed_(RaycastSceneFilter *filters, uint32 filterCount, RaycastSceneResult currentResult, RaycastSceneResult candidateResult) {

    for (int i = 0; i < filterCount; i++) {
        RaycastSceneFilter filter = filters[i];
        switch (filter.filterType) {
            case RaycastSceneFilter_PreferType: {
                // Always take the candidate if it is more "preferred"
                if (candidateResult.componentType == filter.componentType) {
                    return true;
                }
            } break;

            default: {
            } break;
        }
    }

    return candidateResult.t < currentResult.t;
}

uint32 RaycastScene(EntityDB *entityDB, Ray ray, RaycastSceneResult *closestResult, ChunkedArray<RaycastSceneResult> *results, uint32 filterCount, RaycastSceneFilter *filters) {

    b32 intersects = false;
    uint32 intersectCount = 0;

    RaycastSceneResult closest = {};
    closest.t = INFINITY;

    for (int i = 1; i < ENUM_VALUECOUNT(ComponentType); i++) {
        ComponentType type = (ComponentType)i;
        if (!RaycastAllowedForType_(filters, filterCount, closest, type)) {
            continue;
        }

        RaycastSystemFunc *raycastSystem = Game.components.raycastSystemFuncs[type];
        if (raycastSystem != NULL) {
            RaycastSceneResult closestSystemResult = {};
            uint32 systemIntersectCount = raycastSystem(entityDB, entityDB->systems[i], ray, &closestSystemResult, results);
            if (systemIntersectCount > 0) {
                intersectCount += systemIntersectCount;

                if (RaycastAllowed_(filters, filterCount, closest, closestSystemResult)) {
                    closest = closestSystemResult;
                }
            }
        }
    }

    if (intersectCount > 0 && closestResult != NULL) {
        *closestResult = closest;
    }

    return intersectCount;
}


b32 SceneGroundPointAtXZ(EntityDB *entityDB, vec2 point, vec3 *result, RaycastSceneFilter *filters, uint32 filterCount) {
    return false;

    //b32 intersects = false;

    //*result = V3(point.x, 0.0f, point.y);

    //Ray ray;
    //ray.origin = V3(point.x, 128000.0f, point.y);
    //ray.direction = V3(0, -1, 0);

    //RaycastSceneResult raycastResult;
    //if (RaycastScene(entityDB, ray, &raycastResult, filter, filterType)) {
    //    *result = PointAt(ray, raycastResult.t);
    //    intersects = true;
    //}
    //return intersects;
}

b32 SceneGroundPointAtXZ(EntityDB *entityDB, vec2 point, vec3 *result, RaycastSceneFilter filter) {
    return SceneGroundPointAtXZ(entityDB, point, result, &filter, 1);
}

// NameComponent
void AddNameComponent(EntityDB *entityDB, NameSystem *system, uint32 count, EntityID *entityIDs) {
    if (system->entityMap.capacity == 0) {
        AllocateHashTable(&system->entityMap, 1024);
        AllocateHashTable(&system->uniqueNameMap, 1024);
    }

    for (int i = 0; i < count; i++) {
        if (!HashTableContains(&system->entityMap, entityIDs[i])) {
            HashTableInsert(&system->entityMap, entityIDs[i], system->names.count, true);
            AddEntityToSystem(system, entityIDs[i]);

            NameComponent *name = PushBackPtr(&system->names);
            Plat.ClearMem(name->name, MAX_ENTITY_NAME_LENGTH);    
        }
    }
}

void RemoveNameComponent(EntityDB *entityDB, NameSystem *system, EntityID entityID) {
    uint32 index;
    if (HashTableGet(&system->entityMap, entityID, &index)) {
        NameComponent comp = system->names[index];
        uint32 uniqueIndex;

        if (HashTableGet(&system->uniqueNameMap, comp.name, &uniqueIndex)) {
            UniqueName unique = system->uniqueNames[uniqueIndex];

            if (unique.count == 1) {
                RemoveAtIndexBySwap(&system->uniqueNames, uniqueIndex);

                HashTableRemove(&system->uniqueNameMap, unique.name);
                
                if (uniqueIndex < system->entityIDs.count) {
                    HashTableInsert(&system->uniqueNameMap, system->uniqueNames[uniqueIndex].name, uniqueIndex);
                }
            }
            else {
                system->uniqueNames[uniqueIndex].count--;
            }
        }

        
        RemoveAtIndexBySwap(&system->entityIDs, index);
        RemoveAtIndexBySwap(&system->names, index);

        HashTableRemove(&system->entityMap, entityID);

        if (index < system->entityIDs.count) {
            HashTableInsert(&system->entityMap, system->entityIDs[index], index);
        }
    }
}

b32 GetNameComponentValue(EntityDB *entityDB, NameSystem *system, EntityID entityID, NameComponent *component) {
    b32 result = false;

    uint32 index;
    if (HashTableGet(&system->entityMap, entityID, &index)) {
        *component = system->names[index];
        result = true;
    }

    return result;
}

void SetNameInstanceString(uint32 instance, char *instanceStr) {
    stbsp_sprintf(instanceStr, "_%d", instance);
}

void SetNameComponentValue(EntityDB *entityDB, NameSystem *system, EntityID entityID, NameComponent *component) {
    uint32 index;
    if (HashTableGet(&system->entityMap, entityID, &index)) {

        uint32 uniqueIndex;
        if (!HashTableGet(&system->uniqueNameMap, component->name, &uniqueIndex)) {
            component->instanceName[0] = 0;
            
            UniqueName unique = {};
            unique.name = StringCopy(component->name);
            unique.count = 1;
            unique.instance = 1;
            uniqueIndex = PushBack(&system->uniqueNames, unique);
            HashTableInsert(&system->uniqueNameMap, unique.name, uniqueIndex);
        }
        else {
            SetNameInstanceString(system->uniqueNames[uniqueIndex].instance, component->instanceName);
            system->uniqueNames[uniqueIndex].count++;
            system->uniqueNames[uniqueIndex].instance++;
        }

        system->names[index] = *component;
    }
}

EntityID EntityNamed(NameSystem *system, const char *name, int32 *startAt) {
    EntityID result = EntityID_Invalid;

    if (strlen(name) == 0) {
        return result;
    }

    int start = 0;
    if (startAt != NULL) {
        start = *startAt;
    }

    for (int i = start; i < system->names.count; i++) {
        NameComponent comp = system->names[i];
        uint32 nameLength = StringLength(comp.name);
        if (StringEquals(comp.name, name, nameLength) &&
            StringEquals(comp.instanceName, name + nameLength)) {
            //if (StringEquals(system->names[i].name, name)) {
            result = system->entityIDs[i];

            if (startAt != NULL) {
                *startAt = i + 1;
            }
            break;
        }
    }

    return result;
}

char *GetEntityName(NameSystem *system, EntityID id) {
    uint32 index;
    if (HashTableGet(&system->entityMap, id, &index)) {
        return StringConcat(Game.frameMem, 2, system->names[index].name, system->names[index].instanceName);
    }

    return NULL;
}


// TagComponent
void AddTagComponent(EntityDB *entityDB, TagSystem *system, uint32 count, EntityID *entityIDs) {
    if (system->entityMap.capacity == 0) {
        AllocateHashTable(&system->entityMap, 1024);
    }

    for (int i = 0; i < count; i++) {
        if (!HashTableContains(&system->entityMap, entityIDs[i])) {
            HashTableInsert(&system->entityMap, entityIDs[i], system->tags.count, true);
            AddEntityToSystem(system, entityIDs[i]);

            PushBack(&system->tags, {});
        }
    }
}

void RemoveTagComponent(EntityDB *entityDB, TagSystem *system, EntityID entityID) {
    uint32 index;
    if (HashTableGet(&system->entityMap, entityID, &index)) {
        RemoveAtIndexBySwap(&system->entityIDs, index);
        RemoveAtIndexBySwap(&system->tags, index);

        HashTableRemove(&system->entityMap, entityID);

        if (index < system->entityIDs.count) {
            HashTableInsert(&system->entityMap, system->entityIDs[index], index);
        }
    }
}

b32 GetTagComponentValue(EntityDB *entityDB, TagSystem *system, EntityID entityID, TagComponent *component) {
    b32 result = false;

    uint32 index;
    if (HashTableGet(&system->entityMap, entityID, &index)) {
        *component = system->tags[index];
        result = true;
    }

    return result;
}

void SetTagComponentValue(EntityDB *entityDB, TagSystem *system, EntityID entityID, TagComponent *component) {
    uint32 index;
    if (HashTableGet(&system->entityMap, entityID, &index)) {
        system->tags[index] = *component;
    }
}

void SetEntityTag(TagSystem *system, EntityID id, Tag tag) {
    uint32 index;
    if (HashTableGet(&system->entityMap, id, &index)) {
        system->tags[index].tag = tag;
    }
}

uint32 EntitiesTagged(TagSystem *system, Tag tag, EntityID *entityIDs, uint32 entityCount) {
    uint32 foundCount = 0;
    for (int i = 0; i < system->entityIDs.count; i++) {
        TagComponent comp = system->tags[i];
        if (HasFlag(comp.tag, tag) && foundCount < entityCount) {
            entityIDs[foundCount++] = system->entityIDs[i];
        }
    }

    return foundCount;
}

void GetEntitiesTagged(TagSystem *system, Tag tag, ChunkedArray<EntityID> *entityIDs) {
    for (int i = 0; i < system->entityIDs.count; i++) {
        TagComponent comp = system->tags[i];
        if (HasFlag(comp.tag, tag)) {
            PushBack(entityIDs, system->entityIDs[i]);
        }
    }
}


// the bigger the value the better
void FindBestEntityWithTag(EntityDB *entityDB, EntityHeuristicFunc *func, Tag tag, void *data, EntityID *bestID, real32 *bestValue, void *bestDataOut, uint32 dataOutSize) {
    TagSystem *tagSys = GetSystem(entityDB, Tag);
    
    real32 maxValue = -INFINITY;

    void *dataOut = PushSize(Game.frameMem, dataOutSize);
    
    for (int i = 0; i < tagSys->entityIDs.count; i++) {
        TagComponent comp = tagSys->tags[i];

        // @GACK: @NOTE: this is because our tags are not bitfields, but enums. This might change.
        if (comp.tag == tag) {
        //if (HasFlag(comp.tag, tag)) {
            EntityID id = tagSys->entityIDs[i];
            
            real32 value = func(entityDB, id, data, *bestID, bestDataOut, dataOut, dataOutSize);
            
            if (value > maxValue) {
                maxValue = value;
                *bestID = id;
                *bestValue = value;
                Plat.CopyMem(bestDataOut, dataOut, dataOutSize);
            }
        }
    }
}
