
#ifndef _ETC_ECS_H
#define _ETC_ECS_H

const EntityID EntityID_Invalid = 0;

enum ComponentType : uint32;

struct EtcSystem {
    ComponentType type;
    DynamicArray<EntityID> entityIDs;
    // @HASHTABLE?
};

struct EntityDB;

TAGGED_UNION(EntityIdentifier, {
    EntityID id;
    char *name;
});


inline void AddEntityToSystem(EtcSystem *system, EntityID id) {
    PushBack(&system->entityIDs, id);
}

inline void RemoveEntityFromSystem(EtcSystem *system, int32 index) {
    RemoveAtIndexBySwap(&system->entityIDs, index);
}

#include "component_list.generated.h"

const char *ComponentTypeNameFromComponentType(ComponentType type);
Schema *ComponentSchemaFromComponentType(ComponentType type);

union ComponentValue;

// EntityDB
STRUCT()
struct EntityIDRange {
    EntityID base;
    uint64 count;
};

struct EntityDB {
    EntityID nextID;
    DynamicArray<EntityIDRange> deletedRanges;
    DynamicArray<EntityIDRange> deletedTransientRanges;

    EntityID nextTransientID;

    DynamicArray<EntityID> entitiesToDelete;

    uint64 transientEntityCount;
    uint64 entityCount;
    
    uint32 systemCount;
    EtcSystem **systems;
};

void AllocateEntityDB(EntityDB *entityDB);
void DeallocateEntityDB(EntityDB *entityDB);

EntityID PushEntityID(EntityDB *entityDB);
EntityID PushTransientEntityID(EntityDB *entityDB);
void DeleteEntity(EntityDB *entityDB, EntityID entityID);
b32 EntityIsDeleted(EntityDB *entityDB, EntityID entityID);


// @TODO: more overloads!!
EntityID SpawnEntity(EntityDB *entityDB, const char *name, vec3 position, quaternion rotation, vec3 scale);

void AddComponent_(EntityDB *entityDB, ComponentType type, uint32 count, EntityID *entityIDs, ComponentValue *components = NULL);
void RemoveComponent(EntityDB *entityDB, ComponentType type, EntityID entityID);
void RemoveAllComponents(EntityDB *entityDB, EntityID entityID);

b32 GetComponentValue_(EntityDB *entityDB, ComponentType type, EntityID entityID, ComponentValue *component);
void SetComponentValue_(EntityDB *entityDB, ComponentType type, EntityID entityID, ComponentValue *component);

EntityID CopyEntity(EntityDB *entityDB, EntityID entityID, b32 maintainParent);

#define GetSystem(entityDB, name) ((name##System *)(entityDB)->systems[(uint32)ComponentType_##name])

#define GetSystemEntities(entityDB, name) &((name##System *)(entityDB)->systems[(uint32)ComponentType_##name])->entityIDs

ComponentValue GetComponentValueAutocast_(EntityDB *entityDB, ComponentType type, EntityID entityID);
b32 GetComponentValueTest_(EntityDB *entityDB, ComponentType type, EntityID entityID, ComponentValue *componentValue);

#define GetComponentValue(entityDB, type, entityID) ((type##Component)GetComponentValueAutocast_(entityDB, ComponentType_##type, entityID).type)

#define HasComponentValue(entityDB, type, entityID, value) GetComponentValueTest_(entityDB, ComponentType_##type, entityID, (ComponentValue *)value)

#define SetComponentValue(entityDB, type, entityID, ptr) SetComponentValue_(entityDB, ComponentType_##type, entityID, (ComponentValue *)(ptr))

#define AddComponent(entityDB, type, count, entityIDs, components) AddComponent_(entityDB, ComponentType_##type, count, entityIDs, (ComponentValue *)components);


void TickSystems(EntityDB *entityDB, EtcGameMemory *mem);

struct EntityIterator {
    EntityDB *entityDB;
    struct TransformSystem *transformSystem;

    // A rootID of EntityID_Invalid traverses the entire entity tree
    EntityID rootID;
    int32 rootIndex;

    EntityID currentID;
    int32 currentIndex;

    b32 recursive;
    b32 skipNextSubtree;

    EntityIterator(EntityDB *entityDB, EntityID rootID = EntityID_Invalid, b32 recursive = true);
    EntityIterator(struct TransformSystem *transformSystem, EntityID rootID = EntityID_Invalid, b32 recursive = true);
};

EntityID GetFirstEntity(EntityIterator *iterator);
EntityID GetLastEntity(EntityIterator *iterator);

b32 GetNextEntity(EntityIterator *iterator, EntityID *id);
b32 GetPrevEntity(EntityIterator *iterator, EntityID *id);

b32 GetNextEntityWithComponent(EntityIterator *iterator, ComponentType type, EntityID *id, ComponentValue *value = NULL);
b32 GetPrevEntityWithComponent(EntityIterator *iterator, ComponentType type, EntityID *id, ComponentValue *value = NULL);

void SkipNextSubtree(EntityIterator *iterator, b32 shouldSkip = true);
void SetCurrentEntity(EntityIterator *iterator, EntityID id);

b32 FindAncestorWithComponent(EntityDB *entityDB, ComponentType type, EntityID id, EntityID *ancestorID = NULL, ComponentValue *value = NULL);


struct RaycastSceneResult {
    EntityID entityID;
    ComponentType componentType;

    Ray ray;
    real32 t;
    vec3 endPoint;

    vec3 normal;

    int32 index;
};

enum RaycastSceneFilterType {
    RaycastSceneFilter_RequireType,
    RaycastSceneFilter_ExcludeType,
    RaycastSceneFilter_PreferType,
};

struct RaycastSceneFilter {
    RaycastSceneFilterType filterType;
    ComponentType componentType;
};

typedef uint32 RaycastSystemFunc(EntityDB *entityDB, EtcSystem *system, Ray ray,
                                 RaycastSceneResult *closestResult,
                                 ChunkedArray<RaycastSceneResult> *results);

uint32 RaycastScene(EntityDB *entityDB, Ray ray,
                    RaycastSceneResult *closestResult = NULL,
                    ChunkedArray<RaycastSceneResult> *results = NULL,
                    uint32 filterCount = 0,
                    RaycastSceneFilter *filters = NULL);

b32 SceneGroundPointAtXZ(EntityDB *entityDB, vec2 point, vec3 *result, RaycastSceneFilter *filters = NULL, uint32 filterCount = 0);
b32 SceneGroundPointAtXZ(EntityDB *entityDB, vec2 point, vec3 *result, RaycastSceneFilter filter);


#if ETC_DEBUG

#define DEBUGSYS(sysName, _flags, ...)                                  \
    if (Game.GetSchema(DebugMem->systemDebugPreferences.systemDebugFlags[ComponentType_##sysName - 1].schemaName) != NULL && \
        HasFlags(*(uint32 *)DebugMem->systemDebugPreferences.systemDebugFlags[ComponentType_##sysName - 1].data, _flags)) { \
        __VA_ARGS__;                                                    \
    }                                                                   \

#elif
#define DEBUGSYS(sysName, ...)

#endif


DEFINE_FUNCTIONTYPE(QueryEntityFunc, bool, EntityDB *entityDB, EntityID id);


#define MAX_ENTITY_NAME_LENGTH 128

COMPONENT(Name) {
    char name[MAX_ENTITY_NAME_LENGTH];
    FIELD(Serializable = false)
    char instanceName[MAX_ENTITY_NAME_LENGTH];
};

struct UniqueName {
    const char *name;
    uint32 count;
    uint32 instance;
};

SYSTEM(Name, Priority = 0) {
    DynamicArray<NameComponent> names;

    HashTable<EntityID, uint32> entityMap;

    HashTable<const char *, uint32> uniqueNameMap;
    DynamicArray<UniqueName> uniqueNames;
};

EntityID EntityNamed(NameSystem *system, const char *name, int32 *startAt = NULL);
char *GetEntityName(NameSystem *system, EntityID id);

#define TAG(...)

COMPONENT(Tag) {
    Tag tag;
};

SYSTEM(Tag, Priority = 0) {
    // @TODO: to accelerate searching we can have an array per tag type
    DynamicArray<TagComponent> tags;

    HashTable<EntityID, uint32> entityMap;
};

void SetEntityTag(TagSystem *system, EntityID id, Tag tag);
uint32 EntitiesTagged(TagSystem *system, Tag tag, EntityID *entityIDs, uint32 entityCount);

typedef real32 EntityHeuristicFunc(EntityDB *entityDB, EntityID id, void *data, EntityID bestID, void *bestDataOut, void *dataOut, uint32 dataOutSize);
void FindBestEntityWithTag(EntityDB *entityDB, EntityHeuristicFunc *func, Tag tag, void *data, EntityID *bestID, real32 *bestValue, void *bestDataOut, uint32 dataOutSize);

#endif

