
#ifndef _ETC_MEMORY_H
#define _ETC_MEMORY_H

#define Bytes(n) (n)
#define Kilobytes(n) (1024 * Bytes(n))
#define Megabytes(n) (1024 * Kilobytes(n))
#define Gigabytes(n) (1024 * Megabytes(n))

enum EtcAllocatorType {
    EtcAllocator_Heap,
    EtcAllocator_Arena,
    EtcAllocator_ArenaList,
};

struct EtcAllocator {
    EtcAllocatorType type;
};

struct EtcHeap : EtcAllocator {
};


// Memory Arena
struct MemoryArena : EtcAllocator {
    uint64 size;
    uint64 used;

    void *data;
};

inline void AllocateMemoryArena(MemoryArena *arena, uint32 size) {
    arena->type = EtcAllocator_Arena;
    arena->size = size;
    arena->used = 0;
    arena->data = AllocateClearedMemory(size);
}

inline void DeallocateMemoryArena(MemoryArena *arena) {
    arena->size = 0;
    arena->used = 0;
    Plat.DeallocateMem(arena->data);
}

inline uint64 SpaceRemaining(MemoryArena *arena) {
    uint64 result = arena->size - arena->used;
    return result;
}

inline void *PushSize(MemoryArena *arena, uint64 size) {
    if (arena->data == NULL) {
        AllocateMemoryArena(arena, size);
    }

    uint32 bytesLeft = arena->size - arena->used;
    ASSERT(size <= bytesLeft);

    void *result = NULL;
    result = (void *)((intptr)arena->data + arena->used);
    arena->used += size;
    
    int32 align = 8;
    arena->used = (arena->used + (align - 1)) & ~(align - 1);

    Plat.ClearMem(result, size);
    
    return result;
}

inline MemoryArena PushSubArena(MemoryArena *arena, uint64 size) {
    ASSERT(arena->size - arena->used >= size);

    MemoryArena result;
    result.size = size;
    result.used = 0;
    result.data = PushSize(arena, size);

    return result;
}

inline void ClearMemoryArena(MemoryArena *arena) {
    Plat.ClearMem(arena->data, arena->used);
    arena->used = 0;
}

// MemoryArenaList
struct MemoryArenaListNode {
    MemoryArena arena;
    MemoryArenaListNode *next;
};

struct MemoryArenaList : EtcAllocator {
    MemoryArenaListNode *head;
    MemoryArenaListNode *current;

    uint64 arenaSize;
};

inline void AllocateMemoryArenaList(MemoryArenaList *arenaList, uint32 size) {
    ASSERT(size != 0);

    arenaList->type = EtcAllocator_ArenaList;

    arenaList->arenaSize = size;

    arenaList->head = (MemoryArenaListNode *)Plat.AllocateMem(sizeof(MemoryArenaListNode));
    AllocateMemoryArena(&arenaList->head->arena, arenaList->arenaSize);
    arenaList->head->next = NULL;

    arenaList->current = arenaList->head;
}

inline void DeallocateMemoryArenaList(MemoryArenaList *arenaList) {
    MemoryArenaListNode *node = arenaList->head;
    while (node != NULL) {
        DeallocateMemoryArena(&node->arena);
        MemoryArenaListNode *nextNode = node->next;

        Plat.DeallocateMem(node);
        node = nextNode;
    }
}

inline void *PushSize(MemoryArenaList *arenaList, uint64 size) {
    
    ASSERT(size <= arenaList->arenaSize);

    if (arenaList->head == NULL) {
        AllocateMemoryArenaList(arenaList, size);
    }

    // @TODO: this potentially wastes space if a large size is pushed!
    
    if (SpaceRemaining(&arenaList->current->arena) < size) {
        MemoryArenaListNode *next = arenaList->current->next;

        if (next == NULL) {
            next = (MemoryArenaListNode *)Plat.AllocateMem(sizeof(MemoryArenaListNode));
            AllocateMemoryArena(&next->arena, arenaList->arenaSize);
            next->next = NULL;
        }

        arenaList->current->next = next;
        arenaList->current = next;
    }

    void *result = PushSize(&arenaList->current->arena, size);
    return result;
}

inline void ClearMemoryArenaList(MemoryArenaList *arenaList) {
    MemoryArenaListNode *node = arenaList->head;
    while (node != NULL) {
        ClearMemoryArena(&node->arena);
        node = node->next;
    }

    arenaList->current = arenaList->head;
}


// These work for MemoryArena and MemoryArenaList
#define PushStruct(arena, type) ((type *)PushSize(arena, sizeof(type)))
#define PushArray(arena, type, length) ((type *)PushSize(arena, (length) * sizeof(type)))

// @GACK: these should be moved
struct ValueIndices {
    u32 count;
    u32 *indices;
};

ValueIndices GetValueIndices(MemoryArenaList *arena, u32 *array, u32 count, u32 value) {
    ValueIndices indices = {};
    indices.indices = PushArray(arena, uint32, count);
    for (int i = 0; i < count; i++) {
        if (array[i] == value) {
            indices.indices[indices.count] = i;
            indices.count++;
        }
    }

    return indices;
}

// MemoryPool
struct MemoryPoolBlock {
    void *data;
    MemoryPoolBlock *nextBlock;
};

template <typename T>
struct MemoryPool {
    uint32 blockCapacity;

    // The union is so that the free list pointers can be stored in
    // place in the memory pool
    union Slot {
        T value;
        Slot *nextFree;
    };

    MemoryPoolBlock *blockList;
    Slot *nextUnallocated;
    Slot *endOfBlock;

    Slot *freeList;
};

template <typename T>
void AllocateMemoryPool(MemoryPool<T> *pool, uint32 blockCapacity) {
    ASSERT(blockCapacity >= 2);
    pool->blockCapacity = blockCapacity;

    pool->blockList = NULL;
    pool->nextUnallocated = NULL;
    pool->endOfBlock = NULL;

    pool->freeList = NULL;
}

template <typename T>
void DeallocateMemoryPool(MemoryPool<T> *pool) {

    MemoryPoolBlock *block = pool->blockList;
    while (block != NULL) {
        MemoryPoolBlock *next = block->nextBlock;

        // This automatically frees all of the data too
        Plat.DeallocateMem(block);

        block = next;
    }

    pool->blockList = NULL;
    pool->nextUnallocated = NULL;
    pool->endOfBlock = NULL;

    pool->freeList = NULL;
}

template <typename T>
T *Allocate_(MemoryPool<T> *pool) {
    T *ptr = NULL;
    if (pool->freeList != NULL) {
        ptr = (T *)pool->freeList;
        pool->freeList = pool->freeList->nextFree;
    }
    else {
        if (pool->nextUnallocated >= pool->endOfBlock) {
            // Allocate space for MemoryPoolBlock struct and its data field
            size_t align = 8;
            size_t headerSize = (sizeof(MemoryPoolBlock) + (align - 1)) & ~(align - 1);

            uint64 dataSize = sizeof(typename MemoryPool<T>::Slot) * pool->blockCapacity;
            uint64 blockSize = headerSize + dataSize;

            void *memory = Plat.AllocateMem(blockSize);

            MemoryPoolBlock *block = (MemoryPoolBlock *)memory;
            block->data = (uint8 *)memory + headerSize;
            block->nextBlock = pool->blockList;
            pool->blockList = block;

            pool->nextUnallocated = (typename MemoryPool<T>::Slot *)block->data;
            pool->endOfBlock = (typename MemoryPool<T>::Slot *)block->data + pool->blockCapacity;
        }

        ptr = (T *)pool->nextUnallocated;
        pool->nextUnallocated++;
    }

    Plat.ClearMem(ptr, sizeof(T));
    return ptr;
}

template <typename T>
void Deallocate_(MemoryPool<T> *pool, T *ptr) {
    ((typename MemoryPool<T>::Slot *)ptr)->nextFree = pool->freeList;
    pool->freeList = (typename MemoryPool<T>::Slot *)ptr;
}


void *Allocate_(EtcAllocator *allocator, uint32 size) {
    void *result = NULL;
    switch (allocator->type) {
        case EtcAllocator_Heap: {
            EtcHeap *heap = (EtcHeap *)allocator;
            result = Plat.AllocateMem(size);
        } break;

        case EtcAllocator_Arena: {
            MemoryArena *arena = (MemoryArena *)allocator;
            result = PushSize(arena, size);
        } break;

        case EtcAllocator_ArenaList: {
            MemoryArenaList *arenaList = (MemoryArenaList *)allocator;
            result = PushSize(arenaList, size);
        } break;
    }

    return result;
}

void Deallocate_(EtcAllocator *allocator, void *ptr) {
    switch (allocator->type) {
        case EtcAllocator_Heap: {
            EtcHeap *heap = (EtcHeap *)allocator;
            Plat.DeallocateMem(ptr);
        } break;

        case EtcAllocator_Arena: {
            MemoryArena *arena = (MemoryArena *)allocator;
            //DeallocateMemoryArena(arena);
        } break;

        case EtcAllocator_ArenaList: {
            MemoryArenaList *arenaList = (MemoryArenaList *)allocator;
            //DeallocateMemoryArenaList(arenaList);
        } break;
    }
}

#define Allocate(allocator, type, count) ((type *)Allocate_(allocator, sizeof(type) * (count)))

#endif

