
void CreateRenderTargetOpenGL(RenderTarget *renderTarget, uint32 width, uint32 height, uint32 colorTargetCount, ColorFormat *colorFormats, uint8 depthBits) {
    ASSERT(colorTargetCount >= 1);
    ASSERT(colorTargetCount <= MAX_COLOR_ATTACHMENTS);

    renderTarget->width = width;
    renderTarget->height = height;
    renderTarget->colorTargetCount = colorTargetCount;
    renderTarget->depthBits = depthBits;


    glGenFramebuffers(1, &renderTarget->fboID);
    glBindFramebuffer(GL_FRAMEBUFFER, renderTarget->fboID);

    glGenTextures(colorTargetCount, renderTarget->colorTargets);
    for (int i = 0; i < colorTargetCount; i++) {
        glBindTexture(GL_TEXTURE_2D, renderTarget->colorTargets[i]);

        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

        switch (colorFormats[i]) {
            case ColorFormat_RGB8: {
                glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB8, width, height, 0, GL_RGB, GL_UNSIGNED_BYTE, NULL);
            } break;

            case ColorFormat_RGBA8: {
                glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);
            } break;

            case ColorFormat_R32F: {
                glTexImage2D(GL_TEXTURE_2D, 0, GL_R32F, width, height, 0, GL_RED, GL_FLOAT, NULL);
            } break;

            case ColorFormat_RGB32F: {
                glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB32F, width, height, 0, GL_RGB, GL_FLOAT, NULL);
            } break;

            case ColorFormat_RGBA32F: {
                glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA32F, width, height, 0, GL_RGBA, GL_FLOAT, NULL);
            } break;
        }

        glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0 + i, GL_TEXTURE_2D, renderTarget->colorTargets[i], 0); 
        glBindTexture(GL_TEXTURE_2D, 0);
    }

    if (depthBits > 0) {
        uint32 internalDepthFormat = 0;

        switch (depthBits) {
            case 16: {
                internalDepthFormat = GL_DEPTH_COMPONENT16;
            } break;
            case 24: {
                internalDepthFormat = GL_DEPTH_COMPONENT24;
            } break;
            case 32: {
                internalDepthFormat = GL_DEPTH_COMPONENT32;
            } break;
            default: {
                ASSERT(false);
            } break;
        }

        uint32 textureID;
        glGenTextures(1, &renderTarget->depthTarget);
        glBindTexture(GL_TEXTURE_2D, renderTarget->depthTarget);

        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

        glTexImage2D(GL_TEXTURE_2D, 0, internalDepthFormat, width, height, 0, GL_DEPTH_COMPONENT, GL_FLOAT, NULL);
        glBindTexture(GL_TEXTURE_2D, 0);

        glFramebufferTexture2D(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_TEXTURE_2D, renderTarget->depthTarget, 0); 
    }

    GLenum status = glCheckFramebufferStatus(GL_FRAMEBUFFER);
    ASSERT(status == GL_FRAMEBUFFER_COMPLETE);

    glBindFramebuffer(GL_FRAMEBUFFER, 0);
}

void ShareDepthBuffer(RenderTarget *source, RenderTarget *dest) {
    ASSERT(source->fboID != 0);
    ASSERT(dest->fboID != 0);
    ASSERT(source->depthTarget != 0);

    dest->depthBits = source->depthBits;
    dest->depthTarget = source->depthTarget;

    glBindFramebuffer(GL_FRAMEBUFFER, dest->fboID);

    glFramebufferTexture2D(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_TEXTURE_2D, source->depthTarget, 0); 

    glBindFramebuffer(GL_FRAMEBUFFER, 0);
}

void DeleteRenderTargetOpenGL(RenderTarget *renderTarget) {
    glDeleteTextures(renderTarget->colorTargetCount, renderTarget->colorTargets);
    glDeleteTextures(1, &renderTarget->depthTarget);

    glDeleteFramebuffers(1, &renderTarget->fboID);
}


void CreateSamplerOpenGL(Sampler *sampler, TextureFilterMode filterMode, TextureWrapMode wrapMode) {
    CreateSamplerOpenGL(sampler, filterMode, wrapMode, BLACK);
}

void CreateSamplerOpenGL(Sampler *sampler, TextureFilterMode filterMode, TextureWrapMode wrapMode, vec4 borderColor) {
    glGenSamplers(1, &sampler->samplerID);
    glCheckError();

    switch (filterMode) {
        case TextureFilter_Nearest: {
            glSamplerParameteri(sampler->samplerID, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
            glSamplerParameteri(sampler->samplerID, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
        } break;
        case TextureFilter_Linear: {
            glSamplerParameteri(sampler->samplerID, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
            glSamplerParameteri(sampler->samplerID, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        } break;
    }

    switch (wrapMode) {
        case TextureWrap_ClampToEdge: {
            glSamplerParameteri(sampler->samplerID, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
            glSamplerParameteri(sampler->samplerID, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
        } break;
        case TextureWrap_ClampToBorderColor: {
            glSamplerParameteri(sampler->samplerID, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_BORDER);
            glSamplerParameteri(sampler->samplerID, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_BORDER);

            glSamplerParameterfv(sampler->samplerID, GL_TEXTURE_BORDER_COLOR, borderColor.data);
        } break;
        case TextureWrap_Repeat: {
            glSamplerParameteri(sampler->samplerID, GL_TEXTURE_WRAP_S, GL_REPEAT);
            glSamplerParameteri(sampler->samplerID, GL_TEXTURE_WRAP_T, GL_REPEAT);
        } break;
    }

    glCheckError();
}

void DeleteSamplerOpenGL(Sampler *sampler) {
    glDeleteSamplers(1, &sampler->samplerID);
}

void CreateTexture1DOpenGL(Texture *texture) {
    ASSERT(texture->width > 0);
    ASSERT(texture->height == 1);
    ASSERT(texture->data.data != NULL);

    glGenTextures(1, &texture->textureID);
    glCheckError();

    glBindTexture(GL_TEXTURE_1D, texture->textureID);
    glCheckError();

    switch (texture->format) {
        case TextureFormat_R8: {
            glTexImage1D(GL_TEXTURE_1D, 0, GL_R8, texture->width, 0, GL_RED, GL_UNSIGNED_BYTE, texture->data.u8);
        } break;
        case TextureFormat_RG8: {
            glTexImage1D(GL_TEXTURE_1D, 0, GL_RG8, texture->width, 0, GL_RG, GL_UNSIGNED_BYTE, texture->data.u8);
        } break;
        case TextureFormat_RGB8: {
            glTexImage1D(GL_TEXTURE_1D, 0, GL_RGB8, texture->width, 0, GL_RGB, GL_UNSIGNED_BYTE, texture->data.u8);
        } break;
        case TextureFormat_RGBA8: {
            glTexImage1D(GL_TEXTURE_1D, 0, GL_RGBA8, texture->width, 0, GL_RGBA, GL_UNSIGNED_BYTE, texture->data.u8);
        } break;

        case TextureFormat_R32F: {
            glTexImage1D(GL_TEXTURE_1D, 0, GL_R32F, texture->width, 0, GL_RED, GL_FLOAT, texture->data.f32);
        } break;
        case TextureFormat_RG32F: {
            glTexImage1D(GL_TEXTURE_1D, 0, GL_RG32F, texture->width, 0, GL_RG, GL_FLOAT, texture->data.f32);
        } break;
        case TextureFormat_RGB32F: {
            glTexImage1D(GL_TEXTURE_1D, 0, GL_RGB32F, texture->width, 0, GL_RGB, GL_FLOAT, texture->data.f32);
        } break;
        case TextureFormat_RGBA32F: {
            glTexImage1D(GL_TEXTURE_1D, 0, GL_RGBA32F, texture->width, 0, GL_RGBA, GL_FLOAT, texture->data.f32);
        } break;
    }
    glCheckError();

    glBindTexture(GL_TEXTURE_2D, 0);
    glCheckError();
}

void CreateTexture2DOpenGL(Texture *texture) {
    ASSERT(texture->width > 0);
    ASSERT(texture->height > 0);
    ASSERT(texture->data.data != NULL);

    glGenTextures(1, &texture->textureID);
    glCheckError();

    glBindTexture(GL_TEXTURE_2D, texture->textureID);
    glCheckError();

    switch (texture->format) {
        case TextureFormat_R8: {
            glTexImage2D(GL_TEXTURE_2D, 0, GL_R8, texture->width, texture->height, 0, GL_RED, GL_UNSIGNED_BYTE, texture->data.u8);
        } break;
        case TextureFormat_RG8: {
            glTexImage2D(GL_TEXTURE_2D, 0, GL_RG8, texture->width, texture->height, 0, GL_RG, GL_UNSIGNED_BYTE, texture->data.u8);
        } break;
        case TextureFormat_RGB8: {
            glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB8, texture->width, texture->height, 0, GL_RGB, GL_UNSIGNED_BYTE, texture->data.u8);
        } break;
        case TextureFormat_RGBA8: {
            glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, texture->width, texture->height, 0, GL_RGBA, GL_UNSIGNED_BYTE, texture->data.u8);
        } break;

        case TextureFormat_R32F: {
            glTexImage2D(GL_TEXTURE_2D, 0, GL_R32F, texture->width, texture->height, 0, GL_RED, GL_FLOAT, texture->data.f32);
        } break;
        case TextureFormat_RG32F: {
            glTexImage2D(GL_TEXTURE_2D, 0, GL_RG32F, texture->width, texture->height, 0, GL_RG, GL_FLOAT, texture->data.f32);
        } break;
        case TextureFormat_RGB32F: {
            glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB32F, texture->width, texture->height, 0, GL_RGB, GL_FLOAT, texture->data.f32);
        } break;
        case TextureFormat_RGBA32F: {
            glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA32F, texture->width, texture->height, 0, GL_RGBA, GL_FLOAT, texture->data.f32);
        } break;
    }
    glCheckError();

    glBindTexture(GL_TEXTURE_2D, 0);
    glCheckError();
}

void DeleteTextureOpenGL(Texture *texture) {
    glDeleteTextures(1, &texture->textureID);
}

void UpdateTexture1DOpenGL(Texture *texture) {
    glBindTexture(GL_TEXTURE_1D, texture->textureID);
    glCheckError();

    switch (texture->format) {
        case TextureFormat_R8: {
            glTexSubImage1D(GL_TEXTURE_1D, 0, 0, texture->width, GL_RED, GL_UNSIGNED_BYTE, texture->data.u8);
        } break;
        case TextureFormat_RG8: {
            glTexSubImage1D(GL_TEXTURE_1D, 0, 0, texture->width, GL_RG, GL_UNSIGNED_BYTE, texture->data.u8);
        } break;
        case TextureFormat_RGB8: {
            glTexSubImage1D(GL_TEXTURE_1D, 0, 0, texture->width, GL_RGB, GL_UNSIGNED_BYTE, texture->data.u8);
        } break;
        case TextureFormat_RGBA8: {
            glTexSubImage1D(GL_TEXTURE_1D, 0, 0, texture->width, GL_RGBA, GL_UNSIGNED_BYTE, texture->data.u8);
        } break;

        case TextureFormat_R32F: {
            glTexSubImage1D(GL_TEXTURE_1D, 0, 0, texture->width, GL_RED, GL_FLOAT, texture->data.f32);
        } break;
        case TextureFormat_RG32F: {
            glTexSubImage1D(GL_TEXTURE_1D, 0, 0, texture->width, GL_RG, GL_FLOAT, texture->data.f32);
        } break;
        case TextureFormat_RGB32F: {
            glTexSubImage1D(GL_TEXTURE_1D, 0, 0, texture->width, GL_RGB, GL_FLOAT, texture->data.f32);
        } break;
        case TextureFormat_RGBA32F: {
            glTexSubImage1D(GL_TEXTURE_1D, 0, 0, texture->width, GL_RGBA, GL_FLOAT, texture->data.f32);
        } break;
    }
    glCheckError();

    glBindTexture(GL_TEXTURE_2D, 0);
    glCheckError();
}

void UpdateTexture2DOpenGL(Texture *texture) {
    glBindTexture(GL_TEXTURE_2D, texture->textureID);
    glCheckError();

    switch (texture->format) {
        case TextureFormat_R8: {
            glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, texture->width, texture->height, GL_RED, GL_UNSIGNED_BYTE, texture->data.u8);
        } break;
        case TextureFormat_RG8: {
            glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, texture->width, texture->height, GL_RG, GL_UNSIGNED_BYTE, texture->data.u8);
        } break;
        case TextureFormat_RGB8: {
            glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, texture->width, texture->height, GL_RGB, GL_UNSIGNED_BYTE, texture->data.u8);
        } break;
        case TextureFormat_RGBA8: {
            glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, texture->width, texture->height, GL_RGBA, GL_UNSIGNED_BYTE, texture->data.u8);
        } break;

        case TextureFormat_R32F: {
            glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, texture->width, texture->height, GL_RED, GL_FLOAT, texture->data.f32);
        } break;
        case TextureFormat_RG32F: {
            glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, texture->width, texture->height, GL_RG, GL_FLOAT, texture->data.f32);
        } break;
        case TextureFormat_RGB32F: {
            glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, texture->width, texture->height, GL_RGB, GL_FLOAT, texture->data.f32);
        } break;
        case TextureFormat_RGBA32F: {
            glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, texture->width, texture->height, GL_RGBA, GL_FLOAT, texture->data.f32);
        } break;
    }
    glCheckError();

    glBindTexture(GL_TEXTURE_2D, 0);
    glCheckError();
}

void CreateCubemapOpenGL(Cubemap *cubemap) {
    ASSERT(cubemap->width > 0);
    ASSERT(cubemap->height > 0);
    ASSERT(cubemap->components > 0);
    ASSERT(cubemap->components <= 4);
    ASSERT(cubemap->data != NULL);

    glGenTextures(1, &cubemap->textureID);
    glCheckError();
    glBindTexture(GL_TEXTURE_CUBE_MAP, cubemap->textureID);
    glCheckError();
    
    for (int face = 0; face < 6; face++) {
        uint8 *data = cubemap->data[face];

        switch (cubemap->components) {
            case 1: {
                glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_X + face, 0, GL_R8, cubemap->width, cubemap->height, 0, GL_RED, GL_UNSIGNED_BYTE, data);
            } break;
            case 2: {
                glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_X + face, 0, GL_RG8, cubemap->width, cubemap->height, 0, GL_RG, GL_UNSIGNED_BYTE, data);
            } break;
            case 3: {
                glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_X + face, 0, GL_RGB8, cubemap->width, cubemap->height, 0, GL_RGB, GL_UNSIGNED_BYTE, data);
            } break;
            case 4: {
                glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_X + face, 0, GL_RGBA8, cubemap->width, cubemap->height, 0, GL_RGBA, GL_UNSIGNED_BYTE, data);
            } break;
        }
        glCheckError();
    }

    glBindTexture(GL_TEXTURE_CUBE_MAP, 0);
    glCheckError();
}

void DeleteCubemapOpenGL(Cubemap *cubemap) {
    glDeleteTextures(1, &cubemap->textureID);
}

void CreateMeshOpenGL(Mesh *mesh) {
    if (mesh->vboID == 0) {
        glGenBuffers(1, &mesh->vboID);
    }

    glBindBuffer(GL_ARRAY_BUFFER, mesh->vboID);
    glBufferData(GL_ARRAY_BUFFER, mesh->vertexBufferSize, mesh->vertices, GL_STREAM_DRAW);
    glBindBuffer(GL_ARRAY_BUFFER, 0);

    if (mesh->indexCount != 0) {
        if (mesh->iboID == 0) {
            glGenBuffers(1, &mesh->iboID);
        }

        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, mesh->iboID);
        glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(uint32) * mesh->indexCount, mesh->indices, GL_STREAM_DRAW);
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
    }
}

void DeleteMeshOpenGL(Mesh *mesh) {
    glDeleteBuffers(1, &mesh->vboID);
    glDeleteBuffers(1, &mesh->iboID);

    mesh->vboID = 0;
    mesh->iboID = 0;
}

b32 ShaderCompiled(GLuint shader, char **infoLog) {
    b32 isCompiled = 0;
    glGetShaderiv(shader, GL_COMPILE_STATUS, &isCompiled);
    if (isCompiled == GL_FALSE) {
        if (infoLog != NULL) {
            GLint maxLength = 0;
            glGetShaderiv(shader, GL_INFO_LOG_LENGTH, &maxLength);

            *infoLog = (GLchar *)Plat.AllocateMem(sizeof(GLchar) * maxLength);
            glGetShaderInfoLog(shader, maxLength, &maxLength, (GLchar *)*infoLog);
        }

        glDeleteShader(shader);
    }

    return isCompiled;
}

b32 ProgramLinked(GLuint program, char **infoLog) {

    b32 isLinked = 0;
    glGetProgramiv(program, GL_LINK_STATUS, &isLinked);
    if (isLinked == GL_FALSE) {

        if (infoLog != NULL) {
            GLint maxLength = 0;
            glGetProgramiv(program, GL_INFO_LOG_LENGTH, &maxLength);

            *infoLog = (GLchar *)Plat.AllocateMem(sizeof(GLchar) * maxLength);
            glGetProgramInfoLog(program, maxLength, &maxLength, (GLchar *)*infoLog);
        }

        glDeleteProgram(program);
    }

    return isLinked;
}

void CreateShaderError(ShaderProgram *program, const char *infoLog) {
    if (program->programID != 0) {
        glDeleteProgram(program->programID);
    }

    if (program->geometryID != 0) {
        glDeleteShader(program->geometryID);
    }

    if (program->fragmentID != 0) {
        glDeleteShader(program->fragmentID);
    }

    if (program->vertexID != 0) {
        glDeleteShader(program->vertexID);
    }

    program->programID = 0;
    program->vertexID = 0;
    program->fragmentID = 0;
    program->geometryID = 0;

    if (infoLog != NULL) {
        Plat.DebugPrint(infoLog);
    }
}

#define BIND_VERTEXATTRIB(attribName, attribInputString, value)         \
    if (attribInputString) {                                            \
        glBindAttribLocation(program->programID, attribName##_Attrib, attribInputString); \
    }

void CreateShaderProgramOpenGL(ShaderProgram *program) {
    char *infoLog = NULL;

    if (program->vertexSource != NULL) {
        program->vertexID = glCreateShader(GL_VERTEX_SHADER);
        glShaderSource(program->vertexID, 1, (const GLchar**)&program->vertexSource, (GLint *)&program->vertexSourceSize);
        glCheckError();
        glCompileShader(program->vertexID);
        glCheckError();

        if (!ShaderCompiled(program->vertexID, &infoLog)) {
            CreateShaderError(program, infoLog);
            return;
        }
    }

    if (program->fragmentSource != NULL) {
        program->fragmentID = glCreateShader(GL_FRAGMENT_SHADER);
        glShaderSource(program->fragmentID, 1, (const GLchar**)&program->fragmentSource, (GLint *)&program->fragmentSourceSize);
        glCheckError();
        glCompileShader(program->fragmentID);
        glCheckError();

        if (!ShaderCompiled(program->fragmentID, &infoLog)) {
            CreateShaderError(program, infoLog);
            return;
        }
    }

    if (program->geometrySource != NULL) {
        program->geometryID = glCreateShader(GL_GEOMETRY_SHADER);
        glShaderSource(program->geometryID, 1, (const GLchar**)&program->geometrySource, (GLint *)&program->geometrySourceSize);
        glCheckError();
        glCompileShader(program->geometryID);
        glCheckError();

        if (!ShaderCompiled(program->geometryID, &infoLog)) {
            CreateShaderError(program, infoLog);
            return;
        }
    }

    if (program->computeSource != NULL) {
        program->computeID = glCreateShader(GL_COMPUTE_SHADER);
        glShaderSource(program->computeID, 1, (const GLchar**)&program->computeSource, (GLint *)&program->computeSourceSize);
        glCheckError();
        glCompileShader(program->computeID);
        glCheckError();

        if (!ShaderCompiled(program->computeID, &infoLog)) {
            CreateShaderError(program, infoLog);
            return;
        }
    }

        
    program->programID = glCreateProgram();
    glCheckError();

    if (program->vertexID != 0) {
        glAttachShader(program->programID, program->vertexID);
        glCheckError();
    }
    if (program->fragmentID != 0) {
        glAttachShader(program->programID, program->fragmentID);
        glCheckError();
    }
    if (program->geometryID != 0) {
        glAttachShader(program->programID, program->geometryID);
        glCheckError();
    }
    if (program->computeID != 0) {
        glAttachShader(program->programID, program->computeID);
        glCheckError();
    }


    // Specify Shader Inputs
    FOREACH_VERTEXATTRIB(BIND_VERTEXATTRIB)


    // Specify Shader Outputs
    glBindFragDataLocation(program->programID, GBUFFER_DIFFUSE_BUFFER, "fragColor");
    glBindFragDataLocation(program->programID, GBUFFER_NORMAL_BUFFER, "fragNormal");
    glBindFragDataLocation(program->programID, 1, "fragSmear");
    glCheckError();


    glLinkProgram(program->programID);
    glCheckError();

    if (!ProgramLinked(program->programID, &infoLog)) {
        CreateShaderError(program, infoLog);
        return;
    }


    // Build Uniforms Table
    int32 numUniforms;
    glGetProgramiv(program->programID, GL_ACTIVE_UNIFORMS, &numUniforms);

    for (uint32 i = 0; i < numUniforms; i++) {
        int32 blockIndex;
        glGetActiveUniformsiv(program->programID, 1, &i, GL_UNIFORM_BLOCK_INDEX, &blockIndex);
        glCheckError();
        if (blockIndex != -1) {
            continue;
        }

        int32 uniformNameLength;
        glGetActiveUniformsiv(program->programID, 1, &i, GL_UNIFORM_NAME_LENGTH, &uniformNameLength);
        glCheckError();
        
        GLenum type;
        GLint size;

        char *uniformName = (char *)Plat.AllocateMem(sizeof(char) * uniformNameLength);
        glGetActiveUniform(program->programID, i, uniformNameLength, NULL, &size, &type, uniformName);
        glCheckError();

        int32 location = glGetUniformLocation(program->programID, uniformName);
        glCheckError();

        //Plat.DebugPrint("(%d) %s: index=%d, location=%d, type=0x%x, size=%d", program->programID, uniformName, i, location, type, size);

        PushBack(&program->uniforms, {(const char *)uniformName, location});
    }

    {
        int32 blockIndex = glGetUniformBlockIndex(program->programID, "CameraBlock");
        if (blockIndex >= 0) {
            glUniformBlockBinding(program->programID, blockIndex, UNIFORM_BLOCK_BINDING_CAMERA);
        }
    }

    {
        int32 blockIndex = glGetUniformBlockIndex(program->programID, "ScreenBlock");
        if (blockIndex >= 0) {
            glUniformBlockBinding(program->programID, blockIndex, UNIFORM_BLOCK_BINDING_SCREEN);
        }
    }

    {
        int32 blockIndex = glGetUniformBlockIndex(program->programID, "StrokeParamsBlock");
        if (blockIndex >= 0) {
            glUniformBlockBinding(program->programID, blockIndex, UNIFORM_BLOCK_BINDING_STROKEPARAMS);
        }
    }

    {
        int32 blockIndex = glGetUniformBlockIndex(program->programID, "BoneTransformsBlock");
        if (blockIndex >= 0) {
            glUniformBlockBinding(program->programID, blockIndex, UNIFORM_BLOCK_BINDING_BONETRANSFORMS);
        }
    }
}

void DeleteShaderProgramOpenGL(ShaderProgram *program) {
    glDeleteShader(program->vertexID);
    glCheckError();
    glDeleteShader(program->fragmentID);
    glCheckError();
    glDeleteShader(program->geometryID);
    glCheckError();
    glDeleteProgram(program->programID);
    glCheckError();
}

/// RenderCommand Backend Functions
void ClearColorOpenGL(ClearColorCommand *command) {
    glClearBufferfv(GL_COLOR, command->buffer, command->color.data);
    glCheckError();
}

void ClearDepthOpenGL(ClearDepthCommand *command) {
    glClearBufferfv(GL_DEPTH, 0, &command->depth);
    glCheckError();
}

void ClearStencilOpenGL(ClearStencilCommand *command) {
    glClearBufferiv(GL_STENCIL, 0, &command->value);
}

void BindBufferOpenGL(BindBufferCommand *command) {
    glBindBuffer(command->target, command->bufferID);
    glCheckError();
}

void BindBufferRangeOpenGL(BindBufferRangeCommand *command) {
    glBindBufferRange(command->target, command->index, command->bufferID, command->offset, command->size);
    glCheckError();
}

void CopyBufferDataOpenGL(CopyBufferDataCommand *command) {
    glBufferData(command->target, command->size, command->data, command->usage);
    glCheckError();
}

void CopyBufferSubDataOpenGL(CopyBufferSubDataCommand *command) {
    glBufferSubData(command->target, command->offset, command->size, command->data);
    glCheckError();
}

void EnableVertexAttributeOpenGL(EnableVertexAttributeCommand *command) {
    glEnableVertexAttribArray(command->index);
    glVertexAttribPointer(command->index, command->size, command->type, command->normalized, command->stride, (uint8 *)NULL + command->offset);

    glVertexAttribDivisor(command->index, command->divisor);
    glCheckError();
}

void EnableVertexIAttributeOpenGL(EnableVertexIAttributeCommand *command) {
    glEnableVertexAttribArray(command->index);
    glVertexAttribIPointer(command->index, command->size, command->type, command->stride, (uint8 *)NULL + command->offset);

    glVertexAttribDivisor(command->index, command->divisor);
    glCheckError();
}

void DisableVertexAttributeOpenGL(DisableVertexAttributeCommand *command) {
    glDisableVertexAttribArray(command->index);
    glCheckError();
}

void SetDrawRenderTargetOpenGL(SetDrawRenderTargetCommand *command) {
    glBindFramebuffer(GL_DRAW_FRAMEBUFFER, command->fboID);
    glCheckError();
}

void SetReadRenderTargetOpenGL(SetReadRenderTargetCommand *command) {
    glBindFramebuffer(GL_READ_FRAMEBUFFER, command->fboID);
    glCheckError();
}

void SetViewportOpenGL(SetViewportCommand *command) {
    glViewport(command->viewport.pos.x, command->viewport.pos.y,
               command->viewport.size.x, command->viewport.size.y);
    glCheckError();
}

void SetColorBuffersOpenGL(SetColorBuffersCommand *command) {
    glDrawBuffers(command->count, command->buffers);
    glCheckError();
}

void EnableScissorTestOpenGL(EnableScissorTestCommand *command) {
    glEnable(GL_SCISSOR_TEST);
    glScissor(command->region.pos.x, command->region.pos.y, command->region.size.x, command->region.size.y);
    glCheckError();
}

void DisableScissorTestOpenGL(DisableScissorTestCommand *command) {
    glDisable(GL_SCISSOR_TEST);
    glCheckError();
}

void SetColorMaskOpenGL(SetColorMaskCommand *command) {
    glColorMaski(command->buffer, command->red, command->green, command->blue, command->alpha);
    glCheckError();
}

void EnableDepthTestOpenGL(EnableDepthTestCommand *command) {
    glEnable(GL_DEPTH_TEST);
    glCheckError();

    switch (command->type) {
        case DepthTest_Always: {
            glDepthFunc(GL_ALWAYS);
        } break;
        case DepthTest_Never: {
            glDepthFunc(GL_NEVER);
        } break;
        case DepthTest_Less: {
            glDepthFunc(GL_LESS);
        } break;
        case DepthTest_Greater: {
            glDepthFunc(GL_GREATER);
        } break;
        case DepthTest_Equal: {
            glDepthFunc(GL_EQUAL);
        } break;
        case DepthTest_NotEqual: {
            glDepthFunc(GL_NOTEQUAL);
        } break;
        case DepthTest_LessEqual: {
            glDepthFunc(GL_LEQUAL);
        } break;
        case DepthTest_GreaterEqual: {
            glDepthFunc(GL_GEQUAL);
        } break;
    }

    glCheckError();
}

void DisableDepthTestOpenGL(DisableDepthTestCommand *command) {
    glDisable(GL_DEPTH_TEST);
    glCheckError();
}

void EnableDepthWriteOpenGL(EnableDepthWriteCommand *command) {
    glDepthMask(GL_TRUE);
    glCheckError();
}

void DisableDepthWriteOpenGL(DisableDepthWriteCommand *command) {
    glDepthMask(GL_FALSE);
    glCheckError();
}

void SetStencilMaskOpenGL(SetStencilMaskCommand *command) {
    glStencilMask(command->mask);
    glCheckError();
}

void EnableStencilTestOpenGL(EnableStencilTestCommand *command) {
    glEnable(GL_STENCIL_TEST);
    glCheckError();
}

void DisableStencilTestOpenGL(DisableStencilTestCommand *command) {
    glDisable(GL_STENCIL_TEST);
    glCheckError();
}

void SetStencilOpOpenGL(SetStencilOpCommand *command) {
    glStencilOp(command->stencilFail, command->depthFail, command->depthPass);
    glCheckError();
}

void SetStencilFuncOpenGL(SetStencilFuncCommand *command) {
    glStencilFunc(command->func, command->reference, command->mask);
    glCheckError();
}

void EnableBlendingOpenGL(EnableBlendingCommand *command) {
    glEnable(GL_BLEND);
    glBlendEquationSeparate(command->rgbEquation, command->alphaEquation);
    glBlendFuncSeparate(command->srcFactor, command->dstFactor, command->srcAlphaFactor, command->dstAlphaFactor);
    glCheckError();
}

void DisableBlendingOpenGL(DisableBlendingCommand *command) {
    glDisable(GL_BLEND);
    glCheckError();
}

void EnableWireframeOpenGL(EnableWireframeCommand *command) {
    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    glCheckError();
}

void DisableWireframeOpenGL(DisableWireframeCommand *command) {
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    glCheckError();
}

void EnableFaceCullingOpenGL(EnableFaceCullingCommand *command) {
    glEnable(GL_CULL_FACE);
    glCullFace(command->faceMode);
    glCheckError();
}

void DisableFaceCullingOpenGL(DisableFaceCullingCommand *command) {
    glDisable(GL_CULL_FACE);
    glCheckError();
}

void SetShaderOpenGL(SetShaderCommand *command) {
    glUseProgram(command->programID);
    glCheckError();
}

void SetSamplerOpenGL(SetSamplerCommand *command) {
    glBindSampler(command->textureUnit, command->samplerID);
    glCheckError();
}

void SetUniformOpenGL(SetUniformCommand *command) {
    switch (command->type) {
        case UniformType_Integer: {
            glUniform1i(command->location, command->int32Value);
        } break;
        case UniformType_Float: {
            glUniform1f(command->location, command->real32Value);
        } break;
        case UniformType_Vec2: {
            glUniform2fv(command->location, 1, command->vec2Value.data);
        } break;
        case UniformType_Vec3: {
            glUniform3fv(command->location, 1, command->vec3Value.data);
        } break;
        case UniformType_Vec4: {
            glUniform4fv(command->location, 1, command->vec4Value.data);
        } break;
        case UniformType_Mat4: {
            glUniformMatrix4fv(command->location, 1, false, command->mat4Value.data);
        } break;
        case UniformType_Texture: {
            glActiveTexture(GL_TEXTURE0 + command->textureUnit);
            glBindTexture(command->textureTarget, command->textureID);
            glUniform1i(command->location, command->textureUnit);
        } break;
    }

    glCheckError();
}

void DrawArraysOpenGL(DrawArraysCommand *command) {
    glDrawArrays(command->primitiveType, command->first, command->count);
    glCheckError();
}

void DrawArraysInstancedOpenGL(DrawArraysInstancedCommand *command) {
    glDrawArraysInstancedBaseInstance(command->primitiveType, command->first, command->count, command->instanceCount, command->baseInstance);
    glCheckError();
}

void DrawIndexedOpenGL(DrawIndexedCommand *command) {
    glDrawElements(command->primitiveType, command->count, GL_UNSIGNED_INT, (uint8 *)NULL + command->first);
    glCheckError();
}

void DrawIndexedInstancedOpenGL(DrawIndexedInstancedCommand *command) {
    glDrawElementsInstancedBaseInstance(command->primitiveType, command->count, GL_UNSIGNED_INT, (uint8 *)NULL + command->first, command->instanceCount, command->baseInstance);
    glCheckError();
}

void ReadPixelsOpenGL(ReadPixelsCommand *command) {
    glReadPixels(command->x, command->y, command->width, command->height, command->format, command->type, command->data);
    glCheckError();
}


RenderBackendFunction *ClearColorCommand::backendFn = (RenderBackendFunction *)ClearColorOpenGL;
RenderBackendFunction *ClearDepthCommand::backendFn = (RenderBackendFunction *)ClearDepthOpenGL;
RenderBackendFunction *ClearStencilCommand::backendFn = (RenderBackendFunction *)ClearStencilOpenGL;
RenderBackendFunction *BindBufferCommand::backendFn = (RenderBackendFunction *)BindBufferOpenGL;
RenderBackendFunction *BindBufferRangeCommand::backendFn = (RenderBackendFunction *)BindBufferRangeOpenGL;
RenderBackendFunction *CopyBufferDataCommand::backendFn = (RenderBackendFunction *)CopyBufferDataOpenGL;
RenderBackendFunction *CopyBufferSubDataCommand::backendFn = (RenderBackendFunction *)CopyBufferSubDataOpenGL;
RenderBackendFunction *EnableVertexAttributeCommand::backendFn = (RenderBackendFunction *)EnableVertexAttributeOpenGL;
RenderBackendFunction *EnableVertexIAttributeCommand::backendFn = (RenderBackendFunction *)EnableVertexIAttributeOpenGL;
RenderBackendFunction *DisableVertexAttributeCommand::backendFn = (RenderBackendFunction *)DisableVertexAttributeOpenGL;
RenderBackendFunction *SetDrawRenderTargetCommand::backendFn = (RenderBackendFunction *)SetDrawRenderTargetOpenGL;
RenderBackendFunction *SetReadRenderTargetCommand::backendFn = (RenderBackendFunction *)SetReadRenderTargetOpenGL;
RenderBackendFunction *SetViewportCommand::backendFn = (RenderBackendFunction *)SetViewportOpenGL;
RenderBackendFunction *SetColorBuffersCommand::backendFn = (RenderBackendFunction *)SetColorBuffersOpenGL;
RenderBackendFunction *EnableScissorTestCommand::backendFn = (RenderBackendFunction *)EnableScissorTestOpenGL;
RenderBackendFunction *DisableScissorTestCommand::backendFn = (RenderBackendFunction *)DisableScissorTestOpenGL;
RenderBackendFunction *SetColorMaskCommand::backendFn = (RenderBackendFunction *)SetColorMaskOpenGL;
RenderBackendFunction *EnableDepthTestCommand::backendFn = (RenderBackendFunction *)EnableDepthTestOpenGL;
RenderBackendFunction *DisableDepthTestCommand::backendFn = (RenderBackendFunction *)DisableDepthTestOpenGL;
RenderBackendFunction *EnableDepthWriteCommand::backendFn = (RenderBackendFunction *)EnableDepthWriteOpenGL;
RenderBackendFunction *DisableDepthWriteCommand::backendFn = (RenderBackendFunction *)DisableDepthWriteOpenGL;
RenderBackendFunction *SetStencilMaskCommand::backendFn = (RenderBackendFunction *)SetStencilMaskOpenGL;
RenderBackendFunction *EnableStencilTestCommand::backendFn = (RenderBackendFunction *)EnableStencilTestOpenGL;
RenderBackendFunction *DisableStencilTestCommand::backendFn = (RenderBackendFunction *)DisableStencilTestOpenGL;
RenderBackendFunction *SetStencilOpCommand::backendFn = (RenderBackendFunction *)SetStencilOpOpenGL;
RenderBackendFunction *SetStencilFuncCommand::backendFn = (RenderBackendFunction *)SetStencilFuncOpenGL;
RenderBackendFunction *EnableBlendingCommand::backendFn = (RenderBackendFunction *)EnableBlendingOpenGL;
RenderBackendFunction *DisableBlendingCommand::backendFn = (RenderBackendFunction *)DisableBlendingOpenGL;
RenderBackendFunction *EnableWireframeCommand::backendFn = (RenderBackendFunction *)EnableWireframeOpenGL;
RenderBackendFunction *DisableWireframeCommand::backendFn = (RenderBackendFunction *)DisableWireframeOpenGL;
RenderBackendFunction *EnableFaceCullingCommand::backendFn = (RenderBackendFunction *)EnableFaceCullingOpenGL;
RenderBackendFunction *DisableFaceCullingCommand::backendFn = (RenderBackendFunction *)DisableFaceCullingOpenGL;
RenderBackendFunction *SetShaderCommand::backendFn = (RenderBackendFunction *)SetShaderOpenGL;
RenderBackendFunction *SetSamplerCommand::backendFn = (RenderBackendFunction *)SetSamplerOpenGL;
RenderBackendFunction *SetUniformCommand::backendFn = (RenderBackendFunction *)SetUniformOpenGL;
RenderBackendFunction *DrawArraysCommand::backendFn = (RenderBackendFunction *)DrawArraysOpenGL;
RenderBackendFunction *DrawArraysInstancedCommand::backendFn = (RenderBackendFunction *)DrawArraysInstancedOpenGL;
RenderBackendFunction *DrawIndexedCommand::backendFn = (RenderBackendFunction *)DrawIndexedOpenGL;
RenderBackendFunction *DrawIndexedInstancedCommand::backendFn = (RenderBackendFunction *)DrawIndexedInstancedOpenGL;
RenderBackendFunction *ReadPixelsCommand::backendFn = (RenderBackendFunction *)ReadPixelsOpenGL;

