
#include "etcetera.h"
#include "etc_opengl.h"

#define STB_IMAGE_IMPLEMENTATION
#include "stb_image.h"

#define STB_IMAGE_WRITE_IMPLEMENTATION
#include "stb_image_write.h"

#define STB_VORBIS_NO_PUSHDATA_API
#include "stb_vorbis.c"

#define STB_SPRINTF_IMPLEMENTATION
#include "stb_sprintf.h"

#define STB_RECT_PACK_IMPLEMENTATION
#include "stb_rect_pack.h"

#include "fuzzy-native/score_match.cpp"


#if !ETC_RELEASE
PlatformAPI Plat = {};
void RegisterPlatformAPI(PlatformAPI *api) {
    Plat = *api;

#if !ETC_RELEASE && _WIN32
    ASSERT(glewInit() == GLEW_OK);
#endif
}

GameAPI Game = {};
#endif


#include "function_tables.generated.cpp"

#include "io/utf8.cpp"
#include "io/etctxt.cpp"
#include "io/glsl.cpp"
#include "task.cpp"

#include "ui.cpp"

#include "schema.cpp"
#include "ecs.cpp"

#include "transform.cpp"

#include "ecs_core.cpp"

#include "mesh.cpp"
#include "shapes.cpp"

#include "camera.cpp"
#include "light.cpp"


#include "mixer.cpp"

#include "input.cpp"

#include "render.cpp"

#include "assets.cpp"

#include "etc_opengl.cpp"

#include "scenes.cpp"
#include "game.cpp"

#include "game_code.cpp"

#include "schemas.generated.cpp"


#if ETC_DEBUG
#include "debug.cpp"
#endif


// UserPreferences
void SerializeUserPreferences(const char *path, UserPreferences *preferences) {
    // Do some cleanup of the preferences to reflect current data
    if (Game.sceneManager->active.sceneCount > 0) {
        preferences->lastScene = GetScene(GetRootScene(Game.sceneManager->active.scenes[0]))->sceneID;
    }
    else {
        preferences->lastScene = Scene_None;
    }


    // Serialize
    Plat.MakeDirectory("userdata");
    Schema *userPreferencesSchema = Game.GetSchema("UserPreferences");

    EtcTxtWriteState writer;
    if (OpenFileForWrite(&writer, path)) {

        SerializeValue(&writer, userPreferencesSchema, preferences, "UserPreferences");

        CloseFile(&writer);
    }
}

UserPreferences DeserializeUserPreferences(const char *path) {
    UserPreferences preferences = DefaultUserPreferences();

    Schema *userPreferencesSchema = Game.GetSchema("UserPreferences");
    EtcTxtParseState parser;
    if (OpenFileForRead(&parser, path)) {

        const char *identifier;
        if (ParseIdentifier(&parser, &identifier, "UserPreferences")) {
            DeserializeValue(&parser, userPreferencesSchema, &preferences);
        }

        CloseFile(&parser);
    }

    return preferences;
}


void InitGraphics(EtcGameMemory *mem, uint32 width, uint32 height) {

    EtcGraphics *graphics = &mem->graphics;

    graphics->width = width;
    graphics->height = height;
    graphics->dimensions = V2(graphics->width, graphics->height);
    graphics->aspect = (real32)graphics->width / (real32)graphics->height;

    // Camera Uniform Buffer
    glGenBuffers(1, &graphics->cameraUniformBufferID);
    glBindBuffer(GL_UNIFORM_BUFFER, graphics->cameraUniformBufferID);

    CameraUniformBuffer cameraUniformBuffer = {};
    glBufferData(GL_UNIFORM_BUFFER, sizeof(CameraUniformBuffer), &cameraUniformBuffer, GL_DYNAMIC_DRAW);

    glBindBufferRange(GL_UNIFORM_BUFFER, UNIFORM_BLOCK_BINDING_CAMERA, graphics->cameraUniformBufferID, 0, sizeof(CameraUniformBuffer));

    glBindBuffer(GL_UNIFORM_BUFFER, 0);


    // Screen Uniform Buffer
    glGenBuffers(1, &graphics->screenUniformBufferID);
    glBindBuffer(GL_UNIFORM_BUFFER, graphics->screenUniformBufferID);

    ScreenUniformBuffer screenUniformBuffer = {};
    screenUniformBuffer.width = (real32)width;
    screenUniformBuffer.height = (real32)height;

    glBufferData(GL_UNIFORM_BUFFER, sizeof(ScreenUniformBuffer), &screenUniformBuffer, GL_STATIC_DRAW);
    glBindBufferRange(GL_UNIFORM_BUFFER, UNIFORM_BLOCK_BINDING_SCREEN, graphics->screenUniformBufferID, 0, sizeof(ScreenUniformBuffer));

    glBindBuffer(GL_UNIFORM_BUFFER, 0);


    // Samplers
    CreateSamplerOpenGL(&graphics->nearestClampEdge, TextureFilter_Nearest, TextureWrap_ClampToEdge);
    CreateSamplerOpenGL(&graphics->linearClampEdge, TextureFilter_Linear, TextureWrap_ClampToEdge);
    CreateSamplerOpenGL(&graphics->linearRepeat, TextureFilter_Linear, TextureWrap_Repeat);

    // Shaders
    graphics->constantColorShader = LoadShaderProgram(Game.assets, "constantColorShader");
    graphics->skyboxShader = LoadShaderProgram(&mem->assets, "skyboxShader");
    graphics->textShader = LoadShaderProgram(&mem->assets, "textShader");


    graphics->deferredDirectionalLightShader = LoadShaderProgram(&mem->assets, "deferred_directionalLight");
    graphics->deferredPointLightShader = LoadShaderProgram(&mem->assets, "deferred_pointLight");
    graphics->deferredSpotlightShader = LoadShaderProgram(&mem->assets, "deferred_spotlight");

    graphics->defaultUIMeshShader = LoadShaderProgram(&mem->assets, "uiMesh");
    graphics->defaultUITexturedMeshShader = LoadShaderProgram(&mem->assets, "uiTexturedMesh");
    graphics->defaultUITextShader = LoadShaderProgram(&mem->assets, "uiText");


    // Glyph Quad
    AllocatePlane(&graphics->glyphQuad, 2);

    glGenBuffers(1, &graphics->glyphQuad.vboID);
    glBindBuffer(GL_ARRAY_BUFFER, graphics->glyphQuad.vboID);
    glBufferData(GL_ARRAY_BUFFER, graphics->glyphQuad.vertexBufferSize, graphics->glyphQuad.vertices, GL_STATIC_DRAW);

    glGenBuffers(1, &graphics->glyphQuad.iboID);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, graphics->glyphQuad.iboID);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(uint32) * graphics->glyphQuad.indexCount, graphics->glyphQuad.indices, GL_STATIC_DRAW);


    // Render Targets
    ColorFormat gbufferFormats[] = {ColorFormat_RGB8, ColorFormat_RGB32F};
    ColorFormat shadingFormats[] = {ColorFormat_RGB8};

    for (int i = 0; i < MAX_ACTIVE_SCENES; i++) {
        CreateRenderTargetOpenGL(&graphics->gbufferTargets[i], width, height, 2, gbufferFormats, 24);
        CreateRenderTargetOpenGL(&graphics->shadingTargets[i], width, height, 1, shadingFormats, 0);

        ShareDepthBuffer(&graphics->gbufferTargets[i], &graphics->shadingTargets[i]);
    }


    AllocateTexture(&graphics->sceneBlendMap, width, height, TextureFormat_RGBA32F);
    CreateTexture2DOpenGL(&graphics->sceneBlendMap);


    // Command Lists
    AllocateCommandList(&graphics->gbufferPass, 1 << 24, Megabytes(8));

    AllocateCommandList(&graphics->shadingPass, 1024, Kilobytes(256));

    AllocateCommandList(&graphics->textRenderPass, 8192 * 2, Kilobytes(256));

    // Bind globalVAOID
    glGenVertexArrays(1, &graphics->globalVAOID);
    glBindVertexArray(graphics->globalVAOID);

    // @NOTE: _always_ use ccw for front faces!
    glFrontFace(GL_CCW);

}

void DeinitGraphics(EtcGameMemory *mem) {
    EtcGraphics *graphics = &mem->graphics;
}

void InitEtcGameMemory(EtcGameMemory *mem) {

    mem->userPreferences = DeserializeUserPreferences(USERPREFERENCES_PATH);

    AllocateMemoryArena(&mem->permanentArena, PERMANENT_ARENA_STORAGE_SIZE);
    AllocateMemoryArenaList(&mem->frameMem, Megabytes(16));
    // @GACKY:
    mem->heap.type = EtcAllocator_Heap;

    // @GACK: do this here before the init because we may want to run TIMED_BLOCK in one of these
    // which relies on DebugRecords being allocated
#if ETC_DEBUG
    DebugMem = (EtcDebugMemory *)Plat.AllocateMem(sizeof(EtcDebugMemory));
    Plat.ClearMem(DebugMem, sizeof(EtcDebugMemory));
#endif
    
    AllocateAssetDB(&mem->assets, "data");

    InitGraphics(mem, mem->userPreferences.screenWidth, mem->userPreferences.screenHeight);
    InitMixer(&mem->mixer);

    InitSceneManager(&mem->sceneManager);

    mem->inputQueue.capacity = 128;
    mem->inputQueue.queue = PushArray(&mem->permanentArena, EtcInputEvent, mem->inputQueue.capacity);

    {
        AllocateUIManager(&mem->uiManager);
    }
    
    // Loading Keybindings
    SetBindingTypes(mem->inputActions.bindingTypes);
    if (!DeserializeInputActions(mem->userPreferences.keybindingsPath, &mem->inputActions)) {
        DeserializeInputActions(DEFAULT_KEYBINDINGS_PATH, &mem->inputActions);
    }

    // Load starting scene
    SceneHandle sceneHandle = LoadScene(Scene_Breakout);
    SetActiveScene(sceneHandle);


#if ETC_DEBUG
    InitDebug(DebugMem);
#endif

    SeedRand((int32)(int64)mem);
}

void DeinitEtcGameMemory(EtcGameMemory *mem) {
    SerializeUserPreferences(USERPREFERENCES_PATH, &mem->userPreferences);

#if ETC_DEBUG
    DeinitDebug(DebugMem);
#endif

    DeinitMixer(&mem->mixer);

    DeinitGraphics(mem);

    DeinitSceneManager(&mem->sceneManager);
        
    DeallocateAssetDB(&mem->assets);
    DeallocateMemoryArena(&mem->permanentArena);
}

void UpdateAndRender(EtcGameMemory *mem) {
    TIMED_BLOCK();

    // 1) Process Input
#if ETC_DEBUG

#if __linux__
    FindModifiedAssetPaths(Game.assets);
#endif

    ReloadModifiedAssets(Game.assets);
#endif

    UpdateSceneManager(&mem->sceneManager);

    {
        ProcessInputEventQueue(&mem->inputQueue);
        ProcessInputActions(&mem->inputActions, &mem->keyboard, &mem->mouse, mem->gamepads, mem->gamepadCount);

        ProcessMouseLookRequests();
    }


    // 2) Determine which CommandLists to clear
    ClearCommandList(&mem->graphics.gbufferPass);
    ClearCommandList(&mem->graphics.shadingPass);
    ClearCommandList(&mem->graphics.textRenderPass);

    {
        CommandList *commandList = &mem->graphics.textRenderPass;
        RenderTarget *target = NULL;
        SetDrawRenderTarget(commandList, TEXT_SETUP_KEY, target);
        SetViewport(commandList, TEXT_SETUP_KEY, V2i(0, 0), V2i(mem->graphics.width, mem->graphics.height));
    }


#if ETC_DEBUG
    ClearCommandList(&DebugMem->debugDrawPass);

    {
        CommandList *commandList = &DebugMem->debugDrawPass;
        RenderTarget *target = NULL;
        SetDrawRenderTarget(commandList, 0, target);
        SetViewport(commandList, 0, V2i(0, 0), V2i(mem->graphics.width, mem->graphics.height));
    }
#endif


    // 3) scene.update()
    UIManagerStartFrame(&mem->uiManager);
    Plat.ShowMouse();
    

#if ETC_DEBUG
    DebugUpdate(mem);
#endif


    // Unpaused
    if (!mem->paused || mem->steppingFrame) {
        mem->steppingFrame = false;
        mem->deltaTime = mem->platformDeltaTime;

        // @temp - remove  once EtcGameMemory is global
        {
            Game.time = mem->time;
            Game.deltaTime = mem->deltaTime;
            Game.frame = mem->frame;
        }
        
        RunActiveScenesUpdate(Game.sceneManager, Game.mem, false);

        if (InputActionPressed(&mem->inputActions, InputAction_Pause)) {
            PauseGame();
        }

        mem->pausedOnPrevFrame = false;
    }
    // Paused
    else {
        if (!mem->pausedOnPrevFrame) {
            mem->timePaused = mem->platformTime;
            mem->framePaused = mem->platformFrame;
        }

        mem->deltaTime = 0.0f;

        // @temp - remove once EtcGameMemory is global
        {
            Game.time = mem->time;
            Game.deltaTime = mem->deltaTime;
            Game.frame = mem->frame;
        }

        RunActiveScenesUpdate(Game.sceneManager, Game.mem, true);

        if (InputActionPressed(&mem->inputActions, InputAction_Pause)) {
            UnpauseGame();
        }
        else if (InputActionPressed(&mem->inputActions, InputAction_StepFrame)) {
            StepFrame();
        }

        mem->pausedOnPrevFrame = true;
    }


    // 5) Scene Independent Rendering
    RenderTextBuffers(mem->graphics.textRenderBuffers);

#if ETC_DEBUG
    DebugRender(mem);
#endif

    {
        uint32 key = TEXT_SETUP_KEY;
        
        EnableDepthWrite(&mem->graphics.textRenderPass, key);
        ClearDepth(&mem->graphics.textRenderPass, key, 1.0f);
    }


    
    RenderMode effectiveRenderMode = RenderMode_Deferred;
    if (Game.sceneManager->active.sceneCount > 0) {
        if (SceneData *activeScene = GetScene(Game.sceneManager->active.scenes[0])) {
            effectiveRenderMode = activeScene->settings.renderMode;
        }
    }
#if ETC_DEBUG
    if (DebugMem->shouldOverrideRenderMode) {
        effectiveRenderMode = DebugMem->renderModeOverride;
    }
#endif

    switch (effectiveRenderMode) {
        case RenderMode_DeferredNoLighting: {
            uint32 shadingBlitKey = GenerateShadingRenderKey(true, 0, ShadingStage_Blit);

            // Blit gbuffer diffuse directly to screen
            if (mem->sceneManager.active.shouldBlend) {
                for (int i = 0; i < mem->sceneManager.active.sceneCount; i++) {
                    uint32 preLightKey = GenerateShadingRenderKey(false, i, ShadingStage_PreLightAccumulation);
                    uint32 postLightKey = GenerateShadingRenderKey(false, i, ShadingStage_PostLightAccumulation);

                    // Set color mask so no lights can render into accumulation buffer
                    SetColorMask(&mem->graphics.shadingPass, preLightKey, 0, false);
                    SetColorMask(&mem->graphics.shadingPass, postLightKey, 0, true);

                    uint32 renderedDiffuseTarget = mem->graphics.gbufferTargets[i].colorTargets[GBUFFER_DIFFUSE_BUFFER];
                    RenderPassthrough(&mem->graphics.shadingPass, postLightKey, renderedDiffuseTarget);
                }

                uint32 renderedShadingTargets[] = {
                    mem->graphics.shadingTargets[0].colorTargets[0],
                    mem->graphics.shadingTargets[1].colorTargets[0],
                    mem->graphics.shadingTargets[2].colorTargets[0],
                    mem->graphics.shadingTargets[3].colorTargets[0],
                };

                SetDrawRenderTarget(&mem->graphics.shadingPass, shadingBlitKey, NULL);
                RenderPassthroughBlended(&mem->graphics.shadingPass, shadingBlitKey, renderedShadingTargets, mem->sceneManager.active.sceneCount, mem->graphics.sceneBlendMap.textureID);
            }
            else {
                uint32 preLightKey = GenerateShadingRenderKey(false, Game.sceneManager->active.renderedIndex, ShadingStage_PreLightAccumulation);
                uint32 postLightKey = GenerateShadingRenderKey(false, Game.sceneManager->active.renderedIndex, ShadingStage_PostLightAccumulation);

                // Set color mask so no lights can render into accumulation buffer
                SetColorMask(&mem->graphics.shadingPass, preLightKey, 0, false);
                SetColorMask(&mem->graphics.shadingPass, postLightKey, 0, true);

                // Copy gbuffer diffuse to shadingTarget
                uint32 renderedDiffuseTarget = mem->graphics.gbufferTargets[Game.sceneManager->active.renderedIndex].colorTargets[GBUFFER_DIFFUSE_BUFFER];
                RenderPassthrough(&mem->graphics.shadingPass, postLightKey, renderedDiffuseTarget);

                uint32 renderedShadingTarget = mem->graphics.shadingTargets[Game.sceneManager->active.renderedIndex].colorTargets[0];
                SetDrawRenderTarget(&mem->graphics.shadingPass, shadingBlitKey, NULL);
                RenderPassthrough(&mem->graphics.shadingPass, shadingBlitKey, renderedShadingTarget);
            }
        } break;

        case RenderMode_DeferredNormals: {
            uint32 shadingBlitKey = GenerateShadingRenderKey(true, 0, ShadingStage_Blit);

            // Blit gbuffer normal directly to screen
            if (mem->sceneManager.active.shouldBlend) {
                uint32 renderedTargetIDs[] = {
                    mem->graphics.gbufferTargets[0].colorTargets[GBUFFER_NORMAL_BUFFER],
                    mem->graphics.gbufferTargets[1].colorTargets[GBUFFER_NORMAL_BUFFER],
                    mem->graphics.gbufferTargets[2].colorTargets[GBUFFER_NORMAL_BUFFER],
                    mem->graphics.gbufferTargets[3].colorTargets[GBUFFER_NORMAL_BUFFER],
                };

                SetDrawRenderTarget(&mem->graphics.shadingPass, shadingBlitKey, NULL);
                RenderPassthroughNormalBlended(&mem->graphics.shadingPass, shadingBlitKey, renderedTargetIDs, mem->sceneManager.active.sceneCount, mem->graphics.sceneBlendMap.textureID);
            }
            else {
                uint32 renderedNormalTarget = mem->graphics.gbufferTargets[Game.sceneManager->active.renderedIndex].colorTargets[GBUFFER_NORMAL_BUFFER];

                SetDrawRenderTarget(&mem->graphics.shadingPass, shadingBlitKey, NULL);
                RenderPassthroughNormal(&mem->graphics.shadingPass, shadingBlitKey, renderedNormalTarget);
            }
        } break;

        case RenderMode_Deferred: {
            uint32 shadingBlitKey = GenerateShadingRenderKey(true, 0, ShadingStage_Blit);

            if (mem->sceneManager.active.shouldBlend) {
                uint32 renderedShadingTargets[] = {
                    mem->graphics.shadingTargets[0].colorTargets[0],
                    mem->graphics.shadingTargets[1].colorTargets[0],
                    mem->graphics.shadingTargets[2].colorTargets[0],
                    mem->graphics.shadingTargets[3].colorTargets[0],
                };

                SetDrawRenderTarget(&mem->graphics.shadingPass, shadingBlitKey, NULL);
                RenderPassthroughBlended(&mem->graphics.shadingPass, shadingBlitKey, renderedShadingTargets, mem->sceneManager.active.sceneCount, mem->graphics.sceneBlendMap.textureID);
            }
            else {
                uint32 renderedTarget = mem->graphics.shadingTargets[Game.sceneManager->active.renderedIndex].colorTargets[0];
                SetDrawRenderTarget(&mem->graphics.shadingPass, shadingBlitKey, NULL);
                RenderPassthrough(&mem->graphics.shadingPass, shadingBlitKey, renderedTarget);
            }
        } break;

        case RenderMode_Manual: {
        } break;
    }


    // 7) Sort and Submit CommandLists
    switch (effectiveRenderMode) {
        case RenderMode_DeferredNoLighting: {
            PushBack(&mem->graphics.submitted, &mem->graphics.gbufferPass);
            PushBack(&mem->graphics.submitted, &mem->graphics.shadingPass);
            PushBack(&mem->graphics.submitted, &mem->graphics.textRenderPass);
        } break;

        case RenderMode_DeferredNormals: {
            PushBack(&mem->graphics.submitted, &mem->graphics.gbufferPass);
            PushBack(&mem->graphics.submitted, &mem->graphics.shadingPass);
            PushBack(&mem->graphics.submitted, &mem->graphics.textRenderPass);
        } break;

        case RenderMode_Deferred: {
            PushBack(&mem->graphics.submitted, &mem->graphics.gbufferPass);
            PushBack(&mem->graphics.submitted, &mem->graphics.shadingPass);
            PushBack(&mem->graphics.submitted, &mem->graphics.textRenderPass);
        } break;

        case RenderMode_Manual: {
        } break;
    }


#if ETC_DEBUG
    PushBack(&mem->graphics.submitted, &DebugMem->debugDrawPass);
#endif

    PushBack(&mem->graphics.submitted, &mem->uiManager.commandList);

    

    for (int i = 0; i < mem->graphics.submitted.count; i++) {
        SortCommandList(mem->graphics.submitted[i]);
    }

    // 8) Execute CommandLists
    for (int i = 0; i < mem->graphics.submitted.count; i++) {
        ExecuteCommandList(mem->graphics.submitted[i]);
    }


    // 9) Frame Cleanup
    DynamicArrayClear(&mem->graphics.submitted);

    // Flush input
    InputDeviceFlush(&mem->keyboard, mem->deltaTime);
    InputDeviceFlush(&mem->mouse, mem->deltaTime);
    mem->mouse.analogue[Etc_ScrollDirection] = 0.0f;

    for (int i = 0; i < mem->gamepadCount; i++) {
        InputDeviceFlush(mem->gamepads + i, mem->deltaTime);
    }

    ClearMemoryArenaList(&mem->frameMem);


    if (!mem->paused || mem->steppingFrame) {
        mem->time += mem->platformDeltaTime;
        mem->frame++;

        //@NOTE @BUG: anything with a TIMED_BLOCK that still runs during the pause wont have its counter reset
        // producing bogus cycle counts for blocks still running during the pause
        // We need to do something different if paused, the problem is that we dont have the GameAPI in some places
        // where the profiler is needed (like mixer.cpp)
#if ETC_DEBUG
        UpdateDebugRecords();
#endif
    }

    mem->platformTime += mem->platformDeltaTime;
    mem->platformFrame++;


    // Teardown
    if (!mem->running) {
        UnloadAllScenes();
    }
}

void WriteSoundSamples(EtcGameMemory *mem, uint32 sampleCount, real32 *output) {
    if (MixerNode *masterGroup = GetMixerNode(&mem->mixer, mem->mixer.masterGroup)) {
        masterGroup->volume = mem->userPreferences.volume;
    }

    WriteSamples(&mem->mixer, sampleCount, output);
}

