
#ifndef _ETC_ETCETERA_H
#define _ETC_ETCETERA_H


// @-Tag Legend
// @Todo
// @Note
// @Hack/@Guac
// @Bug
// @Maybe
// @Warning
// @Leak

#ifdef _WIN32
#define _CRT_SECURE_NO_WARNINGS

// enumerator 'identifier' in a switch of enum 'enumeration' is not handled
#pragma warning(error : 4062)

#endif

#include "stb_sprintf.h"

#include "etc_types.h"
#include "platform.h"
#include "commonlib.h"
#include "math/math.h"


#define PERMANENT_ARENA_STORAGE_SIZE Megabytes(32)
#define SCENE_ARENA_STORAGE_SIZE Megabytes(4)

#define AUDIO_SAMPLERATE 44100
#define AUDIO_NUM_CHANNELS 2
#define AUDIO_BITS_PER_SAMPLE 32

#define FRAMES_PER_SECOND 60

#define MAX_NUM_GAMEPADS 4

// @GACK
#define DEFAULT_DEPTH_FAIL_SCALE 0.15f


struct EtcGameMemory;
struct EtcGraphics;

#include "metaprogramming.h"

#include "etc_memory.h"

#include "hashtable.h"
#include "dynamic_array.h"
#include "chunked_array.h"
#include "buffered_list.h"

#include "filehandle.h"
#include "task.h"

#include "io/utf8.h"
#include "io/etctxt.h"

#include "schema.h"

#include "serialized_types.generated.h"
#include "tags.generated.h"

#include "input.h"

#include "assets.h"

#include "render.h"
#include "mesh.h"

#include "ecs.h"
#include "ecs_core.h"

#include "transform.h"
#include "shapes.h"

#include "camera.h"

#include "light.h"

#include "ui.h"

#include "scenes.h"

#include "mixer.h"

#include "game.h"

// Serialized functions go at the end because they can reference types defined anywhere!!
#include "serialized_function_types.generated.h"


#if ETC_DEBUG
#include "debug.h"
#endif


// UserPreferences
#define USERDATA_DIRECTORY "userdata"
#define USERPREFERENCES_PATH "userdata/preferences.etctxt"

#define DEFAULT_KEYBINDINGS_PATH "data/keybindings/default.keybindings"

STRUCT()
struct UserPreferences {
    SceneID lastScene;
    const char *keybindingsPath;

    uint32 screenWidth;
    uint32 screenHeight;
    vec2i screenPosition;

    real32 volume;
    real32 cameraSpeed;
};

inline UserPreferences DefaultUserPreferences() {
    UserPreferences preferences = {};

    // @NOTE @LEAK: do a StringCopy() so that if the game code gets
    // reloaded the string is guaranteed valid
    preferences.keybindingsPath = StringCopy(DEFAULT_KEYBINDINGS_PATH);

    preferences.screenWidth = 1600;
    preferences.screenHeight = 900;
    preferences.screenPosition = V2i(128, 128);

    preferences.volume = 1.0f;

    preferences.cameraSpeed = 1;

    return preferences;
}

void SerializeUserPreferences(const char *path, UserPreferences *preferences);
UserPreferences DeserializeUserPreferences(const char *path);


// Graphics Stuff
struct OpenGLInfo {
    const uint8 *vendor;
    const uint8 *renderer;
    const uint8 *version;
    const uint8 *shadingLanguageVersion;
    const uint8 *extensions;
};

#define MAX_SUBMITTED_COMMANDLIST_COUNT 8

struct EtcGraphics {
    OpenGLInfo opengl;

    uint32 width;
    uint32 height;
    vec2 dimensions;
    real32 aspect;

    uint32 cameraUniformBufferID;

    uint32 screenUniformBufferID;

    Sampler nearestClampEdge;
    Sampler linearClampEdge;
    Sampler linearRepeat;

    RenderTarget gbufferTargets[MAX_ACTIVE_SCENES];
    RenderTarget shadingTargets[MAX_ACTIVE_SCENES];

    CommandList gbufferPass;
    CommandList shadingPass;
    CommandList textRenderPass;

    DynamicArray<CommandList *> submitted;

    uint32 globalVAOID;


    ShaderProgramHandle constantColorShader;

    ShaderProgramHandle skyboxShader;

    ShaderProgramHandle textShader;
    Mesh glyphQuad;
    TextRenderBuffer *textRenderBuffers;

    Texture sceneBlendMap;

#if ETC_DEBUG
    ShaderProgramHandle deferredDirectionalLightShader;
    ShaderProgramHandle deferredPointLightShader;
    ShaderProgramHandle deferredSpotlightShader;
#endif

    ShaderProgramHandle defaultUIMeshShader;
    ShaderProgramHandle defaultUITexturedMeshShader;
    ShaderProgramHandle defaultUITextShader;
};

struct EtcGameMemory {
    bool32 running;

    bool restart;

    real32 platformTime;
    real32 platformDeltaTime;
    int32 platformFrame;

    real32 time;
    real64 deltaTime;
    int32 frame;

    bool paused;
    bool pausedOnPrevFrame;
    bool steppingFrame;
    real32 timePaused;
    int32 framePaused;

    EtcGraphics graphics;
    SoundMixer mixer;

    EtcInputEventQueue inputQueue;
    EtcInputDevice keyboard;
    EtcInputDevice mouse;
    uint32 gamepadCount;
    EtcInputDevice gamepads[MAX_NUM_GAMEPADS];
    InputActions inputActions;

    MemoryArena permanentArena;
    MemoryArenaList frameMem;
    EtcHeap heap;

    AssetDB assets;

    SceneManager sceneManager;
    UIManager uiManager;

    UserPreferences userPreferences;

    TaskSchedulerTLSData mainThreadTaskSchedulerTLSData;
};


void InitEtcGameMemory(EtcGameMemory *mem);
void DeinitEtcGameMemory(EtcGameMemory *mem);
void UpdateAndRender(EtcGameMemory *mem);

void WriteSoundSamples(EtcGameMemory *mem, uint32 sampleCount, real32 *output);

#endif

