
#ifndef _ETC_FILEHANDLE_H
#define _ETC_FILEHANDLE_H

#ifndef _WIN32
#include <unistd.h>
#endif

#include <stdarg.h>

struct FileHandle {
    void *data;
    uint64 dataLength;

    uint64 fileSize;
    uint64 offset;

#if _WIN32
    HANDLE handle;
    HANDLE mapHandle;
#else
    int32 fileDescriptor;
#endif
};


inline void FileSeek(FileHandle *handle, uint64 offset) {
    ASSERT(handle->data != NULL);

    if (offset > handle->fileSize) {
        offset = handle->fileSize;
    }

    handle->offset = offset;
}

inline void FileClear(FileHandle *handle) {
    ASSERT(handle->data != NULL);

    handle->fileSize = 0;
    handle->offset = 0;

    Plat.ResizeFile(handle, 0, handle->dataLength);
}

inline uint32 ReadBytes(FileHandle *handle, uint64 count, void *output) {
    ASSERT(handle->data != NULL);

    if (count == 0) {
        return 0;
    }
    
    ASSERT(output != NULL);

    uint64 bytesLeft = handle->fileSize - handle->offset;
    uint32 bytesToRead = count;

    if (bytesLeft < bytesToRead) {
        bytesToRead = bytesLeft;
    }

    Plat.CopyMem(output, (uint8 *)handle->data + handle->offset, bytesToRead);
    handle->offset += bytesToRead;

    return bytesToRead;
}

inline int8 ReadInt8(FileHandle *handle) {
    int8 result = 0;
    ReadBytes(handle, 1, &result);

    return result;
}

inline int16 ReadInt16(FileHandle *handle) {
    int16 result = 0;

    int8 bytes[2];
    uint32 bytesRead = ReadBytes(handle, 2, &bytes);

    if (bytesRead == 2) {
        for (int i = 0; i < 2; i++) {
            int16 byte = bytes[2 - i - 1] & 0xFF;
            result = (result << 8) | byte;
        }
    }

    return result;
}

inline int32 ReadInt32(FileHandle *handle) {
    int32 result = 0;

    int8 bytes[4];
    uint32 bytesRead = ReadBytes(handle, 4, &bytes);

    if (bytesRead == 4) {
        for (int i = 0; i < 4; i++) {
            int32 byte = bytes[4 - i - 1] & 0xFF;
            result = (result << 8) | byte; 
        }
    }

    return result;
}

inline int32 ReadInt64(FileHandle *handle) {
    int32 result = 0;

    int8 bytes[8];
    uint32 bytesRead = ReadBytes(handle, 8, &bytes);

    if (bytesRead == 8) {
        for (int i = 0; i < 8; i++) {
            int32 byte = bytes[8 - i - 1] & 0xFF;
            result = (result << 8) | byte; 
        }
    }

    return result;
}


inline uint8 ReadUInt8(FileHandle *handle) {
    uint8 result = 0;
    ReadBytes(handle, 1, &result);

    return result;
}

inline uint16 ReadUInt16(FileHandle *handle) {
    uint16 result = 0;

    int8 bytes[2];
    uint32 bytesRead = ReadBytes(handle, 2, &bytes);

    if (bytesRead == 2) {
        for (int i = 0; i < 2; i++) {
            int16 byte = bytes[2 - i - 1] & 0xFF;
            result = (result << 8) | byte;
        }
    }

    return result;
}

inline uint32 ReadUInt32(FileHandle *handle) {
    uint32 result = 0;

    int8 bytes[4];
    uint32 bytesRead = ReadBytes(handle, 4, &bytes);

    if (bytesRead == 4) {
        for (int i = 0; i < 4; i++) {
            uint32 byte = bytes[4 - i - 1] & 0xFF;
            result = (result << 8) | byte; 
        }
    }

    return result;
}

inline uint32 ReadUInt64(FileHandle *handle) {
    uint32 result = 0;

    int8 bytes[8];
    uint32 bytesRead = ReadBytes(handle, 8, &bytes);

    if (bytesRead == 8) {
        for (int i = 0; i < 8; i++) {
            uint32 byte = bytes[8 - i - 1] & 0xFF;
            result = (result << 8) | byte; 
        }
    }

    return result;
}

inline real32 ReadReal32(FileHandle *handle) {
    int32 intValue = 0;
    int8 bytes[4];
    uint32 bytesRead = ReadBytes(handle, 4, &bytes);

    if (bytesRead == 4) {
        for (int i = 0; i < 4; i++) {
            int32 byte = bytes[4 - i - 1] & 0xFF;
            intValue = (intValue << 8) | byte; 
        }
    }

    real32 result;
    Plat.CopyMem(&result, &intValue, sizeof(real32));
    return result;
}

inline u32 ReadBytesUntilWhitespace(FileHandle *handle, char *word) {
    u32 len = 0;
    char c = 0;
    while (!IsWhitespace(c)) {
        uint32 bytesRead = ReadBytes(handle, 1, &c);

        if (bytesRead == 0) { break; }

        if (!IsWhitespace(c)) {
            word[len++] = c;    
        }
    }

    return len;
}


inline void WriteInt8(FileHandle *handle, int8 value) {
    ASSERT(handle->data != NULL);

    uint64 newOffset = handle->offset + 1;

    uint64 newFileSize = Max(newOffset, handle->fileSize);
    uint64 newDataLength = Max(newOffset, handle->dataLength);
    b32 resizeSuccess = Plat.ResizeFile(handle, newFileSize, newDataLength);
    ASSERT(resizeSuccess);

    uint8 byte = value & 0xFF;
    *((uint8 *)handle->data + handle->offset++) = byte;
}

inline void WriteInt16(FileHandle *handle, int16 value) {
    ASSERT(handle->data != NULL);

    uint64 newOffset = handle->offset + 2;

    uint64 newFileSize = Max(newOffset, handle->fileSize);
    uint64 newDataLength = Max(newOffset, handle->dataLength);
    b32 resizeSuccess = Plat.ResizeFile(handle, newFileSize, newDataLength);
    ASSERT(resizeSuccess);

    for (int i = 0; i < 2; i++) {
        uint8 byte = (value & (0xFF << (8 * i))) >> (8 * i);
        *((uint8 *)handle->data + handle->offset++) = byte;
    }
}

inline void WriteInt32(FileHandle *handle, int32 value) {
    ASSERT(handle->data != NULL);

    uint64 newOffset = handle->offset + 4;

    uint64 newFileSize = Max(newOffset, handle->fileSize);
    uint64 newDataLength = Max(newOffset, handle->dataLength);
    b32 resizeSuccess = Plat.ResizeFile(handle, newFileSize, newDataLength);
    ASSERT(resizeSuccess);

    for (int i = 0; i < 4; i++) {
        uint8 byte = (value & (0xFF << (8 * i))) >> (8 * i);
        *((uint8 *)handle->data + handle->offset++) = byte;
    }
}

inline void WriteInt64(FileHandle *handle, int64 value) {
    ASSERT(handle->data != NULL);

    uint64 newOffset = handle->offset + 8;

    uint64 newFileSize = Max(newOffset, handle->fileSize);
    uint64 newDataLength = Max(newOffset, handle->dataLength);
    b32 resizeSuccess = Plat.ResizeFile(handle, newFileSize, newDataLength);
    ASSERT(resizeSuccess);

    for (int i = 0; i < 8; i++) {
        uint8 byte = (value & (0xFF << (8 * i))) >> (8 * i);
        *((uint8 *)handle->data + handle->offset++) = byte;
    }
}

inline void WriteUInt8(FileHandle *handle, uint8 value) {
    ASSERT(handle->data != NULL);

    uint64 newOffset = handle->offset + 1;

    uint64 newFileSize = Max(newOffset, handle->fileSize);
    uint64 newDataLength = Max(newOffset, handle->dataLength);
    b32 resizeSuccess = Plat.ResizeFile(handle, newFileSize, newDataLength);
    ASSERT(resizeSuccess);

    uint8 byte = value & 0xFF;
    *((uint8 *)handle->data + handle->offset++) = byte;
}

inline void WriteUInt16(FileHandle *handle, uint16 value) {
    ASSERT(handle->data != NULL);

    uint64 newOffset = handle->offset + 2;

    uint64 newFileSize = Max(newOffset, handle->fileSize);
    uint64 newDataLength = Max(newOffset, handle->dataLength);
    b32 resizeSuccess = Plat.ResizeFile(handle, newFileSize, newDataLength);
    ASSERT(resizeSuccess);

    for (int i = 0; i < 2; i++) {
        uint8 byte = (value & (0xFF << (8 * i))) >> (8 * i);
        *((uint8 *)handle->data + handle->offset++) = byte;
    }
}

inline void WriteUInt32(FileHandle *handle, uint32 value) {
    ASSERT(handle->data != NULL);

    uint64 newOffset = handle->offset + 4;

    uint64 newFileSize = Max(newOffset, handle->fileSize);
    uint64 newDataLength = Max(newOffset, handle->dataLength);
    b32 resizeSuccess = Plat.ResizeFile(handle, newFileSize, newDataLength);
    ASSERT(resizeSuccess);

    for (int i = 0; i < 4; i++) {
        uint8 byte = (value & (0xFF << (8 * i))) >> (8 * i);
        *((uint8 *)handle->data + handle->offset++) = byte;
    }
}

inline void WriteUInt64(FileHandle *handle, uint64 value) {
    ASSERT(handle->data != NULL);

    uint64 newOffset = handle->offset + 8;

    uint64 newFileSize = Max(newOffset, handle->fileSize);
    uint64 newDataLength = Max(newOffset, handle->dataLength);
    b32 resizeSuccess = Plat.ResizeFile(handle, newFileSize, newDataLength);
    ASSERT(resizeSuccess);

    for (int i = 0; i < 8; i++) {
        uint8 byte = (value & (0xFF << (8 * i))) >> (8 * i);
        *((uint8 *)handle->data + handle->offset++) = byte;
    }
}


inline void WriteReal32(FileHandle *handle, real32 value) {
    ASSERT(handle->data != NULL);

    uint64 newOffset = handle->offset + 4;

    uint64 newFileSize = Max(newOffset, handle->fileSize);
    uint64 newDataLength = Max(newOffset, handle->dataLength);
    b32 resizeSuccess = Plat.ResizeFile(handle, newFileSize, newDataLength);
    ASSERT(resizeSuccess);

    uint8 bytes[4];
    Plat.CopyMem(bytes, &value, sizeof(real32));

    for (int i = 0; i < 4; i++) {
        *((uint8 *)handle->data + handle->offset++) = bytes[i];
    }
}

inline void WriteBytes(FileHandle *handle, void *data, uint32 size) {
    ASSERT(handle->data != NULL);

    uint64 newOffset = handle->offset + size;

    uint64 newFileSize = Max(newOffset, handle->fileSize);
    uint64 newDataLength = Max(newOffset, handle->dataLength);
    b32 resizeSuccess = Plat.ResizeFile(handle, newFileSize, newDataLength);
    ASSERT(resizeSuccess);

    Plat.CopyMem((char *)handle->data + handle->offset, data, size);
    handle->offset = newOffset;
}

inline void WriteString(FileHandle *handle, const char *str) {
    WriteBytes(handle, (void *)str, strlen(str));
}

inline char *WriteStringf_Callback(char *buffer, FileHandle *handle, int32 length) {

    uint64 bytesLeftInFile = handle->fileSize - handle->offset;
    uint64 bytesLeftInData = handle->dataLength - handle->offset;

    uint64 newFileSize = handle->fileSize;
    uint64 newDataLength = handle->dataLength;

    if (bytesLeftInFile < length) {
        newFileSize += (length - bytesLeftInFile);
    }

    if (bytesLeftInData < length) {
        newDataLength += (length - bytesLeftInData);
    }

    if (newFileSize != handle->fileSize || newDataLength != handle->dataLength) {
        b32 resizeSuccess = Plat.ResizeFile(handle, newFileSize, newDataLength);
        ASSERT(resizeSuccess);
    }

    // Copy buffer into the file
    Plat.CopyMem((void *)((uint8 *)handle->data + handle->offset), buffer, length);
    handle->offset += length;

    return buffer;
}

inline void WriteStringf(FileHandle *handle, const char *fmt, ...) {
    ASSERT(handle->data != NULL);

    char buffer[STB_SPRINTF_MIN] = {};

    va_list args;
    va_start(args, fmt);

    stbsp_vsprintfcb((STBSP_SPRINTFCB *)WriteStringf_Callback, handle, buffer, fmt, args);

    va_end(args);
}

inline void WriteStringfv(FileHandle *handle, const char *fmt, va_list args) {
    ASSERT(handle->data != NULL);

    char buffer[STB_SPRINTF_MIN] = {};
    stbsp_vsprintfcb((STBSP_SPRINTFCB *)WriteStringf_Callback, handle, buffer, fmt, args);
}


inline void WriteBufferToFile(const char *path, void *buffer, uint32 length) {
    FileHandle file;
    if (Plat.OpenFileForWrite(path, &file, length)) {
        WriteBytes(&file, buffer, length);
        Plat.CloseFile(&file);
    }
}

#endif

