
// ComponentType
const char *GetComponentName(ComponentType type) {
    const char *result = Game.components.humanNames[type];
    return result;
}

const char *GetComponentTypeName(ComponentType type) {
    const char *result = Game.components.typeNames[type];
    return result;
}

ComponentType GetComponentType(const char *componentName) {
    uint32 type = ComponentType_Invalid;

    for (uint32 i = 1; i < Game.components.count; i++) {
        if (StringEquals(Game.components.humanNames[i], componentName) ||
            StringEquals(Game.components.typeNames[i], componentName)) {
            type = i;
            break;
        }
    }

    return (ComponentType)type;
}


// Persistent Functions
uint32 GetPersistentFunctionIndex(PersistentFunctionTable *table, const char *functionName, PersistentFuncType funcType) {
    ASSERT(functionName != NULL);
    ASSERT(funcType != PersistentFuncType_Invalid);
    
    uint32 result = 0;

    if (functionName == NULL) {
        return result;
    }

    if (HashTableGet(&table->nameMap, functionName, &result)) {
        // @NOTE: we dont want to assert, because the UI want to test a name for all possible types
        if (table->table[result].funcType != funcType) {
            result = 0;
        }
    }

    return result;
}

uint32 GetPersistentFunctionIndex(PersistentFunctionTable *table, void *functionPtr, PersistentFuncType funcType) {
    ASSERT(functionPtr != NULL);
    ASSERT(funcType != PersistentFuncType_Invalid);

    uint32 result = 0;

    if (HashTableGet(&table->functionPtrMap, (uintptr)functionPtr, &result)) {
        ASSERT(table->table[result].funcType == funcType);
    }

    return result;
}

void *GetPersistentFunctionPtr(PersistentFunctionTable *table, uint32 index, PersistentFuncType funcType) {
    ASSERT(index != PersistentFunctionIndex_Invalid);
    ASSERT(funcType != PersistentFuncType_Invalid);

    void *functionPtr = table->table[index].functionPtr;
    return functionPtr;
}

const char *GetPersistentFunctionName(PersistentFunctionTable *table, uint32 index, PersistentFuncType funcType) {
    ASSERT(index != PersistentFunctionIndex_Invalid);
    ASSERT(funcType != PersistentFuncType_Invalid);

    const char *name = table->table[index].name;
    return name;
}

const char *GetPersistentFunctionRequiredSchemaDataTypeName(PersistentFunctionTable *table, uint32 index, PersistentFuncType funcType) {
    ASSERT(index != PersistentFunctionIndex_Invalid);
    ASSERT(funcType != PersistentFuncType_Invalid);

    const char *name = table->table[index].requiredSchemaDataTypeName;
    return name;    
}

void ClearPersistentFunctionPointers(PersistentFunctionTable *table) {
    for (int i = 0; i < table->table.count; i++) {
        table->table[i].functionPtr = NULL;
    }
}

void RegisterPersistentFunction(PersistentFunctionTable *table, const char *name, void *functionPtr, PersistentFuncType funcType, const char *requiredSchemaDataTypeName) {

    uint32 index;
    if (HashTableGet(&table->nameMap, name, &index)) {
        table->table[index].name = name;
        table->table[index].functionPtr = functionPtr;
        table->table[index].funcType = funcType;

        table->table[index].requiredSchemaDataTypeName = requiredSchemaDataTypeName;

        HashTableInsert(&table->nameMap, name, index, true);
        HashTableInsert(&table->functionPtrMap, (uintptr)functionPtr, index, true);
    }
    else {
        if (table->table.count == 0) {
            PersistentFunctionTableEntry entry = {};
            PushBack(&table->table, entry);
        }
        
        uint32 index = table->table.count;
        PersistentFunctionTableEntry entry = {};
        entry.name = name;
        entry.functionPtr = functionPtr;
        entry.funcType = funcType;
        entry.requiredSchemaDataTypeName = requiredSchemaDataTypeName;

        PushBack(&table->table, entry);
        HashTableInsert(&table->nameMap, name, index, true);
        HashTableInsert(&table->functionPtrMap, (uintptr)functionPtr, index, true);
    }
}

b32 FuncHandleIsValid_(FuncHandle funcHandle) {
    return (funcHandle.tableIndex != PersistentFunctionIndex_Invalid &&
            funcHandle.funcType != PersistentFuncType_Invalid &&
            funcHandle.funcType < ENUM_VALUECOUNT(PersistentFuncType));
}

// Schemas
Schema *GetSchema(const char *typeName) {
    Schema *result = NULL;
    b32 foundSchema = HashTableGet(&Game.schemas.schemaMap, typeName, &result);

    return result;
}

Schema *GetSchemaNamed_(SchemaTables *tables, const char *typeName) {
    Schema *result = NULL;

    if(!HashTableGet(&tables->schemaMap, typeName, &result)) {
        if (tables->schemas.arenaSize == 0) {
            AllocateMemoryArenaList(&tables->schemas, sizeof(Schema) * 256);
        }

        result = PushStruct(&tables->schemas, Schema);
        HashTableInsert(&tables->schemaMap, typeName, result, true);
    }

    return result;
}

void CopySchemaData(SchemaData *dest, SchemaData *src) {
    if (src->schemaName != NULL) {
        Schema *schema = GetSchema(src->schemaName);
        if (schema != NULL) {
            dest->schemaName = String(src->schemaName);
            dest->data = Plat.AllocateMem(schema->sizeInBytes);
            Plat.CopyMem(dest->data, src->data, schema->sizeInBytes);
        }
    }
}


// API Linkage
ComponentTables GenerateComponentTables() {
    ComponentTables tables = {};
    tables.count = ENUM_VALUECOUNT(ComponentType);

    tables.typeNames = ComponentTypeNameTable;
    tables.humanNames = ComponentHumanNameTable;

    tables.addComponentFuncs = AddComponentFuncTable;
    tables.removeComponentFuncs = RemoveComponentFuncTable;
    tables.getComponentValueFuncs = GetComponentValueFuncTable;
    tables.setComponentValueFuncs = SetComponentValueFuncTable;
    tables.tickSystemFuncs = TickSystemFuncTable;

    tables.raycastSystemFuncs = RaycastSystemFuncTable;

    return tables;
}

void DeallocateComponentTables(ComponentTables *tables) {
}

GameAPI ReloadGameAPI(EtcGameMemory *mem) {
    GameAPI result = {};

    result.mem = mem;

    result.assets = &mem->assets;
    result.graphics = &mem->graphics;
    result.sceneManager = &mem->sceneManager;
    result.uiManager = &mem->uiManager;

    result.mixer = &mem->mixer;
    result.inputActions = &mem->inputActions;
    result.keyboard = &mem->keyboard;
    result.mouse = &mem->mouse;
    result.gamepad0 = &mem->gamepads[0];

    result.frameMem = &mem->frameMem;
    result.heap = &mem->heap;

    result.InitEtcGameMemory = InitEtcGameMemory;
    result.DeinitEtcGameMemory = DeinitEtcGameMemory;
    result.UpdateAndRender = UpdateAndRender;

    result.WriteSoundSamples = WriteSoundSamples;

    result.GetComponentName = GetComponentName;
    result.GetComponentTypeName = GetComponentTypeName;
    result.GetComponentType = GetComponentType;

    result.GetSchema = GetSchema;

    
    ComponentTables componentTables = GenerateComponentTables();

    PersistentFunctionTable persistentFunctionTable = Game.persistentFunctionTable;
    ClearPersistentFunctionPointers(&persistentFunctionTable);

    // @NOTE: This function is generated by the preprocessor
    RegisterPersistentFunctions(&persistentFunctionTable);

    SchemaTables schemaTables = Game.schemas;
    RegisterSchemas(&schemaTables);

    result.components = componentTables;
    result.persistentFunctionTable = persistentFunctionTable;
    result.schemas = schemaTables;

    DeallocateComponentTables(&Game.components);

    
    Game = result;

    return result;
}

void PauseGame() {
    EtcGameMemory *mem = Game.mem;

    if (!mem->paused) {
        mem->paused = true;
    }
}

void PauseGame(b32 condition) {
    if (condition) {
        PauseGame();
    }
}

void UnpauseGame() {
    EtcGameMemory *mem = Game.mem;

    if (mem->paused) {
        mem->paused = false;
    }
}

void StepFrame() {
    EtcGameMemory *mem = Game.mem;

    if (mem->paused) {
        mem->steppingFrame = true;
    }
}

