
#ifndef _ETC_GAME_H
#define _ETC_GAME_H


// ComponentTables
typedef void AddComponentFunc(EntityDB *entityDB, EtcSystem *system, uint32 count, EntityID *entityIDs);
typedef void RemoveComponentFunc(EntityDB *entityDB, EtcSystem *system, EntityID entityID);
typedef b32 GetComponentValueFunc(EntityDB *entityDB, EtcSystem *system, EntityID entityID, ComponentValue *component);
typedef void SetComponentValueFunc(EntityDB *entityDB,EtcSystem *system, EntityID entityID, ComponentValue *component);

typedef void TickSystemFunc(EtcGameMemory *mem, EntityDB *entityDB, EtcSystem *system);


struct ComponentTables {
    uint32 count;

    const char **typeNames;
    const char **humanNames;

    AddComponentFunc **addComponentFuncs;
    RemoveComponentFunc **removeComponentFuncs;
    GetComponentValueFunc **getComponentValueFuncs;
    SetComponentValueFunc **setComponentValueFuncs;

    TickSystemFunc **tickSystemFuncs;
    RaycastSystemFunc **raycastSystemFuncs;
};


// PersistentFunctionTable
struct PersistentFunctionTableEntry {
    const char *name;
    void *functionPtr;
    PersistentFuncType funcType;

    const char *requiredSchemaDataTypeName;
};

struct PersistentFunctionTable {
    DynamicArray<PersistentFunctionTableEntry> table;

    // Maps name/functionPtr -> index in table
    HashTable<const char *, uint32> nameMap;
    HashTable<uintptr, uint32> functionPtrMap;
};

const uint32 PersistentFunctionIndex_Invalid = 0;

uint32 GetPersistentFunctionIndex(PersistentFunctionTable *table, const char *functionName, PersistentFuncType funcType);
uint32 GetPersistentFunctionIndex(PersistentFunctionTable *table, void *functionPtr, PersistentFuncType funcType);
void *GetPersistentFunctionPtr(PersistentFunctionTable *table, uint32 index, PersistentFuncType funcType);

const char *GetPersistentFunctionName(PersistentFunctionTable *table, uint32 index, PersistentFuncType funcType);

const char *GetPersistentFunctionRequiredSchemaDataTypeName(PersistentFunctionTable *table, uint32 index, PersistentFuncType funcType);

void RegisterPersistentFunction(PersistentFunctionTable *table, const char *name, void *functionPtr, PersistentFuncType funcType, const char *requiredSchemaDataTypeName);

b32 FuncHandleIsValid_(FuncHandle funcHandle);
#define FuncHandleIsValid(funcHandle) FuncHandleIsValid_(*((FuncHandle *)&(funcHandle)))

// SchemaTables
struct SchemaTables {
    MemoryArenaList schemas;
    HashTable<const char *, Schema *> schemaMap;
};

void RegisterSchemas(SchemaTables *tables);
Schema *GetSchemaNamed_(SchemaTables *tables, const char *typeName);
void CopySchemaData(SchemaData *dest, SchemaData *src);


// Game API
typedef void InitEtcGameMemoryFunc(EtcGameMemory *mem);
typedef void DeinitEtcGameMemoryFunc(EtcGameMemory *mem);
typedef void UpdateAndRenderFunc(EtcGameMemory *mem);

typedef void WriteSoundSamplesFunc(EtcGameMemory *mem, uint32 sampleCount, real32 *output);

typedef const char *GetComponentNameFunc(ComponentType type);
typedef const char *GetComponentTypeNameFunc(ComponentType type);
typedef ComponentType GetComponentTypeFunc(const char *name);

typedef Schema *GetSchemaFunc(const char *name);


struct GameAPI {
    EtcGameMemory *mem;

    EtcGraphics *graphics;
    AssetDB *assets;
    SceneManager *sceneManager;
    UIManager *uiManager;

    SoundMixer *mixer;

    InputActions *inputActions;
    EtcInputDevice *keyboard;
    EtcInputDevice *mouse;
    EtcInputDevice *gamepad0;

    MemoryArenaList *frameMem;
    EtcHeap *heap;

    // @NOTE: these values won't be correct in the platform layer unless they are specifically updated
    real32 time;
    real32 deltaTime;
    int32 frame;


    // @NOTE: this section should really be on a pointer
    b32 shouldRecenterMousePrevFrame;
    b32 shouldRecenterMouse;

    
    InitEtcGameMemoryFunc *InitEtcGameMemory;
    DeinitEtcGameMemoryFunc *DeinitEtcGameMemory;
    UpdateAndRenderFunc *UpdateAndRender;

    WriteSoundSamplesFunc *WriteSoundSamples;

    GetComponentNameFunc *GetComponentName;
    GetComponentNameFunc *GetComponentTypeName;
    GetComponentTypeFunc *GetComponentType;

    GetSchemaFunc *GetSchema;


    ComponentTables components;
    PersistentFunctionTable persistentFunctionTable;
    SchemaTables schemas;
};

// @TODO: Game should be a pointer to EtcGameMemory instead!
extern "C" GameAPI Game;
extern GameAPI Game;

typedef GameAPI ReloadGameAPIFunc(EtcGameMemory *mem);
//extern "C" DLLEXPORT GameAPI ReloadGameAPI(EtcGameMemory *mem);
extern "C" DLLEXPORT GameAPI ReloadGameAPI(EtcGameMemory *mem);


void PauseGame();
void PauseGame(b32 condition);
void UnpauseGame();
void StepFrame();

#endif

