
// The SCENE() macro is a preprocessor directive that generates some
// boilerplate and registers the scene with the engine. This also
// means there must be a BreakoutInit(), BreakoutDeinit(), and
// BreakoutUpdate() function somwhere in the project. In this case,
// they are defined at the bottom of this file.

SCENE(Breakout);


static real32 BreakoutPhysicsHz = 60.0f;
static real32 BreakoutMSPerTick = 1.0f / BreakoutPhysicsHz;

static int32 BreakoutMaxLevel = 9;

static int32 BreakoutBrickClearPoints = 5;
static int32 BreakoutBrickHitPoints = 1;

static real32 BreakoutPaddlePosY = 1.5f;
static real32 BreakoutPaddleWidth0 = 4.0f;
static real32 BreakoutPaddleWidth1 = 1.0f;
static real32 BreakoutPaddleHeight = 0.25f;
static real32 BreakoutPaddleDepth = 0.0625f;

static real32 BreakoutPaddleAcceleration = 1200.0f;
static real32 BreakoutMaxPaddleSpeed = 24.0f;

static vec3 BreakoutBrickSize = V3(1.0f, 0.25f, 0.0625f);
static vec3 BreakoutBrickHitSize = V3(1.25f, 0.3125f, 0.0625f);
static vec3 BreakoutBrickDeadStartSize = V3(1.25f, 0.3125f, 0.0625f);
static vec3 BreakoutBrickDeadEndSize = V3(1.0f, 0.0f, 0.0f);

static real32 BreakoutMinBallSpeed = 10.0f;
static real32 BreakoutMaxBallSpeed0 = 19.0f;
static real32 BreakoutMaxBallSpeed1 = 24.0f;
static real32 BreakoutMinBallVerticalSpeed = 4.0f;

static real32 BreakoutBallSpinDamping = 0.98f;
static real32 BreakoutBallCollisionSpinDamping = 0.85f;

static vec3 BreakoutBallServePosition = V3(16.0f, 10.0f, 0.0f);
static real32 BreakoutServeSpeed = 13.0f;

static real32 BreakoutBallTrailTicks = 24;
static real32 BreakoutBallTrailMinTime = BreakoutMSPerTick * 4;
static real32 BreakoutBallTrailMaxTime = BreakoutMSPerTick * (BreakoutBallTrailTicks - 1);


static real32 BreakoutBrickRestitution = 0.1f;
static real32 BreakoutWallVelocityDamping = 1.05f;


struct BreakoutAnimationID {
    uint32 index;
    uint32 generation;
};

inline BreakoutAnimationID BreakoutAnimationIDInvalid() {
    BreakoutAnimationID id;
    id.index = 0;
    id.generation = 0;

    return id;
}

union BreakoutAnimationData {
    real32 f32;
    vec2 v2;
    vec3 v3;
    vec4 v4;

    real32 data[4];
};

struct BreakoutAnimation {
    uint32 componentCount;
    real32 delay;
    real32 duration;

    BreakoutAnimationData start;
    BreakoutAnimationData end;

    void *result;
};

struct BreakoutAnimationSlot {
    uint32 generation;
    real32 startTime;
    bool active;
    BreakoutAnimation anim;
};

struct BreakoutAnimationManager {
    ChunkedArray<BreakoutAnimationSlot> animations;
    ChunkedArray<uint32> freeAnimationSlots;
};


struct Paddle {
    EntityID id;

    vec3 position;
    vec3 halfExtents;
    vec4 color;

    vec3 prevTickPosition;
    vec3 prevTickHalfExtents;
    vec4 prevTickColor;

    vec3 velocity;

    Material *material;
    BreakoutAnimationID sizeAnimID;
    BreakoutAnimationID colorAnimID;
};

struct CachedPosition {
    uint32 tick;
    real32 time;
    vec3 position;
};

struct Ball {
    EntityID id;

    real32 radius;

    // Previous ball positions are stored to render a nice trail behind the ball
    DynamicArray<CachedPosition> positions;

    vec3 velocity;
    real32 spin;

    vec3 scale;
    vec4 color;

    vec3 prevTickScale;
    vec4 prevTickColor;

    real32 trailTime;

    int32 trailStartIndex;
    vec3 sampledPosition;
    vec3 sampledScale;
    vec4 sampledColor;

    Material *material;
    BreakoutAnimationID sizeAnimID;
    BreakoutAnimationID colorAnimID;
};

struct Brick {
    EntityID id;

    vec3 position;
    vec3 halfExtents;
    vec4 color;

    vec3 prevTickPosition;
    vec3 prevTickHalfExtents;
    vec4 prevTickColor;

    int32 hitCount;
    int32 hitCountMax;

    b32 invincible;
    real32 invincibleUntilTime;

    Material *material;
    BreakoutAnimationID sizeAnimID;
    BreakoutAnimationID colorAnimID;
};

struct BreakoutLevel {
    EtcRect playArea;
    DynamicArray<Brick> bricks;
    uint32 bricksCleared;

    real32 maxBallSpeed;
};


struct BreakoutPalette {
    // Specifies that the background color is dark and spooky
    bool darkTheme;

    // All colors in the palette are HSV and have hue equal to `hue`
    real32 hue;

    vec4 backgroundColor;
    vec4 textColor;
    vec4 brickColor0;
    vec4 brickColor1;

    vec4 hitColor;
};


// Palettes are generated randomly based on the previous palette to
// ensure there is contrast. Hue is chosen randomly, roughly opposite
// the previous hue on the color wheel, and darkTheme is inverted.

BreakoutPalette GenerateNextPalette(BreakoutPalette prev) {
    BreakoutPalette palette;

    palette.hue = NormalizeAngleDeg(prev.hue + 180 + Randf(90.0f));
    palette.darkTheme = !prev.darkTheme;

    if (palette.darkTheme) {
        palette.backgroundColor = V4(palette.hue, 0.5f, 0.15f, 1.0f);
        palette.textColor = V4(palette.hue, 1.0f, 0.85f, 1.0f);
        palette.brickColor0 = V4(palette.hue, 1.0f, 0.85f, 1.0f);
        palette.brickColor1 = V4(palette.hue, 1.0f, 0.15f, 1.0f);
        palette.hitColor = V4(palette.hue, 0.6f, 0.9f, 1.0f);
    }
    else {
        palette.backgroundColor = V4(palette.hue, 0.5f, 1.0f, 1.0f);
        palette.textColor = V4(palette.hue, 1.0f, 0.85f, 1.0f);
        palette.brickColor0 = V4(palette.hue, 1.0f, 0.85f, 1.0f);
        palette.brickColor1 = V4(palette.hue, 1.0f, 0.15f, 1.0f);
        palette.hitColor = V4(palette.hue, 0.92f, 0.95f, 1.0f);
    }

    return palette;
}


// A simple state machine drives all logic in the main game loop (BreakoutUpdate())
enum BreakoutGameState {
    BreakoutGameState_Serve,
    BreakoutGameState_Play,
    BreakoutGameState_GameOver,
};

struct Breakout {
    EntityID cameraID;

    BreakoutGameState gameState;
    real32 stateStartTime;

    real32 timeAccumulator;
    uint32 tick;

    Paddle paddle;
    Ball ball;
    BreakoutLevel level;

    int32 levelIndex;
    int32 servesLeft;
    real32 gameTime;
    real32 playTime;
    real32 lastServeTime;

    int32 score;
    int32 comboMultiplier;

    BreakoutPalette prevPalette;
    BreakoutPalette palette;

    BreakoutAnimationManager animationManager;

    ShaderProgramHandle trailShader;
};


real32 BreakoutCurrentTime() {
    if (SceneData *scene = GetScene(Game.sceneManager->activeScene)) {
        Breakout *breakout = (Breakout *)scene->data;
        return breakout->gameTime;
    }

    return 0.0f;
}

void SetGameState(Breakout *breakout, BreakoutGameState gameState) {
    breakout->gameState = gameState;
    breakout->stateStartTime = breakout->gameTime;
}


// Animations in Breakout are performed by a fairly simple manager
// that can linearly interpolate groups of up to four floating point
// values at once. An animation is specified by a BreakoutAnimation
// struct and submitted to the StartAnimation() function, which
// returns an id that uniquely identifies an animation.

// The UpdateAnimations() function updates each active animation and
// writes the result into an arbititrary pointer specified by the
// BreakoutAnimation description. Writing to an arbitrary pointer
// submitted by the user is potentially dangerous, however, and should
// be handled with care.

// While admittedly barebones, this system adequately handles the
// sparse animations needed for Breakout without imposing too much
// complexity.

void BreakoutInitAnimationManager(BreakoutAnimationManager *manager, EtcAllocator *allocator) {
    manager->animations = MakeChunkedArray<BreakoutAnimationSlot>(allocator, 16);
    manager->freeAnimationSlots = MakeChunkedArray<uint32>(allocator, 16);
}

uint32 NextAnimationIndex(BreakoutAnimationManager *manager) {
    if (manager->freeAnimationSlots.count > 0) {
        uint32 index = manager->freeAnimationSlots[0];
        RemoveAtIndexBySwap(&manager->freeAnimationSlots, 0);

        ASSERT(!manager->animations[index].active);

        return index;
    }

    BreakoutAnimationSlot slot = {};
    uint32 index = PushBack(&manager->animations, slot);
    return index;
}

BreakoutAnimationID StartAnimation(BreakoutAnimationManager *manager, BreakoutAnimation anim) {
    if (anim.componentCount < 1 || anim.componentCount > 4) {
        return BreakoutAnimationIDInvalid();
    }
    
    uint32 index = NextAnimationIndex(manager);
    BreakoutAnimationSlot *slot = &manager->animations[index];

    slot->generation++;
    slot->startTime = BreakoutCurrentTime();
    slot->anim = anim;
    slot->active = true;

    BreakoutAnimationID id;
    id.index = index;
    id.generation = slot->generation;

    return id;
}

bool StopAnimation(BreakoutAnimationManager *manager, BreakoutAnimationID id) {
    if (id.index >= manager->animations.count) {
        return false;
    }

    BreakoutAnimationSlot *slot = &manager->animations[id.index];

    if (id.generation != slot->generation) {
        return false;
    }

    if (!slot->active) {
        return false;
    }

    slot->active = false;
    PushBack(&manager->freeAnimationSlots, id.index);

    return true;
}

void UpdateAnimations(BreakoutAnimationManager *manager, EntityDB *entityDB) {
    ChunkedArray<BreakoutAnimationSlot> animations = manager->animations;

    real32 gameTime = BreakoutCurrentTime();
    
    for (int i = 0; i < animations.count; i++) {
        BreakoutAnimationSlot slot = manager->animations[i];
        if (!slot.active) {
            continue;
        }

        real32 t = 1.0f;
        if (gameTime < slot.startTime + slot.anim.delay) {
            t = 0.0f;
        }
        else if (slot.anim.duration > 0.0f) {
            t = Clamp((gameTime - slot.startTime - slot.anim.delay) / slot.anim.duration,
                      0.0f, 1.0f);
        }

        // @TODO: currently, all animations are linear. A small set of
        // curve types that are simple to evaluate (quadratic easing,
        // smoothstep, etc) would be a nice extension to the animation
        // system

        BreakoutAnimationData value;

        for (int i = 0; i < slot.anim.componentCount; i++) {
            value.data[i] = Lerp(slot.anim.start.data[i], slot.anim.end.data[i], t);
        }

        if (slot.anim.result != NULL) {
            Plat.CopyMem(slot.anim.result, value.data, sizeof(real32) * slot.anim.componentCount);
        }

        if (gameTime >= slot.startTime + slot.anim.delay + slot.anim.duration) {
            BreakoutAnimationID id;
            id.index = i;
            id.generation = slot.generation;

            bool stopped = StopAnimation(manager, id);
            ASSERT(stopped);
        }
    }
}


// Returns the camera transform needed to ensure playArea is entirely
// in frame. Assumes that the aspect ratio of playArea (width /
// height) is greater than 1.0
Transform BreakoutCameraTransform(EntityDB *entityDB, EntityID cameraID, EtcRect playArea) {
    const real32 padding = 1.0f;

    CameraInfo cameraInfo = GetCameraInfo(entityDB, cameraID);

    Transform transform;
    transform.position.x = 0.5f * (playArea.pos.x + playArea.size.x);
    transform.position.y = 0.5f * (playArea.size.y + padding);
    transform.position.z = 0.5f * (playArea.size.y + padding) / tanf(cameraInfo.fov * 0.5f);
    transform.rotation = AxisAngle(V3(0, 1, 0), _PI);
    transform.scale = V3(1);

    return transform;
}


Paddle SpawnPaddle(Breakout *breakout, EntityDB *entityDB) {
    Paddle paddle = {};
    paddle.id = PushEntityID(entityDB);
    paddle.halfExtents = V3(BreakoutPaddleWidth0, BreakoutPaddleHeight, BreakoutPaddleDepth);
    paddle.prevTickHalfExtents = paddle.halfExtents;
    paddle.position = V3(16.0f, BreakoutPaddlePosY, 0.0f);
    paddle.prevTickPosition = paddle.position;
    paddle.velocity = V3(0);

    TransformComponent transform = {};
    transform.position = paddle.position;
    transform.rotation = IdentityQuaternion();
    transform.scale = paddle.halfExtents;

    AddComponent(entityDB, Transform, 1, &paddle.id, &transform);


    StaticMeshComponent staticMesh = {};
    staticMesh.mesh = DebugMem->cube;
    staticMesh.shader = Game.graphics->constantColorShader;

    Material mat = {};
    PushBack(&staticMesh.overrideMaterials, mat);

    AddComponent(entityDB, StaticMesh, 1, &paddle.id, &staticMesh);

    paddle.material = &staticMesh.overrideMaterials[0];

    return paddle;
}

Ball SpawnBall(Breakout *breakout, EntityDB *entityDB) {
    Ball ball = {};
    ball.id = PushEntityID(entityDB);
    ball.radius = 0.25f;
    ball.scale = V3(ball.radius, 0.0f, ball.radius);

    ball.prevTickScale = ball.scale;

    TransformComponent transform = {};
    transform.position = BreakoutBallServePosition;
    transform.rotation = QuaternionWithXY(V3(1, 0, 0), V3(0, 0, 1));
    transform.scale = ball.scale;

    AddComponent(entityDB, Transform, 1, &ball.id, &transform);


    StaticMeshComponent staticMesh = {};
    staticMesh.mesh = DebugMem->sphere;
    staticMesh.shader = Game.graphics->constantColorShader;

    Material mat = {};
    PushBack(&staticMesh.overrideMaterials, mat);

    AddComponent(entityDB, StaticMesh, 1, &ball.id, &staticMesh);

    ball.material = &staticMesh.overrideMaterials[0];

    ball.sampledPosition = transform.position;
    ball.sampledScale = ball.scale;

    return ball;
}

void DestroyBall(EntityDB *entityDB, Ball ball) {
    DeleteEntity(entityDB, ball.id);
    DeallocateDynamicArray(&ball.positions);
}

Brick SpawnBrick(Breakout *breakout, EntityDB *entityDB, vec3 position, int32 hitCountMax) {
    Brick brick = {};
    brick.id = PushEntityID(entityDB);
    brick.position = position;
    brick.halfExtents = BreakoutBrickSize;

    brick.prevTickPosition = brick.position;
    brick.prevTickHalfExtents = brick.halfExtents;

    brick.hitCount = 0;
    brick.hitCountMax = hitCountMax;

    TransformComponent transform = {};
    transform.position = brick.position;
    transform.rotation = IdentityQuaternion();
    transform.scale = brick.halfExtents;

    AddComponent(entityDB, Transform, 1, &brick.id, &transform);


    StaticMeshComponent staticMesh = {};
    staticMesh.mesh = DebugMem->cube;
    staticMesh.shader = Game.graphics->constantColorShader;

    Material mat = {};
    PushBack(&staticMesh.overrideMaterials, mat);

    AddComponent(entityDB, StaticMesh, 1, &brick.id, &staticMesh);


    brick.material = &staticMesh.overrideMaterials[0];

    return brick;
}

BreakoutLevel SpawnLevel(Breakout *breakout, EntityDB *entityDB, int32 levelIndex) {
    BreakoutLevel level = {};
    level.playArea = {V2(0, 0), V2(32, 19)};

    for (int y = 0; y < 2; y++) {
        for (int x = 0; x < 10; x++) {
            vec3 position = V3(2.5f + 3.0f * x, 17.0f - 1.125f * y, 0.0f);
            Brick brick = SpawnBrick(breakout, entityDB, position, levelIndex);

            int32 index = PushBack(&level.bricks, brick);
        }
    }

    level.maxBallSpeed = Lerp(BreakoutMaxBallSpeed0, BreakoutMaxBallSpeed1,
                              Clamp((levelIndex - 1.0f) / (BreakoutMaxLevel - 1.0f), 0.0f, 1.0f));

    return level;
}


void DestroyLevel(EntityDB *entityDB, BreakoutLevel *level) {
    for (int i = 0; i < level->bricks.count; i++) {
        DeleteEntity(entityDB, level->bricks[i].id);
    }

    DeallocateDynamicArray(&level->bricks);
}


void ApplyPalette(Breakout *breakout, BreakoutPalette palette) {
    vec4 textColorRGB = HSVToRGB(palette.textColor);
    vec4 brickColor0RGB = HSVToRGB(palette.brickColor0);
    
    breakout->paddle.color = palette.textColor;
    breakout->paddle.prevTickColor = palette.textColor;
    StopAnimation(&breakout->animationManager, breakout->paddle.colorAnimID);

    breakout->ball.color = palette.textColor;
    breakout->ball.prevTickColor = palette.textColor;
    breakout->ball.sampledColor = palette.textColor;
    StopAnimation(&breakout->animationManager, breakout->ball.colorAnimID);

    breakout->prevPalette = palette;
}


void StartSpawnLevelAnimations(Breakout *breakout, BreakoutLevel *level) {

    real32 midX = level->playArea.pos.x + 0.5f * level->playArea.size.x;
    
    for (int i = 0; i < level->bricks.count; i++) {
        Brick *brick = &level->bricks[i];

        real32 distance = Absf(brick->position.x - midX) / (level->playArea.size.x * 0.5f);
        real32 t = distance * distance;

        real32 duration = 0.125f;
        real32 delay = 0.5f * t + 0.25f;

        brick->invincible = true;
        brick->invincibleUntilTime = breakout->gameTime + delay + duration;

        BreakoutAnimation sizeAnim = {};
        sizeAnim.componentCount = 3;
        sizeAnim.start.v3 = BreakoutBrickDeadEndSize;
        sizeAnim.end.v3 = BreakoutBrickSize;
        sizeAnim.duration = duration;
        sizeAnim.delay = delay;
        sizeAnim.result = &brick->halfExtents;

        brick->halfExtents = BreakoutBrickDeadEndSize;
        brick->prevTickHalfExtents = BreakoutBrickDeadEndSize;

        StopAnimation(&breakout->animationManager, brick->sizeAnimID);
        brick->sizeAnimID = StartAnimation(&breakout->animationManager, sizeAnim);


        BreakoutAnimation colorAnim = {};
        colorAnim.componentCount = 4;
        colorAnim.start.v4 = breakout->palette.brickColor1;
        colorAnim.end.v4 = breakout->palette.brickColor0;
        colorAnim.duration = duration;
        colorAnim.delay = delay;
        colorAnim.result = &brick->color;

        brick->color = breakout->palette.brickColor1;
        brick->prevTickColor = breakout->palette.brickColor1;

        StopAnimation(&breakout->animationManager, brick->colorAnimID);
        brick->colorAnimID = StartAnimation(&breakout->animationManager, colorAnim);
    }
}

void StartPaddleSizeAnimation(Breakout *breakout, Paddle *paddle, int32 levelIndex) {
    real32 t = Clamp((levelIndex - 1.0f) / (BreakoutMaxLevel - 1.0f), 0.0f, 1.0f);
    t = 1.0f - (1.0f - t) * (1.0f - t);

    real32 targetWidth = Lerp(BreakoutPaddleWidth0, BreakoutPaddleWidth1, t);

    BreakoutAnimation sizeAnim = {};
    sizeAnim.componentCount = 3;
    sizeAnim.start.v3 = V3(paddle->halfExtents.x, BreakoutPaddleHeight, BreakoutPaddleDepth);
    sizeAnim.end.v3 = V3(targetWidth, BreakoutPaddleHeight, BreakoutPaddleDepth);
    sizeAnim.duration = 0.25f;
    sizeAnim.result = &paddle->halfExtents;

    StopAnimation(&breakout->animationManager, paddle->sizeAnimID);
    paddle->sizeAnimID = StartAnimation(&breakout->animationManager, sizeAnim);
}

void StartBallHitAnimation(Breakout *breakout, Ball *ball) {
    real32 speed = Length(ball->velocity);
    real32 speedRatio = speed / breakout->level.maxBallSpeed;

    real32 scale = Max(1.0f, 1.75f * speedRatio * speedRatio);
    
    BreakoutAnimation sizeAnim = {};
    sizeAnim.componentCount = 3;
    sizeAnim.start.v3 = V3(scale * ball->radius,
                           0.0f,
                           1.0f / scale * ball->radius);
    sizeAnim.end.v3 = V3(ball->radius, 0.0f, ball->radius);
    sizeAnim.duration = 0.6f;
    sizeAnim.result = &ball->scale;

    StopAnimation(&breakout->animationManager, ball->sizeAnimID);
    ball->sizeAnimID = StartAnimation(&breakout->animationManager, sizeAnim);


    BreakoutAnimation colorAnim = {};
    colorAnim.componentCount = 4;
    colorAnim.start.v4 = breakout->palette.hitColor;
    colorAnim.end.v4 = breakout->palette.textColor;
    colorAnim.duration = 0.6f;
    colorAnim.result = &ball->color;

    StopAnimation(&breakout->animationManager, ball->colorAnimID);
    ball->colorAnimID = StartAnimation(&breakout->animationManager, colorAnim);
}


void UpdatePaddle(Breakout *breakout, Paddle *paddle, real32 deltaTime) {
    SceneData *scene = GetActiveScene();
    TransformSystem *transformSystem = GetSystem(&scene->entityDB, Transform);

    BreakoutLevel *level = &breakout->level;

    paddle->prevTickPosition = paddle->position;
    paddle->prevTickHalfExtents = paddle->halfExtents;
    paddle->prevTickColor = paddle->color;

    static real32 BreakoutControllerDeadzone = 0.2f;

    vec3 impulse = V3(0, 0, 0);
    if (InputHold(Game.keyboard, Etc_LeftArrow) || InputHold(Game.keyboard, Etc_A) ||
        InputAnalogue(Game.gamepad0, Etc_LeftX) <= -BreakoutControllerDeadzone) {
        impulse.x -= BreakoutPaddleAcceleration;
    }

    if (InputHold(Game.keyboard, Etc_RightArrow) || InputHold(Game.keyboard, Etc_D) ||
        InputAnalogue(Game.gamepad0, Etc_LeftX) >= BreakoutControllerDeadzone) {
        impulse.x += BreakoutPaddleAcceleration;
    }


    paddle->velocity = paddle->velocity + impulse * deltaTime;

    real32 speed = Length(paddle->velocity);
    if (speed > BreakoutMaxPaddleSpeed) {
        paddle->velocity = Normalize(paddle->velocity) * BreakoutMaxPaddleSpeed;
    }

    paddle->position = paddle->position + deltaTime * paddle->velocity;
    paddle->velocity = paddle->velocity * 0.75f;

    real32 leftEdge = level->playArea.pos.x;
    real32 rightEdge = level->playArea.pos.x + level->playArea.size.x;
    if (paddle->position.x < leftEdge + paddle->halfExtents.x) {
        paddle->position.x = leftEdge + paddle->halfExtents.x;
        paddle->velocity = V3(0, 0, 0);
    }

    if (paddle->position.x > rightEdge - paddle->halfExtents.x) {
        paddle->position.x = rightEdge - paddle->halfExtents.x;
        paddle->velocity = V3(0, 0, 0);
    }
}

void UpdateBricks(Breakout *breakout, real32 deltaTime) {
    BreakoutLevel level = breakout->level;
    for (int i = 0; i < level.bricks.count; i++) {
        Brick *brick = &level.bricks[i];

        brick->prevTickPosition = brick->position;
        brick->prevTickHalfExtents = brick->halfExtents;
        brick->prevTickColor = brick->color;

        if (breakout->gameTime >= brick->invincibleUntilTime) {
            brick->invincible = false;
        }
    }
}


// The ball is the only object in the game that needs to do collision
// detection. The ball is modeled as a sphere and it only ever
// collides with straight line segments so the collision routines can
// be greatly simplified by considering the ball as a point and
// insetting all other surfaces by the ball's radius.

enum BreakoutCollisionType {
    BreakoutCollision_Wall,
    BreakoutCollision_Paddle,
    BreakoutCollision_Brick,
};

struct BreakoutCollision {
    BreakoutCollisionType type;
    int32 index;

    vec3 pointOfImpact;
    vec3 normal;
    real32 t;
};


// There are two fundamental collision detection routines:
// 1) PointPlaneCollision() - checks for collision between a moving point and a static plane
// 2) PointEdgeCollision() - checks for collision between a moving point and a translating/scaling (but not rotating) line segment

// Returns true when a collision is found between a point moving
// between p0 and p1 and a plane specified by <a, normal>
bool PointPlaneCollision(vec3 p0, vec3 p1, vec3 a, vec3 normal, BreakoutCollision *collision) {
    collision->normal = normal;

    if (Dot(p1 - a, normal) <= 0) {
        real32 d = Dot(p0 - a, normal);
        if (d > 0) {
            if (LinePlaneIntersection(p0, p1, a, normal, &collision->t, &collision->pointOfImpact)) {
                return true;
            }
        }
        else {
            collision->t = 0;
            collision->pointOfImpact = p0 - normal * d;

            return true;
        }
    }

    return false;
}

// Returns true when a collision is found between a point moving
// between p0 and p1 and a line segment transforming from <a0, b0> and
// <a1, b1>.
// Returns false if the normals of <a0, b0> and <a1, b1> are not equal.
bool PointEdgeCollision(vec3 p0, vec3 p1, vec3 a0, vec3 b0, vec3 a1, vec3 b1, BreakoutCollision *collision) {
    vec3 normal;

    if (!NearlyEquals(DistanceSq(a0, b0), 0.0f)) {
        vec3 edge = Normalize(b0 - a0);
        normal = Cross(V3(0, 0, 1), edge);

        if (!NearlyEquals(Dot(normal, b1 - a1), 0.0f)) {
            return false;
        }
    }
    else if (!NearlyEquals(DistanceSq(a1, b1), 0.0f)){
        vec3 edge = Normalize(b1 - a1);
        normal = Cross(V3(0, 0, 1), edge);

        if (!NearlyEquals(Dot(normal, b0 - a0), 0.0f)) {
            return false;
        }
    }
    else {
        return false;
    }


    if (Dot(p0 - a0, normal) > 0) {
        real32 denom = Dot(p1 - p0, normal) - Dot(a1 - a0, normal);
        real32 t = Dot(a0 - p0, normal) / denom;

        if (0.0f <= t && t <= 1.0f) {
            vec3 pt = Lerp(p0, p1, t);
            vec3 at = Lerp(a0, a1, t);
            vec3 bt = Lerp(b0, b1, t);

            // Ignore the collision if the point of impact is not on the edge
            if (Dot(pt - at, bt - at) < 0 ||
                Dot(pt - bt, at - bt) < 0) {
                return false;
            }

            collision->pointOfImpact = pt;
            collision->normal = normal;
            collision->t = t;
            
            return true;
        }
    }

    return false;
}

// Higher level routines can be built from the two core collision routines:

bool BallPlayAreaCollision(EtcRect playArea, Ball ball, vec3 position, BreakoutCollision *collision) {
    vec3 boundsVerts[] = {
        V3(playArea.pos + V2(playArea.size.x, 0.0f), 0.0f),
        V3(playArea.pos + playArea.size, 0.0f),
        V3(playArea.pos + V2(0.0f, playArea.size.y), 0.0f),
        V3(playArea.pos, 0.0f),
    };

    vec3 prevPosition = ball.positions[0].position;

    bool hitWall = false;
    real32 minT = INFINITY;

    for (int i = 0; i < 3; i++) {
        vec3 a = boundsVerts[i];
        vec3 b = boundsVerts[i + 1];

        vec3 normal = Normalize(Cross(V3(0, 0, 1), b - a));
        a = a + normal * ball.radius;
        b = b + normal * ball.radius;

        BreakoutCollision c;
        if (PointPlaneCollision(prevPosition, position, a, normal, &c)) {
            hitWall = true;

            if (c.t < minT) {
                minT = c.t;
                c.type = BreakoutCollision_Wall;
                *collision = c;
            }
        }
    }

    return hitWall;
}

bool BallPaddleCollision(Paddle paddle, Ball ball, vec3 position, real32 framePercent, BreakoutCollision *collision) {

    vec3 paddlePosition0 = Lerp(paddle.prevTickPosition, paddle.position, framePercent);
    vec3 paddleHalfExtents0 = Lerp(paddle.prevTickHalfExtents, paddle.halfExtents, framePercent);

    vec3 verts0[] = {
        paddlePosition0 + V3(-paddleHalfExtents0.x - ball.radius,
                             -paddleHalfExtents0.y - ball.radius,
                             0.0f),
        paddlePosition0 + V3(-paddleHalfExtents0.x - ball.radius,
                             paddleHalfExtents0.y + ball.radius,
                             0.0f),
        paddlePosition0 + V3(paddleHalfExtents0.x + ball.radius,
                             paddleHalfExtents0.y + ball.radius,
                             0.0f),
        paddlePosition0 + V3(paddleHalfExtents0.x + ball.radius,
                             -paddleHalfExtents0.y - ball.radius,
                             0.0f),
    };

    vec3 verts1[] = {
        paddle.position + V3(-paddle.halfExtents.x - ball.radius,
                             -paddle.halfExtents.y - ball.radius,
                             0.0f),
        paddle.position + V3(-paddle.halfExtents.x - ball.radius,
                             paddle.halfExtents.y + ball.radius,
                             0.0f),
        paddle.position + V3(paddle.halfExtents.x + ball.radius,
                             paddle.halfExtents.y + ball.radius,
                             0.0f),
        paddle.position + V3(paddle.halfExtents.x + ball.radius,
                             -paddle.halfExtents.y - ball.radius,
                             0.0f),
    };

    vec3 prevPosition = ball.positions[0].position;

    for (int i = 0; i < 3; i++) {
        if (PointEdgeCollision(prevPosition, position,
                               verts0[i], verts0[i + 1], verts1[i], verts1[i + 1],
                               collision)) {
            collision->type = BreakoutCollision_Paddle;
            collision->index = i;

            return true;
        }
    }

    return false;
}

bool BallBrickCollision(Brick brick, Ball ball, vec3 position, real32 framePercent, BreakoutCollision *collision) {
    if (brick.invincible) {
        return false;
    }

    if (brick.hitCount >= brick.hitCountMax) {
        return false;
    }

    vec3 brickPosition0 = Lerp(brick.prevTickPosition, brick.position, framePercent);
    vec3 brickHalfExtents0 = Lerp(brick.prevTickHalfExtents, brick.halfExtents, framePercent);

    vec3 verts0[] = {
        brickPosition0 + V3(-brickHalfExtents0.x - ball.radius,
                             -brickHalfExtents0.y - ball.radius,
                             0.0f),
        brickPosition0 + V3(-brickHalfExtents0.x - ball.radius,
                             brickHalfExtents0.y + ball.radius,
                             0.0f),
        brickPosition0 + V3(brickHalfExtents0.x + ball.radius,
                             brickHalfExtents0.y + ball.radius,
                             0.0f),
        brickPosition0 + V3(brickHalfExtents0.x + ball.radius,
                             -brickHalfExtents0.y - ball.radius,
                             0.0f),
    };

    vec3 verts1[] = {
        brick.position + V3(-brick.halfExtents.x - ball.radius,
                             -brick.halfExtents.y - ball.radius,
                             0.0f),
        brick.position + V3(-brick.halfExtents.x - ball.radius,
                             brick.halfExtents.y + ball.radius,
                             0.0f),
        brick.position + V3(brick.halfExtents.x + ball.radius,
                             brick.halfExtents.y + ball.radius,
                             0.0f),
        brick.position + V3(brick.halfExtents.x + ball.radius,
                             -brick.halfExtents.y - ball.radius,
                             0.0f),
    };

    vec3 prevPosition = ball.positions[0].position;

    for (int i = 0; i < 4; i++) {
        vec3 a0 = verts0[i];
        vec3 b0 = verts0[(i + 1) % 4];

        vec3 a1 = verts1[i];
        vec3 b1 = verts1[(i + 1) % 4];

        if (PointEdgeCollision(prevPosition, position, a0, b0, a1, b1, collision)) {
            return true;
        }
    }

    return false;
}

bool BallBricksCollision(DynamicArray<Brick> bricks, Ball ball, vec3 position, real32 framePercent, BreakoutCollision *collision) {

    bool result = false;
    real32 minT = INFINITY;

    for (int32 i = bricks.count - 1; i >= 0; i--) {
        BreakoutCollision c = {};

        if (BallBrickCollision(bricks[i], ball, position, framePercent, &c)) {
            if (c.t < minT) {
                c.type = BreakoutCollision_Brick;
                c.index = i;
                *collision = c;

                minT = c.t;
                result = true;
            }
        }
    }

    return result;
}

// Returns true when a collision is found between the ball and any
// other surface. Writes the closest found collision into `collision`
bool BallNextCollision(Breakout *breakout, Ball ball, vec3 position, real32 framePercent, BreakoutCollision *collision) {
    bool result = false;
    real32 minT = INFINITY;
    BreakoutCollision c = {};

    if (BallPlayAreaCollision(breakout->level.playArea, ball, position, &c)) {
        result = true;

        if (c.t < minT) {
            *collision = c;
        }
    }

    c = {};
    if (BallPaddleCollision(breakout->paddle, ball, position, framePercent, &c)) {
        result = true;

        if (c.t < minT) {
            *collision = c;
        }
    }

    c = {};
    if (BallBricksCollision(breakout->level.bricks, ball, position, framePercent, &c)) {
        result = true;

        if (c.t < minT) {
            *collision = c;
        }
    }

    return result;
}

void UpdateBall(Breakout *breakout, Ball *ball, real32 deltaTime) {
    SceneData *scene = GetActiveScene();
    EntityDB *entityDB = &scene->entityDB;
    TransformSystem *transformSystem = GetSystem(entityDB, Transform);

    BreakoutLevel *level = &breakout->level;

    ball->prevTickScale = ball->scale;
    ball->prevTickColor = ball->color;

    vec3 normal = Cross(Normalize(ball->velocity), V3(0, 0, 1));
    ball->velocity = ball->velocity + normal * ball->spin;
    ball->spin *= BreakoutBallSpinDamping;


    // Because the ball is the only object that needs to check
    // collisions, we can greedily search for collisions and resolve
    // them without worrying about the effects on other
    // objects. Simply "search forward" from the previous location of
    // the ball for the next collision, resolve it, and repeat until
    // the entire frame time has been spent.

    vec3 position = ball->positions[0].position + deltaTime * ball->velocity;

    real32 framePercent = 0.0f;
    uint32 collisionCount = 0;
    BreakoutCollision prevCollision;

    BreakoutCollision collision;
    while (framePercent < 1.0f - FLT_EPSILON &&
           BallNextCollision(breakout, *ball, position, framePercent, &collision)) {

        // Ignore collisions that are too close to the previous
        // collision. This would ocassionally result in an infinite
        // sequence of collisions that would hang the game.
        if (collisionCount > 0 &&
            NearlyEquals(Distance(collision.pointOfImpact, prevCollision.pointOfImpact), 0.0f)) {
            break;
        }
        collisionCount++;
        

        // Reflect the position and velocity about the collision normal
        position = position - 2 * Dot(position - collision.pointOfImpact, collision.normal) * collision.normal;
        ball->velocity = ball->velocity - 2 * Dot(ball->velocity, collision.normal) * collision.normal;
        ball->spin = ball->spin * -BreakoutBallCollisionSpinDamping;


        framePercent = framePercent + (1.0f - framePercent) * collision.t;

        real32 collisionTime = breakout->gameTime + BreakoutMSPerTick * framePercent;

        CachedPosition p;
        p.tick = breakout->tick;
        p.time = collisionTime;
        p.position = collision.pointOfImpact;
        InsertAtIndex(&ball->positions, 0, p);
        
        switch (collision.type) {
            case BreakoutCollision_Wall: {
                ball->velocity = ball->velocity * BreakoutWallVelocityDamping;
            } break;

            case BreakoutCollision_Paddle: {
                if (collision.index == 1) {
                    vec3 velocityNormal = Cross(Normalize(ball->velocity), V3(0, 0, 1));

                    real32 d = Dot(Normalize(ball->velocity), Normalize(breakout->paddle.velocity));
                    real32 spinDirection = Sign(Dot(velocityNormal, V3(0, 1, 0)));

                    real32 ballSpeed = Length(ball->velocity);
                    vec3 ballDirection = Normalize(ball->velocity);

                    real32 paddleSpeed = Length(breakout->paddle.velocity);
                    vec3 paddleDirection = Normalize(breakout->paddle.velocity);
                    real32 speedRatio = paddleSpeed / BreakoutMaxPaddleSpeed;

                    // "Spin" can be added to the ball by moving the
                    // paddle during the collision. A "tight shot" is
                    // achieved if the paddle is moving opposite the
                    // ball's horizontal direction and a "wide shot"
                    // is achieved by moving with the ball's
                    // horizontal direction.
                    
                    if (paddleSpeed >= 0.0f) {
                        // Tight shot
                        if (d <= 0.0f) {
                            ball->spin = (speedRatio - d) * 0.25f * spinDirection;
                            ball->velocity = 1.1f * ballSpeed * NLerp(ballDirection,
                                                                      paddleDirection,
                                                                      0.15f * speedRatio);
                        }
                        // Wide shot
                        else {
                            ball->spin = (speedRatio + d) * 0.1f * spinDirection;
                            ball->velocity = 0.95f * ballSpeed * NLerp(ballDirection,
                                                                       paddleDirection,
                                                                       0.55f * speedRatio);
                        }
                    }
                }
                else {
                    vec3 impulse = collision.normal * Dot(breakout->paddle.velocity, collision.normal);
                    ball->velocity = ball->velocity + impulse;
                }
                
                StartBallHitAnimation(breakout, ball);
                breakout->comboMultiplier = 1;
            } break;

            case BreakoutCollision_Brick: {
                Brick *brick = &level->bricks[collision.index];
                brick->hitCount++;

                vec3 impulse = BreakoutBrickRestitution * collision.normal * Dot(ball->velocity, collision.normal);
                ball->velocity = ball->velocity + impulse;

                if (brick->hitCount >= brick->hitCountMax) {
                    breakout->score += breakout->comboMultiplier * BreakoutBrickClearPoints;
                    level->bricksCleared++;

                    BreakoutAnimation sizeAnim = {};
                    sizeAnim.componentCount = 3;
                    sizeAnim.start.v3 = BreakoutBrickDeadStartSize;
                    sizeAnim.end.v3 = BreakoutBrickDeadEndSize;
                    sizeAnim.duration = 0.125f;
                    sizeAnim.result = &brick->halfExtents;

                    StopAnimation(&breakout->animationManager, brick->sizeAnimID);
                    brick->sizeAnimID = StartAnimation(&breakout->animationManager, sizeAnim);

                    BreakoutAnimation colorAnim = {};
                    colorAnim.componentCount = 4;
                    colorAnim.start.v4 = breakout->palette.hitColor;
                    colorAnim.end.v4 = breakout->palette.brickColor1;
                    colorAnim.duration = 0.125f;
                    colorAnim.result = &brick->color;

                    StopAnimation(&breakout->animationManager, brick->colorAnimID);
                    brick->colorAnimID = StartAnimation(&breakout->animationManager, colorAnim);
                }
                else {
                    breakout->score += breakout->comboMultiplier * BreakoutBrickHitPoints;

                    BreakoutAnimation sizeAnim = {};
                    sizeAnim.componentCount = 3;
                    sizeAnim.start.v3 = BreakoutBrickHitSize;
                    sizeAnim.end.v3 = BreakoutBrickSize;
                    sizeAnim.duration = 0.25f;
                    sizeAnim.result = &brick->halfExtents;

                    StopAnimation(&breakout->animationManager, brick->sizeAnimID);
                    brick->sizeAnimID = StartAnimation(&breakout->animationManager, sizeAnim);


                    BreakoutAnimation colorAnim = {};
                    colorAnim.componentCount = 4;
                    colorAnim.start.v4 = breakout->palette.hitColor;
                    colorAnim.end.v4 = Lerp(breakout->palette.brickColor0,
                                            breakout->palette.brickColor1,
                                            (real32)brick->hitCount / (real32)brick->hitCountMax);
                    colorAnim.duration = 0.25f;
                    colorAnim.result = &brick->color;

                    StopAnimation(&breakout->animationManager, brick->colorAnimID);
                    brick->colorAnimID = StartAnimation(&breakout->animationManager, colorAnim);
                }

                StartBallHitAnimation(breakout, ball);
                breakout->comboMultiplier++;
            } break;
        }

        prevCollision = collision;
    }


    // The ball's maximum speed is governed by the current
    // level. Since collisions can add velocity to the ball it must be
    // clamped each tick.
    real32 ballSpeed = Length(ball->velocity);
    if (ballSpeed < BreakoutMinBallSpeed) {
        ball->velocity = BreakoutMinBallSpeed * Normalize(ball->velocity);
    }
    else if (ballSpeed > level->maxBallSpeed) {
        ball->velocity = level->maxBallSpeed * Normalize(ball->velocity);
    }

    if (Absf(ball->velocity.y) < BreakoutMinBallVerticalSpeed) {
        ball->velocity.y = Sign(ball->velocity.y) * BreakoutMinBallVerticalSpeed;
    }


    CachedPosition p;
    p.tick = breakout->tick;
    p.time = breakout->gameTime + BreakoutMSPerTick;
    p.position = position;
    InsertAtIndex(&ball->positions, 0, p);


    // Remove outdated CachedPositions
    for (int i = ball->positions.count - 1; i >= 1; i--) {
        if (breakout->tick - ball->positions[i].tick > BreakoutBallTrailTicks) {
            RemoveAtIndex(&ball->positions, i);
        }
    }

    ball->trailTime = Lerp(BreakoutBallTrailMinTime, BreakoutBallTrailMaxTime,
                           ballSpeed / level->maxBallSpeed);
}


void BreakoutRestartGame(Breakout *breakout, EntityDB *entityDB) {
    breakout->levelIndex = 1;
    breakout->servesLeft = 3;
    breakout->playTime = 0.0f;
    breakout->score = 0;
    breakout->comboMultiplier = 1;

    breakout->palette = GenerateNextPalette(breakout->prevPalette);

    breakout->ball = SpawnBall(breakout, entityDB);
    breakout->level = SpawnLevel(breakout, entityDB, breakout->levelIndex);

    ApplyPalette(breakout, breakout->palette);

    StartPaddleSizeAnimation(breakout, &breakout->paddle, breakout->levelIndex);
    StartSpawnLevelAnimations(breakout, &breakout->level);

    SetGameState(breakout, BreakoutGameState_Serve);
}


// Rendering is handled by the StaticMeshComponent, which is internal
// to the engine. This system uses the TransformComponent of each
// entity to determine rendering location. Since the frame rate of the
// engine is variable and the physics of Breakout runs at a fixed rate
// the rendering will be choppy unless the rendered positions are
// sampled from the last two known states.

// This function performs the sampling and writes the results into the
// transforms of each entity that will be rendered.
void SampleEntityPositions(Breakout *breakout, EntityDB *entityDB, real32 alpha) {
    TransformSystem *transformSystem = GetSystem(entityDB, Transform);

    Paddle paddle = breakout->paddle;
    if (paddle.id != EntityID_Invalid) {
        SetWorldPosition(transformSystem, paddle.id,
                         Lerp(paddle.prevTickPosition, paddle.position, alpha));
        SetWorldScale(transformSystem, paddle.id,
                      Lerp(paddle.prevTickHalfExtents, paddle.halfExtents, alpha));
        paddle.material->diffuseColor = HSVToRGB(Lerp(paddle.prevTickColor, paddle.color, alpha));
    }
    

    BreakoutLevel level = breakout->level;
    for (int i = 0; i < level.bricks.count; i++) {
        Brick brick = level.bricks[i];
        if (brick.id == EntityID_Invalid) {
            continue;
        }

        SetWorldPosition(transformSystem, brick.id,
                         Lerp(brick.prevTickPosition, brick.position, alpha));
        SetWorldScale(transformSystem, brick.id,
                      Lerp(brick.prevTickHalfExtents, brick.halfExtents, alpha));
        brick.material->diffuseColor = HSVToRGB(Lerp(brick.prevTickColor, brick.color, alpha));
    }

    real32 trailStartTime = breakout->gameTime - BreakoutMSPerTick * (1.0f - alpha);

    Ball *ball = &breakout->ball;
    if (ball->id != EntityID_Invalid) {
        ball->trailStartIndex = -1;

        for (int i = 1; i < ball->positions.count; i++) {
            CachedPosition p1 = ball->positions[i];
            if (p1.time > trailStartTime) {
                continue;
            }

            CachedPosition p0 = ball->positions[i - 1];
            real32 t = InverseLerp(p0.time, p1.time, trailStartTime);

            ball->trailStartIndex = i;
            ball->sampledPosition = Lerp(p0.position, p1.position, t);
            SetWorldPosition(transformSystem, ball->id, ball->sampledPosition);

            quaternion rotation = QuaternionWithXY((p1.position - p0.position), V3(0, 0, 1));
            SetWorldRotation(transformSystem, ball->id, rotation);
        
            break;
        }

        ball->sampledColor = Lerp(ball->prevTickColor, ball->color, alpha);
        ball->material->diffuseColor = HSVToRGB(ball->sampledColor);

        ball->sampledScale = Lerp(ball->prevTickScale, ball->scale, alpha);
        SetWorldScale(transformSystem, ball->id, ball->sampledScale);
    }
}

void RenderBallTrail(Breakout *breakout, Ball ball, real32 alpha) {
    if (ball.trailStartIndex < 1) {
        return;
    }

    ShaderProgram *shader = GetShaderProgram(Game.assets, breakout->trailShader);
    if (shader == NULL) {
        return;
    }

    real32 trailStartTime = breakout->gameTime - BreakoutMSPerTick * (1 - alpha);
    real32 trailEndTime = Max(trailStartTime - ball.trailTime,
                              breakout->lastServeTime);

    vec4 startColor = ball.sampledColor;
    vec4 endColor = breakout->palette.backgroundColor;

    real32 trailWidth = ball.sampledScale.z - 0.01f;

    struct TrailPoint {
        vec3 pos0;
        vec3 pos1;
        vec3 pos2;
        vec4 color;
        real32 time;
    };

    TrailPoint *trailPoints = PushArray(Game.frameMem, TrailPoint, ball.positions.count);
    uint32 trailPointCount = 0;


    {
        vec3 tangent = (ball.positions[ball.trailStartIndex].position -
                        ball.positions[ball.trailStartIndex - 1].position);
        vec3 normal = Cross(Normalize(tangent), V3(0, 0, 1));

        TrailPoint p;
        p.pos0 = ball.sampledPosition;
        p.pos1 = p.pos0 - normal * trailWidth;
        p.pos2 = p.pos0 + normal * trailWidth;
        p.color = HSVToRGB(startColor);
        p.time = trailStartTime;

        trailPoints[trailPointCount++] = p;
    }


    TrailPoint prevPoint = trailPoints[0];
    bool shouldBreak = false;

    for (int i = ball.trailStartIndex; i < ball.positions.count; i++) {
        CachedPosition b = ball.positions[i];
        vec3 normal;

        if (i < ball.positions.count - 1) {
            CachedPosition c = ball.positions[i + 1];

            vec3 tangent0 = b.position - prevPoint.pos0;
            vec3 tangent1 = c.position - b.position;

            vec3 normal0 = Cross(Normalize(tangent0), V3(0, 0, 1));
            vec3 normal1 = Cross(Normalize(tangent1), V3(0, 0, 1));

            normal = Normalize(0.5f * normal0 + 0.5f * normal1);
        }
        else {
            CachedPosition b = ball.positions[i];
            vec3 tangent0 = b.position - prevPoint.pos0;

            normal = Cross(Normalize(tangent0), V3(0, 0, 1));
        }

        TrailPoint p;
        p.time = Max(trailEndTime, b.time);

        if (b.time <= trailEndTime) {
            shouldBreak = true;

            real32 t = InverseLerp(prevPoint.time, b.time, trailEndTime);
            p.pos0 = Lerp(prevPoint.pos0, b.position, t);
        }
        else {
            p.pos0 = b.position;
        }

        real32 t = (p.time - trailStartTime) / (trailEndTime - trailStartTime);
        real32 width = (1.0f - t) * trailWidth;
        
        p.pos1 = p.pos0 - normal * width;
        p.pos2 = p.pos0 + normal * width;
        p.color = HSVToRGB(Lerp(startColor, endColor, (1 - (1 - t) * (1 - t))));

        trailPoints[trailPointCount++] = p;
        prevPoint = p;

        if (shouldBreak) {
            break;
        }
    }


    Mesh mesh = Game.graphics->glyphQuad;

    CommandList *commandList = &Game.graphics->gbufferPass;
    uint32 renderKey = GenerateGBufferRenderKey(Game.sceneManager->activeSceneIndex,
                                                GBufferStage_Render);

    SetShader(commandList, renderKey, shader->programID);
    DisableDepthTest(commandList, renderKey);
    DisableFaceCulling(commandList, renderKey);
    DisableBlending(commandList, renderKey);

    BindBuffer(commandList, renderKey, GL_ARRAY_BUFFER, mesh.vboID);
    BindBuffer(commandList, renderKey, GL_ELEMENT_ARRAY_BUFFER, mesh.iboID);

    EnableVertexAttribute(commandList, renderKey,
                          Position_Attrib, 3, GL_FLOAT, GL_FALSE, 0,
                          POINTER_OFFSET(mesh.vertices, mesh.positions));

    for (int i = trailPointCount - 1; i >= 1; i--) {
        TrailPoint p0 = trailPoints[i];
        TrailPoint p1 = trailPoints[i - 1];

        SetUniform(commandList, renderKey, UniformLocation(shader, "prevPos1"), p0.pos1);
        SetUniform(commandList, renderKey, UniformLocation(shader, "prevPos2"), p0.pos2);
        SetUniform(commandList, renderKey, UniformLocation(shader, "pos1"), p1.pos1);
        SetUniform(commandList, renderKey, UniformLocation(shader, "pos2"), p1.pos2);

        SetUniform(commandList, renderKey, UniformLocation(shader, "prevColor"), p0.color);
        SetUniform(commandList, renderKey, UniformLocation(shader, "color"), p1.color);

        DrawIndexed(commandList, renderKey, mesh.primitiveType, 0, mesh.indexCount);
    }

    DisableVertexAttribute(commandList, renderKey, Position_Attrib);
}

void RenderGameInfo(Breakout *breakout) {
    BreakoutPalette palette = breakout->palette;
    vec4 textColorRGB = HSVToRGB(breakout->palette.textColor);

    uint32 renderKey = GenerateGBufferRenderKey(Game.sceneManager->activeSceneIndex,
                                                GBufferStage_Render);

    EtcRect playArea = breakout->level.playArea;
    vec3 playAreaVerts[] = {
        V3(playArea.pos, 0.0f),
        V3(playArea.pos + V2(playArea.size.x, 0.0f), 0.0f),
        V3(playArea.pos + playArea.size, 0.0f),
        V3(playArea.pos + V2(0.0f, playArea.size.y), 0.0f),
    };
    DebugDrawLine(playAreaVerts[1], playAreaVerts[2], textColorRGB);
    DebugDrawLine(playAreaVerts[2], playAreaVerts[3], textColorRGB);
    DebugDrawLine(playAreaVerts[3], playAreaVerts[0], textColorRGB);

    DebugDrawText(V3(0.0f, 19.25f, 0.0f), 0.5f, textColorRGB, "Serves Left: %d", breakout->servesLeft);
    DebugDrawText(V3(7.0f, 19.25f, 0.0f), 0.5f, textColorRGB, "Level: %d", breakout->levelIndex);
    DebugDrawText(V3(14.0f, 19.25f, 0.0f), 0.5f, textColorRGB, "Score: %d", breakout->score);
    DebugDrawText(V3(21.0f, 19.25f, 0.0f), 0.5f, textColorRGB, "Combo: x%d", breakout->comboMultiplier);

    vec3 timeTextPosition = V3(28.0f, 19.25f, 0.0f);
    if (breakout->playTime < 60.0f) {
        DebugDrawText(timeTextPosition, 0.5f, textColorRGB, "%.2f", breakout->playTime);
    }
    else if (breakout->playTime < 3600.0f) {
        int32 minutes = (int32)Floorf(breakout->playTime / 60.0f);
        real32 seconds = breakout->playTime - 60.0f * minutes;
        if (seconds < 10.0f) {
            DebugDrawText(timeTextPosition, 0.5f, textColorRGB, "%d:0%.2f", minutes, seconds);
        }
        else {
            DebugDrawText(timeTextPosition, 0.5f, textColorRGB, "%d:%.2f", minutes, seconds);
        }
    }
    else {
        int32 hours = (int32)Floorf(breakout->playTime / 3600.0f);
        int32 minutes = (int32)Floorf((breakout->playTime - 3600.0f * hours) / 60.0f);
        real32 seconds = breakout->playTime - (3600 * hours + 60 * minutes);
        if (seconds < 10.0f) {
            DebugDrawText(timeTextPosition, 0.5f, textColorRGB, "%d:%02d:0%.2f", hours, minutes, seconds);
        }
        else {
            DebugDrawText(timeTextPosition, 0.5f, textColorRGB, "%d:%02d:%.2f", hours, minutes, seconds);
        }
    }
}

void BreakoutInit(EtcGameMemory *mem, SceneData *scene) {
    EntityDB *entityDB = &scene->entityDB;
    Breakout *breakout = PushStruct(&scene->arena, Breakout);
    scene->data = breakout;

    BreakoutInitAnimationManager(&breakout->animationManager, &scene->arena);

    breakout->cameraID = GetSystem(entityDB, Camera)->mainCameraID;

    BreakoutPalette dummyPalette;
    dummyPalette.hue = Randf(360.0f);
    dummyPalette.darkTheme = true;
    breakout->prevPalette = dummyPalette;

    breakout->paddle = SpawnPaddle(breakout, entityDB);
    BreakoutRestartGame(breakout, entityDB);

    Transform cameraTransform = BreakoutCameraTransform(entityDB, breakout->cameraID, breakout->level.playArea);
    SetWorldTransform(GetSystem(entityDB, Transform), breakout->cameraID, cameraTransform);

    breakout->trailShader = LoadShaderProgram(Game.assets, "breakoutTrailShader");
}

void BreakoutDeinit(EtcGameMemory *mem, SceneData *scene) {
    // Since Breakout is the only scene in this game, it isn't
    // strictly necessary to cleanup all allocated memory. This is a
    // much bigger deal in a game with a lot of scenes.
}

void BreakoutUpdate(EtcGameMemory *mem, SceneData *scene) {
    Breakout *breakout = (Breakout *)scene->data;
    EntityDB *entityDB = &scene->entityDB;


    // The framerate of the engine's main update loop is variable (so
    // Game.deltaTime is not fixed), which can lead to bizarre and
    // unreliable behavior when resolving the game physics.

    // To combat that, a single physics tick is performed each time
    // `BreakoutMSPerTick` seconds pass. This makes the physics much
    // more stable and leads to a smoother game feel.
    
    breakout->timeAccumulator += Game.deltaTime;

    while (breakout->timeAccumulator >= BreakoutMSPerTick) {
        UpdatePaddle(breakout, &breakout->paddle, BreakoutMSPerTick);
        UpdateBricks(breakout, BreakoutMSPerTick);

        switch (breakout->gameState) {
            case BreakoutGameState_Serve: {
            } break;

            case BreakoutGameState_Play: {
                UpdateBall(breakout, &breakout->ball, BreakoutMSPerTick);
                breakout->playTime += BreakoutMSPerTick;
            } break;

            case BreakoutGameState_GameOver: {
            } break;
        }

        UpdateAnimations(&breakout->animationManager, entityDB);
        breakout->gameTime += BreakoutMSPerTick;
        breakout->tick++;

        breakout->timeAccumulator -= BreakoutMSPerTick;
    }


    switch (breakout->gameState) {
        case BreakoutGameState_Serve: {
            real32 timeInState = breakout->gameTime - breakout->stateStartTime;
            if (((int32)(timeInState / 0.45f)) % 4 != 0) {
                DebugDrawText(V3(12.5f, 8.0f, 0.0f), 0.5, HSVToRGB(breakout->palette.textColor),
                              "Press SPACE to get served");
            }

            if (InputPressed(Game.keyboard, Etc_Space) ||
                InputPressed(Game.gamepad0, Etc_FaceA) ||
                InputPressed(Game.gamepad0, Etc_FaceB) ||
                InputPressed(Game.gamepad0, Etc_FaceX) ||
                InputPressed(Game.gamepad0, Etc_FaceY)) {
                vec3 ballPosition = BreakoutBallServePosition;

                vec3 paddlePosition = breakout->paddle.position;
                real32 paddleHalfWidth = breakout->paddle.halfExtents.x;

                vec3 targetPosition;
                do {
                    targetPosition = paddlePosition + V3(1, 0, 0) * Randf(-paddleHalfWidth, paddleHalfWidth);
                } while (Absf(targetPosition.x - ballPosition.x) < paddleHalfWidth * 0.25f);

                vec3 direction = targetPosition - ballPosition;

                breakout->ball.velocity = Normalize(direction) * BreakoutServeSpeed;
                breakout->ball.spin = Randf(0.15f);
                breakout->servesLeft--;
                breakout->lastServeTime = breakout->gameTime;

                StartBallHitAnimation(breakout, &breakout->ball);

                CachedPosition p;
                p.tick = breakout->tick;
                p.time = breakout->gameTime;
                p.position = ballPosition;

                DynamicArrayClear(&breakout->ball.positions);
                PushBack(&breakout->ball.positions, p);

                SetGameState(breakout, BreakoutGameState_Play);
            }
        } break;

        case BreakoutGameState_Play: {
            if (breakout->ball.positions[0].position.y < breakout->ball.radius) {
                if (breakout->servesLeft > 0) {
                    StopAnimation(&breakout->animationManager, breakout->ball.sizeAnimID);
                    StopAnimation(&breakout->animationManager, breakout->ball.colorAnimID);

                    DestroyBall(entityDB, breakout->ball);

                    breakout->ball = SpawnBall(breakout, entityDB);

                    breakout->ball.color = breakout->palette.textColor;
                    breakout->ball.prevTickColor = breakout->palette.textColor;
                    breakout->ball.material->diffuseColor = HSVToRGB(breakout->ball.color);

                    SetGameState(breakout, BreakoutGameState_Serve);
                }
                else {
                    DestroyLevel(entityDB, &breakout->level);
                    DestroyBall(entityDB, breakout->ball);
                    breakout->ball = {};

                    SetGameState(breakout, BreakoutGameState_GameOver);
                }
            }
            else if (breakout->level.bricksCleared == breakout->level.bricks.count) {
                DestroyLevel(entityDB, &breakout->level);

                breakout->palette = GenerateNextPalette(breakout->prevPalette);

                breakout->levelIndex++;
                breakout->level = SpawnLevel(breakout, entityDB, breakout->levelIndex);

                StartSpawnLevelAnimations(breakout, &breakout->level);
                StartPaddleSizeAnimation(breakout, &breakout->paddle, breakout->levelIndex);
                ApplyPalette(breakout, breakout->palette);
            }
        } break;

        case BreakoutGameState_GameOver: {
            DebugDrawText(V3(6.7f, 10.0f, 0.0f), 4, HSVToRGB(breakout->palette.textColor),
                          "YOU LOSE");
            DebugDrawText(V3(12.7f, 8.0f, 0.0f), 0.5, HSVToRGB(breakout->palette.textColor),
                          "Press SPACE to try again");
            
            if (InputPressed(Game.keyboard, Etc_Space) ||
                InputPressed(Game.gamepad0, Etc_FaceA) ||
                InputPressed(Game.gamepad0, Etc_FaceB) ||
                InputPressed(Game.gamepad0, Etc_FaceX) ||
                InputPressed(Game.gamepad0, Etc_FaceY)) {
                BreakoutRestartGame(breakout, entityDB);
            }
        } break;
    }


    uint32 renderKey = GenerateGBufferRenderKey(Game.sceneManager->activeSceneIndex, GBufferStage_Render);
    ClearColor(&Game.graphics->gbufferPass, renderKey, HSVToRGB(breakout->palette.backgroundColor));


    // The physics is run at a fixed rate, which means it's possible
    // that a physics tick doesn't happen for several frames. In this
    // case the rendering will appear choppy.

    // To prevent this choppiness, the state of various objects at the
    // previous physics tick can be stored and blended with the most
    // recent tick. This results in very smooth rendering at the cost
    // of delaying the rendering by a single physics tick.
    real32 alpha = breakout->timeAccumulator / BreakoutMSPerTick;
    SampleEntityPositions(breakout, entityDB, alpha);

    if (breakout->gameState != BreakoutGameState_GameOver) {
        RenderBallTrail(breakout, breakout->ball, alpha);
    }

    RenderGameInfo(breakout);
}

