
// Input Event Queue
void InputDeviceSetDiscreteState(EtcInputDevice *device, int index, b32 value) {
    if (index >= 0 && index < device->discreteCount) {
        if (!value) {
            if (device->hold[index]) {
                device->released[index] = true;

                device->framesSinceChange[index] = 0;
                device->secondsSinceChange[index] = 0.0f;
            }

            device->pressed[index] = false;
            device->hold[index] = false;
        }
        else {
            if (!device->hold[index]) {
                device->pressed[index] = true;

                device->framesSinceChange[index] = 0;
                device->secondsSinceChange[index] = 0.0f;
            }

            device->hold[index] = true;
            device->released[index] = false;
        }
    }
}

void InputDeviceSetAnalogueState(EtcInputDevice *device, int index, real32 value) {
    if (index >= 0 && index < device->analogueCount) {
        device->analogue[index] = value;
    }
}

void InputDeviceFlush(EtcInputDevice *device, real32 deltaTime) {
    device->lastEventIndex = -1;

    Plat.CopyMem(device->prevAnalogue, device->analogue, sizeof(real32) * device->analogueCount);

    for (int i = 0; i < device->discreteCount; i++) {
        device->pressed[i] = false;
        device->released[i] = false;

        device->framesSinceChange[i]++;
        device->secondsSinceChange[i] += deltaTime;
    }
}

void ProcessInputEventQueue(EtcInputEventQueue *queue) {
    for (int i = 0; i < queue->count; i++) {
        EtcInputEvent event = queue->queue[i];

        switch (event.type) {
            case InputEvent_Discrete: {
                InputDeviceSetDiscreteState(event.device, event.index, event.discreteValue);
                event.device->lastEventIndex = i;
            } break;
            case InputEvent_Analogue: {
                real32 prevValue = event.device->analogue[event.index];
                    
                InputDeviceSetAnalogueState(event.device, event.index, event.analogueValue);

                // @TODO: probably need to deadzone this
                if (event.analogueValue != prevValue) {
                    event.device->lastEventIndex = i;
                }
            } break;
        }
    }

    queue->count = 0;
}

b32 InputPressed(EtcInputDevice *device, int index, b32 force) {
    b32 result = false;

    b32 accessAllowed = force || !device->blocked;

    if (device->connected && accessAllowed) {
        if (index >= 0 && index < device->discreteCount) {
            result = device->pressed[index];
        }
    }

    return result;
}

b32 InputHold(EtcInputDevice *device, int index, b32 force) {
    b32 result = false;

    b32 accessAllowed = force || !device->blocked;

    if (device->connected && accessAllowed) {
        if (index >= 0 && index < device->discreteCount) {
            result = device->hold[index];
        }
    }

    return result;
}

b32 InputReleased(EtcInputDevice *device, int index, b32 force) {
    b32 result = false;

    b32 accessAllowed = force || !device->blocked;

    if (device->connected && accessAllowed) {
        if (index >= 0 && index < device->discreteCount) {
            result = device->released[index];
        }
    }

    return result;
}

real32 InputSecondsSinceChanged(EtcInputDevice *device, int index, b32 force) {
    real32 result = 0.0f;

    b32 accessAllowed = force || !device->blocked;

    if (device->connected && accessAllowed) {
        if (index >= 0 && index < device->discreteCount) {
            result = device->secondsSinceChange[index];
        }
    }

    return result;
}

real32 InputPrevAnalogue(EtcInputDevice *device, int index, b32 force) {
    real32 result = 0;

    b32 accessAllowed = force || !device->blocked;

    if (index >= 0 && index < device->analogueCount) {
        if (device->connected && accessAllowed) {
            result = device->prevAnalogue[index];
        }
    }

    return result;
}

vec2 InputPrevAnalogue2D(EtcInputDevice *device, int index, b32 force) {
    vec2 result = V2(0, 0);

    b32 accessAllowed = force || !device->blocked;

    if (index >= 0 && index < device->analogueCount - 1) {
        if (device->connected && accessAllowed) {
            result = V2(device->prevAnalogue[index],
                        device->prevAnalogue[index + 1]);
        }
    }

    return result;
}

real32 InputAnalogue(EtcInputDevice *device, int index, b32 force) {
    real32 result = 0;

    b32 accessAllowed = force || !device->blocked;

    if (index >= 0 && index < device->analogueCount) {
        if (device->connected && accessAllowed) {
            result = device->analogue[index];
        }
    }

    return result;
}

vec2 InputAnalogue2D(EtcInputDevice *device, int index, b32 force) {
    vec2 result = V2(0, 0);

    b32 accessAllowed = force || !device->blocked;

    if (index >= 0 && index < device->analogueCount - 1) {
        if (device->connected && accessAllowed) {
            result = V2(device->analogue[index],
                        device->analogue[index + 1]);
        }
    }

    return result;
}

void InputSetOutput(EtcInputDevice *device, int index, real32 value) {
    if (device->connected) {
        if (index >= 0 && index < device->outputCount) {
            device->output[index] = value;
        }
    }
}


// Mouse Look Management
b32 RequestMouseLook() {
    if (Game.mouse->blocked) {
        return false;
    }
    b32 grantedMouseLook = true;

    // if (Game.shouldRecenterMousePrevFrame) {
    //     grantedMouseLook = false;
    // }
    Game.shouldRecenterMouse = true;

    return grantedMouseLook;
}

void ProcessMouseLookRequests() {

    if (Game.shouldRecenterMouse) {
        vec2i screenCenter = V2i(Game.graphics->width / 2, Game.graphics->height / 2);
        Plat.MoveMouse(screenCenter.x, screenCenter.y);

        if (!Game.shouldRecenterMousePrevFrame) {
            Plat.HideMouse();
        }
    }
    else if (Game.shouldRecenterMousePrevFrame) {
        Plat.ShowMouse();
    }

    Game.shouldRecenterMousePrevFrame = Game.shouldRecenterMouse;
    Game.shouldRecenterMouse = false;
}


// Input Actions
void SerializeInputActions(const char *path, InputActions *actions) {
    Schema *inputActionBindingSchema = Game.GetSchema("InputActionBinding");

    EtcTxtWriteState writer;
    if (OpenFileForWrite(&writer, path)) {

        for (int type = 0; type < ENUM_VALUECOUNT(InputActionType); type++) {
            DynamicArray<InputActionBinding> bindings = actions->bindings[type];

            if (bindings.count == 0) {
                continue;
            }

            WritePushArrayValue(&writer, ENUM_STRINGVALUE(InputActionType, type));

            for (int i = 0; i < bindings.count; i++) {
                SerializeValue(&writer, inputActionBindingSchema, &bindings[i]);
            }

            WritePopArrayValue(&writer);
        }

        CloseFile(&writer);
    }
}

b32 DeserializeInputActions(const char *path, InputActions *actions) {
    b32 success = true;

    Schema *inputActionBindingSchema = Game.GetSchema("InputActionBinding");

    EtcTxtParseState parser;
    if (OpenFileForRead(&parser, path)) {

        const char *identifier;
        while (ParseIdentifier(&parser, &identifier)) {

            InputActionType actionType;
            if (ENUM_FINDVALUE(InputActionType, identifier, &actionType)) {

                if (ParsePushArrayValue(&parser)) {
                    DynamicArray<InputActionBinding> bindings = {};

                    while (!ParseHasError(&parser)) {
                        if (ParsePopArrayValue(&parser)) {
                            break;
                        }

                        InputActionBinding binding = {};
                        if (DeserializeValue(&parser, inputActionBindingSchema, &binding)) {
                            PushBack(&bindings, binding);
                        }
                        else {
                            continue;
                        }
                    }

                    actions->bindings[actionType] = bindings;
                }
                else if (!ParseSkipValue(&parser)) {
                    success = false;
                    break;
                }

            }
        }

        // @BUG @TODO: memory corruption when deallocating the strings in CloseFile
        CloseFile(&parser);
    }

    return success;
}

b32 BindingSatisfiedByDevice(InputActions *actions, InputActionType actionType, InputActionBinding binding, EtcInputDevice *device, EtcInputDevice *modifierDevice) {
    b32 result = false;

    b32 andModifiers = true;
    b32 orModifiers = false;
    b32 nandModifiers = true;

    uint32 andCount = 0;
    uint32 orCount = 0;
    uint32 nandCount = 0;

    for (int i = 0; i < binding.modifierCount; i++) {
        if (binding.modifierOps[i] == InputModifierOp_And) {
            andModifiers &= InputHold(modifierDevice, binding.modifiers[i], true);
            andCount++;
        }
        else if (binding.modifierOps[i] == InputModifierOp_Or) {
            orModifiers |= InputHold(modifierDevice, binding.modifiers[i], true);
            orCount++;
        }
        else if (binding.modifierOps[i] == InputModifierOp_Nand) {
            nandModifiers &= !InputHold(modifierDevice, binding.modifiers[i], true);
            nandCount++;
        }
    }


    b32 allModifiersSatisfied = true;

    if (andCount > 0) {
        allModifiersSatisfied &= andModifiers;
    }
    
    if (orCount > 0) {
        allModifiersSatisfied &= orModifiers;
    }

    if (nandCount > 0) {
        allModifiersSatisfied &= nandModifiers;
    }


    if (!allModifiersSatisfied) {
        return false;
    }

    InputActionBindingType bindingType = actions->bindingTypes[actionType];

    switch (bindingType) {
        case InputActionBindingType_Discrete: {
            b32 pressed = false;
            b32 hold = false;
            b32 released = false;

            if (binding.inputType == InputActionBindingType_Discrete) {
                pressed = InputPressed(device, binding.discrete0, true);
                hold = InputHold(device, binding.discrete0, true);
                released = InputReleased(device, binding.discrete0, true);
            }
            else if (binding.inputType == InputActionBindingType_Analogue) {

                real32 prev = InputPrevAnalogue(device, binding.analogue, true);
                real32 current = InputAnalogue(device, binding.analogue, true);

                const real32 threshold = DEADZONE_TRIGGER_MIN;
                real32 scale = binding.analogueScale;

                pressed = prev * scale < threshold && current * scale >= threshold;
                hold = current * scale >= threshold;
                released = prev * scale > threshold && current * scale >= threshold;
            }

            result = pressed || hold || released;
        } break;

        case InputActionBindingType_Analogue: {
            if (binding.inputType == InputActionBindingType_Discrete) {
                result = (InputHold(device, binding.discrete0, true) ||
                          InputHold(device, binding.discrete1, true));
            }
            else if (binding.inputType == InputActionBindingType_Analogue) {
                // @MAYBE @BUG: is there a reason we dont want to satisfy binding if the value is 0.0f?
                result = true;
                // if (device->type == Etc_Gamepad) {
                //     result = InputAnalogue(device, binding.analogue, true) != 0.0f; 
                // }
                // else {
                //     result = true;
                // }
                // else if (device->type == Etc_Mouse) {
                //     result = InputAnalogue(device, binding.analogue, true) != 0.5f;    
                // }
                
            }
        } break;
    }

    return result;
}

void WriteDeviceStateForBinding(InputActions *actions, InputActionType actionType, EtcInputDevice *device, InputActionBinding binding) {
        
    actions->device[actionType] = device;
    actions->deviceBlocked[actionType] = device->blocked;

    InputActionBindingType bindingType = actions->bindingTypes[actionType];
    switch (bindingType) {
        case InputActionBindingType_Discrete: {
            if (binding.inputType == InputActionBindingType_Discrete) {
                actions->pressed[actionType] = InputPressed(device, binding.discrete0, true);
                actions->hold[actionType] = InputHold(device, binding.discrete0, true);
                actions->released[actionType] = InputReleased(device, binding.discrete0, true);
            }
            else if (binding.inputType == InputActionBindingType_Analogue) {
                const real32 threshold = DEADZONE_TRIGGER_MIN;

                real32 prev = InputPrevAnalogue(device, binding.analogue, true);
                real32 current = InputAnalogue(device, binding.analogue, true);

                actions->prevAnalogue[actionType] = prev;
                actions->analogue[actionType] = current;

                real32 scale = binding.analogueScale;

                actions->pressed[actionType] = prev * scale < threshold && current * scale >= threshold;
                actions->hold[actionType] = current * scale >= threshold;
                actions->released[actionType] = prev * scale > threshold && current * scale >= threshold;
            }
        } break;

        case InputActionBindingType_Analogue: {
            if (binding.inputType == InputActionBindingType_Discrete) {
                if (binding.discrete0 != binding.discrete1) {
                    if (InputHold(device, binding.discrete0, true)) {
                        actions->analogue[actionType] -= 1.0f;
                    }
                    if (InputHold(device, binding.discrete1, true)) {
                        actions->analogue[actionType] += 1.0f;
                    }
                }
                else {
                    if (InputHold(device, binding.discrete1, true)) {
                        actions->analogue[actionType] += 1.0f;
                    }
                }
            }
            else if (binding.inputType == InputActionBindingType_Analogue) {
                actions->analogue[actionType] = InputAnalogue(device, binding.analogue, true);
            }
        } break;
    }
}

void ProcessInputActions(InputActions *actions, EtcInputDevice *keyboard, EtcInputDevice *mouse, EtcInputDevice *gamepads, uint32 gamepadCount) {
    Plat.ClearMem(actions->pressed, sizeof(actions->pressed));
    Plat.ClearMem(actions->released, sizeof(actions->released));

    Plat.CopyMem(actions->prevAnalogue, actions->analogue, sizeof(actions->analogue));
    Plat.ClearMem(actions->analogue, sizeof(actions->analogue));

    Plat.ClearMem(actions->device, sizeof(actions->device));
    Plat.ClearMem(actions->deviceBlocked, sizeof(actions->deviceBlocked));

    for (int type = 0; type < ENUM_VALUECOUNT(InputActionType); type++) {
        DynamicArray<InputActionBinding> bindings = actions->bindings[type];

        int32 satisfyingBindingIndex = -1;
        EtcInputDevice *satisfyingDevice = NULL;
        EtcInputDevice *satisfyingModifierDevice = NULL;

        for (int i = 0; i < bindings.count; i++) {
            InputActionBinding binding = bindings[i];

            uint32 deviceCount;
            EtcInputDevice *devices;
            EtcInputDevice *modifierDevices;

            switch (binding.deviceType) {
                case Etc_Keyboard: {
                    deviceCount = 1;
                    devices = keyboard;
                    modifierDevices = keyboard;
                } break;
                    
                case Etc_Mouse: {
                    deviceCount = 1;
                    devices = mouse;
                    modifierDevices = keyboard;
                } break;
                    
                case Etc_Gamepad: {
                    deviceCount = gamepadCount;
                    devices = gamepads;
                    modifierDevices = gamepads;
                } break;
            }

            for (int k = 0; k < deviceCount; k++) {
                if (BindingSatisfiedByDevice(actions, (InputActionType)type, binding, &devices[k], &modifierDevices[k])) {
                    if (devices[k].lastEventIndex >= 0 &&
                        (satisfyingDevice == NULL || devices[k].lastEventIndex > satisfyingDevice->lastEventIndex)) {
                        satisfyingDevice = &devices[k];
                        satisfyingBindingIndex = i;
                    }

                    if (devices[k].lastEventIndex >= 0 &&
                        (satisfyingModifierDevice == NULL || devices[k].lastEventIndex > satisfyingModifierDevice->lastEventIndex)) {
                        satisfyingModifierDevice = &devices[k];
                    }
                }
            }
        }

        InputActionPrevDeviceInfo prevInfo = actions->prevDeviceInfo[type];

        if (satisfyingDevice != NULL) {
            WriteDeviceStateForBinding(actions, (InputActionType)type, satisfyingDevice, bindings[satisfyingBindingIndex]);

            InputActionPrevDeviceInfo info;
            info.device = satisfyingDevice;
            info.modifierDevice = satisfyingModifierDevice;
            info.bindingIndex = satisfyingBindingIndex;

            actions->prevDeviceInfo[type] = info;
        }
        else if (prevInfo.device != NULL && prevInfo.bindingIndex < bindings.count &&
                 BindingSatisfiedByDevice(actions, (InputActionType)type, bindings[prevInfo.bindingIndex],
                                          prevInfo.device, prevInfo.modifierDevice)) {
            WriteDeviceStateForBinding(actions, (InputActionType)type, prevInfo.device, bindings[prevInfo.bindingIndex]);
        }
    }
}

EtcInputDevice *GetMostRecentDevice(InputActions *actions, InputActionType type) {
    if (type == InputAction_FreeLookXY) {
        EtcInputDevice *xDevice = actions->device[InputAction_FreeLookX];
        EtcInputDevice *yDevice = actions->device[InputAction_FreeLookY];

        if (xDevice) {
            // @MAYBE @TODO: because we always write in a value for the mouse, and check x first,
            // this will always return the mouse, even if its at rest;
            // so we see if there has been y movement on the gamepad and prefer that
            // if (yDevice && yDevice->type == Etc_Gamepad) {
            //     return yDevice;
            // }
            return xDevice;
        }
        else if (yDevice) {
            return yDevice;
        }
    }

    return actions->device[type];
}

void BlockAllInputActionsExcept(InputActions *actions, InputActionType *types, uint32 typeCount) {
    memset(actions->actionBlocked, 1, sizeof(InputActionType) * ENUM_VALUECOUNT(InputActionType));
    for (int i = 0; i < typeCount; i++) {
        actions->actionBlocked[types[i]] = false;
    }
}

void BlockTheseInputActions(InputActions *actions, InputActionType *types, uint32 typeCount) {
    for (int i = 0; i < typeCount; i++) {
        actions->actionBlocked[types[i]] = true;
    }
}

void UnblockAllInputActions(InputActions *actions) {
    memset(actions->actionBlocked, 0, sizeof(InputActionType) * ENUM_VALUECOUNT(InputActionType));
}

b32 InputActionPressed(InputActions *actions, InputActionType type, b32 force) {
    b32 result = false;

    b32 allowed = (force || !actions->deviceBlocked[type]) && !actions->actionBlocked[type];

    if (allowed) {
        result = actions->pressed[type];
    }

    return result;
}

b32 InputActionHold(InputActions *actions, InputActionType type, b32 force) {
    b32 result = false;

    b32 allowed = (force || !actions->deviceBlocked[type]) && !actions->actionBlocked[type];

    if (allowed) {
        result = actions->hold[type];
    }

    return result;
}

b32 InputActionReleased(InputActions *actions, InputActionType type, b32 force) {
    b32 result = false;

    b32 allowed = (force || !actions->deviceBlocked[type]) && !actions->actionBlocked[type];

    if (allowed) {
        result = actions->released[type];
    }

    return result;
}

real32 InputActionAnalogue(InputActions *actions, InputActionType type, b32 applyDeadzone, b32 force) {
    real32 result = 0.0f;

    b32 allowed = (force || !actions->deviceBlocked[type]) && !actions->actionBlocked[type];

    if (allowed) {
        result = actions->analogue[type];

        // @NOTE: we dont want to apply a trigger deadzone to keyboard or mouse input because
        // one key/button could be -1, which would get clamped to 0 since it's below DEADZONE_TRIGGER_MIN
        if (applyDeadzone && actions->device[type] != NULL && actions->device[type]->type == Etc_Gamepad) {
            result = InputDeadzone(result, DEADZONE_TRIGGER_MIN, DEADZONE_TRIGGER_MAX);
        }
    }
    
    return result;
}

real32 InputActionPrevAnalogue(InputActions *actions, InputActionType type, b32 applyDeadzone, b32 force) {
    real32 result = 0.0f;

    b32 allowed = (force || !actions->deviceBlocked[type]) && !actions->actionBlocked[type];

    if (allowed) {
        result = actions->prevAnalogue[type];

        if (applyDeadzone && actions->device[type] != NULL && actions->device[type]->type == Etc_Gamepad) {
            result = InputDeadzone(result, DEADZONE_TRIGGER_MIN, DEADZONE_TRIGGER_MAX);
        }
    }
    
    return result;
}

b32 InputActionAnaloguePressed(InputActions *actions, InputActionType type, real32 value, b32 applyDeadzone, b32 force) {
    return (InputActionPrevAnalogue(actions, type, applyDeadzone, force) < value &&
            InputActionAnalogue(actions, type, applyDeadzone, force) >= value);
}

b32 InputActionAnalogueHold(InputActions *actions, InputActionType type, real32 value, b32 applyDeadzone, b32 force) {
    return (InputActionPrevAnalogue(actions, type, applyDeadzone, force) >= value &&
            InputActionAnalogue(actions, type, applyDeadzone, force) >= value);
}

b32 InputActionAnalogueReleased(InputActions *actions, InputActionType type, real32 value, b32 applyDeadzone, b32 force) {
    return (InputActionPrevAnalogue(actions, type, applyDeadzone, force) >= value &&
            InputActionAnalogue(actions, type, applyDeadzone, force) < value);
}

// @NOTE @BUG: ifthe 2 analogue values came from different devices then we cant really trust the values
// for instance the stickX could be at 0.0f, but the mouseY 0.5f which would trick us into thinking there's a lot
// of movement
vec2 InputActionAnalogue2D(InputActions *actions, InputActionType type, b32 applyDeadzone, b32 force) {
    vec2 result = {};

    b32 allowed = (force || (!actions->deviceBlocked[type] && !actions->deviceBlocked[type + 1])) &&
        (!actions->actionBlocked[type] && !actions->actionBlocked[type + 1]);
    
    if (allowed) {
        result = V2(actions->analogue[type], actions->analogue[type + 1]);

        // @TODO: only do this for the stick and not the mouse, which means we need to know a) the active
        // device, or we need to know which device satisfied the binding
        if (applyDeadzone) {
            result = InputDeadzone(result, DEADZONE_STICK_MIN, DEADZONE_STICK_MAX);
        }
    }
    
    return result;
}

vec2 InputActionPrevAnalogue2D(InputActions *actions, InputActionType type, b32 applyDeadzone, b32 force) {
    vec2 result = {};

    b32 allowed = (force || (!actions->deviceBlocked[type] && !actions->deviceBlocked[type + 1])) &&
        (!actions->actionBlocked[type] && !actions->actionBlocked[type + 1]);

    if (allowed) {
        result = V2(actions->prevAnalogue[type], actions->prevAnalogue[type + 1]);

        if (applyDeadzone) {
            result = InputDeadzone(result, DEADZONE_STICK_MIN, DEADZONE_STICK_MAX);
        }
    }
    
    return result;
}

b32 AnyInputAction(InputActions *inputActions, InputActionType *type) {
    b32 inputReceived = false;
    for (int i = 0; i < ENUM_VALUECOUNT(InputActionType); i++) {
        if (inputActions->device[i] == NULL) {
            continue;
        }
        
        EtcInputDeviceType deviceType = inputActions->device[i]->type;
        InputActionBindingType bindingType = inputActions->bindingTypes[i];
        switch (deviceType) {
            
            case Etc_Mouse: {
                switch (bindingType) {
                    case InputActionBindingType_Discrete: {
                        inputReceived = inputActions->pressed[i];        
                    } break;
                        
                    case InputActionBindingType_Analogue: {
                        inputReceived = inputActions->analogue[i] != 0.5f; 
                    } break;
                }
            } break;
                
            default: {
                switch (bindingType) {
                    case InputActionBindingType_Discrete: {
                        inputReceived = inputActions->pressed[i];        
                    } break;
                        
                    case InputActionBindingType_Analogue: {
                        inputReceived = inputActions->analogue[i] != 0.0f; 
                    } break;
                }
            } break;
        }

        if (inputReceived) {
            if (type != NULL) {
                *type = (InputActionType)i;    
            }
            
            break;
        }
    }

    return inputReceived;
}

b32 AnyInputActionPressed(InputActions *inputActions, InputActionType *type) {
    b32 inputReceived = false;
    for (int i = 0; i < ENUM_VALUECOUNT(InputActionType); i++) {
        if (inputActions->pressed[i]) {
            if (type != NULL) {
                *type = (InputActionType)i;
            }
            
            inputReceived = true;
            break;
        }
    }

    return inputReceived;
}

b32 InputActionBound(InputActions *actions, InputActionType type) {
    return actions->bindings[type].count > 0;
}

b32 BindInputAction(InputActions *actions, InputActionType type, b32 writeStateWhenBound, InputActionType *exclusives, uint32 exclusiveCount) {
    bool foundBinding = false;
    
    InputActionBinding binding = {};
    binding.inputType = actions->bindingTypes[type];

    EtcInputDevice *keyboard = Game.keyboard;
    EtcInputDevice *mouse = Game.mouse;
    EtcInputDevice *pads = Game.mem->gamepads;

    EtcInputDevice *boundDevice = NULL;

    // @GACK: this code is duplicated for each input device
    switch (binding.inputType) {
        case InputActionBindingType_Discrete : {
            // @NOTE: can only bind a discrete input to a discrete action
            for (int i = 0; i < keyboard->discreteCount; i++) {
                if (InputPressed(keyboard, i)) {

                    if (exclusives != NULL) {
                        bool bindingTaken = false;
                        for (int j = 0; j < exclusiveCount; j++) {
                            InputActionType exclusive = exclusives[j];
                            for (int k = 0; k < actions->bindings[exclusive].count; k++) {
                                if (actions->bindings[exclusive][k].discrete0 == i||
                                    actions->bindings[exclusive][k].discrete1 == i) {
                                    bindingTaken = true;
                                    break;
                                }
                            }

                            if (bindingTaken) { break; }
                        }
                        if (bindingTaken) { continue; }
                    }
                    
                    binding.deviceType = keyboard->type;
                    binding.discrete0 = i;

                    foundBinding = true;
                    boundDevice = keyboard;
                    goto pushBinding;
                }
            }

            for (int i = 0; i < mouse->discreteCount; i++) {
                if (InputPressed(mouse, i)) {

                    if (exclusives != NULL) {
                        bool bindingTaken = false;
                        for (int j = 0; j < exclusiveCount; j++) {
                            InputActionType exclusive = exclusives[j];
                            for (int k = 0; k < actions->bindings[exclusive].count; k++) {
                                if (actions->bindings[exclusive][k].discrete0 == i||
                                    actions->bindings[exclusive][k].discrete1 == i) {
                                    bindingTaken = true;
                                    break;
                                }
                            }

                            if (bindingTaken) { break; }
                        }
                        if (bindingTaken) { continue; }
                    }

                    binding.deviceType = mouse->type;
                    binding.discrete0 = i;

                    foundBinding = true;
                    boundDevice = mouse;
                    goto pushBinding;
                }
            }

            for (int i = 0; i < MAX_NUM_GAMEPADS; i++) {
                for (int j = 0; j < pads[i].discreteCount; j++) {
                    if (InputPressed(&pads[i], j)) {
                        
                        if (exclusives != NULL) {
                            bool bindingTaken = false;
                            for (int k = 0; k < exclusiveCount; k++) {
                                InputActionType exclusive = exclusives[k];
                                for (int l = 0; l < actions->bindings[exclusive].count; l++) {
                                    if (actions->bindings[exclusive][l].discrete0 == j ||
                                        actions->bindings[exclusive][l].discrete1 == j) {
                                        bindingTaken = true;
                                        break;
                                    }
                                }

                                if (bindingTaken) { break; }
                            }
                            if (bindingTaken) { continue; }
                        }
                        
                        binding.deviceType = pads[i].type;
                        binding.discrete0 = j;

                        foundBinding = true;
                        boundDevice = &pads[i];
                        goto pushBinding;
                    }
                }    
            }
            
        } break;

        case InputActionBindingType_Analogue : {
            ASSERT(false);
        } break;
    }

 pushBinding:
    if (foundBinding) {
        PushBack(&actions->bindings[type], binding);
        if (writeStateWhenBound) { WriteDeviceStateForBinding(actions, type, boundDevice, binding); }
    }

    return foundBinding;
}

void ClearBindings(InputActions *actions, InputActionType type) {
    DynamicArrayClear(&actions->bindings[type]);
}
