
EtcTxtParseCursor GetCursor(EtcTxtParseState *parser) {
    EtcTxtParseCursor cursor;

    cursor.byteOffset = parser->byteOffset;
    cursor.column = parser->column;
    cursor.line = parser->line;

    cursor.stackDepth = parser->stackDepth;

    Plat.CopyMem(cursor.hasIdentifier, parser->hasIdentifier, sizeof(cursor.hasIdentifier));
    Plat.CopyMem(cursor.isArray, parser->isArray, sizeof(cursor.isArray));

    return cursor;
}

void SetCursor(EtcTxtParseState *parser, EtcTxtParseCursor cursor) {
    parser->byteOffset = cursor.byteOffset;
    parser->column = cursor.column;
    parser->line = cursor.line;

    parser->stackDepth = cursor.stackDepth;

    Plat.CopyMem(parser->hasIdentifier, cursor.hasIdentifier, sizeof(cursor.hasIdentifier));
    Plat.CopyMem(parser->isArray, cursor.isArray, sizeof(cursor.isArray));
}

b32 OpenFileForRead(EtcTxtParseState *parser, const char *path) {
    Plat.ClearMem(parser, sizeof(EtcTxtParseState));

    b32 result = false;

    if (Plat.OpenFileForRead(path, &parser->file)) {
        result = true;

        parser->buffer = (const char *)parser->file.data;
        parser->bufferSize = parser->file.fileSize;
    }

    return result;
}

void CloseFile(EtcTxtParseState *parser) {
    Plat.CloseFile(&parser->file);

    for (int i = 0; i < parser->allocatedStrings.count; i++) {
        Plat.DeallocateMem((char *)parser->allocatedStrings[i]);
    }

    DeallocateDynamicArray(&parser->allocatedStrings);
}

b32 OpenBufferForRead(EtcTxtParseState *parser, const char *buffer, uint32 bufferSize) {
    Plat.ClearMem(parser, sizeof(EtcTxtParseState));

    if (buffer != NULL && bufferSize > 0) {
        parser->buffer = buffer;
        parser->bufferSize = bufferSize;

        return true;
    }

    return false;
}


b32 OpenFileForWrite(EtcTxtWriteState *writer, const char *path, uint32 initialBytes, uint32 flags) {
    Plat.ClearMem(writer, sizeof(EtcTxtWriteState));

    b32 result = Plat.OpenFileForWrite(path, &writer->file, initialBytes);

    if (result) {
        FileClear(&writer->file);
    }
    
    writer->flags = flags;
    writer->stackDepth = 0;

    return result;
}

void CloseFile(EtcTxtWriteState *writer) {
    Plat.CloseFile(&writer->file);
}

b32 OpenBufferForWrite(EtcTxtWriteState *writer, char *buffer, uint32 bufferSize, EtcTxtWriteCallback *callback, void *callbackUserData, uint32 flags) {
    Plat.ClearMem(writer, sizeof(EtcTxtWriteState));

    if (buffer != NULL && bufferSize > 0) {
        writer->flags = flags;

        writer->currentBuffer = buffer;
        writer->currentBufferSize = bufferSize;
        writer->writeCallback = callback;
        writer->callbackUserData = callbackUserData;

        return true;
    }

    return false;
}

// Parsing
void ParserSetError_(EtcTxtParseState *parser, EtcTxtParseErrorType errorType, const char *fmt = NULL, ...) {

    parser->error.type = errorType;
    parser->error.cursor = GetCursor(parser);

    if (fmt != NULL) {
        va_list args;
        va_start(args, fmt);
        vsnprintf(parser->error.message, ARRAY_LENGTH(char, parser->error.message), fmt, args);
        va_end(args);
    }
    else {
        parser->error.message[0] = '\0';
    }
}

b32 AtEndOfFile_(EtcTxtParseState *parser) {
    b32 result = parser->byteOffset >= parser->bufferSize;
    return result;
}

b32 ConsumeAnyCharacter_(EtcTxtParseState *parser) {
    b32 result = false;
    if (!AtEndOfFile_(parser)) {
        result = true;

        parser->byteOffset++;
        parser->column++;
    }

    return result;
}

b32 ConsumeCharacter_(EtcTxtParseState *parser, char c) {
    b32 result = false;
    if (!AtEndOfFile_(parser) &&
        parser->buffer[parser->byteOffset] == c) {
        result = true;

        parser->byteOffset++;
        parser->column++;
    }

    return result;
}

b32 ConsumeCharacterSequence_(EtcTxtParseState *parser, const char *str) {
    b32 result = true;

    EtcTxtParseCursor startCursor = GetCursor(parser);

    char *at = (char *)str;
    while (*at != '\0') {
        if (ConsumeCharacter_(parser, *at)) {
            if (*at == '\n') {
                parser->column = 1;
                parser->line++;
            }
            at++;
        }
        else {
            result = false;
            break;
        }
    }

    if (!result) {
        SetCursor(parser, startCursor);
    }

    return result;
}

b32 ConsumeAlpha_(EtcTxtParseState *parser) {
    b32 result = false;
    if (!AtEndOfFile_(parser) &&
        isalpha(parser->buffer[parser->byteOffset])) {
        result = true;

        parser->byteOffset++;
        parser->column++;
    }

    return result;
}

b32 ConsumeDigit_(EtcTxtParseState *parser) {
    b32 result = false;
    if (!AtEndOfFile_(parser) &&
        isdigit(parser->buffer[parser->byteOffset])) {
        result = true;

        parser->byteOffset++;
        parser->column++;
    }

    return result;
}

b32 ConsumeHexDigit_(EtcTxtParseState *parser) {
    b32 result = (ConsumeDigit_(parser) ||
                  ConsumeCharacter_(parser, 'A') || ConsumeCharacter_(parser, 'a') ||
                  ConsumeCharacter_(parser, 'B') || ConsumeCharacter_(parser, 'b') ||
                  ConsumeCharacter_(parser, 'C') || ConsumeCharacter_(parser, 'c') ||
                  ConsumeCharacter_(parser, 'D') || ConsumeCharacter_(parser, 'd') ||
                  ConsumeCharacter_(parser, 'E') || ConsumeCharacter_(parser, 'e') ||
                  ConsumeCharacter_(parser, 'F') || ConsumeCharacter_(parser, 'f'));

    return result;
}

b32 ConsumeEndOfLine_(EtcTxtParseState *parser) {
    b32 result = false;

    if (ConsumeCharacter_(parser, '\n') ||
        ConsumeCharacterSequence_(parser, "\r\n")) {
        result = true;

        parser->column = 1;
        parser->line++;
    }

    return result;
}

b32 ConsumeWhiteSpace_(EtcTxtParseState *parser) {
    b32 result = (ConsumeCharacter_(parser, ' ') ||
                  ConsumeCharacter_(parser, '\t') ||
                  ConsumeCharacter_(parser, '\f') ||
                  ConsumeCharacter_(parser, '\v') ||
                  ConsumeEndOfLine_(parser));

    return result;
}

void ConsumeWhiteSpaceComments_(EtcTxtParseState *parser) {
    while (true) {
        if (ConsumeWhiteSpace_(parser)) {
        }
        else if (ConsumeCharacterSequence_(parser, "//")) {
            while (!AtEndOfFile_(parser) && !ConsumeEndOfLine_(parser)) {
                parser->byteOffset++;
            }
        }
        else {
            break;
        }
    }
}

b32 ConsumeBooleanLiteral_(EtcTxtParseState *parser, EtcTxtToken *token) {
    b32 result = false;

    if (ConsumeCharacterSequence_(parser, "false")) {
        token->type = EtcTxtTokenType_Boolean;
        token->text = parser->buffer + (parser->byteOffset - 5);
        token->textLength = 5;

        token->boolValue = 0;

        result = true;
    }
    else if (ConsumeCharacterSequence_(parser, "true")) {
        token->type = EtcTxtTokenType_Boolean;
        token->text = parser->buffer + (parser->byteOffset - 4);
        token->textLength = 4;

        token->boolValue = 1;

        result = true;
    }

    return result;
}

b32 ConsumeStringLiteral_(EtcTxtParseState *parser, EtcTxtToken *token) {
    b32 result = false;

    EtcTxtParseCursor startCursor = GetCursor(parser);

    const char *text = &parser->buffer[parser->byteOffset];

    if (ConsumeCharacter_(parser, '"')) {

        uint32 tokenLength = 1;

        while (true) {
            if (ConsumeCharacter_(parser, '"')) {
                result = true;
                tokenLength += 1;

                token->text = text;
                token->textLength = tokenLength;
                token->type = EtcTxtTokenType_String;

                break;
            }
            else if (ConsumeCharacter_(parser, '\\')) {
                if (ConsumeAlpha_(parser) ||
                    ConsumeCharacter_(parser, '\\') ||
                    ConsumeCharacter_(parser, '\'') ||
                    ConsumeCharacter_(parser, '\"')) {
                    tokenLength += 2;

                    continue;
                }

                break;
            }
            else if (ConsumeEndOfLine_(parser)) {
                // @TODO: set error
                break;
            }
            else if (ConsumeAnyCharacter_(parser)) {
                tokenLength += 1;
            }
            else {
                break;
            }
        }
    }

    if (!result) {
        SetCursor(parser, startCursor);
    }

    return result;
}

b32 ConsumeFloatLiteral_(EtcTxtParseState *parser, EtcTxtToken *token) {
    b32 result = false;

    EtcTxtParseCursor startCursor = GetCursor(parser);

    const char *text = &parser->buffer[parser->byteOffset];

    // @TODO: consume decimal representations too!

    if (ConsumeCharacterSequence_(parser, "0x") ||
        ConsumeCharacterSequence_(parser, "0X")) {

        uint32 textLength = 2;
        while (ConsumeHexDigit_(parser)) {
            textLength++;
        }

        if (textLength > 0) {
            char *endPtr;
            uint64 value = strtoul(text, &endPtr, 16);

            uint32 characterCount = endPtr - text;

            if (characterCount == textLength) {
                token->text = text;
                token->textLength = characterCount;
                token->type = EtcTxtTokenType_Float;

                token->floatValue = *(real32 *)(&value);

                result = true;
            }
        }
    }
    else {
        uint32 textLength = 0;
        if (ConsumeCharacter_(parser, '+') ||
            ConsumeCharacter_(parser, '-')) {
            textLength++;
        }

        b32 hasDigit = false;
        b32 hasExponent = false;
        b32 hasDecimal = false;

        while (true) {
            if (ConsumeCharacter_(parser, '.')) {
                if (!hasDigit || hasDecimal || hasExponent) {
                    goto end;
                }

                hasDecimal = true;
            }
            else if (ConsumeCharacter_(parser, 'e') ||
                     ConsumeCharacter_(parser, 'E')) {
                if (!hasDigit || hasExponent) {
                    goto end;
                }

                if (ConsumeCharacter_(parser, '+') ||
                    ConsumeCharacter_(parser, '-')) {
                    textLength++;
                }

                hasExponent = true;
            }
            else if (ConsumeDigit_(parser)) {
                hasDigit = true;
            }
            else {
                break;
            }

            textLength++;
        }


        if (hasDigit && hasDecimal) {
            char *endPtr;
            real32 value = strtof(text, &endPtr);

            uint32 characterCount = endPtr - text;

            if (characterCount == textLength) {
                token->text = text;
                token->textLength = characterCount;
                token->type = EtcTxtTokenType_Float;

                token->floatValue = value;

                result = true;
            }
        }
    }

 end:
    if (!result) {
        SetCursor(parser, startCursor);
    }

    return result;
}

b32 ConsumeIntegerLiteral_(EtcTxtParseState *parser, EtcTxtToken *token) {
    b32 result = false;

    EtcTxtParseCursor startCursor = GetCursor(parser);

    const char *text = &parser->buffer[parser->byteOffset];

    uint32 textLength = 0;

    if (ConsumeCharacter_(parser, '-')) {
        textLength++;
    }

    while (ConsumeDigit_(parser)) {
        textLength++;
    }

    // @TODO: this _might_ fail for very large unsigned ints!

    if (textLength > 0) {
        char *endPtr;
        int64 value = strtoll(text, &endPtr, 10);

        uint32 characterCount = endPtr - text;

        if (characterCount == textLength) {
            token->text = text;
            token->textLength = textLength;
            token->type = EtcTxtTokenType_Integer;

            token->int64Value = value;

            result = true;
        }
    }

    if (!result) {
        SetCursor(parser, startCursor);
    }

    return result;
}

b32 ConsumeIdentifier_(EtcTxtParseState *parser, EtcTxtToken *token) {
    b32 result = false;

    EtcTxtParseCursor startCursor = GetCursor(parser);

    const char *text = &parser->buffer[parser->byteOffset];
    uint32 textLength = 0;

    if (ConsumeAlpha_(parser) ||
        ConsumeCharacter_(parser, '_')) {

        textLength++;

        while (ConsumeAlpha_(parser) ||
               ConsumeDigit_(parser) ||
               ConsumeCharacter_(parser, '_')) {
            textLength++;
        }

        token->text = text;
        token->textLength = textLength;
        token->type = EtcTxtTokenType_Identifier;

        result = true;
    }

    if (!result) {
        SetCursor(parser, startCursor);
    }

    return result;
}


EtcTxtToken ConsumeNextToken_(EtcTxtParseState *parser) {
    EtcTxtToken token = {};

    if (parser->buffer == NULL) {
        // @TODO: set error
        return token;
    }

    ConsumeWhiteSpaceComments_(parser);

    if (AtEndOfFile_(parser)) {
    }
    else if (ConsumeCharacter_(parser, ':')) {
        token.type = EtcTxtTokenType_Colon;
        token.text = parser->buffer + (parser->byteOffset - 1);
        token.textLength = 1;
    }
    else if (ConsumeCharacter_(parser, ',')) {
        token.type = EtcTxtTokenType_Comma;
        token.text = parser->buffer + (parser->byteOffset - 1);
        token.textLength = 1;
    }
    else if (ConsumeCharacter_(parser, '{')) {
        token.type = EtcTxtTokenType_OpenBrace;
        token.text = parser->buffer + (parser->byteOffset - 1);
        token.textLength = 1;
    }
    else if (ConsumeCharacter_(parser, '}')) {
        token.type = EtcTxtTokenType_CloseBrace;
        token.text = parser->buffer + (parser->byteOffset - 1);
        token.textLength = 1;
    }
    else if (ConsumeCharacter_(parser, '[')) {
        token.type = EtcTxtTokenType_OpenBracket;
        token.text = parser->buffer + (parser->byteOffset - 1);
        token.textLength = 1;
    }
    else if (ConsumeCharacter_(parser, ']')) {
        token.type = EtcTxtTokenType_CloseBracket;
        token.text = parser->buffer + (parser->byteOffset - 1);
        token.textLength = 1;
    }
    else if (ConsumeCharacterSequence_(parser, "NULL")) {
        token.type = EtcTxtTokenType_Null;
        token.text = parser->buffer + (parser->byteOffset - 4);
        token.textLength = 4;
    }
    else if (ConsumeBooleanLiteral_(parser, &token)) {
    }
    else if (ConsumeStringLiteral_(parser, &token)) {
    }
    else if (ConsumeFloatLiteral_(parser, &token)) {
    }
    else if (ConsumeIntegerLiteral_(parser, &token)) {
    }
    else if (ConsumeIdentifier_(parser, &token)) {
    }
    else {
        ParserSetError_(parser, EtcTxtParseErrorType_IllegalCharacter, "Illegal character '%c'", parser->buffer[parser->byteOffset]);
    }

    return token;
}

b32 ConsumeTokenOfType_(EtcTxtParseState *parser, EtcTxtTokenType type, EtcTxtToken *token = NULL) {
    ASSERT(type != EtcTxtTokenType_None);

    b32 result = false;

    EtcTxtParseCursor startCursor = GetCursor(parser);

    EtcTxtToken nextToken = ConsumeNextToken_(parser);

    result = nextToken.type == type;
    
    if (result && token != NULL) {
        *token = nextToken;
    }

    if (!result) {
        SetCursor(parser, startCursor);
    }

    return result;
}

b32 PeekTokenOfType_(EtcTxtParseState *parser, EtcTxtTokenType type, EtcTxtToken *token = NULL) {
    EtcTxtParseCursor startCursor = GetCursor(parser);

    b32 result = ConsumeTokenOfType_(parser, type, token);

    SetCursor(parser, startCursor);

    return result;
}


b32 ParseHasError(EtcTxtParseState *parser) {
    b32 result = parser->error.type != EtcTxtParseErrorType_None;

    return result;
}

b32 ParseCheckHasIdentifier_(EtcTxtParseState *parser) {
    if (!parser->isArray[parser->stackDepth]) {
        if (!parser->hasIdentifier[parser->stackDepth]) {
            // @TODO: set error
            return false;
        }
    }

    return true;
}

b32 ParseCheckIsValidArrayValue_(EtcTxtParseState *parser) {
    if (parser->isArray[parser->stackDepth]) {
        if (ConsumeTokenOfType_(parser, EtcTxtTokenType_Comma)) {
            return true;
        }

        if (!PeekTokenOfType_(parser, EtcTxtTokenType_CloseBracket)) {
            // @TODO: set error
            return false;
        }
    }

    return true;
}

b32 ParseIdentifier(EtcTxtParseState *parser, const char **identifier, const char *matchingString) {
    if (ParseHasError(parser)) {
        return false;
    }

    if (parser->isArray[parser->stackDepth]) {
        // @TODO: set error, can't have identifier in array
        return false;
    }

    if (parser->hasIdentifier[parser->stackDepth]) {
        // @TODO: set error
        return false;
    }

    b32 result = false;

    EtcTxtParseCursor startCursor = GetCursor(parser);

    EtcTxtToken token = {};
    if (ConsumeTokenOfType_(parser, EtcTxtTokenType_Identifier, &token)) {
        if (ConsumeTokenOfType_(parser, EtcTxtTokenType_Colon)) {

            char *identifierString = ALLOCATE_ARRAY(char, token.textLength + 1);
            StringCopy(identifierString, token.textLength + 1, token.text);
            PushBack(&parser->allocatedStrings, (const char *)identifierString);

            if (matchingString == NULL ||
                StringEquals(identifierString, (const char *)matchingString)) {

                if (identifier != NULL) {
                    *identifier = identifierString;
                }

                parser->hasIdentifier[parser->stackDepth] = true;
                result = true;
            }
        }
        else {
            // @TODO: set error
        }
    }

    if (!result) {
        SetCursor(parser, startCursor);
    }

    return result;
}

b32 ParsePushObjectValue(EtcTxtParseState *parser) {
    if (ParseHasError(parser) ||
        !ParseCheckHasIdentifier_(parser)) {
        return false;
    }

    if (parser->stackDepth >= ETCTXT_MAX_STACK_DEPTH - 1) {
        // @TODO: set error
        return false;
    }

    b32 result = false;

    if (ConsumeTokenOfType_(parser, EtcTxtTokenType_OpenBrace)) {
        parser->stackDepth += 1;
        parser->isArray[parser->stackDepth] = false;
        parser->hasIdentifier[parser->stackDepth] = false;

        result = true;
    }

    return result;
}

b32 ParsePopObjectValue(EtcTxtParseState *parser) {
    if (ParseHasError(parser)) {
        return false;
    }

    if (parser->stackDepth == 0) {
        // @TODO: set error
        return false;
    }

    if (parser->hasIdentifier[parser->stackDepth]) {
        // @TODO: set error
        return false;
    }

    b32 result = false;

    if (ConsumeTokenOfType_(parser, EtcTxtTokenType_CloseBrace)) {
        parser->stackDepth -= 1;

        if (ParseCheckIsValidArrayValue_(parser)) {
            parser->hasIdentifier[parser->stackDepth] = false;
            result = true;
        }
    }

    return result;
}

b32 ParsePushArrayValue(EtcTxtParseState *parser) {
    if (ParseHasError(parser) ||
        !ParseCheckHasIdentifier_(parser)) {
        return false;
    }

    if (parser->stackDepth >= ETCTXT_MAX_STACK_DEPTH - 1) {
        // @TODO: set error
        return false;
    }

    b32 result = false;

    if (ConsumeTokenOfType_(parser, EtcTxtTokenType_OpenBracket)) {
        parser->stackDepth += 1;
        parser->isArray[parser->stackDepth] = true;

        result = true;
    }

    return result;
}

b32 ParsePopArrayValue(EtcTxtParseState *parser) {
    if (ParseHasError(parser) ||
        !ParseCheckHasIdentifier_(parser)) {
        return false;
    }

    if (!parser->isArray[parser->stackDepth]) {
        // @TODO: set error
        return false;
    }

    if (parser->stackDepth == 0) {
        // @TODO: set error
        return false;
    }

    b32 result = false;

    if (ConsumeTokenOfType_(parser, EtcTxtTokenType_CloseBracket)) {
        parser->isArray[parser->stackDepth] = false;
        parser->stackDepth -= 1;

        if (ParseCheckIsValidArrayValue_(parser)) {
            parser->hasIdentifier[parser->stackDepth] = false;
            result = true;
        }
    }

    return result;
}

b32 ParseStringValue(EtcTxtParseState *parser, const char **value) {
    if (ParseHasError(parser) ||
        !ParseCheckHasIdentifier_(parser)) {
        return false;
    }

    b32 result = false;

    EtcTxtToken token = {};
    if (ConsumeTokenOfType_(parser, EtcTxtTokenType_Null, &token)) {
        if (ParseCheckIsValidArrayValue_(parser)) {
            if (value != NULL) {
                *value = NULL;
            }

            parser->hasIdentifier[parser->stackDepth] = false;
            result = true;
        }
    }
    else if (ConsumeTokenOfType_(parser, EtcTxtTokenType_String, &token)) {
        if (ParseCheckIsValidArrayValue_(parser)) {
            if (value != NULL) {
                const char *str = UnescapeString(token.text + 1, token.textLength - 2);
                PushBack(&parser->allocatedStrings, str);

                *value = str;
            }

            parser->hasIdentifier[parser->stackDepth] = false;
            result = true;
        }
    }

    return result;
}

b32 ParseStringValue(EtcTxtParseState *parser, char *value, uint32 maxLength) {
    ASSERT(value != NULL);
    ASSERT(maxLength > 0);

    if (ParseHasError(parser) ||
        !ParseCheckHasIdentifier_(parser)) {
        return false;
    }

    b32 result = false;

    EtcTxtToken token = {};
    if (ConsumeTokenOfType_(parser, EtcTxtTokenType_Null, &token)) {
        if (ParseCheckIsValidArrayValue_(parser)) {
            value[0] = 0;

            parser->hasIdentifier[parser->stackDepth] = false;
            result = true;
        }
    }
    else if (ConsumeTokenOfType_(parser, EtcTxtTokenType_String, &token)) {
        if (ParseCheckIsValidArrayValue_(parser)) {
            if (value != NULL) {
                const char *str = UnescapeString(token.text + 1, token.textLength - 2);
                PushBack(&parser->allocatedStrings, str);

                StringCopy(value, maxLength, str);
            }

            parser->hasIdentifier[parser->stackDepth] = false;
            result = true;
        }
    }

    return result;
}

b32 ParseFloatValue(EtcTxtParseState *parser, real32 *value) {
    if (ParseHasError(parser) ||
        !ParseCheckHasIdentifier_(parser)) {
        return false;
    }

    b32 result = false;

    EtcTxtToken token = {};
    if (ConsumeTokenOfType_(parser, EtcTxtTokenType_Float, &token) ||
        ConsumeTokenOfType_(parser, EtcTxtTokenType_Integer, &token)) {

        if (ParseCheckIsValidArrayValue_(parser)) {
            if (value != NULL) {
                switch (token.type) {
                    case EtcTxtTokenType_Float: {
                        *value = token.floatValue;
                    } break;

                    case EtcTxtTokenType_Integer: {
                        *value = (real32)token.int64Value;
                    } break;

                    default: {
                        ASSERT(false);
                    } break;
                }
            }

            parser->hasIdentifier[parser->stackDepth] = false;
            result = true;
        }
    }
    
    return result;
}

b32 ParseBooleanValue(EtcTxtParseState *parser, bool *value) {
    if (ParseHasError(parser) ||
        !ParseCheckHasIdentifier_(parser)) {
        return false;
    }

    b32 result = false;

    EtcTxtToken token = {};
    if (ConsumeTokenOfType_(parser, EtcTxtTokenType_Boolean, &token)) {

        if (ParseCheckIsValidArrayValue_(parser)) {
            if (value != NULL) {
                *value = token.boolValue;
            }

            parser->hasIdentifier[parser->stackDepth] = false;
            result = true;
        }
    }

    return result;
}

b32 ParseBoolean32Value(EtcTxtParseState *parser, b32 *value) {
    if (ParseHasError(parser) ||
        !ParseCheckHasIdentifier_(parser)) {
        return false;
    }

    b32 result = false;

    EtcTxtToken token = {};
    if (ConsumeTokenOfType_(parser, EtcTxtTokenType_Boolean, &token)) {

        if (ParseCheckIsValidArrayValue_(parser)) {
            if (value != NULL) {
                *value = token.boolValue;
            }

            parser->hasIdentifier[parser->stackDepth] = false;
            result = true;
        }
    }

    return result;
}


b32 ParseInt64Value(EtcTxtParseState *parser, int64 *value) {
    if (ParseHasError(parser) ||
        !ParseCheckHasIdentifier_(parser)) {
        return false;
    }

    b32 result = false;

    EtcTxtToken token = {};
    if (ConsumeTokenOfType_(parser, EtcTxtTokenType_Integer, &token)) {

        if (ParseCheckIsValidArrayValue_(parser)) {
            if (value != NULL) {
                *value = (int64)token.int64Value;
            }

            parser->hasIdentifier[parser->stackDepth] = false;
            result = true;
        }
    }
    
    return result;
}

b32 ParseInt8Value(EtcTxtParseState *parser, int8 *value) {
    int64 value64;
    b32 result = ParseInt64Value(parser, &value64);

    if (result && value != NULL) {
        *value = (int8)value64;
    }

    return result;
}

b32 ParseInt16Value(EtcTxtParseState *parser, int16 *value) {
    int64 value64;
    b32 result = ParseInt64Value(parser, &value64);

    if (result && value != NULL) {
        *value = (int16)value64;
    }

    return result;
}

b32 ParseInt32Value(EtcTxtParseState *parser, int32 *value) {
    int64 value64;
    b32 result = ParseInt64Value(parser, &value64);

    if (result && value != NULL) {
        *value = (int32)value64;
    }

    return result;
}

b32 ParseUInt64Value(EtcTxtParseState *parser, uint64 *value) {
    if (ParseHasError(parser) ||
        !ParseCheckHasIdentifier_(parser)) {
        return false;
    }

    b32 result = false;

    EtcTxtToken token = {};
    if (ConsumeTokenOfType_(parser, EtcTxtTokenType_Integer, &token)) {

        if (ParseCheckIsValidArrayValue_(parser)) {
            if (value != NULL) {
                *value = (uint64)token.int64Value;
            }

            parser->hasIdentifier[parser->stackDepth] = false;
            result = true;
        }
    }
    
    return result;
}

b32 ParseUInt8Value(EtcTxtParseState *parser, uint8 *value) {
    uint64 value64;
    b32 result = ParseUInt64Value(parser, &value64);

    if (result && value != NULL) {
        *value = (uint8)value64;
    }

    return result;
}

b32 ParseUInt16Value(EtcTxtParseState *parser, uint16 *value) {
    uint64 value64;
    b32 result = ParseUInt64Value(parser, &value64);

    if (result && value != NULL) {
        *value = (uint16)value64;
    }

    return result;
}

b32 ParseUInt32Value(EtcTxtParseState *parser, uint32 *value) {
    uint64 value64;
    b32 result = ParseUInt64Value(parser, &value64);

    if (result && value != NULL) {
        *value = (uint32)value64;
    }

    return result;
}

b32 ParseEnumValue_(EtcTxtParseState *parser, const char **valueTable, uint32 valueCount, uint32 *value) {
    if (ParseHasError(parser) ||
        !ParseCheckHasIdentifier_(parser)) {
        return false;
    }

    b32 result = false;

    EtcTxtParseCursor startCursor = GetCursor(parser);

    EtcTxtToken token = {};
    if (ConsumeTokenOfType_(parser, EtcTxtTokenType_Identifier, &token)) {

        uint32 index;
        if (FindString(valueTable, valueCount, token.text, token.textLength, &index)) {

            if (ParseCheckIsValidArrayValue_(parser)) {
                if (value != NULL) {
                    *value = index;
                }
            
                parser->hasIdentifier[parser->stackDepth] = false;
                result = true;
            }
        }
    }

    if (!result) {
        SetCursor(parser, startCursor);
    }

    return result;
}

#define ParseEnumValue(parser, enumType, enumValue) ParseEnumValue_(parser, ENUM_STRINGTABLE(enumType), ENUM_VALUECOUNT(enumType), (uint32 *)enumValue)

b32 ParseVec2Value(EtcTxtParseState *parser, vec2 *value) {
    if (ParseHasError(parser) ||
        !ParseCheckHasIdentifier_(parser)) {
        return false;
    }

    EtcTxtParseCursor startCursor = GetCursor(parser);

    b32 result = false;

    vec2 vector;
    if (ParsePushArrayValue(parser) &&
        ParseFloatValue(parser, &vector.x) &&
        ParseFloatValue(parser, &vector.y) &&
        ParsePopArrayValue(parser)) {

        if (value != NULL) {
            *value = vector;
        }

        result = true;
    }

    if (!result) {
        SetCursor(parser, startCursor);
    }

    return result;
}

b32 ParseVec3Value(EtcTxtParseState *parser, vec3 *value) {
    if (ParseHasError(parser) ||
        !ParseCheckHasIdentifier_(parser)) {
        return false;
    }

    EtcTxtParseCursor startCursor = GetCursor(parser);

    b32 result = false;

    vec3 vector;
    if (ParsePushArrayValue(parser) &&
        ParseFloatValue(parser, &vector.x) &&
        ParseFloatValue(parser, &vector.y) &&
        ParseFloatValue(parser, &vector.z) &&
        ParsePopArrayValue(parser)) {

        if (value != NULL) {
            *value = vector;
        }

        result = true;
    }

    if (!result) {
        SetCursor(parser, startCursor);
    }

    return result;
}

b32 ParseVec4Value(EtcTxtParseState *parser, vec4 *value) {
    if (ParseHasError(parser) ||
        !ParseCheckHasIdentifier_(parser)) {
        return false;
    }

    EtcTxtParseCursor startCursor = GetCursor(parser);

    b32 result = false;

    vec4 vector;
    if (ParsePushArrayValue(parser) &&
        ParseFloatValue(parser, &vector.x) &&
        ParseFloatValue(parser, &vector.y) &&
        ParseFloatValue(parser, &vector.z) &&
        ParseFloatValue(parser, &vector.w) &&
        ParsePopArrayValue(parser)) {

        if (value != NULL) {
            *value = vector;
        }

        result = true;
    }

    if (!result) {
        SetCursor(parser, startCursor);
    }

    return result;
}

b32 ParseVec2iValue(EtcTxtParseState *parser, vec2i *value) {
    if (ParseHasError(parser) ||
        !ParseCheckHasIdentifier_(parser)) {
        return false;
    }

    EtcTxtParseCursor startCursor = GetCursor(parser);

    b32 result = false;

    vec2i vector;
    if (ParsePushArrayValue(parser) &&
        ParseInt32Value(parser, &vector.x) &&
        ParseInt32Value(parser, &vector.y) &&
        ParsePopArrayValue(parser)) {

        if (value != NULL) {
            *value = vector;
        }

        result = true;
    }

    if (!result) {
        SetCursor(parser, startCursor);
    }

    return result;
}

b32 ParseVec3iValue(EtcTxtParseState *parser, vec3i *value) {
    if (ParseHasError(parser) ||
        !ParseCheckHasIdentifier_(parser)) {
        return false;
    }

    EtcTxtParseCursor startCursor = GetCursor(parser);

    b32 result = false;

    vec3i vector;
    if (ParsePushArrayValue(parser) &&
        ParseInt32Value(parser, &vector.x) &&
        ParseInt32Value(parser, &vector.y) &&
        ParseInt32Value(parser, &vector.z) &&
        ParsePopArrayValue(parser)) {

        if (value != NULL) {
            *value = vector;
        }

        result = true;
    }

    if (!result) {
        SetCursor(parser, startCursor);
    }

    return result;
}

b32 ParseVec4iValue(EtcTxtParseState *parser, vec4i *value) {
    if (ParseHasError(parser) ||
        !ParseCheckHasIdentifier_(parser)) {
        return false;
    }

    EtcTxtParseCursor startCursor = GetCursor(parser);

    b32 result = false;

    vec4i vector;
    if (ParsePushArrayValue(parser) &&
        ParseInt32Value(parser, &vector.x) &&
        ParseInt32Value(parser, &vector.y) &&
        ParseInt32Value(parser, &vector.z) &&
        ParseInt32Value(parser, &vector.w) &&
        ParsePopArrayValue(parser)) {

        if (value != NULL) {
            *value = vector;
        }

        result = true;
    }

    if (!result) {
        SetCursor(parser, startCursor);
    }

    return result;
}

b32 ParseMat3Value(EtcTxtParseState *parser, mat3 *value) {
    if (ParseHasError(parser) ||
        !ParseCheckHasIdentifier_(parser)) {
        return false;
    }

    EtcTxtParseCursor startCursor = GetCursor(parser);

    b32 result = false;

    mat3 matrix;
    if (ParsePushArrayValue(parser) &&
        ParseVec3Value(parser, &matrix.columns[0]) &&
        ParseVec3Value(parser, &matrix.columns[1]) &&
        ParseVec3Value(parser, &matrix.columns[2]) &&
        ParsePopArrayValue(parser)) {

        if (value != NULL) {
            *value = matrix;
        }

        result = true;
    }

    if (!result) {
        SetCursor(parser, startCursor);
    }

    return result;
}

b32 ParseMat4Value(EtcTxtParseState *parser, mat4 *value) {
    if (ParseHasError(parser) ||
        !ParseCheckHasIdentifier_(parser)) {
        return false;
    }

    EtcTxtParseCursor startCursor = GetCursor(parser);

    b32 result = false;

    mat4 matrix;
    if (ParsePushArrayValue(parser) &&
        ParseVec4Value(parser, &matrix.columns[0]) &&
        ParseVec4Value(parser, &matrix.columns[1]) &&
        ParseVec4Value(parser, &matrix.columns[2]) &&
        ParseVec4Value(parser, &matrix.columns[3]) &&
        ParsePopArrayValue(parser)) {

        if (value != NULL) {
            *value = matrix;
        }

        result = true;
    }

    if (!result) {
        SetCursor(parser, startCursor);
    }

    return result;
}

b32 ParseQuaternionValue(EtcTxtParseState *parser, quaternion *value) {
    if (ParseHasError(parser) ||
        !ParseCheckHasIdentifier_(parser)) {
        return false;
    }

    EtcTxtParseCursor startCursor = GetCursor(parser);

    b32 result = false;

    quaternion vector;
    if (ParsePushArrayValue(parser) &&
        ParseFloatValue(parser, &vector.x) &&
        ParseFloatValue(parser, &vector.y) &&
        ParseFloatValue(parser, &vector.z) &&
        ParseFloatValue(parser, &vector.w) &&
        ParsePopArrayValue(parser)) {

        if (value != NULL) {
            *value = vector;
        }

        result = true;
    }

    if (!result) {
        SetCursor(parser, startCursor);
    }

    return result;
}

b32 ParseSkipValue(EtcTxtParseState *parser) {
    if (ParseHasError(parser) ||
        !ParseCheckHasIdentifier_(parser)) {
        return false;
    }

    b32 result = false;

    if (ParsePushObjectValue(parser)) {
        while (!ParseHasError(parser)) {
            if (ParseIdentifier(parser)) {
                ParseSkipValue(parser);
            }
            else {
                break;
            }
        }

        if (ParsePopObjectValue(parser)) {
            result = true;
        }
        else {
            // @TODO: set error
        }
    }
    else if (ParsePushArrayValue(parser)) {
        while (!ParseHasError(parser)) {
            if (!ParseSkipValue(parser)) {
                break;
            }
        }

        if (ParsePopArrayValue(parser)) {
            result = true;
        }
        else {
            // @TODO: set error
        }
    }
    else if (ParseStringValue(parser) ||
             ParseBooleanValue(parser) ||
             ParseFloatValue(parser) ||
             ParseInt64Value(parser) ||
             ParseInt32Value(parser) ||
             ParseInt16Value(parser) ||
             ParseInt8Value(parser) ||
             ParseUInt64Value(parser) ||
             ParseUInt32Value(parser) ||
             ParseUInt16Value(parser) ||
             ParseUInt8Value(parser)) {
        result = true;
    }
    // Handle enum literal values
    else if (ConsumeTokenOfType_(parser, EtcTxtTokenType_Identifier) &&
             ParseCheckIsValidArrayValue_(parser)) {
        parser->hasIdentifier[parser->stackDepth] = false;
        result = true;
    }
    else {
        // @TODO: set error
    }

    return result;
}

b32 ParseCountArrayValues(EtcTxtParseState *parser, uint32 *arrayValueCount) {
    if (ParseHasError(parser) ||
        !ParseCheckHasIdentifier_(parser)) {
        return false;
    }

    EtcTxtParseCursor startCursor = GetCursor(parser);

    b32 result = false;

    uint32 count = 0;
    if (ParsePushArrayValue(parser)) {
        while (ParseSkipValue(parser)) {
            count++;
        }

        if (ParsePopArrayValue(parser)) {
            if (arrayValueCount != NULL) {
                *arrayValueCount = count;
            }

            result = true;
        }
    }
    
    SetCursor(parser, startCursor);

    return result;
}


// Writing
void EtcTxtCopyToBuffer_(EtcTxtWriteState *writer, char *data, uint32 dataLength) {
    ASSERT(writer->currentBuffer != NULL);
    ASSERT(writer->bytesWrittenToCurrentBuffer <= writer->currentBufferSize);

    uint32 bytesCopied = 0;

    while (bytesCopied < dataLength) {
        uint32 bytesLeft = dataLength - bytesCopied;
        uint32 bytesLeftInCurrentBuffer = writer->currentBufferSize - writer->bytesWrittenToCurrentBuffer;

        if (bytesLeftInCurrentBuffer == 0) {
            ASSERT(writer->writeCallback != NULL);

            char *buffer = writer->currentBuffer;
            uint32 newBufferSize = writer->currentBufferSize;
            char *newBuffer = writer->writeCallback(writer->currentBuffer, &newBufferSize, writer->callbackUserData);

            // @TODO: probably best to have gentler error handling
            ASSERT(newBuffer != NULL);
            ASSERT(newBufferSize != 0);

            writer->currentBuffer = newBuffer;
            writer->currentBufferSize = newBufferSize;
            writer->bytesWrittenToCurrentBuffer = 0;
        }

        uint32 bytesToCopy = Min(bytesLeft, bytesLeftInCurrentBuffer);
        Plat.CopyMem(writer->currentBuffer + writer->bytesWrittenToCurrentBuffer,
                     data + bytesCopied,
                     bytesToCopy);
        writer->bytesWrittenToCurrentBuffer += bytesToCopy;
        writer->totalBytesWritten += bytesToCopy;
        bytesCopied += bytesToCopy;
    }
}

char *EtcTxtWriteStringf_Callback(char *buffer, EtcTxtWriteState *writer, uint32 length) {
    EtcTxtCopyToBuffer_(writer, buffer, length);
    return buffer;
}

void WriteString(EtcTxtWriteState *writer, const char *string) {
    if (writer->file.data != NULL) {
        WriteString(&writer->file, string);
    }
    else {
        EtcTxtCopyToBuffer_(writer, (char *)string, StringLength(string));
    }
}

void WriteStringf(EtcTxtWriteState *writer, const char *fmt, ...) {
    va_list args;
    va_start(args, fmt);

    if (writer->file.data != NULL) {
        WriteStringfv(&writer->file, fmt, args);
    }
    else {
        char buffer[STB_SPRINTF_MIN] = {};
        stbsp_vsprintfcb((STBSP_SPRINTFCB *)EtcTxtWriteStringf_Callback, writer, buffer, fmt, args);
    }

    va_end(args);
}

b32 EtcTxtFieldNameIsValid_(const char *fieldName) {
    if (fieldName == NULL) {
        return false;
    }

    if (StringContainsCharacter(fieldName, ' ')) {
        return false;
    }

    return true;
}

void WritePadSpaces_(EtcTxtWriteState *writer) {
    if (HasFlag(writer->flags, EtcTxtWriteFlag_Minify)) {
        return;
    }

    if (writer->justPoppedGroup && !writer->isArray[writer->stackDepth]) {
        WriteString(writer, "\n");
    }

    writer->justPushedGroup = false;
    writer->justPoppedGroup = false;

    static const char *indentationValue = "    ";
    for (int i = 0; i < writer->stackDepth; i++) {
        WriteString(writer, indentationValue);
    }
}

void WriteFieldName_(EtcTxtWriteState *writer, const char *fieldName) {

    // @NOTE: fieldnames aren't written inside arrays!
    if (writer->isArray[writer->stackDepth]) {
        return;
    }

    ASSERT(EtcTxtFieldNameIsValid_(fieldName));
    
    WriteString(writer, fieldName);
    if (HasFlag(writer->flags, EtcTxtWriteFlag_Minify)) {
        WriteString(writer, ":");
    }
    else {
        WriteString(writer, ": ");
    }
}


void WritePushObjectValue(EtcTxtWriteState *writer, const char *fieldName) {
    ASSERT(writer->stackDepth < ETCTXT_MAX_STACK_DEPTH - 1);

    if (!HasFlag(writer->flags, EtcTxtWriteFlag_Minify) &&
        !writer->justPoppedGroup && !writer->justPushedGroup) {
        WriteString(writer, "\n");
    }

    WritePadSpaces_(writer);
    WriteFieldName_(writer, fieldName);

    if (HasFlag(writer->flags, EtcTxtWriteFlag_Minify)) {
        WriteString(writer, "{");
    }
    else {
        WriteString(writer, "{\n");
    }

    writer->stackDepth += 1;
    writer->justPushedGroup = true;
}

void WritePopObjectValue(EtcTxtWriteState *writer) {
    ASSERT(writer->stackDepth > 0);

    writer->stackDepth -= 1;
    writer->justPoppedGroup = false;

    WritePadSpaces_(writer);
    if (HasFlag(writer->flags, EtcTxtWriteFlag_Minify)) {
        if (writer->isArray[writer->stackDepth]) {
            WriteString(writer, "},");
        }
        else {
            WriteString(writer, "}");
        }
    }
    else {
        if (writer->isArray[writer->stackDepth]) {
            WriteString(writer, "},\n");
        }
        else {
            WriteString(writer, "}\n");
        }
    }

    writer->justPoppedGroup = true;
}

void WritePushArrayValue(EtcTxtWriteState *writer, const char *fieldName) {
    ASSERT(writer->stackDepth < ETCTXT_MAX_STACK_DEPTH - 1);

    if (!HasFlag(writer->flags, EtcTxtWriteFlag_Minify) &&
        !writer->justPoppedGroup && !writer->justPushedGroup) {
        WriteString(writer, "\n");
    }

    WritePadSpaces_(writer);
    WriteFieldName_(writer, fieldName);

    if (HasFlag(writer->flags, EtcTxtWriteFlag_Minify)) {
        WriteString(writer, "[");
    }
    else {
        WriteString(writer, "[\n");
    }

    writer->stackDepth += 1;
    writer->isArray[writer->stackDepth] = true;

    writer->justPushedGroup = true;
}

void WritePopArrayValue(EtcTxtWriteState *writer) {
    ASSERT(writer->stackDepth > 0);
    ASSERT(writer->isArray[writer->stackDepth]);

    writer->isArray[writer->stackDepth] = false;
    writer->stackDepth -= 1;
    writer->justPoppedGroup = false;

    WritePadSpaces_(writer);
    if (HasFlag(writer->flags, EtcTxtWriteFlag_Minify)) {
        if (writer->isArray[writer->stackDepth]) {
            WriteString(writer, "],");
        }
        else {
            WriteString(writer, "]");
        }
    }
    else {
        if (writer->isArray[writer->stackDepth]) {
            WriteString(writer, "],\n");
        }
        else {
            WriteString(writer, "]\n");
        }
    }

    writer->justPoppedGroup = true;
}

void WriteInt8Value(EtcTxtWriteState *writer, int8 value, const char *fieldName) {
    WritePadSpaces_(writer);
    WriteFieldName_(writer, fieldName);

    const char *fmt;
    if (HasFlag(writer->flags, EtcTxtWriteFlag_Minify)) {
        if (writer->isArray[writer->stackDepth]) {
            fmt = "%d,";
        }
        else {
            fmt = "%d";
        }
    }
    else {
        if (writer->isArray[writer->stackDepth]) {
            fmt = "%d,\n";
        }
        else {
            fmt = "%d\n";
        }
    }

    WriteStringf(writer, fmt, value);
}

void WriteInt16Value(EtcTxtWriteState *writer, int16 value, const char *fieldName) {
    WritePadSpaces_(writer);
    WriteFieldName_(writer, fieldName);

    const char *fmt;
    if (HasFlag(writer->flags, EtcTxtWriteFlag_Minify)) {
        if (writer->isArray[writer->stackDepth]) {
            fmt = "%d,";
        }
        else {
            fmt = "%d";
        }
    }
    else {
        if (writer->isArray[writer->stackDepth]) {
            fmt = "%d,\n";
        }
        else {
            fmt = "%d\n";
        }
    }

    WriteStringf(writer, fmt, value);
}

void WriteInt32Value(EtcTxtWriteState *writer, int32 value, const char *fieldName) {
    WritePadSpaces_(writer);
    WriteFieldName_(writer, fieldName);

    const char *fmt;
    if (HasFlag(writer->flags, EtcTxtWriteFlag_Minify)) {
        if (writer->isArray[writer->stackDepth]) {
            fmt = "%d,";
        }
        else {
            fmt = "%d";
        }
    }
    else {
        if (writer->isArray[writer->stackDepth]) {
            fmt = "%d,\n";
        }
        else {
            fmt = "%d\n";
        }
    }

    WriteStringf(writer, fmt, value);
}

void WriteInt64Value(EtcTxtWriteState *writer, int64 value, const char *fieldName) {
    WritePadSpaces_(writer);
    WriteFieldName_(writer, fieldName);

    const char *fmt;
    if (HasFlag(writer->flags, EtcTxtWriteFlag_Minify)) {
        if (writer->isArray[writer->stackDepth]) {
            fmt = "%ll,";
        }
        else {
            fmt = "%ll" ;
        }
    }
    else {
        if (writer->isArray[writer->stackDepth]) {
            fmt = "%ll,\n";
        }
        else {
            fmt = "%ll";
        }
    }

    WriteStringf(writer, fmt, value);
}

void WriteUInt8Value(EtcTxtWriteState *writer, uint8 value, const char *fieldName) {
    WritePadSpaces_(writer);
    WriteFieldName_(writer, fieldName);

    const char *fmt;
    if (HasFlag(writer->flags, EtcTxtWriteFlag_Minify)) {
        if (writer->isArray[writer->stackDepth]) {
            fmt = "%u,";
        }
        else {
            fmt = "%u";
        }
    }
    else {
        if (writer->isArray[writer->stackDepth]) {
            fmt = "%u,\n";
        }
        else {
            fmt = "%u\n";
        }
    }

    WriteStringf(writer, fmt, value);
}

void WriteUInt16Value(EtcTxtWriteState *writer, uint16 value, const char *fieldName) {
    WritePadSpaces_(writer);
    WriteFieldName_(writer, fieldName);

    const char *fmt;
    if (HasFlag(writer->flags, EtcTxtWriteFlag_Minify)) {
        if (writer->isArray[writer->stackDepth]) {
            fmt = "%u,";
        }
        else {
            fmt = "%u";
        }
    }
    else {
        if (writer->isArray[writer->stackDepth]) {
            fmt = "%u,\n";
        }
        else {
            fmt = "%u\n";
        }
    }

    WriteStringf(writer, fmt, value);
}

void WriteUInt32Value(EtcTxtWriteState *writer, uint32 value, const char *fieldName) {
    WritePadSpaces_(writer);
    WriteFieldName_(writer, fieldName);

    const char *fmt;
    if (HasFlag(writer->flags, EtcTxtWriteFlag_Minify)) {
        if (writer->isArray[writer->stackDepth]) {
            fmt = "%u,";
        }
        else {
            fmt = "%u";
        }
    }
    else {
        if (writer->isArray[writer->stackDepth]) {
            fmt = "%u,\n";
        }
        else {
            fmt = "%u\n";
        }
    }

    WriteStringf(writer, fmt, value);
}

void WriteUInt64Value(EtcTxtWriteState *writer, uint64 value, const char *fieldName) {
    WritePadSpaces_(writer);
    WriteFieldName_(writer, fieldName);

    const char *fmt = NULL;
    if (HasFlag(writer->flags, EtcTxtWriteFlag_Minify)) {
        if (writer->isArray[writer->stackDepth]) {
            fmt = "%llu,";
        }
        else {
            fmt = "%llu";
        }
    }
    else {
        if (writer->isArray[writer->stackDepth]) {
            fmt = "%llu,\n";
        }
        else {
            fmt = "%llu\n";
        }
    }

    WriteStringf(writer, fmt, value);
}

void WriteFloatValue(EtcTxtWriteState *writer, real32 value, const char *fieldName) {
    WritePadSpaces_(writer);
    WriteFieldName_(writer, fieldName);

    const char *fmt;
    if (HasFlag(writer->flags, EtcTxtWriteFlag_Minify)) {
        if (writer->isArray[writer->stackDepth]) {
            fmt = "0x%X,";
        }
        else {
            fmt = "0x%X";
        }
    }
    else {
        if (writer->isArray[writer->stackDepth]) {
            fmt = "0x%X, // %.*g,\n";
        }
        else {
            fmt = "0x%X // %.*g\n";
        }
    }

    WriteStringf(writer, fmt, *((uint32 *)&value), FLT_DECIMAL_DIG, value);
}

void WriteBooleanValue(EtcTxtWriteState *writer, bool value, const char *fieldName) {
    WritePadSpaces_(writer);
    WriteFieldName_(writer, fieldName);

    const char *fmt;
    if (HasFlag(writer->flags, EtcTxtWriteFlag_Minify)) {
        if (writer->isArray[writer->stackDepth]) {
            fmt = "%s,";
        }
        else {
            fmt = "%s\n";
        }
    }
    else {
        if (writer->isArray[writer->stackDepth]) {
            fmt = "%s,\n";
        }
        else {
            fmt = "%s\n";
        }
    }

    const char *valueString;
    if (value) {
        valueString = "true";
    }
    else {
        valueString = "false";
    }

    WriteStringf(writer, fmt, valueString);
}

void WriteBoolean32Value(EtcTxtWriteState *writer, b32 value, const char *fieldName) {
    WritePadSpaces_(writer);
    WriteFieldName_(writer, fieldName);

    const char *fmt;
    if (HasFlag(writer->flags, EtcTxtWriteFlag_Minify)) {
        if (writer->isArray[writer->stackDepth]) {
            fmt = "%s,";
        }
        else {
            fmt = "%s\n";
        }
    }
    else {
        if (writer->isArray[writer->stackDepth]) {
            fmt = "%s,\n";
        }
        else {
            fmt = "%s\n";
        }
    }

    const char *valueString;
    if (value) {
        valueString = "true";
    }
    else {
        valueString = "false";
    }

    WriteStringf(writer, fmt, valueString);
}


void WriteStringValue(EtcTxtWriteState *writer, const char *value, uint32 valueLength, const char *fieldName) {
    WritePadSpaces_(writer);
    WriteFieldName_(writer, fieldName);

    if (value == NULL) {
        const char *string;
        if (HasFlag(writer->flags, EtcTxtWriteFlag_Minify)) {
            if (writer->isArray[writer->stackDepth]) {
                string = "NULL,";
            }
            else {
                string = "NULL\n";
            }
        }
        else {
            if (writer->isArray[writer->stackDepth]) {
                string = "NULL,\n";
            }
            else {
                string = "NULL\n";
            }
        }


        WriteStringf(writer, string);
    }
    else {
        const char *fmt;
        if (HasFlag(writer->flags, EtcTxtWriteFlag_Minify)) {
            if (writer->isArray[writer->stackDepth]) {
                fmt = "\"%.*s\",";
            }
            else {
                fmt = "\"%.*s\"";
            }
        }
        else {
            if (writer->isArray[writer->stackDepth]) {
                fmt = "\"%.*s\",\n";
            }
            else {
                fmt = "\"%.*s\"\n";
            }
        }

        WriteStringf(writer, fmt, valueLength, value);
    }
}

void WriteStringValue(EtcTxtWriteState *writer, const char *value, const char *fieldName) {
    if (value == NULL) {
        WriteStringValue(writer, value, 0, fieldName);
    }
    else {
        WriteStringValue(writer, value, strlen(value), fieldName);
    }
}

void WriteEnumValue(EtcTxtWriteState *writer, const char *enumStringValue, const char *fieldName) {
    WritePadSpaces_(writer);
    WriteFieldName_(writer, fieldName);

    const char *fmt;
    if (HasFlag(writer->flags, EtcTxtWriteFlag_Minify)) {
        if (writer->isArray[writer->stackDepth]) {
            fmt = "%s,";
        }
        else {
            fmt = "%s\n";
        }
    }
    else {
        if (writer->isArray[writer->stackDepth]) {
            fmt = "%s,\n";
        }
        else {
            fmt = "%s\n";
        }
    }

    WriteStringf(writer, fmt, enumStringValue);
}

void WriteVec2Value(EtcTxtWriteState *writer, vec2 value, const char *fieldName) {
    WritePadSpaces_(writer);
    WriteFieldName_(writer, fieldName);

    const char *fmt;
    if (HasFlag(writer->flags, EtcTxtWriteFlag_Minify)) {
        const char *fmt;
        if (writer->isArray[writer->stackDepth]) {
            fmt = "[0x%X,0x%X],";
        }
        else {
            fmt = "[0x%X,0x%X]";
        }

        WriteStringf(writer, fmt,
                     *((uint32 *)&value.x), *((uint32 *)&value.y),
                     value.x, value.y);
    }
    else {
        const char *fmt;
        if (writer->isArray[writer->stackDepth]) {
            fmt = "[0x%X, 0x%X], // [%f, %f],\n";
        }
        else {
            fmt = "[0x%X, 0x%X] // [%f, %f]\n";
        }

        WriteStringf(writer, fmt,
                     *((uint32 *)&value.x), *((uint32 *)&value.y),
                     value.x, value.y);
    }
}

void WriteVec3Value(EtcTxtWriteState *writer, vec3 value, const char *fieldName) {
    WritePadSpaces_(writer);
    WriteFieldName_(writer, fieldName);

    const char *fmt;
    if (HasFlag(writer->flags, EtcTxtWriteFlag_Minify)) {
        const char *fmt;
        if (writer->isArray[writer->stackDepth]) {
            fmt = "[0x%X,0x%X,0x%X],";
        }
        else {
            fmt = "[0x%X,0x%X,0x%X]";
        }

        WriteStringf(writer, fmt,
                     *((uint32 *)&value.x), *((uint32 *)&value.y), *((uint32 *)&value.z),
                     value.x, value.y, value.z);
    }
    else {
        const char *fmt;
        if (writer->isArray[writer->stackDepth]) {
            fmt = "[0x%X, 0x%X, 0x%X], // [%f, %f, %f],\n";
        }
        else {
            fmt = "[0x%X, 0x%X, 0x%X] // [%f, %f, %f]\n";
        }

        WriteStringf(writer, fmt,
                     *((uint32 *)&value.x), *((uint32 *)&value.y), *((uint32 *)&value.z),
                     value.x, value.y, value.z);
    }
}

void WriteVec4Value(EtcTxtWriteState *writer, vec4 value, const char *fieldName) {
    WritePadSpaces_(writer);
    WriteFieldName_(writer, fieldName);

    const char *fmt;
    if (HasFlag(writer->flags, EtcTxtWriteFlag_Minify)) {
        const char *fmt;
        if (writer->isArray[writer->stackDepth]) {
            fmt = "[0x%X,0x%X,0x%X,0x%X],";
        }
        else {
            fmt = "[0x%X,0x%X,0x%X,0x%X]";
        }

        WriteStringf(writer, fmt,
                     *((uint32 *)&value.x), *((uint32 *)&value.y), *((uint32 *)&value.z), *((uint32 *)&value.w),
                     value.x, value.y, value.z, value.w);
    }
    else {
        const char *fmt;
        if (writer->isArray[writer->stackDepth]) {
            fmt = "[0x%X, 0x%X, 0x%X, 0x%X], // [%f, %f, %f, %f],\n";
        }
        else {
            fmt = "[0x%X, 0x%X, 0x%X, 0x%X] // [%f, %f, %f, %f]\n";
        }

        WriteStringf(writer, fmt,
                     *((uint32 *)&value.x), *((uint32 *)&value.y), *((uint32 *)&value.z), *((uint32 *)&value.w),
                     value.x, value.y, value.z, value.w);
    }
}

void WriteVec2iValue(EtcTxtWriteState *writer, vec2i value, const char *fieldName) {
    WritePadSpaces_(writer);
    WriteFieldName_(writer, fieldName);

    const char *fmt;
    if (HasFlag(writer->flags, EtcTxtWriteFlag_Minify)) {
        if (writer->isArray[writer->stackDepth]) {
            fmt = "[%d,%d],";
        }
        else {
            fmt = "[%d,%d]";
        }
    }
    else {
        if (writer->isArray[writer->stackDepth]) {
            fmt = "[%d, %d],\n";
        }
        else {
            fmt = "[%d, %d]\n";
        }
    }
       
    WriteStringf(writer, fmt, value.x, value.y);
}

void WriteVec3iValue(EtcTxtWriteState *writer, vec3i value, const char *fieldName) {
    WritePadSpaces_(writer);
    WriteFieldName_(writer, fieldName);

    const char *fmt;

    if (HasFlag(writer->flags, EtcTxtWriteFlag_Minify)) {
        if (writer->isArray[writer->stackDepth]) {
            fmt = "[%d,%d, %d],";
        }
        else {
            fmt = "[%d,%d,%d]";
        }
    }
    else {
        if (writer->isArray[writer->stackDepth]) {
            fmt = "[%d, %d, %d],\n";
        }
        else {
            fmt = "[%d, %d, %d]\n";
        }
    }
       
    WriteStringf(writer, fmt, value.x, value.y, value.z);
}

void WriteVec4iValue(EtcTxtWriteState *writer, vec4i value, const char *fieldName) {
    WritePadSpaces_(writer);
    WriteFieldName_(writer, fieldName);

    const char *fmt;

    if (HasFlag(writer->flags, EtcTxtWriteFlag_Minify)) {
        if (writer->isArray[writer->stackDepth]) {
            fmt = "[%d,%d, %d,%d],";
        }
        else {
            fmt = "[%d,%d,%d,%d]";
        }
    }
    else {
        if (writer->isArray[writer->stackDepth]) {
            fmt = "[%d, %d, %d, %d],\n";
        }
        else {
            fmt = "[%d, %d, %d, %d]\n";
        }
    }
       
    WriteStringf(writer, fmt, value.x, value.y, value.z, value.w);
}

void WriteMat3Value(EtcTxtWriteState *writer, mat3 value, const char *fieldName) {
    WritePushArrayValue(writer, fieldName);

    WriteVec3Value(writer, value.columns[0]);
    WriteVec3Value(writer, value.columns[1]);
    WriteVec3Value(writer, value.columns[2]);

    WritePopArrayValue(writer);
}

void WriteMat4Value(EtcTxtWriteState *writer, mat4 value, const char *fieldName) {
    WritePushArrayValue(writer, fieldName);

    WriteVec4Value(writer, value.columns[0]);
    WriteVec4Value(writer, value.columns[1]);
    WriteVec4Value(writer, value.columns[2]);
    WriteVec4Value(writer, value.columns[3]);

    WritePopArrayValue(writer);
}

void WriteQuaternionValue(EtcTxtWriteState *writer, quaternion value, const char *fieldName) {
    WritePadSpaces_(writer);
    WriteFieldName_(writer, fieldName);

    if (HasFlag(writer->flags, EtcTxtWriteFlag_Minify)) {
        const char *fmt;
        if (writer->isArray[writer->stackDepth]) {
            fmt = "[0x%X,0x%X,0x%X,0x%X],";
        }
        else {
            fmt = "[0x%X,0x%X,0x%X,0x%X]";
        }

        WriteStringf(writer, fmt,
                     *((uint32 *)&value.x), *((uint32 *)&value.y), *((uint32 *)&value.z), *((uint32 *)&value.w),
                     value.x, value.y, value.z, value.w);
    }
    else {
        const char *fmt;
        if (writer->isArray[writer->stackDepth]) {
            fmt = "[0x%X, 0x%X, 0x%X, 0x%X], // [%f, %f, %f, %f],\n";
        }
        else {
            fmt = "[0x%X, 0x%X, 0x%X, 0x%X] // [%f, %f, %f, %f]\n";
        }

        WriteStringf(writer, fmt,
                     *((uint32 *)&value.x), *((uint32 *)&value.y), *((uint32 *)&value.z), *((uint32 *)&value.w),
                     value.x, value.y, value.z, value.w);
    }
}

