
#ifndef _ETC_ETCTXT_H
#define _ETC_ETCTXT_H

// -- Format Example

/* 
NameComponent: {
    name: "Fred"
}

TransformComponent: {
    position: [
        binary_rep, // 1.0,
        binary_rep, // 2.0,
        binary_rep, // 3.0,
    ]

    rotation: [
        binary_rep, // 1.0,
        binary_rep, // 2.0,
        binary_rep, // 3.0,
        binary_rep, // 4.0,
    ]

    scale: [
        binary_rep, // 1.0
        binary_rep, // 2.0
        binary_rep, // 3.0
    ]

    parent: 1230
    firstChild: 1231
    // etc

    vectors: [
        [binary_rep, binary_rep, binary_rep], // [1, 2, 3],
        [binary_rep, binary_rep, binary_rep], // [4, 5, 6],
    ]
}

TerrainComponent: {
    terrainAsset: {
        assetType: AssetType_Terrain
        assetName: "whatever"
    }

    // etc
}
*/

#define ETCTXT_MAX_STACK_DEPTH 16

enum EtcTxtTokenType {
    EtcTxtTokenType_None,

    EtcTxtTokenType_Colon,
    EtcTxtTokenType_Comma,
    EtcTxtTokenType_OpenBrace,
    EtcTxtTokenType_CloseBrace,
    EtcTxtTokenType_OpenBracket,
    EtcTxtTokenType_CloseBracket,
    EtcTxtTokenType_Null,
    EtcTxtTokenType_Boolean,
    EtcTxtTokenType_String,
    EtcTxtTokenType_Float,
    EtcTxtTokenType_Integer,
    EtcTxtTokenType_Identifier,
};

struct EtcTxtToken {
    EtcTxtTokenType type;

    const char *text;
    uint32 textLength;

    union {
        real32 floatValue;
        int64 int64Value;
        bool boolValue;
    };
};

enum EtcTxtParseErrorType {
    EtcTxtParseErrorType_None,

    EtcTxtParseErrorType_IllegalCharacter,
};

struct EtcTxtParseCursor {
    uint32 byteOffset;
    uint32 column;
    uint32 line;

    uint32 stackDepth;

    b32 hasIdentifier[ETCTXT_MAX_STACK_DEPTH];
    b32 isArray[ETCTXT_MAX_STACK_DEPTH];
};

struct EtcTxtParseError {
    EtcTxtParseErrorType type;
    char message[256];

    EtcTxtParseCursor cursor;
};

struct EtcTxtParseState {
    FileHandle file;

    const char *buffer;
    uint32 bufferSize;

    uint32 byteOffset;

    uint32 line;
    uint32 column;

    uint32 stackDepth;

    b32 hasIdentifier[ETCTXT_MAX_STACK_DEPTH];
    b32 isArray[ETCTXT_MAX_STACK_DEPTH];

    EtcTxtParseError error;

    DynamicArray<const char *> allocatedStrings;
};

enum EtcTxtWriteFlags {
    EtcTxtWriteFlag_Minify = 1 << 0,
};

typedef char *EtcTxtWriteCallback(char *buffer, uint32 *bufferSize, void *userData);

struct EtcTxtWriteState {
    FileHandle file;
    uint32 flags;

    char *currentBuffer;
    uint32 currentBufferSize;
    uint32 bytesWrittenToCurrentBuffer;
    uint32 totalBytesWritten;
    EtcTxtWriteCallback *writeCallback;
    void *callbackUserData;

    uint32 stackDepth;

    b32 justPushedGroup;
    b32 justPoppedGroup;

    b32 isArray[ETCTXT_MAX_STACK_DEPTH];
};

b32 OpenFileForRead(EtcTxtParseState *parser, const char *path);
void CloseFile(EtcTxtParseState *parser);

b32 OpenBufferForRead(EtcTxtParseState *parser, const char *buffer, uint32 bufferSize);

b32 OpenFileForWrite(EtcTxtWriteState *writer, const char *path, uint32 initialBytes = 0, uint32 flags = 0);
void CloseFile(EtcTxtWriteState *writer);

b32 OpenBufferForWrite(EtcTxtWriteState *writer, char *buffer, uint32 bufferSize, EtcTxtWriteCallback *callback = NULL, void *callbackUserData = NULL, uint32 flags = 0);


// Parsing
b32 ParseHasError(EtcTxtParseState *parser);

b32 ParseIdentifier(EtcTxtParseState *parser, const char **identifier = NULL, const char *matchingString = NULL);
b32 ParsePushObjectValue(EtcTxtParseState *parser);
b32 ParsePopObjectValue(EtcTxtParseState *parser);
b32 ParsePushArrayValue(EtcTxtParseState *parser);
b32 ParsePopArrayValue(EtcTxtParseState *parser);

b32 ParseStringValue(EtcTxtParseState *parser, const char **value = NULL);
b32 ParseStringValue(EtcTxtParseState *parser, char *value, uint32 maxLength);
b32 ParseFloatValue(EtcTxtParseState *parser, real32 *value = NULL);
b32 ParseBooleanValue(EtcTxtParseState *parser, bool *value = NULL);
b32 ParseBoolean32Value(EtcTxtParseState *parser, b32 *value = NULL);
b32 ParseInt64Value(EtcTxtParseState *parser, int64 *value = NULL);
b32 ParseInt8Value(EtcTxtParseState *parser, int8 *value = NULL);
b32 ParseInt16Value(EtcTxtParseState *parser, int16 *value = NULL);
b32 ParseInt32Value(EtcTxtParseState *parser, int32 *value = NULL);
b32 ParseUInt64Value(EtcTxtParseState *parser, uint64 *value = NULL);
b32 ParseUInt8Value(EtcTxtParseState *parser, uint8 *value = NULL);
b32 ParseUInt16Value(EtcTxtParseState *parser, uint16 *value = NULL);
b32 ParseUInt32Value(EtcTxtParseState *parser, uint32 *value = NULL);

#define ParseEnumValue(parser, enumType, enumValue) ParseEnumValue_(parser, ENUM_STRINGTABLE(enumType), ENUM_VALUECOUNT(enumType), (uint32 *)enumValue)

b32 ParseVec2Value(EtcTxtParseState *parser, vec2 *value = NULL);
b32 ParseVec3Value(EtcTxtParseState *parser, vec3 *value = NULL);
b32 ParseVec4Value(EtcTxtParseState *parser, vec4 *value = NULL);
b32 ParseVec2iValue(EtcTxtParseState *parser, vec2i *value = NULL);
b32 ParseVec3iValue(EtcTxtParseState *parser, vec3i *value = NULL);
b32 ParseVec4iValue(EtcTxtParseState *parser, vec4i *value = NULL);
b32 ParseMat3Value(EtcTxtParseState *parser, mat3 *value = NULL);
b32 ParseMat4Value(EtcTxtParseState *parser, mat4 *value = NULL);
b32 ParseQuaternionValue(EtcTxtParseState *parser, quaternion *value = NULL);

b32 ParseSkipValue(EtcTxtParseState *parser);
b32 ParseCountArrayValues(EtcTxtParseState *parser, uint32 *arrayValueCount);


// Writing
void WriteString(EtcTxtWriteState *writer, const char *string);
void WriteStringf(EtcTxtWriteState *writer, const char *fmt, ...);


void WritePushObjectValue(EtcTxtWriteState *writer, const char *fieldName = NULL);
void WritePopObjectValue(EtcTxtWriteState *writer);
void WritePushArrayValue(EtcTxtWriteState *writer, const char *fieldName = NULL);
void WritePopArrayValue(EtcTxtWriteState *writer);

void WriteInt8Value(EtcTxtWriteState *writer, int8 value, const char *fieldName = NULL);
void WriteInt16Value(EtcTxtWriteState *writer, int16 value, const char *fieldName = NULL);
void WriteInt32Value(EtcTxtWriteState *writer, int32 value, const char *fieldName = NULL);
void WriteInt64Value(EtcTxtWriteState *writer, int64 value, const char *fieldName = NULL);
void WriteUInt8Value(EtcTxtWriteState *writer, uint8 value, const char *fieldName = NULL);
void WriteUInt16Value(EtcTxtWriteState *writer, uint16 value, const char *fieldName = NULL);
void WriteUInt32Value(EtcTxtWriteState *writer, uint32 value, const char *fieldName = NULL);
void WriteUInt64Value(EtcTxtWriteState *writer, uint64 value, const char *fieldName = NULL);
void WriteFloatValue(EtcTxtWriteState *writer, real32 value, const char *fieldName = NULL);
void WriteBooleanValue(EtcTxtWriteState *writer, bool value, const char *fieldName = NULL);
void WriteBoolean32Value(EtcTxtWriteState *writer, b32 value, const char *fieldName = NULL);
void WriteStringValue(EtcTxtWriteState *writer, const char *value, uint32 valueLength, const char *fieldName = NULL);
void WriteStringValue(EtcTxtWriteState *writer, const char *value, const char *fieldName = NULL);
void WriteEnumValue(EtcTxtWriteState *writer, const char *enumStringValue, const char *fieldName = NULL);
void WriteVec2Value(EtcTxtWriteState *writer, vec2 value, const char *fieldName = NULL);
void WriteVec3Value(EtcTxtWriteState *writer, vec3 value, const char *fieldName = NULL);
void WriteVec4Value(EtcTxtWriteState *writer, vec4 value, const char *fieldName = NULL);
void WriteVec2iValue(EtcTxtWriteState *writer, vec2i value, const char *fieldName = NULL);
void WriteVec3iValue(EtcTxtWriteState *writer, vec3i value, const char *fieldName = NULL);
void WriteVec4iValue(EtcTxtWriteState *writer, vec4i value, const char *fieldName = NULL);
void WriteMat3Value(EtcTxtWriteState *writer, mat3 value, const char *fieldName = NULL);
void WriteMat4Value(EtcTxtWriteState *writer, mat4 value, const char *fieldName = NULL);
void WriteQuaternionValue(EtcTxtWriteState *writer, quaternion value, const char *fieldName = NULL);

#endif

