
struct ParsedGLSLSource {
    char *vertexSource;
    uint32 vertexSourceSize;

    char *geometrySource;
    uint32 geometrySourceSize;

    char *fragmentSource;
    uint32 fragmentSourceSize;

    char *computeSource;
    uint32 computeSourceSize;
};

#define GLSL_VERTEX_SHADER_DEF "VERTEX_SHADER"
#define GLSL_GEOMETRY_SHADER_DEF "GEOMETRY_SHADER"
#define GLSL_FRAGMENT_SHADER_DEF "FRAGMENT_SHADER"
#define GLSL_COMPUTE_SHADER_DEF "COMPUTE_SHADER"

#define GLSL_INCLUDE_PARSE_MAX_DEPTH 32

b32 ParseGLSLIncludes(const char *searchDirectory, const char *source, int32 sourceLength, char **expandedSource, uint32 *expandedSourceSize, uint32 stackDepth) {
    ASSERT(stackDepth < GLSL_INCLUDE_PARSE_MAX_DEPTH);

    b32 success = true;

    int32 sourceOffset = 0;

    while (success && sourceOffset < sourceLength) {
        int32 includeIndex = FindSubstring(source + sourceOffset, sourceLength - sourceOffset, "#include ");

        if (includeIndex >= 0) {
            // Copy all of source not yet copied into expandedSource
            if (includeIndex > sourceOffset) {
                uint32 charactersToCopy = includeIndex - sourceOffset;
                uint32 expandedSourceOffset = *expandedSourceSize;
                *expandedSourceSize = expandedSourceOffset + charactersToCopy;

                if (*expandedSource == NULL) {
                    *expandedSource = (char *)Plat.AllocateMem(*expandedSourceSize);
                }
                else {
                    *expandedSource = (char *)Plat.ResizeMem(*expandedSource, *expandedSourceSize);
                }
                Plat.CopyMem((*expandedSource) + expandedSourceOffset, source + sourceOffset, charactersToCopy);
                sourceOffset += charactersToCopy;
            }
            

            // Parse included file
            uint32 includeFileSearchStart = includeIndex + StringLength("#include ");
            uint32 charactersConsumedByIncludeFile;
            const char *relativeIncludePath = ScanQuotedString(source + includeFileSearchStart, sourceLength - includeFileSearchStart, &charactersConsumedByIncludeFile);

            const char *filename;
            const char *extension;

            if (relativeIncludePath != NULL &&
                SplitFileExtension(relativeIncludePath, &filename, &extension)) {

                const char *directory;
                const char *baseFilename;
                SplitPath(filename, &directory, &baseFilename);

                const char *fullDirectory = searchDirectory;
                if (directory != NULL) {
                    fullDirectory = PathConcat(searchDirectory, directory);
                }

                const char *fullIncludePath = Plat.FindFileInDirectory(fullDirectory, baseFilename, 1, &extension);

                FileHandle includeFile;
                if (Plat.OpenFileForRead(fullIncludePath, &includeFile)) {

                    success = ParseGLSLIncludes(searchDirectory, (const char *)includeFile.data, includeFile.fileSize, expandedSource, expandedSourceSize, stackDepth + 1);

                    Plat.CloseFile(&includeFile);

                    sourceOffset = includeFileSearchStart + charactersConsumedByIncludeFile;
                }
                else {
                    success = false;
                }

                Plat.DeallocateMem((char *)filename);
                Plat.DeallocateMem((char *)extension);
                Plat.DeallocateMem((char *)fullIncludePath);
                Plat.DeallocateMem((void *)relativeIncludePath);
            }
            else {
                success = false;
            }
        }
        else {
            // Copy rest of file to expandedSource
            uint32 charactersToCopy = sourceLength - sourceOffset;
            uint32 expandedSourceOffset = *expandedSourceSize;
            *expandedSourceSize = expandedSourceOffset + charactersToCopy;

            if (*expandedSource == NULL) {
                *expandedSource = (char *)Plat.AllocateMem(*expandedSourceSize);
            }
            else {
                *expandedSource = (char *)Plat.ResizeMem(*expandedSource, *expandedSourceSize);
            }
            Plat.CopyMem((*expandedSource) + expandedSourceOffset, source + sourceOffset, charactersToCopy);
            sourceOffset += charactersToCopy;
        }
    }

    return success;
}

b32 ParseGLSL(const char *searchDirectory, const char *sourcePath, ParsedGLSLSource *parsedSource) {
    ASSERT(parsedSource != NULL);

    b32 success = false;

    FileHandle sourceFile;
    if (Plat.OpenFileForRead(sourcePath, &sourceFile)) {

        Plat.ClearMem(parsedSource, sizeof(ParsedGLSLSource));

        const char *source = (const char *)sourceFile.data;
        uint32 sourceSize = sourceFile.fileSize;

        b32 containsVertexShader = FindSubstring(source, sourceSize, "\n#ifdef " GLSL_VERTEX_SHADER_DEF) >= 0;
        b32 containsGeometryShader = FindSubstring(source, sourceSize, "\n#ifdef " GLSL_GEOMETRY_SHADER_DEF) >= 0;
        b32 containsFragmentShader = FindSubstring(source, sourceSize, "\n#ifdef " GLSL_FRAGMENT_SHADER_DEF) >= 0;
        b32 containsComputeShader = FindSubstring(source, sourceSize, "\n#ifdef " GLSL_COMPUTE_SHADER_DEF) >= 0;


        char *expandedSource = NULL;
        uint32 expandedSourceSize = 0;
        if (ParseGLSLIncludes(searchDirectory, source, sourceSize, &expandedSource, &expandedSourceSize, 0)) {

            // Search for #version statement
            int32 startOfVersionStatement = FindSubstring(expandedSource, expandedSourceSize, "#version ");
            if (startOfVersionStatement < 0) {
                goto error;
            }

            b32 containsVersionStatement = true;
            int32 endOfVersionStatement = -1;

            for (int i = startOfVersionStatement; i < expandedSourceSize; i++) {
                if (expandedSource[i] == '\n') {
                    endOfVersionStatement = i + 1;
                    break;
                }
            }

            if (endOfVersionStatement < 0) {
                goto error;
            }


            if (containsVertexShader) {
                const char *shaderDef = "#define " GLSL_VERTEX_SHADER_DEF "\n";
                uint32 shaderDefLength = StringLength(shaderDef);

                parsedSource->vertexSourceSize = shaderDefLength + expandedSourceSize;
                parsedSource->vertexSource = ALLOCATE_ARRAY(char, parsedSource->vertexSourceSize);

                Plat.CopyMem(parsedSource->vertexSource, expandedSource, endOfVersionStatement);
                Plat.CopyMem(parsedSource->vertexSource + endOfVersionStatement, shaderDef, shaderDefLength);
                Plat.CopyMem(parsedSource->vertexSource + endOfVersionStatement + shaderDefLength, expandedSource + endOfVersionStatement, expandedSourceSize - endOfVersionStatement);
            }

            if (containsGeometryShader) {
                const char *shaderDef = "#define " GLSL_GEOMETRY_SHADER_DEF "\n";
                uint32 shaderDefLength = StringLength(shaderDef);

                parsedSource->geometrySourceSize = shaderDefLength + expandedSourceSize;
                parsedSource->geometrySource = ALLOCATE_ARRAY(char, parsedSource->geometrySourceSize);

                Plat.CopyMem(parsedSource->geometrySource, expandedSource, endOfVersionStatement);
                Plat.CopyMem(parsedSource->geometrySource + endOfVersionStatement, shaderDef, shaderDefLength);
                Plat.CopyMem(parsedSource->geometrySource + endOfVersionStatement + shaderDefLength, expandedSource + endOfVersionStatement, expandedSourceSize - endOfVersionStatement);
            }

            if (containsFragmentShader) {
                const char *shaderDef = "#define " GLSL_FRAGMENT_SHADER_DEF "\n";
                uint32 shaderDefLength = StringLength(shaderDef);

                parsedSource->fragmentSourceSize = shaderDefLength + expandedSourceSize;
                parsedSource->fragmentSource = ALLOCATE_ARRAY(char, parsedSource->fragmentSourceSize);

                Plat.CopyMem(parsedSource->fragmentSource, expandedSource, endOfVersionStatement);
                Plat.CopyMem(parsedSource->fragmentSource + endOfVersionStatement, shaderDef, shaderDefLength);
                Plat.CopyMem(parsedSource->fragmentSource + endOfVersionStatement + shaderDefLength, expandedSource + endOfVersionStatement, expandedSourceSize - endOfVersionStatement);
            }

            if (containsComputeShader) {
                const char *shaderDef = "#define " GLSL_COMPUTE_SHADER_DEF "\n";
                uint32 shaderDefLength = StringLength(shaderDef);

                parsedSource->computeSourceSize = shaderDefLength + expandedSourceSize;
                parsedSource->computeSource = ALLOCATE_ARRAY(char, parsedSource->computeSourceSize);

                Plat.CopyMem(parsedSource->computeSource, expandedSource, endOfVersionStatement);
                Plat.CopyMem(parsedSource->computeSource + endOfVersionStatement, shaderDef, shaderDefLength);
                Plat.CopyMem(parsedSource->computeSource + endOfVersionStatement + shaderDefLength, expandedSource + endOfVersionStatement, expandedSourceSize - endOfVersionStatement);
            }

            success = (containsVertexShader && containsFragmentShader) || containsComputeShader;
        }


    error:
        Plat.CloseFile(&sourceFile);
    }

    return success;
}

