
#include "io/utf8.h"

static const uint32 offsetsFromUTF8[6] = {
    0x00000000UL, 0x00003080UL, 0x000E2080UL,
    0x03C82080UL, 0xFA082080UL, 0x82082080UL
};

static const uint8 trailingBytesForUTF8[256] = {
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1, 1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
    2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2, 3,3,3,3,3,3,3,3,4,4,4,4,5,5,5,5
};

b32 utf8_boundary(uint8 src) {
    // Trailing bytes are of the form 0b10xxxxxx, so just check that
    // the first two bits of *src aren't 0b10.

    // NOTE: a null-terminal counts as a utf8_boundary!

    b32 result = (src & 0xC0) != 0x80;
    return result;
}

uint32 utf8_bytelength(uint32 codepoint) {
    uint32 result = 0;

    if (codepoint < 0x0080) {
        result = 1;
    }
    else if (codepoint < 0x0800) {
        result = 2;
    }
    else if (codepoint < 0x10000) {
        result = 3;
    }
    else if (codepoint < 0x200000) {
        result = 4;
    }
    else if (codepoint < 0x4000000) {
        result = 5;
    }
    else if (codepoint <= 0x7FFFFFFF) {
        result = 6;
    }

    return result;
}

uint32 utf8_next(uint8 *src, uint32 *byteOffset) {
    ASSERT(src != NULL);
    ASSERT(byteOffset != NULL);

    uint8 bytesConsumed = 0;
    uint8 codepoint = 0;
    do {
        codepoint <<= 6;
        codepoint += src[(*byteOffset)++];
        bytesConsumed++;
    } while (!utf8_boundary(src[*byteOffset]));

    codepoint -= offsetsFromUTF8[bytesConsumed - 1];
    ASSERT(bytesConsumed <= 4);

    return codepoint;
}

// - srcSize is size of src in bytes
// - destCount is number of wide characters in dest
// - Stops at null-terminals in src, but does not guarantee that dest is null-terminated!
// - Returns number of characters converted
uint32 utf8_toucs(uint8 *src, uint32 srcSize, uint32 *dest, uint32 destCount) {
    ASSERT(src != NULL);
    ASSERT(dest != NULL);

    uint8 *srcEnd = (uint8 *)src + srcSize;
    uint32 charsProcessed = 0;

    while (charsProcessed < destCount) {
        uint8 trailingBytes = trailingBytesForUTF8[(uint8)*src];

        // No 5- or 6-byte characters allowed
        ASSERT(trailingBytes <= 3);

        if (src + trailingBytes >= srcEnd) {
            break;
        }

        if (*src == 0) {
            dest[charsProcessed] = 0;
            break;
        }

        uint32 codepoint = 0;
        switch (trailingBytes) {
            // These fall through deliberately
            case 3: codepoint += (uint8)*src++; codepoint <<= 6;
            case 2: codepoint += (uint8)*src++; codepoint <<= 6;
            case 1: codepoint += (uint8)*src++; codepoint <<= 6;
            case 0: codepoint += (uint8)*src++;
        }
        codepoint -= offsetsFromUTF8[trailingBytes];

        dest[charsProcessed++] = codepoint;
    }

    return charsProcessed;
}


// - srcCount is number of wide characers in src
// - destSize is size of dest in bytes
// - Stops at null-terminals in src, but does not guarantee that dest is
// null-terminated! But dest will be null-terminated if there is enough
// space.
// - Returns number of bytes converted
uint32 ucs_toutf8(uint32 *src, uint32 srcCount, uint8 *dest, uint32 destSize) {
    ASSERT(src != NULL);
    ASSERT(dest != NULL);

    uint32 bytesConverted = 0;

    for (int i = 0; i < srcCount; i++) {
        uint32 codepoint = src[i];
        uint32 numBytesToWrite = utf8_bytelength(codepoint);

        if (numBytesToWrite == 0) {
            break;
        }

        // No 5- or 6-byte characters allowed
        ASSERT(numBytesToWrite <= 4);

        if (codepoint == 0) {
            break;
        }

        if ((bytesConverted + numBytesToWrite) > destSize) {
            break;
        }

        if (numBytesToWrite == 1) {
            dest[i] = (uint8)codepoint;
        }
        else {
            // Start at the last byte to make things easier
            for (int b = numBytesToWrite; b > 1; b--) {
                uint32 mask = (1 << 6) - 1;
                dest[bytesConverted + b - 1] = 0x80 | (codepoint & mask);

                codepoint = codepoint >> 6;
            }

            // The last byte is trickier
            uint32 mask = (1 << (8 - numBytesToWrite - 1)) - 1;
            uint8 leadingByte = 255 << (8 - numBytesToWrite);
            dest[bytesConverted] = leadingByte + (codepoint & mask);
        }

        bytesConverted += numBytesToWrite;
        if (bytesConverted == destSize) {
            break;
        }
    }

    // Null-terminate if there's enough space
    if (destSize > bytesConverted) {
        dest[bytesConverted] = 0;
    }
    return bytesConverted;
}

uint32 ucs16_toutf8(uint16 *src, uint32 srcCount, uint8 *dest, uint32 destSize) {
    ASSERT(src != NULL);
    ASSERT(dest != NULL);

    uint32 bytesConverted = 0;

    for (int i = 0; i < srcCount; i++) {
        uint16 codepoint = src[i];
        uint32 numBytesToWrite = utf8_bytelength(codepoint);

        if (numBytesToWrite == 0) {
            break;
        }

        // No 5- or 6-byte characters allowed
        ASSERT(numBytesToWrite <= 4);

        if (codepoint == 0) {
            break;
        }

        if ((bytesConverted + numBytesToWrite) > destSize) {
            break;
        }

        if (numBytesToWrite == 1) {
            dest[i] = (uint8)codepoint;
        }
        else {
            // Start at the last byte to make things easier
            for (int b = numBytesToWrite; b > 1; b--) {
                uint32 mask = (1 << 6) - 1;
                dest[bytesConverted + b - 1] = 0x80 | (codepoint & mask);

                codepoint = codepoint >> 6;
            }

            // The last byte is trickier
            uint32 mask = (1 << (8 - numBytesToWrite - 1)) - 1;
            uint8 leadingByte = 255 << (8 - numBytesToWrite);
            dest[bytesConverted] = leadingByte + (codepoint & mask);
        }

        bytesConverted += numBytesToWrite;
        if (bytesConverted == destSize) {
            break;
        }
    }

    // Null-terminate if there's enough space
    if (destSize > bytesConverted) {
        dest[bytesConverted] = 0;
    }
    return bytesConverted;
}

