
#ifndef ETC_LOCKFREEQUEUE_H
#define ETC_LOCKFREEQUEUE_H

// References:
// [0] http://www.codeproject.com/Articles/153898/Yet-another-implementation-of-a-lock-free-circul

template <typename T>
struct LockFreeQueue {
    uint32 capacity;
    T *ringBuffer;

    volatile uint32 readIndex;
    volatile uint32 maxReadIndex;
    volatile uint32 writeIndex;
};


template <typename T>
b32 IsEmpty(LockFreeQueue<T> *queue) {
    uint32 currentReadIndex = queue->readIndex;
    uint32 currentWriteIndex = queue->writeIndex;

    b32 isEmpty = currentReadIndex == currentWriteIndex;
    return isEmpty;
}

// NOTE: careful with Peek()! It's not really safe if multiple threads
// are consuming from the queue
template <typename T>
b32 Peek(LockFreeQueue<T> *queue, T *data) {
    ASSERT(data != NULL);

    uint32 currentReadIndex = queue->readIndex;
    uint32 currentWriteIndex = queue->writeIndex;

    b32 isEmpty = currentReadIndex == currentWriteIndex;

    if (!isEmpty) {
        *data = queue->ringBuffer[currentReadIndex];
    }

    return !isEmpty;
}

// NOTE: Both PushBack() and PopFront() are not guaranteed to succeed!
// They will fail early and return false if there's not enough space
// in the queue or there are no elements left to pop.

template <typename T>
b32 PushBack(LockFreeQueue<T> *queue, T element) {

    uint32 currentReadIndex;
    uint32 currentWriteIndex;
    while (true) {
        currentReadIndex = queue->readIndex;
        currentWriteIndex = queue->writeIndex;

        // Make sure the push won't put queue in inconsistent state
        if ((currentWriteIndex + 1) % queue->capacity == currentReadIndex) {
            return false;
        }

        if (currentWriteIndex == EtcCompareAndSwap(&queue->writeIndex, currentWriteIndex, (currentWriteIndex + 1) % queue->capacity)) {
            break;
        }
    }

    queue->ringBuffer[currentWriteIndex] = element;

    while (true) {
        if (currentWriteIndex == EtcCompareAndSwap(&queue->maxReadIndex, currentWriteIndex, (currentWriteIndex + 1) % queue->capacity)) {
            break;
        }

        EtcThreadYield();
    }

    return true;
}

template <typename T>
void PushBackForce(LockFreeQueue<T> *queue, T element) {
    while (true) {
        if (PushBack(queue, element)) {
            break;
        }
    }
}

template <typename T>
b32 PopFront(LockFreeQueue<T> *queue, T *data) {
    ASSERT(data != NULL);
    ASSERT(queue != NULL);

    b32 result = false;

    uint32 currentReadIndex;
    uint32 currentMaxReadIndex;

    uint32 count = 0;
    while (true) {
        currentReadIndex = queue->readIndex;
        currentMaxReadIndex = queue->maxReadIndex;

        if (currentReadIndex == currentMaxReadIndex) {
            return false;
        }

        T element = queue->ringBuffer[currentReadIndex];

        if (currentReadIndex == EtcCompareAndSwap(&queue->readIndex, currentReadIndex, (currentReadIndex + 1) % queue->capacity)) {
            *data = element;
            break;
        }
    }

    return true;
}

template <typename T>
void PopFrontForce(LockFreeQueue<T> *queue, T *element) {
    ASSERT(element != NULL);

    while (true) {
        if (PopFront(queue, element)) {
            break;
        }
    }
}

#endif

