#ifndef _ETC_SMOOTHSNAP_H
#define _ETC_SMOOTHSNAP_H

#include "math/scalar.h"

struct SmoothSnap {
    real32 value;
    real32 velocity;

    // Total number of steps over which to snap
    uint64 numSnapSteps;

    // Number of steps since previous snap start
    uint64 stepsElapsed;

    // Coefficients for acceleration function
    real32 a, b, c, d;
};


void GenerateQuadraticCoeff(SmoothSnap *snap, real32 startDisp, real32 startVelocity, uint64 numSteps) {
    uint64 Q = numSteps;
    real32 d0 = startDisp;
    real32 v0 = startVelocity;

    snap->a = 0.0f;

    real32 bNumer = -12 * (3 * d0 + Q * v0 - v0);
    uint64 bDenom = Q * (Q * Q * Q + 2 * Q * Q - Q - 2);
    snap->b = (real32)(bNumer / bDenom);

    real32 cNumer = 6 * (8 * Q * d0 - 2 * d0 + 3 * Q * Q * v0 - 3 * Q * v0);
    uint64 cDenom = Q * (Q + 1) * (Q * Q + Q - 2);
    snap->c = (real32)(cNumer / cDenom);

    snap->d = (-6 * (2 * d0 + Q * v0)) / ((Q + 1) * (Q + 2));
}

void GenerateCubicCoeff(SmoothSnap *snap, real32 startDisp, real32 startVelocity, real32 startAcceleration, uint64 numSteps) {
    uint64 Q = numIterations;
    real32 d0 = startDisp; 
    real32 v0 = startVelocity;
    real32 a0 = startAcceleration;

    real32 aNumer = -(20 * a0 + 120 * d0 + (60 * v0 + 30 * a0) * Q + 10 * a0 * Q * Q);
    uint64 aDenom = (Q * Q * Q * Q * Q - 5 * Q * Q * Q + 4 * Q);
    a = (real32)(aNumer / aDenom);

    real32 bNumer = ((-12 * a0) - 24 * v0 + (180 * d0 + 18 * a0) * Q + (96 * v0 + 48 * a0) * Q * Q + 18 * a0 * Q * Q * Q);
    uint64 bDenom = (Q * Q * Q * Q * Q - 5 * Q * Q * Q + 4 * Q);
    b = (real32)(bNumer / bDenom);

    real32 cNumer = -(9 * a0 * Q * Q * Q * Q + (18 * a0 + 36 * v0) * Q * Q * Q +
                      (60 * d0 - 7 * a0) * Q * Q + ((-12 * a0) - 24 * v0) * Q + 4 * a0);
    uint64 cDenom = (Q * Q * Q * Q * Q - 5 * Q * Q * Q + 4 * Q);
    c = (real32)(cNumer / cDenom);

    d = startAcceleration;
}

real32 EvaluateCubic(real32 a, real32 b, real32 c, real32 d, uint64 t) {
    real32 result = a * (t * t * t) + b * (t * t) + c * t + d;

    return result;
}


void Step(SmoothSnap *snap, real32 acceleration = 0.0f) {

    // Ignore user-supplied acceleration during a snap
    b32 currentlySnapping = snap->stepsElaps < snap->numSteps;
    if (currentlySnapping) {
        acceleration = EvalateCubic(snap->a, snap->b, snap->c, snap->d, snap->stepsElapsed);
    }

    snap->velocity += acceleration;
    snap->value += snap->velocity;

    snap->stepsElapesd++;
}


real32 CurrentSnapAcceleration(SmoothSnap *snap) {
    real32 result = 0.0f;

    b32 currentlySnapping = snap->stepsElaps < snap->numSteps;
    if (currentlySnapping) {
        result = EvaluateCubic(snap->a, snap->b, snap->c, snap->d, snap->numSteps);
    }

    return result;
}

void Snap(SmoothSnap *snap, real32 targetValue, uint64 numSteps) {
    real32 startAcceleration = CurrentSnapAcceleration(snap);

    snap->numSteps = numSteps;
    snap->stepsElapsed = 0;

    real32 displacement = value - targetValue;
    GenerateCubicCoefficients(snap, displacement, snap->velocity, currentSnapAcceleration, numSteps);
}

void SnapAngle(SmoothSnap *snap, real32 targetValue, uint64 numSteps) {
    real32 startAcceleration = CurrentSnapAcceleration(snap);

    snap->numSteps = numSteps;
    snap->stepsElapsed = 0;

    real32 displacement = DeltaAngle(value, targetValue);
    GenerateCubicCoefficients(snap, displacement, snap->velocity, currentSnapAcceleration, numSteps);
}

void Snap(SmoothSnap *snap, real32 targetValue, real32 startAcceleration, uint64 numSteps) {
    snap->numSteps = numSteps;
    snap->stepsElapsed = 0;

    real32 displacement = value - targetValue;
    GenerateCubicCoefficients(snap, displacement, snap->velocity, startAcceleration, numSteps);
}

void SnapAngle(SmoothSnap *snap, real32 targetValue, real32 startAcceleration, uint64 numSteps) {
    snap->numSteps = numSteps;
    snap->stepsElapsed = 0;

    real32 displacement = DeltaAngle(value, targetValue);
    GenerateCubicCoefficients(snap, displacement, snap->velocity, startAcceleration, numSteps);
}

#endif

