
void AllocateMesh(Mesh *mesh, uint32 vertexCount, uint32 vertexBufferSize, uint32 indexCount, uint32 submeshCount, b32 allocateMaterials) {
    if (vertexCount > 0 && vertexBufferSize > 0) {
        mesh->vertexCount = vertexCount;
        mesh->vertexBufferSize = vertexBufferSize;
        mesh->vertices = Plat.AllocateMem(vertexBufferSize);
    }
    else {
        mesh->vertexCount = 0;
        mesh->vertexBufferSize = 0;
        mesh->vertices = NULL;
    }

    mesh->positions = NULL;
    mesh->texcoords = NULL;
    mesh->normals = NULL;
    mesh->colors = NULL;


    mesh->indexCount = indexCount;
    if (indexCount > 0) {
        mesh->indexBufferSize = sizeof(uint32) * indexCount;
        mesh->indices = (uint32 *)Plat.AllocateMem(mesh->indexBufferSize);
    }
    else {
        mesh->indexBufferSize = 0;
        mesh->indices = NULL;
    }

    mesh->submeshCapacity = submeshCount;
    mesh->submeshCount = submeshCount;
    if (submeshCount > 0) {
        mesh->submeshLengths = (uint32 *)Plat.AllocateMem(sizeof(uint32) * mesh->submeshCapacity);
        Plat.ClearMem(mesh->submeshLengths, sizeof(uint32) * mesh->submeshCapacity);

        // Clear materials because it's reasonable to not initialize them explicitly
        if (allocateMaterials) {
            mesh->materials = (Material *)Plat.AllocateMem(sizeof(Material) * mesh->submeshCapacity);
            Plat.ClearMem(mesh->materials, sizeof(Material) * mesh->submeshCapacity);
        }
        else {
            mesh->materials = NULL;
        }
    }
    else {
        mesh->submeshLengths = NULL;
        mesh->materials = NULL;
    }

    mesh->vboID = 0;
    mesh->iboID = 0;
}

void ReallocateMesh(Mesh *mesh, uint32 vertexCount, uint32 vertexBufferSize, uint32 indexCount, uint32 submeshCount, b32 allocateMaterials) {
    TIMED_BLOCK();

    // If the mesh is uninitialized, just allocate it as normal
    if (mesh->vertices == NULL) {
        AllocateMesh(mesh, vertexCount, vertexBufferSize, indexCount, submeshCount, allocateMaterials);
        return;
    }

    // Otherwise, resize any buffers that need it
    mesh->vertexCount = vertexCount;
    if (mesh->vertexBufferSize < vertexBufferSize) {
        mesh->vertexBufferSize = vertexBufferSize;
        mesh->vertices = Plat.ResizeMem(mesh->vertices, vertexBufferSize);
    }

    uint32 newIndexBufferSize = indexCount * sizeof(uint32);
    mesh->indexCount = indexCount;
    if (mesh->indexBufferSize < newIndexBufferSize) {
        mesh->indexBufferSize = newIndexBufferSize;

        if (mesh->indices == NULL) {
            mesh->indices = (uint32 *)Plat.AllocateMem(newIndexBufferSize);
        }
        else {
            mesh->indices = (uint32 *)Plat.ResizeMem(mesh->indices, newIndexBufferSize);
        }
    }

    
    mesh->submeshCount = submeshCount;
    if (mesh->submeshCapacity < submeshCount) {
        uint32 oldSubmeshCapacity = mesh->submeshCapacity;

        mesh->submeshCapacity = submeshCount;
        uint32 submeshCapacityDiff = mesh->submeshCapacity - oldSubmeshCapacity;

        uint32 submeshLengthsBufferSize = sizeof(uint32) * mesh->submeshCapacity;
        if (mesh->submeshLengths == NULL) {
            mesh->submeshLengths = (uint32 *)Plat.AllocateMem(submeshLengthsBufferSize);
            Plat.ClearMem(mesh->submeshLengths, submeshLengthsBufferSize);
        }
        else {
            mesh->submeshLengths = (uint32 *)Plat.ResizeMem(mesh->submeshLengths, submeshLengthsBufferSize);
            Plat.ClearMem(&mesh->submeshLengths[oldSubmeshCapacity], sizeof(uint32) * submeshCapacityDiff);
        }

        if (mesh->materials != NULL) {
            uint32 materialsBufferSize = sizeof(Material) * mesh->submeshCapacity;
            if (mesh->materials == NULL) {
                mesh->materials = (Material *)Plat.AllocateMem(materialsBufferSize);
            }
            else {
                mesh->materials = (Material *)Plat.ResizeMem(mesh->materials, materialsBufferSize);
            }

            Plat.ClearMem(mesh->materials, materialsBufferSize);
        }
    }
    
    mesh->positions = NULL;
    mesh->texcoords = NULL;
    mesh->normals = NULL;
    mesh->colors = NULL;
}

void DeallocateMesh(Mesh *mesh) {
    Plat.DeallocateMem(mesh->vertices);
    Plat.DeallocateMem(mesh->indices);

    if (mesh->submeshLengths != NULL) {
        Plat.DeallocateMem(mesh->submeshLengths);
    }

    if (mesh->materials != NULL) {
        Plat.DeallocateMem(mesh->materials);
    }
}


// Mesh Loading
b32 MeshLoadMainThread(Mesh *mesh) {
    CreateMeshOpenGL(mesh);
    b32 success = mesh->vboID != 0 && mesh->iboID != 0;
    return success;
}

b32 MeshLoadFile(AssetDB *assets, const char *path, Mesh *mesh) {
    FileHandle file;
    if (Plat.OpenFileForRead(path, &file)) {

        uint32 version = ReadUInt32(&file);
        ASSERT(version == 0);

        uint32 primitiveTypeNameLength = ReadUInt32(&file);
        char *primitiveTypeName = (char *)file.data + file.offset;
        FileSeek(&file, file.offset + primitiveTypeNameLength);

        PrimitiveType primitiveType;
        if (ENUM_FINDVALUE(PrimitiveType, primitiveTypeName, &primitiveType)) {
            mesh->primitiveType = primitiveType;
        }
        else {
            mesh->primitiveType = PrimitiveType_Triangles;
        }

            
        uint32 attribCount = ReadUInt32(&file);
        ASSERT(attribCount > 0);

        // @GACK: main thread should probably have TLS data!
        TaskSchedulerTLSData *tlsData = (TaskSchedulerTLSData *)Plat.GetTLS(Plat.taskSchedulerTLSKey);
        uint32 *attribNameLengths = PushArray(&tlsData->arena, uint32, attribCount);
        char **attribNames = PushArray(&tlsData->arena, char *, attribCount);

        for (int i = 0; i < attribCount; i++) {
            uint32 nameLength = ReadUInt32(&file);
            attribNameLengths[i] = nameLength;
            attribNames[i] = (char *)file.data + file.offset;

            // @BUG: attribNames are sometimes not found, or a garbage pointer
            // Plat.DebugPrint("%s %d %d %x %x", path, nameLength, file.offset, file.data, (char *)file.data + file.offset);
            // Plat.DebugPrint("%.*s", nameLength, attribNames[i]);

            FileSeek(&file, file.offset + nameLength);
        }


        u32 vertCount = ReadUInt32(&file);
        mesh->vertexCount = vertCount;

        uint32 vertexBufferSize = ReadUInt32(&file);
        mesh->vertexBufferSize = vertexBufferSize;

        u32 indexCount = ReadUInt32(&file);
        mesh->indexCount = indexCount;

        u32 submeshCount = ReadUInt32(&file);
        mesh->submeshCount = submeshCount;

        ReallocateMesh(mesh, vertCount, vertexBufferSize, indexCount, submeshCount, true);

        uint32 bytesRead = ReadBytes(&file, vertexBufferSize, mesh->vertices);
        ASSERT(bytesRead == vertexBufferSize);

        u32 offset = 0;

        for (int i = 0; i < attribCount; i++) {
            VertexAttrib attrib;
            if (FindVertexAttrib(attribNames[i], attribNameLengths[i], &attrib)) {
                switch (attrib) {
                    case Position_Attrib: {
                        ASSERT(mesh->positions == NULL);
                        mesh->positions = (vec3 *)BUFFER_OFFSET(mesh->vertices, offset);
                        offset += sizeof(vec3) * vertCount;    
                    } break;
                        
                    case TexCoord_Attrib: {
                        ASSERT(mesh->texcoords == NULL);
                        mesh->texcoords = (vec2 *)BUFFER_OFFSET(mesh->vertices, offset);
                        offset += sizeof(vec2) * vertCount;    
                    } break;

                    case Normal_Attrib: {
                        ASSERT(mesh->normals == NULL);
                        mesh->normals = (vec3 *)BUFFER_OFFSET(mesh->vertices, offset);
                        offset += sizeof(vec3) * vertCount;    
                    } break;

                    case Color_Attrib: {
                        ASSERT(mesh->colors == NULL);
                        mesh->colors = (vec4 *)BUFFER_OFFSET(mesh->vertices, offset);
                        offset += sizeof(vec4) * vertCount;    
                    } break;

                    case BoneIndices_Attrib: {
                        ASSERT(mesh->boneIndices == NULL);
                        mesh->boneIndices = (vec4i *)BUFFER_OFFSET(mesh->vertices, offset);
                        offset += sizeof(vec4i) * vertCount;    
                    } break;

                    case BoneWeights_Attrib: {
                        ASSERT(mesh->boneWeights == NULL);
                        mesh->boneWeights = (vec4 *)BUFFER_OFFSET(mesh->vertices, offset);
                        offset += sizeof(vec4) * vertCount;    
                    } break;

                    default: {
                        ASSERT(false);
                    } break;
                }
            }
        }


        uint32 indexBufferSize = sizeof(uint32) * indexCount;
        bytesRead = ReadBytes(&file, sizeof(uint32) * indexCount, mesh->indices);
        ASSERT(bytesRead == indexBufferSize);

        if (submeshCount > 0) {
            uint32 submeshLengthsBufferSize = sizeof(uint32) * submeshCount;
            bytesRead = ReadBytes(&file, sizeof(uint32) * submeshCount, mesh->submeshLengths);
            ASSERT(bytesRead == submeshLengthsBufferSize);
        }


        for (int i = 0; i < submeshCount; i++) {
            Material mat = {};
            mat.diffuseColor.r = ReadReal32(&file);
            mat.diffuseColor.g = ReadReal32(&file);
            mat.diffuseColor.b = ReadReal32(&file);
            mat.diffuseColor.a = ReadReal32(&file);

            mesh->materials[i] = mat;
        }

        Plat.CloseFile(&file);

        return true;
    }

    return false;
}

void MeshUnloadFile(AssetDB *assets, Mesh *mesh) {
    DeleteMeshOpenGL(mesh);
    DeallocateMesh(mesh);
}

void WriteMeshAsOBJ(const char *path, Mesh *mesh) {

    const char *name;
    const char *directory;
    SplitPath(path, &directory, &name);

    const char *nameSansExt;
    SplitFileExtension(name, &nameSansExt, NULL);

    const char *materialFileName = AppendFileExtension(nameSansExt, "mtl");
    const char *materialPath = PathConcat(directory, materialFileName);

    EtcTxtWriteState writer;
    if (OpenFileForWrite(&writer, materialPath, Megabytes(1))) {
        for (int i = 0; i < mesh->submeshCount; i++) {
            WriteStringf(&writer, "newmtl color_%d \n", i);
            WriteString(&writer, "Ka 1.0 1.0 1.0\n");
            WriteStringf(&writer, "Kd %f %f %f\n", mesh->materials[i].diffuseColor.r, mesh->materials[i].diffuseColor.g, mesh->materials[i].diffuseColor.b);
            WriteString(&writer, "Ks 0.0 0.0 0.0\n");
            //Writestring(&writer, "    Ns 0.0\n");
        }

        CloseFile(&writer);
    }
    
    if (OpenFileForWrite(&writer, path, Megabytes(4))) {
        WriteStringf(&writer, "mtlib %s", materialFileName);
        
        WriteString(&writer, "#verts \n");
        for (int i = 0; i < mesh->vertexCount; i++) {
            vec3 v = mesh->positions[i];
            WriteStringf(&writer, "v %f %f %f 1.0\n", v.x, v.y, v.z);
        }

        WriteString(&writer, "\n");
        WriteString(&writer, "normals \n");
        for (int i = 0; i < mesh->vertexCount; i++) {
            vec3 n = mesh->normals[i];
            WriteStringf(&writer, "vn %f %f %f\n", n.x, n.y, n.z);
        }

        WriteString(&writer, "\n");
        WriteString(&writer, "faces \n");
        uint32 submeshOffset = 0;
        for (int i = 0; i < mesh->submeshCount; i++) {
            u32 submeshLength = mesh->submeshLengths[i];
            for (int j = 0; j < submeshLength; j += 3) {
                WriteStringf(&writer, "usemtl color_%d\n", i);
                WriteStringf(&writer, "f %d %d %d\n", mesh->indices[submeshOffset + j + 0] + 1, mesh->indices[submeshOffset + j + 1] + 1, mesh->indices[submeshOffset + j + 2] + 1);
            }

            submeshOffset += submeshLength;
        }

        CloseFile(&writer);
    }
}


// Primitive Generation
void AllocatePlane(Mesh *mesh, uint32 resolution, mat4 transform) {
    ASSERT(resolution > 1);

    uint32 vertexCount = Square(resolution);
    uint32 indexCount = Square(resolution - 1) * 2 * 3;
    uint32 vertexBufferSize = (sizeof(vec3) + sizeof(vec3) + sizeof(vec2)) * vertexCount;


    ReallocateMesh(mesh, vertexCount, vertexBufferSize, indexCount, 1);
    mesh->primitiveType = PrimitiveType_Triangles;

    mesh->positions = (vec3 *)mesh->vertices;
    mesh->texcoords = (vec2 *)((uint8 *)mesh->vertices + sizeof(vec3) * vertexCount);
    mesh->normals = (vec3 *)((uint8 *)mesh->vertices + (sizeof(vec3) + sizeof(vec2)) * vertexCount);

    real32 stepSize = 1.0f / (real32)(resolution - 1);
    vec3 normal = MultiplyDirection(transform, V3(0, 1, 0));

    for (int y = 0; y < resolution; y++) {
        for (int x = 0; x < resolution; x++) {
            uint32 index = y * resolution + x;

            mesh->positions[index] = MultiplyPoint(transform, V3(x * stepSize, 0.0f, y * stepSize));
            mesh->texcoords[index] = V2(x * stepSize, y * stepSize);
            mesh->normals[index] = normal;
        }
    }

    // Ordered counter-clockwise from top
    for (int y = 0; y < resolution - 1; y++) {
        for (int x = 0; x < resolution - 1; x++) {
            uint32 lowerLeftVertex = y * resolution + x;
            uint32 lowerRightVertex = y * resolution + x + 1;
            uint32 upperLeftVertex = lowerLeftVertex + resolution;
            uint32 upperRightVertex = lowerRightVertex + resolution;

            uint32 lowerRightIndexOffset = (y * (resolution - 1) + x) * 2 * 3;
            mesh->indices[lowerRightIndexOffset + 0] = lowerLeftVertex;
            mesh->indices[lowerRightIndexOffset + 1] = lowerRightVertex;
            mesh->indices[lowerRightIndexOffset + 2] = upperRightVertex;

            uint32 upperLeftIndexOffset = lowerRightIndexOffset + 3;
            mesh->indices[upperLeftIndexOffset + 0] = lowerLeftVertex;
            mesh->indices[upperLeftIndexOffset + 1] = upperRightVertex;
            mesh->indices[upperLeftIndexOffset + 2] = upperLeftVertex;
        }
    }

    mesh->submeshLengths[0] = mesh->indexCount;
}

void AllocatePlane(Mesh *mesh, uint32 resolution) {
    AllocatePlane(mesh, resolution, Identity4());
}


void AllocateInfinitePlane(Mesh *mesh) {
    uint32 vertexCount = 5;
    uint32 vertexBufferSize = 5 * sizeof(vec4);
    uint32 indexCount = 12;

    ReallocateMesh(mesh, vertexCount, vertexBufferSize, indexCount, 1);
    mesh->primitiveType = PrimitiveType_Triangles;

    vec4 *vertices = (vec4 *)mesh->vertices;
    vertices[0] = V4(0, 0, 0, 1);
    vertices[1] = V4(1, 0, 0, 0);
    vertices[2] = V4(0, 0, 1, 0);
    vertices[3] = V4(-1, 0, 0, 0);
    vertices[4] = V4(0, 0, -1, 0);

    uint32 idx = 0;
    mesh->indices[idx++] = 0;
    mesh->indices[idx++] = 1;
    mesh->indices[idx++] = 2;

    mesh->indices[idx++] = 0;
    mesh->indices[idx++] = 2;
    mesh->indices[idx++] = 3;

    mesh->indices[idx++] = 0;
    mesh->indices[idx++] = 3;
    mesh->indices[idx++] = 4;

    mesh->indices[idx++] = 0;
    mesh->indices[idx++] = 4;
    mesh->indices[idx++] = 1;

    mesh->submeshLengths[0] = 12;
}

void AllocateCircle(Mesh *mesh, real32 radius, uint32 segments) {

    uint32 vertexBufferSize = sizeof(vec3) * (segments + 1);
    ReallocateMesh(mesh, segments + 1, vertexBufferSize, segments + 1, 1);
    mesh->primitiveType = PrimitiveType_TriangleFan;

    mesh->positions = (vec3 *)mesh->vertices;

    for (int i = 0; i <= segments; i++) {
        real32 theta = 2 * _PI * i / (real32)segments;
        vec3 position = V3(radius * cosf(theta), 0.0f, radius * sinf(theta));

        mesh->positions[i] = position;
        mesh->indices[i] = i;
    }

    mesh->vertexCount = segments + 1;
    mesh->indexCount = segments + 1;
    mesh->submeshLengths[0] = mesh->indexCount;
}

struct MeshEdge {
    uint32 index0;
    uint32 index1;
};

void AllocateSphere(Mesh *mesh, uint32 subdivisions) {

    // Generate Initial Icosahedron
    DynamicArray<vec3> positions = MakeDynamicArray<vec3>(20);

    DynamicArray<uint32> indices = MakeDynamicArray<uint32>(60);;

    real32 goldenRatio = (1.0f + sqrtf(5.0f)) / 2.0f;

    PushBack(&positions, Normalize(V3(-1.0f, -goldenRatio, 0.0f)));
    PushBack(&positions, Normalize(V3(1.0f, -goldenRatio, 0.0f)));
    PushBack(&positions, Normalize(V3(1.0f, goldenRatio, 0.0f)));
    PushBack(&positions, Normalize(V3(-1.0f, goldenRatio, 0.0f)));

    PushBack(&positions, Normalize(V3(0.0f, -1.0f, goldenRatio)));
    PushBack(&positions, Normalize(V3(0.0f, -1.0f, -goldenRatio)));
    PushBack(&positions, Normalize(V3(0.0f, 1.0f, -goldenRatio)));
    PushBack(&positions, Normalize(V3(0.0f, 1.0f, goldenRatio)));

    PushBack(&positions, Normalize(V3(-goldenRatio, 0.0f, -1.0f)));
    PushBack(&positions, Normalize(V3(goldenRatio, 0.0f, -1.0f)));
    PushBack(&positions, Normalize(V3(goldenRatio, 0.0f, 1.0f)));
    PushBack(&positions, Normalize(V3(-goldenRatio, 0.0f, 1.0f)));


    // These are counterclockwise from the outside
    // Faces around P0
    PushBack(&indices, 0u); PushBack(&indices, 1u); PushBack(&indices, 5u);

    PushBack(&indices, 0u); PushBack(&indices, 5u); PushBack(&indices, 8u);

    PushBack(&indices, 0u); PushBack(&indices, 8u); PushBack(&indices, 11u);

    PushBack(&indices, 0u); PushBack(&indices, 11u); PushBack(&indices, 4u);

    PushBack(&indices, 0u); PushBack(&indices, 4u); PushBack(&indices, 1u);

    // Adjacent faces
    PushBack(&indices, 1u); PushBack(&indices, 5u); PushBack(&indices, 9u);

    PushBack(&indices, 5u); PushBack(&indices, 8u); PushBack(&indices, 6u);

    PushBack(&indices, 8u); PushBack(&indices, 11u); PushBack(&indices, 3u);

    PushBack(&indices, 11u); PushBack(&indices, 4u); PushBack(&indices, 7u);

    PushBack(&indices, 4u); PushBack(&indices, 1u); PushBack(&indices, 10u);

    // Faces around P2
    PushBack(&indices, 2u); PushBack(&indices, 3u); PushBack(&indices, 6u);

    PushBack(&indices, 2u); PushBack(&indices, 6u); PushBack(&indices, 9u);

    PushBack(&indices, 2u); PushBack(&indices, 9u); PushBack(&indices, 10u);

    PushBack(&indices, 2u); PushBack(&indices, 10u); PushBack(&indices, 7u);

    PushBack(&indices, 2u); PushBack(&indices, 7u); PushBack(&indices, 3u);

    // Adjacent faces
    PushBack(&indices, 6u); PushBack(&indices, 3u); PushBack(&indices, 8u);

    PushBack(&indices, 9u); PushBack(&indices, 6u); PushBack(&indices, 5u);

    PushBack(&indices, 10u); PushBack(&indices, 9u); PushBack(&indices, 1u);

    PushBack(&indices, 7u); PushBack(&indices, 10u); PushBack(&indices, 4u);

    PushBack(&indices, 3u); PushBack(&indices, 7u); PushBack(&indices, 11u);
    


    HashTable<MeshEdge, uint32> midpoints;
    uint32 maxNumFaces = 20 * Powi(4, subdivisions);
    AllocateHashTable(&midpoints, 3 * maxNumFaces);

    for (int subdivision = 0; subdivision < subdivisions; subdivision++) {

        HashTableClear(&midpoints);
        DynamicArray<uint32> newIndices = MakeDynamicArray<uint32>(indices.count * 4);

        uint32 faceCount = indices.count / 3;
        for (int face = 0; face < faceCount; face++) {
            // 1) Create new vertices at midpoints (no repeats!)
            uint32 index0 = indices[3 * face + 0];
            uint32 index1 = indices[3 * face + 1];
            uint32 index2 = indices[3 * face + 2];

            MeshEdge edge01 = {index0, index1};
            MeshEdge edge12 = {index1, index2};
            MeshEdge edge20 = {index2, index0};

            uint32 midpoint01, midpoint12, midpoint20;

            if (!HashTableGet(&midpoints, edge01, &midpoint01)) {
                midpoint01 = positions.count;
                PushBack(&positions, Normalize(Lerp(positions[index0], positions[index1], 0.5f)));
            }

            if (!HashTableGet(&midpoints, edge12, &midpoint12)) {
                midpoint12 = positions.count;
                PushBack(&positions, Normalize(Lerp(positions[index1], positions[index2], 0.5f)));
            }

            if (!HashTableGet(&midpoints, edge20, &midpoint20)) {
                midpoint20 = positions.count;
                PushBack(&positions, Normalize(Lerp(positions[index2], positions[index0], 0.5f)));
            }

            // 2) Build new faces in new indices array
            PushBack(&newIndices, index0);
            PushBack(&newIndices, midpoint01);
            PushBack(&newIndices, midpoint20);

            PushBack(&newIndices, midpoint01);
            PushBack(&newIndices, index1);
            PushBack(&newIndices, midpoint12);

            PushBack(&newIndices, midpoint01);
            PushBack(&newIndices, midpoint12);
            PushBack(&newIndices, midpoint20);

            PushBack(&newIndices, midpoint20);
            PushBack(&newIndices, midpoint12);
            PushBack(&newIndices, index2);
        }

        DeallocateDynamicArray(&indices);
        indices = newIndices;
    }

    DeallocateHashTable(&midpoints);

    for (int i = 0; i < indices.count / 3; i++) {
        vec3 normal = Cross(positions[indices[3 * i + 1]] - positions[indices[3 * i + 0]],
                            positions[indices[3 * i + 2]] - positions[indices[3 * i + 0]]);

        vec3 centroid = (positions[indices[3 * i + 0]] + 
                         positions[indices[3 * i + 1]] + 
                         positions[indices[3 * i + 2]]);

        
        if (Dot(normal, centroid) < 0) {
            uint32 temp = indices[3 * i + 0];
            indices[3 * i + 0] = indices[3 * i + 1];
            indices[3 * i + 1] = temp;
        }
    }

    
    uint32 vertexBufferSize = (sizeof(vec3) + sizeof(vec3)) * positions.count;

    ReallocateMesh(mesh, positions.count, vertexBufferSize, indices.count, 1);
    mesh->primitiveType = PrimitiveType_Triangles;

    mesh->positions = (vec3 *)mesh->vertices;
    mesh->normals = (vec3 *)((uint8 *)mesh->vertices + sizeof(vec3) * positions.count);

    // Pack Data
    for (int i = 0; i < positions.count; i++) {
        mesh->positions[i] = positions[i];
        mesh->normals[i] = positions[i];
    }

    Plat.CopyMem(mesh->indices, indices.data, sizeof(uint32) * mesh->indexCount);
    mesh->submeshLengths[0] = mesh->indexCount;

    DeallocateDynamicArray(&positions);
    DeallocateDynamicArray(&indices);
}

void AllocateCylinder(Mesh *mesh, uint32 segments, real32 height = 1.0f) {

    uint32 positionCount = segments * 2 + (segments + 1) * 2;

    u32 triangleCount = 2 * segments + 2 * (segments);

    uint32 indicesCount = triangleCount * 3;

    uint32 vertexBufferSize = (sizeof(vec3) + sizeof(vec3)) * positionCount;
    ReallocateMesh(mesh, positionCount, vertexBufferSize, indicesCount, 1);
    mesh->primitiveType = PrimitiveType_Triangles;

    mesh->positions = (vec3 *)mesh->vertices;
    mesh->normals = (vec3 *)((uint8 *)mesh->vertices + sizeof(vec3) * positionCount);

    uint32 vertIndex = 0;
    uint32 indexCount = 0;

    // Top Face
    {
        uint32 centerIndex = vertIndex;
        mesh->positions[centerIndex] = V3(0, height, 0);
        mesh->normals[centerIndex] = V3(0, 1, 0);
        vertIndex++;

        for (int i = 0; i < segments; i++) {
            real32 theta = 2.0f * _PI * i / (real32)segments;
            mesh->positions[vertIndex] = V3(cosf(theta), height, sinf(theta));
            mesh->normals[vertIndex] = V3(0, 1, 0);
            vertIndex++;
        }

        for (int i = 0; i < segments; i++) {
            mesh->indices[indexCount++] = centerIndex;
            mesh->indices[indexCount++] = centerIndex + 1 + i;
            mesh->indices[indexCount++] = centerIndex + 1 + (i + 1) % segments;
        }
    }

    // Bottom Face
    {
        uint32 centerIndex = vertIndex;
        mesh->positions[centerIndex] = V3(0, 0, 0);
        mesh->normals[centerIndex] = V3(0, -1, 0);
        vertIndex++;

        for (int i = 0; i < segments; i++) {
            real32 theta = 2.0f * _PI * i / (real32)segments;
            mesh->positions[vertIndex] = V3(cosf(theta), 0, sinf(theta));
            mesh->normals[vertIndex] = V3(0, -1, 0);
            vertIndex++;
        }

        for (int i = 0; i < segments; i++) {
            mesh->indices[indexCount++] = centerIndex;
            mesh->indices[indexCount++] = centerIndex + 1 + (i + 1) % segments;
            mesh->indices[indexCount++] = centerIndex + 1 + i;
        }
    }

    // Sides
    {
        uint32 sideStartIndex = vertIndex;
        for (int i = 0; i < segments; i++) {
            real32 theta = 2.0f * _PI * i / (real32)segments;
            vec3 normal = Normalize(V3(cosf(theta), 0, -sinf(theta)));

            mesh->positions[vertIndex] = V3(cosf(theta), height, sinf(theta));
            mesh->normals[vertIndex] = normal;
            vertIndex++;

            mesh->positions[vertIndex] = V3(cosf(theta), 0, sinf(theta));
            mesh->normals[vertIndex] = normal;
            vertIndex++;
        }

        for (int i = 0; i < segments; i++) {
            mesh->indices[indexCount++] = sideStartIndex + 2 * i + 0;
            mesh->indices[indexCount++] = sideStartIndex + 2 * i + 1;
            mesh->indices[indexCount++] = sideStartIndex + 2 * ((i + 1) % segments) + 1;

            mesh->indices[indexCount++] = sideStartIndex + 2 * i + 0;
            mesh->indices[indexCount++] = sideStartIndex + 2 * ((i + 1) % segments) + 1;
            mesh->indices[indexCount++] = sideStartIndex + 2 * ((i + 1) % segments) + 0;
        }
    }

    ASSERT(vertIndex == positionCount);
    ASSERT(indexCount == triangleCount * 3);
    mesh->submeshLengths[0] = indexCount;
}

void AllocateArrow(Mesh *mesh) {
    real32 totalLength = 1.0f;
    real32 headLength = 0.25f;
    real32 shaftLength = totalLength - headLength;

    real32 headRadius = 0.125f;
    real32 shaftRadius = 0.030625f;

    uint32 segments = 16;
    uint32 numFaces = 16 + 2 * 16 + 16 + 16;

    uint32 vertexCount = (16 + 1) + 16 + (16 + 1) + 1;
    uint32 indexCount = 3 * numFaces;

    uint32 vertexBufferSize = sizeof(vec3) * vertexCount;

    ReallocateMesh(mesh, vertexCount, vertexBufferSize, indexCount, 1);
    mesh->primitiveType = PrimitiveType_Triangles;
    mesh->submeshLengths[0] = indexCount;

    mesh->positions = (vec3 *)mesh->vertices;

    uint32 vertexOffset = 0;
    uint32 indexOffset = 0;


    // Shaft bottom
    uint32 shaftBottomCenter = vertexOffset;
    mesh->positions[vertexOffset++] = V3(0, 0, 0);

    uint32 shaftBottomOffset = vertexOffset;

    for (int i = 0; i < segments; i++) {
        real32 z = shaftRadius * cosf(2 * _PI * i / (real32)segments);
        real32 y = shaftRadius * sinf(2 * _PI * i / (real32)segments);

        mesh->positions[vertexOffset++] = V3(0.0f, y, z);
    }

    for (int i = 0; i < segments; i++) {
        mesh->indices[indexOffset++] = shaftBottomCenter;
        mesh->indices[indexOffset++] = shaftBottomOffset + (i % segments);
        mesh->indices[indexOffset++] = shaftBottomOffset + ((i + 1) % segments);
    }


    // Shaft top
    uint32 shaftTopOffset = vertexOffset;

    for (int i = 0; i < segments; i++) {
        real32 z = shaftRadius * cosf(2 * _PI * i / (real32)segments);
        real32 y = shaftRadius * sinf(2 * _PI * i / (real32)segments);

        mesh->positions[vertexOffset++] = V3(shaftLength, y, z);
    }

    for (int i = 0; i < segments; i++) {
        mesh->indices[indexOffset++] = shaftBottomOffset + (i % segments);
        mesh->indices[indexOffset++] = shaftTopOffset + (i % segments);
        mesh->indices[indexOffset++] = shaftBottomOffset + ((i + 1) % segments);

        mesh->indices[indexOffset++] = shaftBottomOffset + ((i + 1) % segments);
        mesh->indices[indexOffset++] = shaftTopOffset + (i % segments);
        mesh->indices[indexOffset++] = shaftTopOffset + ((i + 1) % segments);
    }

    // Cone
    uint32 coneBaseCenter = vertexOffset;
    mesh->positions[vertexOffset++] = V3(shaftLength, 0.0f, 0.0f);

    uint32 coneBaseOffset = vertexOffset;

    for (int i = 0; i < segments; i++) {
        real32 z = headRadius * cosf(2 * _PI * i / (real32)segments);
        real32 y = headRadius * sinf(2 * _PI * i / (real32)segments);

        mesh->positions[vertexOffset++] = V3(shaftLength, y, z);
    }

    uint32 coneTip = vertexOffset;
    mesh->positions[vertexOffset++] = V3(totalLength, 0.0f, 0.0f);

    for (int i = 0; i < segments; i++) {
        mesh->indices[indexOffset++] = coneBaseCenter;
        mesh->indices[indexOffset++] = coneBaseOffset + (i % segments);
        mesh->indices[indexOffset++] = coneBaseOffset + ((i + 1) % segments);
    }

    for (int i = 0; i < segments; i++) {
        mesh->indices[indexOffset++] = coneBaseOffset + (i % segments);
        mesh->indices[indexOffset++] = coneTip;
        mesh->indices[indexOffset++] = coneBaseOffset + ((i + 1) % segments);
    }
}


void AllocateCone(Mesh *mesh, real32 height, real32 radius, uint32 segments = 16) {
    uint32 numFaces = 2 * segments;

    uint32 vertexCount = 2 * segments + 2;
    uint32 indexCount = 3 * numFaces;

    uint32 vertexBufferSize = (sizeof(vec3) * sizeof(vec3)) * vertexCount;

    ReallocateMesh(mesh, vertexCount, vertexBufferSize, indexCount, 1);
    mesh->primitiveType = PrimitiveType_Triangles;
    mesh->submeshLengths[0] = indexCount;

    mesh->positions = (vec3 *)mesh->vertices;
    mesh->normals = (vec3 *)((uint8 *)mesh->vertices + sizeof(vec3) * vertexCount);

    mesh->positions[0] = V3(0, 0, 0);
    mesh->normals[0] = V3(-1, 0, 0);

    mesh->positions[1] = V3(height, 0.0f, 0.0f);
    mesh->normals[1] = V3(1, 0, 0);

    // Bottom face
    uint32 vertexOffset = 2;
    uint32 indexOffset = 0;

    for (int i = 0; i < segments; i++) {
        real32 theta = 2 * _PI * (real32)i / (real32)segments;

        vec3 position = V3(0.0f, radius * cosf(theta), radius * sinf(theta));

        mesh->positions[vertexOffset + i] = position;
        mesh->normals[vertexOffset + i] = V3(-1, 0, 0);
    }

    for (int i = 0; i < segments; i++) {
        int a = vertexOffset + i;
        int b = vertexOffset + ((i + 1) % segments);

        mesh->indices[indexOffset + 3 * i + 0] = 0;
        mesh->indices[indexOffset + 3 * i + 1] = b;
        mesh->indices[indexOffset + 3 * i + 2] = a;
    }

    vertexOffset += segments;
    indexOffset += 3 * segments;


    // Upper face
    for (int i = 0; i < segments; i++) {
        real32 theta = 2 * _PI * (real32)i / (real32)segments;

        vec3 position = V3(0.0f, radius * cosf(theta), radius * sinf(theta));

        vec3 cotangent = mesh->positions[1] - position;
        vec3 tangent = Normalize(Cross(cotangent, V3(1, 0, 0)));
        vec3 normal = Normalize(Cross(tangent, cotangent));

        mesh->positions[vertexOffset + i] = position;
        mesh->normals[vertexOffset + i] = normal;
    }

    for (int i = 0; i < segments; i++) {
        int a = vertexOffset + i;
        int b = vertexOffset + ((i + 1) % segments);

        mesh->indices[indexOffset + 3 * i + 0] = 1;
        mesh->indices[indexOffset + 3 * i + 1] = a;
        mesh->indices[indexOffset + 3 * i + 2] = b;
    }
}

void AllocateCube(Mesh *mesh) {
    mesh->primitiveType = PrimitiveType_Triangles;

    // @TODO: texcoords
    
    uint32 vertexSize = sizeof(vec3) * sizeof(vec3);

    // @NOTE: we're duplicating vertices
    mesh->vertexCount = 24;
    mesh->vertexBufferSize = vertexSize * mesh->vertexCount;
    mesh->vertices = Plat.AllocateMem(mesh->vertexBufferSize);

    mesh->positions = (vec3 *)(mesh->vertices);
    mesh->normals = (vec3 *)((uint8 *)mesh->vertices + sizeof(vec3) * mesh->vertexCount);

    mesh->indexCount = 36;
    mesh->indices = ALLOCATE_ARRAY(uint32, mesh->indexCount);

    // @NOTE: all triangles are counter-clockwise from the outside

    uint32 vertexOffset = 0;
    uint32 indexOffset = 0;
        
    // Pos X
    {
        vec3 normal = V3(1, 0, 0);

        mesh->positions[vertexOffset + 0] = V3( 1.0f, -1.0f,  -1.0f);
        mesh->positions[vertexOffset + 1] = V3( 1.0f, -1.0f, 1.0f);
        mesh->positions[vertexOffset + 2] = V3( 1.0f,  1.0f, 1.0f);
        mesh->positions[vertexOffset + 3] = V3( 1.0f,  1.0f,  -1.0f);

        mesh->normals[vertexOffset + 0] = normal;
        mesh->normals[vertexOffset + 1] = normal;
        mesh->normals[vertexOffset + 2] = normal;
        mesh->normals[vertexOffset + 3] = normal;

        mesh->indices[indexOffset + 0] = vertexOffset + 0;
        mesh->indices[indexOffset + 1] = vertexOffset + 1;
        mesh->indices[indexOffset + 2] = vertexOffset + 2;

        mesh->indices[indexOffset + 3] = vertexOffset + 0;
        mesh->indices[indexOffset + 4] = vertexOffset + 2;
        mesh->indices[indexOffset + 5] = vertexOffset + 3;
    }

    vertexOffset += 4;
    indexOffset += 6;

    // Pos Z
    {
        vec3 normal = V3(0, 0, 1);

        mesh->positions[vertexOffset + 0] = V3(-1.0f, -1.0f,  -1.0f);
        mesh->positions[vertexOffset + 1] = V3( 1.0f, -1.0f,  -1.0f);
        mesh->positions[vertexOffset + 2] = V3( 1.0f,  1.0f,  -1.0f);
        mesh->positions[vertexOffset + 3] = V3(-1.0f,  1.0f,  -1.0f);

        mesh->normals[vertexOffset + 0] = normal;
        mesh->normals[vertexOffset + 1] = normal;
        mesh->normals[vertexOffset + 2] = normal;
        mesh->normals[vertexOffset + 3] = normal;

        mesh->indices[indexOffset + 0] = vertexOffset + 0;
        mesh->indices[indexOffset + 1] = vertexOffset + 1;
        mesh->indices[indexOffset + 2] = vertexOffset + 2;

        mesh->indices[indexOffset + 3] = vertexOffset + 0;
        mesh->indices[indexOffset + 4] = vertexOffset + 2;
        mesh->indices[indexOffset + 5] = vertexOffset + 3;
    }

    vertexOffset += 4;
    indexOffset += 6;

    // Neg X
    {
        vec3 normal = V3(-1, 0, 0);

        mesh->positions[vertexOffset + 0] = V3(-1.0f, -1.0f, 1.0f);
        mesh->positions[vertexOffset + 1] = V3(-1.0f, -1.0f,  -1.0f);
        mesh->positions[vertexOffset + 2] = V3(-1.0f,  1.0f,  -1.0f);
        mesh->positions[vertexOffset + 3] = V3(-1.0f,  1.0f, 1.0f);

        mesh->normals[vertexOffset + 0] = normal;
        mesh->normals[vertexOffset + 1] = normal;
        mesh->normals[vertexOffset + 2] = normal;
        mesh->normals[vertexOffset + 3] = normal;

        mesh->indices[indexOffset + 0] = vertexOffset + 0;
        mesh->indices[indexOffset + 1] = vertexOffset + 1;
        mesh->indices[indexOffset + 2] = vertexOffset + 2;

        mesh->indices[indexOffset + 3] = vertexOffset + 0;
        mesh->indices[indexOffset + 4] = vertexOffset + 2;
        mesh->indices[indexOffset + 5] = vertexOffset + 3;
    }

    vertexOffset += 4;
    indexOffset += 6;

    // Neg Z
    {
        vec3 normal = V3(0, 0, -1);

        mesh->positions[vertexOffset + 0] = V3( 1.0f, -1.0f, 1.0f);
        mesh->positions[vertexOffset + 1] = V3(-1.0f, -1.0f, 1.0f);
        mesh->positions[vertexOffset + 2] = V3(-1.0f,  1.0f, 1.0f);
        mesh->positions[vertexOffset + 3] = V3( 1.0f,  1.0f, 1.0f);

        mesh->normals[vertexOffset + 0] = normal;
        mesh->normals[vertexOffset + 1] = normal;
        mesh->normals[vertexOffset + 2] = normal;
        mesh->normals[vertexOffset + 3] = normal;

        mesh->indices[indexOffset + 0] = vertexOffset + 0;
        mesh->indices[indexOffset + 1] = vertexOffset + 1;
        mesh->indices[indexOffset + 2] = vertexOffset + 2;

        mesh->indices[indexOffset + 3] = vertexOffset + 0;
        mesh->indices[indexOffset + 4] = vertexOffset + 2;
        mesh->indices[indexOffset + 5] = vertexOffset + 3;
    }

    vertexOffset += 4;
    indexOffset += 6;

    // Pos Y
    {
        vec3 normal = V3(0, 1, 0);

        mesh->positions[vertexOffset + 0] = V3(-1.0f,  1.0f,  -1.0f);
        mesh->positions[vertexOffset + 1] = V3( 1.0f,  1.0f,  -1.0f);
        mesh->positions[vertexOffset + 2] = V3( 1.0f,  1.0f, 1.0f);
        mesh->positions[vertexOffset + 3] = V3(-1.0f,  1.0f, 1.0f);

        mesh->normals[vertexOffset + 0] = normal;
        mesh->normals[vertexOffset + 1] = normal;
        mesh->normals[vertexOffset + 2] = normal;
        mesh->normals[vertexOffset + 3] = normal;

        mesh->indices[indexOffset + 0] = vertexOffset + 0;
        mesh->indices[indexOffset + 1] = vertexOffset + 1;
        mesh->indices[indexOffset + 2] = vertexOffset + 2;

        mesh->indices[indexOffset + 3] = vertexOffset + 0;
        mesh->indices[indexOffset + 4] = vertexOffset + 2;
        mesh->indices[indexOffset + 5] = vertexOffset + 3;
    }

    vertexOffset += 4;
    indexOffset += 6;

    // Neg Y
    {
        vec3 normal = V3(1, -1, 0);

        mesh->positions[vertexOffset + 0] = V3(-1.0f, -1.0f, 1.0f);
        mesh->positions[vertexOffset + 1] = V3( 1.0f, -1.0f, 1.0f);
        mesh->positions[vertexOffset + 2] = V3( 1.0f, -1.0f,  -1.0f);
        mesh->positions[vertexOffset + 3] = V3(-1.0f, -1.0f,  -1.0f);

        mesh->normals[vertexOffset + 0] = normal;
        mesh->normals[vertexOffset + 1] = normal;
        mesh->normals[vertexOffset + 2] = normal;
        mesh->normals[vertexOffset + 3] = normal;

        mesh->indices[indexOffset + 0] = vertexOffset + 0;
        mesh->indices[indexOffset + 1] = vertexOffset + 1;
        mesh->indices[indexOffset + 2] = vertexOffset + 2;

        mesh->indices[indexOffset + 3] = vertexOffset + 0;
        mesh->indices[indexOffset + 4] = vertexOffset + 2;
        mesh->indices[indexOffset + 5] = vertexOffset + 3;
    }

    vertexOffset += 4;
    indexOffset += 6;


    mesh->submeshCount = 1;
    mesh->submeshLengths = (uint32 *)Plat.AllocateMem(sizeof(uint32) * mesh->submeshCount);
    mesh->submeshLengths[0] = mesh->indexCount;
}

void AllocateSkyboxCube(Mesh *mesh) {
    mesh->primitiveType = PrimitiveType_Triangles;

    mesh->vertexCount = 8;
    mesh->vertexBufferSize = sizeof(vec3) * mesh->vertexCount;
    mesh->vertices = Plat.AllocateMem(mesh->vertexBufferSize);

    mesh->positions = (vec3 *)mesh->vertices;

    // Bottom
    ((vec3 *)mesh->vertices)[0] = V3(-1.0f, -1.0f,  1.0f);
    ((vec3 *)mesh->vertices)[1] = V3( 1.0f, -1.0f,  1.0f);
    ((vec3 *)mesh->vertices)[2] = V3( 1.0f, -1.0f, -1.0f);
    ((vec3 *)mesh->vertices)[3] = V3(-1.0f, -1.0f, -1.0f);

    // Top
    ((vec3 *)mesh->vertices)[4] = V3(-1.0f,  1.0f,  1.0f);
    ((vec3 *)mesh->vertices)[5] = V3( 1.0f,  1.0f,  1.0f);
    ((vec3 *)mesh->vertices)[6] = V3( 1.0f,  1.0f, -1.0f);
    ((vec3 *)mesh->vertices)[7] = V3(-1.0f,  1.0f, -1.0f);


    mesh->indexCount = 48;
    mesh->indices = ALLOCATE_ARRAY(uint32, mesh->indexCount);

    uint32 faceIndexOffset = 0;

    // @NOTE: all triangles are counter-clockwise from the outside

    // Pos X
    mesh->indices[faceIndexOffset + 0] = 1;
    mesh->indices[faceIndexOffset + 1] = 2;
    mesh->indices[faceIndexOffset + 2] = 5;

    mesh->indices[faceIndexOffset + 3] = 2;
    mesh->indices[faceIndexOffset + 4] = 6;
    mesh->indices[faceIndexOffset + 5] = 5;

    faceIndexOffset += 6;

    // Neg X
    mesh->indices[faceIndexOffset + 0] = 3;
    mesh->indices[faceIndexOffset + 1] = 0;
    mesh->indices[faceIndexOffset + 2] = 7;

    mesh->indices[faceIndexOffset + 3] = 0;
    mesh->indices[faceIndexOffset + 4] = 4;
    mesh->indices[faceIndexOffset + 5] = 7;

    faceIndexOffset += 6;

    // Pos Y
    mesh->indices[faceIndexOffset + 0] = 6;
    mesh->indices[faceIndexOffset + 1] = 7;
    mesh->indices[faceIndexOffset + 2] = 5;

    mesh->indices[faceIndexOffset + 3] = 7;
    mesh->indices[faceIndexOffset + 4] = 4;
    mesh->indices[faceIndexOffset + 5] = 5;

    faceIndexOffset += 6;

    // Neg Y
    mesh->indices[faceIndexOffset + 0] = 1;
    mesh->indices[faceIndexOffset + 1] = 0;
    mesh->indices[faceIndexOffset + 2] = 2;

    mesh->indices[faceIndexOffset + 3] = 0;
    mesh->indices[faceIndexOffset + 4] = 3;
    mesh->indices[faceIndexOffset + 5] = 2;

    faceIndexOffset += 6;

    // Pos Z
    mesh->indices[faceIndexOffset + 0] = 0;
    mesh->indices[faceIndexOffset + 1] = 1;
    mesh->indices[faceIndexOffset + 2] = 4;

    mesh->indices[faceIndexOffset + 3] = 1;
    mesh->indices[faceIndexOffset + 4] = 5;
    mesh->indices[faceIndexOffset + 5] = 4;

    faceIndexOffset += 6;

    // Neg Z
    mesh->indices[faceIndexOffset + 0] = 2;
    mesh->indices[faceIndexOffset + 1] = 3;
    mesh->indices[faceIndexOffset + 2] = 6;

    mesh->indices[faceIndexOffset + 3] = 3;
    mesh->indices[faceIndexOffset + 4] = 7;
    mesh->indices[faceIndexOffset + 5] = 6;

    faceIndexOffset += 6;


    mesh->submeshCount = 1;
    mesh->submeshLengths = (uint32 *)Plat.AllocateMem(sizeof(uint32) * mesh->submeshCount);
    mesh->submeshLengths[0] = mesh->indexCount;
}

void AllocateOutlineCircle(Mesh *mesh, real32 radius, uint32 subdivisions) {
    ASSERT(subdivisions >= 2);

    uint32 vertexCount = subdivisions + 1;
    uint32 vertexBufferSize = sizeof(vec3) * vertexCount;
    uint32 indexCount = 0;

    ReallocateMesh(mesh, vertexCount, vertexBufferSize, indexCount, 0);
    mesh->primitiveType = PrimitiveType_LineStrip;

    mesh->positions = (vec3 *)mesh->vertices;

    for (int i = 0; i < vertexCount; i++) {
        real32 angle = 2 * _PI * i / (real32)subdivisions;
        mesh->positions[i] = V3(radius * cosf(angle), radius * sinf(angle), 0.0f);
    }
}

void DeallocateMeshAdjacencyMap(MeshAdjacencyMap *adjacency) {
    DeallocateDynamicArray(&adjacency->map);
    DeallocateDynamicArray(&adjacency->triangles);
}

void ClearMeshAdjacencyMap(MeshAdjacencyMap *adjacency) {
    DynamicArrayClear(&adjacency->map);
    DynamicArrayClear(&adjacency->triangles);
}

AdjacentTriangleSet GetAdjacentTriangles(MeshAdjacencyMap *adjacency, uint32 centralTriangleIndex) {
    AdjacentTriangleSet result = {};

    if (centralTriangleIndex < adjacency->map.count) {
        result = adjacency->map[centralTriangleIndex];
    }

    return result;
}

b32 EdgeIsAdjacent(uint32 aIndex0, uint32 aIndex1, vec3 aPosition0, vec3 aPosition1,
                   uint32 bIndex0, uint32 bIndex1, vec3 bPosition0, vec3 bPosition1) {
    b32 result = ((aIndex0 == bIndex0 && aIndex1 == bIndex1) ||
                  (aIndex0 == bIndex1 && aIndex1 == bIndex0) ||
                  (aPosition0 == bPosition0 && aPosition1 == bPosition1) ||
                  (aPosition0 == bPosition1 && aPosition1 == bPosition0));
    return result;
}

b32 IsTriangleAdjacent(vec3 *positions, uint32 *indices, uint32 indexStride, uint32 centralTriangleIndex, uint32 candidateTriangleIndex, AdjacentTriangle *adjacentTriangle) {

    if (indexStride == 0) {
        indexStride = 3 * sizeof(uint32);
    }

    uint32 centralIndices[] = {
        //indices[3 * centralTriangleIndex + 0],
        //indices[3 * centralTriangleIndex + 1],
        //indices[3 * centralTriangleIndex + 2],
        ((uint32 *)((uint8 *)indices + indexStride * centralTriangleIndex))[0],
        ((uint32 *)((uint8 *)indices + indexStride * centralTriangleIndex))[1],
        ((uint32 *)((uint8 *)indices + indexStride * centralTriangleIndex))[2],
    };

    vec3 centralPositions[] = {
        positions[centralIndices[0]],
        positions[centralIndices[1]],
        positions[centralIndices[2]],
    };

    uint32 candidateIndices[] = {
        //indices[3 * candidateTriangleIndex + 0],
        //indices[3 * candidateTriangleIndex + 1],
        //indices[3 * candidateTriangleIndex + 2],
        ((uint32 *)((uint8 *)indices + indexStride * candidateTriangleIndex))[0],
        ((uint32 *)((uint8 *)indices + indexStride * candidateTriangleIndex))[1],
        ((uint32 *)((uint8 *)indices + indexStride * candidateTriangleIndex))[2],
    };

    vec3 candidatePositions[] = {
        positions[candidateIndices[0]],
        positions[candidateIndices[1]],
        positions[candidateIndices[2]],
    };

    for (int i = 0; i < 3; i++) {
        for (int j = 0; j < 3; j++) {
            if (EdgeIsAdjacent(centralIndices[i], centralIndices[(i + 1) % 3],
                               centralPositions[i], centralPositions[(i + 1) % 3],
                               candidateIndices[j], candidateIndices[(j + 1) % 3],
                               candidatePositions[j], candidatePositions[(j + 1) % 3])) {
                adjacentTriangle->featureType = FeatureType_Edge;
                adjacentTriangle->centralFeatureIndex = i;
                adjacentTriangle->featureIndex = j;
                return true;
            }
        }
    }

    for (int i = 0; i < 3; i++) {
        for (int j = 0; j < 3; j++) {
            if (centralIndices[i] == candidateIndices[j] || centralPositions[i] == candidatePositions[j]) {
                adjacentTriangle->featureType = FeatureType_Point;
                adjacentTriangle->centralFeatureIndex = i;
                adjacentTriangle->featureIndex = j;
                return true;
            }
        }
    }

    return false;
}

void GenerateAdjacencyMap(vec3 *positions, uint32 *indices, uint32 indicesStride, uint32 indexStart, uint32 indexCount, MeshAdjacencyMap *adjacency) {
    ASSERT(indexStart % 3 == 0);
    ASSERT(indexCount % 3 == 0);

    uint32 triangleCount = indexCount / 3;

    uint32 triangleStart = indexStart / 3; 
    uint32 triangleEnd = triangleStart + triangleCount;

    for (uint32 central = triangleStart; central < triangleEnd; central++) {

        AdjacentTriangleSet set = {};
        set.start = adjacency->triangles.count;
        set.end = set.start;

        set.centralIndex = central;
        set.adjacentIndexAtEdge[0] = -1;
        set.adjacentIndexAtEdge[1] = -1;
        set.adjacentIndexAtEdge[2] = -1;
            
        for (uint32 candidate = triangleStart; candidate < triangleEnd; candidate++) {

            // A triangle can't be adjacent to itself!
            if (central == candidate) {
                continue;
            }

            AdjacentTriangle adjTriangle = {};
            if (IsTriangleAdjacent(positions, indices, indicesStride, central, candidate, &adjTriangle)) {
                adjTriangle.centralIndex = central;
                adjTriangle.index = candidate;
                PushBack(&adjacency->triangles, adjTriangle);

                if (adjTriangle.featureType == FeatureType_Edge) {
                    set.adjacentIndexAtEdge[adjTriangle.centralFeatureIndex] = adjTriangle.index;
                }

                set.end++;
            }
        }

        PushBack(&adjacency->map, set);
    }
}

void GenerateAdjacencyMap(Mesh *mesh, MeshAdjacencyMap *adjacency) {
    ASSERT(mesh != NULL);

    // Pretty sure we only care about triangle adjacency!
    ASSERT(mesh->primitiveType == PrimitiveType_Triangles);

    // Doesn't make sense to compute adjacency without position data
    ASSERT(mesh->positions != NULL);

    // @NOTE: adjacency information is only generated per-submesh. We might want to change this in the future
    uint32 indexOffset = 0;
    for (int submesh = 0; submesh < mesh->submeshCount; submesh++) {

        GenerateAdjacencyMap(mesh->positions, mesh->indices, 0, indexOffset, mesh->submeshLengths[submesh], adjacency);
        indexOffset += mesh->submeshLengths[submesh];
    }
}

vec3 Centroid(ChunkedArray<vec3> verts) {
    vec3 centroid = {};
    for (int i = 0; i < verts.count; i++) {
        centroid = centroid + verts[i];
    }

    if (verts.count > 0) {
        centroid = centroid / (real32)verts.count;
    }

    return centroid;
}
