
MixerNode *GetMixerNode(SoundMixer *mixer, MixerNodeHandle handle) {
    MixerNode *result = NULL;

    if (handle.nodeIndex >= 0 && handle.nodeIndex < mixer->nodeCapacity && handle.generation > 0) {
        MixerNode *node = &mixer->nodes[handle.nodeIndex];

        if (node->generation == handle.generation) {
            result = node;
        }
    }

    return result;
}

MixerNodeHandle AllocateMixerNode(SoundMixer *mixer, MixerNodeType nodeType, MixerNodeHandle parent) {
    ASSERT(nodeType != MixerNodeType_Invalid);

    MixerNodeHandle handle;

    if (PopFront(&mixer->nodeIndexFreeList, &handle.nodeIndex)) {
        MixerNode *node = &mixer->nodes[handle.nodeIndex];

        handle.generation = node->generation + 1;

        Plat.ClearMem(node, sizeof(MixerNode));
        node->generation = handle.generation;

        node->nodeType = nodeType;

        node->buffer.left = Allocate_(&mixer->soundChunks);
        node->buffer.right = Allocate_(&mixer->soundChunks);


        // Insert node as the firstChild of parent, if valid
        node->parent = parent;

        MixerNode *parentNode = GetMixerNode(mixer, node->parent);
        if (parentNode) {
            node->nextSibling = parentNode->firstChild;

            MixerNode *nextSiblingNode = GetMixerNode(mixer, node->nextSibling);
            if (nextSiblingNode) {
                nextSiblingNode->prevSibling = handle;
            }
                
            node->firstChild = MixerNodeHandleInvalid();
            node->prevSibling = MixerNodeHandleInvalid();

            parentNode->firstChild = handle;
        }
        else {
            node->parent = MixerNodeHandleInvalid();
            node->firstChild = MixerNodeHandleInvalid();
            node->prevSibling = MixerNodeHandleInvalid();
            node->nextSibling = MixerNodeHandleInvalid();
        }
    }
    else {
        handle = MixerNodeHandleInvalid();
    }

    return handle;
}

void DeallocateMixerNode(SoundMixer *mixer, MixerNodeHandle handle) {

    MixerNode *node = GetMixerNode(mixer, handle);
    if (node == NULL) {
        return;
    }

    node->playing = false;

    if (node->buffer.left != NULL) {
        Deallocate_(&mixer->soundChunks, node->buffer.left);
    }
    if (node->buffer.right != NULL) {
        Deallocate_(&mixer->soundChunks, node->buffer.right);
    }

    switch (node->nodeType) {
        case MixerNodeType_Group: {
        } break;

        case MixerNodeType_SoundClip: {
        } break;

        case MixerNodeType_SoundStream: {
            stb_vorbis_close(node->soundStreamData.vorbis);
        } break;

        case MixerNodeType_Procedural: {
        } break;

        default: {
            ASSERT(false);
        } break;
    }

    // Remove from MixerNode tree
    MixerNode *parentNode = GetMixerNode(mixer, node->parent);
    if (parentNode) {
        if (parentNode->firstChild == handle) {
            parentNode->firstChild = node->nextSibling;
        }
    }

    MixerNode *prevSiblingNode = GetMixerNode(mixer, node->prevSibling);
    if (prevSiblingNode) {
        prevSiblingNode->nextSibling = node->nextSibling;
    }

    MixerNode *nextSiblingNode = GetMixerNode(mixer, node->nextSibling);
    if (nextSiblingNode) {
        nextSiblingNode->prevSibling = node->prevSibling;
    }

    PushBack(&mixer->nodeIndexFreeList, handle.nodeIndex);
}


void InitMixer(SoundMixer *mixer) {
    mixer->nodeCapacity = 512;
    mixer->nodes = ALLOCATE_CLEARED_ARRAY(MixerNode, mixer->nodeCapacity);

    mixer->nodeIndexFreeList = MakeDynamicArray<int32>(mixer->nodeCapacity);

    for (int i = 0; i < mixer->nodeCapacity; i++) {
        PushBack(&mixer->nodeIndexFreeList, i);
    }

    AllocateMemoryPool(&mixer->soundChunks, 1024);

    // Set up top-level nodes
    mixer->masterGroup = AllocateMixerNode(mixer, MixerNodeType_Group, MixerNodeHandleInvalid());

    MixerNode *masterGroupNode = GetMixerNode(mixer, mixer->masterGroup);
    ASSERT(masterGroupNode != NULL);

    masterGroupNode->parent = MixerNodeHandleInvalid();
    masterGroupNode->firstChild = MixerNodeHandleInvalid();
    masterGroupNode->prevSibling = MixerNodeHandleInvalid();
    masterGroupNode->nextSibling = MixerNodeHandleInvalid();

    mixer->masterBuffer.left = Allocate_(&mixer->soundChunks);
    mixer->masterBuffer.right = Allocate_(&mixer->soundChunks);
    mixer->samplesInMasterBuffer = 0;
}

void DeinitMixer(SoundMixer *mixer) {
    DeallocateDynamicArray(&mixer->nodeIndexFreeList);
    DeallocateMemoryPool(&mixer->soundChunks);
}

MixerNodeHandle Play(SoundMixer *mixer, ProceduralSound sound, real32 volume, bool loop) {
    MixerNodeHandle parent = mixer->masterGroup;
    MixerNodeHandle handle = AllocateMixerNode(mixer, MixerNodeType_Procedural, parent);

    MixerNode *node = GetMixerNode(mixer, handle);
    if (node != NULL) {
        MixerNode *node = GetMixerNode(mixer, handle);
        ASSERT(node != NULL);

        node->sceneHandle = Game.sceneManager->activeScene;

        node->proceduralSound.sound = sound;

        node->samplesRendered = 0;

        node->volume = volume;
        node->pan = 0.0f;
        node->loop = loop;

        node->playing = true;
    }

    return handle;
}

MixerNodeHandle Play(SoundMixer *mixer, SoundClipHandle assetHandle, real32 volume, bool loop) {

    SoundClip *clip = GetSoundClip(Game.assets, assetHandle);

    if (clip == NULL) {
        return MixerNodeHandleInvalid();
    }

    MixerNodeHandle parent = mixer->masterGroup;
    MixerNodeHandle handle = AllocateMixerNode(mixer, MixerNodeType_SoundClip, parent);

    MixerNode *node = GetMixerNode(mixer, handle);
    if (node != NULL) {
        MixerNode *node = GetMixerNode(mixer, handle);
        ASSERT(node != NULL);

        node->sceneHandle = Game.sceneManager->activeScene;

        node->soundClipData.handle = assetHandle;

        node->samplesRendered = 0;

        node->volume = volume;
        node->pan = 0.0f;
        node->loop = loop;

        node->playing = true;
    }

    return handle;
}

MixerNodeHandle Play(SoundMixer *mixer, SoundStreamHandle assetHandle, real32 volume, bool loop) {

    SoundStream *soundStream = GetSoundStream(Game.assets, assetHandle);

    if (soundStream == NULL) {
        return MixerNodeHandleInvalid();
    }

    MixerNodeHandle parent = mixer->masterGroup;
    MixerNodeHandle handle = AllocateMixerNode(mixer, MixerNodeType_SoundStream, parent);

    MixerNode *node = GetMixerNode(mixer, handle);
    if (node != NULL) {
        MixerNode *node = GetMixerNode(mixer, handle);
        ASSERT(node != NULL);

        node->sceneHandle = Game.sceneManager->activeScene;

        node->soundStreamData.handle = assetHandle;

        node->samplesRendered = 0;

        node->volume = volume;
        node->pan = 0.0f;
        node->loop = loop;

        int32 error;
        stb_vorbis *vorbis = stb_vorbis_open_memory(soundStream->data, soundStream->dataSize, &error, NULL);
        ASSERT(vorbis);
        node->soundStreamData.vorbis = vorbis;

        node->playing = true;
    }

    return handle;
}

MixerNodeHandle PlaySpatial(SoundMixer *mixer, SoundClipHandle assetHandle, vec3 soundPos, vec3 listenerPos, r32 soundRange, real32 maxVolume, bool loop) {

    SoundClip *soundClip = GetSoundClip(Game.assets, assetHandle);

    if (soundClip == NULL) {
        return MixerNodeHandleInvalid();
    }

    MixerNodeHandle parent = mixer->masterGroup;
    MixerNodeHandle handle = AllocateMixerNode(mixer, MixerNodeType_SoundClip, parent);

    MixerNode *node = GetMixerNode(mixer, handle);
    if (node != NULL) {
        MixerNode *node = GetMixerNode(mixer, handle);
        ASSERT(node != NULL);

        node->sceneHandle = Game.sceneManager->activeScene;

        node->soundClipData.handle = assetHandle;

        node->samplesRendered = 0;

        node->volume = maxVolume;
        node->pan = 0.0f;
        node->loop = loop;

        node->spatialSound = true;
        node->soundPosition = soundPos;
        node->listenerPosition = listenerPos;
        node->soundRange = soundRange;

        node->playing = true;
    }

    return handle;
}


void Stop(SoundMixer *mixer, MixerNodeHandle handle) {
    DeallocateMixerNode(mixer, handle);
}

uint32 FillProceduralSoundBuffer(SoundMixer *mixer, MixerNode *node, uint32 samplesToRender, MixerNodeOutput output) {
    ProceduralSound sound = node->proceduralSound.sound;
    real32 *data = sound.data;
    if (data == NULL) {
        return 0;
    }

    if (node->samplesRendered == sound.sampleCount) {
        if (node->loop) {
            node->samplesRendered = 0;
        }
        else {
            return 0;
        }
    }

    samplesToRender = Min((uint32)(sound.sampleCount - node->samplesRendered), samplesToRender);

    real32 left;
    real32 right;
    for (int i = 0; i < samplesToRender; i++) {
        int sampleIndex = node->samplesRendered + i;
        
        switch (sound.channels) {
            case 1: {
                left = sound.data[sampleIndex];
                right = left;
            } break;

            case 2: {
                left = sound.data[2 * sampleIndex];
                right = sound.data[2 * sampleIndex + 1];
            } break;

            default: {
                ASSERT(false);
            } break;
        }

        output.left->samples[i] = left;
        output.right->samples[i] = right;
    }

    return samplesToRender;    

}

uint32 FillSoundClipBuffer(SoundMixer *mixer, MixerNode *node, uint32 samplesToRender, MixerNodeOutput output) {
    SoundClip *soundClip = GetSoundClip(Game.assets, node->soundClipData.handle);
    if (soundClip == NULL) {
        return 0;
    }

    if (node->samplesRendered == soundClip->sampleCount) {
        if (node->loop) {
            node->samplesRendered = 0;
        }
        else {
            return 0;
        }
    }

    samplesToRender = Min((uint32)(soundClip->sampleCount - node->samplesRendered), samplesToRender);

    real32 left;
    real32 right;
    for (int i = 0; i < samplesToRender; i++) {
        int sampleIndex = node->samplesRendered + i;

        switch (soundClip->channels) {
            case 1: {
                left = soundClip->data[sampleIndex];
                right = left;
            } break;

            case 2: {
                left = soundClip->data[2 * sampleIndex];
                right = soundClip->data[2 * sampleIndex + 1];
            } break;

            default: {
                ASSERT(false);
            } break;
        }

        output.left->samples[i] = left;
        output.right->samples[i] = right;
    }

    return samplesToRender;
}

uint32 FillSoundStreamBuffer(SoundMixer *mixer, MixerNode *node, uint32 samplesToRender, MixerNodeOutput output) {
    SoundStream *soundStream = GetSoundStream(Game.assets, node->soundStreamData.handle);
    if (soundStream == NULL) {
        return 0;
    }

    if (node->samplesRendered == soundStream->sampleCount) {
        if (node->loop) {
            stb_vorbis_close(node->soundStreamData.vorbis);

            int32 error;
            stb_vorbis *vorbis = stb_vorbis_open_memory(soundStream->data, soundStream->dataSize, &error, NULL);
            ASSERT(vorbis);
            node->soundStreamData.vorbis = vorbis;
            node->samplesRendered = 0;
        }
        else {
            return 0;
        }
    }


    samplesToRender = Min((uint32)(soundStream->sampleCount - node->samplesRendered), samplesToRender);
    int32 samplesRendered = 0;

    switch (soundStream->channels) {
        case 1: {
            samplesRendered = stb_vorbis_get_samples_float(node->soundStreamData.vorbis, 1, (real32 **)&output.left->samples, samplesToRender);
            if (samplesRendered > 0) {
                Plat.CopyMem(output.right->samples, output.left->samples, samplesRendered * sizeof(real32));
            }
        } break;

        case 2: {
            float *buffers[2] = {
                output.left->samples,
                output.right->samples,
            };

            samplesRendered = stb_vorbis_get_samples_float(node->soundStreamData.vorbis, 2, buffers, samplesToRender);
        } break;

        default: {
            ASSERT(false);
        } break;
    }

    return samplesRendered;
}

uint32 FillGroupBuffer(SoundMixer *mixer, MixerNode *node, uint32 samplesToRender, MixerNodeOutput output) {

    Plat.ClearMem(output.left, sizeof(SoundChunk));
    Plat.ClearMem(output.right, sizeof(SoundChunk));

    MixerNodeOutput tempOutput;
    tempOutput.left = Allocate_(&mixer->soundChunks);
    tempOutput.right = Allocate_(&mixer->soundChunks);

    MixerNodeHandle child = node->firstChild;
    MixerNodeHandle nextChild;
    while (true) {
        MixerNode *childNode = GetMixerNode(mixer, child);
        if (childNode == NULL) {
            break;
        }

        // store nextChild in case childNode is stopped
        nextChild = childNode->nextSibling;

        Plat.ClearMem(tempOutput.left, sizeof(SoundChunk));
        Plat.ClearMem(tempOutput.right, sizeof(SoundChunk));
        WriteNodeSamples(mixer, child, samplesToRender, tempOutput);

        for (int i = 0; i < samplesToRender; i++) {
            output.left->samples[i] += tempOutput.left->samples[i];
        }

        for (int i = 0; i < samplesToRender; i++) {
            output.right->samples[i] += tempOutput.right->samples[i];
        }

        child = nextChild;
    }

    Deallocate_(&mixer->soundChunks, tempOutput.left);
    Deallocate_(&mixer->soundChunks, tempOutput.right);

    return samplesToRender;
}

uint32 WriteNodeSamples(SoundMixer *mixer, MixerNodeHandle handle, uint32 sampleCount, MixerNodeOutput output) {
    MixerNode *node = GetMixerNode(mixer, handle);
    ASSERT(node != NULL);


    switch (node->nodeType) {
        case MixerNodeType_SoundClip:
        case MixerNodeType_SoundStream: {
            if (!node->playing || !IsSceneActive(node->sceneHandle)) {
                return 0;
            }
        } break;

        default: {
        } break;
    }


    MixerNodeOutput buffer = node->buffer;

    uint32 samplesRendered = 0;
    while (samplesRendered < SAMPLES_PER_SOUNDCHUNK) {
        if (node->samplesRenderedFromBuffer == node->samplesInBuffer) {

            uint32 samplesFilled = 0;

            switch (node->nodeType) {
                case MixerNodeType_Group: {
                    samplesFilled = FillGroupBuffer(mixer, node, SAMPLES_PER_SOUNDCHUNK, buffer);
                } break;

                case MixerNodeType_SoundClip: {
                    samplesFilled = FillSoundClipBuffer(mixer, node, SAMPLES_PER_SOUNDCHUNK, buffer);
                } break;

                case MixerNodeType_SoundStream: {
                    samplesFilled = FillSoundStreamBuffer(mixer, node, SAMPLES_PER_SOUNDCHUNK, buffer);
                } break;

                case MixerNodeType_Procedural: {
                    samplesFilled = FillProceduralSoundBuffer(mixer, node, SAMPLES_PER_SOUNDCHUNK, buffer);
                } break;

                default: {
                    ASSERT(false);
                } break;
            };

            if (samplesFilled == 0) {
                Stop(mixer, handle);
                break;
            }

            node->samplesInBuffer = samplesFilled;
            node->samplesRenderedFromBuffer = 0;
            node->samplesRendered += samplesFilled;
        }

        uint32 bufferStartSample = node->samplesRenderedFromBuffer;
        uint32 samplesToRender = Min(node->samplesInBuffer - bufferStartSample,
                                     sampleCount - samplesRendered);

        real32 volume;
        real32 panAttenuationLeft;
        real32 panAttenuationRight;
        for (int i = 0; i < samplesToRender; i++) {
            if (node->spatialSound) {
                volume = node->volume -
                    Clamp(Distance(node->soundPosition, node->listenerPosition) / node->soundRange, 0.0f, 1.0f);
            }
            else {
                volume = node->volume;    
            }
            
            EqualPowerPan(node->pan, &panAttenuationLeft, &panAttenuationRight);

            output.left->samples[samplesRendered + i] = volume * panAttenuationLeft * buffer.left->samples[bufferStartSample + i];
            output.right->samples[samplesRendered + i] = volume * panAttenuationRight * buffer.right->samples[bufferStartSample + i];
        }

        samplesRendered += samplesToRender;
        node->samplesRenderedFromBuffer += samplesToRender;
    }

    return samplesRendered;
}

void WriteSamples(SoundMixer *mixer, uint32 sampleCount, real32 *output) {
    TIMED_BLOCK();
    Plat.ClearMem(output, sizeof(real32) * 2 * sampleCount);

    uint32 samplesRendered = 0;
    while (samplesRendered < sampleCount) {
        if (mixer->samplesRenderedFromMasterBuffer == mixer->samplesInMasterBuffer) {
            Plat.ClearMem(mixer->masterBuffer.left, sizeof(SoundChunk));
            Plat.ClearMem(mixer->masterBuffer.right, sizeof(SoundChunk));

            uint32 samplesFilled = WriteNodeSamples(mixer, mixer->masterGroup, SAMPLES_PER_SOUNDCHUNK, mixer->masterBuffer);
            
            mixer->samplesInMasterBuffer = samplesFilled;
            mixer->samplesRenderedFromMasterBuffer = 0;
        }

        uint32 bufferStartSample = mixer->samplesRenderedFromMasterBuffer;
        uint32 samplesToRender = Min(mixer->samplesInMasterBuffer - bufferStartSample,
                                     sampleCount - samplesRendered);

        for (int i = 0; i < samplesToRender; i++) {
            int32 outputIndex = samplesRendered + i;
            
            output[2 * outputIndex] = mixer->masterBuffer.left->samples[bufferStartSample + i];
            output[2 * outputIndex + 1] = mixer->masterBuffer.right->samples[bufferStartSample + i];
        }

        mixer->samplesRenderedFromMasterBuffer += samplesToRender;
        samplesRendered += samplesToRender;
    }

    mixer->samplesRendered += samplesRendered;
}

