
#ifndef _ETC_MIXER_H
#define _ETC_MIXER_H

struct stb_vorbis;


// Audio processing happens in discrete chunks of a fixed size. The
// mixer works only at that size internally, but buffers output and
// potentially renders several chunks per frame
#define SAMPLES_PER_SOUNDCHUNK 1024

struct SoundChunk {
    real32 samples[SAMPLES_PER_SOUNDCHUNK];
};

struct MixerNodeOutput {
    SoundChunk *left;
    SoundChunk *right;
};

enum MixerNodeType {
    MixerNodeType_Invalid,

    MixerNodeType_Group,

    MixerNodeType_SoundClip,
    MixerNodeType_SoundStream,

    MixerNodeType_Procedural,
};

STRUCT(Edit = false)
struct MixerNodeHandle {
    int32 nodeIndex;
    int32 generation;
};

inline MixerNodeHandle MixerNodeHandleInvalid() {
    MixerNodeHandle handle;

    handle.nodeIndex = -1;
    handle.generation = 0;

    return handle;
}

inline b32 operator==(MixerNodeHandle a, MixerNodeHandle b) {
    b32 result = (a.nodeIndex == b.nodeIndex &&
                  a.generation == b.generation);

    return result;
}

//inline b32 MixerNodeHandleIsValid(MixerNodeHandle handle) {
//    return (handle.nodeIndex >= 0 && handle.generation > 0);
//}

struct ProceduralSound {
    uint8 channels;
    uint32 sampleCount;
    real32 *data;
};

struct MixerNode {
    SceneHandle sceneHandle;
    int32 generation;

    // MixerNodes are nodes in a tree. The masterGroup is the
    // default parent for all MixerNodes. When rendering sound, the
    // tree is walked depth-first, started from the
    // masterGroup. Therefore, MixerNodes that aren't ancestered by
    // masterGroup won't play!
    MixerNodeHandle parent;
    MixerNodeHandle firstChild;
    MixerNodeHandle prevSibling;
    MixerNodeHandle nextSibling;

    MixerNodeType nodeType;

    real32 volume;
    real32 pan;
    bool loop;

    bool playing;

    MixerNodeOutput buffer;
    uint32 samplesInBuffer;
    uint32 samplesRenderedFromBuffer;
    uint64 samplesRendered;

    bool spatialSound;
    r32 soundRange;
    vec3 soundPosition;
    vec3 listenerPosition;

    union {
        struct {
            SoundClipHandle handle;
        } soundClipData;

        struct {
            SoundStreamHandle handle;
            stb_vorbis *vorbis;
        } soundStreamData;

        struct {
            ProceduralSound sound;
        } proceduralSound;
    };
};

struct SoundMixer {
    uint32 nodeCapacity;
    MixerNode *nodes;
    DynamicArray<int32> nodeIndexFreeList;

    MemoryPool<SoundChunk> soundChunks;

    uint64 samplesRendered;

    MixerNodeHandle masterGroup;
    MixerNodeOutput masterBuffer;
    uint32 samplesInMasterBuffer;
    uint32 samplesRenderedFromMasterBuffer;
};

void InitMixer(SoundMixer *mixer);
void DeinitMixer(SoundMixer *mixer);

MixerNode *GetMixerNode(SoundMixer *mixer, MixerNodeHandle handle);


MixerNodeHandle Play(SoundMixer *mixer, ProceduralSound sound, real32 volume = 1.0f, bool loop = false);
MixerNodeHandle Play(SoundMixer *mixer, SoundClipHandle asset, real32 volume = 1.0f, bool loop = false);
MixerNodeHandle Play(SoundMixer *mixer, SoundStreamHandle asset, real32 volume = 1.0f, bool loop = false);


MixerNodeHandle PlaySpatial(SoundMixer *mixer, SoundClipHandle asset, vec3 soundPos, vec3 listenerPos, r32 soundRange, real32 maxVolume = 1.0f, bool loop = false);

void Stop(SoundMixer *mixer, MixerNodeHandle source);

void WriteSamples(SoundMixer *mixer, uint32 sampleCount, real32 *output);
uint32 WriteNodeSamples(SoundMixer *mixer, MixerNodeHandle handle, uint32 sampleCount, MixerNodeOutput output);

// DSP stuff
inline void EqualPowerPan(real32 pan, real32 *leftAttenuation, real32 *rightAttenuation) {
    *leftAttenuation = 1.0f;
    *rightAttenuation = 1.0f;
}

#endif

