
#include <stdio.h>
#include <string.h>

#include <fcntl.h>  
#include <sys/stat.h>
#include <io.h>


void WinDebugPrint(const char *fmt, ...) {
    const uint32 maxBufferLength = 1024;
    char buffer[maxBufferLength];

    va_list args;
    va_start (args, fmt);

    vsnprintf(buffer, maxBufferLength, fmt, args);

    OutputDebugString(buffer);
    OutputDebugString("\n");

    va_end(args);
}

b32 WinOpenFileForRead(const char *path, FileHandle *handle) {
    b32 result = false;

    handle->handle = CreateFile(path, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);

    if (handle->handle == INVALID_HANDLE_VALUE) {
        DWORD error = GetLastError();
        Plat.DebugPrint("Could not open file %s", path);

        return false;
    }

    handle->fileSize = GetFileSize(handle->handle, NULL);
    handle->dataLength = handle->fileSize;
    handle->offset = 0;

    handle->mapHandle = CreateFileMapping(handle->handle, NULL, PAGE_READONLY, 0, 0, NULL);
    if (handle->mapHandle == NULL) {
        int32 error = GetLastError();
        Plat.DebugPrint("%s mapping error %d", path, error);

        return false;
    }

    handle->data = MapViewOfFile(handle->mapHandle, FILE_MAP_READ, 0, 0, handle->fileSize);

    if (handle->data == NULL) {
        int32 error = GetLastError();

        CloseHandle(handle->handle);
        return false;
    }

    return true;
}

b32 WinOpenFileForWrite(const char *path, FileHandle *handle, uint64 initialLength) {
    b32 result = false;

    Plat.ClearMem(handle, sizeof(FileHandle));

    // why is the size always 4294967295 if we use CREATE_NEW?
    handle->handle = CreateFile(path, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);

    if (handle->handle == INVALID_HANDLE_VALUE) {
        DWORD error = GetLastError();

        Plat.DebugPrint("Could not open file %s", path);
        return false;
    }
    
    handle->fileSize = GetFileSize(handle->handle, NULL);
    handle->offset = handle->fileSize;
    
    SYSTEM_INFO system_info;
    GetSystemInfo(&system_info);
    int32 pageSize = system_info.dwPageSize;

    if (handle->dataLength == 0) {
        handle->dataLength = pageSize;
    }
    if (handle->fileSize > handle->dataLength) {
        handle->dataLength = handle->fileSize;
    }
    if (initialLength > handle->dataLength) {
        handle->dataLength = initialLength;
    }

    handle->mapHandle = CreateFileMapping(handle->handle, NULL, PAGE_READWRITE, 0, handle->dataLength, NULL);
    // The volume for a file has been externally altered so that the opened file is no longer valid.
    if (handle->mapHandle == NULL) {
        DWORD error = GetLastError();
        Plat.DebugPrint("%s mapping error %d", path, error);

        CloseHandle(handle->handle);
        return false;
    }

    handle->data = MapViewOfFile(handle->mapHandle, FILE_MAP_WRITE, 0, 0, handle->fileSize);

    if (handle->data == NULL) {
        DWORD error = GetLastError();

        CloseHandle(handle->mapHandle);
        CloseHandle(handle->handle);
        return false;
    }

    return true;
}


// @TODO: need to look at this some more
b32 WinResizeFile(FileHandle *handle, uint64 fileSize, uint64 dataLength) {
    ASSERT(handle != NULL);

    b32 result = false;

    if (fileSize != handle->fileSize) {
        handle->fileSize = fileSize;
        
        if (handle->offset > fileSize) {
            handle->offset = fileSize;
        }
    }

    SYSTEM_INFO system_info;
    GetSystemInfo (&system_info);
    int32 pageSize = system_info.dwPageSize;

    if (dataLength == 0) {
        dataLength = pageSize;
    }
    if (fileSize > dataLength) {
        dataLength = fileSize;
    }

    if (dataLength != handle->dataLength) {
        b32 success = UnmapViewOfFile(handle->data);
        if (!success) {
            DWORD error = GetLastError();
            return false;
        }
        
        success = CloseHandle(handle->mapHandle);
        if (!success) {
            DWORD error = GetLastError();
            return false;
        }
        
        handle->mapHandle = CreateFileMapping(handle->handle, NULL, PAGE_READWRITE, 0, dataLength, NULL);
        if (handle->mapHandle == INVALID_HANDLE_VALUE) {
            DWORD error = GetLastError();
            return false;
        }

        handle->data = MapViewOfFile(handle->mapHandle, FILE_MAP_WRITE, 0, 0, handle->fileSize);
        if (handle->data == NULL) {
            DWORD error = GetLastError();
            return false;
        }

        handle->dataLength = dataLength;
    }

    return true;
}

void WinCloseFile(FileHandle *handle) {
    ASSERT(handle != NULL);

    b32 success = UnmapViewOfFile(handle->data);
    if (!success) {
        DWORD error = GetLastError();
        return;
    }

    success = CloseHandle(handle->mapHandle);
    if (!success) {
        DWORD error = GetLastError();
        return;
    }

    DWORD filePtr = SetFilePointer(handle->handle, handle->fileSize, NULL, FILE_CURRENT);
    if (filePtr == INVALID_SET_FILE_POINTER) {
        DWORD error = GetLastError();
        return;
    }

    success = SetEndOfFile(handle->handle);
    if (!success) {
        DWORD error = GetLastError();
        if (error != ERROR_ACCESS_DENIED) {
            return;
        }
    }

    success = CloseHandle(handle->handle);
    if (!success) {
        DWORD error = GetLastError();
        return;
    }
}

int WinSystem(const char *command) {
    return system(command);
}

void WinMakeDirectory(const char *path) {
    // @TODO: this might not work?
    CreateDirectory(path, NULL);
}

void WinRemoveFile(const char *path) {
    DeleteFile(path);
}

char *_WinFindFileInDirectory(const char *directory, const char *name, uint32 extensionCount, const char **extensions) {
    char *result = NULL;

    WIN32_FIND_DATAA findData = {};

    const char *searchPattern = PathConcat(directory, "*");
    HANDLE handle = FindFirstFileA(searchPattern, &findData);

    int32 nameLength = strlen(name);

    if (handle != INVALID_HANDLE_VALUE) {
        do {
            if (findData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
                // @note: FindNextFile can take us one directory out if we dont check.
                if (!StringEquals((const char *)findData.cFileName, ".") &&
                    !StringEquals((const char *)findData.cFileName, "..")) {
                    const char *nextDirectory = PathConcat(directory, findData.cFileName);
                    char *recursiveResult = _WinFindFileInDirectory(nextDirectory, name, extensionCount, extensions);
                    Plat.DeallocateMem((char *)nextDirectory);

                    if (recursiveResult != NULL) {
                        result = recursiveResult;
                        break;
                    }
                }
            }
            else {
                int32 filenameLengthWithoutExtension = strlen(findData.cFileName);
                filenameLengthWithoutExtension = FindLastSubstring(findData.cFileName, filenameLengthWithoutExtension, ".");

                if (filenameLengthWithoutExtension != nameLength) {
                    continue;
                }
                
                if (extensionCount > 0) {
                    b32 hasExtension = false;
                    for (int i = 0; i < extensionCount; i++) {
                        if (HasFileExtension(findData.cFileName, extensions[i])) {
                            hasExtension = true;
                            break;
                        }
                    }

                    if (!hasExtension) {
                        continue;
                    }
                }

                if (StringEquals(findData.cFileName, name, nameLength)) {
                    result = (char *)PathConcat(directory, findData.cFileName);
                    break;
                }
            }
        } while (FindNextFileA(handle, &findData));

        FindClose(handle);
    }
    else {
        int32 error = GetLastError();
        Plat.DebugPrint("error code %d", error);
    }

    Plat.DeallocateMem((char *)searchPattern);

    return result;
}

char *WinFindFileInDirectory(const char *directory, const char *name, uint32 extensionCount, const char **extensions) {
    char *result = _WinFindFileInDirectory(directory, name, extensionCount, extensions);

    return result;
}

bool _WinGetAllFilesWithExtensions(const char *directory, uint32 extensionCount, const char **extensions, DynamicArray<const char *> *allFiles) {
    bool result = false;

    WIN32_FIND_DATAA findData = {};

    const char *dir = PathConcat(directory, "*");
    HANDLE handle = FindFirstFileA(dir, &findData);

    if (handle != INVALID_HANDLE_VALUE) {
        do {
            if (findData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
                // @note: FindNextFile can take us one directory out if we dont check.
                if (!StringEquals((const char *)findData.cFileName, ".") &&
                    !StringEquals((const char *)findData.cFileName, "..")) {
                    const char *nextDirectory = PathConcat(directory, findData.cFileName);
                    _WinGetAllFilesWithExtensions(nextDirectory, extensionCount, extensions, allFiles);
                    Plat.DeallocateMem((char *)nextDirectory);
                }
            }
            else {
                for (int i = 0; i < extensionCount; i++) {
                    if (HasFileExtension(findData.cFileName, extensions[i])) {
                        const char *filePath = PathConcat(directory, findData.cFileName);
                        PushBack(allFiles, filePath);
                        break;
                    }
                }
            }
        } while (FindNextFileA(handle, &findData));

        FindClose(handle);
    }
    else {
        int32 error = GetLastError();
        Plat.DebugPrint("error code %d", error);
    }

    Plat.DeallocateMem((char *)dir);

    return result;
}

DynamicArray<const char *> WinGetAllFilesWithExtensions(const char *directory, uint32 extensionCount, const char **extensions) {
    DynamicArray<const char *> files = {};

    _WinGetAllFilesWithExtensions(TEXT(directory), extensionCount, extensions, &files);

    return files;
}

uint64 WinFileLastModifiedTime(const char *path) {
    uint64 result = 0;

    WIN32_FILE_ATTRIBUTE_DATA Data;
    if(GetFileAttributesEx(path, GetFileExInfoStandard, &Data))
    {
        FILETIME filetime = Data.ftLastWriteTime;
        ULARGE_INTEGER ulargeint;
        ulargeint.LowPart = filetime.dwLowDateTime;
        ulargeint.HighPart = filetime.dwHighDateTime;

        result = ulargeint.QuadPart;
    }

    return result;
}

void *WinAllocateMem(uint32 size) {
    return malloc(size);
}

void *WinResizeMem(void *mem, uint32 size) {
    return realloc(mem, size);
}

void WinDeallocateMem(void *mem) {
    free(mem);
}

void *WinCopyMem(void *destination, const void *source, uint32 size) {
    return memcpy(destination, source, size);
}

void *WinMoveMem(void *destination, const void *source, uint32 size) {
    return memmove(destination, source, size);
}

void WinClearMem(void *mem, uint32 size) {
    memset(mem, 0, size);
}

void WinSemaSignal(void *sema) {
    ReleaseSemaphore(*((HANDLE *)sema), 1, NULL);
}

b32 WinSemaWait(void *sema, b32 block) {
    b32 result = false;

    if (block) {
        DWORD returnValue = WaitForSingleObjectEx(*((HANDLE *)sema), INFINITE, FALSE);
        result = returnValue == WAIT_OBJECT_0;
    }
    else {
        DWORD returnValue = WaitForSingleObjectEx(*((HANDLE *)sema), 0, FALSE);
        result = returnValue == WAIT_OBJECT_0;
    }

    return result;
}

ThreadLocalStorageKey WinCreateTLSKey() {
    ThreadLocalStorageKey key;

    key.key = TlsAlloc();
    ASSERT(key.key != TLS_OUT_OF_INDEXES);

    return key;
}

void *WinGetTLS(ThreadLocalStorageKey key) {
    return TlsGetValue(key.key);
}

void WinSetTLS(ThreadLocalStorageKey key, void *data) {
    TlsSetValue(key.key, data);
}

#ifdef ETC_GAME
void WinMoveMouse(int32 screenX, int32 screenY) {
    POINT screenPoint;
    screenPoint.x = screenX;
    screenPoint.y = (int32)Game.graphics->height - screenY;

    //POINT screenPoint;
    ClientToScreen(GlobalPlatformMemory->window, &screenPoint);

    SetCursorPos(screenPoint.x, (int32)screenPoint.y);
}

void WinShowMouse() {
    ShowCursor(true);
}

void WinHideMouse() {
    ShowCursor(false);
}

void WinSetMouseWrap(bool wrapMouse) {
}


// @NOTE: time in milliseconds
real64 WinGetTime() {
    LARGE_INTEGER systemTime;
    QueryPerformanceCounter(&systemTime);
    return (real64)systemTime.QuadPart / (real64)GlobalPlatformMemory->PerfCountFrequency.QuadPart;
}

#endif


PlatformAPI WindowsPlatformAPI() {
    PlatformAPI result = {};
    
    result.DebugPrint = WinDebugPrint;

    result.OpenFileForRead = WinOpenFileForRead;
    result.OpenFileForWrite = WinOpenFileForWrite;
    result.ResizeFile = WinResizeFile;
    result.CloseFile = WinCloseFile;

    result.System = WinSystem;

    result.MakeDirectory = WinMakeDirectory;
    result.RemoveFile = WinRemoveFile;
    
    result.FindFileInDirectory = WinFindFileInDirectory;
    result.FileLastModifiedTime = WinFileLastModifiedTime;
    result.GetAllFilesWithExtensions = WinGetAllFilesWithExtensions;
    
    result.AllocateMem = WinAllocateMem;
    result.ResizeMem = WinResizeMem;
    result.CopyMem = WinCopyMem;
    result.MoveMem = WinMoveMem;
    result.DeallocateMem = WinDeallocateMem;
    result.ClearMem = WinClearMem;

    result.SemaSignal = WinSemaSignal;
    result.SemaWait = WinSemaWait;

    result.CreateTLSKey = WinCreateTLSKey;
    result.GetTLS = WinGetTLS;
    result.SetTLS = WinSetTLS;

#ifdef ETC_GAME
    result.MoveMouse = WinMoveMouse;
    result.ShowMouse = WinShowMouse;
    result.HideMouse = WinHideMouse;

    result.SetMouseWrap = WinSetMouseWrap;

    result.GetTime = WinGetTime;
#endif

    result.taskSchedulerTLSKey = WinCreateTLSKey();

    return result;
};
