
void WriteWhitespace(FileHandle *file, uint32 spaces) {
    for (int i = 0; i < spaces; i++) {
        WriteString(file, " ");
    }
}

void WriteTemplateTypes(FileHandle *file, DynamicArray<ParsedTypeDescription> templateTypes) {
    WriteString(file, "<");

    for (int i = 0; i < templateTypes.count; i++) {

        ParsedTypeDescription type = templateTypes[i];

        if (type.isConst) {
            WriteString(file, "const ");
        }

        WriteBytes(file, type.typeName.text, type.typeName.length);

        if (type.templateTypes.count > 0) {
            WriteTemplateTypes(file, type.templateTypes);
        }

        if (type.pointerLevels > 0) {
            WriteString(file, " ");
            for (int j = 0; j < type.pointerLevels; j++) {
                WriteString(file, "*");
            }
        }
        
        if (i < templateTypes.count - 1) {
            WriteString(file, ", ");
        }
    }

    WriteString(file, ">");
}

void WriteTemplateTypes(FileHandle *file, DynamicArray<SchemaTypeDescription> templateTypes) {
    WriteString(file, "<");

    for (int i = 0; i < templateTypes.count; i++) {

        SchemaTypeDescription type = templateTypes[i];

        if (type.isConst) {
            WriteString(file, "const ");
        }

        switch (type.dataType) {
            case SchemaDataType_Struct:
            case SchemaDataType_Enum:
            case SchemaDataType_TaggedUnion: {
                WriteString(file, type.typeName);
            } break;

            case SchemaDataType_String: {
                WriteString(file, "char *");
            } break;

            case SchemaDataType_PersistentFunctionHandle: {
                WriteString(file, type.typeName);
            } break;

            case SchemaType_AssetHandle: {
                WriteString(file, type.typeName);
            } break;

            default: {
                WriteString(file, type.typeName);
            } break;
        }

        if (type.templateTypes.count > 0) {
            WriteTemplateTypes(file, type.templateTypes);
        }

        if (i < templateTypes.count - 1) {
            WriteString(file, ", ");
        }
    }

    WriteString(file, ">");
}

void WriteStruct(FileHandle *file, PPToken name, DynamicArray<ParsedField> fields, const char *nameDecorator) {
    WriteString(file, "struct ");
    WriteBytes(file, name.text, name.length);

    if (nameDecorator != NULL) {
        WriteString(file, nameDecorator);
    }

    WriteString(file, " {\n");
    for (int i = 0; i < fields.count; i++) {
        ParsedField field = fields[i];

        WriteWhitespace(file, 4);
        if (field.type.isConst) {
            WriteString(file, "const ");
        }

        WriteBytes(file, field.type.typeName.text, field.type.typeName.length);

        if (field.type.templateTypes.count > 0) {
            WriteTemplateTypes(file, field.type.templateTypes);
        }
                
        WriteString(file, " ");
        for (int k = 0; k < field.type.pointerLevels; k++) {
            WriteString(file, "*");
        }

        WriteBytes(file, field.name.text, field.name.length);

        if (field.type.arrayLength.type != PPTokenType_Unknown) {
            WriteStringf(file, "[%d]", field.type.arrayLengthValue);
        }

        WriteString(file, ";\n");
    }
    WriteString(file, "};\n");
}

void WriteFunctionArguments(FileHandle *file, DynamicArray<ParsedFunctionArgument> arguments) {
    for (int j = 0; j < arguments.count; j++) {
        ParsedFunctionArgument arg = arguments[j];

        if (arg.type.isConst) {
            WriteStringf(file, "const ");
        }

        WriteStringf(file, "%.*s", arg.type.typeName.length, arg.type.typeName.text);

        if (arg.type.templateTypes.count > 0) {
            WriteTemplateTypes(file, arg.type.templateTypes);
        }

        WriteString(file, " ");

        for (int k = 0; k < arg.type.pointerLevels; k++) {
            WriteStringf(file, "*");
        }
            
        WriteStringf(file, "%.*s", arg.name.length, arg.name.text);

        if (j < arguments.count - 1) {
            WriteStringf(file, ", ");
        }
    }
}

void WriteEnum(FileHandle *file, PPToken name, DynamicArray<PPToken> values) {
    WriteStringf(file, "enum %.*s : uint32 {\n", name.length, name.text);

    for (int i = 0; i < values.count; i++) {
        WriteWhitespace(file, 4);
        WriteStringf(file, "%.*s,\n", values[i].length, values[i].text);
    }

    WriteString(file, "};\n");
}

void WriteBitfield(FileHandle *file, PPToken name, DynamicArray<PPToken> values) {
    WriteStringf(file, "enum %.*s : uint32 {\n", name.length, name.text);

    for (int i = 0; i < values.count; i++) {
        WriteWhitespace(file, 4);
        WriteStringf(file, "%.*s = 1 << %d,\n", values[i].length, values[i].text, i);
    }

    WriteString(file, "};\n");

}

void WriteBitarray(FileHandle *file, PPToken name, DynamicArray<PPToken> values) {
    WriteStringf(file, "enum %.*s_Indices : uint32 {\n", name.length, name.text);

    for (int i = 0; i < values.count; i++) {
        WriteWhitespace(file, 4);
        WriteStringf(file, "%.*s,\n", values[i].length, values[i].text, i);
    }

    WriteString(file, "};\n");

    WriteStringf(file, "struct %.*s {\n", name.length, name.text);

    WriteWhitespace(file, 4);
    WriteStringf(file, "uint32 bits[%d];\n", ((values.count) / 32) + 1);
    WriteStringf(file, "};\n");
}

void WriteEnumStringValues(FileHandle *file, PPToken name, DynamicArray<PPToken> values) {
    WriteStringf(file, "const char *%.*s_StringValues[] = {\n", name.length, name.text);

    for (int i = 0; i < values.count; i++) {
        WriteWhitespace(file, 4);
        WriteStringf(file, "\"%.*s\",\n", values[i].length, values[i].text);
    }
    
    WriteString(file, "};\n");
};

void WriteTaggedUnion(FileHandle *file, PPToken name, DynamicArray<ParsedField> fields) {
    // Write TypeTag enum
    WriteStringf(file, "enum %.*s_TypeTag : uint32 {\n", name.length, name.text);

    for (int i = 0; i < fields.count; i++) {
        ParsedField field = fields[i];

        WriteWhitespace(file, 4);
        WriteStringf(file, "%.*s_%.*s,\n", name.length, name.text, field.name.length, field.name.text);
    }

    WriteStringf(file, "};\n\n");

    // Write TypeTag StringValues
    WriteStringf(file, "const char *%.*s_TypeTag_StringValues[] = {\n", name.length, name.text);

    for (int i = 0; i < fields.count; i++) {
        ParsedField field = fields[i];

        WriteWhitespace(file, 4);
        WriteStringf(file, "\"%.*s_%.*s\",\n", name.length, name.text, field.name.length, field.name.text);
    }

    WriteString(file, "};\n\n");


    // Write TaggedUnion struct
#if 0
    WriteStringf(file, "struct %.*s {\n", name.length, name.text);

    WriteWhitespace(file, 4);
    WriteStringf(file, "%.*s_TypeTag type;\n\n", name.length, name.text);

    WriteWhitespace(file, 4);
    WriteStringf(file, "union {\n");

    WriteWhitespace(file, 8);
    WriteStringf(file, "void *_data;\n");

    for (int i = 0; i < fields.count; i++) {
        ParsedField field = fields[i];

        WriteWhitespace(file, 8);
        if (field.type.isConst) {
            WriteString(file, "const ");
        }

        WriteBytes(file, field.type.typeName.text, field.type.typeName.length);

        if (field.type.templateTypes.count > 0) {
            WriteTemplateTypes(file, field.type.templateTypes);
        }
                
        WriteString(file, " ");
        for (int k = 0; k < field.type.pointerLevels; k++) {
            WriteString(file, "*");
        }

        WriteBytes(file, field.name.text, field.name.length);

        if (field.type.arrayLength.type != PPTokenType_Unknown) {
            WriteStringf(file, "[%d]", field.type.arrayLengthValue);
        }

        WriteString(file, ";\n");
    }

    WriteWhitespace(file, 4);
    WriteStringf(file, "};\n");

    WriteStringf(file, "};\n");
#endif
}

void WriteComponentNameTables(FileHandle *file, PreprocessorData *data) {
    WriteStringf(file, "static const char *ComponentTypeNameTable[] = {\n");
    WriteWhitespace(file, 4);
    WriteStringf(file, "NULL,\n");

    for (int i = 0; i < data->definedComponents.count; i++) {
        ParsedComponent component = data->definedComponents[i];
        WriteStringf(file, "    \"%.*sComponent\",\n", component.name.length, component.name.text);
    }

    WriteStringf(file, "};\n\n");


    WriteStringf(file, "static const char *ComponentHumanNameTable[] = {\n");
    WriteWhitespace(file, 4);
    WriteStringf(file, "NULL,\n");

    for (int i = 0; i < data->definedComponents.count; i++) {
        ParsedComponent component = data->definedComponents[i];
        WriteWhitespace(file, 4);
        WriteStringf(file, "\"%.*s\",\n", component.name.length, component.name.text);
    }

    WriteStringf(file, "};\n\n");
}


void WriteComponentList(const char *filepath, PreprocessorData *data) {
    FileHandle file;
    if (Plat.OpenFileForWrite(filepath, &file, 4096)) {
        FileClear(&file);

        WriteString(&file, "#define FOREACH_COMPONENT_ARGS(OP, ...) \\\n");

        for (int i = 0; i < data->definedComponents.count; i++) {
            ParsedComponent component = data->definedComponents[i];

            WriteWhitespace(&file, 4);
            WriteString(&file, "OP(");
            WriteBytes(&file, component.name.text, component.name.length);
            WriteString(&file, ", __VA_ARGS__) \\\n");
        }

        WriteString(&file, "\n");
        WriteString(&file, "#define FOREACH_COMPONENT(OP) \\\n");

        for (int i = 0; i < data->definedComponents.count; i++) {
            ParsedComponent component = data->definedComponents[i];

            WriteWhitespace(&file, 4);
            WriteString(&file, "OP(");
            WriteBytes(&file, component.name.text, component.name.length);
            WriteString(&file, ") \\\n");
        }

        WriteString(&file, "\n");

        WriteComponentNameTables(&file, data);


        // ComponentType_StringValues
        WriteStringf(&file, "const char *ComponentType_StringValues[] = {\n");
        WriteWhitespace(&file, 4);
        WriteStringf(&file, "\"ComponentType_Invalid\",\n");
        for (int i = 0; i < data->definedComponents.count; i++) {
            ParsedComponent component = data->definedComponents[i];

            WriteWhitespace(&file, 4);
            WriteStringf(&file, "\"ComponentType_%.*s\",\n", component.name.length, component.name.text);
        }
        WriteString(&file, "};\n\n");

        // ComponentType enum
        WriteStringf(&file, "enum ComponentType : uint32 {\n");
        WriteWhitespace(&file, 4);
        WriteStringf(&file, "ComponentType_Invalid,\n");
        for (int i = 0; i < data->definedComponents.count; i++) {
            ParsedComponent component = data->definedComponents[i];

            WriteWhitespace(&file, 4);
            WriteStringf(&file, "ComponentType_%.*s,\n", component.name.length, component.name.text);
        }
        WriteString(&file, "};\n\n");

        Plat.CloseFile(&file);
    }
}

void WriteScenesList(const char *sceneListPath, PreprocessorData *data) {
    FileHandle listFile;
    if (Plat.OpenFileForWrite(sceneListPath, &listFile, 4096)) {
        FileClear(&listFile);

        // Generate Scene Lists
        WriteString(&listFile, "#define FOREACH_SCENE_ARGS(OP, ...) \\\n");
        for (int i = 0; i < data->definedScenes.count; i++) {
            ParsedScene scene = data->definedScenes[i];

#if !ETC_DEBUG
            if (HasProperty(scene.properties, "Editor") ||
                HasProperty(scene.properties, "Test")) {
                continue;
            }
#endif

            WriteWhitespace(&listFile, 4);
            WriteString(&listFile, "OP(");
            WriteBytes(&listFile, scene.name.text, scene.name.length);
            WriteString(&listFile, ", __VA_ARGS__) \\\n");
        }

        WriteString(&listFile, "\n");
        WriteString(&listFile, "#define FOREACH_SCENE(OP) \\\n");
        for (int i = 0; i < data->definedScenes.count; i++) {
            ParsedScene scene = data->definedScenes[i];

#if !ETC_DEBUG
            if (HasProperty(scene.properties, "Editor") ||
                HasProperty(scene.properties, "Test")) {
                continue;
            }
#endif

            WriteWhitespace(&listFile, 4);
            WriteString(&listFile, "OP(");
            WriteBytes(&listFile, scene.name.text, scene.name.length);
            WriteString(&listFile, ") \\\n");
        }

        WriteString(&listFile, "\n");


        // Write SceneID enum
        WriteString(&listFile, "\n");
        WriteString(&listFile, "enum SceneID : uint32 {\n");
        WriteWhitespace(&listFile, 4);
        WriteString(&listFile, "Scene_None,\n");

        for (int i = 0; i < data->definedScenes.count; i++) {
            ParsedScene scene = data->definedScenes[i];

            WriteWhitespace(&listFile, 4);
            WriteStringf(&listFile, "Scene_%.*s,\n", scene.name.length, scene.name.text);
        }

        WriteString(&listFile, "};\n\n");

        WriteString(&listFile, "const char *SceneID_StringValues[] = {\n");
        WriteWhitespace(&listFile, 4);
        WriteString(&listFile, "\"Scene_None\",\n");

        for (int i = 0; i < data->definedScenes.count; i++) {
            ParsedScene scene = data->definedScenes[i];

            WriteWhitespace(&listFile, 4);
            WriteStringf(&listFile, "\"Scene_%.*s\",\n", scene.name.length, scene.name.text);
        }

        WriteString(&listFile, "};\n\n");


        Plat.CloseFile(&listFile);
    }
}

void WriteSerializableTypeHeader(FileHandle *file, PreprocessorData *data, DynamicArray<ParsedSerializable> serializableTypes) {

    // Persistent Function Handles
    WriteStringf(file, "\n");

    for (int i = 0; i < data->definedFunctionTypes.count; i++) {
        ParsedPersistentFunctionType func = data->definedFunctionTypes[i];
        WriteStringf(file, "struct %.*sHandle : FuncHandle {", func.name.length, func.name.text);
        WriteStringf(file, "};\n\n");

        WriteStringf(file, "struct %.*sDataPair {\n", func.name.length, func.name.text);
        WriteWhitespace(file, 4);
        WriteStringf(file, "%.*sHandle funcHandle;\n", func.name.length, func.name.text);
        WriteWhitespace(file, 4);
        WriteStringf(file, "SchemaData data;\n");
        WriteStringf(file, "};\n\n");
    }

    
    // Write Serializable Types
    for (int i = 0; i < serializableTypes.count; i++) {
        ParsedSerializable serializable = serializableTypes[i];

        switch (serializable.type) {
            case SerializableType_Struct: {
#if 0
                WriteStruct(file, serializable.name, serializable._struct.fields, NULL);
                WriteString(file, "\n");
#endif
            } break;

            case SerializableType_Enum: {
                WriteEnumStringValues(file, serializable.name, serializable._enum.values);
                WriteString(file, "\n");

                //WriteEnum(file, serializable.name, serializable._enum.values);
                //WriteString(file, "\n");
            } break;

            case SerializableType_Bitfield: {
                WriteEnumStringValues(file, serializable.name, serializable._bitfield.values);
                WriteString(file, "\n");

                WriteBitfield(file, serializable.name, serializable._bitfield.values);
                WriteString(file, "\n");
            } break;

            case SerializableType_Bitarray: {
                WriteEnumStringValues(file, serializable.name, serializable._bitarray.values);
                WriteString(file, "\n");

                WriteBitarray(file, serializable.name, serializable._bitarray.values);
                WriteString(file, "\n");
            } break;
                    
            case SerializableType_TaggedUnion: {
                WriteTaggedUnion(file, serializable.name, serializable._tagged_union.fields);
                WriteString(file, "\n");
            } break;
        }
    }
}

void WriteSchemaTypeDescription(FileHandle *file, SchemaTypeDescription type, uint32 spaces) {
    const char *fieldTypeName = GetSchemaDataTypeName(type.dataType);
    
    WriteWhitespace(file, spaces);
    if (type.dataType != SchemaDataType_Invalid) {
        WriteStringf(file, "type->dataType = SchemaDataType_%s;\n", fieldTypeName);
    }
    else {
        WriteStringf(file, "type->dataType = SchemaDataType_Invalid;\n");
    }

    WriteWhitespace(file, spaces);
    WriteStringf(file, "type->typeName = \"%s\";\n", type.typeName);

    WriteWhitespace(file, spaces);
    WriteStringf(file, "type->pointerLevels = %d;\n", type.pointerLevels);

    WriteWhitespace(file, spaces);
    WriteStringf(file, "type->isConst = %s;\n", type.isConst ? "true" : "false");

    WriteWhitespace(file, spaces);
    WriteStringf(file, "type->arrayLength = %d;\n", type.arrayLength);

    WriteWhitespace(file, spaces);
    WriteStringf(file, "type->maxStringLength = %d;\n", type.maxStringLength);

    WriteWhitespace(file, spaces);
    if (type.dataType == SchemaDataType_DynamicArray) {
        WriteStringf(file, "type->sizeInBytes = sizeof(DynamicArray");

        WriteTemplateTypes(file, type.templateTypes);
    }
    else if (type.dataType == SchemaDataType_ChunkedArray) {
        WriteStringf(file, "type->sizeInBytes = sizeof(ChunkedArray");

        WriteTemplateTypes(file, type.templateTypes);
    }
    else if (type.dataType == SchemaDataType_HashTable) {
        WriteStringf(file, "type->sizeInBytes = sizeof(HashTable");

        WriteTemplateTypes(file, type.templateTypes);
    }
    else if (type.dataType == SchemaDataType_String) {
        WriteStringf(file, "type->sizeInBytes = sizeof(char\n");
    }
    else if (type.typeName != NULL) {
        WriteStringf(file, "type->sizeInBytes = sizeof(%s", type.typeName);
    }
    else {
        WriteStringf(file, "type->sizeInBytes = sizeof(%s", fieldTypeName);
    }

    if (type.pointerLevels > 0) {
        WriteStringf(file, " ");

        for (int i = 0; i < type.pointerLevels; i++) {
            WriteStringf(file, "*");
        }
    }

    WriteStringf(file, ");\n");


    if (type.templateTypes.count > 0) {
        WriteStringf(file, "\n");

        WriteWhitespace(file, spaces);
        WriteStringf(file, "type->templateTypes = MakeDynamicArray<SchemaTypeDescription>(%d);\n", type.templateTypes.count);

        for (int k = 0; k < type.templateTypes.count; k++) {
            WriteWhitespace(file, spaces);
            WriteStringf(file, "{\n");

            WriteWhitespace(file, spaces + 4);
            WriteStringf(file, "SchemaTypeDescription *type_ = type;\n");

            WriteWhitespace(file, spaces + 4);
            WriteStringf(file, "type = PushBackPtr(&type->templateTypes);\n", k);
            WriteSchemaTypeDescription(file, type.templateTypes[k], spaces + 4);

            WriteWhitespace(file, spaces + 4);
            WriteStringf(file, "type = type_;\n");
    
            WriteWhitespace(file, spaces);
            WriteStringf(file, "}\n");

            if (k < type.templateTypes.count - 1) {
                WriteStringf(file, "\n");
            }
        }
    }
    else {
        WriteWhitespace(file, spaces);
        WriteStringf(file, "type->templateTypes = {};\n");
    }


    WriteWhitespace(file, spaces);
    WriteStringf(file, "type->canSerialize = %s;\n", type.canSerialize ? "true" : "false");
}

void WriteGenerateSchemasFunction(FileHandle *file, DynamicArray<Schema> schemas) {
    WriteString(file, "void RegisterSchemas(SchemaTables *tables) {\n");

    for (int i = 0; i < schemas.count; i++) {
        Schema schema = schemas[i];

        WriteWhitespace(file, 4);
        WriteStringf(file, "// %s\n", schema.name);
        WriteWhitespace(file, 4);
        WriteStringf(file, "{\n");

        WriteWhitespace(file, 8);
        WriteStringf(file, "Schema *schema = GetSchemaNamed_(tables, \"%s\");\n", schema.name);

        WriteWhitespace(file, 8);
        WriteStringf(file, "schema->name = \"%s\";\n", schema.name);
        WriteWhitespace(file, 8);
        WriteStringf(file, "schema->editable = %s;\n", schema.editable ? "true" : "false");
        WriteWhitespace(file, 8);
        WriteStringf(file, "schema->serializable = %s;\n", schema.serializable ? "true" : "false");
        
        switch (schema.type) {
            case SchemaType_Struct: {
                WriteWhitespace(file, 8);
                WriteStringf(file, "schema->type = SchemaType_Struct;\n");
                WriteWhitespace(file, 8);
                WriteStringf(file, "schema->sizeInBytes = sizeof(%s);\n", schema.name);

                WriteWhitespace(file, 8);
                WriteStringf(file, "schema->_struct.isAsset = %s;\n", schema._struct.isAsset ? "true" : "false");

                if (schema._struct.fields.count > 0) {
                    WriteWhitespace(file, 8);
                    WriteStringf(file, "schema->_struct.fields = MakeDynamicArray<SchemaField>(%d);\n\n", schema._struct.fields.count);

                    for (int j = 0; j < schema._struct.fields.count; j++) {
                        SchemaField field = schema._struct.fields[j];

                        WriteWhitespace(file, 8);
                        WriteStringf(file, "{\n");

                        WriteWhitespace(file, 12);
                        WriteStringf(file, "SchemaField *field = PushBackPtr(&schema->_struct.fields);\n\n");

                        WriteWhitespace(file, 12);
                        WriteStringf(file, "field->name = \"%s\";\n", field.name);
                        WriteWhitespace(file, 12);
                        WriteStringf(file, "field->byteOffset = FIELD_OFFSET(%s, %s);\n", schema.name, field.name);

                        WriteWhitespace(file, 12);
                        WriteStringf(file, "field->editable = %s;\n", field.editable ? "true" : "false");
                        WriteWhitespace(file, 12);
                        WriteStringf(file, "field->serializable = %s;\n", field.serializable ? "true" : "false");

                        WriteWhitespace(file, 12);
                        WriteStringf(file, "{\n");
                        WriteWhitespace(file, 16);
                        WriteStringf(file, "SchemaTypeDescription *type = &field->type;\n");
                        WriteSchemaTypeDescription(file, field.type, 16);

                        WriteWhitespace(file, 12);
                        WriteStringf(file, "}\n");

                        WriteWhitespace(file, 8);
                        WriteStringf(file, "}\n\n");
                    }
                }
                else {
                    WriteWhitespace(file, 8);
                    WriteStringf(file, "schema->_struct.fields = {};\n\n");
                }
            } break;

            case SchemaType_Enum: {
                WriteWhitespace(file, 8);
                WriteStringf(file, "schema->type = SchemaType_Enum;\n");
                WriteWhitespace(file, 8);
                WriteStringf(file, "schema->sizeInBytes = sizeof(%s);\n", schema.name);

                WriteWhitespace(file, 8);
                WriteStringf(file, "schema->_enum.count = ENUM_VALUECOUNT(%s);\n", schema.name);
                WriteWhitespace(file, 8);
                WriteStringf(file, "schema->_enum.stringValues = ENUM_STRINGTABLE(%s);\n\n", schema.name);
            } break;

            case SchemaType_Bitfield: {
                WriteWhitespace(file, 8);
                WriteStringf(file, "schema->type = SchemaType_Bitfield;\n");
                WriteWhitespace(file, 8);
                WriteStringf(file, "schema->sizeInBytes = sizeof(%s);\n", schema.name);

                WriteWhitespace(file, 8);
                WriteStringf(file, "schema->_bitfield.count = ENUM_VALUECOUNT(%s);\n", schema.name);
                WriteWhitespace(file, 8);
                WriteStringf(file, "schema->_bitfield.stringValues = ENUM_STRINGTABLE(%s);\n\n", schema.name);
            } break;
                
            case SchemaType_Bitarray: {
                WriteWhitespace(file, 8);
                WriteStringf(file, "schema->type = SchemaType_Bitarray;\n");
                WriteWhitespace(file, 8);
                WriteStringf(file, "schema->sizeInBytes = (((ENUM_VALUECOUNT(%s) / 32) + 1) * sizeof(uint32));\n", schema.name);

                WriteWhitespace(file, 8);
                WriteStringf(file, "schema->_bitarray.count = ENUM_VALUECOUNT(%s);\n", schema.name);
                WriteWhitespace(file, 8);
                WriteStringf(file, "schema->_bitarray.stringValues = ENUM_STRINGTABLE(%s);\n\n", schema.name);
            } break;

            case SchemaType_TaggedUnion: {
                WriteWhitespace(file, 8);
                WriteStringf(file, "schema->type = SchemaType_TaggedUnion;\n");
                WriteWhitespace(file, 8);
                WriteStringf(file, "schema->sizeInBytes = sizeof(%s);\n", schema.name);

                WriteStringf(file, "schema->_tagged_union.typeTagStringValues = ENUM_STRINGTABLE(%s_TypeTag);\n\n", schema.name);


                if (schema._tagged_union.fields.count > 0) {
                    WriteWhitespace(file, 8);
                    WriteStringf(file, "schema->_tagged_union.fields = MakeDynamicArray<SchemaField>(%d);\n\n", schema._tagged_union.fields.count);

                    for (int j = 0; j < schema._tagged_union.fields.count; j++) {
                        SchemaField field = schema._tagged_union.fields[j];

                        WriteWhitespace(file, 8);
                        WriteStringf(file, "{\n");

                        WriteWhitespace(file, 12);
                        WriteStringf(file, "SchemaField *field = PushBackPtr(&schema->_tagged_union.fields);\n\n");

                        WriteWhitespace(file, 12);
                        WriteStringf(file, "field->name = \"%s\";\n", field.name);
                        WriteWhitespace(file, 12);
                        WriteStringf(file, "field->byteOffset = FIELD_OFFSET(%s, _data.%s);\n", schema.name, field.name);

                        WriteWhitespace(file, 12);
                        WriteStringf(file, "field->editable = %s;\n", field.editable ? "true" : "false");
                        WriteWhitespace(file, 12);
                        WriteStringf(file, "field->serializable = %s;\n", field.serializable ? "true" : "false");

                        WriteWhitespace(file, 12);
                        WriteStringf(file, "{\n");
                        WriteWhitespace(file, 16);
                        WriteStringf(file, "SchemaTypeDescription *type = &field->type;\n");

                        WriteSchemaTypeDescription(file, field.type, 16);

                        WriteWhitespace(file, 12);
                        WriteStringf(file, "}\n");

                        WriteWhitespace(file, 8);
                        WriteStringf(file, "}\n\n");
                    }
                }
                else {
                    WriteWhitespace(file, 8);
                    WriteStringf(file, "schema->_tagged_union.fields = {};\n\n");
                }
            } break;

            case SchemaType_PersistentFunctionHandle: {
                WriteWhitespace(file, 8);
                WriteStringf(file, "schema->type = SchemaType_PersistentFunctionHandle;\n");

                uint32 nameLength = StringLength(schema.name) - StringLength("Handle");
                WriteWhitespace(file, 8);
                WriteStringf(file, "schema->_func_handle.funcType = PersistentFuncType_%.*s;\n", nameLength, schema.name);

                WriteWhitespace(file, 8);
                WriteStringf(file, "schema->sizeInBytes = sizeof(%s);\n", schema.name);
            } break;

            case SchemaType_PersistentFunctionDataPair: {
                WriteWhitespace(file, 8);
                WriteStringf(file, "schema->type = SchemaType_PersistentFunctionDataPair;\n");

                uint32 nameLength = StringLength(schema.name) - StringLength("DataPair");
                WriteWhitespace(file, 8);
                WriteStringf(file, "schema->sizeInBytes = sizeof(%s);\n", schema.name);
                
                for (int j = 0; j < 2; j++) {
                    SchemaField field = schema._func_data_pair.fields[j];

                    WriteWhitespace(file, 8);
                    WriteStringf(file, "{\n");

                    WriteWhitespace(file, 12);
                    WriteStringf(file, "SchemaField *field = &schema->_func_data_pair.fields[%d];\n\n", j);

                    WriteWhitespace(file, 12);
                    WriteStringf(file, "field->name = \"%s\";\n", field.name);
                    WriteWhitespace(file, 12);
                    WriteStringf(file, "field->byteOffset = FIELD_OFFSET(%s, %s);\n", schema.name, field.name);

                    WriteWhitespace(file, 12);
                    WriteStringf(file, "field->editable = %s;\n", field.editable ? "true" : "false");
                    WriteWhitespace(file, 12);
                    WriteStringf(file, "field->serializable = %s;\n", field.serializable ? "true" : "false");

                    WriteWhitespace(file, 12);
                    WriteStringf(file, "{\n");
                    WriteWhitespace(file, 16);
                    WriteStringf(file, "SchemaTypeDescription *type = &field->type;\n");
                    WriteSchemaTypeDescription(file, field.type, 16);

                    WriteWhitespace(file, 12);
                    WriteStringf(file, "}\n");

                    WriteWhitespace(file, 8);
                    WriteStringf(file, "}\n\n");
                }
            } break;

            case SchemaType_AssetHandle: {
                WriteWhitespace(file, 8);
                WriteStringf(file, "schema->type = SchemaType_AssetHandle;\n");
                WriteWhitespace(file, 8);
                WriteStringf(file, "schema->sizeInBytes = sizeof(%s);\n", schema.name);

                uint32 nameLength = StringLength(schema.name) - StringLength("Handle");
                
                WriteWhitespace(file, 8);
                WriteStringf(file, "schema->_assetHandle.assetTypeName = \"AssetType_%.*s\";\n", nameLength, schema.name);
                WriteWhitespace(file, 8);
                WriteStringf(file, "schema->_assetHandle.assetStructName = \"%.*s\";\n", nameLength, schema.name);

            } break;
        }

        WriteWhitespace(file, 4);
        WriteStringf(file, "}\n");

        if (i < schemas.count - 1) {
            WriteStringf(file, "\n");
        }
    }

    WriteString(file, "}\n\n");
}

void WriteFunctionDeclaration(FileHandle *file, ParsedFunction *function) {
    //WriteStringf(file, "%s %.*s;\n", typedefName, functions[i].name.length, functions[i].name.text);

    WriteStringf(file, "%.*s %.*s", function->returnType.typeName.length, function->returnType.typeName.text, function->name.length, function->name.text);

    WriteStringf(file, "(");

    WriteFunctionArguments(file, function->arguments);

    WriteStringf(file, ");\n");
}

void WriteFunctionTable(FileHandle *file, PreprocessorData *data, const char *typedefName, ParsedFunction *functions, uint32 count, b32 required, b32 editorOnly, const char *zeroValue = NULL) {

    if (editorOnly) {
        WriteStringf(file, "#if ETC_DEBUG\n");
    }

    // Forward declare
    for (int i = 0; i < count; i++) {
        if (functions[i].name.text != NULL) {
            WriteFunctionDeclaration(file, &functions[i]);
        }
    }
    WriteStringf(file, "\n");

    WriteStringf(file, "static %s *%sTable[] = {\n", typedefName, typedefName);

    if (zeroValue) {
        WriteWhitespace(file, 4);
        WriteStringf(file, "(%s *)%s,\n", typedefName, zeroValue);
    }
    else {
        // @NOTE: assumes that 0 represents an Invalid type
        WriteWhitespace(file, 4);
        WriteStringf(file, "NULL,\n");
    }

    for (int i = 0; i < count; i++) {
        if (functions[i].name.text != NULL) {
            WriteWhitespace(file, 4);
            WriteStringf(file, "(%s *)%.*s,\n", typedefName, functions[i].name.length, functions[i].name.text);
        }
        else {
            if (required) {
                PreprocessorSetError(data, PreprocessorError_MissingFunction,
                                     "Missing required %s function for component types '%.*s'\n",
                                     typedefName, data->definedComponents[i].name.length, data->definedComponents[i].name.text);
                return;
            }

            WriteWhitespace(file, 4);
            WriteStringf(file, "NULL,\n");
        }
    }

    if (editorOnly) {
        WriteStringf(file, "};\n#endif\n\n");
    }
    else {
        WriteStringf(file, "};\n\n");
    }
}

void WriteFunctionTables(FileHandle *file, PreprocessorData *data) {

    WriteStringf(file, "\n");

    WriteFunctionTable(file, data, "AddComponentFunc", data->addComponentFunctions, data->definedComponents.count, true, false);
    WriteFunctionTable(file, data, "RemoveComponentFunc", data->removeComponentFunctions, data->definedComponents.count, true, false);
    WriteFunctionTable(file, data, "GetComponentValueFunc", data->getComponentValueFunctions, data->definedComponents.count, true, false);
    WriteFunctionTable(file, data, "SetComponentValueFunc", data->setComponentValueFunctions, data->definedComponents.count, true, false);
    WriteFunctionTable(file, data, "TickSystemFunc", data->tickSystemFunctions, data->definedComponents.count, false, false);
    WriteFunctionTable(file, data, "RaycastSystemFunc", data->raycastSystemFunctions, data->definedComponents.count, false, false);

    WriteFunctionTable(file, data, "SceneInitFunc", data->sceneInitFunctions, data->definedScenes.count, true, false, "SceneNoneInit");
    WriteFunctionTable(file, data, "SceneDeinitFunc", data->sceneDeinitFunctions, data->definedScenes.count, true, false, "SceneNoneDeinit");
    WriteFunctionTable(file, data, "SceneUpdateFunc", data->sceneUpdateFunctions, data->definedScenes.count, true, false, "SceneNoneUpdate");
}

void WriteDefinedFunctionTypes(FileHandle *file, PreprocessorData *data) {

    WriteStringf(file, "\n");

    WriteString(file, "enum PersistentFuncType : uint32 {\n");
    WriteWhitespace(file, 4);
    WriteString(file, "PersistentFuncType_Invalid,\n");
    for (int i = 0; i < data->definedFunctionTypes.count; i++) {
        ParsedPersistentFunctionType func = data->definedFunctionTypes[i];
        WriteWhitespace(file, 4);
        WriteStringf(file, "PersistentFuncType_%.*s,\n", func.name.length, func.name.text);
    }
    WriteString(file, "};\n\n");

    WriteString(file, "const char *PersistentFuncType_StringValues[] = {\n");
    WriteWhitespace(file, 4);
    WriteString(file, "\"PersistentFuncType_Invalid\",\n");
    for (int i = 0; i < data->definedFunctionTypes.count; i++) {
        ParsedPersistentFunctionType func = data->definedFunctionTypes[i];
        WriteWhitespace(file, 4);
        WriteStringf(file, "\"PersistentFuncType_%.*s\",\n", func.name.length, func.name.text);
    }
    WriteString(file, "};\n\n");
    
    for (int i = 0; i < data->definedFunctionTypes.count; i++) {
        ParsedPersistentFunctionType func = data->definedFunctionTypes[i];

        WriteStringf(file, "typedef %.*s %.*sType_(", func.returnType.typeName.length, func.returnType.typeName.text,
                     func.name.length, func.name.text);

        WriteFunctionArguments(file, func.arguments);

        
        WriteStringf(file, ");");
        WriteStringf(file, "\n\n");

        WriteStringf(file, "inline %.*sHandle Get%.*sHandle(%.*sType_ *functionPtr) {\n", func.name.length, func.name.text, func.name.length, func.name.text, func.name.length, func.name.text);

        WriteWhitespace(file, 4);
        WriteStringf(file, "%.*sHandle handle;\n", func.name.length, func.name.text);

        WriteWhitespace(file, 4);
        WriteStringf(file, "handle.funcType = PersistentFuncType_%.*s;\n", func.name.length, func.name.text);
        
        WriteWhitespace(file, 4);
        WriteStringf(file, "handle.tableIndex = GetPersistentFunctionIndex(&Game.persistentFunctionTable, (void *)functionPtr, PersistentFuncType_%.*s);\n", func.name.length, func.name.text);

        WriteWhitespace(file, 4);
        WriteStringf(file, "return handle;\n");
        WriteStringf(file, "}\n\n");


        WriteStringf(file, "inline %.*s CallFunc(%.*sHandle handle", func.returnType.typeName.length, func.returnType.typeName.text, func.name.length, func.name.text);

        if (func.arguments.count > 0) {
            WriteStringf(file, ", ");
        }

        WriteFunctionArguments(file, func.arguments);

        WriteStringf(file, ") {\n");

        WriteWhitespace(file, 4);
        WriteStringf(file, "%.*sType_ *functionPtr = (%.*sType_ *)GetPersistentFunctionPtr(&Game.persistentFunctionTable, handle.tableIndex, handle.funcType);\n", func.name.length, func.name.text, func.name.length, func.name.text);

        WriteWhitespace(file, 4);
        WriteStringf(file, "return functionPtr(");

        for (int j = 0; j < func.arguments.count; j++) {
            ParsedFunctionArgument arg = func.arguments[j];
            WriteStringf(file, "%.*s", arg.name.length, arg.name.text);

            if (j < func.arguments.count - 1) {
                WriteStringf(file, ", ");
            }
        }

        WriteStringf(file, ");\n");

        WriteStringf(file, "}\n");

        WriteStringf(file, "\n\n");
    }

    WriteStringf(file, "// Forward Declare Persistent Functions\n");
    for (int i = 0; i < data->definedPersistentFunctions.count; i++) {
        ParsedFunction func = data->definedPersistentFunctions[i];
        ASSERT(func.isPersistent);

        WriteStringf(file, "%.*sType_ %.*s;\n", func.persistentFunctionType.length, func.persistentFunctionType.text, func.name.length, func.name.text);
    }
}

void WritePersistentFunctionTable(FileHandle *file, PreprocessorData *data) {
    WriteStringf(file, "void RegisterPersistentFunctions(PersistentFunctionTable *table) {\n");

    for (int i = 0; i < data->definedPersistentFunctions.count; i++) {
        ParsedFunction func = data->definedPersistentFunctions[i];
        WriteWhitespace(file, 4);
        WriteStringf(file, "RegisterPersistentFunction(table, \"%.*s\", (void *)&%.*s, PersistentFuncType_%.*s",
                     func.name.length, func.name.text, func.name.length, func.name.text, func.persistentFunctionType.length, func.persistentFunctionType.text);

        if (func.requiredSchemaDataType.length > 0) {
            WriteStringf(file, ", \"%.*s\");\n",  func.requiredSchemaDataType.length, func.requiredSchemaDataType.text); 
        }
        else {
            WriteStringf(file, ", NULL);\n");
        }
                     
    }

    WriteStringf(file, "}\n");
}

void WriteAssetList(FileHandle *file, PreprocessorData *data) {
    // FOREACH macros
    WriteStringf(file, "#define FOREACH_ASSET_ARGS(OP, ...) \\\n");

    for (int i = 0; i < data->definedAssets.count; i++) {
        ParsedSerializable asset = data->definedAssets[i];

        WriteWhitespace(file, 4);
        WriteStringf(file, "OP(%.*s, __VA_ARGS__) \\\n", asset.name.length, asset.name.text);
    }

    WriteStringf(file, "\n");


    WriteStringf(file, "#define FOREACH_ASSET(OP) \\\n");

    for (int i = 0; i < data->definedAssets.count; i++) {
        ParsedSerializable asset = data->definedAssets[i];

        WriteWhitespace(file, 4);
        WriteStringf(file, "OP(%.*s) \\\n", asset.name.length, asset.name.text);
    }

    WriteStringf(file, "\n\n");


    // AssetNameTable
    WriteStringf(file, "static const char *AssetNameTable[] = {\n");
    WriteWhitespace(file, 4);
    WriteStringf(file, "NULL,\n");

    for (int i = 0; i < data->definedAssets.count; i++) {
        ParsedSerializable asset = data->definedAssets[i];

        WriteWhitespace(file, 4);
        WriteStringf(file, "\"%.*s\",\n", asset.name.length, asset.name.text);
    }

    WriteStringf(file, "};\n\n");


    // AssetType_StringValues
    WriteStringf(file, "const char *AssetType_StringValues[] = {\n");
    WriteWhitespace(file, 4);
    WriteStringf(file, "\"AssetType_Invalid\",\n");
    for (int i = 0; i < data->definedAssets.count; i++) {
        ParsedSerializable asset = data->definedAssets[i];

        WriteWhitespace(file, 4);
        WriteStringf(file, "\"AssetType_%.*s\",\n", asset.name.length, asset.name.text);
    }
    WriteStringf(file, "};\n\n");

    // AssetType enum
    WriteStringf(file, "enum AssetType : uint32 {\n");
    WriteWhitespace(file, 4);
    WriteStringf(file, "AssetType_Invalid,\n");
    for (int i = 0; i < data->definedAssets.count; i++) {
        ParsedSerializable asset = data->definedAssets[i];

        WriteWhitespace(file, 4);
        WriteStringf(file, "AssetType_%.*s,\n", asset.name.length, asset.name.text);
    }
    WriteStringf(file, "};\n\n");


    // AssetHandles
    for (int i = 0; i < data->definedAssets.count; i++) {
        ParsedSerializable asset = data->definedAssets[i];

        WriteStringf(file, "struct %.*sHandle : AssetHandle {\n", asset.name.length, asset.name.text);
        WriteStringf(file, "};\n\n", asset.name.length, asset.name.text);
    }


    // Forward Declare Asset Types
    for (int i = 0; i < data->definedAssets.count; i++) {
        ParsedSerializable asset = data->definedAssets[i];
        WriteStringf(file, "struct %.*s;\n", asset.name.length, asset.name.text);
    }
    
    WriteStringf(file, "\n\n");


    // Load*/LoadProcedural*(), Get*()
    for (int i = 0; i < data->definedAssets.count; i++) {
        ParsedSerializable asset = data->definedAssets[i];

        WriteStringf(file, "inline %.*sHandle Load%.*s(AssetDB *assets, const char *name, b32 forceReload = false) {\n", asset.name.length, asset.name.text, asset.name.length, asset.name.text);
        WriteWhitespace(file, 4);
        WriteStringf(file, "AssetHandle handle = LoadAsset(assets, AssetType_%.*s, name, forceReload);\n", asset.name.length, asset.name.text);
        WriteWhitespace(file, 4);
        WriteStringf(file, "return *(%.*sHandle *)&handle;\n", asset.name.length, asset.name.text);
        WriteStringf(file, "}\n\n");


        WriteStringf(file, "inline %.*sHandle LoadProcedural%.*s(AssetDB *assets, const char *name, %.*s *data) {\n", asset.name.length, asset.name.text, asset.name.length, asset.name.text, asset.name.length, asset.name.text);
        WriteWhitespace(file, 4);
        WriteStringf(file, "ASSERT(data != NULL);\n");
        WriteWhitespace(file, 4);
        WriteStringf(file, "AssetHandle handle = LoadProceduralAsset(assets, AssetType_%.*s, name, data);\n", asset.name.length, asset.name.text, asset.name.length, asset.name.text);
        WriteWhitespace(file, 4);
        WriteStringf(file, "return *(%.*sHandle *)&handle;\n", asset.name.length, asset.name.text);
        WriteStringf(file, "}\n\n");

        
        WriteStringf(file, "inline %.*sHandle SyncLoadProcedural%.*s(AssetDB *assets, const char *name, %.*s *data) {\n", asset.name.length, asset.name.text, asset.name.length, asset.name.text, asset.name.length, asset.name.text);
        WriteWhitespace(file, 4);
        WriteStringf(file, "ASSERT(data != NULL);\n");
        WriteWhitespace(file, 4);
        WriteStringf(file, "AssetHandle handle = SyncLoadProceduralAsset(assets, AssetType_%.*s, name, data);\n", asset.name.length, asset.name.text, asset.name.length, asset.name.text);
        WriteWhitespace(file, 4);
        WriteStringf(file, "return *(%.*sHandle *)&handle;\n", asset.name.length, asset.name.text);
        WriteStringf(file, "}\n\n");

        WriteStringf(file, "inline %.*sHandle SyncLoad%.*s(AssetDB *assets, const char *name) {\n", asset.name.length, asset.name.text, asset.name.length, asset.name.text);
        WriteWhitespace(file, 4);
        WriteStringf(file, "AssetHandle handle = SyncLoadAsset(assets, AssetType_%.*s, name);\n", asset.name.length, asset.name.text);
        WriteWhitespace(file, 4);
        WriteStringf(file, "return *(%.*sHandle *)&handle;\n", asset.name.length, asset.name.text);
        WriteStringf(file, "}\n\n");


        WriteStringf(file, "inline %.*s *Get%.*s(AssetDB *assets, %.*sHandle handle) {\n", asset.name.length, asset.name.text, asset.name.length, asset.name.text, asset.name.length, asset.name.text);
        WriteWhitespace(file, 4);
        WriteStringf(file, "%.*s *result = (%.*s *)GetAssetDataData(assets, handle.asset, AssetType_%.*s);\n", asset.name.length, asset.name.text, asset.name.length, asset.name.text, asset.name.length, asset.name.text);
        WriteWhitespace(file, 4);
        WriteStringf(file, "return result;\n");
        WriteStringf(file, "}\n\n");
    }
}

void WriteAssetTables(FileHandle *file, PreprocessorData *data) {

    // RegisterAssetTypeExtensions()
    WriteStringf(file, "void RegisterAssetTypeExtensions(AssetDB *assets) {\n");

    for (int i = 0; i < data->definedAssets.count; i++) {
        ParsedSerializable asset = data->definedAssets[i];
        DynamicArray<PPToken> extensions = data->definedAssetExtensions[i];

        WriteStringf(file, "\n");
        WriteWhitespace(file, 4);
        WriteStringf(file, "// %.*s\n", asset.name.length, asset.name.text);

        for (int j = 0; j < extensions.count; j++) {

            WriteWhitespace(file, 4);
            WriteStringf(file, "PushBack(&assets->assetTypeExtensions[AssetType_%.*s], %.*s);\n", asset.name.length, asset.name.text, extensions[j].length, extensions[j].text);
        }
    }

    WriteStringf(file, "}\n\n");


    WriteFunctionTable(file, data, "AssetLoadFileFunc", data->assetLoadFileFunctions, data->definedAssets.count, false, false);
    WriteFunctionTable(file, data, "AssetLoadMainThreadFunc", data->assetLoadMainThreadFunctions, data->definedAssets.count, false, false);
    WriteFunctionTable(file, data, "AssetUnloadFileFunc", data->assetUnloadFileFunctions, data->definedAssets.count, false, false);
}

void WriteTags(FileHandle *file, PreprocessorData *data) {
    WriteStringf(file, "enum Tag : uint32 {\n");

    WriteWhitespace(file, 4);
    WriteStringf(file, "Tag_Invalid,\n");

    for (int i = 0; i < data->definedTags.count; i++) {
        WriteWhitespace(file, 4);
        WriteStringf(file, "Tag_%.*s,\n", data->definedTags[i].name.length, data->definedTags[i].name.text);
    }
    WriteStringf(file, "};\n");
    WriteStringf(file, "\n");

    
    WriteStringf(file, "const char *Tag_StringValues[] = {\n");

    WriteWhitespace(file, 4);
    WriteStringf(file, "\"Tag_Invalid\",\n");

    for (int i = 0; i < data->definedTags.count; i++) {
        WriteWhitespace(file, 4);
        WriteStringf(file, "\"Tag_%.*s\",\n", data->definedTags[i].name.length, data->definedTags[i].name.text);
    }

    WriteStringf(file, "};\n");
    WriteStringf(file, "\n");
}

void WriteInputActions(FileHandle *file, PreprocessorData *data) {
    WriteStringf(file, "enum InputActionType : uint32 {\n");
    WriteWhitespace(file, 4);
    WriteStringf(file, "InputAction_Invalid,\n");
    
    for (int i = 0; i < data->definedActions.count; i++) {
        WriteWhitespace(file, 4);
        WriteStringf(file, "InputAction_%.*s,\n", data->definedActions[i].name.length, data->definedActions[i].name.text);
    }

    WriteStringf(file, "};\n");
    WriteStringf(file, "\n");


    WriteStringf(file, "const char *InputActionType_StringValues[] = {\n");
    WriteWhitespace(file, 4);
    WriteStringf(file, "\"InputAction_Invalid\",\n");

    for (int i = 0; i < data->definedActions.count; i++) {
        WriteWhitespace(file, 4);
        WriteStringf(file, "\"InputAction_%.*s\",\n", data->definedActions[i].name.length, data->definedActions[i].name.text);
    }
    WriteStringf(file, "};\n");
    WriteStringf(file, "\n");
    

    WriteStringf(file, "inline void SetBindingTypes(InputActionBindingType *bindingTypes) {\n");
    for (int i = 0; i < data->definedActions.count; i++) {
        WriteWhitespace(file, 4);
        WriteStringf(file, "bindingTypes[InputAction_%.*s] = ", data->definedActions[i].name.length, data->definedActions[i].name.text);
        if (data->definedActions[i].analogue) {
            WriteStringf(file, "InputActionBindingType_Analogue;\n");
        }
        else {
            WriteStringf(file, "InputActionBindingType_Discrete;\n");
        }
    }

    WriteStringf(file, "}\n");
}

