
// Tokenization
struct Tokenizer {
    FileHandle filehandle;
    const char *filePath;

    char *buffer;

    uint32 index;
    uint32 line;
    uint32 column;
};

b32 IsAlpha(char c) {
    return (((c >= 'a') && (c <= 'z')) ||
            ((c >= 'A') && (c <= 'Z')));
}

b32 IsDigit(char c) {
    return ((c >= '0') && (c <= '9'));
}

b32 AtEndOfFile(Tokenizer *tokenizer) {
    b32 result = tokenizer->index >= tokenizer->filehandle.fileSize;
    return result;
}

b32 ConsumeAnyCharacter(Tokenizer *tokenizer) {

    b32 result = false;

    if (!AtEndOfFile(tokenizer)) {
        result = true;
        tokenizer->index++;
        tokenizer->column++;
    }

    return result;
}

b32 ConsumeCharacter(Tokenizer *tokenizer, char c) {
    b32 result = false;

    if (!AtEndOfFile(tokenizer) && 
        tokenizer->buffer[tokenizer->index] == c) {
        result = true;
        tokenizer->index++;
        tokenizer->column++;
    }

    return result;
}

b32 ConsumeCharacterSequence(Tokenizer *tokenizer, const char *str) {
    b32 result = true;

    uint32 startIndex = tokenizer->index;
    uint32 startColumn = tokenizer->column;
    uint32 startLine = tokenizer->line;

    char *at = (char *)str;
    while (*at != '\0') {
        if (ConsumeCharacter(tokenizer, *at)) {
            if (*at == '\n') {
                tokenizer->column = 1;
                tokenizer->line++;
            }
            at++;
        }
        else {
            result = false;
            break;
        }
    }

    if (!result) {
        tokenizer->index = startIndex;
        tokenizer->column = startColumn;
        tokenizer->line = startLine;
    }

    return result;
}

b32 ConsumeAlphaCharacter(Tokenizer *tokenizer) {
    b32 result = false;

    if (!AtEndOfFile(tokenizer) && IsAlpha(tokenizer->buffer[tokenizer->index])) {
        result = true;
        tokenizer->index++;
        tokenizer->column++;
    }

    return result;
}

b32 ConsumeNumericCharacter(Tokenizer *tokenizer) {
    b32 result = false;

    if (!AtEndOfFile(tokenizer) && IsDigit(tokenizer->buffer[tokenizer->index])) {
        result = true;
        tokenizer->index++;
        tokenizer->column++;
    }

    return result;
}

b32 ConsumeAlphaNumericCharacter(Tokenizer *tokenizer) {
    b32 result = ConsumeAlphaCharacter(tokenizer) || ConsumeNumericCharacter(tokenizer);
    return result;
}

b32 ConsumeWhiteSpace(Tokenizer *tokenizer) {
    b32 result = (ConsumeCharacter(tokenizer, ' ') ||
                  ConsumeCharacter(tokenizer, '\t') ||
                  ConsumeCharacter(tokenizer, '\f') ||
                  ConsumeCharacter(tokenizer, '\v'));

    return result;
}

b32 ConsumeEndOfLine(Tokenizer *tokenizer) {
    b32 result = false;

    if (ConsumeCharacterSequence(tokenizer, "\r\n") ||
        ConsumeCharacter(tokenizer, '\n')) {

        tokenizer->column = 1;
        tokenizer->line++;
        result = true;
    }

    return result;
}

void ConsumeWhitespaceAndComments(Tokenizer *tokenizer) {
    while (true) {
        if (ConsumeWhiteSpace(tokenizer)) {
        }
        else if (ConsumeEndOfLine(tokenizer)) {
        }
        else if (ConsumeCharacterSequence(tokenizer, "//")) {
            while (!AtEndOfFile(tokenizer) && !ConsumeEndOfLine(tokenizer)) {
                ConsumeAnyCharacter(tokenizer);
            }
        }
        else if (ConsumeCharacterSequence(tokenizer, "/*")) {
            while (!AtEndOfFile(tokenizer) && !ConsumeCharacterSequence(tokenizer, "*/")) {
                if (!ConsumeEndOfLine(tokenizer)) {
                    ConsumeAnyCharacter(tokenizer);
                }
            }
        }
        else {
            break;
        }
    }
}

b32 ConsumeBool(Tokenizer *tokenizer, uint32 *length, bool *boolValue) {
    b32 result = false;

    if (ConsumeCharacterSequence(tokenizer, "true")) {
        *length = 4;
        *boolValue = true;
        result = true;
    }
    else if (ConsumeCharacterSequence(tokenizer, "false")) {
        *length = 5;
        *boolValue = false;
        result = true;
    }
    
    return result;
}

b32 ConsumeFloat(Tokenizer *tokenizer, uint32 *length, real32 *real32Value) {
    b32 result = false;

    // TODO: ...

    return result;
}

b32 ConsumeInteger(Tokenizer *tokenizer, uint32 *length, int32 *int32Value) {
    b32 result = false;

    int32 value = 0;
    *length = 0;
    while (ConsumeNumericCharacter(tokenizer)) {
        *length = *length + 1;

        int32 lastValue = tokenizer->buffer[tokenizer->index - 1] - '0';
        value = value * 10 + lastValue;
        result = true;
    }

    if (result) {
        *int32Value = value;
    }

    return result;
}

b32 ConsumeQuotedChar(Tokenizer *tokenizer, uint32 *length, char *charValue) {
    b32 result = false;

    // TODO: ...

    return result;
}

b32 ConsumeString(Tokenizer *tokenizer, uint32 *length, char **stringValue) {
    b32 result = false;

    char *buffer = &tokenizer->buffer[tokenizer->index];
    uint32 charactersLeft = tokenizer->filehandle.fileSize - tokenizer->index;

    uint32 charactersConsumed = 0;
    const char *scannedString = ScanQuotedString(&tokenizer->buffer[tokenizer->index], charactersLeft, &charactersConsumed);

    if (scannedString != NULL) {
        const char *unescapedString = UnescapeString(scannedString, charactersConsumed);
        ASSERT(unescapedString != NULL);

        *stringValue = (char *)unescapedString;
        *length = charactersConsumed;
        tokenizer->index += charactersConsumed;
        result = true;
    }

    return result;
}

b32 ConsumeIdentifier(Tokenizer *tokenizer, uint32 *length) {
    b32 result = false;
    *length = 0;

    if (ConsumeCharacter(tokenizer, '_') || ConsumeAlphaCharacter(tokenizer)) {
        *length = 1;

        while (ConsumeCharacter(tokenizer, '_') || ConsumeAlphaNumericCharacter(tokenizer)) {
            *length = *length + 1;
        }

        result = true;
    }
    
    return result;
}

PPToken GetNextToken(Tokenizer *tokenizer) {
    ConsumeWhitespaceAndComments(tokenizer);

    PPToken token = {};
    token.filePath = tokenizer->filePath;
    token.line = tokenizer->line;
    token.column = tokenizer->column;

    token.text = tokenizer->buffer + tokenizer->index;

    if (AtEndOfFile(tokenizer)) {
        token.type = PPTokenType_EndOfFile;
        token.length = 0;
    }
    else if (ConsumeCharacter(tokenizer, ':')) {
        token.type = PPTokenType_Colon;
        token.length = 1;
    }
    else if (ConsumeCharacter(tokenizer, ';')) {
        token.type = PPTokenType_Semicolon;
        token.length = 1;
    }
    else if (ConsumeCharacter(tokenizer, '.')) {
        token.type = PPTokenType_Period;
        token.length = 1;
    }
    else if (ConsumeCharacter(tokenizer, ',')) {
        token.type = PPTokenType_Comma;
        token.length = 1;
    }
    else if (ConsumeCharacterSequence(tokenizer, "->")) {
        token.type = PPTokenType_Arrow;
        token.length = 2;
    }
    else if (ConsumeCharacter(tokenizer, '(')) {
        token.type = PPTokenType_OpenParen;
        token.length = 1;
    }
    else if (ConsumeCharacter(tokenizer, ')')) {
        token.type = PPTokenType_CloseParen;
        token.length = 1;
    }
    else if (ConsumeCharacter(tokenizer, '[')) {
        token.type = PPTokenType_OpenBracket;
        token.length = 1;
    }
    else if (ConsumeCharacter(tokenizer, ']')) {
        token.type = PPTokenType_CloseBracket;
        token.length = 1;
    }
    else if (ConsumeCharacter(tokenizer, '{')) {
        token.type = PPTokenType_OpenBrace;
        token.length = 1;
    }
    else if (ConsumeCharacter(tokenizer, '}')) {
        token.type = PPTokenType_CloseBrace;
        token.length = 1;
    }
    else if (ConsumeCharacter(tokenizer, '+')) {
        token.type = PPTokenType_Plus;
        token.length = 1;
    }
    else if (ConsumeCharacter(tokenizer, '-')) {
        token.type = PPTokenType_Minus;
        token.length = 1;
    }
    else if (ConsumeCharacter(tokenizer, '/')) {
        token.type = PPTokenType_Slash;
        token.length = 1;
    }
    else if (ConsumeCharacter(tokenizer, '*')) {
        token.type = PPTokenType_Asterisk;
        token.length = 1;
    }
    else if (ConsumeCharacter(tokenizer, '&')) {
        token.type = PPTokenType_Ampersand;
        token.length = 1;
    }
    else if (ConsumeCharacter(tokenizer, '|')) {
        token.type = PPTokenType_Pipe;
        token.length = 1;
    }
    else if (ConsumeCharacter(tokenizer, '!')) {
        token.type = PPTokenType_Exclamation;
        token.length = 1;
    }
    else if (ConsumeCharacter(tokenizer, '<')) {
        token.type = PPTokenType_LessThan;
        token.length = 1;
    }
    else if (ConsumeCharacter(tokenizer, '>')) {
        token.type = PPTokenType_GreaterThan;
        token.length = 1;
    }
    else if (ConsumeCharacter(tokenizer, '=')) {
        token.type = PPTokenType_Equal;
        token.length = 1;
    }
    else if (ConsumeCharacter(tokenizer, '?')) {
        token.type = PPTokenType_Question;
        token.length = 1;
    }
    else if (ConsumeCharacter(tokenizer, '#')) {
        token.type = PPTokenType_Hash;
        token.length = 1;
    }
    else if (ConsumeBool(tokenizer, &token.length, &token.boolValue)) {
        token.type = PPTokenType_Bool;
    }
    else if (ConsumeFloat(tokenizer, &token.length, &token.real32Value)) {
        token.type = PPTokenType_Float;
    }
    else if (ConsumeInteger(tokenizer, &token.length, &token.int32Value)) {
        token.type = PPTokenType_Integer;
    }
    else if (ConsumeQuotedChar(tokenizer, &token.length, &token.charValue)) {
        token.type = PPTokenType_Char;
    }
    else if (ConsumeString(tokenizer, &token.length, &token.stringValue)) {
        token.type = PPTokenType_String;
    }
    else if (ConsumeIdentifier(tokenizer, &token.length)) {
        token.type = PPTokenType_Identifier;
    }
    else {
        ConsumeAnyCharacter(tokenizer);
        token.type = PPTokenType_Unknown;
        token.length = 1;
    }

    return token;
}

DynamicArray<PPToken> TokenizeFile(Tokenizer *tokenizer) {
    DynamicArray<PPToken> result = MakeDynamicArray<PPToken>(4096);

    while (true) {
        PPToken token = GetNextToken(tokenizer);
        if (token.type == PPTokenType_EndOfFile) {
            break;
        }

        PushBack(&result, token);
    }

    return result;
}


// Parsing
struct ParseCursor {
    DynamicArray<PPToken> tokens;
    uint32 index;
};

b32 TokenEquals(PPToken token, const char *match) {
    b32 result = false;
    if (match != NULL) {

        uint32 matchLength = StringLength(match);
        if (token.length == matchLength) {
            result = memcmp(token.text, match, token.length) == 0;
        }
    }

    return result;
}

b32 TokenEquals(PPToken a, PPToken b) {
    b32 result = false;

    if (a.type == b.type && a.length == b.length) {
        result = memcmp(a.text, b.text, a.length) == 0;
    }

    return result;
}

b32 PreviousTokenMatches(ParseCursor *cursor, const char *match, int32 offset = 1) {
    b32 result = false;

    if ((int32)cursor->index - offset >= 0 &&
        TokenEquals(cursor->tokens[cursor->index - offset], match)) {

        result = true;
    }

    return result;
}

b32 ParseTokenOfType(ParseCursor *cursor, PPTokenType type, PPToken *value = NULL) {
    b32 result = false;

    if (cursor->index < cursor->tokens.count &&
        cursor->tokens[cursor->index].type == type) {

        if (value != NULL) {
            *value = cursor->tokens[cursor->index];
        }

        cursor->index++;
        result = true;
    }

    return result;
}

b32 ParseIdentifierToken(ParseCursor *cursor, const char *match, PPToken *value = NULL) {
    b32 result = false;

    if (cursor->index < cursor->tokens.count &&
        cursor->tokens[cursor->index].type == PPTokenType_Identifier &&
        TokenEquals(cursor->tokens[cursor->index], match)) {

        if (value != NULL) {
            *value = cursor->tokens[cursor->index];
        }

        result = true;
        cursor->index++;
    }

    return result;
}

b32 ParseArrayLength(ParseCursor *cursor, PPToken *token) {
    uint32 startIndex = cursor->index;
    b32 result = false;

    if (ParseTokenOfType(cursor, PPTokenType_OpenBracket) &&
        (ParseTokenOfType(cursor, PPTokenType_Identifier, token) ||
         ParseTokenOfType(cursor, PPTokenType_Integer, token)) &&
        ParseTokenOfType(cursor, PPTokenType_CloseBracket)) {
        result = true;
    }
    
    if (!result) {
        cursor->index = startIndex;
    }

    return result;
}

b32 ParseTypeDescription(ParseCursor *cursor, ParsedTypeDescription *type);

b32 ParseTemplateTypes(ParseCursor *cursor, DynamicArray<ParsedTypeDescription> *types) {
    uint32 startIndex = cursor->index;
    b32 result = false;

    if (ParseTokenOfType(cursor, PPTokenType_LessThan)) {
        while (true) {
            if (ParseTokenOfType(cursor, PPTokenType_GreaterThan)) {
                if (types->count > 0) {
                    result = true;
                }
                break;
            }
            else if (types->count > 0 &&
                     !ParseTokenOfType(cursor, PPTokenType_Comma)) {
                break;
            }

            ParsedTypeDescription typeDescription = {};
            if (ParseTypeDescription(cursor, &typeDescription)) {
                PushBack(types, typeDescription);
            }
            else {
                break;
            }
        }
    }
    else {
        result = true;
    }

    if (!result) {
        cursor->index = startIndex;
    }

    return result;
}

b32 ParseTypeDescription(ParseCursor *cursor, ParsedTypeDescription *type) {
    uint32 startIndex = cursor->index;
    b32 result = false;

    ParsedTypeDescription typeValue = {};
    typeValue.isConst = ParseIdentifierToken(cursor, "const");

    if (ParseTokenOfType(cursor, PPTokenType_Identifier, &typeValue.typeName) &&
        ParseTemplateTypes(cursor, &typeValue.templateTypes)) {

        while (ParseTokenOfType(cursor, PPTokenType_Asterisk)) {
            typeValue.pointerLevels++;
        }

        ParseArrayLength(cursor, &typeValue.arrayLength);
        result = true;
    }

    if (!result) {
        DeallocateParsedTypeDescription(type);

        cursor->index = startIndex;
    }
    else {
        *type = typeValue;
    }

    return result;
}

b32 ParseFunctionArguments(ParseCursor *cursor, DynamicArray<ParsedFunctionArgument> *arguments) {
    b32 result = false;
    uint32 startIndex = cursor->index;

    while (true) {
        ParsedFunctionArgument argument = {};

        if (ParseTokenOfType(cursor, PPTokenType_CloseParen)) {
            result = true;
            break;
        }
        else if (ParseTypeDescription(cursor, &argument.type)) {
            if (ParseTokenOfType(cursor, PPTokenType_Identifier, &argument.name)) {

                // Default value
                if (ParseTokenOfType(cursor, PPTokenType_Equal)) {
                    if (ParseTokenOfType(cursor, PPTokenType_String, &argument.defaultValue) ||
                        ParseTokenOfType(cursor, PPTokenType_Integer, &argument.defaultValue) ||
                        ParseTokenOfType(cursor, PPTokenType_Bool, &argument.defaultValue) ||
                        ParseTokenOfType(cursor, PPTokenType_Float, &argument.defaultValue) ||
                        ParseTokenOfType(cursor, PPTokenType_Identifier, &argument.defaultValue)) {
                        argument.hasDefaultValue = true;
                    }
                    else {
                        Plat.DebugPrint("Failed to parse default argument of type %.*s", argument.type.typeName.length, argument.type.typeName.text);
                    }
                }

                if (ParseTokenOfType(cursor, PPTokenType_Comma)) {
                    PushBack(arguments, argument);
                    continue;
                }
                else if (ParseTokenOfType(cursor, PPTokenType_CloseParen)) {
                    PushBack(arguments, argument);
                    result = true;
                    break;
                }
                else {
                    break;
                }
            }
        }

        result = false;
        break;
    }

    if (!result) {
        for (int i = 0; i < arguments->count; i++) {
            DeallocateParsedTypeDescription(&(*arguments)[i].type);
        }

        DeallocateDynamicArray(arguments);

        cursor->index = startIndex;
    }

    return result;
}

b32 ParseIncludeFile(ParseCursor *cursor, PreprocessorData *data) {
    b32 result = false;
    uint32 startIndex = cursor->index;

    PPToken includedPath;
    if (ParseTokenOfType(cursor, PPTokenType_Hash) &&
        ParseIdentifierToken(cursor, "include") &&
        ParseTokenOfType(cursor, PPTokenType_String, &includedPath)) {

        result = ParseSourceFile(data, (const char *)includedPath.stringValue);
    }

    if (!result) {
        cursor->index = startIndex;
    }

    return result;
}

b32 ParseProperties(ParseCursor *cursor, DynamicArray<ParsedProperty> *properties) {
    uint32 startIndex = cursor->index;
    b32 result = true;

    while (true) {
        ParsedProperty property = {};

        if (ParseTokenOfType(cursor, PPTokenType_Identifier, &property.name)) {

            if (ParseTokenOfType(cursor, PPTokenType_Equal)) {
                if (ParseTokenOfType(cursor, PPTokenType_String, &property.value) ||
                    ParseTokenOfType(cursor, PPTokenType_Integer, &property.value) ||
                    ParseTokenOfType(cursor, PPTokenType_Bool, &property.value)) {

                    property.hasValue = true;
                }
                else {
                    result = false;
                    break;
                }
            }

            PushBack(properties, property);

            if (!ParseTokenOfType(cursor, PPTokenType_Comma)) {
                break;
            }
        }
        else {
            break;
        }
    }

    if (!result) {
        cursor->index = startIndex;
        DynamicArrayClear(properties);
    }

    return result;
}

b32 ParseDefineFunction(ParseCursor *cursor, PreprocessorData *data) {
    b32 result = false;
    uint32 startIndex = cursor->index;

    ParsedFunction function = {};

    PPToken funcTypeToken = {};
    PPToken requiredDataTypeToken = {};
    
    bool registerFunction = false;
    if (!PreviousTokenMatches(cursor, "define") &&
        ParseIdentifierToken(cursor, "FUNCTION") &&
        ParseTokenOfType(cursor, PPTokenType_OpenParen) &&
        ParseTokenOfType(cursor, PPTokenType_Identifier, &funcTypeToken)) {

        if (ParseTokenOfType(cursor, PPTokenType_Comma)) {
            ParseTokenOfType(cursor, PPTokenType_Identifier, &requiredDataTypeToken);
        }
        ParseTokenOfType(cursor, PPTokenType_CloseParen);

        registerFunction = true;
        function.isPersistent = true;
        function.persistentFunctionType = funcTypeToken;
        function.requiredSchemaDataType = requiredDataTypeToken;
    }

    if (!TokenEquals(cursor->tokens[cursor->index], "define") &&
        ParseTypeDescription(cursor, &function.returnType)) {

        if (ParseTokenOfType(cursor, PPTokenType_Identifier, &function.name) &&
            ParseTokenOfType(cursor, PPTokenType_OpenParen) &&
            ParseFunctionArguments(cursor, &function.arguments) &&
            ParseTokenOfType(cursor, PPTokenType_OpenBrace)) {

            PushBack(&data->definedFunctions, function);

            uint32 functionNameLength = function.name.length;
            char * functionName = (char *)Plat.AllocateMem(functionNameLength + 1);
            snprintf(functionName, functionNameLength + 1, "%.*s", function.name.length, function.name.text);
            HashTableInsert(&data->symbolSourceFileMap, functionName, functionNameLength, function.name.filePath, true);
            
            result = true;
        }
    }

    if (registerFunction && result) {
        PushBack(&data->definedPersistentFunctions, function);
    }

    if (!result) {
        cursor->index = startIndex;
    }

    return result;
}

b32 ParseDefineFunctionType(ParseCursor *cursor, PreprocessorData *data) {
    b32 result = false;
    uint32 startIndex = cursor->index;

    ParsedPersistentFunctionType function = {};
    if (!PreviousTokenMatches(cursor, "define") &&
        ParseIdentifierToken(cursor, "DEFINE_FUNCTIONTYPE")) {
        
        if (ParseTokenOfType(cursor, PPTokenType_OpenParen) &&
            ParseTokenOfType(cursor, PPTokenType_Identifier, &function.name)) {

            if (ParseTokenOfType(cursor, PPTokenType_Comma) &&
                ParseTypeDescription(cursor, &function.returnType)) {

                if (ParseTokenOfType(cursor, PPTokenType_Comma) && 
                    // @NOTE: ParseFunctionArguments() parses the parens
                    ParseFunctionArguments(cursor, &function.arguments) &&
                    ParseTokenOfType(cursor, PPTokenType_Semicolon)) {

                    // Disallow default values for function arguments
                    b32 anyArgsHaveDefaultValues = false;
                    for (int i = 0; i < function.arguments.count; i++) {
                        if (function.arguments[i].hasDefaultValue) {
                            anyArgsHaveDefaultValues = true;
                            // @TODO: error message
                        }
                    }

                    if (!anyArgsHaveDefaultValues) {
                        PushBack(&data->definedFunctionTypes, function);
                        result = true;
                    }
                }
            }
        }
        else {
            //Plat.DebugPrint("No return type specified for %.*s", function.name.length, function.name.text);
            // @TODO: error message
        }
    }

end:
    if (!result) {
        cursor->index = startIndex;
    }

    return result;
}

b32 ParseStructFields(ParseCursor *cursor, DynamicArray<ParsedField> *fields) {
    uint32 startIndex = cursor->index;
    b32 result = false;

    if (ParseTokenOfType(cursor, PPTokenType_OpenBrace)) {
        while (true) {
            if (ParseTokenOfType(cursor, PPTokenType_CloseBrace)) {
                result = true;
                break;
            }

            ParsedField parsedField = {};

            if (
                // Either parse with FIELD() properties...
                (!PreviousTokenMatches(cursor, "define") &&
                 ParseIdentifierToken(cursor, "FIELD") &&
                 ParseTokenOfType(cursor, PPTokenType_OpenParen) &&
                 ParseProperties(cursor, &parsedField.properties) &&
                 ParseTokenOfType(cursor, PPTokenType_CloseParen) &&
                 ParseTypeDescription(cursor, &parsedField.type))

                ||

                // Or just the field itself
                ParseTypeDescription(cursor, &parsedField.type)) {

                if (!ParseTokenOfType(cursor, PPTokenType_Identifier, &parsedField.name)) {
                    break;
                }

                ParseArrayLength(cursor, &parsedField.type.arrayLength);

                if (ParseTokenOfType(cursor, PPTokenType_Semicolon)) {
                    PushBack(fields, parsedField);
                }
                else {
                    break;
                }
            }
            else {
                break;
            }
        }
    }

    if (!result) {
        for (int i = 0; i < fields->count; i++) {
            DeallocateParsedTypeDescription(&(*fields)[i].type);
        }

        DeallocateDynamicArray(fields);

        cursor->index = startIndex;
    }

    return result;
}

b32 ParseEnumValues(ParseCursor *cursor, DynamicArray<PPToken> *values) {
    uint32 startIndex = cursor->index;
    b32 result = false;

    if (ParseTokenOfType(cursor, PPTokenType_OpenBrace)) {
        while (true) {
            if (ParseTokenOfType(cursor, PPTokenType_CloseBrace)) {
                result = true;
                break;
            }
            else {
                if (!ParseTokenOfType(cursor, PPTokenType_Identifier)) {
                    break;
                }

                PushBack(values, cursor->tokens[cursor->index - 1]);

                if (ParseTokenOfType(cursor, PPTokenType_Comma)) {
                    continue;
                }
                else if (ParseTokenOfType(cursor, PPTokenType_CloseBrace)) {
                    result = true;
                    break;
                }
                else {
                    break;
                }
            }
        }
    }

    if (!result) {
        cursor->index = startIndex;
        DeallocateDynamicArray(values);
    }

    return result;
}

b32 ParseDefineComponent(ParseCursor *cursor, PreprocessorData *data) {
    PPToken token = cursor->tokens[cursor->index];
    uint32 startIndex = cursor->index;
    b32 result = false;

    if (!PreviousTokenMatches(cursor, "define") &&
        ParseIdentifierToken(cursor, "COMPONENT")) {

        ParsedComponent component = {};

        if (ParseTokenOfType(cursor, PPTokenType_OpenParen)) {
            if (ParseTokenOfType(cursor, PPTokenType_Identifier, &component.name) &&
                ParseTokenOfType(cursor, PPTokenType_CloseParen)) {

                if (ParseStructFields(cursor, &component.fields) &&
                    ParseTokenOfType(cursor, PPTokenType_Semicolon)) {

                    PushBack(&data->definedComponents, component);
                    result = true;

                    uint32 componentNameLength = component.name.length + 9;
                    char * componentName = (char *)Plat.AllocateMem(componentNameLength + 1);
                    snprintf(componentName, componentNameLength + 1, "%.*sComponent", component.name.length, component.name.text);
                    HashTableInsert(&data->symbolSourceFileMap, componentName, componentNameLength, component.name.filePath);
                }
            }
        }
    }

    if (!result) {
        cursor->index = startIndex;
    }

    return result;
}

b32 ParseDefineSystem(ParseCursor *cursor, PreprocessorData *data) {
    uint32 startIndex = cursor->index;
    b32 result = false;

    if (!PreviousTokenMatches(cursor, "define") &&
        ParseIdentifierToken(cursor, "SYSTEM")) {

        ParsedSystem system = {};
        system.priority = DEFAULT_SYSTEM_PRIORITY;

        if (ParseTokenOfType(cursor, PPTokenType_OpenParen) &&
            ParseTokenOfType(cursor, PPTokenType_Identifier, &system.name)) {
            
            if (ParseTokenOfType(cursor, PPTokenType_Comma) &&
                (!ParseProperties(cursor, &system.properties) || system.properties.count == 0)) {
                return false;
            }

            // @TODO: probably do this in a separate property analysis pass
            for (int i = 0; i < system.properties.count; i++) {
                ParsedProperty property = system.properties[i];

                if (TokenEquals(property.name, "Priority") &&
                    property.hasValue &&
                    property.value.type == PPTokenType_Integer) {

                    system.priority = property.value.int32Value;
                }
            }

            if (ParseTokenOfType(cursor, PPTokenType_CloseParen)) {
                PushBack(&data->definedSystems, system);

                uint32 systemNameLength = system.name.length + 6;
                char * systemName = (char *)Plat.AllocateMem(systemNameLength + 1);
                snprintf(systemName, systemNameLength + 1, "%.*sSystem", system.name.length, system.name.text);
                HashTableInsert(&data->symbolSourceFileMap, systemName, systemNameLength, system.name.filePath);
                result = true;
            }
            else {
                DeallocateDynamicArray(&system.properties);
            }                
        }
    }

    if (!result) {
        cursor->index = startIndex;
    }

    return result;
}

b32 ParseDefineScene(ParseCursor *cursor, PreprocessorData *data) {
    uint32 startIndex = cursor->index;
    b32 result = false;

    if (!PreviousTokenMatches(cursor, "define") &&
        ParseIdentifierToken(cursor, "SCENE")) {

        ParsedScene scene = {};

        if (ParseTokenOfType(cursor, PPTokenType_OpenParen) &&
            ParseTokenOfType(cursor, PPTokenType_Identifier, &scene.name)) {

            if (ParseTokenOfType(cursor, PPTokenType_Comma) &&
                (!ParseProperties(cursor, &scene.properties) || scene.properties.count == 0)) {
                return false;
            }

            if (ParseTokenOfType(cursor, PPTokenType_CloseParen) &&
                ParseTokenOfType(cursor, PPTokenType_Semicolon)) {
                
                PushBack(&data->definedScenes, scene);
                result = true;
            }
            else {
                DeallocateDynamicArray(&scene.properties);
            }
        }
    }

    if (!result) {
        cursor->index = startIndex;
    }

    return result;
}

b32 ParseSerializableStruct(ParseCursor *cursor, PreprocessorData *data) {
    uint32 startIndex = cursor->index;
    b32 result = false;

    ParsedSerializable serializable = {};
    serializable.type = SerializableType_Struct;

    if (!PreviousTokenMatches(cursor, "define") &&
        ParseIdentifierToken(cursor, "STRUCT") &&
        ParseTokenOfType(cursor, PPTokenType_OpenParen) &&
        ParseProperties(cursor, &serializable.properties) &&
        ParseTokenOfType(cursor, PPTokenType_CloseParen)) {

        if (ParseIdentifierToken(cursor, "struct") &&
            ParseTokenOfType(cursor, PPTokenType_Identifier, &serializable.name) &&
            ParseStructFields(cursor, &serializable._struct.fields) &&
            ParseTokenOfType(cursor, PPTokenType_Semicolon)) {

            result = true;
            PushBack(&data->serializableTypes, serializable);

            uint32 typeNameLength = serializable.name.length;
            char * typeName = (char *)Plat.AllocateMem(typeNameLength + 1);
            snprintf(typeName, typeNameLength + 1, "%.*s",
                     serializable.name.length, serializable.name.text);
            HashTableInsert(&data->symbolSourceFileMap, typeName, typeNameLength,
                            serializable.name.filePath);
        }
    }

    if (!result) {
        cursor->index = startIndex;
    }

    return result;
}

b32 ParseSerializableEnum(ParseCursor *cursor, PreprocessorData *data) {
    uint32 startIndex = cursor->index;
    b32 result = false;

    ParsedSerializable serializable = {};
    serializable.type = SerializableType_Enum;
        

    if (!PreviousTokenMatches(cursor, "define") &&
        ParseIdentifierToken(cursor, "ENUM") &&
        ParseTokenOfType(cursor, PPTokenType_OpenParen) &&
        ParseProperties(cursor, &serializable.properties) &&
        ParseTokenOfType(cursor, PPTokenType_CloseParen)) {

        PPToken backingType;

        if (ParseIdentifierToken(cursor, "enum") &&
            ParseTokenOfType(cursor, PPTokenType_Identifier, &serializable.name) &&
            ParseTokenOfType(cursor, PPTokenType_Colon) &&
            ParseTokenOfType(cursor, PPTokenType_Identifier, &backingType)) {

            // @NOTE: only uint32 backing type allowed right now
            if (TokenEquals(backingType, "uint32") &&
                ParseEnumValues(cursor, &serializable._enum.values) &&
                ParseTokenOfType(cursor, PPTokenType_Semicolon)) {

                result = true;
                PushBack(&data->serializableTypes, serializable);

                uint32 typeNameLength = serializable.name.length;
                char * typeName = (char *)Plat.AllocateMem(typeNameLength + 1);
                snprintf(typeName, typeNameLength + 1, "%.*s",
                         serializable.name.length, serializable.name.text);
                HashTableInsert(&data->symbolSourceFileMap, typeName, typeNameLength,
                                serializable.name.filePath);

                uint32 valueCount = serializable._enum.values.count;
                for (int value = 0; value < valueCount; value++) {
                    PPToken valueToken = serializable._enum.values[value];

                    ParsedDefinedInt definedInt = {};

                    definedInt.name = valueToken;
                    definedInt.value = value;

                    PushBack(&data->definedInts, definedInt);
                }
            }
        }
    }

    if (!result) {
        cursor->index = startIndex;
    }

    return result;
}

b32 ParseSerializableBitfield(ParseCursor *cursor, PreprocessorData *data) {
    uint32 startIndex = cursor->index;
    b32 result = false;

    if (!PreviousTokenMatches(cursor, "define") &&
        ParseIdentifierToken(cursor, "BITFIELD")) {

        ParsedSerializable serializable = {};
        serializable.type = SerializableType_Bitfield;
        
        if (ParseTokenOfType(cursor, PPTokenType_OpenParen) &&
            ParseTokenOfType(cursor, PPTokenType_Identifier, &serializable.name) &&
            ParseTokenOfType(cursor, PPTokenType_Comma)) {

            if (ParseEnumValues(cursor, &serializable._bitfield.values) &&
                ParseTokenOfType(cursor, PPTokenType_CloseParen) &&
                ParseTokenOfType(cursor, PPTokenType_Semicolon)) {
                result = true;
                PushBack(&data->serializableTypes, serializable);

                uint32 typeNameLength = serializable.name.length;
                char * typeName = (char *)Plat.AllocateMem(typeNameLength + 1);
                snprintf(typeName, typeNameLength + 1, "%.*s",
                         serializable.name.length, serializable.name.text);
                HashTableInsert(&data->symbolSourceFileMap, typeName, typeNameLength,
                                serializable.name.filePath);

                uint32 valueCount = serializable._bitfield.values.count;
                for (int value = 0; value < valueCount; value++) {
                    PPToken valueToken = serializable._bitfield.values[value];

                    ParsedDefinedInt definedInt = {};

                    definedInt.name = valueToken;
                    definedInt.value = 1 << value;

                    PushBack(&data->definedInts, definedInt);
                }
            }
        }   
    }

    if (!result) {
        cursor->index = startIndex;
    }

    return result;
}

b32 ParseSerializableBitarray(ParseCursor *cursor, PreprocessorData *data) {
    uint32 startIndex = cursor->index;
    b32 result = false;

    if (!PreviousTokenMatches(cursor, "define") &&
        ParseIdentifierToken(cursor, "BITARRAY")) {

        ParsedSerializable serializable = {};
        serializable.type = SerializableType_Bitarray;
        
        if (ParseTokenOfType(cursor, PPTokenType_OpenParen) &&
            ParseTokenOfType(cursor, PPTokenType_Identifier, &serializable.name) &&
            ParseTokenOfType(cursor, PPTokenType_Comma)) {

            if (ParseEnumValues(cursor, &serializable._bitarray.values) &&
                ParseTokenOfType(cursor, PPTokenType_CloseParen) &&
                ParseTokenOfType(cursor, PPTokenType_Semicolon)) {
                result = true;
                PushBack(&data->serializableTypes, serializable);

                uint32 typeNameLength = serializable.name.length;
                char * typeName = (char *)Plat.AllocateMem(typeNameLength + 1);
                snprintf(typeName, typeNameLength + 1, "%.*s",
                         serializable.name.length, serializable.name.text);
                HashTableInsert(&data->symbolSourceFileMap, typeName, typeNameLength,
                                serializable.name.filePath);

                uint32 valueCount = serializable._bitarray.values.count;
                for (int value = 0; value < valueCount; value++) {
                    PPToken valueToken = serializable._bitarray.values[value];

                    ParsedDefinedInt definedInt = {};

                    definedInt.name = valueToken;
                    definedInt.value = 1 << value;

                    PushBack(&data->definedInts, definedInt);
                }
            }
        }   
    }

    if (!result) {
        cursor->index = startIndex;
    }

    return result;
}

b32 ParseSerializableTaggedUnion(ParseCursor *cursor, PreprocessorData *data) {
    uint32 startIndex = cursor->index;
    b32 result = false;

    if (!PreviousTokenMatches(cursor, "define") &&
        ParseIdentifierToken(cursor, "TAGGED_UNION")) {

        ParsedSerializable serializable = {};
        serializable.type = SerializableType_TaggedUnion;
        
        if (ParseTokenOfType(cursor, PPTokenType_OpenParen) &&
            ParseTokenOfType(cursor, PPTokenType_Identifier, &serializable.name) &&
            ParseTokenOfType(cursor, PPTokenType_Comma)) {

            if (ParseStructFields(cursor, &serializable._tagged_union.fields) &&
                ParseTokenOfType(cursor, PPTokenType_CloseParen) &&
                ParseTokenOfType(cursor, PPTokenType_Semicolon)) {
                result = true;
                PushBack(&data->serializableTypes, serializable);

                uint32 typeNameLength = serializable.name.length;
                char * typeName = (char *)Plat.AllocateMem(typeNameLength + 1);
                snprintf(typeName, typeNameLength + 1, "%.*s",
                         serializable.name.length, serializable.name.text);
                HashTableInsert(&data->symbolSourceFileMap, typeName, typeNameLength,
                                serializable.name.filePath);
            }
        }   
    }

    if (!result) {
        cursor->index = startIndex;
    }

    return result;
}

b32 ParseDefineTag(ParseCursor *cursor, PreprocessorData *data) {
    uint32 startIndex = cursor->index;
    b32 result = false;

    if (!PreviousTokenMatches(cursor, "define") &&
        ParseIdentifierToken(cursor, "TAG") &&
        ParseTokenOfType(cursor, PPTokenType_OpenParen)) {

        
        while (!ParseTokenOfType(cursor, PPTokenType_CloseParen)) {

            ParsedTag tag = {};
            PPToken tagName = {};
            if (ParseTokenOfType(cursor, PPTokenType_Identifier, &tagName)) {
                tag.name = tagName;
            }
            else {
                ASSERT(false);
            }

            PushBack(&data->definedTags, tag);

            if (ParseTokenOfType(cursor, PPTokenType_Comma)) {
                continue;
            }
        }
    }

    if (!result) {
        cursor->index = startIndex;
    }

    return result;
}

b32 ParseDefineInputAction(ParseCursor *cursor, PreprocessorData *data) {
    uint32 startIndex = cursor->index;
    b32 result = false;

    if (!PreviousTokenMatches(cursor, "define") &&
        ParseIdentifierToken(cursor, "ACTION") &&
        ParseTokenOfType(cursor, PPTokenType_OpenParen)) {

        ParsedInputAction action = {};
        PPToken actionName = {};
        if (ParseTokenOfType(cursor, PPTokenType_Identifier, &actionName)) {
            action.name = actionName;
            
            if (ParseTokenOfType(cursor, PPTokenType_Comma)) {
                if (ParseIdentifierToken(cursor, "Analogue")) {
                    action.analogue = true;
                }
                else {
                    ASSERT(false);
                }
            }
        }
        else {
            ASSERT(false);
        }

        PushBack(&data->definedActions, action);

        if (!ParseTokenOfType(cursor, PPTokenType_CloseParen)) {
            ASSERT(false);
        }
    }

    if (!result) {
        cursor->index = startIndex;
    }

    return result;
}

b32 ParseSerializable(ParseCursor *cursor, PreprocessorData *data) {

    b32 result = (ParseSerializableStruct(cursor, data) ||
                  ParseSerializableEnum(cursor, data) ||
                  ParseSerializableBitfield(cursor, data) ||
                  ParseSerializableBitarray(cursor, data) ||
                  ParseSerializableTaggedUnion(cursor, data));

    return result;
}

b32 ParseAssetFileExtensions(ParseCursor *cursor, DynamicArray<PPToken> *fileExtensions) {
    b32 result = false;
    uint32 startIndex = cursor->index;

    if (ParseTokenOfType(cursor, PPTokenType_OpenBrace)) {
        while (true) {
            PPToken extension = {};
            if (ParseTokenOfType(cursor, PPTokenType_String, &extension)) {
                PushBack(fileExtensions, extension);
                ParseTokenOfType(cursor, PPTokenType_Comma);
            }
            else if (ParseTokenOfType(cursor, PPTokenType_CloseBrace)) {
                result = fileExtensions->count > 0;
                break;
            }
            else {
                break;
            }
        }
    }

    if (!result) {
        DynamicArrayClear(fileExtensions);
        cursor->index = startIndex;
    }

    return result;
}

b32 ParseDefineAsset(ParseCursor *cursor, PreprocessorData *data) {
    b32 result = false;
    uint32 startIndex = cursor->index;

    ParsedSerializable asset = {};
    DynamicArray<PPToken> extensions = {};

    if (!PreviousTokenMatches(cursor, "define") &&
        ParseIdentifierToken(cursor, "ASSET") &&
        ParseTokenOfType(cursor, PPTokenType_OpenParen) &&
        ParseAssetFileExtensions(cursor, &extensions)) {

        if (ParseTokenOfType(cursor, PPTokenType_Comma) &&
            (!ParseProperties(cursor, &asset.properties) || asset.properties.count == 0)) {
            return false;
        }

        if (ParseTokenOfType(cursor, PPTokenType_CloseParen) &&
            ParseIdentifierToken(cursor, "struct") &&
            ParseTokenOfType(cursor, PPTokenType_Identifier, &asset.name) &&
            ParseStructFields(cursor, &asset._struct.fields) &&
            ParseTokenOfType(cursor, PPTokenType_Semicolon)) {

            asset.type = SerializableType_Struct;

            result = true;
            PushBack(&data->definedAssets, asset);
            PushBack(&data->definedAssetExtensions, extensions);
            PushBack(&data->serializableTypes, asset);

            uint32 typeNameLength = asset.name.length;
            char * typeName = (char *)Plat.AllocateMem(typeNameLength + 1);
            snprintf(typeName, typeNameLength + 1, "%.*s", asset.name.length, asset.name.text);
            HashTableInsert(&data->symbolSourceFileMap, typeName, typeNameLength, asset.name.filePath);
        }   
    }

    if (!result) {
        cursor->index = startIndex;
    }

    return result;
}

b32 ParseDefineInt(ParseCursor *cursor, PreprocessorData *data) {

    uint32 startIndex = cursor->index;
    b32 result = false;

    ParsedDefinedInt definedInt = {};

    if (ParseIdentifierToken(cursor, "define") &&
        ParseTokenOfType(cursor, PPTokenType_Identifier, &definedInt.name) &&
        ParseTokenOfType(cursor, PPTokenType_Integer)) {

        PPToken value = cursor->tokens[cursor->index - 1];

        char *valueStr = (char *)Plat.AllocateMem(sizeof(char) * (value.length + 1));
        StringCopy(valueStr, value.length + 1, value.text);

#if _WIN32
        sscanf_s(valueStr, "%d", &definedInt.value);
#else
        sscanf(valueStr, "%d", &definedInt.value);
#endif


        Plat.DeallocateMem(valueStr);

        PushBack(&data->definedInts, definedInt);
    }
    

    if (!result) {
        cursor->index = startIndex;
    }

    return result;
}

typedef b32 ParseFunc(ParseCursor *cursor, PreprocessorData *data);

b32 ParseFile(DynamicArray<PPToken> tokens, PreprocessorData *data) {

    ParseCursor cursor = {};
    cursor.tokens = tokens;

    b32 succeeded = true;

    static ParseFunc *parseFuncs[] = {
        ParseIncludeFile,
        ParseDefineFunction,
        ParseDefineComponent,
        ParseDefineSystem,
        ParseDefineScene,
        ParseSerializable,
        ParseDefineInt,
        ParseDefineFunctionType,
        ParseDefineAsset,
        ParseDefineInputAction,
        ParseDefineTag,
    };
    static const uint32 parseFuncCount = ARRAY_LENGTH(ParseFunc *, parseFuncs);

    while (cursor.index < cursor.tokens.count) {
        b32 successfulParse = false;

        for (int i = 0; i < parseFuncCount; i++) {
            if (parseFuncs[i](&cursor, data)) {
                successfulParse = true;
                break;
            }
            else if (data->error.type != PreprocessorError_None) {
                return false;
            }
        }

        if (!successfulParse) {
            cursor.index++;
        }
    }

    return succeeded;
}

