
#include <stdio.h>
#include <stdlib.h>
#include <memory.h>

#if _WIN32
#include <windows.h>
#elif __linux__
#include <unistd.h>
#include <fts.h>
#endif

#define STB_SPRINTF_IMPLEMENTATION
#include "stb_sprintf.h"


#include "etc_types.h"
#include "platform.h"
#include "commonlib.h"
#include "math/math.h"
#include "etc_memory.h"
#include "filehandle.h"
#include "dynamic_array.h"
#include "hashtable.h"

#include "metaprogramming.h"


#ifdef _WIN32
#include "platform/windows_api.cpp"
PlatformAPI Plat = WindowsPlatformAPI();

// enumerator 'identifier' in a switch of enum 'enumeration' is not handled
#pragma warning(error : 4062)

#elif __APPLE__
#include "platform/osx_api.m"
PlatformAPI Plat = OSXPlatformAPI();

#elif __linux__
#include "platform/linux_api.cpp"
PlatformAPI Plat = LinuxPlatformAPI();
#endif


#include "regexp/regexp.h"
#include "regexp/regexp.c"

#include "io/etctxt.h"

enum PersistentFuncType : uint32 {
};
#include "schema.h"


#include "preprocessor/preprocessor.h"

#include "preprocessor/parse.cpp"
#include "preprocessor/generate.cpp"


b32 ParsedTypeDescriptionEquals(ParsedTypeDescription a, ParsedTypeDescription b) {
    if (!(TokenEquals(a.typeName, b.typeName) &&
          a.pointerLevels == b.pointerLevels &&
          TokenEquals(a.arrayLength, b.arrayLength) &&
          a.isConst == b.isConst &&
          a.templateTypes.count == b.templateTypes.count)) {
        return false;
    }

    for (int i = 0; i < a.templateTypes.count; i++) {
        if (!ParsedTypeDescriptionEquals(a.templateTypes[i], b.templateTypes[i])) {
            return false;
        }
    }

    return true;
}


// Properties
b32 HasProperty(DynamicArray<ParsedProperty> properties, const char *propertyName, ParsedProperty *outProperty)
{
    for (int i = 0; i < properties.count; i++) {
        if (TokenEquals(properties[i].name, propertyName)) {
            if (outProperty != NULL) {
                *outProperty = properties[i];
            }
            return true;
        }
    }
    return false;
}

b32 HasPropertyValue(DynamicArray<ParsedProperty> properties, const char *propertyName, b32 testValue, b32 defaultValue) {

    ParsedProperty property;
    if (HasProperty(properties, propertyName, &property)) {
        if (property.hasValue) {
            if (property.value.type == PPTokenType_Bool) {
                return property.value.boolValue == (bool)testValue;
            }

            return false;
        }

        if (!property.hasValue && testValue) {
            return true;
        }

        return false;
    }
    
    return defaultValue;
}


// Systems are sorted to ensure execution order at runtime
int32 ParsedSystemComparator(const void *a, const void *b) {
    ParsedSystem *aSystem = (ParsedSystem *)a;
    ParsedSystem *bSystem = (ParsedSystem *)b;

    int32 result = (int32)aSystem->priority - (int32)bSystem->priority;
    return result;
}

b32 FindParsedSerializableWithName(PreprocessorData *data, const char *name, uint32 nameLength, ParsedSerializable *serializable) {
    b32 result = false;

    for (int i = 0; i < data->serializableTypes.count; i++) {
        uint32 checkLength = Max(data->serializableTypes[i].name.length, nameLength);
        if (StringEquals(data->serializableTypes[i].name.text, name, checkLength)) {
            if (serializable != NULL) {
                *serializable = data->serializableTypes[i];
            }
            result = true;
            break;
        }
    }

    return result;
}

// Preprocessor Data Access
b32 ComponentTypeExists(PreprocessorData *data, const char *humanName, uint32 nameLength, int32 *index = NULL) {
    b32 result = false;

    PPToken inputToken;
    inputToken.type = PPTokenType_Identifier;
    inputToken.text = (char *)humanName;
    inputToken.length = nameLength;

    for (int i = 0; i < data->definedComponents.count; i++) {
        if (TokenEquals(inputToken, data->definedComponents[i].name)) {
            if (index != NULL) {
                *index = i;
            }
            result = true;
            break;
        }
    }

    return result;
}

b32 SceneExists(PreprocessorData *data, const char *sceneName, uint32 nameLength, int32 *index = NULL) {
    b32 result = false;

    PPToken inputToken;
    inputToken.type = PPTokenType_Identifier;
    inputToken.text = (char *)sceneName;
    inputToken.length = nameLength;

    for (int i = 0; i < data->definedScenes.count; i++) {
        if (TokenEquals(inputToken, data->definedScenes[i].name)) {
            if (index != NULL) {
                *index = i;
            }
            result = true;
            break;
        }
    }

    return result;
}

b32 AssetTypeExists(PreprocessorData *data, const char *assetTypeName, uint32 nameLength, int32 *index = NULL) {
    b32 result = false;

    PPToken inputToken;
    inputToken.type = PPTokenType_Identifier;
    inputToken.text = (char *)assetTypeName;
    inputToken.length = nameLength;

    for (int i = 0; i < data->definedAssets.count; i++) {
        if (TokenEquals(inputToken, data->definedAssets[i].name)) {
            if (index != NULL) {
                *index = i;
            }
            result = true;
            break;
        }
    }

    return result;
}

static int32 handleStringLength = StringLength("Handle");

b32 FindParsedPersistentFunctionTypeWithName(PreprocessorData *data, const char *name, uint32 nameLength, ParsedPersistentFunctionType *type) {
    b32 result = false;
    

    if (nameLength <= handleStringLength ||
        !StringEquals(name + nameLength - handleStringLength, "Handle", handleStringLength)) {
        return false;
    }

    for (int i = 0; i < data->definedFunctionTypes.count; i++) {
        uint32 checkLength = Max(data->definedFunctionTypes[i].name.length, nameLength - handleStringLength);
        if (StringEquals(data->definedFunctionTypes[i].name.text, name, checkLength)) {
            if (type != NULL) {
                *type = data->definedFunctionTypes[i];
            }
            result = true;
            break;
        }
    }

    return result;
}

static int32 dataPairStringLength = StringLength("DataPair");

b32 FindParsedPersistentFunctionDataPairTypeWithName(PreprocessorData *data, const char *name, uint32 nameLength, ParsedPersistentFunctionType *type) {
    b32 result = false;
    

    if (nameLength <= dataPairStringLength ||
        !StringEquals(name + nameLength - dataPairStringLength, "DataPair", dataPairStringLength)) {
        return false;
    }

    for (int i = 0; i < data->definedFunctionTypes.count; i++) {
        uint32 checkLength = Max(data->definedFunctionTypes[i].name.length, nameLength - dataPairStringLength);
        if (StringEquals(data->definedFunctionTypes[i].name.text, name, checkLength)) {
            if (type != NULL) {
                *type = data->definedFunctionTypes[i];
            }
            result = true;
            break;
        }
    }

    return result;
}

b32 FindParsedAssetHandleWithName(PreprocessorData *data, const char *name, uint32 nameLength, ParsedSerializable *asset) {
    b32 result = false;
    
    if (nameLength <= handleStringLength ||
        !StringEquals(name + nameLength - handleStringLength, "Handle", handleStringLength)) {
        return false;
    }

    for (int i = 0; i < data->definedAssets.count; i++) {
        uint32 checkLength = Max(data->definedAssets[i].name.length, nameLength - handleStringLength);
        if (StringEquals(data->definedAssets[i].name.text, name, checkLength)) {
            if (asset != NULL) {
                *asset = data->definedAssets[i];
            }
            result = true;
            break;
        }
    }

    return result;
}

b32 FindDefinedIntWithName(PreprocessorData *data, PPToken name, int32 *value) {
    b32 result = false;

    for (int i = 0; i < data->definedInts.count; i++) {
        if (TokenEquals(data->definedInts[i].name, name)) {
            *value = data->definedInts[i].value;
            result = true;
            break;
        }
    }

    return result;
}

b32 IsValidPersistentFunction(PreprocessorData *data, ParsedFunction function) {
    b32 result = false;

    ASSERT(function.isPersistent);

    for (int i = 0; i < data->definedFunctionTypes.count; i++) {
        ParsedPersistentFunctionType persistentFunctionType = data->definedFunctionTypes[i];

        if (!TokenEquals(function.persistentFunctionType, persistentFunctionType.name)) {
            continue;
        }

        // Make sure function's argument match the persistentFunctionType arguments
        if (function.arguments.count != persistentFunctionType.arguments.count) {
            goto end;
        }

        for (int i = 0; i < function.arguments.count; i++) {
            if (!ParsedTypeDescriptionEquals(persistentFunctionType.arguments[i].type,
                                             function.arguments[i].type)) {
                goto end;
            }
        }
            
        result = true;
        goto end;
    }

 end:
    if (!result) {
        PreprocessorSetError(data, PreprocessorError_MissingFunction,
                             "%s:%d:%d: Invalid persistent function '%.*s'",
                             function.name.filePath, function.name.line, function.name.column,
                             function.name.length, function.name.text);
    }

    return result;
}

b32 ConvertTokenToIntegerValue(PreprocessorData *data, PPToken token, int32 *value) {
    b32 result = false;

    if (token.length > 0) {
        if (token.type == PPTokenType_Integer) {
            const char *str = StringCopy(token.text, token.length);

#if _WIN32
            sscanf_s(str, "%d", value);
#else
            sscanf(str, "%d", value);
#endif

            Plat.DeallocateMem((char *)str);
            result = true;
        }
        else if (FindDefinedIntWithName(data, token, value)) {
            result = true;
        }
    }

    return result;
}

b32 ParsedTypeIsValid(PreprocessorData *data, ParsedTypeDescription type) {
    int32 arrayLength;
    if (type.arrayLength.type != PPTokenType_Unknown &&
             !ConvertTokenToIntegerValue(data, type.arrayLength, &arrayLength)) {
        return false;
    }

    if (StringEquals("DynamicArray", type.typeName.text, type.typeName.length)) {

        if (type.templateTypes.count != 1) {
            return false;
        }

        if (!ParsedTypeIsValid(data, type.templateTypes[0])) {
            return false;
        }
    }
    else if (StringEquals("ChunkedArray", type.typeName.text, type.typeName.length)) {

        return false;

        if (type.templateTypes.count != 1) {
            return false;
        }

        if (!ParsedTypeIsValid(data, type.templateTypes[0])) {
            return false;
        }
    }
    else if (StringEquals("HashTable", type.typeName.text, type.typeName.length)) {

        // @TODO: once this is actually supported we can allow it

        return false;

        if (type.templateTypes.count != 2) {
            return false;
        }

        if (!ParsedTypeIsValid(data, type.templateTypes[0]) ||
            !ParsedTypeIsValid(data, type.templateTypes[1])) {
            return false;
        }
    }
    else if (StringEquals("Schema", type.typeName.text, type.typeName.length)) {
        if (type.pointerLevels != 1) {
            return false;
        }
    }
    else if (StringEquals("char", type.typeName.text, type.typeName.length)) {
        if (type.pointerLevels > 1) {
            return false;
        }
    }
    else if (type.pointerLevels != 0) {
        return false;
    }
    else if (!HashTableContains(&data->validTypeSet, (const char *)type.typeName.text, type.typeName.length)) {
        return false;
    }

    return true;
}


// Schema Construction
SchemaTypeDescription SchemaTypeDescriptionFromParsedTypeDescription(PreprocessorData *data, ParsedTypeDescription parsedType) {
    SchemaTypeDescription type = {};

    // Resolve typeName alias
    const char *typeName = parsedType.typeName.text;
    int32 typeNameLength = parsedType.typeName.length;

    if (HashTableGet(&data->builtinTypeAliases, typeName, typeNameLength, &typeName)) {
        typeNameLength = StringLength(typeName);
    }


    int32 arrayLength = 0;
    ConvertTokenToIntegerValue(data, parsedType.arrayLength, &arrayLength);
    ASSERT(arrayLength >= 0);


    ParsedSerializable serializable;
    if (TokenEquals(parsedType.typeName, "char") &&
        (parsedType.pointerLevels == 1 || arrayLength > 0)) {
        type.dataType = SchemaDataType_String;
        type.typeName = "char";
        type.maxStringLength = arrayLength;

        arrayLength = 0;
    }
    else if (FindParsedPersistentFunctionTypeWithName(data, typeName, typeNameLength, NULL)) {
        type.dataType = SchemaDataType_PersistentFunctionHandle;
        type.typeName = StringCopy(typeName, typeNameLength);
    }
    else if (FindParsedPersistentFunctionDataPairTypeWithName(data, typeName, typeNameLength, NULL)) {
        type.dataType = SchemaDataType_PersistentFunctionDataPair;
        type.typeName = StringCopy(typeName, typeNameLength);
    }
    else if (FindParsedAssetHandleWithName(data, typeName, typeNameLength, NULL)) {
        type.dataType = SchemaDataType_AssetHandle;
        type.typeName = StringCopy(typeName, typeNameLength);
    }
    else if (FindParsedSerializableWithName(data, typeName, typeNameLength, &serializable)) {
        switch (serializable.type) {
            case SerializableType_Struct: {
                type.dataType = SchemaDataType_Struct;
            } break;

            case SerializableType_Enum: {
                type.dataType = SchemaDataType_Enum;
            } break;

            case SerializableType_Bitfield: {
                type.dataType = SchemaDataType_Bitfield;
            } break;

            case SerializableType_Bitarray: {
                type.dataType = SchemaDataType_Bitarray;
            } break;

            case SerializableType_TaggedUnion: {
                type.dataType = SchemaDataType_TaggedUnion;
            } break;

            default: {
                ASSERT(false);
            } break;
        }
            
        type.typeName = StringCopy(typeName, typeNameLength);
    }
    else if (StringEquals("AssetType", typeName, typeNameLength)) {
        type.dataType = SchemaDataType_Enum;
        type.typeName = "AssetType";
    }
    else if (StringEquals("ComponentType", typeName, typeNameLength)) {
        type.dataType = SchemaDataType_Enum;
        type.typeName = "ComponentType";
    }
    else if (StringEquals("SceneID", typeName, typeNameLength)) {
        type.dataType = SchemaDataType_Enum;
        type.typeName = "SceneID";
    }
    else if (StringEquals("InputActionType", typeName, typeNameLength)) {
        type.dataType = SchemaDataType_Enum;
        type.typeName = "InputActionType";
    }
    else if (StringEquals("Tag", typeName, typeNameLength)) {
        type.dataType = SchemaDataType_Enum;
        type.typeName = "Tag";
    }
    else {
        type.dataType = GetSchemaDataTypeFromString(typeName, typeNameLength);
        type.typeName = StringCopy(typeName, typeNameLength);
    }

    if (parsedType.templateTypes.count > 0) {
        type.templateTypes = MakeDynamicArray<SchemaTypeDescription>(parsedType.templateTypes.count);

        for (int i = 0; i < parsedType.templateTypes.count; i++) {
            SchemaTypeDescription *templateType = PushBackPtr(&type.templateTypes);
            *templateType = SchemaTypeDescriptionFromParsedTypeDescription(data, parsedType.templateTypes[i]);
        }
    }

    type.arrayLength = arrayLength;

    type.pointerLevels = parsedType.pointerLevels;
    type.canSerialize = ParsedTypeIsValid(data, parsedType);

    return type;
}

Schema SchemaFromStructFields(PreprocessorData *data, DynamicArray<ParsedField> fields) {
    Schema schema = {};
    schema.type = SchemaType_Struct;

    schema._struct.fields = MakeDynamicArray<SchemaField>(fields.count);

    for (int i = 0; i < fields.count; i++) {
        SchemaField *schemaField = PushBackPtr(&schema._struct.fields);

        schemaField->name = StringCopy(fields[i].name.text, fields[i].name.length);
        schemaField->type = SchemaTypeDescriptionFromParsedTypeDescription(data, fields[i].type);
        schemaField->editable = HasPropertyValue(fields[i].properties, "Edit", true, true);
        schemaField->serializable = HasPropertyValue(fields[i].properties, "Serializable", true, true);
    }

    return schema;
}

Schema SchemaFromEnumValues(PreprocessorData *data, DynamicArray<PPToken> values) {
    Schema schema = {};
    schema.type = SchemaType_Enum;

    schema._enum.count = values.count;
    schema._enum.stringValues = (const char **)Plat.AllocateMem(sizeof(const char *) * schema._enum.count);

    for (int i = 0; i < schema._enum.count; i++) {
        schema._enum.stringValues[i] = StringCopy(values[i].text, values[i].length);
    }

    return schema;
}

Schema SchemaFromBitfieldValues(PreprocessorData *data, DynamicArray<PPToken> values) {
    Schema schema = {};
    schema.type = SchemaType_Bitfield;

    schema._bitfield.count = values.count;
    schema._bitfield.stringValues = (const char **)Plat.AllocateMem(sizeof(const char *) * schema._enum.count);

    for (int i = 0; i < schema._bitfield.count; i++) {
        schema._bitfield.stringValues[i] = StringCopy(values[i].text, values[i].length);
    }

    return schema;
}

Schema SchemaFromBitarrayValues(PreprocessorData *data, DynamicArray<PPToken> values) {
    Schema schema = {};
    schema.type = SchemaType_Bitarray;

    schema._bitarray.count = values.count;
    schema._bitarray.stringValues = (const char **)Plat.AllocateMem(sizeof(const char *) * schema._enum.count);

    for (int i = 0; i < schema._bitarray.count; i++) {
        schema._bitarray.stringValues[i] = StringCopy(values[i].text, values[i].length);
    }

    return schema;
}


Schema SchemaFromTaggedUnionFields(PreprocessorData *data, DynamicArray<ParsedField> fields) {
    Schema schema = {};
    schema.type = SchemaType_TaggedUnion;

    schema._tagged_union.fields = MakeDynamicArray<SchemaField>(fields.count);

    for (int i = 0; i < fields.count; i++) {
        SchemaField *schemaField = PushBackPtr(&schema._tagged_union.fields);

        schemaField->name = StringCopy(fields[i].name.text, fields[i].name.length);
        schemaField->type = SchemaTypeDescriptionFromParsedTypeDescription(data, fields[i].type);

        schemaField->editable = HasPropertyValue(fields[i].properties, "Edit", true, true);
        schemaField->serializable = HasPropertyValue(fields[i].properties, "Serializable", true, true);
    }

    return schema;
}

Schema SchemaFromParsedComponent(PreprocessorData *data, ParsedComponent *component) {
    
    const char *append = "Component";
    uint32 nameLength = component->name.length + strlen(append);
    char *name = ALLOCATE_ARRAY(char, nameLength + 1);
    snprintf(name, nameLength + 1, "%.*s%s", component->name.length, component->name.text, append);
    
    Schema schema = SchemaFromStructFields(data, component->fields);

    schema.name = name;

    schema.editable = true;
    schema.serializable = true;

    return schema;
}

Schema SchemaFromParsedSerializable(PreprocessorData *data, ParsedSerializable *serializable) {
    Schema schema;

    b32 shouldSerialize = HasPropertyValue(serializable->properties, "Serializable", true, true);

    switch (serializable->type) {
        case SerializableType_Struct: {
            schema = SchemaFromStructFields(data,serializable->_struct.fields);

            // SchemaFromStructFields() doesn't take into account
            // whether the entire struct is serializable, so force it
            // after the fact
            if (!shouldSerialize) {
                for (int i = 0; i < schema._struct.fields.count; i++) {
                    schema._struct.fields[i].serializable = false;
                }
            }
        } break;

        case SerializableType_Enum: {
            schema = SchemaFromEnumValues(data, serializable->_enum.values);
        } break;

        case SerializableType_Bitfield: {
            schema = SchemaFromBitfieldValues(data, serializable->_bitfield.values);
        } break;
            
        case SerializableType_Bitarray: {
            schema = SchemaFromBitarrayValues(data, serializable->_bitarray.values);
        } break;

        case SerializableType_TaggedUnion: {
            schema = SchemaFromTaggedUnionFields(data, serializable->_tagged_union.fields);

            if (!shouldSerialize) {
                for (int i = 0; i < schema._tagged_union.fields.count; i++) {
                    schema._tagged_union.fields[i].serializable = false;
                }
            }
        } break;

        default: {
            ASSERT(false);
        } break;
    }


    schema.name = StringCopy(serializable->name.text, serializable->name.length);

    schema.editable = HasPropertyValue(serializable->properties, "Edit", true, true); 
    schema.serializable = shouldSerialize;

    return schema;
}

Schema SchemaFromParsedPersistentFunctionType(PreprocessorData *data, ParsedPersistentFunctionType *type, Schema *funcDataPairSchema) {
    Schema schema = {};

    schema.type = SchemaType_PersistentFunctionHandle;
    schema.name = StringConcat(type->name.text, type->name.length, "Handle");

    schema.editable = true;
    schema.serializable = true;

    funcDataPairSchema->type = SchemaType_PersistentFunctionDataPair;
    funcDataPairSchema->name = StringConcat(type->name.text, type->name.length, "DataPair");
    funcDataPairSchema->editable = true;
    funcDataPairSchema->serializable = true;

    Plat.ClearMem(funcDataPairSchema->_func_data_pair.fields, sizeof(SchemaField) * 2);
    {
        SchemaField *schemaField = &funcDataPairSchema->_func_data_pair.fields[0];

        schemaField->name = "funcHandle";
        schemaField->type.dataType = SchemaDataType_PersistentFunctionHandle;
        schemaField->type.typeName = StringConcat(type->name.text, type->name.length, "Handle");
        schemaField->type.sizeInBytes = sizeof(FuncHandle);
        schemaField->type.canSerialize = true;
        
        schemaField->editable = true;
        schemaField->serializable = true;
        
    }
    {
        SchemaField *schemaField = &funcDataPairSchema->_func_data_pair.fields[1];

        schemaField->name = "data";
        schemaField->type.dataType = SchemaDataType_SchemaData;
        schemaField->type.typeName = "SchemaData";
        schemaField->type.sizeInBytes = sizeof(SchemaData);
        schemaField->type.canSerialize = true;
        
        schemaField->editable = true;
        schemaField->serializable = true;
    }

    return schema;
}

Schema SchemaFromParsedAsset(PreprocessorData *data, ParsedSerializable *asset) {
    Schema schema = {};

    ASSERT(asset->type == SerializableType_Struct);

    schema = SchemaFromStructFields(data, asset->_struct.fields);
    schema.name = StringCopy(asset->name.text, asset->name.length);

    schema.editable = HasPropertyValue(asset->properties, "Edit", true, true); 
    schema.serializable = HasPropertyValue(asset->properties, "Serializable", true, true);

    // @NOTE: don't bother setting _func_handle.funcType. That code is
    // generated by the preprocessor, so it's not defined in the
    // preprocessor itself
    if (!schema.serializable) {
        for (int i = 0; i < schema._struct.fields.count; i++) {
            schema._struct.fields[i].serializable = false;
        }
    }

    schema._struct.isAsset = true;

    return schema;
}

Schema SchemaFromParsedAssetHandle(PreprocessorData *data, ParsedSerializable *asset) {
    Schema schema = {};

    schema.type = SchemaType_AssetHandle;
    schema.name = StringConcat(asset->name.text, asset->name.length, "Handle");

    schema.editable = true;
    schema.serializable = true;

    return schema;
}


b32 ParseSourceFile(PreprocessorData *data, const char *sourceFileName) {
    const char *directoryName;
    const char *fileNameWithExtension;
    const char *fileNameNoExtension;
    const char *extension;
    SplitPath(sourceFileName, &directoryName, &fileNameWithExtension);
    SplitFileExtension(fileNameWithExtension, &fileNameNoExtension, &extension);

    directoryName = PathConcat(data->srcDirectory, directoryName);
        
    const char *filePath = Plat.FindFileInDirectory(directoryName, fileNameNoExtension, 1, &extension);

    if (filePath == NULL) {
        return false;
    }

    Tokenizer tokenizer = {};

    if (!Plat.OpenFileForRead(filePath, &tokenizer.filehandle)) {
        PreprocessorSetError(data, PreprocessorError_FileLoadFailed, "Couldn't open file '%s'", filePath);
        return false;
    }

    PushBack(&data->openedFiles, tokenizer.filehandle);

    tokenizer.buffer = (char *)tokenizer.filehandle.data;
    tokenizer.filePath = filePath;
    tokenizer.line = 1;
    tokenizer.column = 1;

    DynamicArray<PPToken> tokens = TokenizeFile(&tokenizer);

    b32 result = ParseFile(tokens, data);

    return result;
}

void GenerateStep(PreprocessorData *data) {

    Plat.MakeDirectory(data->generatedDirectory);

    // Code Generation Step
    HashTable<char *, int32> componentIndices;
    AllocateHashTable(&componentIndices, data->definedComponents.count);

    uint32 numBuiltinTypes = data->builtinTypes.count;

    // Convert all arrayLengths to integer values
    for (int i = 0; i < data->definedComponents.count; i++) {
        ParsedComponent *component = data->definedComponents.data + i;
        for (int j = 0; j < component->fields.count; j++) {
            if (component->fields[j].type.arrayLength.type != PPTokenType_Unknown) {
                ConvertTokenToIntegerValue(data, component->fields[j].type.arrayLength,
                                           &component->fields[j].type.arrayLengthValue);
            }
        }
    }

    for (int i = 0; i < data->definedSystems.count; i++) {
        ParsedSystem *system = data->definedSystems.data + i;
        for (int j = 0; j < system->fields.count; j++) {
            if (system->fields[j].type.arrayLength.type != PPTokenType_Unknown) {
                ConvertTokenToIntegerValue(data, system->fields[j].type.arrayLength,
                                           &system->fields[j].type.arrayLengthValue);
            }
        }
    }

    for (int i = 0; i < data->serializableTypes.count; i++) {
        ParsedSerializable *serializable = data->serializableTypes.data + i;
        if (serializable->type == SerializableType_Struct) {
            for (int j = 0; j < serializable->_struct.fields.count; j++) {
                if (serializable->_struct.fields[j].type.arrayLength.type != PPTokenType_Unknown) {
                    ConvertTokenToIntegerValue(data,
                                               serializable->_struct.fields[j].type.arrayLength,
                                               &serializable->_struct.fields[j].type.arrayLengthValue);
                }
            }
        }
    }


    // @TODO: evaluate properties on all serializable types


    // Make sure there are no duplicate Components
    for (int32 i = 0; i < data->definedComponents.count; i++) {
        ParsedComponent component = data->definedComponents[i];

        if (HashTableContains(&componentIndices, component.name.text, component.name.length)) {
            PreprocessorSetError(data, PreprocessorError_DuplicateComponent, 
                                 "%s:%d:%d: Duplicate component '%.*s'",
                                 component.name.filePath, component.name.line, component.name.column,
                                 component.name.length, component.name.text);
            return;
        }

        HashTableInsert(&componentIndices, component.name.text, component.name.length, i);
    }

    // Make sure all Components are paired with a System and that there are no duplicate Systems
    for (int i = 0; i < data->definedSystems.count; i++) {
        ParsedSystem *system = &data->definedSystems[i];
        int32 index;

        if (HashTableGet(&componentIndices, system->name.text, system->name.length, &index)) {
            if (index >= 0) {
                // Write -1 into the HashTable so it's clear if a system name is duplicated
                HashTableRemove(&componentIndices, system->name.text, system->name.length);
                HashTableInsert(&componentIndices, system->name.text, system->name.length, -1);
            }
            else {
                PreprocessorSetError(data, PreprocessorError_DuplicateSystem,
                                     "%s:%d:%d: Duplicate system '%.*s'",
                                     system->name.filePath, system->name.line, system->name.column,
                                     system->name.length, system->name.text);
                return;
            }
        }
        else {
            PreprocessorSetError(data, PreprocessorError_MissingComponent,
                                 "No component matching %.*sSystem", system->name.length, system->name.text);
            return;
        }
    }

    // Sort Systems and Components by priority
    qsort(data->definedSystems.data, data->definedSystems.count, sizeof(ParsedSystem), ParsedSystemComparator);

    DynamicArray<ParsedComponent> sortedComponents = MakeDynamicArray<ParsedComponent>(data->definedComponents.count);

    for (int i = 0; i < data->definedSystems.count; i++) {
        ParsedSystem system = data->definedSystems[i];

        for (int j = 0; j < data->definedComponents.count; j++) {
            ParsedComponent component = data->definedComponents[j];

            if (memcmp(system.name.text, component.name.text, system.name.length) == 0) {
                PushBack(&sortedComponents, component);
                break;
            }
        }
    }

    DeallocateDynamicArray(&data->definedComponents);
    data->definedComponents = sortedComponents;


    // Construct Initial validTypeSet
    for (int i = 0; i < numBuiltinTypes; i++) {
        const char *name = data->builtinTypes[i];
        HashTableInsert(&data->validTypeSet, name, strlen(name), 0, true);
    }

    for (int i = 0; i < data->definedFunctionTypes.count; i++) {
        ParsedPersistentFunctionType type = data->definedFunctionTypes[i];
        const char *name = StringConcat(type.name.text, type.name.length, "Handle");
        HashTableInsert(&data->validTypeSet, name, strlen(name), 0, true);

        const char *funcDataPairName = StringConcat(type.name.text, type.name.length, "DataPair");
        HashTableInsert(&data->validTypeSet, funcDataPairName, strlen(funcDataPairName), 0, true);
    }

    for (int i = 0; i < data->definedAssets.count; i++) {
        ParsedSerializable asset = data->definedAssets[i];
        const char *handleName = StringConcat(asset.name.text, asset.name.length, "Handle");
        HashTableInsert(&data->validTypeSet, handleName, strlen(handleName), 0, true);

        HashTableInsert(&data->validTypeSet, (const char *)asset.name.text, asset.name.length, 0, true);
    }

    HashTableInsert(&data->validTypeSet, "InputActionType", 0, true);
    HashTableInsert(&data->validTypeSet, "Tag", 0, true);
    HashTableInsert(&data->validTypeSet, "ComponentType", 0, true);
    HashTableInsert(&data->validTypeSet, "AssetType", 0, true);
    HashTableInsert(&data->validTypeSet, "SceneID", 0, true);


    DynamicArray<ParsedField> invalidFields = {};

    for (int i = 0; i < data->serializableTypes.count; i++) {
        ParsedSerializable serializable = data->serializableTypes[i];

        if (HasPropertyValue(serializable.properties, "Serializable", true, true)) {
            b32 allFieldsValid = true;
            switch (serializable.type) {
                case SerializableType_Struct: {
                    for (int j = 0; j < serializable._struct.fields.count; j++) {
                        ParsedField field = serializable._struct.fields[j];
                        if (HasPropertyValue(field.properties, "Serializable", true, true) &&
                            !ParsedTypeIsValid(data, field.type)) {

                            PushBack(&invalidFields, field);
                            allFieldsValid = false;
                        }
                    }
                } break;

                case SerializableType_TaggedUnion: {
                    for (int j = 0; j < serializable._tagged_union.fields.count; j++) {
                        ParsedField field = serializable._tagged_union.fields[j];
                        if (!ParsedTypeIsValid(data, field.type)) {

                            PushBack(&invalidFields, field);
                            allFieldsValid = false;
                        }
                    }
                } break;

                default: {
                } break;
            }

            if (allFieldsValid) {
                HashTableInsert(&data->validTypeSet, (const char *)serializable.name.text, serializable.name.length, 0, true);
            }
        }
    }

    // Ensure that all field types in defined components are serializable
    // @TODO: it's probably best to combine this with the loop over
    // serializable types to solve some potential ordering issues
    for (int i = 0; i < data->definedComponents.count; i++) {
        ParsedComponent component = data->definedComponents[i];

        for (int j = 0; j < component.fields.count; j++) {
            ParsedField field = component.fields[j];
            if (!ParsedTypeIsValid(data, field.type)) {
                PushBack(&invalidFields, field);
            }
        }
    }

    if (invalidFields.count > 0) {
        printf("Unable to resolve dependencies in serializable types:\n");

        for (int i = 0; i < invalidFields.count; i++) {
            ParsedField field = invalidFields[i];

            printf("%s:%d:%d: field '%.*s' of type %.*s is not serializable\n",
                   field.type.typeName.filePath, field.type.typeName.line,
                   field.type.typeName.column,
                   field.name.length, field.name.text,
                   field.type.typeName.length, field.type.typeName.text);
            
        }

        PreprocessorSetError(data, PreprocessorError_UnserializableField);
        return;
    }


    // Make sure Persistent Functions only reference serializable types
    for (int i = 0; i < data->definedPersistentFunctions.count; i++) {
        ParsedFunction function = data->definedPersistentFunctions[i];

        if (!IsValidPersistentFunction(data, function)) {
            return;
        }
    }


    // Construct Schemas for all types
    DynamicArray<Schema> schemas = {};

    {
        Schema schema;
        schema.name = "InputActionType";
        schema.type = SchemaType_Enum;
        schema.serializable = true;
        schema.editable = true;

        PushBack(&schemas, schema);
    }

    
    {
        Schema schema;
        schema.name = "Tag";
        schema.type = SchemaType_Enum;
        schema.serializable = true;
        schema.editable = true;

        PushBack(&schemas, schema);
    }

    {
        Schema schema;
        schema.name = "AssetType";
        schema.type = SchemaType_Enum;
        schema.serializable = true;
        schema.editable = true;

        PushBack(&schemas, schema);
    }

    {
        Schema schema;
        schema.name = "ComponentType";
        schema.type = SchemaType_Enum;
        schema.serializable = true;
        schema.editable = true;

        PushBack(&schemas, schema);
    }

    {
        Schema schema;
        schema.name = "SceneID";
        schema.type = SchemaType_Enum;
        schema.serializable = true;
        schema.editable = true;

        PushBack(&schemas, schema);
    }

    for (int i = 0; i < data->definedAssets.count; i++) {
        ParsedSerializable *asset = &data->definedAssets[i];
        Schema handleSchema = SchemaFromParsedAssetHandle(data, asset);
        PushBack(&schemas, handleSchema);

        Schema schema = SchemaFromParsedAsset(data, asset);
        PushBack(&schemas, schema);
    }

    for (int i = 0; i < data->serializableTypes.count; i++) {
        ParsedSerializable *serializable = &data->serializableTypes[i];

        if (schemas.count > 0 &&
            FindString(&schemas[0].name, schemas.count, serializable->name.text, serializable->name.length, NULL, sizeof(Schema)
)) {
            continue;
        }
        
        Schema schema = SchemaFromParsedSerializable(data, serializable);

        PushBack(&schemas, schema);
    }

    for (int i = 0; i < data->definedComponents.count; i++) {
        ParsedComponent *component = &data->definedComponents[i];
        Schema schema = SchemaFromParsedComponent(data, component);

        PushBack(&schemas, schema);
    }

    for (int i = 0; i < data->definedFunctionTypes.count; i++) {
        ParsedPersistentFunctionType *type = &data->definedFunctionTypes[i];
        Schema funcDataPairSchema = {};
        Schema schema = SchemaFromParsedPersistentFunctionType(data, type, &funcDataPairSchema);

        PushBack(&schemas, schema);
        PushBack(&schemas, funcDataPairSchema);
    }


    // Find all functions to put in the function tables
    data->addComponentFunctions = ALLOCATE_CLEARED_ARRAY(ParsedFunction, data->definedComponents.count);
    data->removeComponentFunctions = ALLOCATE_CLEARED_ARRAY(ParsedFunction, data->definedComponents.count);
    data->getComponentValueFunctions = ALLOCATE_CLEARED_ARRAY(ParsedFunction, data->definedComponents.count);
    data->setComponentValueFunctions = ALLOCATE_CLEARED_ARRAY(ParsedFunction, data->definedComponents.count);
    data->tickSystemFunctions = ALLOCATE_CLEARED_ARRAY(ParsedFunction, data->definedComponents.count);
    data->raycastSystemFunctions = ALLOCATE_CLEARED_ARRAY(ParsedFunction, data->definedComponents.count);
    data->customInputComponentValueFunctions = ALLOCATE_CLEARED_ARRAY(ParsedFunction, data->definedComponents.count);
    data->inspectSystemValueFunctions = ALLOCATE_CLEARED_ARRAY(ParsedFunction, data->definedComponents.count);
    data->sceneInitFunctions = ALLOCATE_CLEARED_ARRAY(ParsedFunction, data->definedScenes.count);
    data->sceneDeinitFunctions = ALLOCATE_CLEARED_ARRAY(ParsedFunction, data->definedScenes.count);
    data->sceneUpdateFunctions = ALLOCATE_CLEARED_ARRAY(ParsedFunction, data->definedScenes.count);

    data->assetLoadFileFunctions = ALLOCATE_CLEARED_ARRAY(ParsedFunction, data->definedAssets.count);
    data->assetLoadMainThreadFunctions = ALLOCATE_CLEARED_ARRAY(ParsedFunction, data->definedAssets.count);
    data->assetUnloadFileFunctions = ALLOCATE_CLEARED_ARRAY(ParsedFunction, data->definedAssets.count);


    enum ParsedFunctionCategory {
        FunctionCategory_Component,
        FunctionCategory_Scene,
        FunctionCategory_Asset,
    };

    struct ParsedFunctionPattern {
        const char *pattern;
        ParsedFunctionCategory category;
        ParsedFunction *functionArray;
    };

    ParsedFunctionPattern functionPatterns[] = {
        {"Add(.+)Component$", FunctionCategory_Component, data->addComponentFunctions},
        {"Remove(.+)Component$", FunctionCategory_Component, data->removeComponentFunctions},
        {"Get(.+)ComponentValue$", FunctionCategory_Component, data->getComponentValueFunctions},
        {"Set(.+)ComponentValue$", FunctionCategory_Component, data->setComponentValueFunctions},
        {"Tick(.+)System$", FunctionCategory_Component, data->tickSystemFunctions},
        {"Raycast(.+)System$", FunctionCategory_Component, data->raycastSystemFunctions},

        {"Input(.+)ComponentValue$", FunctionCategory_Component, data->customInputComponentValueFunctions},
        {"Inspect(.+)SystemValue$", FunctionCategory_Component, data->inspectSystemValueFunctions},

        {"(.+)Init$", FunctionCategory_Scene, data->sceneInitFunctions},
        {"(.+)Deinit$", FunctionCategory_Scene, data->sceneDeinitFunctions},
        {"(.+)Update$", FunctionCategory_Scene, data->sceneUpdateFunctions},

        {"(.+)LoadFile", FunctionCategory_Asset, data->assetLoadFileFunctions},
        {"(.+)LoadMainThread", FunctionCategory_Asset, data->assetLoadMainThreadFunctions},
        {"(.+)UnloadFile", FunctionCategory_Asset, data->assetUnloadFileFunctions},
    };

    DynamicArray<Reprog *> functionRegexes = {};
    for (int i = 0; i < ARRAY_LENGTH(ParsedFunctionPattern, functionPatterns); i++) {
        PushBack(&functionRegexes, regcomp(functionPatterns[i].pattern, 0, NULL));
    }

    for (int i = 0; i < data->definedFunctions.count; i++) {
        ParsedFunction function = data->definedFunctions[i];

        char functionName[2048];
        StringCopy(functionName, function.name.length + 1, function.name.text);

        for (int j = 0; j < functionRegexes.count; j++) {
            Reprog *regex = functionRegexes[j];
            Resub match = {};

            if (regexec(regex, functionName, &match, 0) == 0) {
                if (match.nsub > 0) {
                    const char *matchString = match.sub[1].sp;
                    uint32 matchLength = match.sub[1].ep - match.sub[1].sp;

                    switch (functionPatterns[j].category) {
                        case FunctionCategory_Component: {
                            int32 index;
                            if (ComponentTypeExists(data, matchString, matchLength, &index)) {

                                if (functionPatterns[j].functionArray[index].name.text == NULL) {
                                    functionPatterns[j].functionArray[index] = function;
                                }
                                else {
                                    PreprocessorSetError(data, PreprocessorError_DuplicateFunction,
                                                         "%s:%d:%d: Duplicate function '%.*s'",
                                                         function.name.filePath, function.name.line, function.name.column,
                                                         function.name.length, function.name.text);
                                    return;
                                }
                            }
                        } break;

                        case FunctionCategory_Scene: {
                            int32 index;
                            if (SceneExists(data, matchString, matchLength, &index)) {
                                if (functionPatterns[j].functionArray[index].name.text == NULL) {
                                    functionPatterns[j].functionArray[index] = function;
                                }
                                else {
                                    PreprocessorSetError(data, PreprocessorError_DuplicateFunction,
                                                         "%s:%d:%d: Duplicate function '%.*s'",
                                                         function.name.filePath, function.name.line, function.name.column,
                                                         function.name.length, function.name.text);
                                    return;
                                }
                            }
                        } break;

                        case FunctionCategory_Asset: {
                            int32 index;
                            if (AssetTypeExists(data, matchString, matchLength, &index)) {

                                if (functionPatterns[j].functionArray[index].name.text == NULL) {
                                    functionPatterns[j].functionArray[index] = function;
                                }
                                else {
                                    PreprocessorSetError(data, PreprocessorError_DuplicateFunction,
                                                         "%s:%d:%d: Duplicate function '%.*s'",
                                                         function.name.filePath, function.name.line, function.name.column,
                                                         function.name.length, function.name.text);
                                    return;
                                }
                            }
                        } break;
                    }
                }
            }
        }
    }

    for (int i = 0; i < functionRegexes.count; i++) {
        regfree(functionRegexes[i]);
    }

    DeallocateDynamicArray(&functionRegexes);


    for (int i = 0; i < data->definedScenes.count; i++) {
        ParsedScene scene = data->definedScenes[i];

#if !ETC_DEBUG
        if (HasProperty(scene.properties, "Editor") ||
            HasProperty(scene.properties, "Test")) {
            continue;
        }
#endif
    }


    // Write generated files
    {
        char *path = PathConcat(data->generatedDirectory, "component_list.generated.h");
        WriteComponentList(path, data);
        Plat.DeallocateMem(path);
    }

    {
        char *scenesListPath = PathConcat(data->generatedDirectory, "scene_list.generated.h");
        WriteScenesList(scenesListPath, data);
        Plat.DeallocateMem(scenesListPath);
    }

    {
        char *headerPath = PathConcat(data->generatedDirectory, "serialized_types.generated.h");

        FileHandle file;
        if (Plat.OpenFileForWrite(headerPath, &file, Kilobytes(512))) {
            FileClear(&file);
        
            WriteSerializableTypeHeader(&file, data, data->serializableTypes);

            Plat.CloseFile(&file);
        
            Plat.DeallocateMem(headerPath);
        }
    }

    {
        char *headerPath = PathConcat(data->generatedDirectory, "serialized_function_types.generated.h");

        FileHandle file;
        if (Plat.OpenFileForWrite(headerPath, &file, Kilobytes(512))) {
            FileClear(&file);
        
            WriteDefinedFunctionTypes(&file, data);

            Plat.CloseFile(&file);
        
            Plat.DeallocateMem(headerPath);
        }
    }

    
    {
        char *headerPath = PathConcat(data->generatedDirectory, "input.generated.h");

        FileHandle file;
        if (Plat.OpenFileForWrite(headerPath, &file, Kilobytes(512))) {
            FileClear(&file);
        
            WriteInputActions(&file, data);

            Plat.CloseFile(&file);
        
            Plat.DeallocateMem(headerPath);
        }
    }

    {
        char *headerPath = PathConcat(data->generatedDirectory, "tags.generated.h");

        FileHandle file;
        if (Plat.OpenFileForWrite(headerPath, &file, Kilobytes(512))) {
            FileClear(&file);
        
            WriteTags(&file, data);

            Plat.CloseFile(&file);
        
            Plat.DeallocateMem(headerPath);
        }
    }

    {
        char *path = PathConcat(data->generatedDirectory, "schemas.generated.cpp");

        FileHandle file;
        if (Plat.OpenFileForWrite(path, &file, Kilobytes(512))) {
            FileClear(&file);

            WriteGenerateSchemasFunction(&file, schemas);

            Plat.CloseFile(&file);
        }

        Plat.DeallocateMem(path);
    }

    {
        char *path = PathConcat(data->generatedDirectory, "function_tables.generated.cpp");

        FileHandle file;
        if (Plat.OpenFileForWrite(path, &file, Kilobytes(512))) {
            FileClear(&file);

            WriteFunctionTables(&file, data);

            WritePersistentFunctionTable(&file, data);

            Plat.CloseFile(&file);
        }

        Plat.DeallocateMem(path);
    }

    {
        char *headerPath = PathConcat(data->generatedDirectory, "assets.generated.h");

        FileHandle file;
        if (Plat.OpenFileForWrite(headerPath, &file, Kilobytes(512))) {
            FileClear(&file);
        
            WriteAssetList(&file, data);

            Plat.CloseFile(&file);
        
            Plat.DeallocateMem(headerPath);
        }
    }

    {
        char *headerPath = PathConcat(data->generatedDirectory, "assets.generated.cpp");

        FileHandle file;
        if (Plat.OpenFileForWrite(headerPath, &file, Kilobytes(512))) {
            FileClear(&file);
        
            WriteAssetTables(&file, data);

            Plat.CloseFile(&file);
        
            Plat.DeallocateMem(headerPath);
        }
    }

}

int main(int argc, const char **args) {
    ASSERT(argc == 3);

    PreprocessorData preprocessorData = {};
    AllocatePreprocessorData(&preprocessorData);

    preprocessorData.srcDirectory = args[1];
    preprocessorData.generatedDirectory = args[2];

    RemoveTrailingSlashes((char *)preprocessorData.srcDirectory);
    RemoveTrailingSlashes((char *)preprocessorData.generatedDirectory);

    b32 parseSuccess = ParseSourceFile(&preprocessorData, "etcetera.cpp");
    
    if (preprocessorData.error.type != PreprocessorError_None) {
        goto error;
    }

    GenerateStep(&preprocessorData);
    if (preprocessorData.error.type != PreprocessorError_None) {
        goto error;
    }


    // Cleanup
 error:
    DeallocatePreprocessorData(&preprocessorData);

    int32 errorCode = 0;

    if (preprocessorData.error.type != PreprocessorError_None) {
        errorCode = -(int32)preprocessorData.error.type;

        if (strlen(preprocessorData.error.message) > 0) {
            printf("%s\n", preprocessorData.error.message);
        }
    }

    return errorCode;
}

