#ifndef ETC_PREPROCESSOR_H
#define ETC_PREPROCESSOR_H


#define DEFAULT_SYSTEM_PRIORITY 50

// Parsing
enum PPTokenType {
    PPTokenType_Unknown,
    PPTokenType_EndOfFile,
    PPTokenType_Colon,
    PPTokenType_Semicolon,
    PPTokenType_Period,
    PPTokenType_Comma,
    PPTokenType_Arrow,
    PPTokenType_OpenParen,
    PPTokenType_CloseParen,
    PPTokenType_OpenBracket,
    PPTokenType_CloseBracket,
    PPTokenType_OpenBrace,
    PPTokenType_CloseBrace,

    PPTokenType_Plus,
    PPTokenType_Minus,
    PPTokenType_Slash,
    PPTokenType_Asterisk,
    PPTokenType_Ampersand,
    PPTokenType_Pipe,
    PPTokenType_Exclamation,
    PPTokenType_LessThan,
    PPTokenType_GreaterThan,
    PPTokenType_Equal,
    PPTokenType_Question,
    PPTokenType_Hash,

    PPTokenType_Bool,
    PPTokenType_Float,
    PPTokenType_Integer,
    PPTokenType_Char,
    PPTokenType_String,
    PPTokenType_Identifier,
};

struct PPToken {
    PPTokenType type;
    char *text;
    uint32 length;

    const char *filePath;
    uint32 line;
    uint32 column;

    union {
        bool boolValue;
        real32 real32Value;
        int32 int32Value;
        char charValue;
        char *stringValue;
    };
};


// Properties
struct ParsedProperty {
    PPToken name;

    b32 hasValue;
    PPToken value;
};

b32 HasProperty(DynamicArray<ParsedProperty> *properties, const char *propertyName, ParsedProperty *outProperty = NULL);
b32 PropertyIsTrue(DynamicArray<ParsedProperty> *properties, const char *propertyName, b32 testValue, b32 defaultValue = false);


struct ParsedTypeDescription {
    PPToken typeName;
    uint32 pointerLevels;

    PPToken arrayLength;
    int32 arrayLengthValue;

    b32 isConst;

    DynamicArray<ParsedTypeDescription> templateTypes;
};

b32 ParsedTypeDescriptionEquals(ParsedTypeDescription a, ParsedTypeDescription b);

struct ParsedFunctionArgument {
    PPToken name;
    ParsedTypeDescription type;

    b32 hasDefaultValue;
    PPToken defaultValue;
};

struct ParsedFunction {
    PPToken name;

    ParsedTypeDescription returnType;
    DynamicArray<ParsedFunctionArgument> arguments;

    b32 isPersistent;
    PPToken persistentFunctionType;
    PPToken requiredSchemaDataType;
};

struct ParsedField {
    PPToken name;

    ParsedTypeDescription type;
    DynamicArray<ParsedProperty> properties;
};

struct ParsedComponent {
    PPToken name;
    DynamicArray<ParsedField> fields;

    DynamicArray<ParsedProperty> properties;
};

struct ParsedSystem {
    PPToken name;
    DynamicArray<ParsedField> fields;

    DynamicArray<ParsedProperty> properties;
    uint32 priority;
};

struct ParsedScene {
    PPToken name;
    DynamicArray<ParsedProperty> properties;
};

struct ParsedInputAction {
    PPToken name;
    b32 analogue;
    // @MAYBE @TODO: editor/gameplay property
};

struct ParsedTag {
    PPToken name;
};

enum SerializableType {
    SerializableType_Struct,
    SerializableType_Enum,
    SerializableType_Bitfield,
    SerializableType_Bitarray,
    SerializableType_TaggedUnion,
};

struct ParsedSerializable {
    SerializableType type;
    PPToken name;
    DynamicArray<ParsedProperty> properties;

    union {
        struct {
            DynamicArray<ParsedField> fields;
        } _struct;

        struct {
            DynamicArray<PPToken> values;
        } _enum;

        struct {
            DynamicArray<PPToken> values;
        } _bitfield;

        struct {
            DynamicArray<PPToken> values;
        } _bitarray;

        struct {
            DynamicArray<ParsedField> fields;
        } _tagged_union;
    };
};

struct ParsedDefinedInt {
    PPToken name;
    int32 value;
};

struct ParsedPersistentFunctionType {
    PPToken name;

    ParsedTypeDescription returnType;
    DynamicArray<ParsedFunctionArgument> arguments;
};


inline void DeallocateParsedTypeDescription(ParsedTypeDescription *type) {
    for (int i = 0; i < type->templateTypes.count; i++) {
        DeallocateParsedTypeDescription(&type->templateTypes[i]);
    }

    DeallocateDynamicArray(&type->templateTypes);
}


// Preprocessor
#define MAX_ERROR_MESSAGE_LENGTH 2048

enum PreprocessorErrorType {
    PreprocessorError_None,

    PreprocessorError_FileLoadFailed,
    PreprocessorError_UnexpectedToken,

    PreprocessorError_DuplicateComponent = 16,
    PreprocessorError_DuplicateSystem,
    PreprocessorError_MissingComponent,
    PreprocessorError_UnserializableField,
    PreprocessorError_DuplicateFunction,
    PreprocessorError_MissingFunction,

    PreprocessorError_FileGenerationFailed,
};

struct PreprocessorError {
    PreprocessorErrorType type;

    uint32 messageLength;
    char *message;
};

struct PreprocessorData {
    const char *srcDirectory;
    const char *generatedDirectory;


    DynamicArray<const char *> sourceFilePaths;
    DynamicArray<FileHandle> openedFiles;

    DynamicArray<const char *> builtinTypes;
    HashTable<const char *, const char *> builtinTypeAliases;

    DynamicArray<ParsedFunction> definedFunctions;
    DynamicArray<ParsedFunction> definedPersistentFunctions;
    DynamicArray<ParsedPersistentFunctionType> definedFunctionTypes;
    DynamicArray<ParsedSerializable> definedAssets;
    DynamicArray<DynamicArray<PPToken>> definedAssetExtensions;
    DynamicArray<ParsedComponent> definedComponents;
    DynamicArray<ParsedSystem> definedSystems;
    DynamicArray<ParsedScene> definedScenes;
    DynamicArray<ParsedInputAction> definedActions;
    DynamicArray<ParsedTag> definedTags;

    HashTable<const char *, int32> processedFilePaths;

    DynamicArray<ParsedSerializable> serializableTypes;
    DynamicArray<ParsedDefinedInt> definedInts;

    HashTable<char *, const char *> symbolSourceFileMap;

    ParsedFunction *addComponentFunctions;
    ParsedFunction *removeComponentFunctions;
    ParsedFunction *getComponentValueFunctions;
    ParsedFunction *setComponentValueFunctions;
    ParsedFunction *tickSystemFunctions;
    ParsedFunction *raycastSystemFunctions;

    ParsedFunction *customInputComponentValueFunctions;
    ParsedFunction *inspectSystemValueFunctions;

    ParsedFunction *sceneInitFunctions;
    ParsedFunction *sceneDeinitFunctions;
    ParsedFunction *sceneUpdateFunctions;

    ParsedFunction *assetLoadFileFunctions;
    ParsedFunction *assetLoadMainThreadFunctions;
    ParsedFunction *assetUnloadFileFunctions;

    HashTable<const char *, int32> validTypeSet;

    PreprocessorError error;

    MemoryArenaList arena;
};


#define INSERT_BUILTIN_TYPE(builtin, size, list) PushBack(list, #builtin);
#define INSERT_BUILTIN_TYPE_ALIAS(alias, builtin, list) PushBack(list, #alias);

#define INSERT_BUILTIN_TYPE_ALIAS_MAP(alias, builtin, table) HashTableInsert(table, #alias, #builtin, true);


inline void AllocatePreprocessorData(PreprocessorData *data) {
    AllocateHashTable(&data->symbolSourceFileMap, 4096);

    FOREACH_SCHEMADATATYPE(INSERT_BUILTIN_TYPE, &data->builtinTypes);
    FOREACH_SCHEMADATATYPE_ALIAS(INSERT_BUILTIN_TYPE_ALIAS, &data->builtinTypes);
    
    AllocateHashTable(&data->builtinTypeAliases, 1024);
    FOREACH_SCHEMADATATYPE_ALIAS(INSERT_BUILTIN_TYPE_ALIAS_MAP, &data->builtinTypeAliases);


    data->error.messageLength = 2048;
    data->error.message = ALLOCATE_ARRAY(char, data->error.messageLength);

    AllocateMemoryArenaList(&data->arena, Kilobytes(512));
}

inline void DeallocatePreprocessorData(PreprocessorData *data) {
    for (int i = 0; i < data->openedFiles.count; i++) {
        Plat.CloseFile(&data->openedFiles[i]);
    }

    DeallocateDynamicArray(&data->sourceFilePaths);
    DeallocateDynamicArray(&data->openedFiles);
    DeallocateDynamicArray(&data->builtinTypes);
    DeallocateHashTable(&data->builtinTypeAliases);

    DeallocateDynamicArray(&data->definedComponents);
    DeallocateDynamicArray(&data->definedSystems);
    DeallocateDynamicArray(&data->definedScenes);
    DeallocateDynamicArray(&data->serializableTypes);
    DeallocateDynamicArray(&data->definedInts);

    DeallocateHashTable(&data->symbolSourceFileMap);

    Plat.DeallocateMem(data->error.message);
}

inline void PreprocessorSetError(PreprocessorData *data, PreprocessorErrorType type, const char *fmt = NULL, ...) {
    data->error.type = type;

    if (fmt != NULL) {
        va_list args;
        va_start(args, fmt);
        vsnprintf(data->error.message, data->error.messageLength, fmt, args);
        va_end(args);
    }
    else {
        data->error.message[0] = '\0';
    }
}

inline void PreprocessorClearError(PreprocessorData *data) {
    data->error.type = PreprocessorError_None;
    data->error.message[0] = '\0';
}

b32 ParseSourceFile(PreprocessorData *data, const char *fileName);


// The SchemaDataTypeMap is used to map from string type names to a SchemaDataType enum value.
struct SchemaDataTypeMap {
    const char *name;
    SchemaDataType dataType;
};

#define GENERATE_SCHEMADATATYPEMAP(type, size, _) {#type, SchemaDataType_##type},

const SchemaDataTypeMap SchemaDataTypeMap_List[] = {
    {NULL, SchemaDataType_Invalid},

    FOREACH_SCHEMADATATYPE(GENERATE_SCHEMADATATYPEMAP, args)

    {"String", SchemaDataType_String},
    {"Struct", SchemaDataType_Struct},
    {"Enum", SchemaDataType_Enum},
    {"Bitfield", SchemaDataType_Bitfield},
    {"Bitarray", SchemaDataType_Bitarray},
    {"TaggedUnion", SchemaDataType_TaggedUnion},
    {"PersistentFunctionHandle", SchemaDataType_PersistentFunctionHandle},
    {"PersistentFunctionDataPair", SchemaDataType_PersistentFunctionDataPair},
    {"AssetHandle", SchemaDataType_AssetHandle},
};

inline const char *GetSchemaDataTypeName(SchemaDataType type) {
    const char *result = SchemaDataTypeMap_List[type].name;
    return result;
}

inline SchemaDataType GetSchemaDataTypeFromString(const char *typeName, uint32 length) {
    SchemaDataType result = SchemaDataType_Invalid;

    for (int i = 1; i < ARRAY_LENGTH(SchemaDataTypeMap, SchemaDataTypeMap_List); i++) {
        if (StringEquals(typeName, SchemaDataTypeMap_List[i].name, length, StringLength(SchemaDataTypeMap_List[i].name))) {
            result = SchemaDataTypeMap_List[i].dataType;
            break;
        }
    }

    return result;
}

inline SchemaDataType GetSchemaDataTypeFromString(const char *typeName) {
    return GetSchemaDataTypeFromString(typeName, strlen(typeName) + 1);
}

#endif

