
void AllocateCommandList(CommandList *commandList, uint32 capacity, uint32 arenaSize) {
    AllocateMemoryArena(&commandList->arena, arenaSize);
    commandList->commandCapacity = capacity;
    commandList->commands = (EnqueuedRenderCommand *)Plat.AllocateMem(sizeof(EnqueuedRenderCommand) * capacity);
    commandList->sortedCommandIndices = (uint32 *)Plat.AllocateMem(sizeof(uint32) * capacity);
    commandList->tempSortedCommandIndices = (uint32 *)Plat.AllocateMem(sizeof(uint32) * capacity);
}

void *AllocateRenderCommand_(CommandList *commandList, uint32 key, uint32 commandSize, RenderBackendFunction *backendFn) {
    ASSERT(commandList->commandCount < commandList->commandCapacity);

    // TODO @hz: yo, this isn't threadsafe yet

    void *command = PushSize(&commandList->arena, commandSize);
    
    uint32 index = commandList->commandCount;
    commandList->commands[index].key = key;
    commandList->commands[index].backendFn = backendFn;
    commandList->commands[index].data = command;

    // This insertion is so the sort doesn't have to do a prepass
    commandList->sortedCommandIndices[index] = index;

    commandList->commandCount++;

    return command;
}

void ClearCommandList(CommandList *commandList) {
    commandList->commandCount = 0;
    commandList->arena.used = 0;
}

void CommandListRadixSort8Pass(CommandList *commandList, uint32 pass) {
    EnqueuedRenderCommand *commands = commandList->commands;

    // The sorted values from this pass are written into `temp`
    uint32 *sorted = commandList->sortedCommandIndices;
    uint32 *temp = commandList->tempSortedCommandIndices;

    uint32 shift = 8 * pass;
    const uint32 mask = 0xFF;
    const uint32 numBuckets = 1 << 8;

    uint32 histogram[numBuckets] = {};

    for (int i = 0; i < commandList->commandCount; i++) {
        uint32 commandIndex = sorted[i];
        uint32 commandKey = commands[commandIndex].key;

        uint32 bucket = (commandKey >> shift) & mask;
        histogram[bucket]++;
    }

    uint32 bucketOffsets[numBuckets];
    uint32 sum = 0;
    for (int bucket = 0; bucket < numBuckets; bucket++) {
        bucketOffsets[bucket] = sum;
        sum += histogram[bucket];
    }

    for (int i = 0; i < commandList->commandCount; i++) {
        uint32 commandIndex = sorted[i];
        uint32 commandKey = commands[commandIndex].key;

        uint32 bucket = (commandKey >> shift) & mask;
        uint32 offset = bucketOffsets[bucket];

        temp[offset] = commandIndex;
        bucketOffsets[bucket]++;
    }

    commandList->sortedCommandIndices = temp;
    commandList->tempSortedCommandIndices = sorted;
}

void SortCommandList(CommandList *commandList) {
    // At this point, sortedCommandIndices will contain
    // 0..(commandCount) in order. _AllocateRenderCommand() does the
    // insertion
    
    for (int pass = 0; pass < 4; pass++) {
        CommandListRadixSort8Pass(commandList, pass);
    }
}

void ExecuteCommandList(CommandList *commandList) {
    uint32 prevKey = 0;

    for (int command = 0; command < commandList->commandCount; command++) {
        uint32 commandIndex = commandList->sortedCommandIndices[command];
        EnqueuedRenderCommand enqueuedCommand = commandList->commands[commandIndex];
        enqueuedCommand.backendFn(enqueuedCommand.data);

        ASSERT(prevKey <= enqueuedCommand.key);
        prevKey = enqueuedCommand.key;
    }
}


// Framebuffer blitting
void RenderPassthrough(uint32 textureID) {
    glBindFramebuffer(GL_FRAMEBUFFER, 0);
    uint32 colorBuffer = GL_COLOR_ATTACHMENT0;
    glDrawBuffers(1, &colorBuffer);


    ShaderProgram *program = GetShaderProgram(Game.assets, DebugMem->passthroughShader);
    Mesh *quad = GetMesh(Game.assets, DebugMem->quad);

    if (program != NULL && quad != NULL) {
        glViewport(0, 0, Game.graphics->width, Game.graphics->height);
        glDisable(GL_BLEND);
        glDisable(GL_DEPTH_TEST);

        glUseProgram(program->programID);

        glActiveTexture(GL_TEXTURE0);
        glBindTexture(GL_TEXTURE_2D, textureID);
        glUniform1i(UniformLocation(program, "in_texture"), 0);
        glBindSampler(0, Game.graphics->linearClampEdge.samplerID);

        glBindBuffer(GL_ARRAY_BUFFER, quad->vboID);
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, quad->iboID);

        glEnableVertexAttribArray(Position_Attrib);
        glVertexAttribPointer(Position_Attrib, 3, GL_FLOAT, GL_FALSE, 0,
                              (uint8 *)NULL + POINTER_OFFSET(quad->vertices, quad->positions));

        glEnableVertexAttribArray(TexCoord_Attrib);
        glVertexAttribPointer(TexCoord_Attrib, 2, GL_FLOAT, GL_FALSE, 0,
                              (uint8 *)NULL + POINTER_OFFSET(quad->vertices, quad->texcoords));

        glDrawElements(GL_TRIANGLES, quad->submeshLengths[0], GL_UNSIGNED_INT, 0);

        glDisableVertexAttribArray(Position_Attrib);
        glDisableVertexAttribArray(TexCoord_Attrib);

        glBindBuffer(GL_ARRAY_BUFFER, 0);
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
    }
    else {
        glClearColor(1, 0, 1, 1);
        glClear(GL_COLOR_BUFFER_BIT);
    }
}

void RenderPassthrough(CommandList *commandList, uint32 key, uint32 textureID, bool enableBlending) {
    ShaderProgram *program = GetShaderProgram(Game.assets, DebugMem->passthroughShader);
    Mesh *quad = GetMesh(Game.assets, DebugMem->quad);
    
    if (program != NULL && quad != NULL) {
        DisableDepthWrite(commandList, key);
        DisableDepthTest(commandList, key);

        if (enableBlending) {
            EnableBlending(commandList, key, GL_FUNC_ADD, GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
        }
        else {
            DisableBlending(commandList, key);
        }

        SetShader(commandList, key, program->programID);

        SetUniform(commandList, key, UniformLocation(program, "in_texture"), 0, textureID, GL_TEXTURE_2D);
        SetSampler(commandList, key, 0, Game.graphics->linearClampEdge);

        BindBuffer(commandList, key, GL_ARRAY_BUFFER, quad->vboID);
        BindBuffer(commandList, key, GL_ELEMENT_ARRAY_BUFFER, quad->iboID);

        EnableVertexAttribute(commandList, key, Position_Attrib, 3, GL_FLOAT, GL_FALSE, 0,
                              POINTER_OFFSET(quad->vertices, quad->positions));

        EnableVertexAttribute(commandList, key, TexCoord_Attrib, 2, GL_FLOAT, GL_FALSE, 0,
                              POINTER_OFFSET(quad->vertices, quad->texcoords));

        DrawIndexed(commandList, key, quad->primitiveType, 0, quad->indexCount);

        DisableVertexAttribute(commandList, key, Position_Attrib);
        DisableVertexAttribute(commandList, key, TexCoord_Attrib);
    }
}

void RenderPassthroughDepth(CommandList *commandList, real32 nearClip, uint32 key, uint32 textureID) {
    ShaderProgram *program = GetShaderProgram(Game.assets, DebugMem->passthroughDepthShader);
    Mesh *quad = GetMesh(Game.assets, DebugMem->quad);
    
    if (program != NULL && quad != NULL) {
        DisableDepthWrite(commandList, key);
        DisableDepthTest(commandList, key);
        DisableBlending(commandList, key);

        SetShader(commandList, key, program->programID);

        SetUniform(commandList, key, UniformLocation(program, "in_depth"), 0, textureID, GL_TEXTURE_2D);
        SetSampler(commandList, key, 0, Game.graphics->nearestClampEdge);

        SetUniform(commandList, key, UniformLocation(program, "nearClip"), nearClip);

        BindBuffer(commandList, key, GL_ARRAY_BUFFER, quad->vboID);
        BindBuffer(commandList, key, GL_ELEMENT_ARRAY_BUFFER, quad->iboID);

        EnableVertexAttribute(commandList, key, Position_Attrib, 3, GL_FLOAT, GL_FALSE, 0,
                              POINTER_OFFSET(quad->vertices, quad->positions));

        EnableVertexAttribute(commandList, key, TexCoord_Attrib, 2, GL_FLOAT, GL_FALSE, 0,
                              POINTER_OFFSET(quad->vertices, quad->texcoords));

        DrawIndexed(commandList, key, quad->primitiveType, 0, quad->indexCount);

        DisableVertexAttribute(commandList, key, Position_Attrib);
        DisableVertexAttribute(commandList, key, TexCoord_Attrib);
    }
}

void RenderPassthroughNormal(CommandList *commandList, uint32 key, uint32 textureID) {
    ShaderProgram *program = GetShaderProgram(Game.assets, DebugMem->passthroughNormalShader);
    Mesh *quad = GetMesh(Game.assets, DebugMem->quad);
    
    if (program != NULL && quad != NULL) {
        DisableDepthWrite(commandList, key);
        DisableDepthTest(commandList, key);
        DisableBlending(commandList, key);

        SetShader(commandList, key, program->programID);

        SetUniform(commandList, key, UniformLocation(program, "in_normal"), 0, textureID, GL_TEXTURE_2D);
        SetSampler(commandList, key, 0, Game.graphics->nearestClampEdge);

        BindBuffer(commandList, key, GL_ARRAY_BUFFER, quad->vboID);
        BindBuffer(commandList, key, GL_ELEMENT_ARRAY_BUFFER, quad->iboID);

        EnableVertexAttribute(commandList, key, Position_Attrib, 3, GL_FLOAT, GL_FALSE, 0,
                              POINTER_OFFSET(quad->vertices, quad->positions));

        EnableVertexAttribute(commandList, key, TexCoord_Attrib, 2, GL_FLOAT, GL_FALSE, 0,
                              POINTER_OFFSET(quad->vertices, quad->texcoords));

        DrawIndexed(commandList, key, quad->primitiveType, 0, quad->indexCount);

        DisableVertexAttribute(commandList, key, Position_Attrib);
        DisableVertexAttribute(commandList, key, TexCoord_Attrib);
    }
}

void RenderPassthroughBlended(CommandList *commandList, uint32 key, uint32 *textureIDs, uint32 textureCount, uint32 blendMapTextureID) {
    ShaderProgram *program = GetShaderProgram(Game.assets, DebugMem->passthroughBlendedShader);
    Mesh *quad = GetMesh(Game.assets, DebugMem->quad);
    
    if (program != NULL && quad != NULL) {
        DisableDepthWrite(commandList, key);
        DisableDepthTest(commandList, key);
        DisableBlending(commandList, key);

        SetShader(commandList, key, program->programID);


        const char *textureUniforms[] = {
            "in_texture0",
            "in_texture1",
            "in_texture2",
            "in_texture3",
        };
        for (int i = 0; i < 4; i++) {
            if (i < textureCount) {
                SetUniform(commandList, key, UniformLocation(program, textureUniforms[i]), i, textureIDs[i], GL_TEXTURE_2D);
                SetSampler(commandList, key, i, Game.graphics->linearClampEdge);
            }
            else {
                // @TODO: default black texture
            }
        }

        SetUniform(commandList, key, UniformLocation(program, "in_blendMap"), 4, blendMapTextureID, GL_TEXTURE_2D);
        SetSampler(commandList, key, 4, Game.graphics->linearClampEdge);

        // @TEMP
        SetUniform(commandList, key, UniformLocation(program, "GameTime"), Game.time);

        BindBuffer(commandList, key, GL_ARRAY_BUFFER, quad->vboID);
        BindBuffer(commandList, key, GL_ELEMENT_ARRAY_BUFFER, quad->iboID);

        EnableVertexAttribute(commandList, key, Position_Attrib, 3, GL_FLOAT, GL_FALSE, 0,
                              POINTER_OFFSET(quad->vertices, quad->positions));

        EnableVertexAttribute(commandList, key, TexCoord_Attrib, 2, GL_FLOAT, GL_FALSE, 0,
                              POINTER_OFFSET(quad->vertices, quad->texcoords));

        DrawIndexed(commandList, key, quad->primitiveType, 0, quad->indexCount);

        DisableVertexAttribute(commandList, key, Position_Attrib);
        DisableVertexAttribute(commandList, key, TexCoord_Attrib);
    }
}

void RenderPassthroughNormalBlended(CommandList *commandList, uint32 key, uint32 *textureIDs, uint32 textureCount, uint32 blendMapTextureID) {
    // @TODO: ...
}

void RenderSkybox(CommandList *commandList, uint32 key, uint32 textureID) {
    ShaderProgram *program = GetShaderProgram(Game.assets, Game.graphics->skyboxShader);
    Mesh *cube = GetMesh(Game.assets, DebugMem->skyboxCube);

    if (program != NULL && cube != NULL) {

        SetShader(commandList, key, program->programID);

        EnableDepthTest(commandList, key, DepthTest_Equal);

        SetUniform(commandList, key, UniformLocation(program, "skybox"), 0, textureID, GL_TEXTURE_CUBE_MAP);
        SetSampler(commandList, key, 0, Game.graphics->linearClampEdge);

            
        BindBuffer(commandList, key, GL_ARRAY_BUFFER, cube->vboID);
        BindBuffer(commandList, key, GL_ELEMENT_ARRAY_BUFFER, cube->iboID);

        EnableVertexAttribute(commandList, key, Position_Attrib, 3, GL_FLOAT, GL_FALSE, 0,
                              (uint8 *)cube->positions - (uint8 *)cube->vertices);

        DrawIndexed(commandList, key, PrimitiveType_Triangles, 0, cube->indexCount);

        DisableVertexAttribute(commandList, key, Position_Attrib);
    }
}


// Text Rendering
TextRenderBuffer *AllocateTextRenderBuffer() {
    TextRenderBuffer *buffer = (TextRenderBuffer *)Plat.AllocateMem(sizeof(TextRenderBuffer));

    buffer->nextBuffer = NULL;

    buffer->glyphCapacity = 1024;
    buffer->glyphCount = 0;

    buffer->glyphs = ALLOCATE_ARRAY(GlyphAttributes, buffer->glyphCapacity);

    glGenBuffers(1, &buffer->bufferID);
    glBindBuffer(GL_ARRAY_BUFFER, buffer->bufferID);

    buffer->bufferSize = sizeof(GlyphAttributes) * buffer->glyphCapacity;
    glBufferData(GL_ARRAY_BUFFER, buffer->bufferSize, NULL, GL_STREAM_DRAW);

    glBindBuffer(GL_ARRAY_BUFFER, 0);

    return buffer;
};

void DeallocateTextRenderBuffers(TextRenderBuffer *buffer) {

    while (buffer != NULL) {
        TextRenderBuffer *nextBuffer = buffer->nextBuffer;

        Plat.DeallocateMem(buffer->glyphs);
        glDeleteBuffers(1, &buffer->bufferID);

        Plat.DeallocateMem(buffer);

        buffer = nextBuffer;
    }
}

void RenderTextBuffers(TextRenderBuffer *buffer) {
    TIMED_BLOCK();    
    
    CommandList *commandList = &Game.graphics->textRenderPass;
    uint32 key = TEXT_RENDER_KEY;

    ShaderProgram *shader = GetShaderProgram(Game.assets, Game.graphics->textShader);
    if (shader == NULL) {
        return;
    }

    SetShader(commandList, key, shader->programID);

    DisableDepthWrite(commandList, key);
    DisableDepthTest(commandList, key);
    
    // EnableDepthWrite(commandList, key);
    // EnableDepthTest(commandList, key, DepthTest_LessEqual);

    EnableBlending(commandList, key, GL_FUNC_ADD, GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    // Bind GlyphQuad
    BindBuffer(commandList, key, GL_ARRAY_BUFFER, Game.graphics->glyphQuad.vboID);
    BindBuffer(commandList, key, GL_ELEMENT_ARRAY_BUFFER, Game.graphics->glyphQuad.iboID);

    EnableVertexAttribute(commandList, key, Position_Attrib, 3, GL_FLOAT, GL_FALSE, 0,
                          POINTER_OFFSET(Game.graphics->glyphQuad.vertices, Game.graphics->glyphQuad.positions));

    EnableVertexAttribute(commandList, key, TexCoord_Attrib, 2, GL_FLOAT, GL_FALSE, 0,
                          POINTER_OFFSET(Game.graphics->glyphQuad.vertices, Game.graphics->glyphQuad.texcoords));

    while (buffer != NULL) {
        FontTable *font = GetFontTable(Game.assets, buffer->font);
        if (font == NULL) {
            return;
        }

        BindBuffer(commandList, key, GL_ARRAY_BUFFER, buffer->bufferID);

        CopyBufferData(commandList, key, GL_ARRAY_BUFFER, buffer->bufferSize, buffer->glyphs, GL_STREAM_DRAW);

                
        EnableVertexIAttribute(commandList, key, GlyphCodepoint_Attrib, 1, GL_UNSIGNED_INT,
                               sizeof(GlyphAttributes), FIELD_OFFSET(GlyphAttributes, codepoint), 1);

        EnableVertexAttribute(commandList, key, GlyphPosition_Attrib, 3, GL_FLOAT, GL_FALSE,
                              sizeof(GlyphAttributes), FIELD_OFFSET(GlyphAttributes, position), 1);

        EnableVertexAttribute(commandList, key, Rotation_Attrib, 4, GL_FLOAT, GL_FALSE,
                              sizeof(GlyphAttributes), FIELD_OFFSET(GlyphAttributes, rotation), 1);

        EnableVertexAttribute(commandList, key, GlyphDimensions_Attrib, 2, GL_FLOAT, GL_FALSE,
                              sizeof(GlyphAttributes), FIELD_OFFSET(GlyphAttributes, dimensions), 1);

        EnableVertexAttribute(commandList, key, Color_Attrib, 4, GL_FLOAT, GL_FALSE,
                              sizeof(GlyphAttributes), FIELD_OFFSET(GlyphAttributes, color), 1);
            
        EnableVertexAttribute(commandList, key, Scale_Attrib, 2, GL_FLOAT, GL_FALSE,
                              sizeof(GlyphAttributes), FIELD_OFFSET(GlyphAttributes, scale), 1);

        EnableVertexAttribute(commandList, key, GlyphAnchor_Attrib, 2, GL_FLOAT, GL_FALSE,
                              sizeof(GlyphAttributes), FIELD_OFFSET(GlyphAttributes, anchor), 1);

        EnableVertexAttribute(commandList, key, GlyphAnchor_Attrib, 2, GL_FLOAT, GL_FALSE,
                              sizeof(GlyphAttributes), FIELD_OFFSET(GlyphAttributes, anchor), 1);

        EnableVertexAttribute(commandList, key, GlyphThickness_Attrib, 1, GL_FLOAT, GL_FALSE,
                              sizeof(GlyphAttributes), FIELD_OFFSET(GlyphAttributes, thickness), 1);

        EnableVertexAttribute(commandList, key, GlyphOutline_Attrib, 1, GL_FLOAT, GL_FALSE,
                              sizeof(GlyphAttributes), FIELD_OFFSET(GlyphAttributes, outline), 1);

        EnableVertexAttribute(commandList, key, GlyphOutlineColor_Attrib, 4, GL_FLOAT, GL_FALSE,
                              sizeof(GlyphAttributes), FIELD_OFFSET(GlyphAttributes, outlineColor), 1);
        
        EnableVertexAttribute(commandList, key, GlyphShadowThickness_Attrib, 1, GL_FLOAT, GL_FALSE,
                              sizeof(GlyphAttributes), FIELD_OFFSET(GlyphAttributes, shadowThickness), 1);
        EnableVertexAttribute(commandList, key, GlyphShadowSoftness_Attrib, 1, GL_FLOAT, GL_FALSE,
                              sizeof(GlyphAttributes), FIELD_OFFSET(GlyphAttributes, shadowSoftness), 1);
        EnableVertexAttribute(commandList, key, GlyphShadowColor_Attrib, 4, GL_FLOAT, GL_FALSE,
                              sizeof(GlyphAttributes), FIELD_OFFSET(GlyphAttributes, shadowColor), 1);
        EnableVertexAttribute(commandList, key, GlyphShadowOffset_Attrib, 2, GL_FLOAT, GL_FALSE,
                              sizeof(GlyphAttributes), FIELD_OFFSET(GlyphAttributes, shadowOffset), 1);

        SetUniform(commandList, key, UniformLocation(shader, "model"), buffer->model);

        SetUniform(commandList, key, UniformLocation(shader, "texcoordsMap"), 0, font->texcoordsMapID, GL_TEXTURE_1D);
        SetSampler(commandList, key, 0, Game.graphics->nearestClampEdge);

        SetUniform(commandList, key, UniformLocation(shader, "atlas"), 1, font->atlas.textureID, GL_TEXTURE_2D);
        SetUniform(commandList, key, UniformLocation(shader, "msdfRange"), font->msdfRange);
        SetSampler(commandList, key, 1, Game.graphics->linearClampEdge);

        DrawIndexedInstanced(commandList, key, PrimitiveType_Triangles, 0, 6, buffer->glyphCount);

        DisableVertexAttribute(commandList, key, GlyphCodepoint_Attrib);
        DisableVertexAttribute(commandList, key, GlyphPosition_Attrib);
        DisableVertexAttribute(commandList, key, Rotation_Attrib);
        DisableVertexAttribute(commandList, key, GlyphDimensions_Attrib);
        DisableVertexAttribute(commandList, key, Color_Attrib);
        DisableVertexAttribute(commandList, key, Scale_Attrib);
        DisableVertexAttribute(commandList, key, GlyphAnchor_Attrib);

        buffer->glyphCount = 0;
        buffer = buffer->nextBuffer;
    }

    DisableVertexAttribute(commandList, key, Position_Attrib);
    DisableVertexAttribute(commandList, key, TexCoord_Attrib);
}

TextRenderBuffer *FindFirstTextRenderBuffer(TextRenderBuffer **buffers, FontTableHandle fontHandle) {

    TextRenderBuffer *buffer = *buffers;
    if (buffer == NULL) {
        buffer = AllocateTextRenderBuffer();
        *buffers = buffer;
    }

    while (true) {
        if (buffer->glyphCount > 0) {
            if (buffer->nextBuffer == NULL) {
                TextRenderBuffer *nextBuffer = AllocateTextRenderBuffer();
                buffer->nextBuffer = nextBuffer;
            }

            buffer = buffer->nextBuffer;
            continue;
        }

        break;
    }
    
    return buffer;
}

void DrawGlyphs(TextRenderBuffer **buffers, FontTableHandle fontHandle, uint32 glyphCount, GlyphAttributes *attribs, mat4 model, bool *renderMask) {
    FontTable *font = GetFontTable(Game.assets, fontHandle);
    if (font == NULL) {
        return;
    }

    // @BUG: if asked to draw more glyphs than will fit in buffer, things go boom!
    TextRenderBuffer *buffer = FindFirstTextRenderBuffer(buffers, fontHandle);
    if (buffer == NULL) {
        return;
    }

    buffer->font = fontHandle;
    buffer->model = model;
    
    for (int i = 0; i < glyphCount; i++) {
        if (renderMask != NULL) {
            if (renderMask[i]) {
                buffer->glyphs[buffer->glyphCount] = attribs[i];
                buffer->glyphCount += 1;
            }
        }
        else {
            buffer->glyphs[buffer->glyphCount] = attribs[i];
            buffer->glyphCount += 1;    
        }
    }
}


void DrawGlyphs(FontTableHandle fontHandle, uint32 glyphCount, GlyphAttributes *attribs, mat4 model, bool *renderMask) {
    DrawGlyphs(&Game.graphics->textRenderBuffers, fontHandle, glyphCount, attribs, model, renderMask);
}
