
#ifndef _ETC_RENDER_H
#define _ETC_RENDER_H

#define MAX_COLOR_ATTACHMENTS 8

ENUM()
enum PrimitiveType : uint32 {
    PrimitiveType_Points,
    PrimitiveType_Lines,
    PrimitiveType_LineLoop,
    PrimitiveType_LineStrip,
    PrimitiveType_Triangles,
    PrimitiveType_TriangleStrip,
    PrimitiveType_TriangleFan,
};

enum ColorFormat {
    ColorFormat_RGB8,
    ColorFormat_RGBA8,
    ColorFormat_R32F,
    ColorFormat_RGB32F,
    ColorFormat_RGBA32F,
};

enum TextureFilterMode {
    TextureFilter_Nearest,
    TextureFilter_Linear,
};

enum TextureWrapMode {
    TextureWrap_ClampToEdge,
    TextureWrap_ClampToBorderColor,
    TextureWrap_Repeat,
};

enum CubemapFace {
    Cubemap_PosX,
    Cubemap_NegX,
    Cubemap_PosY,
    Cubemap_NegY,
    Cubemap_PosZ,
    Cubemap_NegZ,
};

enum UniformType {
    UniformType_Integer,
    UniformType_Float,
    UniformType_Vec2,
    UniformType_Vec3,
    UniformType_Vec4,
    UniformType_Mat4,
    UniformType_Texture,
};

enum DepthTestType {
    DepthTest_Always,
    DepthTest_Never,
    DepthTest_Less,
    DepthTest_Greater,
    DepthTest_Equal,
    DepthTest_NotEqual,
    DepthTest_LessEqual,
    DepthTest_GreaterEqual,
};


#define FOREACH_VERTEXATTRIB(OP)                                \
    OP(Position, "in_position", 0)                              \
    OP(TexCoord, "in_texcoord", 1)                              \
    OP(Normal, "in_normal", 3)                                  \
    OP(Color, "in_color", 4)                                    \
    OP(Rotation, "in_rotation", 5)                              \
    OP(Scale, "in_scale", 6)                                    \
    OP(BoneIndices, "in_boneIndices", 8)                        \
    OP(BoneWeights, "in_boneWeights", 9)                        \
    OP(InstanceMat4_Col0, "in_instance_mat4", 8)                \
    OP(InstanceMat4_Col1, NULL, 9)                              \
    OP(InstanceMat4_Col2, NULL, 10)                             \
    OP(InstanceMat4_Col3, NULL, 11)                             \
    OP(GlyphCodepoint, "in_codepoint", 8)                       \
    OP(GlyphFontIndex, "in_fontIndex", 9)                       \
    OP(GlyphPosition, "in_glyphPosition", 10)                   \
    OP(GlyphDimensions, "in_glyphDimensions", 12)               \
    OP(GlyphAnchor, "in_anchor", 13)                            \
    OP(GlyphThickness, "in_thickness", 14)                      \
    OP(GlyphShadowThickness, "in_shadowThickness", 15)          \
    OP(GlyphShadowSoftness, "in_shadowSoftness", 16)            \
    OP(GlyphShadowOffset, "in_shadowOffset", 17)                \
    OP(GlyphShadowColor, "in_shadowColor", 18)                  \
    OP(GlyphOutline, "in_outline", 19)                          \
    OP(GlyphOutlineColor, "in_outlineColor", 20)                \
    OP(StrokeVertexPositionX, "in_strokePositionX", 8)          \
    OP(StrokeVertexPositionY, "in_strokePositionY", 9)          \
    OP(StrokeVertexPositionZ, "in_strokePositionZ", 10)         \
    OP(StrokeVertexNormalX, "in_strokeNormalX", 11)             \
    OP(StrokeVertexNormalY, "in_strokeNormalY", 12)             \
    OP(StrokeVertexNormalZ, "in_strokeNormalZ", 13)             \
    OP(StrokeVertexBrushStrokeIndex, "in_brushStrokeIndex", 14) \
    OP(SplatPosition, "in_splatPosition", 2)                    \
    OP(SplatColor, "in_splatColor", 3)                          \
    OP(SplatSize, "in_splatSize", 4)                            \
    OP(SplatAngle, "in_splatAngle", 5)                          \
    OP(SplatStrokeIndex, "in_splatStrokeIndex", 6)              \
    OP(SplatSmearDirection, "in_splatSmearDirection", 7)        \
    OP(SplatSmearAmount, "in_splatSmearAmount", 8)              \
    OP(SplatWeight, "in_splatWeight", 9)                        \
    OP(DashParam, "in_dashParam", 5)                            \
    OP(DepthFailScale, "in_depthFailScale", 6)                  \


#define DEFINE_VERTEXATTRIB_ENUM(attribName, attribInputString, value) attribName##_Attrib = value,

enum VertexAttrib {
    FOREACH_VERTEXATTRIB(DEFINE_VERTEXATTRIB_ENUM)
};


struct VertexAttribEntry {
    const char *name;
    VertexAttrib value;
};

#define DEFINE_VERTEXATTRIB_ENTRY(attribName, attribInputString, value) {#attribName, attribName##_Attrib},

static VertexAttribEntry VertexAttribTable[] = {
    FOREACH_VERTEXATTRIB(DEFINE_VERTEXATTRIB_ENTRY)
};

inline b32 FindVertexAttrib(const char *name, uint32 length, VertexAttrib *result) {
    b32 success = false;

    uint32 index;
    if (FindString(&VertexAttribTable[0].name, ARRAY_LENGTH(VertexAttribEntry, VertexAttribTable), name, length, &index, sizeof(VertexAttribEntry))) {
        if (result != NULL) {
            *result = VertexAttribTable[index].value;
        }
        success = true;
    }

    return success;
}


struct RenderTarget {
    uint32 fboID;
    
    ColorFormat colorFormat;
    uint32 depthBits;

    uint16 width;
    uint16 height;

    uint32 colorTargetCount;
    uint32 colorTargets[MAX_COLOR_ATTACHMENTS];
    uint32 depthTarget;
    uint32 stencilTarget;
};

struct Sampler {
    uint32 samplerID;
};


// Low Level Render Commands
typedef void RenderBackendFunction(void *data);

struct ClearColorCommand {
    static RenderBackendFunction *backendFn;

    uint32 buffer;
    vec4 color;
};

struct ClearDepthCommand {
    static RenderBackendFunction *backendFn;

    real32 depth;
};

struct ClearStencilCommand {
    static RenderBackendFunction *backendFn;

    int32 value;
};

struct BindBufferCommand {
    static RenderBackendFunction *backendFn;

    uint32 target;
    uint32 bufferID;
};

struct BindBufferRangeCommand {
    static RenderBackendFunction *backendFn;

    uint32 target;
    uint32 index;
    uint32 bufferID;

    uint32 offset;
    uint32 size;
};

struct CopyBufferDataCommand {
    static RenderBackendFunction *backendFn;

    uint32 target;
    uint32 size;
    void *data;
    uint32 usage;
};

struct CopyBufferSubDataCommand {
    static RenderBackendFunction *backendFn;

    uint32 target;
    uint32 offset;
    uint32 size;
    void *data;
};

struct EnableVertexAttributeCommand {
    static RenderBackendFunction *backendFn;

    uint32 index;
    uint32 size;
    uint32 type;
    b32 normalized;
    uint32 stride;
    uint32 offset;

    uint32 divisor;
};

struct EnableVertexIAttributeCommand {
    static RenderBackendFunction *backendFn;

    uint32 index;
    uint32 size;
    uint32 type;
    uint32 stride;
    uint32 offset;

    uint32 divisor;
};

struct DisableVertexAttributeCommand {
    static RenderBackendFunction *backendFn;

    uint32 index;
};

struct SetDrawRenderTargetCommand {
    static RenderBackendFunction *backendFn;

    uint32 fboID;
};

struct SetReadRenderTargetCommand {
    static RenderBackendFunction *backendFn;

    uint32 fboID;
};

struct SetViewportCommand {
    static RenderBackendFunction *backendFn;

    EtcRecti viewport;
};

struct SetColorBuffersCommand {
    static RenderBackendFunction *backendFn;

    uint32 count;
    uint32 buffers[4];
};

struct EnableScissorTestCommand {
    static RenderBackendFunction *backendFn;

    EtcRecti region;
};

struct DisableScissorTestCommand {
    static RenderBackendFunction *backendFn;
};

struct SetColorMaskCommand {
    static RenderBackendFunction *backendFn;

    uint32 buffer;
    b32 red;
    b32 green;
    b32 blue;
    b32 alpha;
};

struct EnableDepthTestCommand {
    static RenderBackendFunction *backendFn;

    DepthTestType type;
};

struct DisableDepthTestCommand {
    static RenderBackendFunction *backendFn;
};

struct EnableDepthWriteCommand {
    static RenderBackendFunction *backendFn;
};

struct DisableDepthWriteCommand {
    static RenderBackendFunction *backendFn;
};

struct SetStencilMaskCommand {
    static RenderBackendFunction *backendFn;

    uint32 mask;
};

struct EnableStencilTestCommand {
    static RenderBackendFunction *backendFn;
};

struct DisableStencilTestCommand {
    static RenderBackendFunction *backendFn;
};

struct SetStencilOpCommand {
    static RenderBackendFunction *backendFn;

    uint32 stencilFail;
    uint32 depthFail;
    uint32 depthPass;
};

struct SetStencilFuncCommand {
    static RenderBackendFunction *backendFn;

    uint32 func;
    int32 reference;
    uint32 mask;
};

struct EnableBlendingCommand {
    static RenderBackendFunction *backendFn;

    uint32 rgbEquation;
    uint32 alphaEquation;

    uint32 srcFactor;
    uint32 dstFactor;
    uint32 srcAlphaFactor;
    uint32 dstAlphaFactor;
};

struct DisableBlendingCommand {
    static RenderBackendFunction *backendFn;
};

struct EnableWireframeCommand {
    static RenderBackendFunction *backendFn;
};

struct DisableWireframeCommand {
    static RenderBackendFunction *backendFn;
};

struct EnableFaceCullingCommand {
    static RenderBackendFunction *backendFn;

    uint32 faceMode;
};

struct DisableFaceCullingCommand {
    static RenderBackendFunction *backendFn;
};

struct SetShaderCommand {
    static RenderBackendFunction *backendFn;

    uint32 programID;
};

struct SetSamplerCommand {
    static RenderBackendFunction *backendFn;

    uint32 textureUnit;
    uint32 samplerID;
};

struct SetUniformCommand {
    static RenderBackendFunction *backendFn;

    UniformType type;
    uint32 location;

    union {
        real32 real32Value;
        int32 int32Value;

        vec2 vec2Value;
        vec3 vec3Value;
        vec4 vec4Value;
        mat4 mat4Value;

        void *ptrValue;

        struct {
            uint32 textureUnit;
            uint32 textureID;
            uint32 textureTarget;
        };
    };
};

struct DrawArraysCommand {
    static RenderBackendFunction *backendFn;

    PrimitiveType primitiveType;
    uint32 first;
    uint32 count;
};

struct DrawArraysInstancedCommand {
    static RenderBackendFunction *backendFn;

    PrimitiveType primitiveType;
    uint32 first;
    uint32 count;
    uint32 instanceCount;
    uint32 baseInstance;
};

struct DrawIndexedCommand {
    static RenderBackendFunction *backendFn;

    PrimitiveType primitiveType;
    uint32 first;
    uint32 count;
};

struct DrawIndexedInstancedCommand {
    static RenderBackendFunction *backendFn;

    PrimitiveType primitiveType;
    uint32 first;
    uint32 count;
    uint32 instanceCount;
    uint32 baseInstance;
};

struct ReadPixelsCommand {
    static RenderBackendFunction *backendFn;

    int32 x;
    int32 y;
    uint32 width;
    uint32 height;
    uint32 format;
    uint32 type;
    void *data;
};

struct EnqueuedRenderCommand {
    uint32 key;
    RenderBackendFunction *backendFn;
    void *data;
};

struct CommandList {
    MemoryArena arena;

    uint32 commandCapacity;
    uint32 commandCount;
    EnqueuedRenderCommand *commands;

    // Sorting method puts `commands` indices here
    uint32 *sortedCommandIndices;
    uint32 *tempSortedCommandIndices;
};

void *AllocateRenderCommand_(CommandList *commandList, uint32 key, uint32 commandSize, RenderBackendFunction *backendFn);
#define AllocateRenderCommand(commandList, key, type) (type *)AllocateRenderCommand_(commandList, key, sizeof(type), type::backendFn)

void AllocateCommandList(CommandList *commandList, uint32 capacity, uint32 arenaSize);
void ClearCommandList(CommandList *commandList);
void SortCommandList(CommandList *commandList);
void ExecuteCommandList(CommandList *commandList);


// Render Command Convenience Functions
inline void ClearColor(CommandList *commandList, uint32 key, vec4 color, uint32 buffer = 0) {
    ClearColorCommand *cmd = AllocateRenderCommand(commandList, key, ClearColorCommand);
    cmd->buffer = buffer;
    cmd->color = color;
}

inline void ClearDepth(CommandList *commandList, uint32 key, real32 depth = 1.0f) {
    ClearDepthCommand *cmd = AllocateRenderCommand(commandList, key, ClearDepthCommand);
    cmd->depth = depth;
}

inline void ClearStencil(CommandList *commandList, uint32 key, int32 value) {
    ClearStencilCommand *cmd = AllocateRenderCommand(commandList, key, ClearStencilCommand);
    cmd->value = value;
}

inline void BindBuffer(CommandList *commandList, uint32 key, uint32 target, uint32 bufferID) {
    BindBufferCommand *cmd = AllocateRenderCommand(commandList, key, BindBufferCommand);
    cmd->target = target;
    cmd->bufferID = bufferID;
}

inline void BindBufferRange(CommandList *commandList, uint32 key, uint32 target, uint32 index, uint32 bufferID, uint32 offset, uint32 size) {
    BindBufferRangeCommand *cmd = AllocateRenderCommand(commandList, key, BindBufferRangeCommand);
    cmd->target = target;
    cmd->index = index;
    cmd->bufferID = bufferID;

    cmd->offset = offset;
    cmd->size = size;
}

inline void CopyBufferData(CommandList *commandList, uint32 key, uint32 target, uint32 size, void *data, uint32 usage) {
    CopyBufferDataCommand *cmd = AllocateRenderCommand(commandList, key, CopyBufferDataCommand);
    cmd->target = target;
    cmd->size = size;
    cmd->data = data;
    cmd->usage = usage;
}
inline void CopyBufferSubData(CommandList *commandList, uint32 key, uint32 target, uint32 offset, uint32 size, void *data) {
    CopyBufferSubDataCommand *cmd = AllocateRenderCommand(commandList, key, CopyBufferSubDataCommand);
    cmd->target = target;
    cmd->offset = offset;
    cmd->size = size;
    cmd->data = data;
}

inline void EnableVertexAttribute(CommandList *commandList, uint32 key, uint32 index, uint32 size, uint32 type, b32 normalized, uint32 stride, uint32 offset, uint32 divisor = 0) {
    EnableVertexAttributeCommand *cmd = AllocateRenderCommand(commandList, key, EnableVertexAttributeCommand);
    cmd->index = index;
    cmd->size = size;
    cmd->type = type;
    cmd->normalized = normalized;
    cmd->stride = stride;
    cmd->offset = offset;
    cmd->divisor = divisor;
}

inline void EnableVertexIAttribute(CommandList *commandList, uint32 key, uint32 index, uint32 size, uint32 type, uint32 stride, uint32 offset, uint32 divisor = 0) {
    EnableVertexIAttributeCommand *cmd = AllocateRenderCommand(commandList, key, EnableVertexIAttributeCommand);
    cmd->index = index;
    cmd->size = size;
    cmd->type = type;
    cmd->stride = stride;
    cmd->offset = offset;
    cmd->divisor = divisor;
}

inline void DisableVertexAttribute(CommandList *commandList, uint32 key, uint32 index) {
    DisableVertexAttributeCommand *cmd = AllocateRenderCommand(commandList, key, DisableVertexAttributeCommand);
    cmd->index = index;
}

inline void SetDrawRenderTarget(CommandList *commandList, uint32 key, RenderTarget *target) {
    SetDrawRenderTargetCommand *cmd = AllocateRenderCommand(commandList, key, SetDrawRenderTargetCommand);
    if (target != NULL) {
        cmd->fboID = target->fboID;
    }
    else {
        cmd->fboID = 0;
    }
}

inline void SetReadRenderTarget(CommandList *commandList, uint32 key, RenderTarget *target) {
    SetReadRenderTargetCommand *cmd = AllocateRenderCommand(commandList, key, SetReadRenderTargetCommand);
    if (target != NULL) {
        cmd->fboID = target->fboID;
    }
    else {
        cmd->fboID = 0;
    }
}

inline void SetViewport(CommandList *commandList, uint32 key, EtcRecti viewport) {
    SetViewportCommand *cmd = AllocateRenderCommand(commandList, key, SetViewportCommand);
    cmd->viewport = viewport;
}

inline void SetViewport(CommandList *commandList, uint32 key, vec2i pos, vec2i size) {
    SetViewportCommand *cmd = AllocateRenderCommand(commandList, key, SetViewportCommand);
    cmd->viewport.pos = pos;
    cmd->viewport.size = size;
}

inline void SetColorBuffers(CommandList *commandList, uint32 key, uint32 count, uint32 *attachments) {
    SetColorBuffersCommand *cmd = AllocateRenderCommand(commandList, key, SetColorBuffersCommand);
    cmd->count = count;
    Plat.CopyMem(cmd->buffers, attachments, sizeof(uint32) * count);
}

inline void EnableScissorTest(CommandList *commandList, uint32 key, EtcRecti region) {
    EnableScissorTestCommand *cmd = AllocateRenderCommand(commandList, key, EnableScissorTestCommand);
    cmd->region = region;
}

inline void DisableScissorTest(CommandList *commandList, uint32 key) {
    DisableScissorTestCommand *cmd = AllocateRenderCommand(commandList, key, DisableScissorTestCommand);
}

inline void SetColorMask(CommandList *commandList, uint32 key, uint32 buffer, b32 allChannels) {
    SetColorMaskCommand *cmd = AllocateRenderCommand(commandList, key, SetColorMaskCommand);

    cmd->buffer = buffer;
    cmd->red = allChannels;
    cmd->green = allChannels;
    cmd->blue = allChannels;
    cmd->alpha = allChannels;
}

inline void EnableDepthTest(CommandList *commandList, uint32 key, DepthTestType type) {
    EnableDepthTestCommand *cmd = AllocateRenderCommand(commandList, key, EnableDepthTestCommand);
    cmd->type = type;
}

inline void DisableDepthTest(CommandList *commandList, uint32 key) {
    DisableDepthTestCommand *cmd = AllocateRenderCommand(commandList, key, DisableDepthTestCommand);
}

inline void EnableDepthWrite(CommandList *commandList, uint32 key) {
    EnableDepthWriteCommand *cmd = AllocateRenderCommand(commandList, key, EnableDepthWriteCommand);
}

inline void DisableDepthWrite(CommandList *commandList, uint32 key) {
    DisableDepthWriteCommand *cmd = AllocateRenderCommand(commandList, key, DisableDepthWriteCommand);
}

inline void SetStencilMask(CommandList *commandList, uint32 key, uint32 mask) {
    SetStencilMaskCommand *cmd = AllocateRenderCommand(commandList, key, SetStencilMaskCommand);
    cmd->mask = mask;
}

inline void EnableStencilTest(CommandList *commandList, uint32 key) {
    EnableStencilTestCommand *cmd = AllocateRenderCommand(commandList, key, EnableStencilTestCommand);
}

inline void DisableStencilTest(CommandList *commandList, uint32 key) {
    DisableStencilTestCommand *cmd = AllocateRenderCommand(commandList, key, DisableStencilTestCommand);
}

inline void SetStencilOp(CommandList *commandList, uint32 key, uint32 stencilFail, uint32 depthFail, uint32 depthPass) {
    SetStencilOpCommand *cmd = AllocateRenderCommand(commandList, key, SetStencilOpCommand);
    cmd->stencilFail = stencilFail;
    cmd->depthFail = depthFail;
    cmd->depthPass = depthPass;
}

inline void SetStencilFunc(CommandList *commandList, uint32 key, uint32 func, int32 reference, uint32 mask) {
    SetStencilFuncCommand *cmd = AllocateRenderCommand(commandList, key, SetStencilFuncCommand);
    cmd->func = func;
    cmd->reference = reference;
    cmd->mask = mask;
}

inline void EnableBlending(CommandList *commandList, uint32 key, uint32 equation, uint32 srcFactor, uint32 dstFactor) {
    EnableBlendingCommand *cmd = AllocateRenderCommand(commandList, key, EnableBlendingCommand);
    cmd->rgbEquation = equation;
    cmd->alphaEquation = equation;

    cmd->srcFactor = srcFactor;
    cmd->dstFactor = dstFactor;
    cmd->srcAlphaFactor = srcFactor;
    cmd->dstAlphaFactor = dstFactor;
}

inline void EnableBlending(CommandList *commandList, uint32 key, uint32 equation, uint32 srcFactor, uint32 dstFactor, uint32 srcAlphaFactor, uint32 dstAlphaFactor) {
    EnableBlendingCommand *cmd = AllocateRenderCommand(commandList, key, EnableBlendingCommand);
    cmd->rgbEquation = equation;
    cmd->alphaEquation = equation;

    cmd->srcFactor = srcFactor;
    cmd->dstFactor = dstFactor;
    cmd->srcAlphaFactor = srcAlphaFactor;
    cmd->dstAlphaFactor = dstAlphaFactor;
}

inline void EnableBlending(CommandList *commandList, uint32 key, uint32 rgbEquation, uint32 srcFactor, uint32 dstFactor, uint32 alphaEquation, uint32 srcAlphaFactor, uint32 dstAlphaFactor) {
    EnableBlendingCommand *cmd = AllocateRenderCommand(commandList, key, EnableBlendingCommand);
    cmd->rgbEquation = rgbEquation;
    cmd->alphaEquation = alphaEquation;

    cmd->srcFactor = srcFactor;
    cmd->dstFactor = dstFactor;
    cmd->srcAlphaFactor = srcAlphaFactor;
    cmd->dstAlphaFactor = dstAlphaFactor;
}

inline void DisableBlending(CommandList *commandList, uint32 key) {
    DisableBlendingCommand *cmd = AllocateRenderCommand(commandList, key, DisableBlendingCommand);
}

inline void EnableWireframe(CommandList *commandList, uint32 key) {
    EnableWireframeCommand *cmd = AllocateRenderCommand(commandList, key, EnableWireframeCommand);
}

inline void DisableWireframe(CommandList *commandList, uint32 key) {
    DisableWireframeCommand *cmd = AllocateRenderCommand(commandList, key, DisableWireframeCommand);
}

inline void EnableFaceCulling(CommandList *commandList, uint32 key, uint32 faceMode) {
    EnableFaceCullingCommand *cmd = AllocateRenderCommand(commandList, key, EnableFaceCullingCommand);
    cmd->faceMode = faceMode;
};

inline void DisableFaceCulling(CommandList *commandList, uint32 key) {
    DisableFaceCullingCommand *cmd = AllocateRenderCommand(commandList, key, DisableFaceCullingCommand);
};

inline void SetShader(CommandList *commandList, uint32 key, uint32 programID) {
    SetShaderCommand *cmd = AllocateRenderCommand(commandList, key, SetShaderCommand);
    cmd->programID = programID;
}

inline void SetSampler(CommandList *commandList, uint32 key, uint32 textureUnit, Sampler sampler) {
    SetSamplerCommand *cmd = AllocateRenderCommand(commandList, key, SetSamplerCommand);
    cmd->textureUnit = textureUnit;
    cmd->samplerID = sampler.samplerID;
}

inline void SetUniform(CommandList *commandList, uint32 key, uint32 location, int32 int32Value) {
    SetUniformCommand *cmd = AllocateRenderCommand(commandList, key, SetUniformCommand);
    cmd->type = UniformType_Integer;
    cmd->location = location;
    cmd->int32Value = int32Value;
}

inline void SetUniform(CommandList *commandList, uint32 key, uint32 location, real32 real32Value) {
    SetUniformCommand *cmd = AllocateRenderCommand(commandList, key, SetUniformCommand);
    cmd->type = UniformType_Float;
    cmd->location = location;
    cmd->real32Value = real32Value;
}

inline void SetUniform(CommandList *commandList, uint32 key, uint32 location, vec2 vec2Value) {
    SetUniformCommand *cmd = AllocateRenderCommand(commandList, key, SetUniformCommand);
    cmd->type = UniformType_Vec2;
    cmd->location = location;
    cmd->vec2Value = vec2Value;
}

inline void SetUniform(CommandList *commandList, uint32 key, uint32 location, vec3 vec3Value) {
    SetUniformCommand *cmd = AllocateRenderCommand(commandList, key, SetUniformCommand);
    cmd->type = UniformType_Vec3;
    cmd->location = location;
    cmd->vec3Value = vec3Value;
}

inline void SetUniform(CommandList *commandList, uint32 key, uint32 location, vec4 vec4Value) {
    SetUniformCommand *cmd = AllocateRenderCommand(commandList, key, SetUniformCommand);
    cmd->type = UniformType_Vec4;
    cmd->location = location;
    cmd->vec4Value = vec4Value;
}

inline void SetUniform(CommandList *commandList, uint32 key, uint32 location, mat4 mat4Value) {
    SetUniformCommand *cmd = AllocateRenderCommand(commandList, key, SetUniformCommand);
    cmd->type = UniformType_Mat4;
    cmd->location = location;
    cmd->mat4Value = mat4Value;
}

inline void SetUniform(CommandList *commandList, uint32 key, uint32 location, uint32 textureUnit, uint32 textureID, uint32 textureTarget) {
    SetUniformCommand *cmd = AllocateRenderCommand(commandList, key, SetUniformCommand);
    cmd->type = UniformType_Texture;
    cmd->location = location;
    cmd->textureUnit = textureUnit;
    cmd->textureID = textureID;
    cmd->textureTarget = textureTarget;
}

inline void DrawArrays(CommandList *commandList, uint32 key, PrimitiveType primitiveType, uint32 first, uint32 count) {
    DrawArraysCommand *cmd = AllocateRenderCommand(commandList, key, DrawArraysCommand);
    cmd->primitiveType = primitiveType;
    cmd->first = first;
    cmd->count = count;
}

inline void DrawArraysInstanced(CommandList *commandList, uint32 key, PrimitiveType primitiveType, uint32 first, uint32 count, uint32 instanceCount, uint32 baseInstance = 0) {
    DrawArraysInstancedCommand *cmd = AllocateRenderCommand(commandList, key, DrawArraysInstancedCommand);
    cmd->primitiveType = primitiveType;
    cmd->first = first;
    cmd->count = count;
    cmd->instanceCount = instanceCount;
    cmd->baseInstance = baseInstance;
}

inline void DrawIndexed(CommandList *commandList, uint32 key, PrimitiveType primitiveType, uint32 first, uint32 count) {
    DrawIndexedCommand *cmd = AllocateRenderCommand(commandList, key, DrawIndexedCommand);
    cmd->primitiveType = primitiveType;
    cmd->first = first;
    cmd->count = count;
}

inline void DrawIndexedInstanced(CommandList *commandList, uint32 key, PrimitiveType primitiveType, uint32 first, uint32 count, uint32 instanceCount, uint32 baseInstance = 0) {
    DrawIndexedInstancedCommand *cmd = AllocateRenderCommand(commandList, key, DrawIndexedInstancedCommand);
    cmd->primitiveType = primitiveType;
    cmd->first = first;
    cmd->count = count;
    cmd->instanceCount = instanceCount;
    cmd->baseInstance = baseInstance;
}

inline void ReadPixels(CommandList *commandList, uint32 key, int32 x, int32 y, uint32 width, uint32 height, uint32 format, uint32 type, void *data) {
    ReadPixelsCommand *cmd = AllocateRenderCommand(commandList, key, ReadPixelsCommand);
    cmd->x = x;
    cmd->y = y;
    cmd->width = width;
    cmd->height = height;
    cmd->format = format;
    cmd->type = type;
    cmd->data = data;
}


// High Level Rendering
ENUM()
enum RenderMode : uint32 {
    RenderMode_DeferredNoLighting,
    RenderMode_DeferredNormals,
    RenderMode_Deferred,

    RenderMode_Manual,
};

#define UNIFORM_BLOCK_BINDING_CAMERA 0
#define UNIFORM_BLOCK_BINDING_SCREEN 1
#define UNIFORM_BLOCK_BINDING_STROKEPARAMS 2
#define UNIFORM_BLOCK_BINDING_BONETRANSFORMS 3

struct CameraUniformBuffer {
    mat4 view;
    mat4 projection;
    mat4 viewProjection;

    mat4 invView;
    mat4 invProjection;

    real32 vFov;
    real32 aspect;
    real32 nearClip;
    real32 farClip;
};

struct ScreenUniformBuffer {
    real32 width;
    real32 height;
};


#define GBUFFER_DIFFUSE_BUFFER 0
#define GBUFFER_NORMAL_BUFFER 1



inline void AddRenderKeyValue(uint32 *key, uint32 *bitsLeft, uint32 value, uint32 sizeInBits) {
    *bitsLeft -= sizeInBits;
    *key |= (value & ((1 << sizeInBits) - 1)) << (*bitsLeft);
}

enum GBufferRenderStage {
    GBufferStage_Setup,
    GBufferStage_Render,
};

inline uint32 GenerateGBufferRenderKey(uint32 sceneIndex, GBufferRenderStage stage, uint32 *bitsLeft = NULL) {
    uint32 key = 0;
    uint32 tempBitsLeft = 32;

    AddRenderKeyValue(&key, &tempBitsLeft, sceneIndex, 2);
    AddRenderKeyValue(&key, &tempBitsLeft, stage, 8);

    if (bitsLeft != NULL) {
        *bitsLeft = tempBitsLeft;
    }

    return key;
}

enum ShadingRenderStage {
    ShadingStage_Setup,
    
    ShadingStage_PreLightAccumulation,
    ShadingStage_LightAccumulation,
    ShadingStage_PostLightAccumulation,

    ShadingStage_PreSkybox,
    ShadingStage_Skybox,
    ShadingStage_PostSkybox,

    ShadingStage_Blit,
    ShadingStage_PostBlit,
};

inline uint32 GenerateShadingRenderKey(b32 isPostScene, uint32 sceneIndex, ShadingRenderStage stage, uint32 *bitsLeft = NULL) {
    uint32 key = 0;
    uint32 tempBitsLeft = 32;

    AddRenderKeyValue(&key, &tempBitsLeft, (bool)isPostScene, 1);
    AddRenderKeyValue(&key, &tempBitsLeft, sceneIndex, 2);
    AddRenderKeyValue(&key, &tempBitsLeft, stage, 8);

    if (bitsLeft != NULL) {
        *bitsLeft = tempBitsLeft;
    }

    return key;
}


#define PAINTERLY_DEPTH_PREPASS_SETUP_KEY 1
#define PAINTERLY_DEPTH_PREPASS_RENDER_KEY 2
#define PAINTERLY_DEPTH_PREPASS_TEARDOWN_KEY 3

#define PAINTERLY_SETUP_KEY 4

#define PAINTERLY_SPLAT_SETUP_KEY 10
#define PAINTERLY_SPLAT_RENDER_KEY 11

#define PAINTERLY_BLEND_KEY 1 << 29
#define PAINTERLY_SMEAR_KEY 1 << 30
#define PAINTERLY_BLIT_KEY 1 << 31


#define TEXT_SETUP_KEY 0
#define TEXT_RENDER_KEY 1
#define TEXT_BLIT_KEY 1 << 31


// Blitting to screen
void RenderPassthrough(uint32 textureID);
void RenderPassthrough(CommandList *commandList, uint32 key, uint32 textureID, bool enableBlending = false);
void RenderPassthroughDepth(CommandList *commandList, real32 nearClip, uint32 key, uint32 textureID);
void RenderPassthroughNormal(CommandList *commandList, uint32 key, uint32 textureID);

void RenderPassthroughBlended(CommandList *commandList, uint32 key, uint32 *textureIDs, uint32 textureCount, uint32 blendMapTextureID);

void RenderPassthroughNormalBlended(CommandList *commandList, uint32 key, uint32 *textureIDs, uint32 textureCount, uint32 blendMapTextureID);

void RenderSkybox(CommandList *commandList, uint32 key, uint32 cubemapID);


// Text Rendering
struct GlyphAttributes {
    uint32 codepoint;

    vec3 position;
    quaternion rotation;

    vec2 dimensions;

    vec4 color;

    vec2 scale;
    vec2 anchor;

    real32 thickness;
    real32 outline;
    vec4 outlineColor;

    vec4 shadowColor;
    real32 shadowSoftness;
    real32 shadowThickness;
    vec2 shadowOffset;
};

// @TODO: build the modelMatrix instead of storing it in the renderbuffer,
// and just use it in the shader instead of building it there.

struct TextRenderBuffer {
    FontTableHandle font;

    TextRenderBuffer *nextBuffer;

    uint32 glyphCapacity;
    uint32 glyphCount;
    GlyphAttributes *glyphs;

    // @Note: each call to DrawGlyphs() produces a new
    // TextRenderBuffer()! This might be fine, but if it becomes a
    // performance problem, a separate buffer of headers that point
    // into a large buffer can be used. Then there would be a single
    // draw call per header w/ larger buffers, rather than a single
    // draw call per buffer w/ smaller buffers. -- hz feb 5 2017
    mat4 model;

    uint32 bufferID;
    uint32 bufferSize;
};

void DrawGlyphs(TextRenderBuffer **buffers, FontTableHandle font, uint32 glyphCount, GlyphAttributes *attribs, mat4 model, bool *renderMask = NULL);

void DrawGlyphs(FontTableHandle font, uint32 glyphCount, GlyphAttributes *attribs, mat4 model, bool *renderMask = NULL);

// @TODO: there should be a non-copy DrawGlyphs() API that gives the user a GPU buffer to write into


// CPU Image Routines
inline void RectilinearBlit(void *dest, uint32 destWidth, uint32 destHeight,
                            uint32 destRegionX, uint32 destRegionY,
                            void *src, uint32 srcWidth, uint32 srcHeight,
                            uint32 srcRegionX, uint32 srcRegionY,
                            uint32 regionWidth, uint32 regionHeight,
                            uint8 bytesPerPixel) {

    ASSERT(destRegionX + regionWidth <= destWidth);
    ASSERT(destRegionY + regionHeight <= destHeight);

    ASSERT(srcRegionX + regionWidth <= srcWidth);
    ASSERT(srcRegionY + regionHeight <= srcHeight);

    // @NOTE: assumes there are always 4 components
    uint32 lineWidthBytes = bytesPerPixel * regionWidth;
        
    for (int y = 0; y < regionHeight; y++) {
        void *destLine = BUFFER_OFFSET(dest, bytesPerPixel * (destWidth * (destRegionY + y) + destRegionX));
        void *srcLine = BUFFER_OFFSET(src, bytesPerPixel * (srcWidth * (srcRegionY + y) + srcRegionX));

        Plat.CopyMem(destLine, srcLine, lineWidthBytes);
    }
}

template <typename T>
inline T PointSample(T *image, uint32 imageWidth, uint32 imageHeight, vec2 sampleLocation) {
    vec2 texelCoord = V2(sampleLocation.x * imageWidth, sampleLocation.y * imageHeight);

    int32 coordX = Clamp((int32)Roundf(texelCoord.x), 0, imageWidth - 1);
    int32 coordY = Clamp((int32)Roundf(texelCoord.y), 0, imageHeight - 1);

    T sample = image[coordY * imageWidth + coordX];

    return sample;
}

template <typename T>
inline T BilinearSample(T *image, uint32 imageWidth, uint32 imageHeight, vec2 sampleLocation) {
    vec2 texelCoord = V2(sampleLocation.x * imageWidth, sampleLocation.y * imageHeight);
    vec2 texelCoordFract = V2(texelCoord.x - Floorf(texelCoord.x),
                              texelCoord.y - Floorf(texelCoord.y));

    int32 coordX0 = Clamp((int32)Floorf(texelCoord.x), 0, imageWidth - 1);
    int32 coordX1 = Clamp((int32)Ceilf(texelCoord.x), 0, imageWidth - 1);
    int32 coordY0 = Clamp((int32)Floorf(texelCoord.y), 0, imageHeight - 1);
    int32 coordY1 = Clamp((int32)Ceilf(texelCoord.y), 0, imageHeight - 1);

    T sample00 = image[coordY0 * imageWidth + coordX0];
    T sample10 = image[coordY0 * imageWidth + coordX1];
    T sample01 = image[coordY1 * imageWidth + coordX0];
    T sample11 = image[coordY1 * imageWidth + coordX1];

    T result = Lerp(Lerp(sample00, sample10, texelCoordFract.x),
                    Lerp(sample01, sample11, texelCoordFract.x),
                    texelCoordFract.y);
    return result;
}

#endif

