
// Scene_None
void SceneNoneInit(EtcGameMemory *mem, SceneData *data) {
}

void SceneNoneDeinit(EtcGameMemory *mem, SceneData *data) {
}

void SceneNoneUpdate(EtcGameMemory *mem, SceneData *data) {
    uint32 key = GenerateGBufferRenderKey(0, GBufferStage_Render);
    ClearColor(&mem->graphics.gbufferPass, key, V4(0.1f, 0.1f, 0.1f, 1.0f), 0);
}


// SceneManager
void InitSceneManager(SceneManager *sceneManager) {
    sceneManager->scenes = MakeChunkedArray<SceneData>(Game.heap, 16);
    sceneManager->scenesFreeList = MakeDynamicArray<uint32>(16);
}

void DeinitSceneManager(SceneManager *sceneManager) {
    DeallocateChunkedArray(&sceneManager->scenes);
    DeallocateDynamicArray(&sceneManager->scenesFreeList);
    DeallocateDynamicArray(&sceneManager->loadingScenes);
}

void LoadAssetList(SceneData *scene);

void UnloadMarkedScenes_(SceneManager *sceneManager) {
    for (uint32 i = 0; i < sceneManager->scenes.count; i++) {
        SceneData *scene = &sceneManager->scenes[i];
        if (scene->state != SceneState_MarkedForUnload) {
            continue;
        }

        SceneID sceneID = scene->sceneID;
        SceneState prevState = scene->state;

        SceneDeinitFunc *deinit = SceneDeinitFuncTable[(uint32)sceneID];
        if (deinit != NULL) {
            sceneManager->activeScene = MakeSceneHandle(i, scene->generation);
            sceneManager->activeSceneIndex = -1;

            deinit(Game.mem, scene);

            sceneManager->activeScene = SceneHandleInvalid();
            sceneManager->activeSceneIndex = -1;
        }

        DeallocateMemoryArenaList(&scene->arena);
        DeallocateEntityDB(&scene->entityDB);
        DeallocateDynamicArray(&scene->children);

        PushBack(&sceneManager->scenesFreeList, i);
        scene->state = SceneState_Free;
    }
}

void RemoveFreedActiveScenes_(SceneManager *sceneManager) {
    if (sceneManager->nextActive.sceneCount > 0) {
        return;
    }

    ActiveSceneSet newSet = sceneManager->active;
    newSet.sceneCount = 0;

    b32 useNewSet = false;

    for (int i = 0; i < sceneManager->active.sceneCount; i++) {
        SceneHandle sceneHandle = sceneManager->active.scenes[i];
        SceneData *scene = GetScene(sceneHandle);

        if (scene->state == SceneState_Free) {
            useNewSet = true;
        }
        else {
            newSet.scenes[newSet.sceneCount++] = sceneHandle;
        }
    }

    if (useNewSet) {
        sceneManager->nextActive = newSet;
    }
}

void UpdateSceneManager(SceneManager *sceneManager) {
    static const uint32 assetsToLoadPerFrame = 4;
    static const uint32 assetsToCheckPerFrame = 128;

    for (int i = sceneManager->loadingScenes.count - 1; i >= 0; i--) {
        SceneHandle sceneHandle = sceneManager->loadingScenes[i];
        SceneData *scene = GetScene(sceneHandle);

        if (scene == NULL) {
            RemoveAtIndex(&sceneManager->loadingScenes, i);
            continue;
        }

        switch (scene->state) {
            case SceneState_LoadingAssets: {

                if (!scene->hasLoadedAssetList) {
                    LoadAssetList(scene);
                }

                // Start asset loads
                for (int j = 0; j < assetsToLoadPerFrame; j++) {
                    uint32 assetIndex = scene->assetLoadsStarted;
                    if (assetIndex >= scene->assetListTypes.count) {
                        break;
                    }

                    AssetType assetType = scene->assetListTypes[assetIndex];
                    const char *assetName = scene->assetListNames[assetIndex];

                    PushBack(&scene->assets, LoadAsset(Game.assets, assetType, assetName));
                    PushBack(&scene->assetLoaded, (b32)false);

                    scene->assetLoadsStarted++;
                }

                // Check if started loads have finished
                if (scene->assets.count > 0) {
                    uint32 startCheckIndex = scene->assetLoadedCheckOffset;

                    for (int j = 0; j < assetsToCheckPerFrame; j++) {
                        uint32 assetIndex = scene->assetLoadedCheckOffset;
                        if (!scene->assetLoaded[assetIndex]) {
                            AssetHandle handle = scene->assets[assetIndex];
                            AssetData *head = handle.asset->assetDataQueueHead;
                            if (head != NULL &&
                                (head->state == AssetDataState_Loaded ||
                                 head->state == AssetDataState_LoadFailed ||
                                 head->state == AssetDataState_Active)) {
                                scene->assetLoaded[assetIndex] = true;
                                scene->assetLoadsCompleted++;
                            }
                        }

                        scene->assetLoadedCheckOffset = (assetIndex + 1) % scene->assets.count;
                        if (scene->assetLoadedCheckOffset == startCheckIndex) {
                            break;
                        }
                    }
                }

                if (scene->assetLoadsCompleted == scene->assetListNames.count) {
                    scene->state = SceneState_LoadingEntities;
                    CloseFile(&scene->assetFile);
                }
            } break;

            case SceneState_LoadingEntities: {
                DeserializeScene(scene->sceneID, scene);

                SceneInitFunc *init = SceneInitFuncTable[scene->sceneID];
                if (init != NULL) {
                    sceneManager->activeScene = sceneHandle;
                    sceneManager->activeSceneIndex = -1;

                    init(Game.mem, scene);

                    sceneManager->activeScene = SceneHandleInvalid();
                    sceneManager->activeSceneIndex = -1;
                }

                if (scene->children.count > 0) {
                    scene->state = SceneState_LoadingChildren;
                }
                else {
                    scene->state = SceneState_Loaded;
                }
            } break;

            case SceneState_LoadingChildren: {
                bool allScenesLoaded = true;
                for (int i = 0; i < scene->children.count; i++) {
                    if (SceneData *child = GetScene(scene->children[i])) {
                        if (child->state != SceneState_Loaded) {
                            allScenesLoaded = false;
                            break;
                        }
                    }
                }

                if (allScenesLoaded) {
                    scene->state = SceneState_Loaded;
                }
            } break;

            default: {
                ASSERT(false);
            } break;
        }

        if (scene->state == SceneState_Loaded) {
            RemoveAtIndex(&sceneManager->loadingScenes, i);
        }
    }


    UnloadMarkedScenes_(sceneManager);
    RemoveFreedActiveScenes_(sceneManager);


    // Set new activeScene list
    if (sceneManager->nextActive.sceneCount > 0) {
        sceneManager->active = sceneManager->nextActive;
        sceneManager->nextActive = {};
    }
}

void UpdateDebugCamera(SceneData *activeScene) {
    EntityDB *entityDB = &activeScene->entityDB;

    CameraSystem *camSys = GetSystem(entityDB, Camera);
    TransformSystem *transSys = GetSystem(entityDB, Transform);

    if (InputActionPressed(Game.inputActions, InputAction_ToggleDebugCamera)) {
        EditorCamera *editorCamera = &activeScene->editorCamera;

        EntityID currentCameraID = camSys->mainCameraID;

        if (editorCamera->entityID == EntityID_Invalid) {
            vec3 mainCameraPos;
            quaternion mainCameraRot;

            if (camSys->mainCameraID == EntityID_Invalid) {
                mainCameraPos = V3(0);
                mainCameraRot = IdentityQuaternion();
            }
            else {
                mainCameraPos = GetWorldPosition(transSys, camSys->mainCameraID);
                mainCameraRot = GetWorldRotation(transSys, camSys->mainCameraID);
            }
            
            InitEditorCamera(editorCamera, entityDB, mainCameraPos, mainCameraRot);
        }

        if (editorCamera->entityID != currentCameraID) {
            editorCamera->cameraIDToReturnTo = currentCameraID;
            activeScene->usingEditorCamera = true;

            InputActionType actionsToBlock[] = {
                InputAction_MoveX,
                InputAction_MoveY,
                InputAction_FreeLookX,
                InputAction_FreeLookY,
                InputAction_LeftHandUse,
                InputAction_RightHandUse,
                InputAction_LeftHandUseAnalogue,
                InputAction_RightHandUseAnalogue,
                InputAction_Action1,
                InputAction_Action2,
                InputAction_Action3,
                InputAction_Action4,
                InputAction_FreeLookXY,
            };
            //BlockTheseInputActions(Game.inputActions, actionsToBlock, ARRAY_LENGTH(InputActionType, actionsToBlock));
            // @TODO: what we want isnt to block the action, but to block the device for that action
            // so you can still look, but it cant come from the mouse
        }
        else {
            if (editorCamera->cameraIDToReturnTo != EntityID_Invalid) {
                activeScene->usingEditorCamera = false;
                SetMainCamera(camSys, editorCamera->cameraIDToReturnTo);
                UnblockAllInputActions(Game.inputActions);
            }
        }
    }

    if (activeScene->usingEditorCamera) {
        UpdateEditorCamera(&activeScene->editorCamera);
    }
}

void RunActiveScenesUpdate(SceneManager *sceneManager, EtcGameMemory *mem, b32 paused) {
    b32 allScenesLoaded = true;

    for (int i = 0; i < sceneManager->active.sceneCount; i++) {
        SceneData *scene = GetScene(sceneManager->active.scenes[i]);
        if (scene->state != SceneState_Loaded) {
            allScenesLoaded = false;
        }
    }

    if (allScenesLoaded) {
        for (int32 i = 0; i < sceneManager->active.sceneCount; i++) {
            SceneHandle activeSceneHandle = sceneManager->active.scenes[i];
            SceneData *activeScene = GetScene(activeSceneHandle);
            EntityDB *entityDB = &activeScene->entityDB;

            sceneManager->activeScene = activeSceneHandle;
            sceneManager->activeSceneIndex = i;


            // Setup gbuffer
            {
                uint32 setupKey = GenerateGBufferRenderKey(Game.sceneManager->activeSceneIndex, GBufferStage_Setup);

                CommandList *commandList = &Game.graphics->gbufferPass;
                RenderTarget *gbufferTarget = &Game.graphics->gbufferTargets[i];

                SetDrawRenderTarget(commandList, setupKey, gbufferTarget);
                SetViewport(commandList, setupKey, V2i(0, 0), V2i(gbufferTarget->width, gbufferTarget->height));

                uint32 colorBuffers[] = {GL_COLOR_ATTACHMENT0, GL_COLOR_ATTACHMENT1};
                SetColorBuffers(commandList, setupKey, 2, colorBuffers);

                ClearColor(commandList, setupKey, CLEAR, GBUFFER_DIFFUSE_BUFFER);
                ClearColor(commandList, setupKey, CLEAR, GBUFFER_NORMAL_BUFFER);

                EnableDepthWrite(commandList, setupKey);
                ClearDepth(commandList, setupKey, 1.0f);


                // CameraUniformBlock
                BindBuffer(commandList, setupKey, GL_UNIFORM_BUFFER, Game.graphics->cameraUniformBufferID);
                CopyBufferSubData(commandList, setupKey, GL_UNIFORM_BUFFER, 0, sizeof(CameraUniformBuffer), &activeScene->cameraUniformBuffer);
                BindBuffer(commandList, setupKey, GL_UNIFORM_BUFFER, 0);
            }

            // Setup shading accumulation buffer
            {
                uint32 setupKey = GenerateShadingRenderKey(false, Game.sceneManager->activeSceneIndex, ShadingStage_Setup);
                uint32 lightAccumSetupKey = GenerateShadingRenderKey(false, Game.sceneManager->activeSceneIndex, ShadingStage_PreLightAccumulation);
                uint32 lightAccumTeardownKey = GenerateShadingRenderKey(false, Game.sceneManager->activeSceneIndex, ShadingStage_PostLightAccumulation);

                CommandList *commandList = &Game.graphics->shadingPass;
                RenderTarget *shadingTarget = &Game.graphics->shadingTargets[i];

                SetDrawRenderTarget(commandList, setupKey, shadingTarget);
                SetViewport(commandList, setupKey, V2i(0, 0), V2i(shadingTarget->width, shadingTarget->height));

                uint32 colorBuffers[] = {GL_COLOR_ATTACHMENT0};
                SetColorBuffers(commandList, setupKey, 2, colorBuffers);

                ClearColor(commandList, setupKey, CLEAR);

                EnableBlending(commandList, lightAccumSetupKey, GL_FUNC_ADD, GL_ONE, GL_ONE);
                DisableBlending(commandList, lightAccumTeardownKey);


                // CameraUniformBlock
                BindBuffer(commandList, setupKey, GL_UNIFORM_BUFFER, Game.graphics->cameraUniformBufferID);
                CopyBufferSubData(commandList, setupKey, GL_UNIFORM_BUFFER, 0, sizeof(CameraUniformBuffer), &activeScene->cameraUniformBuffer);
                BindBuffer(commandList, setupKey, GL_UNIFORM_BUFFER, 0);
            }


#if ETC_DEBUG
            // @NOTE: this can go here because SetMainCamera will not
            // override the editorCamera as the main camera if
            // 'SceneData.usingEditorCamera' is true so we dont have
            // to worry about something in update or tick changing the
            // camera away from the editorCamera
            if (i == sceneManager->active.renderedIndex) {
                UpdateDebugCamera(activeScene);
            }
#endif

            // Scene Update 
            if (!paused) {
                SceneUpdateFunc *updateFunc = SceneUpdateFuncTable[activeScene->sceneID];
                updateFunc(mem, activeScene);
                TickSystems(entityDB, mem);

                ProcessDeletedEntities(entityDB);
            }
            else {
                // @TODO: it would be nice if this was just a property on the system
                TickCameraSystem(mem, entityDB, GetSystem(entityDB, Camera));
                TickStaticMeshSystem(mem, entityDB, GetSystem(entityDB, StaticMesh));
                TickLightSystem(mem, entityDB, GetSystem(entityDB, Light));
            }
        }
    }
    else {
        SceneUpdateFunc *updateFunc = SceneUpdateFuncTable[Scene_None];
        updateFunc(mem, NULL);
    }

    sceneManager->activeScene = SceneHandleInvalid();
    sceneManager->activeSceneIndex = -1;
}

SceneData *GetScene(SceneHandle handle) {
    SceneManager *sceneManager = Game.sceneManager;

    if (!SceneHandleIsValid(handle)) {
        return NULL;
    }

    if (handle.index < sceneManager->scenes.count) {
        SceneData *sceneData = &sceneManager->scenes[handle.index];
        if (sceneData->generation == handle.generation) {
            return sceneData;
        }
    }

    return NULL;
}

void *GetSceneData(SceneHandle handle) {
    if (SceneData *sceneData = GetScene(handle)) {
        return sceneData->data;
    }

    return NULL;
}

SceneData *GetActiveScene() {
    return GetScene(Game.sceneManager->activeScene);
}

SceneData *GetSceneFromActiveSceneSet(u32 index) {
    return GetScene(Game.sceneManager->active.scenes[index]);
}

SceneHandle GetRootScene(SceneHandle handle) {
    SceneData *scene = GetScene(handle);
    if (scene == NULL) {
        return SceneHandleInvalid();
    }

    while ((scene = GetScene(scene->owner)) != NULL) {
        handle = scene->handle;
    }

    return handle;
}

bool FindLoadedScene(SceneManager *sceneManager, SceneID sceneID, SceneHandle owner, SceneHandle *handle) {
    for (int i = 0; i < sceneManager->scenes.count; i++) {
        SceneData *scene = &sceneManager->scenes[i];
        if (scene->state == SceneState_Free) {
            continue;
        }

        if (scene->sceneID == sceneID && scene->owner == owner) {
            handle->index = i;
            handle->generation = scene->generation;
            return true;
        }
    }

    return false;
}

SceneHandle LoadScene(SceneID sceneID, SceneHandle owner) {
    // Attempt to find an already loaded version of the scene that
    // isn't opened by an editor and has the same owner
    SceneHandle handle;
    if (FindLoadedScene(Game.sceneManager, sceneID, owner, &handle)) {
        if (SceneData *scene = GetScene(handle)) {
            if (scene->state != SceneState_MarkedForUnload) {
                return handle;
            }
        }
    }

    SceneData *scene;
    uint32 sceneSlot;
    if (PopFront(&Game.sceneManager->scenesFreeList, &sceneSlot)) {
        scene = &Game.sceneManager->scenes[sceneSlot];
        ASSERT(scene->state == SceneState_Free);
    }
    else {
        sceneSlot = Game.sceneManager->scenes.count;
        scene = PushBackPtr(&Game.sceneManager->scenes);
    }

    uint32 prevGeneration = scene->generation;

    Plat.ClearMem(scene, sizeof(SceneData));
    scene->sceneID = sceneID;
    scene->generation = prevGeneration + 1;

    handle = {};
    handle.index = sceneSlot;
    handle.generation = scene->generation;

    scene->handle = handle;

    scene->owner = owner;
    if (SceneData *ownerScene = GetScene(owner)) {
        PushBack(&ownerScene->children, handle);
    }

    scene->state = SceneState_LoadingAssets;

    AllocateMemoryArenaList(&scene->arena, SCENE_ARENA_STORAGE_SIZE);
    AllocateEntityDB(&scene->entityDB);

    PushBack(&Game.sceneManager->loadingScenes, handle);

    return handle;
}

void UnloadScene(SceneHandle handle) {
    if (SceneData *scene = GetScene(handle)) {
        scene->state = SceneState_MarkedForUnload;

        for (int i = 0; i < scene->children.count; i++) {
            UnloadScene(scene->children[i]);
        }
    }
}

void UnloadAllScenes() {
    for (int i = 0; i < Game.sceneManager->scenes.count; i++) {
        SceneData *scene = &Game.sceneManager->scenes[i];
        if (scene->state == SceneState_Loaded) {
            scene->state = SceneState_MarkedForUnload;
        }
    }

    UnloadMarkedScenes_(Game.sceneManager);
}


b32 IsSceneActive(SceneHandle handle) {
    for (int i = 0; i < Game.sceneManager->active.sceneCount; i++) {
        if (handle == Game.sceneManager->active.scenes[i]) {
            return true;
        }
    }

    return false;
}

b32 SetActiveScene(SceneHandle handle) {
    return SetActiveScenes(&handle, 1);
}

b32 SetActiveScenes(SceneHandle *handles, uint32 handleCount, b32 shouldBlend, uint32 renderedIndex) {
    ASSERT(handleCount <= MAX_ACTIVE_SCENES);

    ActiveSceneSet newSet = {};
    newSet.shouldBlend = shouldBlend;
    newSet.renderedIndex = renderedIndex;

    b32 useNewSet = true;

    for (int i = 0; i < handleCount; i++) {
        SceneData *scene = GetScene(handles[i]);
        if (scene == NULL ||
            scene->state == SceneState_Free || scene->state == SceneState_MarkedForUnload) {

            useNewSet = false;
            break;
        }

        newSet.scenes[newSet.sceneCount++] = handles[i];
    }

    if (useNewSet) {
        Game.sceneManager->nextActive = newSet;
    }

    return useNewSet;
}


// Serialization
const char *SceneDirectoryPath(SceneID sceneID) {
    const char *sceneName = GetSceneName(sceneID);
    const char *path = PathConcat("scenes", sceneName);

    return path;
}

const char *SceneManifestPath(SceneID sceneID, const char *sceneDirectory) {
    const char *sceneName = GetSceneName(sceneID);

    char sceneManifestName[256] = {};
    snprintf(sceneManifestName, sizeof(sceneManifestName), "%s.manifest", sceneName);
    const char *manifestPath = PathConcat(sceneDirectory, sceneManifestName);

    return manifestPath;
}

const char *SceneAssetListPath(SceneID sceneID, const char *sceneDirectory) {
    const char *sceneName = GetSceneName(sceneID);

    char assetListName[256] = {};
    snprintf(assetListName, sizeof(assetListName), "%s.assets", sceneName);
    const char *assetListPath = PathConcat(sceneDirectory, assetListName);

    return assetListPath;
}

const char *SceneTransformTreePath(SceneID sceneID, const char *sceneDirectory) {
    const char *sceneName = GetSceneName(sceneID);

    char transformTreeName[256] = {};
    snprintf(transformTreeName, sizeof(transformTreeName), "%s.transform_tree", sceneName);
    const char *transformTreePath = PathConcat(sceneDirectory, transformTreeName);

    return transformTreePath;
}

const char *SceneEntityPath(const char *sceneDirectory, EntityID entityID) {
    char entityFileName[64];
    snprintf(entityFileName, sizeof(entityFileName), "%" PRIu64 ".entity", entityID);
    const char *entityFilePath = PathConcat(sceneDirectory, entityFileName);

    return entityFilePath;
}

void WriteComponentAssetDependencies(EtcTxtWriteState *writer, const char *sceneName, const char *typeName, void *value, ChunkedArray<const char *> *assetNamesWritten) {
    Schema *schema = Game.GetSchema(typeName);
    if (schema != NULL) {
        if (schema->type == SchemaType_Struct) {
            for (int i = 0; i < schema->_struct.fields.count; i++) {
                SchemaField field = schema->_struct.fields[i];
                void *offsetValue = (void *)((u8 *)value + field.byteOffset);
                
                if (field.type.dataType == SchemaDataType_Struct) {
                    WriteComponentAssetDependencies(writer, sceneName, field.type.typeName, offsetValue, assetNamesWritten);
                }
                else if (field.type.dataType == SchemaDataType_DynamicArray) {
                    DynamicArray_Untyped *array = (DynamicArray_Untyped *)offsetValue;

                    SchemaTypeDescription elementType = field.type.templateTypes[0];
                    uint32 elementSize = elementType.sizeInBytes;

                    if (elementType.dataType == SchemaDataType_Struct) {
                        for (int j = 0; j < array->count; j++) {
                            void *element = (void *)((uint8 *)array->data + j * elementSize);
                            WriteComponentAssetDependencies(writer, sceneName, elementType.typeName, element, assetNamesWritten);
                        }
                    }
                }
                else if (field.type.dataType == SchemaDataType_SchemaData) {
                    
                }
                else if (field.type.dataType == SchemaDataType_AssetHandle) {
                    AssetHandle *handle = (AssetHandle *)((uint8 *)value + field.byteOffset);

                    if (AssetHandleIsProcedural(*handle)) {
                        continue;
                    }

                    if (handle->type != AssetType_Invalid && handle->asset != NULL) {
                        if (!Contains(assetNamesWritten, handle->asset->name)) {
                            WriteEnumValue(writer, ENUM_STRINGVALUE(AssetType, handle->type), handle->asset->name);
                            PushBack(assetNamesWritten, handle->asset->name);
                        }
                    }
                }
            }
        }
    }
}

void SortEntityIDsInTransformTree(EntityDB *entityDB, ChunkedArray<EntityID> *sortedIDs) {
    EntityIterator it = EntityIterator(entityDB);
    EntityID id;
    while (GetNextEntityWithComponent(&it, ComponentType_Transform, &id)) {
        PushBack(sortedIDs, id);
    }
}

int32 TEMPEntityIDComparator(const void *a, const void *b) {
    EntityID idA = *(EntityID *)a;
    EntityID idB = *(EntityID *)b;
    if (idA < idB) {
        return -1;
    }
    if (idA > idB) {
        return 1;
    }

    return 0;
}

void SerializeEntity(EtcTxtWriteState *writer, EntityDB *entityDB, EntityID id) {
    if (EntityIDIsTransient(id)) {
        return;
    }
    
    static uint32 sceneSerializationVersion = 0;
    WriteUInt32Value(writer, sceneSerializationVersion, "sceneSerializationVersion");

    WriteUInt64Value(writer, id, "entityID");

    for (int type = 1; type < ENUM_VALUECOUNT(ComponentType); type++) {
        ComponentValue componentValue = {};
        if (GetComponentValue_(entityDB, (ComponentType)type, id, &componentValue)) {
            const char *componentName = Game.GetComponentTypeName((ComponentType)type);
            Schema *schema = ComponentSchemaFromComponentType((ComponentType)type);
            ASSERT(schema != NULL);

            SerializeValue(writer, schema, &componentValue, componentName);
        }
    }
}

void SerializeScene(SceneID sceneID, SceneData *scene) {
    ComponentValue componentValues[ENUM_VALUECOUNT(ComponentType)] = {};

    const char *sceneDirectory = SceneDirectoryPath(sceneID);
    const char *manifestPath = SceneManifestPath(sceneID, sceneDirectory);
    const char *assetListPath = SceneAssetListPath(sceneID, sceneDirectory);
    const char *transformPath = SceneTransformTreePath(sceneID, sceneDirectory);

    static uint32 sceneSerializationVersion = 0;

#if ETC_DEBUG
    EditorCamera editorCamera = scene->editorCamera;

    vec3 editorCameraPrevPos = GetWorldPosition(GetSystem(&scene->entityDB, Transform), editorCamera.entityID);
    quaternion editorCameraPrevRot = GetWorldRotation(GetSystem(&scene->entityDB, Transform), editorCamera.entityID);
    if (editorCamera.entityID != EntityID_Invalid) {
        DeleteEntity(&scene->entityDB, editorCamera.entityID);
    }
#endif

    Plat.MakeDirectory("scenes");
    Plat.MakeDirectory(sceneDirectory);

    const char *sceneName = GetSceneName(sceneID);

    EntityDB *entityDB = &scene->entityDB;

    // Write asset list
    EtcTxtWriteState assetList;
    if (OpenFileForWrite(&assetList, assetListPath)) {
        WriteUInt32Value(&assetList, sceneSerializationVersion, "sceneSerializationVersion");

        ChunkedArray<const char *> assetNamesWritten = MakeChunkedArray<const char *>(Game.frameMem, 1024);
        
        uint32 assetCount = 0;
        for (uint64 id = 1; id < entityDB->nextID; id++) {
            b32 hasComponent[ENUM_VALUECOUNT(ComponentType)] = {};
            ComponentValue componentValues[ENUM_VALUECOUNT(ComponentType)] = {};
            uint32 componentCount = 0;

            for (int type = 1; type < entityDB->systemCount; type++) {
                ComponentValue value = {};
                if (GetComponentValue_(entityDB, (ComponentType)type, id, &value)) {
                    WriteComponentAssetDependencies(&assetList, sceneName, Game.components.typeNames[type], &value, &assetNamesWritten);
                }
            }   
        }

        // @NOTE: we do need to do this for the script belonging to the scene I think
        // for (int node = 0; node < script->nodes.count; node++) {

        //     char *textDataName = (char *)StringConcat(Game.frameMem, 3, script->nodes[node].name, "_", sceneName);
        //     WriteEnumValue(&assetList, ENUM_STRINGVALUE(AssetType, AssetType_TextData), textDataName);
        // }

        CloseFile(&assetList);
    }

    EntityID *entityIDs = PushArray(Game.frameMem, EntityID, entityDB->entityCount);

    uint64 nonTransientEntityCount = entityDB->entityCount - entityDB->transientEntityCount;

    EtcTxtWriteState manifestFile;
    if (OpenFileForWrite(&manifestFile, manifestPath)) {
        WriteUInt32Value(&manifestFile, sceneSerializationVersion, "sceneSerializationVersion");

        // Write entityDB state
        Schema *entityIDRangeSchema = Game.GetSchema("EntityIDRange");
        ASSERT(entityIDRangeSchema != NULL);

        WritePushArrayValue(&manifestFile, "deletedRanges");

        for (int i = 0; i < entityDB->deletedRanges.count; i++) {
            SerializeValue(&manifestFile, entityIDRangeSchema, &entityDB->deletedRanges[i], NULL);
        }

        WritePopArrayValue(&manifestFile);
        
        
        WriteUInt64Value(&manifestFile, entityDB->nextID, "nextID");
        WriteUInt64Value(&manifestFile, nonTransientEntityCount, "entityCount");

        CloseFile(&manifestFile);
    }

    // Write transform tree
    EtcTxtWriteState transformTree;
    if (OpenFileForWrite(&transformTree, transformPath)) {
        ChunkedArray<EntityID> sortedIDs = MakeChunkedArray<EntityID>(Game.frameMem, sizeof(EntityID) * 64);

        SortEntityIDsInTransformTree(entityDB, &sortedIDs);

        WriteUInt64Value(&transformTree, sortedIDs.count, "entityCount");

        for (int i = 0; i < sortedIDs.count; i++) {
            // @WARNING: things could go badly if we have non-transient entities parented to transient ones
            // because then the transform tree will look like they are parented to the last non transient entity
            if (EntityIDIsTransient(sortedIDs[i])) {
                continue;
            }
            WriteUInt64Value(&transformTree, sortedIDs[i], "entityID");
        }

        CloseFile(&transformTree);
    }
        
    // Write Entity files
    for (uint64 id = 1; id < entityDB->nextID; id++) {
        const char *entityFilePath = SceneEntityPath(sceneDirectory, id);

        if (EntityIsDeleted(entityDB, id)) {
            continue;
        }

        EtcTxtWriteState entityFile;
        if (OpenFileForWrite(&entityFile, entityFilePath)) {
            SerializeEntity(&entityFile, entityDB, id);
            CloseFile(&entityFile);
       }

        Plat.DeallocateMem((char *)entityFilePath);
    }


#if ETC_DEBUG
    if (editorCamera.entityID != EntityID_Invalid) {
        InitEditorCamera(&scene->editorCamera, &scene->entityDB, editorCameraPrevPos, editorCameraPrevRot);
    }
#endif
    
    Plat.DeallocateMem((char *)sceneDirectory);
    Plat.DeallocateMem((char *)manifestPath);
}

void LoadAssetList(SceneData *scene) {
    const char *sceneDirectory = SceneDirectoryPath(scene->sceneID);
    const char *assetListPath = SceneAssetListPath(scene->sceneID, sceneDirectory);

    scene->assetListTypes = {};
    scene->assetListNames = {};

    // We need to keep this file open until asset loading has finished
    if (OpenFileForRead(&scene->assetFile, assetListPath)) {

        uint32 sceneSerializationVersion;
        ParseIdentifier(&scene->assetFile, NULL, "sceneSerializationVersion");
        ParseUInt32Value(&scene->assetFile, &sceneSerializationVersion);
        ASSERT(sceneSerializationVersion == 0);
        
        const char *assetName;
        while (ParseIdentifier(&scene->assetFile, &assetName)) {
            PushBack(&scene->assetListNames, assetName);

            AssetType *assetType = PushBackPtr(&scene->assetListTypes);
            ParseEnumValue(&scene->assetFile, AssetType, assetType);
        }
    }

    scene->hasLoadedAssetList = true;
}

b32 DeserializeSceneManifest(const char *path, EntityDB *entityDB) {
    b32 success = true;

    EtcTxtParseState manifestFile;
    if (OpenFileForRead(&manifestFile, path)) {

        uint32 sceneSerializationVersion;
        if (!(ParseIdentifier(&manifestFile, NULL, "sceneSerializationVersion") &&
              ParseUInt32Value(&manifestFile, &sceneSerializationVersion))) {
            success = false;
            goto error;
        }

        // Read entityDB state
        Schema *entityIDRangeSchema = Game.GetSchema("EntityIDRange");
        ASSERT(entityIDRangeSchema != NULL);

        entityDB->deletedRanges = {};

        if (ParseIdentifier(&manifestFile, NULL, "deletedRanges") &&
            ParsePushArrayValue(&manifestFile)) {

            while (!ParseHasError(&manifestFile)) {
                EntityIDRange range;

                if (ParsePopArrayValue(&manifestFile)) {
                    break;
                }
                else if (DeserializeValue(&manifestFile, entityIDRangeSchema, &range)) {
                    PushBack(&entityDB->deletedRanges, range);
                }
                else {
                    success = false;
                    goto error;
                }
            }

        }
        else {
            success = false;
            goto error;
        }

        if (!(ParseIdentifier(&manifestFile, NULL, "nextID") &&
              ParseUInt64Value(&manifestFile, &entityDB->nextID))) {
            success = false;
            goto error;
        }

        if (!(ParseIdentifier(&manifestFile, NULL, "entityCount") &&
              ParseUInt64Value(&manifestFile, &entityDB->entityCount))) {
            success = false;
            goto error;
        }
    }

 error:
    CloseFile(&manifestFile);
    
    return success;
}

bool ReadComponentValueFromEntityFile(EtcTxtParseState *entityFile, EntityDB *entityDB, EntityID id, ComponentType type, ComponentValue *componentValue) {
    bool success = true;

    EtcTxtParseState prevState = *entityFile;
    
    const char *componentName;
    while (ParseIdentifier(entityFile, &componentName)) {

        ComponentType componentType = Game.GetComponentType(componentName);
        Schema *schema = Game.GetSchema(componentName);

        if (componentType == ComponentType_Invalid || schema == NULL || componentType != type) {
            if (ParseSkipValue(entityFile)) {
                continue;
            }
            else {
                success = false;
                goto error;
            }
        }

        if (DeserializeValue(entityFile, schema, componentValue)) {
            break;
        }
        else if (ParseSkipValue(entityFile)) {
            continue;
        }
        else {
            success = false;
            goto error;
        }
    }

 error:
    
    return success;
}

bool ReadComponentValuesFromEntityFile(const char *path, EntityDB *entityDB, EntityID id, ComponentType *types, u32 typeCount, ComponentValue *componentValues) {
    b32 success = true;

    EtcTxtParseState entityFile;
    if (OpenFileForRead(&entityFile, path)) {

        uint32 sceneSerializationVersion;
        if (!(ParseIdentifier(&entityFile, NULL, "sceneSerializationVersion") &&
              ParseUInt32Value(&entityFile, &sceneSerializationVersion))) {
            success = false;
            goto error;
        }

        uint64 deserializedID;
        if (!(ParseIdentifier(&entityFile, NULL, "entityID") &&
              ParseUInt64Value(&entityFile, &deserializedID) &&
              deserializedID == id)) {
            success = false;
            goto error;
        }

        EtcTxtParseState prevState = entityFile;
        for (u32 i = 0; i < typeCount; i++) {
            ReadComponentValueFromEntityFile(&entityFile, entityDB, id, types[i], &componentValues[i]);

            entityFile.byteOffset = prevState.byteOffset;
            entityFile.line = prevState.line;
            entityFile.column = prevState.column;
            entityFile.stackDepth = prevState.stackDepth;
        }
    }

 error:
    CloseFile(&entityFile);
    
    return success;
}

b32 DeserializeEntity(EtcTxtParseState *parser, EntityID id, ChunkedArray<ComponentType> *componentTypes, ChunkedArray<ComponentValue> *componentValues) {
    uint32 sceneSerializationVersion;
    if (!(ParseIdentifier(parser, NULL, "sceneSerializationVersion") &&
          ParseUInt32Value(parser, &sceneSerializationVersion))) {
        return false;
    }

    uint64 deserializedID;
    if (!(ParseIdentifier(parser, NULL, "entityID") &&
          ParseUInt64Value(parser, &deserializedID) &&
          deserializedID == id)) {
        return false;
    }

    const char *componentName;
    while (ParseIdentifier(parser, &componentName)) {

        ComponentType componentType = Game.GetComponentType(componentName);
        Schema *schema = Game.GetSchema(componentName);

        if (componentType == ComponentType_Invalid || schema == NULL) {
            if (ParseSkipValue(parser)) {
                continue;
            }
            else {
                return false;
            }
        }

        ComponentValue componentValue = {};
        if (DeserializeValue(parser, schema, &componentValue)) {
            PushBack(componentTypes, componentType);
            PushBack(componentValues, componentValue);
        }
        else if (ParseSkipValue(parser)) {
            continue;
        }
        else {
            return false;
        }
    }

    return true;
}

b32 DeserializeEntityFile(const char *path, EntityDB *entityDB, EntityID id) {
    b32 success = false;

    EtcTxtParseState entityFile;
    if (OpenFileForRead(&entityFile, path)) {
        ChunkedArray<ComponentType> componentTypes = MakeChunkedArray<ComponentType>(Game.frameMem, 16);
        ChunkedArray<ComponentValue> componentValues = MakeChunkedArray<ComponentValue>(Game.frameMem, 16);

        if (DeserializeEntity(&entityFile, id, &componentTypes, &componentValues)) {
            for (int i = 0; i < componentTypes.count; i++) {
                AddComponent_(entityDB, componentTypes[i], 1, &id, &componentValues[i]);
            }

            success = true;
        }

        CloseFile(&entityFile);
    }
    
    return success;
}

bool DeserializeTransformTree(const char *sceneDirectory, const char *path, EntityDB *entityDB) {
    b32 success = true;

    EtcTxtParseState transformFile;

    TransformSystem *system = GetSystem(entityDB, Transform);
    
    if (OpenFileForRead(&transformFile, path)) {

        u32 entityCount;
        if (!ParseIdentifier(&transformFile, NULL, "entityCount")) {
            success = false;
            goto error;
        }
        ParseUInt32Value(&transformFile, &entityCount);

        while (ParseIdentifier(&transformFile, NULL, "entityID")) {
            EntityID id;
            if (!ParseUInt64Value(&transformFile, &id)) {
                success = false;
                break;
            }

            const char *entityFilePath = SceneEntityPath(sceneDirectory, id);

            ComponentType type = ComponentType_Transform;
            ComponentValue component = {};
            ReadComponentValuesFromEntityFile(entityFilePath, entityDB, id, &type, 1, &component);

            //TransformComponent transform = component.Transform;
            //AddTransformComponent(entityDB, system, 1, &id);
            AddComponent_(entityDB, ComponentType_Transform, 1, &id, &component);
        }
    }

 error:
    CloseFile(&transformFile);
    return success;
}

void DeserializeScene(SceneID sceneID, SceneData *scene) {

    const char *sceneDirectory = SceneDirectoryPath(sceneID);
    const char *manifestPath = SceneManifestPath(sceneID, sceneDirectory);

    EntityDB *entityDB = &scene->entityDB;

    if (DeserializeSceneManifest(manifestPath, entityDB)) {

        const char *transformPath = SceneTransformTreePath(sceneID, sceneDirectory);

        // Deserialize transform tree
        DeserializeTransformTree(sceneDirectory, transformPath, entityDB);

        // Read Entity Files
        for (uint64 id = 1; id < entityDB->nextID; id++) {
            if (EntityIsDeleted(entityDB, id)) {
                continue;
            }

            const char *entityFilePath = SceneEntityPath(sceneDirectory, id);

            DeserializeEntityFile(entityFilePath, entityDB, id);

            Plat.DeallocateMem((char *)entityFilePath);
        }
    }

    Plat.DeallocateMem((char *)sceneDirectory);
    Plat.DeallocateMem((char *)manifestPath);
}

const char *CreateUniqueAssetName(SceneData *scene, const char *name, EntityID entityID) {
    ASSERT(scene->sceneID != Scene_None);
    ASSERT(entityID != EntityID_Invalid);
    ASSERT(name != NULL);

    EntityDB *entityDB = &scene->entityDB;
    char *entityName = GetEntityName(GetSystem(entityDB, Name), entityID);

    char entityIDString[8];
    stbsp_sprintf(entityIDString, "%" PRIu64, entityID);
    
    if (entityName != NULL) {
        return StringConcat(Game.frameMem, 7, GetSceneName(scene->sceneID), "__", name, "__", entityName, "_", entityIDString);
    }
    else {
        return StringConcat(Game.frameMem, 5, GetSceneName(scene->sceneID), "__", name, "__", entityIDString);
    }
}

const char *GetSceneAssetsPath(SceneData *scene) {
    ASSERT(scene->sceneID != Scene_None);
    char *scenePath = PathConcat(Game.frameMem, 3, Game.assets->searchPath, "scenes", GetSceneName(scene->sceneID));
    char *assetsPath = NULL;
    if (scenePath != NULL) {
        Plat.MakeDirectory(scenePath);
        assetsPath = PathConcat(Game.frameMem, 2, scenePath, "assets");

        if (assetsPath != NULL) {
            Plat.MakeDirectory(assetsPath);
        }
    }

    return assetsPath;
}

void GetUniqueAssetNameAndPath(SceneData *scene, EntityID entityID, const char *name, const char *ext, const char **assetName, const char **path) {
    ASSERT(name != NULL && assetName != NULL && path != NULL);
    *assetName = CreateUniqueAssetName(scene, name, entityID);
    const char *assetNameExt = AppendFileExtension(*assetName, ext);
    *path = PathConcat(Game.frameMem, 2, GetSceneAssetsPath(scene), assetNameExt);
}
