
#ifndef _ETC_ALL_SCENES_H
#define _ETC_ALL_SCENES_H

#include "scene_list.generated.h"

inline const char *GetSceneName(SceneID sceneID) {
    if ((uint32)sceneID < ENUM_VALUECOUNT(SceneID)) {
        return ENUM_STRINGVALUE(SceneID, sceneID);
    }

    return NULL;
}

inline SceneID GetSceneID(const char *name) {
    SceneID result = Scene_None;
    ENUM_FINDVALUE(SceneID, name, &result);

    return result;
}

inline SceneID GetSceneID(const char *name, uint32 nameLength) {
    SceneID result = Scene_None;
    ENUM_FINDVALUELENGTH(SceneID, name, nameLength, &result);

    return result;
}


struct SceneHandle {
    int32 index;
    uint32 generation;
};

inline SceneHandle MakeSceneHandle(int32 index, uint32 generation) {
    SceneHandle handle;

    handle.index = index;
    handle.generation = generation;

    return handle;
}

inline SceneHandle SceneHandleInvalid() {
    SceneHandle handle;

    handle.index = -1;
    handle.generation = 0;

    return handle;
}

inline b32 SceneHandleIsValid(SceneHandle handle) {
    return handle.index >= 0 && handle.generation > 0;
}

inline b32 operator==(SceneHandle a, SceneHandle b) {
    return a.index == b.index && a.generation == b.generation;
}

enum SceneState {
    SceneState_Free,
    SceneState_LoadingAssets,
    SceneState_LoadingEntities,
    SceneState_LoadingChildren,
    SceneState_Loaded,
    SceneState_MarkedForUnload,
};

struct SceneSettings {
    RenderMode renderMode;
};

struct SceneData {
    SceneID sceneID;
    uint32 generation;
    SceneHandle handle;

    SceneHandle owner;
    DynamicArray<SceneHandle> children;

    SceneState state;

    void *data;
    MemoryArenaList arena;

    EntityDB entityDB;
    SceneSettings settings;

    uint32 activeSceneIndex;

    b32 hasLoadedAssetList;
    EtcTxtParseState assetFile;
    DynamicArray<AssetType> assetListTypes;
    DynamicArray<const char *> assetListNames;
    DynamicArray<AssetHandle> assets;
    DynamicArray<b32> assetLoaded;
    uint32 assetLoadsStarted;
    uint32 assetLoadsCompleted;
    uint32 assetLoadedCheckOffset;

    CameraUniformBuffer cameraUniformBuffer;


    // @TODO: put pointers to the systems here!


#if ETC_DEBUG
    bool usingEditorCamera;
    EditorCamera editorCamera;
#endif
};

typedef void SceneInitFunc(EtcGameMemory *mem, SceneData *data);
typedef void SceneDeinitFunc(EtcGameMemory *mem, SceneData *data);
typedef void SceneUpdateFunc(EtcGameMemory *mem, SceneData *data);


void SceneNoneInit(EtcGameMemory *mem, SceneData *data);
void SceneNoneDeinit(EtcGameMemory *mem, SceneData *data);
void SceneNoneUpdate(EtcGameMemory *mem, SceneData *data);


// Scene Manager
#define MAX_ACTIVE_SCENES 4

struct ActiveSceneSet {
    b32 shouldBlend;
    int32 renderedIndex;
    
    uint32 sceneCount;
    SceneHandle scenes[MAX_ACTIVE_SCENES];
};

struct SceneManager {
    // @NOTE: this is only a valid handle when a scene is in its Update call
    SceneHandle activeScene;
    int32 activeSceneIndex;

    ActiveSceneSet active;
    ActiveSceneSet nextActive;

    ChunkedArray<SceneData> scenes;
    DynamicArray<uint32> scenesFreeList;
    DynamicArray<SceneHandle> loadingScenes;
};

void InitSceneManager(SceneManager *sceneManager);
void DeInitSceneManager(SceneManager *sceneManager);

void UpdateSceneManager(SceneManager *sceneManager);
void RunActiveScenesUpdate(SceneManager *sceneManager, EtcGameMemory *mem, b32 paused);

SceneData *GetScene(SceneHandle handle);
void *GetSceneData(SceneHandle handle);
SceneData *GetActiveScene();
void *GetActiveSceneData();

SceneData *GetSceneFromActiveSceneSet(u32 index);

SceneHandle GetRootScene(SceneHandle handle);


SceneHandle LoadScene(SceneID sceneID, SceneHandle owner = SceneHandleInvalid());
void UnloadScene(SceneHandle handle);
void UnloadAllScenes();


b32 IsSceneActive(SceneHandle handle);

b32 SetActiveScene(SceneHandle handle);
b32 SetActiveScenes(SceneHandle *handles, uint32 handleCount, b32 shouldBlendScenes = false, uint32 renderedIndex = 0);


// Serialization
void SerializeScene(SceneID sceneID, SceneData *scene);
void DeserializeScene(SceneID sceneID, SceneData *scene);

const char *CreateUniqueAssetName(SceneData *scene, const char *name, EntityID id);
const char *GetSceneAssetsPath(SceneData *scene);
void GetUniqueAssetNameAndPath(SceneData *scene, EntityID id, const char *name, const char *ext, const char **assetName, const char **path);

#endif

