
#ifndef _ETC_SCHEMA_H
#define _ETC_SCHEMA_H

#define FOREACH_SCHEMADATATYPE(OP, ...)                  \
    OP(void, 0, __VA_ARGS__)                             \
    OP(bool32, sizeof(uint32), __VA_ARGS__)              \
    OP(bool, sizeof(bool), __VA_ARGS__)                  \
    OP(char, sizeof(char), __VA_ARGS__)                  \
    OP(int8, sizeof(int8), __VA_ARGS__)                  \
    OP(int16, sizeof(int16), __VA_ARGS__)                \
    OP(int32, sizeof(int64), __VA_ARGS__)                \
    OP(int64, sizeof(int64), __VA_ARGS__)                \
    OP(uint8, sizeof(uint8), __VA_ARGS__)                \
    OP(uint16, sizeof(uint16), __VA_ARGS__)              \
    OP(uint32, sizeof(uint32), __VA_ARGS__)              \
    OP(uint64, sizeof(uint64), __VA_ARGS__)              \
    OP(real32, sizeof(real32), __VA_ARGS__)              \
    OP(vec2, sizeof(vec2), __VA_ARGS__)                  \
    OP(vec3, sizeof(vec3), __VA_ARGS__)                  \
    OP(vec4, sizeof(vec4), __VA_ARGS__)                  \
    OP(Color, sizeof(Color), __VA_ARGS__)                \
    OP(vec2i, sizeof(vec2i), __VA_ARGS__)                \
    OP(vec3i, sizeof(vec3i), __VA_ARGS__)                \
    OP(vec4i, sizeof(vec4i), __VA_ARGS__)                \
    OP(mat3, sizeof(mat3), __VA_ARGS__)                  \
    OP(mat4, sizeof(mat4), __VA_ARGS__)                  \
    OP(quaternion, sizeof(quaternion), __VA_ARGS__)      \
    OP(Schema, sizeof(Schema), __VA_ARGS__)              \
    OP(SchemaData, sizeof(SchemaData), __VA_ARGS__)      \
    OP(EntityID, sizeof(uint64), __VA_ARGS__)            \
    OP(DynamicArray, 0, __VA_ARGS__)                     \
    OP(ChunkedArray, 0, __VA_ARGS__)                     \
    OP(HashTable, 0, __VA_ARGS__)                        \


#define FOREACH_SCHEMADATATYPE_ALIAS(OP, ...)           \
    OP(b32, bool32, __VA_ARGS__)                        \


#define GENERATE_SCHEMADATATYPE(type, size, _) SchemaDataType_##type,

enum SchemaDataType : uint32 {
    SchemaDataType_Invalid,

    FOREACH_SCHEMADATATYPE(GENERATE_SCHEMADATATYPE, args)

    SchemaDataType_String,
    SchemaDataType_Struct,
    SchemaDataType_Enum,
    SchemaDataType_Bitfield,
    SchemaDataType_Bitarray,
    SchemaDataType_TaggedUnion,
    SchemaDataType_PersistentFunctionHandle,
    SchemaDataType_PersistentFunctionDataPair,
    SchemaDataType_AssetHandle,
};

enum SchemaType : uint32 {
    SchemaType_Struct,
    SchemaType_Enum,
    SchemaType_Bitfield,
    SchemaType_Bitarray,
    SchemaType_TaggedUnion,
    SchemaType_PersistentFunctionHandle,
    SchemaType_PersistentFunctionDataPair,
    SchemaType_AssetHandle,
};

struct SchemaTypeDescription {
    SchemaDataType dataType;
    const char *typeName;

    uint32 pointerLevels;
    b32 isConst;

    uint32 arrayLength;
    uint32 maxStringLength;
    uint32 sizeInBytes;

    DynamicArray<SchemaTypeDescription> templateTypes;

    b32 canSerialize;
};

struct SchemaField {
    const char *name;
    SchemaTypeDescription type;
    uint32 byteOffset;

    b32 editable;
    b32 serializable;
};

struct SchemaStruct {
    DynamicArray<SchemaField> fields;

    b32 isAsset;
};

struct SchemaEnum {
    uint32 count;
    const char **stringValues;
};

struct SchemaBitfield {
    uint32 count;
    const char **stringValues;
};

struct SchemaBitarray {
    uint32 count;
    const char **stringValues;
};

struct SchemaTaggedUnion {
    DynamicArray<SchemaField> fields;

    const char **typeTagStringValues;
};

struct SchemaPersistentFuncHandle {
    PersistentFuncType funcType;
};

struct SchemaPersistentFuncDataPair {
    SchemaField fields[2];
};

struct SchemaAssetHandle {
    const char *assetTypeName;
    const char *assetStructName;
};

struct Schema {
    const char *name;

    SchemaType type;
    union {
        SchemaStruct _struct;
        SchemaEnum _enum;
        SchemaBitfield _bitfield;
        SchemaBitfield _bitarray;
        SchemaTaggedUnion _tagged_union;
        SchemaPersistentFuncHandle _func_handle;
        SchemaPersistentFuncDataPair _func_data_pair;
        SchemaAssetHandle _assetHandle;
    };

    uint32 sizeInBytes;

    // Properties
    b32 editable;
    b32 serializable;
};

struct SchemaData {
    //Schema *schema;
    char *schemaName;
    void *data;
};

struct FuncDataPair {
    FuncHandle funcHandle;
    SchemaData data;
};

inline void DeallocateSchemaTypeDescription(SchemaTypeDescription *type) {
    for (int i = 0; i < type->templateTypes.count; i++) {
        DeallocateSchemaTypeDescription(&type->templateTypes[i]);
    }

    DeallocateDynamicArray(&type->templateTypes);
}

inline void DeallocateSchemaField(SchemaField *field) {
    DeallocateSchemaTypeDescription(&field->type);
}

inline void DeallocateSchema(Schema *schema) {
    switch (schema->type) {
        case SchemaType_Struct: {
            for (int j = 0; j < schema->_struct.fields.count; j++) {
                DeallocateSchemaField(&schema->_struct.fields[j]);
            }

            DeallocateDynamicArray(&schema->_struct.fields);
        } break;

        case SchemaType_TaggedUnion: {
            for (int j = 0; j < schema->_tagged_union.fields.count; j++) {
                DeallocateSchemaField(&schema->_tagged_union.fields[j]);
            }

            DeallocateDynamicArray(&schema->_tagged_union.fields);
        } break;

        default: {
        } break;
    }
}


void SerializeValue(EtcTxtWriteState *writer, Schema *schema, void *value, const char *fieldName = NULL);

// @NOTE: assumes `value` is cleared already
b32 DeserializeValue(EtcTxtParseState *parser, Schema *schema, void *value);


// Schema Comparison
#define SchemaIsOfType(schema, type) StringEquals(schema->name, #type)

STRUCT()
struct Real32EqualityCompare {
    real32 value;
};

STRUCT()
struct Real32RangeCompare {
    real32 min;
    real32 max;
};

STRUCT()
struct Real32GreaterCompare {
    real32 min;
};

// @GACK: this is necessary to include "schema.h" without the generated type
enum CompareData_TypeTag : uint32;

TAGGED_UNION(CompareData, {
    Real32EqualityCompare real32Equality;
    Real32RangeCompare real32Range;
    Real32GreaterCompare real32Greater;
});

STRUCT()
struct StructCompareData {
    char *fieldName;
    CompareData compareData;
};


b32 StructCompareField(Schema *schema, void *data, StructCompareData structCompareData);

#endif

