

bool PointInAABB(MinMaxAABB aabb, vec3 point, real32 epsilon) {
    bool result = point.x >= aabb.min.x - epsilon && point.x <= aabb.max.x + epsilon;
    result &= point.y >= aabb.min.y - epsilon && point.y <= aabb.max.y + epsilon;
    result &= point.z >= aabb.min.z - epsilon && point.z <= aabb.max.z + epsilon;
    return result;
}

r32 AABBVolume(MinMaxAABB aabb) {
    vec3 diff = Absf(aabb.max - aabb.min);
    return diff.x * diff.y * diff.z;
}

bool AABBsOverlap(MinMaxAABB a, MinMaxAABB b) {
    if (a.max.x < b.min.x || a.min.x > b.max.x) return 0;
    if (a.max.y < b.min.y || a.min.y > b.max.y) return 0;
    if (a.max.z < b.min.z || a.min.z > b.max.z) return 0;

    return 1;
}

bool AABBContainsAABB(MinMaxAABB a, MinMaxAABB b) {
    return a.max.x > b.max.x && a.max.y > b.max.y && a.max.z > b.max.z &&
        a.min.x < b.min.x && a.min.y < b.min.y && a.min.z < b.min.z;
}

MinMaxAABB AABBUnion(MinMaxAABB a, MinMaxAABB b) {
    MinMaxAABB result = {};

    for (int i = 0; i < 3; i++) {
        result.min.data[i] = Min(a.min.data[i], b.min.data[i]);
        result.max.data[i] = Max(a.max.data[i], b.max.data[i]);
    }
    
    return result;
}


bool PointInAABB(AABB shape, vec3 shapePos, vec3 point) {
    bool result = false;

    vec3 p = shapePos - point;


    if ((p.x > -shape.halfExtents.x && p.x < shape.halfExtents.x) &&
        (p.y > -shape.halfExtents.y && p.y < shape.halfExtents.y) &&
        (p.z > -shape.halfExtents.z && p.z < shape.halfExtents.z)) {
        result = true;
    }

    return result;
}

AABB CalculateAABB(vec3 *points, uint32 pointCount, Transform transform, uint32 stride = 0) {
    ASSERT(pointCount != 0);
    
    if (stride == 0) {
        stride = sizeof(vec3);
    }
    
    vec3 min = {INFINITY, INFINITY, INFINITY};
    vec3 max = {-INFINITY, -INFINITY, -INFINITY};

    for (int i = 0; i < pointCount; i++) {
        
        vec3 point = MultiplyPoint(transform, STRIDED_READ(vec3, points, stride, i));

        min = Min(min, point);
        max = Max(max, point);
    }

    AABB aabb;
    aabb.center = (min + max) * 0.5f;
    aabb.halfExtents = (max - min) * 0.5f;

    return aabb;
}

AABB CalculateAABB(vec3 *points, uint32 pointCount, uint32 stride = 0) {
    ASSERT(pointCount != 0);
    
    if (stride == 0) {
        stride = sizeof(vec3);
    }
    
    vec3 min = {INFINITY, INFINITY, INFINITY};
    vec3 max = {-INFINITY, -INFINITY, -INFINITY};

    for (int i = 0; i < pointCount; i++) {
        
        vec3 point = STRIDED_READ(vec3, points, stride, i);

        min = Min(min, point);
        max = Max(max, point);
    }

    AABB aabb;
    aabb.center = (min + max) * 0.5f;
    aabb.halfExtents = (max - min) * 0.5f;

    return aabb;
}

bool TestAABBAABB(AABB a, AABB b) {
    if (Absf(a.center.x - b.center.x) > (a.halfExtents.x + b.halfExtents.x)) { return false; }
    if (Absf(a.center.y - b.center.y) > (a.halfExtents.y + b.halfExtents.y)) { return false; }
    if (Absf(a.center.z - b.center.z) > (a.halfExtents.z + b.halfExtents.z)) { return false; }

    return true;
}

bool PointInOBB(OBB obb, vec3 obbPos, quaternion rot, vec3 point) {
    bool result = false;

    vec3 p = obbPos - point;
    
    vec3 dimensions = Rotate(rot * obb.rot, obb.dimensions);
    dimensions.x = Absf(dimensions.x);
    dimensions.y = Absf(dimensions.y);
    dimensions.z = Absf(dimensions.z);

    if ((p.x > -dimensions.x * 0.5f && p.x < dimensions.x * 0.5f) &&
        (p.y > -dimensions.y * 0.5f && p.y < dimensions.y * 0.5f) &&
        (p.z > -dimensions.z * 0.5f && p.z < dimensions.z * 0.5f)) {
        result = true;
    }

    return result;
}

bool PointInSphere(Sphere sphere, vec3 shapePos, vec3 point) {
    bool result = false;

    real32 radiusSq = sphere.radius * sphere.radius;
    if (DistanceSq(shapePos, point) < radiusSq) {
        result = true;
    }

    return result;
}

bool PointInColliderShape(Shape shape, vec3 pos, quaternion rot, vec3 point) {
    TIMED_BLOCK();
    switch (shape.type) {
        case Shape_sphere : {
            Sphere sphere;
            TaggedUnionGetValue(Shape, &shape, sphere, &sphere, sizeof(Sphere));
            return PointInSphere(sphere, pos + sphere.center, point);
        } break;

        case Shape_obb : {
            OBB obb;
            TaggedUnionGetValue(Shape, &shape, obb, &obb, sizeof(OBB));
            return PointInOBB(obb, pos + obb.center, rot, point);
        };

        case Shape_aabb : {
            AABB aabb;
            TaggedUnionGetValue(Shape, &shape, aabb, &aabb, sizeof(AABB));
            return PointInAABB(aabb, pos + aabb.center, point);
        };
    }

    return false;
}

            
// Grids

inline vec3i IndexTo3D(uint32 i, uint32 maxX, uint32 maxY, uint32 maxZ) {
    vec3i result;
    result.z = i / (maxX * maxY);
    i -= result.z * maxX * maxY;
    result.y = i / maxX;
    result.x = i % maxX;
    return result;
}

inline uint32 IndexTo1D(vec3i index3D, int32 width, int32 height) {
    return index3D.x + width * (index3D.y + height * index3D.z);
    //return index3D.x + (grid.width * index3D.y) + (index3D.z * grid.width * grid.height);
}

inline vec3 RelativeAABBPosition(vec3 pos, MinMaxAABB aabb) {
    return pos - aabb.min;
}

inline int32 PositionToGridCellIndex(vec3 pos, MinMaxAABB aabb, real32 cellDimensions, uint32 width, uint32 height, uint32 depth, r32 epsilon, vec3i *index3D) {
    // @PERF: we should just optimize this out because we should always have a valid grid
    // @NOTE: need position in worldspace for this test, 
    if (!PointInAABB(aabb, pos, epsilon)) {
        return -1;
    }

    vec3 extents = aabb.max - aabb.min;
    
    vec3 posRel = RelativeAABBPosition(pos, aabb);

    posRel.x = Clamp(posRel.x, 0.0f, extents.x);
    posRel.y = Clamp(posRel.y, 0.0f, extents.y);
    posRel.z = Clamp(posRel.z, 0.0f, extents.z);

    int32 x = (int32)floorf(posRel.x / cellDimensions);
    int32 y = (int32)floorf(posRel.y / cellDimensions);
    int32 z = (int32)floorf(posRel.z / cellDimensions);

    if (posRel.x == extents.x) {
        x = width - 1;
    }
    if (posRel.y == extents.y) {
        y = height - 1;
    }
    if (posRel.z == extents.z) {
        z = depth - 1;
    }

    ASSERT(x >= 0 && x < width);
    ASSERT(y >= 0 && y < height);
    ASSERT(z >= 0 && z < depth);

    vec3i index3D_ = {x, y, z};

    if (index3D != NULL) {
        *index3D = index3D_;
    }

    // @WTU: why are IndexTo1D and this computing the index differently????
    //return x + (width * y) + (z * width * height);
    return IndexTo1D(index3D_, width, height);
}


MinMaxAABB ComputeAABBForGridCell(MinMaxAABB aabb, uint32 cellDimension, vec3i index3D) {
    vec3 worldPt = V3(cellDimension * index3D.x,
                      cellDimension * index3D.y,
                      cellDimension * index3D.z);
    worldPt = worldPt + aabb.min;

    MinMaxAABB result;
    result.min = worldPt;
    result.max = worldPt + V3(cellDimension);
    return result;
}


uint32 RaycastGrid(Ray ray, MinMaxAABB aabb, u32 width, u32 height, u32 depth, r32 cellSize, ChunkedArray<u32> *hitCells) {
    TIMED_BLOCK();
    uint32 intersectCount = 0;

    vec3i index3D;
    int32 rayCellIndex = PositionToGridCellIndex(ray.origin, aabb, cellSize, width, height, depth, 0.0f, &index3D);

    if (rayCellIndex == -1) {
        return 0;
    }

    uint32 cellIndex = rayCellIndex;
    PushBack(hitCells, cellIndex);

    // how do we get the last cell in a direction?
    // we take the n planes of the bounding box, raycast against each of them, get the point
    r32 t;
    vec3 halfExtents = (aabb.max - aabb.min) * 0.5f;
    vec3 center = Lerp(aabb.max, aabb.min, 0.5f);
    RaycastCube(center, halfExtents, ray, &t);
    
    vec3 rayEndPoint = PointAt(ray, t);
    if (rayEndPoint.x > aabb.max.x) {
        rayEndPoint.x = aabb.max.x;
    }
    if (rayEndPoint.y > aabb.max.y) {
        rayEndPoint.y = aabb.max.y;
    }
    if (rayEndPoint.z > aabb.max.z) {
        rayEndPoint.z = aabb.max.z;
    }

    if (rayEndPoint.x < aabb.min.x) {
        rayEndPoint.x = aabb.min.x;
    }
    if (rayEndPoint.y < aabb.min.y) {
        rayEndPoint.y = aabb.min.y;
    }
    if (rayEndPoint.z < aabb.min.z) {
        rayEndPoint.z = aabb.min.z;
    }
    // DebugDrawPoint(rayEndPoint, CYAN, 2);

    // DebugDrawLine(rayEndPoint, ray.origin);

    r32 epsilon = 1.0e-4;
    vec3i endCellIndex3D;
    int32 endCellIndex = PositionToGridCellIndex(rayEndPoint, aabb, cellSize, width, height, depth, epsilon, &endCellIndex3D);

    if (endCellIndex >= 0) {
        //DebugDrawAABB(grid->cells[endCellIndex].aabb, GREEN);
    }
    else {
        endCellIndex = PositionToGridCellIndex(rayEndPoint, aabb, cellSize, width, height, depth, epsilon, &endCellIndex3D);
        ASSERT(false);
    }

    vec3i stepDirection = {};
    if (ray.origin.x < rayEndPoint.x) { stepDirection.x = 1; }
    else if (ray.origin.x > rayEndPoint.x) { stepDirection.x = -1; }

    if (ray.origin.y < rayEndPoint.y) { stepDirection.y = 1; }
    else if (ray.origin.y > rayEndPoint.y) { stepDirection.y = -1; }

    if (ray.origin.z < rayEndPoint.z) { stepDirection.z = 1; }
    else if (ray.origin.z > rayEndPoint.z) { stepDirection.z = -1; }

    r32 minX = cellSize * floorf(ray.origin.x / cellSize);
    r32 maxX = minX + cellSize;
    r32 tx = SafeRatio(((ray.origin.x > rayEndPoint.x) ? (ray.origin.x - minX) : (maxX - ray.origin.x)), Absf(rayEndPoint.x - ray.origin.x), INFINITY);

    r32 minY = cellSize * floorf(ray.origin.y / cellSize);
    r32 maxY = minY + cellSize;
    r32 ty = SafeRatio(((ray.origin.y > rayEndPoint.y) ? (ray.origin.y - minY) : (maxY - ray.origin.y)), Absf(rayEndPoint.y - ray.origin.y), INFINITY);

    r32 minZ = cellSize * floorf(ray.origin.z / cellSize);
    r32 maxZ = minZ + cellSize;
    r32 tz = SafeRatio(((ray.origin.z > rayEndPoint.z) ? (ray.origin.z - minZ) : (maxZ - ray.origin.z)), Absf(rayEndPoint.z - ray.origin.z), INFINITY);

    // @NOTE: "for unifrom-size cells the distance moved along the ray is constant from one vertical boundary to the next, and thus can be pre-computed"
    r32 deltaX = SafeRatio(cellSize, Absf(rayEndPoint.x - ray.origin.x), INFINITY);
    r32 deltaY = SafeRatio(cellSize, Absf(rayEndPoint.y - ray.origin.y), INFINITY);
    r32 deltaZ = SafeRatio(cellSize, Absf(rayEndPoint.z - ray.origin.z), INFINITY);

    

    while (true) {
        if (tx <= ty && tx <= tz) {
            if (index3D.x == endCellIndex3D.x) { break; }
            tx += deltaX;
            index3D.x += stepDirection.x;
        }
        else if (ty <= tx && ty <= tz) {
            if (index3D.y == endCellIndex3D.y) { break; }
            ty += deltaY;
            index3D.y += stepDirection.y;
        }
        else if (tz <= tx && tz <= ty) {
            if (index3D.z == endCellIndex3D.z) { break; }
            tz += deltaZ;
            index3D.z += stepDirection.z;
        }

        if (index3D == endCellIndex3D) {
            break;
        }

        uint32 i = IndexTo1D(index3D, width, height);

        PushBack(hitCells, i);
    }

    return intersectCount;
}




#if ETC_DEBUG
void DebugDrawShape(Shape shape, vec3 position, quaternion rotation, vec3 scale, vec4 color) {
    switch (shape.type) {
        case Shape_sphere: {
            DebugDrawMeshWireframe(DebugMem->sphere, position + Rotate(rotation, shape._data.sphere.center), rotation, scale * shape._data.sphere.radius, color);
        } break;
        case Shape_obb: {
            vec3 newScale = Hadamard(scale, (0.5f * shape._data.obb.dimensions));
            DebugDrawMeshWireframe(DebugMem->cube, position + Rotate(rotation * shape._data.obb.rot, shape._data.obb.center), rotation * shape._data.obb.rot, newScale, color);
        } break;

        case Shape_aabb: {
            vec3 newScale = Hadamard(scale, shape._data.aabb.halfExtents);
            DebugDrawMeshWireframe(DebugMem->cube, position + shape._data.aabb.center, IdentityQuaternion(), newScale, color);
        } break;
    }
}

void DebugDrawShape(Shape shape, Transform trans, vec4 color) {
    DebugDrawShape(shape, trans.position, trans.rotation, trans.scale, color);
}
#endif


// ShapeComponent
void AddShapeComponent(EntityDB *entityDB, ShapeSystem *system, uint32 count, EntityID *entityIDs) {
    if (system->entityMap.capacity == 0) {
        AllocateHashTable(&system->entityMap, 1024);
    }

    for (int i = 0; i < count; i++) {
        AddEntityToSystem(system, entityIDs[i]);
        HashTableInsert(&system->entityMap, entityIDs[i], system->entityIDs.count - 1);
        
        ShapeComponent *shape = PushBackPtr(&system->shapes);
        Plat.ClearMem(shape, sizeof(ShapeComponent));
    }
}

void RemoveShapeComponent(EntityDB *entityDB, ShapeSystem *system, EntityID entityID) {
    uint32 index;
    if (HashTableGet(&system->entityMap, entityID, &index)) {
        
        uint32 endIndex = system->shapes.count - 1;

        RemoveAtIndexBySwap(&system->shapes, index);

        HashTableInsert(&system->entityMap, system->entityIDs[endIndex], index);
        HashTableRemove(&system->entityMap, entityID);

        RemoveAtIndexBySwap(&system->entityIDs, index);
    }
}

b32 GetShapeComponentValue(EntityDB *entityDB, ShapeSystem *system, EntityID entityID, ShapeComponent *component) {
    b32 result = false;

    uint32 index;
    if (HashTableGet(&system->entityMap, entityID, &index)) {
        *component = system->shapes[index];
        result = true;
    }

    return result;
}

void SetShapeComponentValue(EntityDB *entityDB, ShapeSystem *system, EntityID entityID, ShapeComponent *component) {
    uint32 index;
    if (HashTableGet(&system->entityMap, entityID, &index)) {
        system->shapes[index] = *component;
    }
}

void TickShapeSystem(EtcGameMemory *mem, EntityDB *entityDB, ShapeSystem *system) {
}

uint32 RaycastShapeSystem(EntityDB *entityDB, ShapeSystem *system, Ray ray, RaycastSceneResult *closestResult, ChunkedArray<RaycastSceneResult> *results) {
    TIMED_BLOCK();
    b32 intersectCount = 0;

    RaycastSceneResult closest = {};
    closest.t = INFINITY;

    for (int i = 0; i < system->entityIDs.count; i++) {
        EntityID id = system->entityIDs[i];

        vec3 pos = GetWorldPosition(GetSystem(entityDB, Transform), id);
        quaternion rot = GetWorldRotation(GetSystem(entityDB, Transform), id);
        // @TODO: scale matters, do eet too!

        ShapeComponent shape = system->shapes[i];

        b32 intersects = false;
        real32 t;
        
        switch (shape.shape.type) {
            case Shape_sphere : {
                Sphere sphere = shape.shape._data.sphere;
                intersects = RaycastSphere(pos + sphere.center, sphere.radius, ray, &t);
            } break;
            case Shape_obb : {
                OBB obb = shape.shape._data.obb;
                vec3 rotDimensions = Rotate(rot * obb.rot, obb.dimensions * 0.5f);
                rotDimensions.x = Absf(rotDimensions.x);
                rotDimensions.y = Absf(rotDimensions.y);
                rotDimensions.z = Absf(rotDimensions.z);

                intersects = RaycastCube(pos + Rotate(rot * obb.rot, obb.center), rotDimensions, ray, &t);
            } break;

            case Shape_aabb: {
                AABB aabb = shape.shape._data.aabb;
                intersects = RaycastCube(pos + Rotate(rot, aabb.center), aabb.halfExtents, ray, &t);
            } break;
        }


        if (intersects) {
            RaycastSceneResult raycastResult = {};
            raycastResult.entityID = id;
            raycastResult.componentType = ComponentType_Shape;
            raycastResult.ray = ray;
            raycastResult.t = t;
            raycastResult.endPoint = PointAt(ray, t);

            if (t < closest.t) {
                closest = raycastResult;
            }

            if (results != NULL) {
                PushBack(results, raycastResult);
            }

            intersectCount++;
        }
    }

    if (intersectCount > 0 && closestResult != NULL) {
        *closestResult = closest;
    }

    return intersectCount;
}


// Trigger Volumes
void AddTriggerVolumeComponent(EntityDB *entityDB, TriggerVolumeSystem *system, uint32 count, EntityID *entityIDs) {
    if (system->entityMap.capacity == 0) {
        AllocateHashTable(&system->entityMap, 1024);
    }

    for (int i = 0; i < count; i++) {
        AddEntityToSystem(system, entityIDs[i]);
        HashTableInsert(&system->entityMap, entityIDs[i], system->entityIDs.count - 1);
     
        TriggerVolumeComponent *triggerVolume = PushBackPtr(&system->triggers);
        Plat.ClearMem(triggerVolume, sizeof(TriggerVolumeComponent));

        TriggerVolumeState state = {};
        PushBack(&system->states, state);
    }
}

void RemoveTriggerVolumeComponent(EntityDB *entityDB, TriggerVolumeSystem *system, EntityID entityID) {
    uint32 index;
    if (HashTableGet(&system->entityMap, entityID, &index)) {
     
        uint32 endIndex = system->triggers.count - 1;

        RemoveAtIndexBySwap(&system->triggers, index);

        HashTableInsert(&system->entityMap, system->entityIDs[endIndex], index);
        HashTableRemove(&system->entityMap, entityID);

        RemoveAtIndexBySwap(&system->entityIDs, index);
    }
}

b32 GetTriggerVolumeComponentValue(EntityDB *entityDB, TriggerVolumeSystem *system, EntityID entityID, TriggerVolumeComponent *component) {
    b32 result = false;

    uint32 index;
    if (HashTableGet(&system->entityMap, entityID, &index)) {
        *component = system->triggers[index];
        result = true;
    }

    return result;
}

void SetTriggerVolumeComponentValue(EntityDB *entityDB, TriggerVolumeSystem *system, EntityID entityID, TriggerVolumeComponent *component) {
    uint32 index;
    if (HashTableGet(&system->entityMap, entityID, &index)) {
        system->triggers[index] = *component;
    }
}

void TickTriggerVolumeSystem(EtcGameMemory *mem, EntityDB *entityDB, TriggerVolumeSystem *system) {
    TransformSystem *transSys = GetSystem(entityDB, Transform);
    for (int i = 0; i < system->entityIDs.count; i++) {
        EntityID triggerID = system->entityIDs[i];
        TriggerVolumeComponent trigger = system->triggers[i];
        TriggerVolumeState *state = &system->states[i];

        vec3 triggerPos = GetWorldPosition(transSys, triggerID);
        quaternion triggerRot = GetWorldRotation(transSys, triggerID);

        for (int j = 0; j < trigger.listeners.count; j++) {
            TriggerVolumeListener *listener = &trigger.listeners[j];

            if (listener->filter != EntityID_Invalid) {
                EntityID id = listener->filter;

                vec3 pos = GetWorldPosition(transSys, id);
                
                uint32 index;
                bool contains = Contains(&state->entitiesInside, id, &index);

                if (PointInColliderShape(trigger.shape, triggerPos, triggerRot, pos)) {
                    if (!contains) {
                        PushBack(&state->entitiesInside, id);
                    }

                    if (FuncHandleIsValid(listener->enterFunc)) {
                        CallFunc(listener->enterFunc, entityDB, triggerID, id, listener->data);
                    }
                }
                else if (contains) {
                    RemoveAtIndex(&state->entitiesInside, index);
                        
                    if (FuncHandleIsValid(listener->exitFunc)) {
                        CallFunc(listener->exitFunc, entityDB, triggerID, id, listener->data);
                    }
                }

                continue;
            }

            EtcSystem *filterSystem = entityDB->systems[ComponentType_Transform];
            
            if (listener->componentTypeFilter != ComponentType_Invalid) {
                filterSystem = entityDB->systems[listener->componentTypeFilter];    
            }
            
            DynamicArray<EntityID> *entityIDs = &filterSystem->entityIDs;

            TagComponent tag;
            //bool filteringOnTags = listener->
            for (int k = 0; k < entityIDs->count; k++) {
                EntityID id = (*entityIDs)[k];

                vec3 pos = GetWorldPosition(transSys, id);

                uint32 index;
                bool contains = Contains(&state->entitiesInside, id, &index);

                if (PointInColliderShape(trigger.shape, triggerPos, triggerRot, pos)) {
                    if (!contains) {
                        if (FuncHandleIsValid(listener->enterFunc)) {
                            CallFunc(listener->enterFunc, entityDB, triggerID, id, listener->data);
                        }
                        PushBack(&state->entitiesInside, id);
                    }
                }
                else {
                    if (contains) {
                        RemoveAtIndex(&state->entitiesInside, index);
                        
                        if (FuncHandleIsValid(listener->exitFunc)) {
                            CallFunc(listener->exitFunc, entityDB, triggerID, id, listener->data);
                        }
                    }
                }
            }
        }
    }
}

