
void AddTransformComponent(EntityDB *entityDB, TransformSystem *system, uint32 count, EntityID *entityIDs) {
    if (system->entityMap.capacity == 0) {
        AllocateHashTable(&system->entityMap, 4096);
    }

    for (int i = 0; i < count; i++) {
        if (!HashTableContains(&system->entityMap, entityIDs[i])) {
            int32 index = system->entityIDs.count;
            HashTableInsert(&system->entityMap, entityIDs[i], index, true);

            //AddEntityToSystem(system, entityIDs[i]);
            AddEntityToSystem(system, entityIDs[i]);
            PushBack(&system->positions, V3(0, 0, 0));
            PushBack(&system->rotations, IdentityQuaternion());
            PushBack(&system->scales, V3(1, 1, 1));

            PushBack(&system->worldPositions, V3(0, 0, 0));
            PushBack(&system->worldRotations, IdentityQuaternion());
            PushBack(&system->worldScales, V3(1, 1, 1));

            PushBack(&system->parentIndices, -1);
            PushBack(&system->firstChildIndices, -1);
            PushBack(&system->prevSiblingIndices, -1);
            PushBack(&system->nextSiblingIndices, -1);

            PushBack(&system->worldTransforms, Identity4());
        }
    }
}

void RemoveTransformComponent(EntityDB *entityDB, TransformSystem *system, EntityID entityID) {
    int32 index;
    if (HashTableGet(&system->entityMap, entityID, &index)) {

        int32 parentIndex = system->parentIndices[index];
        int32 prevSiblingIndex = system->prevSiblingIndices[index];
        int32 nextSiblingIndex = system->nextSiblingIndices[index];
        int32 firstChildIndex = system->firstChildIndices[index];

        if (parentIndex >= 0 && system->firstChildIndices[parentIndex] == index) {
            if (firstChildIndex >= 0) {
                system->firstChildIndices[parentIndex] = firstChildIndex;
            }
            else {
                system->firstChildIndices[parentIndex] = nextSiblingIndex;

                if (nextSiblingIndex >= 0) {
                    system->prevSiblingIndices[nextSiblingIndex] = -1;
                }
            }
        }
        else if (firstChildIndex >= 0) {
            if (prevSiblingIndex >= 0) {
                system->nextSiblingIndices[prevSiblingIndex] = firstChildIndex;
            }

            system->prevSiblingIndices[firstChildIndex] = prevSiblingIndex;
        }
        else {
            if (prevSiblingIndex >= 0) {
                system->nextSiblingIndices[prevSiblingIndex] = nextSiblingIndex;
            }

            if (nextSiblingIndex >= 0) {
                system->prevSiblingIndices[nextSiblingIndex] = prevSiblingIndex;
            }
        }


        // Reparent children
        int32 childIndex = firstChildIndex;
        while (childIndex >= 0) {
            system->parentIndices[childIndex] = parentIndex;

            int32 nextReparentedSiblingIndex = system->nextSiblingIndices[childIndex];

            if (parentIndex < 0) {
                system->prevSiblingIndices[childIndex] = -1;
                system->nextSiblingIndices[childIndex] = -1;
            }
            // Set last child's nextSibling to removed transform's nextSibling
            else if (nextReparentedSiblingIndex < 0) {
                system->nextSiblingIndices[childIndex] = nextSiblingIndex;

                if (nextSiblingIndex >= 0) {
                    system->prevSiblingIndices[nextSiblingIndex] = childIndex;
                }
            }

            childIndex = nextReparentedSiblingIndex;
        }


        RemoveAtIndexBySwap(&system->entityIDs, index);

        RemoveAtIndexBySwap(&system->positions, index);
        RemoveAtIndexBySwap(&system->rotations, index);
        RemoveAtIndexBySwap(&system->scales, index);

        RemoveAtIndexBySwap(&system->worldPositions, index);
        RemoveAtIndexBySwap(&system->worldRotations, index);
        RemoveAtIndexBySwap(&system->worldScales, index);

        RemoveAtIndexBySwap(&system->parentIndices, index);
        RemoveAtIndexBySwap(&system->firstChildIndices, index);
        RemoveAtIndexBySwap(&system->prevSiblingIndices, index);
        RemoveAtIndexBySwap(&system->nextSiblingIndices, index);
        RemoveAtIndexBySwap(&system->worldTransforms, index);

        HashTableRemove(&system->entityMap, entityID);

        // Update refererences to the entity that got swapped into index
        if (index < system->entityIDs.count) {

            int32 parentIndex = system->parentIndices[index];
            int32 prevSiblingIndex = system->prevSiblingIndices[index];
            int32 nextSiblingIndex = system->nextSiblingIndices[index];
            int32 firstChildIndex = system->firstChildIndices[index];

            if (prevSiblingIndex >= 0) {
                system->nextSiblingIndices[prevSiblingIndex] = index;
            }

            if (nextSiblingIndex >= 0) {
                system->prevSiblingIndices[nextSiblingIndex] = index;
            }

            if (parentIndex >= 0 &&
                system->firstChildIndices[parentIndex] == system->entityIDs.count) {
                system->firstChildIndices[parentIndex] = index;
            }

            int32 childIndex = firstChildIndex;
            while (childIndex >= 0) {
                system->parentIndices[childIndex] = index;
                childIndex = system->nextSiblingIndices[childIndex];
            }

            HashTableInsert(&system->entityMap, system->entityIDs[index], index);
        }
    }
}

EntityID LastChildOf(TransformSystem *system, EntityID entityID) {
    EntityID result = EntityID_Invalid;

    int32 index;
    if (HashTableGet(&system->entityMap, entityID, &index)) {
        
        int32 childIndex = system->firstChildIndices[index];
        while (childIndex >= 0) {
            result = system->entityIDs[childIndex];
            childIndex = system->nextSiblingIndices[childIndex];
        }
    }

    return result;
}

void UpdateWorldTransform(TransformSystem *system, int32 rootIndex) {
    // It is assumed that the direct parent (if there is one) already
    // has a valid set of cached world transform values. To compute
    // rootIndex's worldTransform we only have to look at the direct
    // parent
    vec3 relativePosition = system->positions[rootIndex];
    quaternion relativeRotation = system->rotations[rootIndex];
    vec3 relativeScale = system->scales[rootIndex];

    mat4 relativeTransform = TRS(relativePosition, relativeRotation, relativeScale);

    int32 parentIndex = system->parentIndices[rootIndex];
    if (parentIndex >= 0) {
        vec3 parentWorldPosition = system->worldPositions[parentIndex];
        quaternion parentWorldRotation = system->worldRotations[parentIndex];
        vec3 parentWorldScale = system->worldScales[parentIndex];

        system->worldTransforms[rootIndex] = system->worldTransforms[parentIndex] * relativeTransform;

        system->worldPositions[rootIndex] = PointToReferenceSpace(parentWorldPosition, parentWorldRotation, parentWorldScale, relativePosition);
        system->worldRotations[rootIndex] = parentWorldRotation * relativeRotation;
        system->worldScales[rootIndex] = Hadamard(parentWorldScale, relativeScale);
    }
    else {
        system->worldTransforms[rootIndex] = relativeTransform;

        system->worldPositions[rootIndex] = relativePosition;
        system->worldRotations[rootIndex] = relativeRotation;
        system->worldScales[rootIndex] = relativeScale;
    }

    EntityIterator iter(system, system->entityIDs[rootIndex], true);
    EntityID childID;
    while (GetNextEntity(&iter, &childID)) {
        int32 index = iter.currentIndex;
        int32 parentIndex = system->parentIndices[index];

        vec3 relativePosition = system->positions[index];
        quaternion relativeRotation = system->rotations[index];
        vec3 relativeScale = system->scales[index];

        mat4 relativeTransform = TRS(relativePosition, relativeRotation, relativeScale);

        vec3 parentWorldPosition = system->worldPositions[parentIndex];
        quaternion parentWorldRotation = system->worldRotations[parentIndex];
        vec3 parentWorldScale = system->worldScales[parentIndex];
        
        system->worldTransforms[index] = system->worldTransforms[parentIndex] * relativeTransform;

        system->worldPositions[index] = PointToReferenceSpace(parentWorldPosition, parentWorldRotation, parentWorldScale, relativePosition);
        system->worldRotations[index] = parentWorldRotation * relativeRotation;
        system->worldScales[index] = Hadamard(parentWorldScale, relativeScale);
    }
}

mat4 GetWorldTransformMatrix(TransformSystem *system, EntityID entityID) {
    mat4 result;
    int32 index;

    if (HashTableGet(&system->entityMap, entityID, &index)) {
        result = system->worldTransforms[index];
    }
    else {
        result = Identity4();
    }
         
    return result;
}

Transform GetWorldTransform(TransformSystem *system, EntityID entityID) {
    Transform result;
    int32 index;

    if (HashTableGet(&system->entityMap, entityID, &index)) {
        result.position = system->worldPositions[index];
        result.rotation = system->worldRotations[index];
        result.scale = system->worldScales[index];
    }
    else {
        result = IdentityTransform();
    }
         
    return result;
}

Transform GetLocalTransform(TransformSystem *system, EntityID entityID) {
    Transform result;
    int32 index;

    if (HashTableGet(&system->entityMap, entityID, &index)) {
        result.position = system->positions[index];
        result.rotation = system->rotations[index];
        result.scale = system->scales[index];
    }
    else {
        result = IdentityTransform();
    }
         
    return result;
}


vec3 GetWorldPosition(TransformSystem *system, EntityID entityID) {
    vec3 result = {};

    int32 index;
    if (HashTableGet(&system->entityMap, entityID, &index)) {
        result = system->worldPositions[index];
    }

    return result;
}

vec3 GetLocalPosition(TransformSystem *system, EntityID entityID) {
    vec3 result = {};

    int32 index;
    if (HashTableGet(&system->entityMap, entityID, &index)) {
        result = system->positions[index];
    }

    return result;
}

quaternion GetLocalRotation(TransformSystem *system, EntityID entityID) {
    quaternion result = IdentityQuaternion();

    int32 index;
    if (HashTableGet(&system->entityMap, entityID, &index)) {
        result = system->rotations[index];
    }

    return result;
}

quaternion GetWorldRotation(TransformSystem *system, EntityID entityID) {
    quaternion result = IdentityQuaternion();

    int32 index;
    if (HashTableGet(&system->entityMap, entityID, &index)) {
        result = system->worldRotations[index];
    }

    return result;
}

vec3 GetLocalScale(TransformSystem *system, EntityID entityID) {
    vec3 result = V3(1, 1, 1);

    int32 index;
    if (HashTableGet(&system->entityMap, entityID, &index)) {
        result = system->scales[index];
    }

    return result;
}

vec3 GetWorldScale(TransformSystem *system, EntityID entityID) {
    vec3 result = V3(1, 1, 1);

    int32 index;
    if (HashTableGet(&system->entityMap, entityID, &index)) {
        result = system->worldScales[index];
    }

    return result;
}

vec3 GetForward(TransformSystem *system, EntityID entityID) {
    vec3 result = FORWARD;

    int32 index;
    if (HashTableGet(&system->entityMap, entityID, &index)) {
        result = Rotate(system->worldRotations[index], result);
    }
    return result;
}

void SetLocalPosition(TransformSystem *system, EntityID entityID, vec3 position) {
    int32 index;
    if (HashTableGet(&system->entityMap, entityID, &index)) {
        system->positions[index] = position;

        UpdateWorldTransform(system, index);
    }
    else {
        ASSERT(false);
    }
}

void SetWorldPosition(TransformSystem *system, EntityID entityID, vec3 position) {
    int32 index;
    if (HashTableGet(&system->entityMap, entityID, &index)) {
        int32 parentIndex = system->parentIndices[index];

        if (parentIndex >= 0) {
            Transform parentWorldTransform = {system->worldPositions[parentIndex],
                                              system->worldRotations[parentIndex],
                                              system->worldScales[parentIndex]};
            Transform worldTransform = {position,
                                        system->worldRotations[index],
                                        system->worldScales[index]};

            Transform localTransform = RelativeTransform(parentWorldTransform, worldTransform);

            system->positions[index] = localTransform.position;
        }
        else {
            system->positions[index] = position;
        }

        UpdateWorldTransform(system, index);
    }
    else {
        ASSERT(false);
    }
}

EntityID GetParentID(TransformSystem *system, EntityID entityID) {
    int32 index;
    if (HashTableGet(&system->entityMap, entityID, &index)) {
        s32 parentIndex = system->parentIndices[index];
        if (parentIndex >= 0) {
            return system->entityIDs[parentIndex];    
        }
        return EntityID_Invalid;
    }

    return EntityID_Invalid;
}

void SetLocalRotation(TransformSystem *system, EntityID entityID, quaternion rotation) {
    int32 index;
    if (HashTableGet(&system->entityMap, entityID, &index)) {
        system->rotations[index] = rotation;

        UpdateWorldTransform(system, index);
    }
    else {
        ASSERT(false);
    }
}

void SetWorldRotation(TransformSystem *system, EntityID entityID, quaternion rotation) {
    int32 index;
    if (HashTableGet(&system->entityMap, entityID, &index)) {
        int32 parentIndex = system->parentIndices[index];

        if (parentIndex >= 0) {
            quaternion parentWorldRotation = system->worldRotations[parentIndex];
            quaternion localRotation = RelativeQuaternion(rotation, parentWorldRotation);

            system->rotations[index] = localRotation;
        }
        else {
            system->rotations[index] = rotation;
        }

        UpdateWorldTransform(system, index);
    }
    else {
        ASSERT(false);
    }
}

void SetLocalScale(TransformSystem *system, EntityID entityID, vec3 scale) {
    int32 index;
    if (HashTableGet(&system->entityMap, entityID, &index)) {
        system->scales[index] = scale;

        UpdateWorldTransform(system, index);
    }
    else {
        ASSERT(false);
    }
}

void SetWorldScale(TransformSystem *system, EntityID entityID, vec3 scale) {
    int32 index;
    if (HashTableGet(&system->entityMap, entityID, &index)) {
        int32 parentIndex = system->parentIndices[index];

        if (parentIndex >= 0) {
            Transform parentWorldTransform = {system->worldPositions[parentIndex],
                                              system->worldRotations[parentIndex],
                                              system->worldScales[parentIndex]};
            Transform worldTransform = {system->worldPositions[index],
                                        system->worldRotations[index],
                                        scale};

            Transform localTransform = RelativeTransform(parentWorldTransform, worldTransform);

            system->scales[index] = localTransform.scale;
        }
        else {
            system->scales[index] = scale;
        }

        UpdateWorldTransform(system, index);
    }
    else {
        ASSERT(false);
    }
}

void SetLocalTransform(TransformSystem *system, EntityID entityID, vec3 position, quaternion rotation, vec3 scale) {
    int32 index;
    if (HashTableGet(&system->entityMap, entityID, &index)) {
        system->positions[index] = position;
        system->rotations[index] = rotation;
        system->scales[index] = scale;

        UpdateWorldTransform(system, index);
    }
    else {
        ASSERT(false);
    }
}

void SetLocalTransform(TransformSystem *system, EntityID entityID, Transform localTransform) {
    SetLocalTransform(system, entityID, localTransform.position, localTransform.rotation, localTransform.scale);
}

void SetWorldTransform(TransformSystem *system, EntityID entityID, vec3 position, quaternion rotation, vec3 scale) {
    int32 index;
    if (HashTableGet(&system->entityMap, entityID, &index)) {
        int32 parentIndex = system->parentIndices[index];

        if (parentIndex >= 0) {
            Transform parentWorldTransform = {system->worldPositions[parentIndex],
                                              system->worldRotations[parentIndex],
                                              system->worldScales[parentIndex]};
            Transform worldTransform = {position, rotation, scale};

            Transform localTransform = RelativeTransform(parentWorldTransform, worldTransform);

            system->positions[index] = localTransform.position;
            system->rotations[index] = localTransform.rotation;
            system->scales[index] = localTransform.scale;
        }
        else {
            system->positions[index] = position;
            system->rotations[index] = rotation;
            system->scales[index] = scale;
        }

        UpdateWorldTransform(system, index);
    }
}

void SetWorldTransform(TransformSystem *system, EntityID entityID, Transform worldTransform) {
    SetWorldTransform(system, entityID, worldTransform.position, worldTransform.rotation, worldTransform.scale);
}

b32 GetTransformComponentValue(EntityDB *entityDB, TransformSystem *system, EntityID entityID, TransformComponent *component) {
    b32 result = false;

    int32 index;
    if (HashTableGet(&system->entityMap, entityID, &index)) {

        component->position = system->positions[index];
        component->rotation = system->rotations[index];
        component->scale = system->scales[index];

        int32 parentIndex = system->parentIndices[index];
        int32 firstChildIndex = system->firstChildIndices[index];
        int32 prevSiblingIndex = system->prevSiblingIndices[index];
        int32 nextSiblingIndex = system->nextSiblingIndices[index];

        component->parent = (parentIndex >= 0) ? system->entityIDs[parentIndex] : EntityID_Invalid;
        component->firstChild = (firstChildIndex >= 0) ? system->entityIDs[firstChildIndex] : EntityID_Invalid;
        component->prevSibling = (prevSiblingIndex >= 0) ? system->entityIDs[prevSiblingIndex] : EntityID_Invalid;
        component->nextSibling = (nextSiblingIndex >= 0) ? system->entityIDs[nextSiblingIndex] : EntityID_Invalid;
        
        result = true;
    }

    return result;
}

void SetTransformComponentValue(EntityDB *entityDB, TransformSystem *system, EntityID entityID, TransformComponent *component) {
    ASSERT(component->parent != entityID);

    int32 index;
    if (HashTableGet(&system->entityMap, entityID, &index)) {
        TransformComponent prevTransformComponent = {};
        GetTransformComponentValue(entityDB, system, entityID, &prevTransformComponent);

        if (prevTransformComponent.parent != component->parent ||
            prevTransformComponent.nextSibling != component->nextSibling) {

            // Clean up hole in previous spot in the tree
            int32 prevParentIndex = system->parentIndices[index];
            int32 prevPrevSiblingIndex = system->prevSiblingIndices[index];
            int32 prevNextSiblingIndex = system->nextSiblingIndices[index];

            if (prevPrevSiblingIndex >= 0) {
                system->nextSiblingIndices[prevPrevSiblingIndex] = prevNextSiblingIndex;
            }

            if (prevNextSiblingIndex >= 0) {
                system->prevSiblingIndices[prevNextSiblingIndex] = prevPrevSiblingIndex;
            }

            if (prevParentIndex >= 0 && system->firstChildIndices[prevParentIndex] == index) {
                ASSERT(prevPrevSiblingIndex < 0);
                system->firstChildIndices[prevParentIndex] = prevNextSiblingIndex;
            }


            int32 targetParentIndex = -1;
            HashTableGet(&system->entityMap, component->parent, &targetParentIndex);

            if (targetParentIndex >= 0) {
                EntityID targetNextSibling = component->nextSibling;
                int32 targetNextSiblingIndex = -1;

                int32 prevChildIndex = -1;
                int32 currentChildIndex = system->firstChildIndices[targetParentIndex];
                while (currentChildIndex >= 0) {
                    if (system->entityIDs[currentChildIndex] == targetNextSibling) {
                        targetNextSiblingIndex = currentChildIndex;
                        break;
                    }

                    prevChildIndex = currentChildIndex;
                    currentChildIndex = system->nextSiblingIndices[currentChildIndex];
                }

                if (prevChildIndex >= 0) {
                    // Insert between prevChildIndex and currentChildIndex
                    if (currentChildIndex >= 0) {
                        system->prevSiblingIndices[currentChildIndex] = index;
                    }
                    system->nextSiblingIndices[index] = currentChildIndex;

                    system->nextSiblingIndices[prevChildIndex] = index;
                    system->prevSiblingIndices[index] = prevChildIndex;
                }
                else {
                    // Insert as first child (before currentChildIndex)
                    ASSERT(system->firstChildIndices[targetParentIndex] == currentChildIndex);

                    if (currentChildIndex >= 0) {
                        system->prevSiblingIndices[currentChildIndex] = index;
                    }

                    system->nextSiblingIndices[index] = currentChildIndex;
                    system->prevSiblingIndices[index] = -1;
                    system->firstChildIndices[targetParentIndex] = index;
                }

                system->parentIndices[index] = targetParentIndex;
            }
            else {
                // Unparent case
                system->parentIndices[index] = -1;
                system->prevSiblingIndices[index] = -1;
                system->nextSiblingIndices[index] = -1;
            }
        }

        SetLocalTransform(system, entityID, component->position, component->rotation, component->scale);
    }
}

void SetParent(EntityDB *entityDB, EntityID entityID, EntityID parentID) {
    TransformSystem *sys = GetSystem(entityDB, Transform);
    int32 index;
    int32 parentIndex;
    if (HashTableGet(&sys->entityMap, entityID, &index)) {
        TransformComponent trans = {};
        trans.parent = parentID;

        Transform worldTrans = {sys->worldPositions[index],
                                sys->worldRotations[index],
                                sys->worldScales[index],
        };
            
        if (HashTableGet(&sys->entityMap, parentID, &parentIndex)) {
            Transform worldTransParent = {sys->worldPositions[parentIndex],
                                          sys->worldRotations[parentIndex],
                                          sys->worldScales[parentIndex],
            };
        
            Transform newLocalTrans = RelativeTransform(worldTrans, worldTransParent);
            trans.position = newLocalTrans.position;
            trans.rotation = newLocalTrans.rotation;
            trans.scale = newLocalTrans.scale;
        }
        else {
            Transform worldTrans = {sys->worldPositions[index],
                                    sys->worldRotations[index],
                                    sys->worldScales[index],
            };
            trans.position = worldTrans.position;
            trans.rotation = worldTrans.rotation;
            trans.scale = worldTrans.scale;
        }

        SetTransformComponentValue(entityDB, sys, entityID, &trans);
    }
}

void Unparent(EntityDB *entityDB, EntityID entityID) {
    SetParent(entityDB, entityID, EntityID_Invalid);
}

uint32 CountChildren(EntityDB *entityDB, TransformSystem *system, EntityID parentID) {
    uint32 count = 0;
    int32 index;
    if (HashTableGet(&system->entityMap, parentID, &index)) {
        int32 childIndex = system->firstChildIndices[index];
        if (childIndex < 0) {
            return count;
        }
        
        while (childIndex >= 0) {
            EntityID childID = system->entityIDs[childIndex];
            if (childID != EntityID_Invalid) {
                count++;
            }
            childIndex = system->nextSiblingIndices[childIndex];
        }
    }

    return count;
}

void GetChildren(EntityDB *entityDB, TransformSystem *system, EntityID parentID, EntityID *entityIDs, uint32 count) {

    int32 index;
    uint32 entityIndex = 0;
    if (HashTableGet(&system->entityMap, parentID, &index)) {
        int32 childIndex = system->firstChildIndices[index];
        while (childIndex >= 0) {
            EntityID childID = system->entityIDs[childIndex];
            if (childID != EntityID_Invalid) {
                entityIDs[entityIndex] = childID;
                entityIndex++;
            }
            childIndex = system->nextSiblingIndices[childIndex];
        }
    }
}

EntityID *GetChildren(EntityDB *entityDB, TransformSystem *system, EntityID parentID, uint32 *childCount) {
    *childCount = CountChildren(entityDB, system, parentID);
    EntityID *children = PushArray(Game.frameMem, EntityID, *childCount);

    if (*childCount > 0) {
        GetChildren(entityDB, system, parentID, children, *childCount);
    }

    return children;
}

uint32 CountChildrenWithComponent(EntityDB *entityDB, TransformSystem *system, EntityID parentID, ComponentType type) {
    uint32 count = 0;
    int32 index;
    if (HashTableGet(&system->entityMap, parentID, &index)) {
        int32 childIndex = system->firstChildIndices[index];
        if (childIndex < 0) {
            return count;
        }
        
        while (childIndex >= 0) {
            ComponentValue component = {};
            EntityID childID = system->entityIDs[childIndex];
            if (GetComponentValue_(entityDB, type, childID, &component)) {
                count++;
            }
            childIndex = system->nextSiblingIndices[childIndex];
        }
    }

    return count;
}

void GetChildrenWithComponent(EntityDB *entityDB, TransformSystem *system, EntityID parentID, ComponentType type, EntityID *entityIDs, uint32 count) {

    int32 index;
    uint32 entityIndex = 0;
    if (HashTableGet(&system->entityMap, parentID, &index)) {
        int32 childIndex = system->firstChildIndices[index];
        while (childIndex >= 0) {
            EntityID childID = system->entityIDs[childIndex];
            ComponentValue component = {};
            if (GetComponentValue_(entityDB, type, childID, &component)) {
                entityIDs[entityIndex] = childID;
                entityIndex++;
            }
            childIndex = system->nextSiblingIndices[childIndex];
        }
    }
}

EntityID *GetChildrenWithComponent(EntityDB *entityDB, TransformSystem *system, EntityID parentID, ComponentType type, uint32 *childCount) {
    *childCount = CountChildrenWithComponent(entityDB, system, parentID, type);
    EntityID *children = PushArray(Game.frameMem, EntityID, *childCount);

    if (*childCount > 0) {
        GetChildrenWithComponent(entityDB, system, parentID, type, children, *childCount);
    }

    return children;
}

EntityID *GetChildrenNamed(EntityDB *entityDB, TransformSystem *system, EntityID parentID, const char *name, uint32 *childCountPtr) {
    EntityID *childrenNamed = NULL;
    
    EntityID *children = NULL;

    uint32 childCount = CountChildrenWithComponent(entityDB, system, parentID, ComponentType_Name);

    if (childCount > 0) {
        children = PushArray(Game.frameMem, EntityID, childCount);

        GetChildrenWithComponent(entityDB, system, parentID, ComponentType_Name, children, childCount);

        NameComponent *names = (NameComponent *)PushArray(Game.frameMem, NameComponent, childCount);
        uint32 *hasName = (uint32 *)PushArray(Game.frameMem, bool, childCount);
        
        uint32 childrenNamedCount = 0;
        for (int i = 0; i < childCount; i++) {
            NameComponent nameComp;
            if (HasComponentValue(entityDB, Name, children[i], &nameComp) &&
                StringEquals(name, nameComp.name)) {
                names[i] = nameComp;
                hasName[i] = true;
                childrenNamedCount++;
            }
        }

        childrenNamed = (EntityID *)PushArray(Game.frameMem, EntityID, childrenNamedCount);

        uint32 childNamedIndex = 0;
        for (int i = 0; i < childCount; i++) {
            if (hasName[i]) {
                childrenNamed[childNamedIndex++] = children[i];
            }
        }

        if (childCountPtr != NULL) {
            *childCountPtr = childrenNamedCount;
        }
    }

    
    return childrenNamed;
}

