
STRUCT()
struct Transform {
    vec3 position;
    quaternion rotation;
    vec3 scale;
};

inline Transform IdentityTransform() {
    Transform result;

    result.position = V3(0, 0, 0);
    result.rotation = IdentityQuaternion();
    result.scale = V3(1, 1, 1);

    return result;
};

inline mat4 ToMat4(Transform transform) {
    return TRS(transform.position, transform.rotation, transform.scale);
}

inline vec3 MultiplyPoint(Transform transform, vec3 point) {
    return PointToReferenceSpace(transform.position, transform.rotation, transform.scale, point);
}

inline vec3 MultiplyDirection(Transform transform, vec3 direction) {
    return DirectionToReferenceSpace(transform.rotation, direction);
}

inline vec3 InverseMultiplyPoint(Transform transform, vec3 point) {
    return PointToLocalSpace(transform.position, transform.rotation, transform.scale, point);
}

inline vec3 InverseMultiplyDirection(Transform transform, vec3 direction) {
    return DirectionToLocalSpace(transform.rotation, direction);
}

inline vec3 ScalePoint(vec3 scale, vec3 point) {
    mat4 mat = Scale4(scale);
    return MultiplyPoint(mat, point);
}

inline vec3 RelativeScale(vec3 fromScale, vec3 toScale) {
    vec3 invFromScale = V3(SafeRatio(1.0f, fromScale.x, 0.0f),
                           SafeRatio(1.0f, fromScale.y, 0.0f),
                           SafeRatio(1.0f, fromScale.z, 0.0f));
    return Hadamard(invFromScale, toScale);
}

inline Transform RelativeTransform(Transform from, Transform to) {
    Transform result;

    vec3 invFromScale = V3(SafeRatio(1.0f, from.scale.x, 0.0f),
                           SafeRatio(1.0f, from.scale.y, 0.0f),
                           SafeRatio(1.0f, from.scale.z, 0.0f));
    result.scale = Hadamard(invFromScale, to.scale);

    quaternion invFromRotation = Inverse(from.rotation);
    result.rotation = invFromRotation * to.rotation;

    result.position = Hadamard(Rotate(invFromRotation, (to.position - from.position)), invFromScale);

    return result;
}


COMPONENT(Transform) {
    vec3 position;
    quaternion rotation;
    vec3 scale;

    EntityID parent;
    EntityID firstChild;
    EntityID prevSibling;
    EntityID nextSibling;
};

SYSTEM(Transform, Priority = 1) {
    DynamicArray<vec3> positions;
    DynamicArray<quaternion> rotations;
    DynamicArray<vec3> scales;

    DynamicArray<vec3> worldPositions;
    DynamicArray<quaternion> worldRotations;
    DynamicArray<vec3> worldScales;

    DynamicArray<int32> parentIndices;
    DynamicArray<int32> firstChildIndices;
    DynamicArray<int32> prevSiblingIndices;
    DynamicArray<int32> nextSiblingIndices;
        
    DynamicArray<mat4> worldTransforms;
    HashTable<EntityID, int32> entityMap;
};

inline TransformComponent IdentityTransformComp() {
    TransformComponent trans = {};
    trans.rotation = IdentityQuaternion();
    trans.scale = V3(1, 1, 1);
    return trans;
}


mat4 GetWorldTransformMatrix(TransformSystem *system, EntityID entityID);
Transform GetWorldTransform(TransformSystem *system, EntityID entityID);
Transform GetLocalTransform(TransformSystem *system, EntityID entityID);
//@TODO: mat4 GetLocalTransform(TransformSystem *system, EntityID entityID);

vec3 GetWorldPosition(TransformSystem *system, EntityID entityID);
vec3 GetLocalPosition(TransformSystem *system, EntityID entityID);

quaternion GetLocalRotation(TransformSystem *system, EntityID entityID);
quaternion GetWorldRotation(TransformSystem *system, EntityID entityID);

EntityID GetParentID(TransformSystem *system, EntityID entityID);

vec3 GetForward(TransformSystem *system, EntityID entityID);

vec3 GetLocalScale(TransformSystem *system, EntityID entityID);
vec3 GetWorldScale(TransformSystem *system, EntityID entityID);

void SetLocalPosition(TransformSystem *system, EntityID entityID, vec3 position);
void SetWorldPosition(TransformSystem *system, EntityID entityID, vec3 position);

void SetLocalRotation(TransformSystem *system, EntityID entityID, quaternion rotation);
void SetWorldRotation(TransformSystem *system, EntityID entityID, quaternion rotation);

void SetLocalScale(TransformSystem *system, EntityID entityID, vec3 scale);
void SetWorldScale(TransformSystem *system, EntityID entityID, vec3 scale);

void SetLocalTransform(TransformSystem *system, EntityID entityID, vec3 position, quaternion rotation, vec3 scale);
void SetLocalTransform(TransformSystem *system, EntityID entityID, Transform localTransform);
void SetWorldTransform(TransformSystem *system, EntityID entityID, Transform worldTransform);
void SetWorldTransform(TransformSystem *system, EntityID entityID, vec3 position, quaternion rotation, vec3 scale);


void SetParent(EntityDB *entityDB, EntityID entityID, EntityID parentID);
void Unparent(EntityDB *entityDB, EntityID entityID);

uint32 CountChildren(EntityDB *entityDB, TransformSystem *system, EntityID parentID);
void GetChildren(EntityDB *entityDB, TransformSystem *system, EntityID parentID, EntityID *entityIDs, uint32 count);

EntityID LastChildOf(TransformSystem *system, EntityID entityID);
EntityID *GetChildren(EntityDB *entityDB, TransformSystem *system, EntityID parentID, uint32 *childCount);

uint32 CountChildrenWithComponent(EntityDB *entityDB, TransformSystem *system, EntityID parentID, ComponentType type);

void GetChildrenWithComponent(EntityDB *entityDB, TransformSystem *system, EntityID parentID, ComponentType type, EntityID *entityIDs, uint32 count);

EntityID *GetChildrenWithComponent(EntityDB *entityDB, TransformSystem *system, EntityID parentID, ComponentType type, uint32 *childCount);

EntityID *GetChildrenNamed(EntityDB *entityDB, TransformSystem *system, EntityID parentID, const char *name, uint32 *childCountPtr = NULL);

