        
#ifndef _ETC_UI_H
#define _ETC_UI_H

struct UIManager;

// Render Key Generation
enum UIRenderPriority : uint32 {
    UIRenderPriority_First,
    UIRenderPriority_Default,
    UIRenderPriority_Last,
    UIRenderPriority_Teardown,
};

enum UIBatchType : uint8 {
    UIBatchType_Mesh,
    UIBatchType_Font,

    UIBatchType_Count,
};

enum UIBatchRenderMode : uint8 {
    UIBatchRenderMode_Setup,
    UIBatchRenderMode_PriorityRender,
    UIBatchRenderMode_Render,
    UIBatchRenderMode_Teardown,
};

#define UI_MAX_SHADERS_PER_GROUP (1 << 3)
#define UI_MAX_BATCHES_PER_GROUP (1 << 7)

// Listed from most- to least-significant, with (n) bits reserved
// - priority (2)
// - group index (15)
// - group setup (1)
// - batch type (1)
// - shader (3)
// - shader setup (1)
// - batch (7)
// - batch render mode (2)
inline uint32 GenerateUIRenderKey(UIRenderPriority priority, uint32 groupIndex = 0, b32 isGroupSetup = false, uint32 shaderIndex = 0, b32 isShaderSetup = false, UIBatchType batchType = UIBatchType_Mesh, uint32 batchIndex = 0, UIBatchRenderMode batchRenderMode = UIBatchRenderMode_Setup) {
    ASSERT(priority < (1 << 2));
    ASSERT(groupIndex < (1 << 15));
    ASSERT(shaderIndex < (1 << 3));
    ASSERT(batchType < (1 << 1));
    ASSERT(batchIndex < (1 << 7));
    ASSERT(batchRenderMode < (1 << 2));
    
    uint32 bitsLeft = 32;
    uint32 key = 0;

    AddRenderKeyValue(&key, &bitsLeft, priority, 2);
    AddRenderKeyValue(&key, &bitsLeft, groupIndex, 15);
    AddRenderKeyValue(&key, &bitsLeft, !isGroupSetup, 1);
    AddRenderKeyValue(&key, &bitsLeft, batchType, 1);
    AddRenderKeyValue(&key, &bitsLeft, shaderIndex, 3);
    AddRenderKeyValue(&key, &bitsLeft, !isShaderSetup, 1);
    AddRenderKeyValue(&key, &bitsLeft, batchIndex, 7);
    AddRenderKeyValue(&key, &bitsLeft, batchRenderMode, 2);
    
    return key;
}


// UIStyle
enum UIFlowDirection : uint32 {
    UIFlowDirection_Vertical,
    UIFlowDirection_Horizontal,
};

struct UIPadding {
    real32 left;
    real32 right;
    real32 top;
    real32 bottom;
};

enum UIStyleTag {
    UIStyle_WindowBackgroundColor,

    UIStyle_Font,
    UIStyle_FontSize,
    UIStyle_FontColor,

    UIStyle_ButtonColor,
    UIStyle_ButtonActiveColor,

    UIStyle_DisplayName,
    UIStyle_DisplayNameWidgetFunc,

    // Padding affects the spacing between widgets
    UIStyle_Padding,
    UIStyle_FlowDirection,

    // Inset margin affects spacing of elements inside widgets
    UIStyle_InsetMargin,

    UIStyle_MeshShader,
    UIStyle_TexturedMeshShader,
    UIStyle_TextShader,

    UIStyle_StructPushTree,
    UIStyle_BitfieldPushTree,

    UIStyle_InputStringBlinkRate,
    UIStyle_InputStringSelectAllOnActive,
    UIStyle_InputStringNoSelectAllWhenActivated,
    UIStyle_DoubleClickThreshold,

    UIStyle_UseShortEnumNames,

    UIStyleTag_Count
};

#define FOREACH_STYLETYPE(OP)                   \
    OP(void *, Pointer)                         \
    OP(bool32, b)                               \
    OP(int32, i)                                \
    OP(real32, f)                               \
    OP(vec2, 2)                                 \
    OP(vec3, 3)                                 \
    OP(vec4, 4)                                 \
    OP(FontTableHandle, Font)                   \
    OP(UIFlowDirection, FlowDirection)          \
    OP(UIPadding, Padding)                      \
    OP(ShaderProgramHandle, ShaderProgram)      \

#define DEFINE_STYLETYPE_FIELD(type, nameSuffix) type value##nameSuffix;

struct UIStyle {
    UIStyleTag tag;
    union {
        FOREACH_STYLETYPE(DEFINE_STYLETYPE_FIELD)
    };
};

struct UIStyleGroup {
    uint32 first;
    uint32 count;
};

void PushStyles(UIStyle *styles, uint32 count, UIManager *manager = NULL);
void PushStyle(UIStyle style, UIManager *manager = NULL);
void PopStyles(UIManager *manager = NULL);


// Widget Management
typedef uint32 WidgetID;
const WidgetID WidgetID_Invalid = 0;


// @TODO: must be able to filter widgets in a window eventually,
// probably based on widget type and filter-string. This is probably
// some data that's set in the widget stack node!

enum WidgetGroupFlags {
    WidgetGroupFlag_CanActivateParent = 1 << 0,
    WidgetGroupFlag_OverridePriority = 1 << 1,
    WidgetGroupFlag_Is3D = 1 << 2,
    WidgetGroupFlag_BlocksMouse = 1 << 3,

    WidgetGroupFlag_ResizeToContentHorizontal = 1 << 4,
    WidgetGroupFlag_ResizeToContentVertical = 1 << 5,
    WidgetGroupFlag_ResizeToContent = WidgetGroupFlag_ResizeToContentHorizontal | WidgetGroupFlag_ResizeToContentVertical,

    WidgetGroupFlag_Absolute = 1 << 6,

    WidgetGroupFlag_IsTreeHeaderGroup = 1 << 7,
    WidgetGroupFlag_IsTreeContentGroup = 1 << 8,

    WidgetGroupFlag_IsScrollGroup = 1 << 9,
    WidgetGroupFlag_IsScrollContentGroup = 1 << 10,

    WidgetGroupFlag_DrawBackground = 1 << 11,
};


struct WidgetGroup {
    WidgetID id;
    uint32 groupIndex;
    uint32 groupDepth;

    // anchor is where local (0, 0) is located in the parent group
    vec2 anchor;
    EtcRect localBounds;

    // this is the "visible" portion of the group in local space
    EtcRect localClipRect;


    vec2 screenPosition;
    vec2 screenSize;

    mat4 view;
    mat4 projection;
    mat4 viewProjection;

    CameraInfo cameraInfo;

    // This is the clipRect in screen space, clipped to the parent's clipRect
    EtcRecti clipRect;


    uint32 flags;
    UIRenderPriority overridePriority;


    uint32 widgetCount;

    WidgetID prevID;
    EtcRect prevWidgetBounds;
    EtcRect totalWidgetBounds;

    b32 nextWidgetBoundsOverride;
    EtcRect nextWidgetBounds;

    vec2 localMousePosition;
    b32 isHoveringGroup;
    Ray mouseWorldRay;


    uint32 shaderCount;
    ShaderProgram *shaders[UI_MAX_SHADERS_PER_GROUP];
    UIBatchType shaderBatchTypes[UI_MAX_SHADERS_PER_GROUP];

    uint32 batchAssetCount[UIBatchType_Count];
    void *batchAssets[UIBatchType_Count][UI_MAX_BATCHES_PER_GROUP];

    // Each shader that uses a batched asset needs to go through the
    // batch setup phase again, so store which shaders each asset has
    // used in a bitfield
    uint8 batchAssetUsedShader[UIBatchType_Count][UI_MAX_BATCHES_PER_GROUP];
};


struct PersistentAllocatedData {
    PersistentAllocatedData *next;
    uint32 dataID;

    uint32 dataSize;
    uint8 data[0];
};

struct PersistentWidgetData {
    b32 accessedLastFrame;

    PersistentAllocatedData *dataHead;
};

struct UIManager {
    UIRenderPriority hoveredRenderPriority;
    uint32 hoveredGroupIndex;
    uint32 hoveredGroupDepth;
    real32 hoveredDepth;
    int32 hoveredIndex;
    WidgetID hoveredID;
    b32 hoveredIDAccessedLastFrame;

    WidgetID activeID;
    b32 activeIDAccessedLastFrame;

    void *activeWidgetData;
    uint32 activeWidgetDataSize;


    WidgetID prevSelectableWidgetID;
    WidgetID nextSelectableWidgetID;
    WidgetID shouldActivateWidgetID;
    b32 shouldActivateNextWidgetID;
    

    vec2 prevScreenMousePosition;
    vec2 screenMousePosition;
    vec2 screenMousePositionNorm;
    
    bool wrappingMouse;
    vec2i wrapScreenCoords;
    vec2 screenMouseVirtualPosition;
    vec2 screenMouseVirtualPositionNorm;
    

    DynamicArray<uint32> inputCharactersFrontBuffer;
    DynamicArray<uint32> inputCharactersBackBuffer;


    // @TODO: I think we need to know _what_ is blocking input
    b32 blockingKeyboard;
    b32 blockingMouse;
    bool showMouse;

    uint32 nextGroupIndex;
    ChunkedArray<WidgetGroup> groupStack;
    HashTable<WidgetID, PersistentWidgetData> persistentDataMap;

    DynamicArray<UIStyleGroup> styleGroups;
    DynamicArray<UIStyle> styles;
    UIStyle *cachedStyle[UIStyleTag_Count];

    CommandList commandList;

    uint32 textBufferID;
};

void AllocateUIManager(UIManager *manager);
void DeallocateUIManager(UIManager *manager);

#define GetCurrentUIManager() (Game.uiManager)
#define ResolveUIManager(manager) (manager != NULL ? manager : GetCurrentUIManager())

void UIManagerStartFrame(UIManager *manager);

inline void AddInputCharacter(UIManager *manager, uint32 codepoint) {
    PushBack(&manager->inputCharactersFrontBuffer, codepoint);
}


inline uint32 GenerateDataID_(const char *string) {
    uint32 dataID = Hash((void *)string, StringLength(string));
    return dataID;
}

#if __linux__
#define GenerateDataID(name) GenerateDataID_(StringConcat(Game.frameMem, 3, (const char *)__FUNCTION__, ":", name))

#else
#define GenerateDataID(name) GenerateDataID_(__FUNCTION__ ":" name)
#endif

#define GenerateDataIDString(name) GenerateDataID_(name)

WidgetID PushWidgetID(UIManager *manager, const char *name, b32 recordID = true);

b32 RegisterSelectableWidget(UIManager *manager, WidgetID id);
void ActivateNextWidget(UIManager *manager = NULL);

void *AllocatePersistentData(UIManager *manager, WidgetID id, uint32 dataID, uint32 size);
void DeallocatePersistentData(UIManager *manager, WidgetID id, uint32 dataID);
void DeallocateAllPersistentData(UIManager *manager, WidgetID id);
b32 GetPersistentData(UIManager *manager, WidgetID id, uint32 dataID, void **data);

void SetHoveredWidget(UIManager *manager, WidgetGroup *group, WidgetID id, real32 depth = INFINITY, int32 index = -1);
b32 IsHovered(UIManager *manager, WidgetID id);

void SetActiveWidget(UIManager *manager, WidgetID id, uint32 activeWidgetDataSize = 0);
b32 IsActive(UIManager *manager, WidgetID id);

EtcRect GetNextWidgetBounds(UIManager *manager, WidgetGroup *group);
void SetNextWidgetBounds(UIManager *manager, WidgetGroup *group, EtcRect bounds);

WidgetGroup *PushWidgetGroup(WidgetID id, vec2 anchor, EtcRect localBounds, EtcRect localClipRect, uint32 flags = 0, UIManager *manager = NULL);
WidgetGroup *PushWidgetGroup3D(WidgetID id, vec2 anchor, EtcRect localBounds, CameraInfo info, uint32 flags = 0, UIManager *manager = NULL);
WidgetGroup *PushWidgetGroup3D(WidgetID id, CameraInfo info, uint32 flags = 0, UIManager *manager = NULL);
void PopWidgetGroup(UIManager *manager = NULL);
WidgetGroup *PeekWidgetGroup(UIManager *manager = NULL);

WidgetGroup *GetParentGroup(UIManager *manager, WidgetGroup *group);
WidgetGroup *GetAncestorGroupWithFlags(UIManager *manager, WidgetGroup *group, uint32 flags);

b32 RenderMesh(UIManager *manager, WidgetGroup *group, Mesh *mesh, Texture *texture0, mat4 model, vec4 color, b32 priorityRender = false);
b32 RenderMesh(UIManager *manager, WidgetGroup *group, Mesh *mesh, mat4 model, vec4 color, b32 priorityRender = false);
b32 RenderText(UIManager *manager, WidgetGroup *group, FontTable *font, GlyphAttributes *glyphs, uint32 glyphCount);

// UIStyle Function Definitions
inline UIStyle *GetCachedStyle(UIManager *manager, UIStyleTag tag) {
    UIStyle *result = manager->cachedStyle[tag];
    return result;
}

#define DECLARE_STYLETYPE(type, nameSuffix)                             \
    UIStyle MakeStyle##nameSuffix(UIStyleTag tag, type value);          \
    type GetStyle##nameSuffix(UIStyleTag tag, UIManager *manager = NULL); \

#define DEFINE_STYLETYPE(type, nameSuffix)                              \
    UIStyle MakeStyle##nameSuffix(UIStyleTag tag, type value) {         \
        UIStyle result;                                                 \
        result.tag = tag;                                               \
        result.value##nameSuffix = value;                               \
        return result;                                                  \
    }                                                                   \
    type GetStyle##nameSuffix(UIStyleTag tag, UIManager *manager) {     \
        type result = {};                                               \
        UIStyle *style = GetCachedStyle(ResolveUIManager(manager), tag); \
        if (style != NULL) {                                            \
            result = style->value##nameSuffix;                          \
        }                                                               \
        return result;                                                  \
    }

FOREACH_STYLETYPE(DECLARE_STYLETYPE)


// Widgets
enum UIWindowFlags {
    UIWindowFlag_ShowTitle = 1 << 0,
    UIWindowFlag_ResizeToContentVertical = 1 << 1,
    UIWindowFlag_ResizeToContentHorizontal = 1 << 2,
    UIWindowFlag_ResizeToContent = UIWindowFlag_ResizeToContentVertical | UIWindowFlag_ResizeToContentHorizontal,

    UIWindowFlag_Absolute = 1 << 3,
    UIWindowFlag_Freemove = 1 << 4,
    UIWindowFlag_Freescale = 1 << 5,
    UIWindowFlag_ReuseWidgetID_ = 1 << 8,

    UIWindowFlag_Collapsable = 1 << 16,
};

void PushWindow(const char *name, vec2 position, vec2 size, bool *collapsed, uint32 flags = 0, const char *title = NULL);
void PushWindow(const char *name, vec2 *position, vec2 size, bool *collapsed, uint32 flags = 0, const char *title = NULL);

void PushWindow(const char *name, vec2 position, vec2 size, uint32 flags = 0, const char *title = NULL);
void PushWindow(const char *name, vec2 *position, vec2 size, uint32 flags = 0, const char *title = NULL);
void PushWindow(const char *name, vec2 *position, vec2 *size, uint32 flags = 0, const char *title = NULL);

void PopWindow();


enum UIScrollBarFlags {
    UIScrollBarFlag_Vertical = 1 << 0,
    UIScrollBarFlag_Horizontal = 1 << 1,
};

const uint32 DefaultScrollGroupFlags = UIScrollBarFlag_Vertical;

void PushScrollGroup(const char *name, EtcRect localBounds, uint32 flags = DefaultScrollGroupFlags);
void PushScrollGroup(const char *name, uint32 flags = DefaultScrollGroupFlags);
void PopScrollGroup();


enum UITreeGroupFlags {
    UITreeGroupFlag_StartCollapsed = 1 << 0,
};


typedef b32 InputTreeHeaderFunc(const char *name, b32 collapsed, void *data);

struct DefaultTreeHeaderData {
    b32 selected;
    b32 clicked;
};

b32 DefaultTreeHeader(const char *name, b32 collapsed, void *data);

b32 PushTreeGroup(const char *name, InputTreeHeaderFunc *headerFunc, void *headerData, uint32 flags = 0);
b32 PushTreeGroup(const char *name, uint32 flags = 0);
void PopTreeGroup();

enum UIMenuFlags {
    UIMenuFlag_UsePreviousWidget = 1 << 0,
};

struct ContextMenuData {
    WidgetID ownerID;

    WidgetID windowID;
    b32 renderedThisFrame;
    vec2 pos;
    vec2 size;
};

b32 PushContextMenu(const char *name, WidgetID ownerID, ContextMenuData *menuData, uint32 flags = 0);
b32 PushContextMenuPrevWidget(const char *name, ContextMenuData *menuData, uint32 flags = 0);

void CloseContextMenu(ContextMenuData *menuData);
void PopContextMenu(ContextMenuData *menuData);


typedef bool ListEntryWidgetFunc(uint32 index, void *userData);

void ListWidget(const char *name, real32 entryHeight, uint32 *listCount, void *userData, ListEntryWidgetFunc *widgetFunc);


typedef void DisplayNameWidgetFunc(const char *name, void *userData);
void DefaultDisplayNameWidget(const char *name, void *userData);

inline DisplayNameWidgetFunc *GetDisplayNameWidgetFunc(UIManager *manager) {
    DisplayNameWidgetFunc *func = (DisplayNameWidgetFunc *)GetStylePointer(UIStyle_DisplayNameWidgetFunc);
    if (func == NULL) {
        func = &DefaultDisplayNameWidget;
    }

    return func;
}

// @TODO: this function does nonsense layout, and it should push a widget ID
void PushText(const char *name, uint32 textLength);
void PushText(const char *text);
void PushTextf(const char *fmt, ...);

void PushSchemaValue(const char *name, EntityDB *entityDB, Schema *schema, void *value);

enum UIButtonFlags {
    UIButtonFlag_TrueOnRelease = 1 << 1,
};

b32 Button(const char *name, uint32 flags = 0, const char *displayText = NULL);
b32 ButtonConfirm(const char *text);

enum UIToggleFlags {
    UIToggleFlag_UseAllHorizontalSpace = 1 << 1
};

b32 Toggle(const char *name, uint32 nameLength, bool *value, uint32 flags = 0);
b32 Toggle(const char *name, b32 *value, uint32 flags = 0);
b32 Toggle(const char *name, bool *value, uint32 flags = 0);
b32 RadioToggle(const char *name, uint32 valueCount, uint32 *value, const char **valueNames, uint32 strOffset = 0, uint32 flags = 0);
b32 RadioToggleEnum(const char *name, const char *enumName, uint32 *value, uint32 flags = 0);


enum UISelectableFlag {
    UISelectableFlag_TrueOnDoubleClick = 1 << 0,
};

b32 Selectable(const char *name, b32 *selected = NULL, uint32 flags = 0);



enum InputStringFlags {
    InputStringFlag_ReturnTrueOnEnter = 1 << 0,

    InputStringFlag_ReuseWidgetID_ = 1 << 8,
};


enum InputStringState {
    InputStringState_None,

    InputStringState_InsertedChar,

    InputStringState_MoveLeft,
    InputStringState_MoveRight,
    InputStringState_MoveBeginningOfLine,
    InputStringState_MoveEndOfLine,

    InputStringState_MoveUp,
    InputStringState_MoveDown,

    InputStringState_Backspace,
    InputStringState_ForwardDelete,

    InputStringState_Completion,

    InputStringState_MouseClick,
    InputStringState_MouseDrag,
};

struct InputStringPersistentData {
    b32 bufferIsValid;
    b32 wasActive;

    // Store the contents of the buffer when the widget becomes active
    // so it can be restored if necessary
    uint32 initialBufferLength;
    char *initialBuffer;
};

struct InputStringActiveData {
    int32 cursorIndex;

    InputStringState state;
    real32 nextActionTime;

    b32 hasSelection;
    int32 selectionStart;
    int32 selectionEnd;

    real32 blinkStartTime;
    real32 prevMouseReleaseTime;
    int32 initialClickedIndex;
};

typedef b32 ValidateInputTextFunc(const char *string, InputStringActiveData *acitveData, void *data);

b32 InputString(const char *name, char *buffer, uint32 bufferLength, uint32 flags = 0, ValidateInputTextFunc *validationFunc = NULL, void *validateFuncData = NULL);

b32 InputInt(const char *name, int32 *value, uint32 flags = 0);
b32 InputUInt(const char *name, uint32 *value, uint32 flags = 0);
b32 InputFloat(const char *name, real32 *value, uint32 flags = 0);
b32 InputIntOrNormFloat(const char *name, real32 *value, int32 intMin, int32 intMax, uint32 flags = 0);


b32 Sliderf(const char *name, real32 *value, real32 minValue, real32 maxValue, real32 snapValue = 0.0f, uint32 flags = 0);


b32 InputVec2(const char *name, vec2 *value, uint32 flags = 0);
b32 InputVec3(const char *name, vec3 *value, uint32 flags = 0);
b32 InputVec4(const char *name, vec4 *value, uint32 flags = 0);
b32 InputQuaternion(const char *name, quaternion *value, uint32 flags = 0);

b32 InputVec2i(const char *name, vec2i *value, uint32 flags = 0);

enum UIColorPickerFlags {
    UIColorPickerFlag_InputHSV = 1 << 0,
    UIColorPickerFlag_RenderRYB = 1 << 1,
};

b32 InputColor(const char *name, vec4 *value, uint32 flags = 0);

b32 ColorPicker(const char *name, vec3 *value, EtcRect bounds, uint32 flags = 0);
b32 ColorPicker(const char *name, vec4 *value, EtcRect bounds, uint32 flags = 0);

struct ColorPickerWindowData {
    WidgetID ownerID;
    b32 renderedThisFrame;
    vec4 originalColor;
    vec3 prevHSV;
};

b32 ColorPickerWindow(const char *name, vec4 *value, WidgetID ownerID, ColorPickerWindowData *windowData, uint32 flags);

b32 InputTransform(const char *name, Transform *value, uint32 flags = 0);


b32 InputFuncHandle(const char *name, PersistentFuncType funcType, FuncHandle *funcHandle, uint32 flags = 0);
b32 InputBitfield(const char *name, Schema *schema, uint32 *value, uint32 flags = 0);

b32 Dropdown(const char *name, const char **strings, uint32 count, uint32 *index, uint32 flags = 0);
b32 DropdownEnum(const char *name, const char *enumName, uint32 *index, uint32 flags = 0);

b32 InputAssetHandle(const char *name, AssetType type, AssetHandle *handle, uint32 flags = 0);

#define DEFINE_INPUTASSETHANDLE(type)                                   \
    inline b32 Input##type##Handle(const char *name, type##Handle *handle, uint32 flags) { \
        AssetHandle handle_ = (AssetHandle)(*handle);                   \
        if (InputAssetHandle(name, AssetType_##type, &handle_, flags)) { \
            *handle = *(type##Handle *)&handle_;                        \
            return true;                                                \
        }                                                               \
        return false;                                                   \
    }                                                                   \

FOREACH_ASSET(DEFINE_INPUTASSETHANDLE)

b32 InputEntityName(const char *name, EntityDB *entityDB, EntityID *entityID, uint32 ignoredIDCount, EntityID *ignoredIDs, uint32 flags = 0);
b32 InputEntityName(const char *name, EntityDB *entityDB, EntityID *entityID, uint32 flags = 0);

b32 InputDynamicArray(const char *name, EntityDB *entityDB, Schema *schema, uint32 *arrayCapacity, uint32 *arrayCount, void **arrayData, uint32 flags = 0);
b32 InputDynamicArray(const char *name, EntityDB *entityDB, Schema *schema, DynamicArray_Untyped *array, uint32 flags = 0);


// @TODO: entityDB should maybe be an optional parameter rather than a required one
b32 InputSchemaDataValue_(const char *name, EntityDB *entityDB, SchemaTypeDescription type, void *value, uint32 flags = 0);

b32 InputSchemaStructFieldValue(const char *name, const char *fieldName, EntityDB *entityDB, Schema *schema, void *value, u32 flags = 0);
b32 InputSchemaValue(const char *name, EntityDB *entityDB, Schema *schema, void *value, uint32 flags = 0);

b32 InputSchemaValue(const char *name, EntityDB *entityDB, const char *schemaName, void *data, uint32 flags = 0);

b32 InputSchemaData(const char *name, EntityDB *entityDB, SchemaData *schemaData, uint32 flags = 0);
b32 InputSchemaData(const char *name, EntityDB *entityDB, const char *schemaName, SchemaData *schemaData, uint32 flags = 0);

b32 InputComponentValue(const char *name, EntityDB *entityDB, ComponentType type, ComponentValue *value, uint32 flags = 0);

#endif
